/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/


#ifndef ASN1_CERT_H
#define ASN1_CERT_H

#include <openssl/x509.h>
#include <openssl/asn1.h>
#include <openssl/asn1t.h>
#include <openssl/x509v3.h>
#include <openssl/evp.h>
#include <openssl/pem.h>
#include <openssl/ossl_typ.h>

#include "my_asn1.h"
#include "ASN1_err.h"


/*!
  This enumeration defines the possible states of a certificate
*/
typedef enum en_CERT_STATE
{
	CERT_STATE_VALID = 1,		//!< Certificate is valid
	CERT_STATE_REVOKED = 2,		//!< Certificate is revoked
	CERT_STATE_UNKNOWN = 3		//!< Unknown state
}CERT_STATE;



/*!
  This structure represents an certification request
*/
typedef struct st_NEWPKI_CERT_REQUEST
{
	ASN1_INTEGER * id;					//!< An internal ID from the requesting entity, this id is sent back with the response
	X509_REQ * request;					//!< The certificate request
	ASN1_INTEGER * validity;			//!< The number of days the certificate should be valid for
	ASN1_UTF8STRING * ldap_uid;			//!< The LDAP UID of the user this request id for, the LDAP UID is latter used for publication
}NEWPKI_CERT_REQUEST;
DECLARE_ASN1_FUNCTIONS(NEWPKI_CERT_REQUEST)


/*!
  This structure represents an revocation request
*/
typedef struct st_NEWPKI_REV_REQUEST
{
	ASN1_INTEGER * id;				//!< An internal ID from the requesting entity, this id is sent back with the response
	ASN1_INTEGER * serial;			//!< The serial of the certificate to revoke
	ASN1_INTEGER * reason;			//!< The reason for this revocation
	ASN1_UTF8STRING * ldap_uid;
}NEWPKI_REV_REQUEST;
DECLARE_ASN1_FUNCTIONS(NEWPKI_REV_REQUEST)


/*!
  This structure represents the body of a revocation publication request
*/
typedef struct st_NEWPKI_PUB_REQUEST_BODY_REV
{
	X509 * cert;					//!< The certificate that has been revoked
	ASN1_UTCTIME * revdate;			//!< The revocation date
}NEWPKI_PUB_REQUEST_BODY_REV;
DECLARE_ASN1_FUNCTIONS(NEWPKI_PUB_REQUEST_BODY_REV)

#define NEWPKI_PUB_REQUEST_CERT 0	//!< The publication request is for a certification
#define NEWPKI_PUB_REQUEST_REV 1	//!< The publication request is for a revocation
#define NEWPKI_PUB_REQUEST_CRL 2	//!< The publication request is for a CRL

/*!
  This structure represents the body of a publication request
*/
typedef struct st_NEWPKI_PUB_REQUEST_BODY
{
	int type;
	union
	{
		X509 * cert;							//!< A certificate when publishing a new certificate, type = NEWPKI_PUB_REQUEST_CERT
		NEWPKI_PUB_REQUEST_BODY_REV * rev;		//!< Revocation info when publishing a revocation, type = NEWPKI_PUB_REQUEST_REV
		X509_CRL * crl;							//!< A CRL when publishing a CRL, type = NEWPKI_PUB_REQUEST_CRL
	}d;
}NEWPKI_PUB_REQUEST_BODY;
DECLARE_ASN1_FUNCTIONS(NEWPKI_PUB_REQUEST_BODY)

/*!
  This structure represents a publication request
*/
typedef struct st_NEWPKI_PUB_REQUEST
{
	ASN1_UTF8STRING * ldap_uid;			//!< The UID for a cert or a rev, Empty for a CRL 
	STACK_OF(X509) * ParentCerts;		//!< The list of parents certs for a cert or a rev
	ASN1_UTF8STRING * object;			//!< A string name for the published object
	NEWPKI_PUB_REQUEST_BODY * body;		//!< The body
}NEWPKI_PUB_REQUEST;
DECLARE_ASN1_FUNCTIONS(NEWPKI_PUB_REQUEST)



#define NEWPKI_REQUEST_TYPE_CERT	0	//!< The request is for a certification
#define NEWPKI_REQUEST_TYPE_REV		1	//!< The request is for a revocation
#define NEWPKI_REQUEST_TYPE_PUB		2	//!< The request is for a publication

/*!
  This structure represents an inter-entity request
*/
typedef struct st_NEWPKI_REQUEST
{
	int type;
	union
	{
		NEWPKI_CERT_REQUEST * cert_request;	//!< A certification request
		NEWPKI_REV_REQUEST * rev_request;	//!< A revocation request
		NEWPKI_PUB_REQUEST * pub_request;	//!< A publication request
	}d;
}NEWPKI_REQUEST;
DECLARE_ASN1_FUNCTIONS(NEWPKI_REQUEST)
DECLARE_ASN1_ENCRYPT_SIGN(NEWPKI_REQUEST)



/*!
  This structure represents an inter-entity request signed and crypted. Requests are signed by the sender and crypted for the recipient.
*/
typedef struct st_CRYPTED_NEWPKI_REQUEST
{
	X509_PUBKEY * sender;				//!< The public key of the sender
	X509_PUBKEY * recipient;			//!< The public key of the recipient
	ASN1_ENCRYPT_SIGN * c_request;		//!< The Crypted representation of a NEWPKI_REQUEST
}CRYPTED_NEWPKI_REQUEST;
DECLARE_ASN1_FUNCTIONS(CRYPTED_NEWPKI_REQUEST)
DECLARE_ASN1_CPY_FUNCTIONS(CRYPTED_NEWPKI_REQUEST)
DECLARE_PEM_CONVERT(CRYPTED_NEWPKI_REQUEST)

/*!
	This structure is used between the repositories and the other
	entities.
	When a repository received a new CRYPTED_NEWPKI_REQUESTS
	it generates a STORED_NEWPKI_REQUESTS, for the request,
	assigns it a unique transactionID (PKI Wide), the repository
	returns the transactionID to the RA, it allows it to check 
	the status for its request.
	Once the STORED_NEWPKI_REQUESTS is all set and the RA informed
	the repository sends the STORED_NEWPKI_REQUESTS to 
	the others repositories.

	RA1  --- CRYPTED_NEWPKI_REQUESTS ---->
	                                         REP1
	RA1  <---------- transactionID ---------
	                                         REP1
	REP2 <--- STORED_NEWPKI_REQUEST -----
	                                         REP1
	CA1  <--- STORED_NEWPKI_REQUESTS ----
*/
typedef struct st_STORED_NEWPKI_REQUEST
{
	ASN1_OCTET_STRING * transactionID;	//!< The transaction ID of the request
	CRYPTED_NEWPKI_REQUEST * c_info;	//!< The signed and crypted request
}STORED_NEWPKI_REQUEST;
DECLARE_ASN1_FUNCTIONS(STORED_NEWPKI_REQUEST)
DECLARE_ASN1_CPY_FUNCTIONS(STORED_NEWPKI_REQUEST)
DECLARE_PEM_CONVERT(STORED_NEWPKI_REQUEST)

/*!
	This structure  represents a list of inter-entity requests.
*/
typedef struct st_STORED_NEWPKI_REQUESTS
{
	STACK_OF(STORED_NEWPKI_REQUEST) * requests;	//!< The list of requests
}STORED_NEWPKI_REQUESTS;
DECLARE_ASN1_FUNCTIONS(STORED_NEWPKI_REQUESTS)


#define CERT_RESPONSE_OK			1	//!< The requested certificate has been successfuly generated
#define CERT_RESPONSE_ERROR			2	//!< There was an error generation the requested certificate

/*!
  This structure represents a certificate response,
  it might either have the certificate or an error
*/
typedef struct st_NEWPKI_CERT_RESPONSE
{
	ASN1_INTEGER * id;				//!< The internal ID from the requesting entity
	ASN1_INTEGER * status;			//!< The status, CERT_RESPONSE_OK or CERT_RESPONSE_ERROR
	X509_CRL * last_crl;			//!< The last CRL
	X509 * certificate;				//!< The optional generated certificate
	STACK_OF(X509) * ParentCerts;	//!< The parents certs
	STACK_OF(ERROR_ENTRY) * errors;	//!< The optional errors
}NEWPKI_CERT_RESPONSE;
DECLARE_ASN1_FUNCTIONS(NEWPKI_CERT_RESPONSE)



#define REV_RESPONSE_OK				1	//!< The requested revocation has been successfuly processed
#define REV_RESPONSE_ERROR			2	//!< There was an error processing the requested revocation

/*!
  This structure represents a revocation response
*/
typedef struct st_NEWPKI_REV_RESPONSE
{
	ASN1_INTEGER * id;				//!< The internal ID from the requesting entity
	ASN1_INTEGER * status;			//!< The status, REV_RESPONSE_OK or REV_RESPONSE_ERROR
	ASN1_INTEGER * cert_status;		//!< The status of the certificate
	X509_CRL * last_crl;			//!< The last CRL
	STACK_OF(ERROR_ENTRY) * errors;	//!< The optional errors
}NEWPKI_REV_RESPONSE;
DECLARE_ASN1_FUNCTIONS(NEWPKI_REV_RESPONSE)


#define PUB_RESPONSE_PUBLISHED		1	//!< The requested publication has been successfuly processed
#define PUB_RESPONSE_ERROR			2	//!< There was an error processing the requested publication

typedef struct st_NEWPKI_PUB_RESPONSE
{
	ASN1_INTEGER * status;			//!< The status, PUB_RESPONSE_PUBLISHED or PUB_RESPONSE_ERROR
	ASN1_INTEGER * type;			//!< The publication type, NEWPKI_PUB_REQUEST_CERT or NEWPKI_PUB_REQUEST_REV or NEWPKI_PUB_REQUEST_CRL
	ASN1_UTF8STRING * object;		//!< A string name for the published object
	STACK_OF(ERROR_ENTRY) * errors;	//!< The optional errors
}NEWPKI_PUB_RESPONSE;
DECLARE_ASN1_FUNCTIONS(NEWPKI_PUB_RESPONSE)



#define NEWPKI_RESPONSE_TYPE_CERT		0		//!< The response is for a certification
#define NEWPKI_RESPONSE_TYPE_REV		1		//!< The response is for a revocation
#define NEWPKI_RESPONSE_TYPE_ERR		2		//!< There was an error processing the request
#define NEWPKI_RESPONSE_TYPE_PUB		3		//!< The response is for a publication

/*!
  This structure represents an inter-entity response.
*/
typedef struct st_NEWPKI_RESPONSE
{
	int type;
	union
	{
		NEWPKI_CERT_RESPONSE * cert_response;			//!< A certification response
		NEWPKI_REV_RESPONSE * rev_response;				//!< A revocation response
		NEWPKI_PUB_RESPONSE * pub_response;				//!< A publication response
		STACK_OF(ERROR_ENTRY) * errors;					//!< The was an error before even processing the requests (ACCESS DENIED, etc)
	}d;
}NEWPKI_RESPONSE;
DECLARE_ASN1_FUNCTIONS(NEWPKI_RESPONSE)
DECLARE_ASN1_ENCRYPT_SIGN(NEWPKI_RESPONSE)

/*!
  This structure represents an inter-entity response signed and crypted. Responses are signed by the sender and crypted for the recipient.
*/
typedef struct st_CRYPTED_NEWPKI_RESPONSE
{
	X509_PUBKEY * sender;				//!< The public key of the sender
	X509_PUBKEY * recipient;			//!< The public key of the recipient
	ASN1_ENCRYPT_SIGN * c_response;		//!< The Crypted representation of NEWPKI_RESPONSE
}CRYPTED_NEWPKI_RESPONSE;
DECLARE_ASN1_FUNCTIONS(CRYPTED_NEWPKI_RESPONSE)



/*!
	This structure is used between the repositories and the other
	entities.
*/
typedef struct st_STORED_NEWPKI_RESPONSE
{
	ASN1_OCTET_STRING * transactionID;		//!< The transaction ID
	CRYPTED_NEWPKI_RESPONSE * c_info;		//!< The signed and crypted response
}STORED_NEWPKI_RESPONSE;
DECLARE_ASN1_FUNCTIONS(STORED_NEWPKI_RESPONSE)
DECLARE_ASN1_CPY_FUNCTIONS(STORED_NEWPKI_RESPONSE)
DECLARE_PEM_CONVERT(STORED_NEWPKI_RESPONSE)


/*!
	This structure represents a list of inter-entity responses.
*/
typedef struct st_STORED_NEWPKI_RESPONSES
{
	STACK_OF(STORED_NEWPKI_RESPONSE) * responses;	//!< The list of responses
}STORED_NEWPKI_RESPONSES;
DECLARE_ASN1_FUNCTIONS(STORED_NEWPKI_RESPONSES)





#define NEWPKI_OBJECTS_TYPE_REQ		0		//!< The newpki object is a request
#define NEWPKI_OBJECTS_TYPE_RESP	1		//!< The newpki object is a response


/*!
	This structure represents the body of a newpi object.
*/
typedef struct st_WAITING_NEWPKI_OBJECT_BODY
{
	int type;
	union
	{
		STORED_NEWPKI_REQUEST * request;		//!< The object is a request
		STORED_NEWPKI_RESPONSE * response;		//!< The object is a response
	}d;
}WAITING_NEWPKI_OBJECT_BODY;
DECLARE_ASN1_FUNCTIONS(WAITING_NEWPKI_OBJECT_BODY)


/*!
	This structure represents a list of transaction IDs
*/
typedef struct st_TRANSACTION_IDS
{
	STACK_OF(ASN1_OCTET_STRING) * transactionIds;	//!< The list
}TRANSACTION_IDS;
DECLARE_ASN1_FUNCTIONS(TRANSACTION_IDS)
DECLARE_ASN1_CPY_FUNCTIONS(TRANSACTION_IDS)


/*!
	This structure is used by repositories for synchronization
	purposes, it lists the known requests and responses of a
	repository.
*/
typedef struct st_OBJECTS_LIST_REQUEST
{
	TRANSACTION_IDS * requests;			//!< The list of known requets
	TRANSACTION_IDS * responses;		//!< The list of known responses
}OBJECTS_LIST_REQUEST;
DECLARE_ASN1_FUNCTIONS(OBJECTS_LIST_REQUEST)
DECLARE_ASN1_CPY_FUNCTIONS(OBJECTS_LIST_REQUEST)


/*!
	This structure is used by repositories for synchronization
	purposes, it lists the known requests and responses of a
	repository, as well as a list of objects not known by
	the requesting repository.
*/
typedef struct st_OBJECTS_LIST_RESPONSE
{
	TRANSACTION_IDS * requests;						//!< The list of known requets
	TRANSACTION_IDS * responses;					//!< The list of known responses
	STACK_OF(WAITING_NEWPKI_OBJECT) * objects;		//!< The list of objects the requesting repository doesn't know
}OBJECTS_LIST_RESPONSE;
DECLARE_ASN1_FUNCTIONS(OBJECTS_LIST_RESPONSE)
DECLARE_ASN1_CPY_FUNCTIONS(OBJECTS_LIST_RESPONSE)


/*!
	This structure is used by repositories to propagate a newpki object.
*/
typedef struct st_WAITING_NEWPKI_OBJECT
{
	WAITING_NEWPKI_OBJECT_BODY * object;		//!< The body of the object
	STACK_OF(X509_PUBKEY) * rep_path;			//!< The list of all the repositories that have already processed this object
}WAITING_NEWPKI_OBJECT;
DECLARE_ASN1_FUNCTIONS(WAITING_NEWPKI_OBJECT)
DECLARE_ASN1_CPY_FUNCTIONS(WAITING_NEWPKI_OBJECT)

/*!
	This structure represents a profile.
*/
typedef struct st_NEWPKI_PROFILE
{
	ASN1_INTEGER * EE_Validation;		//!< Should the profile be validated by an EE entity
	X509_NAME * dn;						//!< The DN of the profile
	ASN1_INTEGER * owner_group_serial;	//!< The serial of the owning group, or 0 if the owner is a user
	ASN1_UTF8STRING * ldap_uid;			//!< The LDAP UID
}NEWPKI_PROFILE;
DECLARE_ASN1_FUNCTIONS(NEWPKI_PROFILE)
DECLARE_PEM_CONVERT(NEWPKI_PROFILE)

#define NEWPKI_PROFILE_CERT_STATE_WAITING 1			//!< The certification request is waiting for the return from the CA
#define NEWPKI_PROFILE_CERT_STATE_ACTIVE 2			//!< The certificate is valid
#define NEWPKI_PROFILE_CERT_STATE_REVOKED 3			//!< The certificate is revoked
#define NEWPKI_PROFILE_CERT_STATE_ERROR 4			//!< There was an error during last request for this certificate
#define NEWPKI_PROFILE_CERT_STATE_WAITING_REV 5		//!< The revocation request is waiting for the return from the CA


#define NEWPKI_PROFILE_CERT_TYPE_PKCS10 1			//!< The certificate request has been imported from a PKCS#10
#define NEWPKI_PROFILE_CERT_TYPE_PKCS12 2			//!< The certificate request has been generated by the server
#define NEWPKI_PROFILE_CERT_TYPE_SC		3			//!< The certificate request has been generated by the admin

/*!
	This structure represents a certificate.
*/
typedef struct st_NEWPKI_PROFILE_DATAS_CERT
{
	ASN1_INTEGER * Id;					//!< The internal ID of the cerificate
	ASN1_INTEGER * state;				//!< The state of the certificate
	ASN1_INTEGER * type;				//!< The type of certificate
	ASN1_UTF8STRING * ca_name;			//!< The name the certificate has been requested from
	ASN1_BIT_STRING * flags;			//!< The flags, see REQUEST_CERT_FLAG_SEND_ADMIN_NOTIFICATION, etc.
	ASN1_UTF8STRING * admin_mail;		//!< The email address of the admin that requetsed the certificate
	X509 * cert;						//!< The optional certificate
	ASN1_UTF8STRING * error;			//!< The optional errors
	PKCS12 *p12;						//!< The optional PKCS#12
}NEWPKI_PROFILE_DATAS_CERT;
DECLARE_ASN1_FUNCTIONS(NEWPKI_PROFILE_DATAS_CERT)

#define NEWPKI_PROFILE_STATE_WAITING_FOR_VALIDATION	1	//!< The profile hasn't been validated by the EE entity yet
#define NEWPKI_PROFILE_STATE_VALIDATED				2	//!< The profile has been validated

/*!
	This structure represents a complete profile.
*/
typedef struct st_NEWPKI_PROFILE_DATAS
{
	ASN1_INTEGER * State;							//!< The state of the profile
	ASN1_INTEGER * Id;								//!< The internal ID of the profile
	NEWPKI_PROFILE * Profile;						//!< The profile
	STACK_OF(NEWPKI_PROFILE_DATAS_CERT) * Certs;	//!< All the associated certificates
}NEWPKI_PROFILE_DATAS;
DECLARE_ASN1_FUNCTIONS(NEWPKI_PROFILE_DATAS)


#endif
