/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/


#ifndef ASN1_CONF_H
#define ASN1_CONF_H

#include <assert.h>
#include <openssl/x509.h>
#include <openssl/asn1.h>
#include <openssl/asn1t.h>
#include <openssl/x509v3.h>
#include <openssl/rand.h>
#include <openssl/evp.h>
#include <openssl/err.h>

#include "my_asn1.h"

#define PKI_ACL_RESOURCE				"PkiAcls"			//!< This is the resource name for the PKI entity's ACLs conf
#define PKI_EMAIL_RESOURCE				"PkiEmail"			//!< This is the resource name for the PKI entity's email conf
#define PKI_AUDITS_RESOURCE				"PkiAudits"			//!< This is the resource name for the PKI entity's audits conf
#define PKI_REPOSITORIES_RESOURCE		"PkiRepositories"	//!< This is the resource name for the PKI entity's repositories conf
#define PKI_ENTITIES_LINKS_RESOURCE		"PkiEntitiesLinks"	//!< This is the resource name for the PKI entity's links
#define PKI_ENTITIES_GROUPS_RESOURCE	"PkiGroupsLinks"	//!< This is the resource name for the PKI entity's groups

#define CRYPTO_LOCK_REPS_INFO CRYPTO_NUM_LOCKS+1	//!< This is the lock ID for the repositories structure
#define CRYPTO_LOCK_INTERNAL_PKI_CRL CRYPTO_NUM_LOCKS+2		//!< This is the lock ID for the internal CRL structure
#define CRYPTO_LOCK_INTERNAL_PKI_CA CRYPTO_NUM_LOCKS+3		//!< This is the lock ID for the internal CA structure
#define CRYPTO_LOCK_ENTITY_CONF_CRYPTED CRYPTO_NUM_LOCKS+4	//!< This is the lock ID for the PKI conf structure

#define NEWPKI_NUM_LOCKS CRYPTO_NUM_LOCKS+4					//!< This is the number of locks

#define ENTITY_TYPE_RA			0	//!< The entity type RA
#define ENTITY_TYPE_CA			1	//!< The entity type CA				
#define ENTITY_TYPE_REPOSITORY	2	//!< The entity type Repository
#define ENTITY_TYPE_PUBLICATION	3	//!< The entity type Publication
#define ENTITY_TYPE_KEY_STORE	4	//!< The entity type Key Store
#define ENTITY_TYPE_PKI			5	//!< The entity type PKI

/*!
	This structure defines a DN specifications
*/
typedef struct st_DN_SPECS
{
	ASN1_UTF8STRING * Name;		//!< The name of the DN field
	ASN1_UTF8STRING * Default;	//!< The default value of the DN field
	ASN1_UTF8STRING * Comment;	//!< The comment of the DN field
	ASN1_INTEGER * Min;			//!< The min length of the DN field
	ASN1_INTEGER * Max;			//!< The max length of the DN field
}DN_SPECS;
DECLARE_ASN1_FUNCTIONS(DN_SPECS)
DECLARE_ASN1_STACK_CPY_FUNCTIONS(DN_SPECS)
DECLARE_ASN1_CPY_FUNCTIONS(DN_SPECS)

/*!
	This structure defines an extension
*/
typedef struct st_EXTENSION_VALUE
{
	ASN1_UTF8STRING * name;		//!< The name
	ASN1_UTF8STRING * value;	//!< The value
}EXTENSION_VALUE;
DECLARE_ASN1_FUNCTIONS(EXTENSION_VALUE)

/*!
	This enumeration defines the type of DN policy
*/
typedef enum DN_TYPE
{
	DN_TYPE_OPTIONAL = 0,		//!< The DN field is optional
	DN_TYPE_SUPPLIED = 1,		//!< The DN must be supplied
	DN_TYPE_MATCH = 2,			//!< The DN must a value declared in the list
	DN_TYPE_UNKNOWN = 3			//!< Unknown policy type
};

#define STR_OPTIONAL	"optional"	//!< The string representation of DN_TYPE_OPTIONAL
#define STR_SUPPLIED	"supplied"	//!< The string representation of DN_TYPE_SUPPLIED
#define STR_MATCH		"match:"	//!< The string representation of DN_TYPE_MATCH
#define NUM_DN_TYPE 3				//!< The number of policy representation strings

/*!
	This structure defines an policy
*/
typedef struct st_POLICY_VALUE
{
	ASN1_UTF8STRING * name;		//!< The name
	ASN1_UTF8STRING * value;	//!< The value
}POLICY_VALUE;
DECLARE_ASN1_FUNCTIONS(POLICY_VALUE)
DECLARE_ASN1_CPY_FUNCTIONS(POLICY_VALUE)



#define FORMAT_MIGRATION_FUNC_NAME(stname, version) \
stname##_To_##version

typedef void * (CONF_MIGRATION_FUNC)(void *);

typedef struct st_MIGRATION_ITEM
{
	CONF_MIGRATION_FUNC * cb_migr;
	int version;
}MIGRATION_ITEM;

#define DECLARE_CONF_MIGRATION_FUNCTIONS(stname, lastversion) \
int stname##_Upgrade(stname * a); \
const MIGRATION_ITEM * stname##_migr();

#define BEGIN_MIGRATION_FUNCTION(stname, lastversion) \
int stname##_Upgrade(stname * a) \
{ \
	int i; \
	int tmpType; \
	const MIGRATION_ITEM * migr = stname##_migr(); \
	if(!migr) \
		return 0; \
	for(i=0; migr[i].cb_migr; i++) \
	{ \
		tmpType = a->type + 1; \
		if(migr[i].version == tmpType) \
		{ \
			migr[i].cb_migr((void*)a); \
			i=-1; \
		} \
	} \
	assert( (lastversion == a->type) ); \
	return 1; \
} \
const MIGRATION_ITEM * stname##_migr() \
{ \
	static MIGRATION_ITEM stname##_migr_it[] = {

#define MIGRATION_FUNCTION_ITEM(stname, version) \
		{ & FORMAT_MIGRATION_FUNC_NAME(stname, version), version },

#define END_MIGRATION_FUNCTION(stname) \
		{ NULL, 0} \
	}; \
	return stname##_migr_it; \
}

#define CONF_VERSION_0 0
#define CONF_VERSION_1 1

#define REP_CONF_VERSION CONF_VERSION_0
#define REP_CONF_STRUCT_new ENTITY_CONF_BODY_REP_0_new
#define REP_CONF_PTR conf0

/*!
	This structure represents the body version 0 of a repository conf
*/
typedef struct st_ENTITY_CONF_BODY_REP_0
{
	ASN1_INTEGER * Enabled;
}ENTITY_CONF_BODY_REP_0;
DECLARE_ASN1_FUNCTIONS(ENTITY_CONF_BODY_REP_0)


/*!
	This structure represents a repository conf
*/
typedef struct st_ENTITY_CONF_BODY_REP
{
	int type;
	union{
		ENTITY_CONF_BODY_REP_0 * conf0;	//!< Conf version 0
		void * ptr;
	}d;
}ENTITY_CONF_BODY_REP;
DECLARE_ASN1_FUNCTIONS(ENTITY_CONF_BODY_REP)
DECLARE_CONF_MIGRATION_FUNCTIONS(ENTITY_CONF_BODY_REP, CONF_VERSION_0)


#define PKI_CONF_VERSION CONF_VERSION_0
#define PKI_CONF_STRUCT_new ENTITY_CONF_BODY_PKI_0_new
#define PKI_CONF_PTR conf0

/*!
	This structure represents the body version 0 of a PKI conf
*/
typedef struct st_ENTITY_CONF_BODY_PKI_0
{
	ASN1_INTEGER * offline;			//!< The PKI is offline
}ENTITY_CONF_BODY_PKI_0;
DECLARE_ASN1_FUNCTIONS(ENTITY_CONF_BODY_PKI_0)

/*!
	This structure represents a PKI conf
*/
typedef struct st_ENTITY_CONF_BODY_PKI
{
	int type;
	union{
		ENTITY_CONF_BODY_PKI_0 * conf0; //!< Conf version 0
		void * ptr;
	}d;
}ENTITY_CONF_BODY_PKI;
DECLARE_ASN1_FUNCTIONS(ENTITY_CONF_BODY_PKI)
DECLARE_CONF_MIGRATION_FUNCTIONS(ENTITY_CONF_BODY_PKI, CONF_VERSION_0)


/*!
	This structure represents a plugin option
*/
typedef struct st_PLUG_OPTION
{
	ASN1_UTF8STRING * name;		//!< The name
	ASN1_UTF8STRING * value;	//!< The value
}PLUG_OPTION;
DECLARE_ASN1_FUNCTIONS(PLUG_OPTION)

/*!
	This structure represents a plugin
*/
typedef struct st_EXTERNAL_ENTITY_PLUG
{
	ASN1_UTF8STRING * LibraryPath;		//!< The path to the plugin
	STACK_OF(PLUG_OPTION) * Options;	//!< The options
}EXTERNAL_ENTITY_PLUG;
DECLARE_ASN1_FUNCTIONS(EXTERNAL_ENTITY_PLUG)



#define CA_CONF_VERSION CONF_VERSION_0
#define CA_CONF_STRUCT_new ENTITY_CONF_BODY_CA_0_new
#define CA_CONF_PTR conf0

/*!
	This structure represents the body version 0 of a CA conf
*/
typedef struct st_ENTITY_CONF_BODY_CA_0
{
	STACK_OF(EXTENSION_VALUE) * Exts;		//!< The certificate extensions
	STACK_OF(EXTENSION_VALUE) * CrlExts;	//!< The CRL extensions
	ASN1_INTEGER * CrlValidityDays;			//!< The validity in days for the generated CRLs
	ASN1_INTEGER * CrlValidityHours;		//!< The validity in hours for the generated CRLs
	EXTERNAL_ENTITY_PLUG * PlugCa;			//!< A validation plugin
}ENTITY_CONF_BODY_CA_0;
DECLARE_ASN1_FUNCTIONS(ENTITY_CONF_BODY_CA_0)

/*!
	This structure represents a CA conf
*/
typedef struct st_ENTITY_CONF_BODY_CA
{
	int type;
	union{
		ENTITY_CONF_BODY_CA_0 * conf0;	//!< Conf version 0
		void * ptr;
	}d;
}ENTITY_CONF_BODY_CA;
DECLARE_ASN1_FUNCTIONS(ENTITY_CONF_BODY_CA)
DECLARE_CONF_MIGRATION_FUNCTIONS(ENTITY_CONF_BODY_CA, CONF_VERSION_0)

#define PUBLICATION_METHOD_TYPE_CERT 1		//!< The publication method is for a certificate
#define PUBLICATION_METHOD_TYPE_CRL 2		//!< The publication method is for a CRL

/*!
	This structure represents a publication method
*/
typedef struct st_PUBLICATION_METHOD
{
	ASN1_UTF8STRING * Name;				//!< The publication method's name
	ASN1_INTEGER * Type;				//!< The publication method's type
	ASN1_UTF8STRING * LibraryPath;		//!< The path to the publication plugin
	STACK_OF(PLUG_OPTION) * Options;	//!< The options for the publication method
}PUBLICATION_METHOD;
DECLARE_ASN1_FUNCTIONS(PUBLICATION_METHOD)

/*!
	This structure represents a list of publication methods
*/
typedef struct st_PUBLICATION_ENTRY
{
	ASN1_UTF8STRING * CaName;					//!< The name of the CA concerned by the methods
	STACK_OF(PUBLICATION_METHOD) * Methods;		//!< The publication methods
}PUBLICATION_ENTRY;
DECLARE_ASN1_FUNCTIONS(PUBLICATION_ENTRY)


#define PUB_CONF_VERSION CONF_VERSION_0
#define PUB_CONF_STRUCT_new ENTITY_CONF_BODY_PUB_0_new
#define PUB_CONF_PTR conf0

/*!
	This structure represents the body version 0 of a Publication conf
*/
typedef struct st_ENTITY_CONF_BODY_PUB_0
{
	STACK_OF(PUBLICATION_ENTRY) * Publications;		//!< The list of publication methods
	ASN1_INTEGER * OcspEnabled;						//!< Should the OCSP server be enabled
	ASN1_UTF8STRING * BindAddress;					//!< The OCSP server bind address
	ASN1_INTEGER * Port;							//!< The OCSP server port
}ENTITY_CONF_BODY_PUB_0;
DECLARE_ASN1_FUNCTIONS(ENTITY_CONF_BODY_PUB_0)

/*!
	This structure represents a Publication conf
*/
typedef struct st_ENTITY_CONF_BODY_PUB
{
	int type;
	union{
		ENTITY_CONF_BODY_PUB_0 * conf0;		//!< Conf version 0
		void * ptr;
	}d;
}ENTITY_CONF_BODY_PUB;
DECLARE_ASN1_FUNCTIONS(ENTITY_CONF_BODY_PUB)
DECLARE_CONF_MIGRATION_FUNCTIONS(ENTITY_CONF_BODY_PUB, CONF_VERSION_0)






#define ALLOW_SERVER_SIDE_RA_KEY_GEN 1		//!< RA conf flag to allow the RA entity to generate end-users private keys
#define ALLOW_PKCS10_IMPORT 2				//!< RA conf flag to allow the RA to import an end-user PKCS#10
#define ALLOW_RA_OP_SC_KEY_GEN 4			//!< RA conf flag to allow the RA Operator to generate the end-user's private key into a Smart Card
#define PUSH_KEYS_TO_KEY_ESCROW 8			//!< RA conf flag to tell the RA to push the key it generated to a key store
#define REMOVE_UNKNOWN_DN_FIELDS 16			//!< RA conf flag to tell the RA to remove the DN fields that would not be present into DnSpecs
#define ALLOW_LDAP_PROFILE_IMPORT 32		//!< RA conf flag to allow LDAP profiles importation
#define ALLOW_LDAP_AUTO_SYNCH 64			//!< RA conf flag to tell the RA that it should synchronize automatically with the LDAP


#define RA_CONF_VERSION CONF_VERSION_0
#define RA_CONF_STRUCT_new ENTITY_CONF_BODY_RA_0_new
#define RA_CONF_PTR conf0

/*!
	This structure represents the body version 0 of a RA conf
*/
typedef struct st_ENTITY_CONF_BODY_RA_0
{
	ASN1_BIT_STRING * Flags;				//!< The Flags
	STACK_OF(DN_SPECS) * DnSpecs;			//!< The DN specs
	STACK_OF(POLICY_VALUE) * Policies;		//!< The DN policies
	ASN1_INTEGER * MinKeyLen;				//!< The minimum key length
	ASN1_INTEGER * MinPwdLen;				//!< The minimum PKCS#12 password
	ASN1_INTEGER * default_validity;		//!< The default validity for a certificate
	ASN1_UTF8STRING * ldap_server;			//!< The address of the LDAP server
	ASN1_INTEGER * ldap_port;				//!< The port of the LDAP server
	ASN1_UTF8STRING * ldap_username;		//!< The username to use during the LDAP connection
	ASN1_UTF8STRING * ldap_password;		//!< The password to use during the LDAP connection
	ASN1_UTF8STRING * ldap_base;			//!< The base of the LDAP
	ASN1_INTEGER * ldap_profiles_group;		//!< Delegate the ownership on newly imported profiles to this group
	ASN1_UTF8STRING * ldap_filters;			//!< Some filters to add during LDAP auto synchro search
	ASN1_UTF8STRING * ldap_attr_name;		//!< The name of the LDAP attribute that represents the user's UID
}ENTITY_CONF_BODY_RA_0;
DECLARE_ASN1_FUNCTIONS(ENTITY_CONF_BODY_RA_0)

/*!
	This structure represents a RA conf
*/
typedef struct st_ENTITY_CONF_BODY_RA
{
	int type;
	union{
		ENTITY_CONF_BODY_RA_0 * conf0;	//!< Conf version 0
		void * ptr;
	}d;
}ENTITY_CONF_BODY_RA;
DECLARE_ASN1_FUNCTIONS(ENTITY_CONF_BODY_RA)
DECLARE_CONF_MIGRATION_FUNCTIONS(ENTITY_CONF_BODY_RA, CONF_VERSION_0)




/*!
	This structure respresents the conf body
*/
typedef struct st_ENTITY_CONF_BODY
{
	int type;
	union{
		ASN1_OCTET_STRING * conf;			//!< An undefined conf
		ENTITY_CONF_BODY_RA * ra_conf;		//!< A RA conf
		ENTITY_CONF_BODY_CA * ca_conf;		//!< A CA conf
		ENTITY_CONF_BODY_REP * rep_conf;	//!< A Repository conf
		ENTITY_CONF_BODY_PKI * pki_conf;	//!< A PKI conf
		ENTITY_CONF_BODY_PUB * pub_conf;	//!< A Publication conf
	}d;
}ENTITY_CONF_BODY;
DECLARE_ASN1_FUNCTIONS(ENTITY_CONF_BODY)
DECLARE_ASN1_CPY_FUNCTIONS(ENTITY_CONF_BODY)


#define REP_ENTRY_INFO_FLAG_FIREWALLED			1	//!< The repository is firewalled
#define REP_ENTRY_INFO_FLAG_SHOULD_FULL_SYNCHRO	2	//!< The repository should do a full synchro

/*!
	This structure respresents a repository information
*/
typedef struct st_REP_ENTRY_INFO
{
	X509 * RepositorySSL;			//!< The repository certificate
	ASN1_UTF8STRING * Name;			//!< The repository name
	ASN1_UTF8STRING * Address;		//!< The repository address
	ASN1_INTEGER * Port;			//!< The repository port
	ASN1_BIT_STRING	* Flags;		//!< The flags
}REP_ENTRY_INFO;
DECLARE_ASN1_FUNCTIONS(REP_ENTRY_INFO)
DECLARE_ASN1_CPY_FUNCTIONS(REP_ENTRY_INFO)


/*!
	This structure respresents a list of repositories
*/
typedef struct st_REPS_INFO
{
	STACK_OF(REP_ENTRY_INFO) * list;		//!< The list
}REPS_INFO;
DECLARE_ASN1_FUNCTIONS(REPS_INFO)
DECLARE_ASN1_CPY_FUNCTIONS(REPS_INFO)


/*!
	This structure respresents a CA information
*/
typedef struct st_CA_ENTRY_INFO
{
	X509 * CaSSL;				//!< The CA certificate
	ASN1_UTF8STRING * Name;		//!< The CA name
}CA_ENTRY_INFO;
DECLARE_ASN1_FUNCTIONS(CA_ENTRY_INFO)

/*!
	This structure respresents a list of CAs
*/
typedef struct st_CAS_INFO
{
	STACK_OF(CA_ENTRY_INFO) * list;		//!< The list
}CAS_INFO;
DECLARE_ASN1_FUNCTIONS(CAS_INFO)

/*!
	This structure respresents a RA information
*/
typedef struct st_RA_ENTRY_INFO
{
	X509 * RaSSL;				//!< The RA certificate
	ASN1_UTF8STRING * Name;		//!< The RA name
}RA_ENTRY_INFO;
DECLARE_ASN1_FUNCTIONS(RA_ENTRY_INFO)

/*!
	This structure respresents a list of RAs
*/
typedef struct st_RAS_INFO
{
	STACK_OF(RA_ENTRY_INFO) * list;		//!< The list
}RAS_INFO;
DECLARE_ASN1_FUNCTIONS(RAS_INFO)


/*!
	This structure respresents a Publication information
*/
typedef struct st_PUB_ENTRY_INFO
{
	ASN1_UTF8STRING * Name;			//!< The Publication name
	X509 * PubSSL;					//!< The Publication certificate
	X509 * PubOCSP;					//!< The Publication OCSP certificate
}PUB_ENTRY_INFO;
DECLARE_ASN1_FUNCTIONS(PUB_ENTRY_INFO)
DECLARE_ASN1_CPY_FUNCTIONS(PUB_ENTRY_INFO)

/*!
	This structure respresents a list of Publications
*/
typedef struct st_PUBS_INFO
{
	STACK_OF(PUB_ENTRY_INFO) * list;		//!< The list
}PUBS_INFO;
DECLARE_ASN1_FUNCTIONS(PUBS_INFO)

/*!
	This structure respresents a Key Store information
*/
typedef struct st_KEY_ENTRY_INFO
{
	X509 * KeySSL;				//!< The Key Store certificate
	ASN1_UTF8STRING * Name;		//!< The Key Store name
}KEY_ENTRY_INFO;
DECLARE_ASN1_FUNCTIONS(KEY_ENTRY_INFO)
DECLARE_ASN1_CPY_FUNCTIONS(KEY_ENTRY_INFO)

/*!
	This structure respresents a list of Key Stores
*/
typedef struct st_KEYS_INFO
{
	STACK_OF(KEY_ENTRY_INFO) * list;		//!< The list
}KEYS_INFO;
DECLARE_ASN1_FUNCTIONS(KEYS_INFO)
DECLARE_ASN1_CPY_FUNCTIONS(KEYS_INFO)


/*!
	This enumeration defines the possible ACLs
*/
typedef enum en_ACL_TYPE
{
	ACL_TYPE_AUTHENTICATE_ON_ENTITY			= 1,		//!< The user is allowed to authenticate on the entity
	ACL_TYPE_MANAGE_USERS					= 2,		//!< The user is allowed to manage the users
	ACL_TYPE_MANAGE_ENTITIES				= 3,		//!< The user is allowed to manage the entities
	ACL_TYPE_VIEW_LOGS						= 4,		//!< The user is allowed to view the logs
	ACL_TYPE_MANAGE_ACLS					= 5,		//!< The user is allowed to manage the ACLs
	ACL_TYPE_MODIFY_PKI_STATUS				= 6,		//!< The user is allowed to set the PKI online/offline
	ACL_TYPE_IMPORT_CONF					= 7,		//!< The user is allowed to import a conf
	ACL_TYPE_SEND_ADMIN_MAIL				= 8,		//!< The user is allowed to send admin mails
	ACL_TYPE_CREATE_ROOT_CA					= 9,		//!< The user is allowed to create a root CA
	ACL_TYPE_CREATE_CHILD_CA				= 10,		//!< The user is allowed to create a child CA
	ACL_TYPE_CA_SIGN_CERT					= 11,		//!< The user is allowed to sign a certificate
	ACL_TYPE_CA_REVOKE_CERT					= 12,		//!< The user is allowed to revoke a certificate
	ACL_TYPE_CA_GENERATE_CRL				= 13,		//!< The user is allowed to generate a CRL
	ACL_TYPE_IMPORT_NEWPKI_OBJECTS			= 14,		//!< The user is allowed to imports newpki objects
	ACL_TYPE_RA_IMPORT_PROFILE				= 15,		//!< The user is allowed to imprt a profile
	ACL_TYPE_RA_DELEGATE_PROFILE_OWNERSHIP	= 16,		//!< The user is allowed to delegate a profiles ownership
	ACL_TYPE_RA_REQUEST_CERT				= 17,		//!< The user is allowed to request a certificate
	ACL_TYPE_RA_REVOKE_CERT					= 18		//!< The user is allowed to requets a certificate revocation
}ACL_TYPE;


/*!
	This structure respresents a group
*/
typedef struct st_USERS_GROUP
{
	ASN1_UTF8STRING * name;					//!< The name of the group
	ASN1_INTEGER * serial;					//!< The unique id of the group (will be used for the ACL)
	STACK_OF(ASN1_INTEGER) * users_serial;	//!< The list of users being part of the group
}USERS_GROUP;
DECLARE_ASN1_FUNCTIONS(USERS_GROUP)

/*!
	This structure respresents a list of groups
*/
typedef struct st_USERS_GROUPS
{
	STACK_OF(USERS_GROUP) * list;		//!< The list
}USERS_GROUPS;
DECLARE_ASN1_FUNCTIONS(USERS_GROUPS)
DECLARE_ASN1_CPY_FUNCTIONS(USERS_GROUPS)


#define ACL_ENTRY_TYPE_USER		1		//!< The ACL concerns a user
#define ACL_ENTRY_TYPE_GROUP	2		//!< The ACL concerns a group

/*!
	This structure respresents an ACL entry
*/
typedef struct st_ACL_ENTRY
{
	ASN1_UTF8STRING * name;			//!< The name of user/group
	ASN1_INTEGER * type;			//!< The type of ACL 
	ASN1_INTEGER * serial;			//!< The serial of the user/group
	ASN1_BIT_STRING * acls;			//!< The list of allowed actions
}ACL_ENTRY;
DECLARE_ASN1_FUNCTIONS(ACL_ENTRY)

/*!
	This structure respresents an PKI admin
*/
typedef struct st_PKI_ADMIN_ENTRY
{
	ASN1_INTEGER * serial;			//!< The serial of its certificate
	ASN1_UTF8STRING * email;		//!< Its email address
}PKI_ADMIN_ENTRY;
DECLARE_ASN1_FUNCTIONS(PKI_ADMIN_ENTRY)
DECLARE_ASN1_CPY_FUNCTIONS(PKI_ADMIN_ENTRY)


/*!
	This structure represents the list of pki admin
*/
typedef struct st_PKI_ADMIN_SERIALS
{
	STACK_OF(PKI_ADMIN_ENTRY) * admins;		//!< The list
}PKI_ADMIN_SERIALS;
DECLARE_ASN1_FUNCTIONS(PKI_ADMIN_SERIALS)
DECLARE_ASN1_CPY_FUNCTIONS(PKI_ADMIN_SERIALS)



/*!
	This structure represents a complete ACL information for an entity
*/
typedef struct st_X509_ACL
{
	STACK_OF(ACL_ENTRY) *acl_entries;		//!< The list of ACLs
	STACK_OF(ASN1_INTEGER) *acl_type;		//!< The list supported ACLs by the entity
	PKI_ADMIN_SERIALS * AdminSerials;		//!< The list of PKI admins
}X509_ACL;
DECLARE_ASN1_FUNCTIONS(X509_ACL)
DECLARE_ASN1_CPY_FUNCTIONS(X509_ACL)


/*!
	This structure represents the CRLs for the PKI internal CA
*/
typedef struct st_INTERNAL_PKI_CRL
{
	X509_CRL * RootCaCrl;		//!< The CRL for the ROOT internal CA
	X509_CRL * UsersCaCrl;		//!< The CRL for the Users internal CA
	X509_CRL * EntitiesCaCrl;	//!< The CRL for the Entitis internal CA
	X509_CRL * OcspCaCrl;		//!< The CRL for the OCSP internal CA
	int references;
}INTERNAL_PKI_CRL;
DECLARE_ASN1_FUNCTIONS(INTERNAL_PKI_CRL)
DECLARE_ASN1_CPY_FUNCTIONS(INTERNAL_PKI_CRL)

/*!
	This structure represents the Certificates for the PKI internal CA
*/
typedef struct st_INTERNAL_PKI_CA
{
	X509 * RootCa;			//!< The certificate for the ROOT internal CA
	X509 * UsersCa;			//!< The certificate for the Users internal CA
	X509 * EntitiesCa;		//!< The certificate for the Entitis internal CA
	X509 * OcspCa;			//!< The certificate for the OCSP internal CA
	X509 * PkiCert;			//!< The certificate for the PKI entity
	int references;
}INTERNAL_PKI_CA;
DECLARE_ASN1_FUNCTIONS(INTERNAL_PKI_CA)
DECLARE_ASN1_CPY_FUNCTIONS(INTERNAL_PKI_CA)


/*!
	This structure an audit entry
*/
typedef struct st_ENTITY_AUDIT_ENTRY
{
	ASN1_UTF8STRING * object;		//!< The object name
	ASN1_INTEGER * type;			//!< The log type
	ASN1_INTEGER * status;			//!< The log status
	ASN1_UTF8STRING * user;			//!< The user that generated the log
}ENTITY_AUDIT_ENTRY;
DECLARE_ASN1_FUNCTIONS(ENTITY_AUDIT_ENTRY)
DECLARE_ASN1_CPY_FUNCTIONS(ENTITY_AUDIT_ENTRY)


/*!
	This structure represents a list of audits
*/
typedef struct st_ENTITY_AUDIT
{
	STACK_OF(ENTITY_AUDIT_ENTRY) * audits;	//!< The list
}ENTITY_AUDIT;
DECLARE_ASN1_FUNCTIONS(ENTITY_AUDIT)
DECLARE_ASN1_CPY_FUNCTIONS(ENTITY_AUDIT)

/*!
	This structure represents the smtp configuration
*/
typedef struct st_EMAIL_CONF
{
	ASN1_UTF8STRING * server;		//!< The SMTP server
	ASN1_INTEGER * port;			//!< The SMTP port
}EMAIL_CONF;
DECLARE_ASN1_FUNCTIONS(EMAIL_CONF)
DECLARE_ASN1_CPY_FUNCTIONS(EMAIL_CONF)


/*!
	This structure is an entity conf but not crypted for it
*/
typedef struct st_ENTITY_CONF
{
	ASN1_UTF8STRING * Name;					//!< The name of the entity
	ASN1_INTEGER * version;					//!< The version of the entity conf
	X509_ACL * acls;						//!< The ACLs for the entity
	ENTITY_CONF_BODY * body;				//!< The body of the entity conf
	REPS_INFO * Repositories;		//!< The repositories list is exported as well
	RAS_INFO * Ras;							//!< The RAs list is exported as well
	CAS_INFO * Cas;							//!< The CAs list is exported as well
	KEYS_INFO * KeyStores;					//!< The Key stores list is exported as well
	PUBS_INFO * Publications;				//!< The Publications list is exported as well
	INTERNAL_PKI_CRL * crls;				//!< The current CRLs for the internal CAs
	EMAIL_CONF * mail_conf;					//!< The mail conf
	ENTITY_AUDIT * audits;					//!< The audits
	USERS_GROUPS * groups;					//!< The groups
}ENTITY_CONF;
DECLARE_ASN1_ENCRYPT_SIGN(ENTITY_CONF)
DECLARE_ASN1_FUNCTIONS(ENTITY_CONF)
DECLARE_ASN1_CPY_FUNCTIONS(ENTITY_CONF)


/*!
	This structure represents an entity conf, the conf is crypted 
	for the destination entity and signed by us.
	The entity will decrypt its own conf and verify the signature.
*/
typedef struct st_ENTITY_CONF_CRYPTED
{
	X509_PUBKEY * recipient;			//!< The public key of the destination entity	
	ASN1_ENCRYPT_SIGN * crypted;		//!< The crypted and signed datas
	int references;
}ENTITY_CONF_CRYPTED;
DECLARE_PEM_CONVERT(ENTITY_CONF_CRYPTED)
DECLARE_ASN1_FUNCTIONS(ENTITY_CONF_CRYPTED)
DECLARE_ASN1_CPY_FUNCTIONS(ENTITY_CONF_CRYPTED)


/*!
	This structure represents a global PKI conf.
	Necessary for the signature generation of all the confs together.
*/
typedef struct st_EXPORTED_PKI_CONF_BODY
{
	ASN1_INTEGER * version;								//!< The version of the "alltogether" configuration
	STACK_OF(ENTITY_CONF_CRYPTED) * all_confs;			//!< All the crypted and signed confs
}EXPORTED_PKI_CONF_BODY;
DECLARE_ASN1_FUNCTIONS(EXPORTED_PKI_CONF_BODY)


/*!
	This structure represents the whole pki configuration for each
	entity, it is signed by the PKI entity priv key.
	This structure is sent to the repositories, they're
	in charge of verifying the "global" signature
*/
typedef struct st_EXPORTED_PKI_CONF
{
	EXPORTED_PKI_CONF_BODY * confs;			//!< The global conf
	X509_SIG * sig;							//!< The signature of the global conf
}EXPORTED_PKI_CONF;
DECLARE_ASN1_FUNCTIONS(EXPORTED_PKI_CONF)
DECLARE_PEM_CONVERT(EXPORTED_PKI_CONF)
DECLARE_ASN1_CPY_FUNCTIONS(EXPORTED_PKI_CONF)


/*!
	This enumeration defines the list of internal CAs.
*/
typedef enum en_INTERNAL_CA_TYPE
{
	INTERNAL_CA_TYPE_ROOT = 1,		//!< The ROOT internal CA
	INTERNAL_CA_TYPE_ENTITY = 2,	//!< The Entities internal CA
	INTERNAL_CA_TYPE_USER = 3,		//!< The Users internal CA
	INTERNAL_CA_TYPE_OCSP = 4		//!< The OCSP internal CA
}INTERNAL_CA_TYPE;


/*! \brief This function returns the string representation of an ACL.
 *  \param Type [IN] The ACL.
 *  \return The string representation of an ACL.
 */
char * GetACL_String(ACL_TYPE Type);

#endif
