 /*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/


#ifndef ASN1_ENTITY_H
#define ASN1_ENTITY_H

#include <openssl/x509.h>
#include <openssl/asn1.h>
#include <openssl/asn1t.h>
#include <openssl/x509v3.h>
#include <openssl/evp.h>
#include <openssl/pem.h>

#include "my_asn1.h"
#include "ASN1_conf.h"

#define GEN_PRIVATE_KEY_TYPE_KEYLEN 0		//!< The key should be generated
#define GEN_PRIVATE_KEY_TYPE_ENGINE 1		//!< The key should be imported from an ENGINE

/*!
	This structure represents the information needed to generate a private key
*/
typedef struct st_GEN_PRIVATE_KEY
{
	int type;
	union
	{
		ASN1_INTEGER * keylen;			//!< The key length
		ASN1_UTF8STRING * keyid;		//!< The key id when imported from an ENGINE
	}d;
}GEN_PRIVATE_KEY;
DECLARE_ASN1_FUNCTIONS(GEN_PRIVATE_KEY)

/*!
	This structure represents a PKI creation request
*/
typedef struct st_PKI_CREATION_REQ
{
	ASN1_UTF8STRING * Email;			//!< The email address of the PKI
	ASN1_UTF8STRING * AdminEmail;		//!< The email address of the admin
	ASN1_INTEGER * validity;			//!< The validity in days for the internal CAs
	X509_PUBKEY * AdminPubKey;			//!< The public key for the Admin
	X509_NAME * dn;						//!< The DN for the internal CAs
	GEN_PRIVATE_KEY * root_ca;			//!< The private key info for the internal ROOT CA certificate
	GEN_PRIVATE_KEY * entities_ca;		//!< The private key info for the internal Entities CA certificate
	GEN_PRIVATE_KEY * users_ca;			//!< The private key info for the internal Users CA certificate
	GEN_PRIVATE_KEY * ocsp_ca;			//!< The private key info for the internal OCSP CA certificate
	GEN_PRIVATE_KEY * entity_key;		//!< The private key info for the PKI entity certificate
}PKI_CREATION_REQ;
DECLARE_ASN1_FUNCTIONS(PKI_CREATION_REQ)

/*!
	This structure represents a Publication creation request
*/
typedef struct st_PUB_CREATION_REQ
{
	GEN_PRIVATE_KEY * entity_key;	//!< The private key info for the Publication entity certificate
	GEN_PRIVATE_KEY * ocsp_key;		//!< The private key info for the Publication OCSP certificate
}PUB_CREATION_REQ;
DECLARE_ASN1_FUNCTIONS(PUB_CREATION_REQ)


/*!
	This structure represents an entity creation request body
*/
typedef struct st_ENTITY_CREATION_DATAS
{
	int type;
	union
	{
		PKI_CREATION_REQ * pki_create;		//!< Creating a PKI
		PUB_CREATION_REQ * pub_create;		//!< Creating a Publication
		GEN_PRIVATE_KEY * entity_key;		//!< Creating a RA, CA, KeyStore, Repository...
	}d;
}ENTITY_CREATION_DATAS;
DECLARE_ASN1_FUNCTIONS(ENTITY_CREATION_DATAS)

/*!
	This structure represents an entity creation request
*/
typedef struct st_ENTITY_CREATION_REQ
{
	ASN1_UTF8STRING * Name;				//!< The entity name
	ENTITY_CREATION_DATAS * Datas;		//!< The body
}ENTITY_CREATION_REQ;
DECLARE_ASN1_FUNCTIONS(ENTITY_CREATION_REQ)


/*!
	This structure represents a Publication creation response
*/
typedef struct st_PUB_ENTITY_CREATION_RESP
{
	X509_PUBKEY * entity_key;			//!< The Publication entity public key
	X509_PUBKEY * ocsp_key;				//!< The Publication OCSP public key
}PUB_ENTITY_CREATION_RESP;
DECLARE_ASN1_FUNCTIONS(PUB_ENTITY_CREATION_RESP)


/*!
	This structure represents an entity creation response
*/
typedef struct st_ENTITY_CREATION_RESP
{
	int type;
	union{
		X509_PUBKEY * entity_pub_key;				//!< The entity public key for a RA, CA, KeyStore, Repository...
		PUB_ENTITY_CREATION_RESP * pub_create;		//!< The creation response for a Publication
		STACK_OF(X509) * pki_create;				//!< The certificate of the PKI admin with all its parents
	}d;
}ENTITY_CREATION_RESP;
DECLARE_ASN1_FUNCTIONS(ENTITY_CREATION_RESP)



/*!
	This structure represents an entity info
*/
typedef struct st_ENTITY_ENTRY_INFO
{
	ASN1_UTF8STRING * Name;			//!< The entity name
	ASN1_INTEGER * Type;			//!< The entity type
	X509 * certificate;				//!< The entity certificate
	ASN1_BIT_STRING * flags;		//!< The flags (only used for the links)
}ENTITY_ENTRY_INFO;
DECLARE_ASN1_FUNCTIONS(ENTITY_ENTRY_INFO)
DECLARE_ASN1_CPY_FUNCTIONS(ENTITY_ENTRY_INFO)


/*!
	This strucutre represents the links between entities
*/
typedef struct st_ENTITY_LINKS
{
	ENTITY_ENTRY_INFO * src;				//!< The source of the link
	STACK_OF(ENTITY_ENTRY_INFO) * dsts;		//!< All the destinations
}ENTITY_LINKS;
DECLARE_ASN1_FUNCTIONS(ENTITY_LINKS)

/*!
	This strucutre represents a list of links
*/
typedef struct st_ENTITIES_LINKS
{
	STACK_OF(ENTITY_LINKS) * links;		//!< The list
}ENTITIES_LINKS;
DECLARE_ASN1_FUNCTIONS(ENTITIES_LINKS)
DECLARE_ASN1_CPY_FUNCTIONS(ENTITIES_LINKS)



/*!
	This strucutre represents a signature request body for a Publication entity
*/
typedef struct st_ENTITY_SIGNATURE_REQ_PUB
{
	X509_PUBKEY * EntityPubKey;		//!< The Publication public key
	X509_PUBKEY * OcspPubKey;		//!< The Publication OCSP public key
}ENTITY_SIGNATURE_REQ_PUB;
DECLARE_ASN1_FUNCTIONS(ENTITY_SIGNATURE_REQ_PUB)


/*!
	This strucutre represents a signature request body for a Repository entity
*/
typedef struct st_ENTITY_SIGNATURE_REQ_REP
{
	X509_PUBKEY * EntityPubKey;	//!< The Repository public key
	ASN1_UTF8STRING * Address;	//!< The Repository address
	ASN1_INTEGER * Port;		//!< The Repository port
}ENTITY_SIGNATURE_REQ_REP;
DECLARE_ASN1_FUNCTIONS(ENTITY_SIGNATURE_REQ_REP)


/*!
	This strucutre represents an entity signature request body
*/
typedef struct st_ENTITY_SIGNATURE_REQ_BODY
{
	int type;
	union{
		X509_PUBKEY * EntityPubKey;				//!< The entity public key, for RA, CA, KeyStore
		ENTITY_SIGNATURE_REQ_REP * sign_rep;	//!< The body for a repository
		ENTITY_SIGNATURE_REQ_PUB * sign_pub;	//!< The body for a publication
	}d;
}ENTITY_SIGNATURE_REQ_BODY;
DECLARE_ASN1_FUNCTIONS(ENTITY_SIGNATURE_REQ_BODY)


/*!
	This strucutre represents an entity signature request
*/
typedef struct st_ENTITY_SIGNATURE_REQ
{
	ASN1_UTF8STRING * Name;					//!< The name of the entity
	ASN1_UTF8STRING * Email;				//!< The email address of the entity
	ENTITY_SIGNATURE_REQ_BODY * body;		//!< The body
}ENTITY_SIGNATURE_REQ;
DECLARE_ASN1_FUNCTIONS(ENTITY_SIGNATURE_REQ)
DECLARE_PEM_CONVERT(ENTITY_SIGNATURE_REQ);





/*!
	This strucutre represents a signature response body for a Publication entity
*/
typedef struct st_ENTITY_SIGNATURE_RESP_PUB
{
	X509 * EntityCert;			//!< The entity certificate
	X509 * OcspCert;			//!< The OCSP certificate
}ENTITY_SIGNATURE_RESP_PUB;
DECLARE_ASN1_FUNCTIONS(ENTITY_SIGNATURE_RESP_PUB)


/*!
	This strucutre represents an entity signature response body
*/
typedef struct st_ENTITY_SIGNATURE_RESP_BODY
{
	int type;
	union{
		X509 * EntityCert;						//!< The entity certificate for a RA, CA, Repository, KeyStore
		ENTITY_SIGNATURE_RESP_PUB * sign_pub;	//!< The response for a publication entity
	}d;
}ENTITY_SIGNATURE_RESP_BODY;
DECLARE_ASN1_FUNCTIONS(ENTITY_SIGNATURE_RESP_BODY)


/*!
	This strucutre represents an entity signature response
*/
typedef struct st_ENTITY_SIGNATURE_RESP
{
	ENTITY_SIGNATURE_RESP_BODY * body;		//!< The body
	ENTITY_CONF_CRYPTED * conf;				//!< The entity conf
	INTERNAL_PKI_CA * cas;					//!< The internal CAs info
}ENTITY_SIGNATURE_RESP;
DECLARE_ASN1_FUNCTIONS(ENTITY_SIGNATURE_RESP)
DECLARE_PEM_CONVERT(ENTITY_SIGNATURE_RESP);
DECLARE_ASN1_ENCRYPT_SIGN(ENTITY_SIGNATURE_RESP)


/*!
	This strucutre represents an entity initialization response
*/
typedef struct st_ENTITY_INIT_REQ
{
	ASN1_UTF8STRING * name;					//!< The name of the entity to intialize
	ENTITY_SIGNATURE_RESP * sign_resp;		//!< The entity signature response
}ENTITY_INIT_REQ;
DECLARE_ASN1_FUNCTIONS(ENTITY_INIT_REQ)


/*!
	This strucutre represents a ROOT CA creation request
*/
typedef struct st_REQ_CREATE_ROOT_CA
{
	GEN_PRIVATE_KEY * privKey;					//!< The private key
	X509_NAME * dn;								//!< The DN
	ASN1_INTEGER * validity;					//!< The validity in days
	STACK_OF(EXTENSION_VALUE) * extensions;		//!< The self-signed certificate extensions
}REQ_CREATE_ROOT_CA;
DECLARE_ASN1_FUNCTIONS(REQ_CREATE_ROOT_CA)

/*!
	This strucutre represents a CHILD CA creation request
*/
typedef struct st_REQ_CREATE_CHILD_CA
{
	GEN_PRIVATE_KEY * privKey;	//!< The private key
	X509_NAME * dn;				//!< The DN
}REQ_CREATE_CHILD_CA;
DECLARE_ASN1_FUNCTIONS(REQ_CREATE_CHILD_CA)




#endif

