/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/


#ifndef ASN1_LOGS_H
#define ASN1_LOGS_H

#include <openssl/x509.h>
#include <openssl/asn1.h>
#include <openssl/asn1t.h>
#include <openssl/x509v3.h>
#include <openssl/evp.h>
#include <openssl/pem.h>

#include "my_asn1.h"


/*!
  This structure represents the body of a Log.
*/
typedef struct st_LOG_ENTRY_BODY
{
	ASN1_INTEGER * log_status;		//!< The log status
	ASN1_INTEGER * log_type;		//!< The type of log
	ASN1_UTF8STRING * user;			//!< The user that generated the log
	ASN1_UTF8STRING * object_name;	//!< An optional obect name
	ASN1_UTF8STRING * error;		//!< An optional error message
	ASN1_INTEGER * log_date;		//!< The date of logging
}LOG_ENTRY_BODY;
DECLARE_ASN1_FUNCTIONS(LOG_ENTRY_BODY)



/*!
  This structure represents a Log signature.
*/
typedef struct st_LOG_ENTRY_SIG
{
	ASN1_BIT_STRING * signature;		//!< The signature
	X509_ALGOR * sig_alg;				//!< The signature algo
}LOG_ENTRY_SIG;
DECLARE_ASN1_FUNCTIONS(LOG_ENTRY_SIG)
DECLARE_PEM_CONVERT(LOG_ENTRY_SIG)

/*!
  This structure represents a Log entry.
*/
typedef struct st_LOG_ENTRY
{
	ASN1_INTEGER * log_id;			//!< The log id
	LOG_ENTRY_BODY * body;			//!< The log entry body
	LOG_ENTRY_SIG * sig;			//!< The log entry signature
}LOG_ENTRY;
DECLARE_ASN1_FUNCTIONS(LOG_ENTRY)

/*!
  This structure is used to enumerate logs
*/
typedef struct st_ADMIN_REQ_ENUM_LOGS
{
	ASN1_INTEGER * index;			//!< The index
	ASN1_INTEGER * max;				//!< The maximum number of logs to return
	ASN1_UTF8STRING * object;		//!< The object name
	ASN1_INTEGER * type;			//!< The log type
	ASN1_INTEGER * date_from;		//!< The starting date of logs
	ASN1_INTEGER * date_to;			//!< The ending date of logs
	ASN1_INTEGER * status;			//!< The log status
	ASN1_UTF8STRING * user;			//!< The user that generated the log
}ADMIN_REQ_ENUM_LOGS;
DECLARE_ASN1_FUNCTIONS(ADMIN_REQ_ENUM_LOGS)


/*!
  This enumeration defines the possible type of logs
*/
typedef enum LOG_MESSAGE_TYPE
{
	LOG_MESSAGE_TYPE_USER_LOGIN				= 1,	//!< User login
	LOG_MESSAGE_TYPE_USER_ADD				= 2,	//!< The creation of a user
	LOG_MESSAGE_TYPE_USER_MOD				= 3,	//!< The modification of a user
	LOG_MESSAGE_TYPE_PWD_MOD				= 4,	//!< The modification of a user's paswword
	LOG_MESSAGE_TYPE_ENTITY_ADD				= 5,	//!< The creation of a entity
	LOG_MESSAGE_TYPE_ENTITY_DEL				= 6,	//!< The deletion of a ca
	LOG_MESSAGE_TYPE_SIGN_ENTITY			= 7,	//!< The signature of an entity certificate
	LOG_MESSAGE_TYPE_USER_REVOCATION		= 8,	//!< The revocation of an user certificate
	LOG_MESSAGE_TYPE_ENTITY_REVOCATION		= 9,	//!< The revocation of an entity certificate
	LOG_MESSAGE_TYPE_CONFIG_PUSH			= 10,	//!< The push of the configuration on a repository
	LOG_MESSAGE_TYPE_SET_OFFLINE			= 11,	//!< Set offline status
	LOG_MESSAGE_TYPE_GET_CONFIGURATION		= 12,	//!< Get configuration
	LOG_MESSAGE_TYPE_CONFIG_IMPORT			= 13,	//!< The importation of the configuration in the repository
	LOG_MESSAGE_TYPE_ENTITY_CONFIG_GET		= 14,	//!< The exportation of an entity conf
	LOG_MESSAGE_TYPE_ENTITY_INIT			= 15,	//!< The initialisation of an entity
	LOG_MESSAGE_TYPE_SET_REPOSITORIES		= 16,	//!< The modification of the repositories
	LOG_MESSAGE_TYPE_ENTITY_LOAD			= 17,	//!< The loading of an entity
	LOG_MESSAGE_TYPE_LOCK_RESOURCE			= 18,	//!< The locking of a resource
	LOG_MESSAGE_TYPE_UNLOCK_RESOURCE		= 19,	//!< The unlocking of a resource
	LOG_MESSAGE_TYPE_SET_ENTITY_ACL			= 20,	//!< The modification of an entity acl
	LOG_MESSAGE_TYPE_SET_ENTITY_MAIL_CONF	= 21,	//!< The modification of an entity mail configuration
	LOG_MESSAGE_TYPE_SET_ENTITY_CONF		= 22,	//!< The modification of an entity configuration
	LOG_MESSAGE_TYPE_SEND_ADMIN_MAIL		= 24,	//!< Sending an admin notification mail
	LOG_MESSAGE_TYPE_SEND_MAIL				= 25,	//!< Sending a notification mail
	LOG_MESSAGE_TYPE_ADD_ADMIN_MAIL_QUEUE	= 26,	//!< Adding an admin notification mail to queue
	LOG_MESSAGE_TYPE_ADD_MAIL_QUEUE			= 27,	//!< Adding a notification mail to queue
	LOG_MESSAGE_TYPE_SET_ENTITY_AUDITS		= 28,	//!< The modification of an entity audits
	LOG_MESSAGE_TYPE_SET_ENTITIES_LINKS		= 29,	//!< The modification of the entities links
	LOG_MESSAGE_TYPE_ENTITY_GET_MY_CONF		= 30,	//!< The retreival of my conf from a rep
	LOG_MESSAGE_TYPE_OBJECTS_IMPORT			= 31,	//!< The importation of the CA objects in the repository
	LOG_MESSAGE_TYPE_OBJECTS_PUSH			= 32,	//!< The push of the CA objects on a repository
	LOG_MESSAGE_TYPE_CREATE_ROOT_CA			= 33,	//!< Creating a ROOT CA
	LOG_MESSAGE_TYPE_GET_MY_REQS			= 34,	//!< An enity is retreiving its requests
	LOG_MESSAGE_TYPE_SIGN_CSR				= 35,	//!< An administrator is signing a req from a CA
	LOG_MESSAGE_TYPE_GEN_CRL				= 36,	//!< A crl Generation
	LOG_MESSAGE_TYPE_CREATE_CHILD_CA		= 37,	//!< Creating a Child CA
	LOG_MESSAGE_TYPE_IMPORT_CHILD_CA_CERT	= 38,	//!< Importing child CA cert
	LOG_MESSAGE_TYPE_CERT_REVOCATION		= 39,	//!< The revocation of a certificate
	LOG_MESSAGE_TYPE_SET_GROUPS				= 40,	//!< The modification of the groups
	LOG_MESSAGE_TYPE_ADD_GROUP				= 41,	//!< Adding a group
	LOG_MESSAGE_TYPE_DEL_GROUP				= 42,	//!< Deleting a group
	LOG_MESSAGE_TYPE_REN_GROUP				= 43,	//!< Renaming a group
	LOG_MESSAGE_TYPE_GET_LOCAL_CONF			= 44,	//!< When a user request the entity conf
	LOG_MESSAGE_TYPE_IMPORT_PROFILE			= 45,	//!< When a user imports a profile
	LOG_MESSAGE_TYPE_GET_MY_RESPS			= 46,	//!< An entity is retreiving its responses
	LOG_MESSAGE_TYPE_INSERT_REQ				= 47,	//!< An enity is retreiving its requests
	LOG_MESSAGE_TYPE_INSERT_RESP			= 48,	//!< An entity is retreiving its responses
	LOG_MESSAGE_TYPE_CERT_REQUEST			= 49,	//!< A new cert request for a profile
	LOG_MESSAGE_TYPE_DELETE_PKCS12			= 50,	//!< Delete a pkcs12
	LOG_MESSAGE_TYPE_NEW_REQUET				= 51,	//!< A new request is received
	LOG_MESSAGE_TYPE_CA_CERT				= 52,	//!< A new certification request
	LOG_MESSAGE_TYPE_CA_REV					= 53,	//!< A new revocation request
	LOG_MESSAGE_TYPE_LOADING_PUB_METH		= 54,	//!< Loading publication method
	LOG_MESSAGE_TYPE_CERT_PUBLICATION		= 55,	//!< A cert publication request
	LOG_MESSAGE_TYPE_REV_PUBLICATION		= 56,	//!< A rev publication request
	LOG_MESSAGE_TYPE_CRL_PUBLICATION		= 57,	//!< A crl publication request
	LOG_MESSAGE_TYPE_LDAP_CONNECTION		= 58,	//!< A ldap connction
	LOG_MESSAGE_TYPE_LDAP_AUTO_SYNCH_START	= 59,	//!< A ldap synch start
	LOG_MESSAGE_TYPE_CERT_REV_RESP			= 60,	//!< The response for a revocation
	LOG_MESSAGE_TYPE_CERT_SIG_RESP			= 61,	//!< The response for a signature
	LOG_MESSAGE_TYPE_CERT_RNW_RESP			= 62,	//!< The response for a renewal
	LOG_MESSAGE_TYPE_LDAP_AUTO_SYNCH_IMPRT	= 63,	//!< A ldap synch profile import
	LOG_MESSAGE_TYPE_OCSP_START				= 64,	//!< Starting a OCSP responder
	LOG_MESSAGE_TYPE_OCSP_STOP				= 65,	//!< Stoping a OCSP responder
	LOG_MESSAGE_TYPE_OCSP_REQUEST			= 66,	//!< A new OCSP request
	LOG_MESSAGE_TYPE_ADMIN_REQUEST			= 67	//!< A new Admin request
};


/*!
  This enumeration defines the possible status of logs
*/
typedef enum LOG_MESSAGE_STATUS
{
	LOG_STATUS_TYPE_REQUEST = 1,	//!< Logging a request event
	LOG_STATUS_TYPE_SUCCESS = 2,	//!< Logging a success event
	LOG_STATUS_TYPE_FAILURE = 3		//!< Logging an error event
};


#define LOG_NO_OBJECTID ((unsigned long)~0)		//!< There is no object ID

/*! \brief This function returns the string representation of a log type.
 *  \param log_type [IN] The log type.
 *  \return The string representation of the log type.
 */
char * ASN1_logs_get_TypeString(LOG_MESSAGE_TYPE log_type);	

/*! \brief This function returns the string representation of a log status.
 *  \param log_status [IN] The log status.
 *  \return The string representation of the log status.
 */
char * ASN1_logs_get_StatusString(LOG_MESSAGE_STATUS log_status);


#define LOG_ENTRY_sign(x,pkey) \
	ASN1_sign((int (*)())i2d_LOG_ENTRY_BODY, x->sig->sig_alg, NULL, x->sig->signature, (char *)x->body, pkey, EVP_sha1())

#define LOG_ENTRY_verify(x,pkey) \
	ASN1_verify((int (*)())i2d_LOG_ENTRY_BODY,x->sig->sig_alg, x->sig->signature,(char *)x->body, pkey)


#endif
