/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/


#ifndef ASN1_REQ_H
#define ASN1_REQ_H

#include <openssl/x509.h>
#include <openssl/asn1.h>
#include <openssl/asn1t.h>
#include <openssl/x509v3.h>
#include <openssl/rand.h>
#include <openssl/evp.h>
#include <openssl/err.h>

#include "my_asn1.h"


#include "ASN1_logs.h"
#include "ASN1_entity.h"
#include "ASN1_user.h"
#include "ASN1_conf.h"
#include "ASN1_cert.h"



/*!
	This structure represents a login request.
*/
typedef struct st_ADMIN_REQ_LOGIN
{
	ASN1_UTF8STRING * entity;		//!< The entity name to log into
	ASN1_UTF8STRING * username;		//!< The optional username
	ASN1_UTF8STRING * password;		//!< The optional password
}ADMIN_REQ_LOGIN;
DECLARE_ASN1_FUNCTIONS(ADMIN_REQ_LOGIN)


/*!
	This structure represents an entity email configuration modification.
*/
typedef struct st_SET_ENTITY_MAIL_CONF
{
	X509 * entity_cert;			//!< The certificate of the entity to modify the conf from
	EMAIL_CONF * mail_conf;		//!< The new email conf
}SET_ENTITY_MAIL_CONF;
DECLARE_ASN1_FUNCTIONS(SET_ENTITY_MAIL_CONF)

/*!
	This structure represents an entity ACLs configuration modification.
*/
typedef struct st_SET_ENTITY_ACL
{
	X509 * entity_cert;						//!< The certificate of the entity to modify the conf from
	STACK_OF(ACL_ENTRY) * acl_entries;		//!< The ACLs
}SET_ENTITY_ACL;
DECLARE_ASN1_FUNCTIONS(SET_ENTITY_ACL)


/*!
	This structure represents an entity configuration modification.
*/
typedef struct st_SET_ENTITY_CONF
{
	X509 * entity_cert;						//!< The certificate of the entity to modify the conf from
	ENTITY_CONF_BODY * conf;				//!< The conf
}SET_ENTITY_CONF;
DECLARE_ASN1_FUNCTIONS(SET_ENTITY_CONF)


/*!
	This structure represents an entity Audits configuration modification.
*/
typedef struct st_SET_ENTITY_AUDITS
{
	X509 * entity_cert;			//!< The certificate of the entity to modify the conf from
	ENTITY_AUDIT * audits;		//!< The audits
}SET_ENTITY_AUDITS;
DECLARE_ASN1_FUNCTIONS(SET_ENTITY_AUDITS)


/*!
	This structure represents an email.
*/
typedef struct st_MAIL_INFO
{
	ASN1_UTF8STRING * Subject;		//!< The subject
	ASN1_UTF8STRING * Body;			//!< The body
	ASN1_INTEGER * SignMail;		//!< Should the email be signed by the entity
	ASN1_BIT_STRING * Attach;		//!< An attachement
	ASN1_UTF8STRING * AttachName;	//!< The file name of the attachement
	ASN1_UTF8STRING * AttachType;	//!< The MIME type of the attachement
}MAIL_INFO;
DECLARE_ASN1_FUNCTIONS(MAIL_INFO)
DECLARE_ASN1_CPY_FUNCTIONS(MAIL_INFO)


/*!
	This structure represents an object enumeration.
*/
typedef struct st_ENUM_OBJECTS
{
	ASN1_INTEGER * state;		//!< An optional state
	ASN1_INTEGER * index;		//!< The index
	ASN1_INTEGER * num;			//!< The maximum number of objects to return
}ENUM_OBJECTS;
DECLARE_ASN1_FUNCTIONS(ENUM_OBJECTS)


/*!
	[CA] This structure represents a certificate signature
*/
typedef struct st_SIGN_CSR
{
	ASN1_INTEGER * days;		//!< The validity in days for the certificate
	X509_REQ * request;			//!< The certificat request
}SIGN_CSR;
DECLARE_ASN1_FUNCTIONS(SIGN_CSR)


/*!
	This structure represents a PKI user creation request.
*/
typedef struct st_CREATE_PKI_USER
{
	ASN1_INTEGER * pkiadmin;	//!< Is the new user a PKI admin ?
	X509_REQ * csr;				//!< The certificate request
}CREATE_PKI_USER;
DECLARE_ASN1_FUNCTIONS(CREATE_PKI_USER)


/*!
	This structure represents a group renaming request.
*/
typedef struct st_RENAME_GROUP
{
	ASN1_INTEGER * id;			//!< The ID of the group to rename
	ASN1_UTF8STRING * name;		//!< The new name
}RENAME_GROUP;
DECLARE_ASN1_FUNCTIONS(RENAME_GROUP)



/*!
	[RA] This structure represents a certificat request for a PKCS#12 generation.
	
*/
typedef struct st_REQUEST_CERT_BODY_P12
{
	ASN1_INTEGER * keylen;			//!< The length of the key to generate
	ASN1_UTF8STRING * passwd; 		//!< The PKCS#12 password
}REQUEST_CERT_BODY_P12;
DECLARE_ASN1_FUNCTIONS(REQUEST_CERT_BODY_P12)

#define REQUEST_CERT_BODY_PKCS10	0	//!< Requesting a certificate from a PKCS#10
#define REQUEST_CERT_BODY_PKCS12	1	//!< Requesting a certificate for a PKCS#12

/*!
	[RA] This structure represents the body of a certificat request.
*/
typedef struct st_REQUEST_CERT_BODY
{
	int type;
	union
	{
		X509_REQ * pkcs10;							//!< The PKCS#10 certificate request
		REQUEST_CERT_BODY_P12 * p12_privdatas;		//!< The info to generate the PKCS#12
		
	}d;
}REQUEST_CERT_BODY;
DECLARE_ASN1_FUNCTIONS(REQUEST_CERT_BODY)

#define REQUEST_CERT_FLAG_SEND_ADMIN_NOTIFICATION		1	//!< This flag is used to request that the request creator admin should be notified when the RA gets the response back from the CA
#define REQUEST_CERT_FLAG_SEND_CERT_TO_USER				2	//!< This flag is used to request that the certificate/PKCS#12 should be sent to the end-user
#define REQUEST_CERT_FLAG_SEND_P12_PASSWD_TO_USER		4	//!< This flag is used to request that the RA sends the PKCS#12 password to the end-user
#define REQUEST_CERT_FLAG_RA_GENERATE_P12_PASSWD		8	//!< This flag is used to request that the RA generates the PKCS#12 password

/*!
	[RA] This structure represents a certificat request.
*/
typedef struct st_REQUEST_CERT
{
	ASN1_INTEGER * profile_id;			//!< The owner profile
	ASN1_UTF8STRING * ca_name;			//!< The name of the CA
	REQUEST_CERT_BODY * request;		//!< The request body
	ASN1_INTEGER * validity;			//!< The number of days the certificate should be valid for
	ASN1_INTEGER * type;				//!< The certificate type, see NEWPKI_PROFILE_CERT_TYPE_PKCS10
	ASN1_BIT_STRING * flags;			//!< The flags, see REQUEST_CERT_FLAG_SEND_ADMIN_NOTIFICATION
}REQUEST_CERT;
DECLARE_ASN1_FUNCTIONS(REQUEST_CERT)


#define ADMIN_REQ_TYPE_LOGIN					0	//!< The request is a login
#define ADMIN_REQ_TYPE_CREATE_ENTITY			1	//!< The request is to create an entity
#define ADMIN_REQ_TYPE_DELETE_ENTITY			2	//!< The request is to delete an entity
#define ADMIN_REQ_TYPE_ENUM_ENTITIES			3	//!< The request is to list the entities
#define ADMIN_REQ_TYPE_CREATE_USER				4	//!< The request is to create a server user
#define ADMIN_REQ_TYPE_ENUM_USERS				5	//!< The request is to list the users
#define ADMIN_REQ_TYPE_UPDATE_USER				6	//!< The request is to update a server user
#define ADMIN_REQ_TYPE_ENUM_LOGS				7	//!< The request is to enumerate the logs
#define ADMIN_REQ_TYPE_CHG_USER_PASSWD			8	//!< The request is for changing a server user's password
#define ADMIN_REQ_TYPE_CHG_PASSWD				9	//!< The request is for a server user to change its password
#define ADMIN_REQ_TYPE_SIGN_ENTITY				10	//!< The request is to sign an entity
#define ADMIN_REQ_TYPE_CREATE_PKI_USER			11	//!< The request is to create a PKI user
#define ADMIN_REQ_TYPE_REVOKE_USER				12	//!< The request is for revoking a PKI user
#define ADMIN_REQ_TYPE_REVOKE_ENTITY			13	//!< The request is revoking an entity
#define ADMIN_REQ_TYPE_PUSH_CONFIG				14	//!< The request is for pushing the PKI Conf on a repository
#define ADMIN_REQ_TYPE_SET_OFFLINE				15	//!< The request is setting the PKI Entity offline
#define ADMIN_REQ_TYPE_GET_OFFLINE				16	//!< The request is getting the PKI Offline Status
#define ADMIN_REQ_TYPE_GET_MY_ACL				17	//!< The request is user to get its ACL
#define ADMIN_REQ_TYPE_GET_CONFIG				18	//!< The request is for exporting the PKI Conf
#define ADMIN_REQ_TYPE_GET_MY_CONFIG			19	//!< The request is for an entity to get its conf from a repository
#define ADMIN_REQ_TYPE_INIT_ENTITY				20	//!< The request is intializing an entity
#define ADMIN_REQ_TYPE_GET_ENTITY_CONF			21	//!< The request is for an entity GUI to get the entity Conf
#define ADMIN_REQ_TYPE_GET_REPOSITORIES			22	//!< The request is to get the list of known repositories
#define ADMIN_REQ_TYPE_SET_REPOSITORIES			23	//!< The request is to modify the repositories information (host:port)
#define ADMIN_REQ_TYPE_GET_LOGS_TYPE			24	//!< The request is getting the list of known logs type on a server entity
#define ADMIN_REQ_TYPE_LOCK_RESOURCE			25	//!< The request is for locking a PKI resource
#define ADMIN_REQ_TYPE_UNLOCK_RESOURCE			26	//!< The request is for unlocking a resource
#define ADMIN_REQ_TYPE_GET_ENTITY_ACL			27	//!< The request is to get the ACLs of an entity
#define ADMIN_REQ_TYPE_GET_ENTITY_MAIL_CONF		28	//!< The request is to get the Email Conf of an entity
#define ADMIN_REQ_TYPE_SET_ENTITY_ACL			29	//!< The request is to modify the ACLs of an entity
#define ADMIN_REQ_TYPE_SET_ENTITY_MAIL_CONF		30	//!< The request is to modify the Email Conf of an entity
#define ADMIN_REQ_TYPE_SET_ENTITY_CONF			31	//!< The request is to modify the Conf of an entity
#define ADMIN_REQ_TYPE_SEND_ADMIN_MAIL			32	//!< The request is for sending a administrative email
#define ADMIN_REQ_TYPE_GET_ENTITY_AUDITS		33	//!< The request is to get the Audits of an entity
#define ADMIN_REQ_TYPE_SET_ENTITY_AUDITS		34	//!< The request is to modify the Audits of an entity
#define ADMIN_REQ_TYPE_GET_ENTITY_LOGS_TYPE		35	//!< The request is to get the list of known logs type on a entity
#define ADMIN_REQ_TYPE_GET_ENTITIES_LINKS		36	//!< The request is to get the entities links
#define ADMIN_REQ_TYPE_SET_ENTITIES_LINKS		37	//!< The request is to modify the entities links
#define ADMIN_REQ_TYPE_PUSH_NEWPKI_OBJECTS		38	//!< The request is for pushing a newpki object between two repositories
#define ADMIN_REQ_TYPE_CREATE_ROOT_CA			39	//!< The request is to create a ROOT CA
#define ADMIN_REQ_TYPE_GET_CA_STATUS			40	//!< The request is to get the status of a CA
#define ADMIN_REQ_TYPE_GET_MY_REQS				41	//!< The request is to get the new inter-entities requests that were sent to an entity
#define ADMIN_REQ_TYPE_ENUM_CERTS				42	//!< The request is for enumerating the CA certs
#define ADMIN_REQ_TYPE_ENUM_CRLS				43	//!< The request is for enumerating the CA crls
#define ADMIN_REQ_TYPE_SIGN_CSR					44	//!< The request is to sign a CSR on the CA
#define ADMIN_REQ_TYPE_CREATE_CHILD_CA			45	//!< The request is to create a Child CA
#define ADMIN_REQ_TYPE_IMPORT_CHILD_CA_CERT		46	//!< The request is for importing the Child CA certificate
#define ADMIN_REQ_TYPE_REVOKE_CERT				47	//!< The request is for revoking a cert in the CA
#define ADMIN_REQ_TYPE_GET_GROUPS				48	//!< The request is to get the groups
#define ADMIN_REQ_TYPE_SET_GROUPS				49	//!< The request is to set the groups
#define ADMIN_REQ_TYPE_ADD_GROUP				50	//!< The request is to add a group
#define ADMIN_REQ_TYPE_DEL_GROUP				51	//!< The request is to delete a group
#define ADMIN_REQ_TYPE_REN_GROUP				52	//!< The request is to rename a group
#define ADMIN_REQ_TYPE_CHECK_LOGS				53	//!< The request is to verify the integrity of the logs
#define ADMIN_REQ_TYPE_IMPORT_PROFILE			54	//!< The request is to import a profile
#define ADMIN_REQ_TYPE_GET_LOCAL_CONF			55	//!< The request is to get the complete conf of an entity
#define ADMIN_REQ_TYPE_PUSH_REQ					56	//!< The request is to send a inter-entities request to a repository
#define ADMIN_REQ_TYPE_PUSH_RESP				57	//!< The request is to send a inter-entities response to a repository
#define ADMIN_REQ_TYPE_GET_MY_RESPS				58	//!< The request is for an entity to get the inter-entities responses
#define ADMIN_REQ_TYPE_ENUM_PROFILES			59	//!< The request is for enumerating the profiles
#define ADMIN_REQ_TYPE_REQUEST_CERT				60	//!< The request is for a RA Operator to request a certificate for a profile
#define ADMIN_REQ_TYPE_GENERATE_CRL				61	//!< The request is for a CA Operator to request the generation of a CRL
#define ADMIN_REQ_TYPE_SYNCH_OBJECTS			62	//!< The request is for two repositories to synchronize their objects
#define ADMIN_REQ_TYPE_GET_CA_CERT				63	//!< The request is to get the certificate of a CA
#define ADMIN_REQ_TYPE_DELETE_PKCS12			64	//!< The request is to delete the PKCS#12
#define ADMIN_REQ_TYPE_GET_LOGS_COUNT			65	//!< The request is to get the number of logs
#define ADMIN_REQ_TYPE_SEARCH_LDAP				66	//!< The request is to search the LDAP





/*!
	This structure represents the body of a request.
*/
typedef struct st_ADMIN_REQUEST_BODY
{
	int type;
	union
	{
		ADMIN_REQ_LOGIN * login;					//!< Login request
		ENTITY_CREATION_REQ * cre_entity;			//!< Entity creation request
		ASN1_UTF8STRING * del_entity;				//!< Entity suppresion request
		USER_INFO * cre_user;						//!< Server user creation request
		USER_ENTRY * upd_user;						//!< Server user update request
		ADMIN_REQ_ENUM_LOGS * enum_logs;			//!< Logs enumeration request
		CHANGE_USER_PASSWD * chg_user_pwd;			//!< Server user password changing request
		CHANGE_PASSWD * chg_pwd;					//!< Server user password changing request
		ENTITY_SIGNATURE_REQ * sign_entity;			//!< Entity signature request
		CREATE_PKI_USER * create_pki_user;			//!< PKI User creation request
		EXPORTED_PKI_CONF * conf;					//!< PKI conf synchronization request
		ASN1_INTEGER * serial;						//!< A serial related request
		ASN1_INTEGER * offline;						//!< Setting PKI offline status request
		ENTITY_INIT_REQ * init_entity;				//!< Entity initialization request
		X509 * entity_cert;							//!< Entity certificate related request
		REPS_INFO * repositories;			//!< Setting the repositories info request
		ASN1_UTF8STRING * resource_name;			//!< A resource name related request
		SET_ENTITY_ACL * entity_acl;				//!< Setting the ACLs request
		SET_ENTITY_MAIL_CONF * entity_mail_conf;	//!< Setting the email conf request
		SET_ENTITY_CONF * entity_conf;				//!< Setting the entity Conf request
		SET_ENTITY_AUDITS * entity_audits;			//!< Setting the entity audits request
		MAIL_INFO * admin_mail;						//!< Sending a admin mail request
		ASN1_INTEGER * entity_type;					//!< Getting the logs type request
		ENTITIES_LINKS * entities_links;			//!< Setting the entities links requets
		WAITING_NEWPKI_OBJECT * waiting_obj;		//!< Synchronization the repositories objects request
		REQ_CREATE_ROOT_CA * create_root_ca;		//!< Creating a ROOT CA request
		TRANSACTION_IDS * transactions_ids;			//!< Transaction IDs related request
		ENUM_OBJECTS * enum_objects;				//!< Objects enumeration request
		SIGN_CSR * sign_csr;						//!< CSR signature request
		REQ_CREATE_CHILD_CA * create_child_ca;		//!< Creating a CHILD CA request
		USERS_GROUPS * groups;						//!< Setting the groups request
		ASN1_UTF8STRING * group_name;				//!< Group name related request
		ASN1_INTEGER * group_id;					//!< Group ID related request
		RENAME_GROUP * rename_group;				//!< Renaming a group request
		NEWPKI_PROFILE * profile;					//!< Profile importation request
		CRYPTED_NEWPKI_REQUEST * new_req;			//!< Inter-entities request publication request
		STORED_NEWPKI_RESPONSE * new_resp;			//!< Inter-entities response publication request
		REQUEST_CERT * cert_request;				//!< RA CSR request
		OBJECTS_LIST_REQUEST * known_objects;		//!< Inter repositories synchronization request
		PKCS7 * p7b;								//!< Child CA certificate importation request
		ASN1_UTF8STRING * ldap_search;				//!< A LDAP search request
		ASN1_NULL * other;							//!< Any other request
	}d;
}ADMIN_REQUEST_BODY;
DECLARE_ASN1_FUNCTIONS(ADMIN_REQUEST_BODY)

/*!
	This structure represents a request.
*/
typedef struct st_ADMIN_REQUEST
{
	ADMIN_REQUEST_BODY * body;			//!< The body
}ADMIN_REQUEST;
DECLARE_ASN1_FUNCTIONS(ADMIN_REQUEST)




#define d2i_ADMIN_REQUEST_bio(bp,msg) (ADMIN_REQUEST *)ASN1_d2i_bio((char *(*)())ADMIN_REQUEST_new, \
		(char *(*)())d2i_ADMIN_REQUEST, (bp),(unsigned char **)(msg))

#define i2d_ADMIN_REQUEST_bio(bp,msg) ASN1_i2d_bio((int (*)(void))i2d_ADMIN_REQUEST,bp,(unsigned char *)msg)



#endif
