/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/


#ifndef ASN1_RESP_H
#define ASN1_RESP_H

#include <openssl/x509.h>
#include <openssl/asn1.h>
#include <openssl/asn1t.h>
#include <openssl/x509v3.h>
#include <openssl/rand.h>
#include <openssl/evp.h>
#include <openssl/err.h>
#include <openssl/pkcs7.h>

#include "my_asn1.h"


#include "ASN1_logs.h"
#include "ASN1_err.h"
#include "ASN1_user.h"
#include "ASN1_entity.h"
#include "ASN1_conf.h"
#include "ASN1_cert.h"

/*!
	This structure represents the certificate generated by a CA.
*/
typedef struct st_INTERNAL_CA_CERT
{
	ASN1_INTEGER * ca_type;			//!< The CA type
	ASN1_INTEGER * serial;			//!< The serial
	ASN1_INTEGER * state;			//!< The certificate state
	ASN1_UTF8STRING * dn;			//!< The certificate DN
	ASN1_INTEGER * begin_date;		//!< The begin date
	ASN1_INTEGER * end_date;		//!< The expiration date
	ASN1_INTEGER * rev_date;		//!< The optional revocation date
	X509 * cert;					//!< The certificate
}INTERNAL_CA_CERT;
DECLARE_ASN1_FUNCTIONS(INTERNAL_CA_CERT)

/*!
	This structure represents a LDAP attribute.
*/
typedef struct st_LDAP_OBJECT
{
	ASN1_UTF8STRING * name;			//!< The attribute name
	ASN1_UTF8STRING * value;		//!< The attribute value
}LDAP_OBJECT;
DECLARE_ASN1_FUNCTIONS(LDAP_OBJECT)


/*!
	This structure represents a LDAP search result.
*/
typedef struct st_LDAP_RESULT
{
	ASN1_UTF8STRING * rdn;				//!< The result RDN
	ASN1_UTF8STRING * uid;				//!< The result UID
	STACK_OF(LDAP_OBJECT) * objects;	//!< The LDAP attributes
}LDAP_RESULT;
DECLARE_ASN1_FUNCTIONS(LDAP_RESULT)



#define STATUS_CA_OK			1		//!< The status of the CA is OK
#define STATUS_CA_AWAITING_CERT	2		//!< The status of the CA is waiting for its certificate
#define STATUS_CA_NO_CERT		3		//!< The status of the CA is not initialized



#define ADMIN_RESP_TYPE_ERRORS				0	//!< The response is an error
#define ADMIN_RESP_TYPE_NONE				1	//!< There is no result
#define ADMIN_RESP_TYPE_USER_TYPE			2	//!< The response is a user type
#define ADMIN_RESP_TYPE_ENUM_USERS			3	//!< The response is a list of Server users
#define ADMIN_RESP_TYPE_ENUM_ENTITIES		4	//!< The response is a list of entities
#define ADMIN_RESP_TYPE_ENUM_LOGS			5	//!< The response is a list of logs
#define ADMIN_RESP_TYPE_CREATE_ENTITY		6	//!< The response is the public key(s) of the created entity
#define ADMIN_RESP_TYPE_SIGN_ENTITY			7	//!< The response is the certificate(s) of the sign entity
#define ADMIN_RESP_TYPE_ENUM_USERS_CERT		8	//!< The response is a list of PKI users
#define ADMIN_RESP_TYPE_CREATE_PKI_USER		9	//!< The response is the certificate of the created user
#define ADMIN_RESP_TYPE_STATUS				10	//!< The response is the status (online/offline) de l'entit PKI
#define ADMIN_RESP_TYPE_ACLS				11	//!< The response is a list of ACLs types
#define ADMIN_RESP_TYPE_MY_ACL				12	//!< The response is the ACLs of the logged in user
#define ADMIN_RESP_TYPE_CONF				13	//!< The response is the global PKI conf
#define ADMIN_RESP_TYPE_MY_CONF				14	//!< The response is an entity conf (crypted)
#define ADMIN_RESP_TYPE_GET_REPOSITORIES	15	//!< The response is the list of repositories
#define ADMIN_RESP_TYPE_LOGS_TYPE			16	//!< The response is a supported loggs type
#define ADMIN_RESP_TYPE_ENTITY_CONF			17	//!< The response is a the conf of an entity (not crypted)
#define ADMIN_RESP_TYPE_ENTITY_ACL			18	//!< The response is a list of ACLs
#define ADMIN_RESP_TYPE_MAIL_CONF			19	//!< The response is the email conf
#define ADMIN_RESP_TYPE_AUDITS				20	//!< The response is a list of audits
#define ADMIN_RESP_TYPE_ENTITIES_LINKS		21	//!< The response is the list of links
#define ADMIN_RESP_TYPE_REQS				22	//!< The response is a list of requests
#define ADMIN_RESP_TYPE_CERTS				23	//!< The response is a list of certificates
#define ADMIN_RESP_TYPE_P7B					24	//!< The response is a P7B
#define ADMIN_RESP_TYPE_CSR					25	//!< The response is a CSR
#define ADMIN_RESP_TYPE_CRLS				26	//!< The response is a list of CRLs
#define ADMIN_RESP_TYPE_GROUPS				27	//!< The response is a list of groups
#define ADMIN_RESP_TYPE_PROFILE_ID			28	//!< The response is a profile ID
#define ADMIN_RESP_TYPE_TRANSACTION_ID		29	//!< The response is a transaction ID
#define ADMIN_RESP_TYPE_RESPS				30	//!< The response is a list of responses
#define ADMIN_RESP_TYPE_PROFILES			31	//!< The response is a list of profiles
#define ADMIN_RESP_TYPE_LOCAL_ENTITY_CONF	32	//!< The response is the conf of an entity
#define ADMIN_RESP_TYPE_KNOWN_OBJECTS		33	//!< The response is a list of known objects
#define ADMIN_RESP_TYPE_REQ_ID				34	//!< The response is a request ID
#define ADMIN_RESP_TYPE_CERT				35	//!< The response is a certificate
#define ADMIN_RESP_TYPE_LOGS_COUNT			36	//!< The response is the number of logs an entity has
#define ADMIN_RESP_TYPE_LDAP_RESULTS		37	//!< The response is a LDAP search result


/*!
	This structure represents the body of a response.
*/
typedef struct st_ADMIN_RESPONSE_BODY
{
	int type;
	union
	{
		STACK_OF(ERROR_ENTRY) * errors;					//!< A list of errors
		ASN1_INTEGER * usertype;						//!< The user type
		STACK_OF(ENTITY_ENTRY_INFO) * entities;			//!< A list of entities
		STACK_OF(USER_ENTRY) * users;					//!< A list of server users
		STACK_OF(LOG_ENTRY) * logs;						//!< A list of logs
		ENTITY_CREATION_RESP * cre_entity;				//!< A creation response
		ENTITY_SIGNATURE_RESP * sign_entity;			//!< A signature response
		STACK_OF(USER_ENTRY_CERT) * users_cert;			//!< A list of PKI users
		STACK_OF(X509) * create_pki_user;				//!< A list of certificates
		ASN1_INTEGER * status;							//!< The PKI entity status
		X509_ACL * acls_conf;							//!< The ACLs
		ASN1_BIT_STRING * my_acl;						//!< A user's ACLs
		EXPORTED_PKI_CONF * conf;						//!< The global PKI conf
		ENTITY_CONF_CRYPTED * my_conf;					//!< The crypted entity conf
		REPS_INFO * repositories;				//!< The list of repositories
		STACK_OF(ASN1_INTEGER) * logs_type;				//!< A list of entity logs type
		ENTITY_CONF_BODY * entity_conf;					//!< An entity conf body
		EMAIL_CONF * mail_conf;							//!< An entity email conf
		ENTITY_AUDIT * audits;							//!< An entity audits conf
		ENTITIES_LINKS * entities_links;				//!< The PKI links
		STORED_NEWPKI_REQUESTS * object_reqs;			//!< The inter-entities requests
		STORED_NEWPKI_RESPONSES * object_resps;			//!< The inter-entities responses
		STACK_OF(INTERNAL_CA_CERT) * certs;				//!< A list of internal certificates
		PKCS7 * p7b;									//!< A P7B
		X509_REQ * csr;									//!< A CSR
		STACK_OF(X509_CRL) * crls;						//!< A list of CRLs
		USERS_GROUPS * groups;							//!< A list of groups
		ASN1_INTEGER * id;								//!< An ID
		ASN1_OCTET_STRING * transactionID;				//!< A transaction ID
		STACK_OF(NEWPKI_PROFILE_DATAS) * profiles;		//!< A list of profiles
		ENTITY_CONF * local_entity_conf;				//!< An entity conf (not crypted)
		OBJECTS_LIST_RESPONSE * known_objects;					//!< The list of repositories' known objects
		X509 * cert;									//!< A certificate
		STACK_OF(LDAP_RESULT) * ldap_results;			//!< A list of LDAP results
		ASN1_NULL * other;								//!< Any other type of response
	}d;
}ADMIN_RESPONSE_BODY;
DECLARE_ASN1_FUNCTIONS(ADMIN_RESPONSE_BODY)


/*!
	This structure represents a response.
*/
typedef struct st_ADMIN_RESPONSE
{
	ADMIN_RESPONSE_BODY * body;				//!< The response body
	X509 * signing_cert;					//!< The signing certificate
	X509_ALGOR * sig_algo;					//!< The signature algo
	ASN1_BIT_STRING * signature;			//!< The signature
}ADMIN_RESPONSE;
DECLARE_ASN1_FUNCTIONS(ADMIN_RESPONSE)


#define d2i_ADMIN_RESPONSE_bio(bp,msg) (ADMIN_RESPONSE *)ASN1_d2i_bio((char *(*)())ADMIN_RESPONSE_new, \
		(char *(*)())d2i_ADMIN_RESPONSE, (bp),(unsigned char **)(msg))

#define i2d_ADMIN_RESPONSE_bio(bp,msg) ASN1_i2d_bio((int (*)(void))i2d_ADMIN_RESPONSE,bp,(unsigned char *)msg)



#endif
