/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "my_asn1.h"

ASN1_SEQUENCE(ASN1_ENCRYPT_SIGN) = {
	ASN1_SIMPLE(ASN1_ENCRYPT_SIGN, sym_algo, ASN1_OBJECT),
	ASN1_SIMPLE(ASN1_ENCRYPT_SIGN, crypteddatas, ASN1_OCTET_STRING),
	ASN1_SIMPLE(ASN1_ENCRYPT_SIGN, sessionkey, ASN1_OCTET_STRING),
	ASN1_SIMPLE(ASN1_ENCRYPT_SIGN, sig, X509_SIG),
}ASN1_SEQUENCE_END(ASN1_ENCRYPT_SIGN)
IMPLEMENT_ASN1_FUNCTIONS(ASN1_ENCRYPT_SIGN)
IMPLEMENT_PEM_CONVERT(ASN1_ENCRYPT_SIGN, "NEWPKI SIGNED ENCRYPTED DATAS")
IMPLEMENT_ASN1_CPY_FUNCTIONS(ASN1_ENCRYPT_SIGN)

#define CIPHER_IV (unsigned char*)"NEWPKI_IV"

IMPLEMENT_ASN1_CPY_FUNCTIONS(X509)
IMPLEMENT_ASN1_CPY_FUNCTIONS(X509_REQ)
IMPLEMENT_ASN1_CPY_FUNCTIONS(X509_CRL)
IMPLEMENT_ASN1_CPY_FUNCTIONS(ASN1_BIT_STRING)
IMPLEMENT_ASN1_CPY_FUNCTIONS(X509_PUBKEY)
IMPLEMENT_ASN1_CPY_FUNCTIONS(ASN1_OCTET_STRING)
IMPLEMENT_ASN1_CPY_FUNCTIONS(PKCS12)
IMPLEMENT_PEM_CONVERT(X509_PUBKEY, "X509 PUBLIC KEY")
IMPLEMENT_ASN1_STACK_CPY_FUNCTIONS(ASN1_UTF8STRING)



int ASN1_ENCRYPT_SIGN_i2d(ASN1_ENCRYPT_SIGN * cryptinfo, int (*i2d)(char *a, unsigned char **out), char * a, const EVP_PKEY * sig_pkey, const EVP_PKEY * crypt_pkey, const EVP_MD * sig_md, const EVP_CIPHER * crypt_cypher)
{
	unsigned char rnd_datas[50];
	unsigned char * c_key;
	int c_key_len;
	unsigned char key[EVP_MAX_KEY_LENGTH];
	int keylen;
	EVP_CIPHER_CTX ctx;
	unsigned char * der_body;
	unsigned char * p;
	int der_body_len;
	unsigned char * crypted_body;
	int crypted_body_len;
	unsigned char iv[EVP_MAX_IV_LENGTH+1];
	strcpy((char*)iv, (char*)CIPHER_IV);

	if(!cryptinfo || !i2d || !a || !sig_pkey || !crypt_pkey || !sig_md || !crypt_cypher)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_BAD_PARAM);
		return 0;
	}

	ASN1_OBJECT_free(cryptinfo->sym_algo);
	cryptinfo->sym_algo = OBJ_nid2obj(EVP_CIPHER_nid(crypt_cypher));
	if(!cryptinfo->sym_algo)
	{
		ASN1err(ASN1_F_ASN1_SIGN,ASN1_R_UNKNOWN_OBJECT_TYPE);
		return 0;
	}

	//generate session key, that will encrypt the body
	RAND_bytes(rnd_datas, sizeof(rnd_datas));
	if( (keylen = EVP_BytesToKey(crypt_cypher, sig_md, NULL, rnd_datas, sizeof(rnd_datas), 1, key, iv)) <= 0 )
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		return 0;
	}
	//Not needed anymore
	OPENSSL_cleanse(rnd_datas, sizeof(rnd_datas));

	//Encrypt session key with public key
	c_key = (unsigned char*)malloc(EVP_PKEY_size((EVP_PKEY*)crypt_pkey));
	if(!c_key)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		return 0;
	}
	if( (c_key_len = EVP_PKEY_encrypt(c_key, key, keylen, (EVP_PKEY*)crypt_pkey)) <= 0 )
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		free(c_key);
		return 0;
	}
	if(ASN1_OCTET_STRING_set(cryptinfo->sessionkey, c_key, c_key_len) <= 0)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		free(c_key);
		return 0;
	}
	OPENSSL_cleanse(c_key, c_key_len);
	free(c_key);





	//get the body in DER
	der_body_len = i2d(a, NULL);
	if(der_body_len <= 0)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		return 0;
	}

	der_body = (unsigned char*)malloc(der_body_len);
	if(!der_body)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		return 0;
	}
	p = der_body;

	der_body_len = i2d(a, &p);
	if(der_body_len <= 0)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		free(der_body);
		return 0;
	}

	//Allocate decryption buffer
	crypted_body = (unsigned char *)malloc(keylen + der_body_len + 10);
	if(!crypted_body)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_MALLOC);
		free(der_body);
		return 0;
	}


	//Do the encryption
	EVP_CIPHER_CTX_init(&ctx);
	if(EVP_EncryptInit(&ctx, crypt_cypher, key, CIPHER_IV) <= 0)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		free(der_body);
		free(crypted_body);
		EVP_CIPHER_CTX_cleanup(&ctx);
		return 0;
	}
	if(EVP_EncryptUpdate(&ctx, crypted_body, &crypted_body_len, der_body, der_body_len) <= 0)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		free(der_body);
		free(crypted_body);
		EVP_CIPHER_CTX_cleanup(&ctx);
		return 0;
	}
	if(EVP_EncryptFinal(&ctx, &crypted_body[crypted_body_len], &keylen) <= 0)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		free(der_body);
		free(crypted_body);
		EVP_CIPHER_CTX_cleanup(&ctx);
		return 0;
	}
	crypted_body_len += keylen;
	EVP_CIPHER_CTX_cleanup(&ctx);
	free(der_body);
	OPENSSL_cleanse(key, keylen);
	
	if(ASN1_OCTET_STRING_set(cryptinfo->crypteddatas, crypted_body, crypted_body_len) <= 0)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		free(crypted_body);
		return 0;
	}
	free(crypted_body);

	//Do signature
	if(ASN1_sign((int (*)())i2d_ASN1_OCTET_STRING, cryptinfo->sig->algor, NULL, cryptinfo->sig->digest, (char*)cryptinfo->crypteddatas, (EVP_PKEY*)sig_pkey, sig_md) <= 0)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		return 0;
	}	
	return 1;
}

int ASN1_ENCRYPT_SIGN_d2i(const ASN1_ENCRYPT_SIGN * cryptinfo, int (*i2d)(char *a, unsigned char **out), char* (*d2i)(char **a, unsigned char **in, long len), char ** a, const EVP_PKEY * sig_pkey, const EVP_PKEY * crypt_pkey)
{
	const EVP_MD * sig_md;
	const EVP_CIPHER * crypt_cypher;
	unsigned char * c_key;
	unsigned char key[EVP_MAX_KEY_LENGTH];
	int keylen;
	EVP_CIPHER_CTX ctx;
	unsigned char * der_body;
	unsigned char * p;
	int der_body_len;
	unsigned char iv[EVP_MAX_IV_LENGTH+1];
	strcpy((char*)iv, (char*)CIPHER_IV);
	
	if(!cryptinfo || !i2d || !d2i || !sig_pkey)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_BAD_PARAM);
		return 0;
	}

	sig_md = EVP_get_digestbyobj(cryptinfo->sig->algor->algorithm);
	if(!sig_md)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_BAD_DATAS);
		return 0;
	}
	crypt_cypher = EVP_get_cipherbyobj(cryptinfo->sym_algo);
	if(!sig_md)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_BAD_DATAS);
		return 0;
	}

	//We can verify the signature
	if (ASN1_verify((int(*)())i2d_ASN1_OCTET_STRING, cryptinfo->sig->algor, cryptinfo->sig->digest, (char*)cryptinfo->crypteddatas, (EVP_PKEY*)sig_pkey) <= 0)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		return 0;
	}

	//We can be asked to only verify the datas
	if(crypt_pkey)
	{
		//We now decrypt the session key
		c_key = (unsigned char*)malloc(EVP_PKEY_size((EVP_PKEY*)crypt_pkey)+30);
		if(!c_key)
		{
			NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
			return 0;
		}
		if( (keylen = EVP_PKEY_decrypt(c_key, cryptinfo->sessionkey->data, cryptinfo->sessionkey->length, (EVP_PKEY*)crypt_pkey)) <= 0 )
		{
			NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
			free(c_key);
			return 0;
		}

		if(keylen > (int)sizeof(key))
		{
			NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_BAD_DATAS);
			free(c_key);
			return 0;
		}
		memcpy(key, c_key, keylen);
		OPENSSL_cleanse(c_key, keylen);
		free(c_key);

		//With the session key we can decrypt the body !

		//Allocate decryption buffer
		der_body = (unsigned char *)malloc(EVP_MAX_KEY_LENGTH + cryptinfo->crypteddatas->length + 50);
		if(!der_body)
		{
			OPENSSL_cleanse(key, keylen);
			NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_MALLOC);
			return 0;
		}


		//Do the decryption
		EVP_CIPHER_CTX_init(&ctx);
		if(EVP_DecryptInit(&ctx, crypt_cypher, key, CIPHER_IV) <= 0)
		{
			NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
			EVP_CIPHER_CTX_cleanup(&ctx);
			free(der_body);
			OPENSSL_cleanse(key, keylen);
			return 0;
		}
		if(EVP_DecryptUpdate(&ctx, der_body, &der_body_len, cryptinfo->crypteddatas->data, cryptinfo->crypteddatas->length) <= 0)
		{
			NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
			EVP_CIPHER_CTX_cleanup(&ctx);
			free(der_body);
			OPENSSL_cleanse(key, keylen);
			return 0;
		}
		if(EVP_DecryptFinal(&ctx, &der_body[der_body_len], &keylen) <= 0)
		{
			NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
			EVP_CIPHER_CTX_cleanup(&ctx);
			free(der_body);
			OPENSSL_cleanse(key, keylen);
			return 0;
		}
		der_body_len += keylen;
		EVP_CIPHER_CTX_cleanup(&ctx);
		OPENSSL_cleanse(key, keylen);
		
		p = der_body;
		*a = d2i(NULL, &p, der_body_len);
		if(!*a)
		{
			NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
			free(der_body);
			return 0;
		}
		free(der_body);
	}

	return 1;
}

void ASN1_CHOICE_VALUE_free(ASN1_VALUE * pval, const ASN1_ITEM *it)
{
	const ASN1_TEMPLATE *tt = NULL;
	ASN1_VALUE **pchval;
	const ASN1_AUX *aux = (const ASN1_AUX *)it->funcs;
	ASN1_aux_cb *asn1_cb;
	int i;

	if(!pval) return;
	if((it->itype != ASN1_ITYPE_CHOICE))
		return;

	if(aux && aux->asn1_cb)
		asn1_cb = aux->asn1_cb;
	else
		asn1_cb = 0;

	
	if(asn1_cb)
	{
		i = asn1_cb(ASN1_OP_FREE_PRE, &pval, it);
		if(i == 2)
			return;
	}
	
	i = asn1_get_choice_selector(&pval, it);

	if(asn1_cb)
		asn1_cb(ASN1_OP_FREE_PRE, &pval, it);

	if((i >= 0) && (i < it->tcount)) 
	{
		tt = it->templates + i;
		pchval = asn1_get_field_ptr(&pval, tt);
		ASN1_template_free(pchval, tt);
	}
	if(asn1_cb) asn1_cb(ASN1_OP_FREE_POST, &pval, it);
}
