/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/

#ifndef MY_ASN1_H
#define MY_ASN1_H

#include <openssl/asn1.h>
#include <openssl/pkcs12.h>
#include <openssl/asn1t.h>
#include <PKI_ERR.h>


#define DECLARE_ASN1_STACK_CPY_FUNCTIONS(stname) STACK_OF(stname) * stname##_stack_cpy(STACK_OF(stname) ** out, const STACK_OF(stname) * in);

#define IMPLEMENT_ASN1_STACK_CPY_FUNCTIONS(stname) \
STACK_OF(stname) * stname##_stack_cpy(STACK_OF(stname) ** out, const STACK_OF(stname) * in) \
{ \
	STACK_OF(stname) * ret; \
	stname * currValue; \
	int i; \
	ret = SKM_sk_new_null(stname); \
	if(!ret) \
	{ \
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_MALLOC); \
		return NULL; \
	} \
	for(i=0; i<SKM_sk_num(stname, in); i++) \
	{ \
		currValue = SKM_sk_value(stname, in, i); \
		if(!currValue) \
		{ \
			NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_UNKNOWN); \
			SKM_sk_pop_free(stname, ret, stname##_free); \
			return NULL; \
		} \
		currValue = (stname*)ASN1_item_dup(ASN1_ITEM_rptr(stname), (void*)currValue); \
		if(!currValue) \
		{ \
			NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC); \
			SKM_sk_pop_free(stname, ret, stname##_free); \
			return NULL; \
		} \
		if(SKM_sk_push(stname, ret, currValue) == -1) \
		{ \
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT); \
			SKM_sk_pop_free(stname, ret, stname##_free); \
			stname##_free(currValue); \
			return NULL; \
		} \
	} \
	if(out) \
	{ \
		if(*out) SKM_sk_pop_free(stname, *out, stname##_free); \
		*out = ret; \
	} \
	return ret; \
}




#define DECLARE_ASN1_CPY_FUNCTIONS(stname) stname * stname##_cpy(stname ** out, const stname * in);

#define IMPLEMENT_ASN1_CPY_FUNCTIONS(stname) \
stname * stname##_cpy(stname ** out, const stname * in) \
{ \
	stname * ret; \
	ret = (stname*)ASN1_item_dup(ASN1_ITEM_rptr(stname), (void*)in); \
	if(out) \
	{ \
		if(*out) stname##_free(*out); \
		*out = ret; \
	} \
	return ret; \
}

#define IMPLEMENT_ASN1_CPY_FUNCTIONS_REF(stname, LOCKID) \
stname * stname##_cpy(stname ** out, const stname * in) \
{ \
	stname * ret; \
	ret = (stname *)in; \
	CRYPTO_add((int*)&((in)->references), 1, LOCKID); \
	if(out) \
	{ \
		if(*out) stname##_free(*out); \
		*out = ret; \
	} \
	return ret; \
}


#define ASN1_STRING_SET(stdest, stsource) \
if(stsource) \
{ \
	if(!*stdest) *stdest = ASN1_STRING_new(); \
	if(*stdest) \
	{ \
		ASN1_STRING_set(*stdest, stsource, strlen(stsource)); \
	} \
}

#define ASN1_INTEGER_SET(stdest, stsource) \
if(!*stdest) *stdest = ASN1_INTEGER_new(); \
if(*stdest) \
{ \
	ASN1_INTEGER_set(*stdest, stsource); \
}


#define DECLARE_PEM_CONVERT(stname) \
DECLARE_PEM_rw(stname, stname) \
char * stname##_To_PEM(const stname * datas); \
int stname##_From_PEM(stname ** datas, const char * pem_datas);


#define IMPLEMENT_PEM_CONVERT(stname, pemstr) \
char * stname##_To_PEM(const stname * datas) \
{ \
	char * pem_datas; \
	BIO * membio = BIO_new(BIO_s_mem()); \
	if(!membio) \
		return NULL; \
	if(!PEM_write_bio_##stname(membio, (stname *)datas)) \
	{ \
		BIO_free_all(membio); \
		return NULL; \
	} \
	pem_datas = (char *)malloc(BIO_number_written(membio) + 1); \
	if(!pem_datas) \
	{ \
		BIO_free_all(membio); \
		return NULL; \
	} \
	BIO_seek(membio, 0); \
	BIO_read(membio, pem_datas, BIO_number_written(membio)); \
	pem_datas[BIO_number_written(membio)]=0; \
	BIO_free_all(membio); \
	return pem_datas; \
} \
int stname##_From_PEM(stname ** datas, const char * pem_datas) \
{ \
	BIO * membio = BIO_new(BIO_s_mem()); \
	if(!membio) \
		return 0; \
	BIO_printf(membio, pem_datas); \
	BIO_seek(membio, 0); \
	if(!PEM_read_bio_##stname(membio, datas, NULL, NULL)) \
	{ \
		BIO_free_all(membio); \
		return 0; \
	} \
	BIO_free_all(membio); \
	return 1; \
} \
IMPLEMENT_PEM_rw(stname, stname, pemstr, stname)

/*!
	This structure represents some signed and crypted datas
*/
typedef struct st_ASN1_ENCRYPT_SIGN
{
	ASN1_OBJECT * sym_algo;				//!< The symetical encryption algo
	ASN1_OCTET_STRING * crypteddatas;	//!< The symetrically crypted datas
	ASN1_OCTET_STRING * sessionkey;		//!< The asymetrically crypted symetrical key
	X509_SIG * sig;						//!< The signature
}ASN1_ENCRYPT_SIGN;
DECLARE_ASN1_FUNCTIONS(ASN1_ENCRYPT_SIGN)
DECLARE_PEM_CONVERT(ASN1_ENCRYPT_SIGN)
DECLARE_ASN1_CPY_FUNCTIONS(ASN1_ENCRYPT_SIGN)

/*! \brief This function does the signature en encipherment.
 *  \param cryptinfo [OUT] The result.
 *  \param i2d [IN] The pointer to the function capable of converting the datas to a buffer.
 *  \param a [IN] The datas.
 *  \param sig_pkey [IN] The signature private key.
 *  \param crypt_pkey [IN] The encipherment public key.
 *  \param sig_md [IN] The hash algorythm.
 *  \param crypt_cypher [IN] The symetrical encipherment hash.
 *  \return 1 on success, 0 on error.
 */
int ASN1_ENCRYPT_SIGN_i2d(ASN1_ENCRYPT_SIGN * cryptinfo, int (*i2d)(char *a, unsigned char **out), char * a, const EVP_PKEY * sig_pkey, const EVP_PKEY * crypt_pkey, const EVP_MD * sig_md, const EVP_CIPHER * crypt_cypher);

/*! \brief This function decrypts datas and verifies the signature.
 *  \param cryptinfo [IN] The signed and crypted datas.
 *  \param i2d [IN] The pointer to the function capable of converting the datas to a buffer.
 *  \param d2i [IN] The pointer to the function capable of converting the buffer to datas.
 *  \param a [OUT] The datas.
 *  \param sig_pkey [IN] The signature verification public key.
 *  \param crypt_pkey [IN] The decipher private key.
 *  \return 1 on success, 0 on error.
 */
int ASN1_ENCRYPT_SIGN_d2i(const ASN1_ENCRYPT_SIGN * cryptinfo, int (*i2d)(char *a, unsigned char **out), char* (*d2i)(char **a, unsigned char **in, long len), char ** a, const EVP_PKEY * sig_pkey, const EVP_PKEY * crypt_pkey);

#define DECLARE_ASN1_ENCRYPT_SIGN(stname) \
int stname##_sign_encrypt(ASN1_ENCRYPT_SIGN * cryptinfo, const stname * a, const EVP_PKEY * sig_pkey, const EVP_PKEY * crypt_pkey, const EVP_MD * sig_md, const EVP_CIPHER * crypt_cypher); \
int stname##_decrypt_verify(const ASN1_ENCRYPT_SIGN * cryptinfo, stname ** a, const EVP_PKEY * sig_pkey, const EVP_PKEY * crypt_pkey);

#define IMPLEMENT_ASN1_ENCRYPT_SIGN(stname) \
int stname##_sign_encrypt(ASN1_ENCRYPT_SIGN * cryptinfo, const stname * a, const EVP_PKEY * sig_pkey, const EVP_PKEY * crypt_pkey, const EVP_MD * sig_md, const EVP_CIPHER * crypt_cypher) \
{ \
	return ASN1_ENCRYPT_SIGN_i2d(cryptinfo, (int (*)(char *a, unsigned char **out))i2d_##stname, (char*)a, sig_pkey, crypt_pkey, sig_md, crypt_cypher); \
} \
int stname##_decrypt_verify(const ASN1_ENCRYPT_SIGN * cryptinfo, stname ** a, const EVP_PKEY * sig_pkey, const EVP_PKEY * crypt_pkey) \
{ \
	int ret; \
	*a = NULL; \
	ret = ASN1_ENCRYPT_SIGN_d2i(cryptinfo, (int (*)(char *a, unsigned char **out))i2d_##stname, (char* (*)(char **a, unsigned char **in, long len))d2i_##stname, (char**)a, sig_pkey, crypt_pkey); \
	if(!ret && *a) \
		stname##_free(*a); \
	return ret; \
} \




#define ASN1_STRING_GET(stname) (char*)((stname && (stname)->data && (stname)->length)?(stname)->data:NULL)
#define ASN1_INTEGER_GET(stname) ((stname)?ASN1_INTEGER_get(stname):0)

void ASN1_CHOICE_VALUE_free(ASN1_VALUE * pval, const ASN1_ITEM *it);

DECLARE_ASN1_CPY_FUNCTIONS(X509)
DECLARE_ASN1_CPY_FUNCTIONS(X509_REQ)
DECLARE_ASN1_CPY_FUNCTIONS(X509_CRL)
DECLARE_ASN1_CPY_FUNCTIONS(ASN1_BIT_STRING)
DECLARE_ASN1_CPY_FUNCTIONS(ASN1_OCTET_STRING)
DECLARE_ASN1_CPY_FUNCTIONS(X509_PUBKEY)
DECLARE_ASN1_CPY_FUNCTIONS(PKCS12)
DECLARE_PEM_CONVERT(X509_PUBKEY)
DECLARE_ASN1_STACK_CPY_FUNCTIONS(ASN1_UTF8STRING)

#endif

