/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/


// HashCorrelation.cpp: implementation of the HashCorrelation class.
//
//////////////////////////////////////////////////////////////////////
#include <NewPKI.h>
#include "HashCorrelation.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
ASN1_SEQUENCE(INTERNAL_HASH) = {
	ASN1_SIMPLE(INTERNAL_HASH, hash, ASN1_UTF8STRING),
}ASN1_SEQUENCE_END(INTERNAL_HASH)
IMPLEMENT_ASN1_FUNCTIONS(INTERNAL_HASH);
IMPLEMENT_ASN1_ENCRYPT_SIGN(INTERNAL_HASH);


HashCorrelation::HashCorrelation()
{
	ActualHash = NULL;
}

HashCorrelation::~HashCorrelation()
{
	if(ActualHash)
		INTERNAL_HASH_free(ActualHash);
}

void HashCorrelation::SetCert(const PKI_CERT & cert)
{
	m_cert = cert;
}

bool HashCorrelation::SetCurrentHash(const mString & hash)
{
	if(!hash.size())
		return true;

	ASN1_ENCRYPT_SIGN * crypted_hash;
	if(!m_cert)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}
	
	crypted_hash = NULL;
	if(!ASN1_ENCRYPT_SIGN_From_PEM(&crypted_hash, hash.c_str()))
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_BAD_DATAS);
		ASN1_ENCRYPT_SIGN_free(crypted_hash);
		crypted_hash = NULL;
		return false;
	}

	//We now have the hash, let's decrypt it
	const PKI_RSA & rootKey = m_cert.GetPrivateKey();
	if(!rootKey)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_UNKNOWN);
		ASN1_ENCRYPT_SIGN_free(crypted_hash);
		crypted_hash = NULL;
		return false;
	}

	//Decrypt the body and verify signature
	if(INTERNAL_HASH_decrypt_verify(crypted_hash, &ActualHash, rootKey.GetRsaKey(), rootKey.GetRsaKey()) <= 0)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		ASN1_ENCRYPT_SIGN_free(crypted_hash);
		crypted_hash = NULL;
		return false;
	}
	ASN1_ENCRYPT_SIGN_free(crypted_hash);

	return true;
}

bool HashCorrelation::AddEntry(const char * id, int id_len, const char * datas, int dataslen) const
{
	char * entry_hash;
	char * full_hash;

	// We first compute the id and the datas together
	entry_hash = HashDatas(id, id_len, datas, dataslen, id_len);
	if(!entry_hash)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!ActualHash)
	{
		((HashCorrelation*)this)->ActualHash = INTERNAL_HASH_new();
		if(!ActualHash)
		{
			NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_MALLOC);
			free(entry_hash);
			return false;
		}
		if(!ASN1_STRING_set(ActualHash->hash, entry_hash, id_len))
		{
			NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_MALLOC);
			free(entry_hash);
			return false;
		}
	}
	else
	{
		full_hash = HashDatas(entry_hash, id_len, (char*)ActualHash->hash->data, ActualHash->hash->length, id_len);
		if(!full_hash)
		{
			NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
			free(entry_hash);
			return false;
		}
		if(ASN1_STRING_set(ActualHash->hash, full_hash, id_len) <= 0)
		{
			NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_UNKNOWN);
			free(full_hash);
			free(entry_hash);
			return false;
		}
		free(full_hash);
	}
	free(entry_hash);
	return true;
}

bool HashCorrelation::AddEntry(unsigned long id, const char * datas, int dataslen) const
{
	int str_id_len;
	char str_id[STR_LONG_LEN];

	str_id_len = sprintf(str_id, "%ld", id);
	return AddEntry(str_id, str_id_len, datas, dataslen);
}

char * HashCorrelation::HashDatas(const char *datas1, int len1, const char *datas2, int len2, int & resultlen) const
{
	int i;
	const char * dst_datas;
	const char * src_datas;
	char * str_hash;
	int min;

	if(len1 > len2)
	{
		dst_datas = datas1;
		src_datas = datas2;
		resultlen = len1;
		min = len2;
	}
	else
	{
		dst_datas = datas2;
		src_datas = datas1;
		resultlen = len2;
		min = len1;
	}

	str_hash = (char*)malloc(resultlen);
	if(!str_hash)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_MALLOC);
		return NULL;
	}
	memcpy(str_hash, dst_datas, resultlen);

	for(i=0; i<min; i++)
	{
		str_hash[i] ^= src_datas[i];
	}

	return str_hash;
}

bool HashCorrelation::operator==(const HashCorrelation &other)
{
	if(!ActualHash || !other.ActualHash)
		return false;

	return (ASN1_STRING_cmp(ActualHash->hash, other.ActualHash->hash)==0)?true:false;
}

HashCorrelation::operator int()
{
	if(!m_cert)
		return 0;
	else
		return 1;
}


bool HashCorrelation::GetHash(mString & pem_hash) const
{
	char * pem_datas;
	ASN1_ENCRYPT_SIGN * crypted_hash;

	if(!m_cert)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	
	pem_hash = "";
	if(!ActualHash)
		return true;


	const PKI_RSA & rootKey = m_cert.GetPrivateKey();
	if(!rootKey)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}

	crypted_hash = ASN1_ENCRYPT_SIGN_new();	
	if(!crypted_hash)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	//Ok, now we're all ready to sign and crypt for ourself the whole think
	//Encrypt the body and generate signature
	if( !INTERNAL_HASH_sign_encrypt(crypted_hash, ActualHash, rootKey.GetRsaKey(), rootKey.GetRsaKey(), EVP_sha1(), EVP_des_ede3_cbc()) ) 
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		ASN1_ENCRYPT_SIGN_free(crypted_hash);
		return false;
	}

	//We have everything we can now convert it to PEM
	pem_datas = ASN1_ENCRYPT_SIGN_To_PEM(crypted_hash);
	if(!pem_datas)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		ASN1_ENCRYPT_SIGN_free(crypted_hash);		
		return false;
	}
	ASN1_ENCRYPT_SIGN_free(crypted_hash);

	pem_hash = pem_datas;
	free(pem_datas);

	return true;
}

void HashCorrelation::Clear()
{
	if(ActualHash)
	{
		INTERNAL_HASH_free(ActualHash);
		ActualHash = NULL;
	}

}
