/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/


// StringHashTable_String.cpp: implementation of the StringHashTable_String class.
//
//////////////////////////////////////////////////////////////////////

#include "HashTable_String.h"




HashTable_String::HashTable_String()
{
}


HashTable_String::~HashTable_String()
{
	Clear();
}


const char * HashTable_String::Get(const char * Name) const
{
	return (const char *)m_Get(Name);
}


const char * HashTable_String::Get(long Pos) const
{
	return (const char *)m_GetPTR(Pos);
}


bool HashTable_String::Add(const char *Name, const char *Value)
{
	return m_Add(Name, Value, strlen(Value)+1);
}


bool HashTable_String::Modify(const char *Name, const char *Value)
{
	return m_Modify(Name, Value, strlen(Value)+1);
}


bool HashTable_String::operator =(const HashTable_String &other)
{
	return PKI_HashTable::operator=(other);
}

bool HashTable_String::From_EXTENSION_VALUE(STACK_OF(EXTENSION_VALUE) * Exts)
{
	Clear();
	EXTENSION_VALUE * currValue;
	int i;
	const char * value;
	const char * name;
	
	AllowDuplicateNames();
	for(i=0; i<SKM_sk_num(EXTENSION_VALUE, Exts); i++)
	{
		currValue = SKM_sk_value(EXTENSION_VALUE, Exts, i);
		if(!currValue) continue;

		value = ASN1_STRING_GET(currValue->value);
		name = ASN1_STRING_GET(currValue->name);
		if(!name || !value) continue;

		if(!Add(name, value)) continue;
	}

	return true;
}

bool HashTable_String::To_EXTENSION_VALUE(STACK_OF(EXTENSION_VALUE) * Exts)
{
	EXTENSION_VALUE * currValue;
	int i;
	const char * value;
	const char * name;

	while( (currValue = SKM_sk_pop(EXTENSION_VALUE, Exts)) )
	{
		EXTENSION_VALUE_free(currValue);
	}

	for(i=0; i<EntriesCount(); i++)
	{
		name = GetName(i);
		value = Get(i);
		if(!name || !value) continue;

		currValue = EXTENSION_VALUE_new();
		if(!currValue) continue;

		ASN1_STRING_SET(&currValue->name, name);
		ASN1_STRING_SET(&currValue->value, value);

		if(SKM_sk_push(EXTENSION_VALUE, Exts, currValue) == -1)
			EXTENSION_VALUE_free(currValue);

	}

	return true;
}

bool HashTable_String::From_PLUG_OPTION(STACK_OF(PLUG_OPTION) * Options)
{
	Clear();
	PLUG_OPTION * currValue;
	int i;
	const char * value;
	const char * name;
	
	AllowDuplicateNames();
	for(i=0; i<SKM_sk_num(PLUG_OPTION, Options); i++)
	{
		currValue = SKM_sk_value(PLUG_OPTION, Options, i);
		if(!currValue) continue;

		value = ASN1_STRING_GET(currValue->value);
		name = ASN1_STRING_GET(currValue->name);
		if(!name || !value) continue;

		if(!Add(name, value)) continue;
	}

	return true;
}

bool HashTable_String::To_PLUG_OPTION(STACK_OF(PLUG_OPTION) * Options)
{
	PLUG_OPTION * currValue;
	int i;
	const char * value;
	const char * name;

	while( (currValue = SKM_sk_pop(PLUG_OPTION, Options)) )
	{
		PLUG_OPTION_free(currValue);
	}

	for(i=0; i<EntriesCount(); i++)
	{
		name = GetName(i);
		value = Get(i);
		if(!name || !value) continue;

		currValue = PLUG_OPTION_new();
		if(!currValue) continue;

		ASN1_STRING_SET(&currValue->name, name);
		ASN1_STRING_SET(&currValue->value, value);

		if(SKM_sk_push(PLUG_OPTION, Options, currValue) == -1)
			PLUG_OPTION_free(currValue);

	}

	return true;
}

bool HashTable_String::From_POLICY_VALUE(STACK_OF(POLICY_VALUE) * Policies)
{
	Clear();
	POLICY_VALUE * currValue;
	int i;
	const char * value;
	const char * name;
	
	AllowDuplicateNames();
	for(i=0; i<SKM_sk_num(POLICY_VALUE, Policies); i++)
	{
		currValue = SKM_sk_value(POLICY_VALUE, Policies, i);
		if(!currValue) continue;

		value = ASN1_STRING_GET(currValue->value);
		name = ASN1_STRING_GET(currValue->name);
		if(!name || !value) continue;

		if(!Add(name, value)) continue;
	}

	return true;
}

bool HashTable_String::To_POLICY_VALUE(STACK_OF(POLICY_VALUE) * Policies)
{
	POLICY_VALUE * currValue;
	int i;
	const char * value;
	const char * name;

	while( (currValue = SKM_sk_pop(POLICY_VALUE, Policies)) )
	{
		POLICY_VALUE_free(currValue);
	}

	for(i=0; i<EntriesCount(); i++)
	{
		name = GetName(i);
		value = Get(i);
		if(!name || !value) continue;

		currValue = POLICY_VALUE_new();
		if(!currValue) continue;

		ASN1_STRING_SET(&currValue->name, name);
		ASN1_STRING_SET(&currValue->value, value);

		if(SKM_sk_push(POLICY_VALUE, Policies, currValue) == -1)
			POLICY_VALUE_free(currValue);

	}

	return true;
}


DN_TYPE HashTable_String::IsValidPolicyField(const char *Name, const char *Value)
{
	int k;
	static char * DN_TYPE_VALUE[NUM_DN_TYPE] = {STR_OPTIONAL, STR_SUPPLIED, STR_MATCH};

	if(!Name || !Value)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_BAD_PARAM);
		ERR_set_error_data("(null)", ERR_TXT_STRING);
		return DN_TYPE_UNKNOWN;
	}

	if(!IsValidObject(Name))
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_UNKNOWN_NID);
		ERR_set_error_data(strdup(Name), ERR_TXT_MALLOCED | ERR_TXT_STRING);
		return DN_TYPE_UNKNOWN;
	}		
	
	
	//On vrifie que le champ ait une bonne valeur
	for(k=0; k<NUM_DN_TYPE; k++)
	{
		if(strncmp(Value, DN_TYPE_VALUE[k], strlen(DN_TYPE_VALUE[k]))==0)
		{
			return (DN_TYPE)k;
		}
	}

	NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_BAD_DATAS_FORMAT);
	ERR_set_error_data(strdup(Name), ERR_TXT_MALLOCED | ERR_TXT_STRING);
	return DN_TYPE_UNKNOWN;
}
