/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/


// PKI_CSR.cpp: implementation of the PKI_CSR class.
//
//////////////////////////////////////////////////////////////////////

#include "PKI_CSR.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

PKI_CSR::PKI_CSR()
{
	Reset();
}

PKI_CSR::PKI_CSR(const char * PEM_CSR)
{
	Reset();
	if(!SetCSR(PEM_CSR))
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		Clear();
		throw ExceptionNewPKI();
	}
}

PKI_CSR::PKI_CSR(const HashTable_Dn & RequestDN, const PKI_RSA & mRsaKey)
{
	Reset();
	if(!GenerateCSR(RequestDN, mRsaKey))
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		Clear();
		throw ExceptionNewPKI();
	}
}

PKI_CSR::~PKI_CSR()
{
	Clear();
}

bool PKI_CSR::MakeRequest(const HashTable_Dn & RequestDN)
{
	X509_NAME * subj;

	if (!X509_REQ_set_version(req,0L))  /* version 1 */
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_GENERATE_CSR);
		return false;
	}

	subj = X509_REQ_get_subject_name(req);
	if(!subj)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_GENERATE_CSR);
		return false;
	}

	if(!RequestDN.To_X509_NAME(subj))
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}
	
	if(!LoadDN())
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}


bool PKI_CSR::X509ReqToString()
{

	int dsize=0;
	unsigned char *p,*data=NULL;
	int dataslen;
	PEM_DER pem_cvrt;
	char * tmpPEM;
	
	if ((dsize=i2d_X509_REQ(req,NULL)) < 0)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_CONVERT_X509REQ_CSR);
		return false;
	}

	/* dzise + 8 bytes are needed */
	data=(unsigned char *)malloc(dsize+20);
	if (!data)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	p=data;
	if((dsize=i2d_X509_REQ(req,&p))<0)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_CONVERT_X509REQ_CSR);
		return false;
	}
	if(!pem_cvrt.Der2Pem((char *)data, dsize, &tmpPEM, &dataslen))
	{
		free(data);
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	lePEM_CSR = tmpPEM;
	free(tmpPEM);
	free(data);
	return true;
}

bool PKI_CSR::StringToX509Req(const char * pem)
{

	int dsize=0;
	unsigned char *data=NULL;
	unsigned char *tmp_data=NULL;
	PEM_DER der_cvrt;
	BIO * sout;

	if(*pem == '-')
	{
		sout = BIO_new_mem_buf((char*)pem, strlen(pem));
		if(!sout)
		{
			NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_MALLOC);
			return false;
		}
		if(!PEM_read_bio_X509_REQ(sout, &req, NULL, NULL))
		{
			NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_CONVERT_CSR_X509REQ);
			BIO_free_all(sout);
			return false;
		}
		BIO_free_all(sout);
		return true;
	}

	if(!der_cvrt.Pem2Der(pem, strlen(pem), (char **)&data, &dsize))
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	tmp_data=data;
	if(!d2i_X509_REQ(&req,&data,dsize))
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_CONVERT_CSR_X509REQ);
		free(tmp_data);
		return false;
	}

	free(tmp_data);
	return true;
}

const mString & PKI_CSR::GetPemCSR() const
{
	return lePEM_CSR;
}

bool PKI_CSR::LoadDN()
{
	X509_NAME * name;
	name=X509_REQ_get_subject_name(req);
	if(!name)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}
	if(!laRequestDN.From_X509_NAME(name))
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}
	return true;
}

const PKI_RSA & PKI_CSR::GetRsaKey() const
{
	return RsaKey;
}

const HashTable_Dn & PKI_CSR::GetRequestDN() const
{
	return laRequestDN;
}

X509_REQ * PKI_CSR::GetX509_REQ(bool duplicate) const
{
	if(duplicate)
	{
		CRYPTO_add(&req->references, 1, CRYPTO_LOCK_X509_REQ);
	}
	return req;
}

bool PKI_CSR::CheckSignature() const
{
	if(!req) return false;
	EVP_PKEY * pubkey = X509_REQ_get_pubkey(req);
	if(!pubkey) return false;


	if(X509_REQ_verify(req, pubkey)<0)
	{
		EVP_PKEY_free(pubkey);
		return false;
	}
	else
	{
		EVP_PKEY_free(pubkey);
		return true;
	}
}

bool PKI_CSR::SetCSR(const char *PEM_CSR)
{
	Clear();

	//On teste les paramtres
	if(!PEM_CSR)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	req=X509_REQ_new();
	if(!req)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	if(!StringToX509Req(PEM_CSR))
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!X509ReqToString())
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!LoadDN())
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	pubKey = X509_PUBKEY_get(req->req_info->pubkey);
	if(!pubKey)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool PKI_CSR::GenerateCSR(const HashTable_Dn & RequestDN, const PKI_RSA & mRsaKey)
{
	Clear();

	const EVP_MD * digest;

	//On teste les paramtres
	if(!mRsaKey)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	
	req=X509_REQ_new();
	if(!req)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	
	if(!MakeRequest(RequestDN))
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!X509_REQ_set_pubkey(req, (EVP_PKEY*)mRsaKey.GetRsaKey()))
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_GENERATE_CSR);
		return false;
	}

	digest=(EVP_MD *)EVP_md5();
	if(!digest)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_GENERATE_CSR);
		return false;
	}

	if(!X509_REQ_sign(req, (EVP_PKEY*)mRsaKey.GetRsaKey(), digest))
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_GENERATE_CSR);
		return false;
	}

	if(!X509ReqToString())
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	RsaKey = mRsaKey;

	pubKey = X509_PUBKEY_get(req->req_info->pubkey);

	return true;
}

void PKI_CSR::Reset()
{
	pubKey = NULL;
	req=NULL;
}

void PKI_CSR::Clear()
{
	RsaKey.Clear();
	if(req) X509_REQ_free(req);
	if(pubKey) EVP_PKEY_free(pubKey);
	Reset();
}


bool PKI_CSR::SetCSR(X509_REQ *x509csr)
{
	Clear();

	//On teste les paramtres
	if(!x509csr)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	req = X509_REQ_dup(x509csr);
	if(!req)
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	if(!X509ReqToString())
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!LoadDN())
	{
		NEWPKIerr(CRYPTO_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	pubKey = X509_PUBKEY_get(req->req_info->pubkey);

	return true;
}

const EVP_PKEY * PKI_CSR::GetPublicKey() const
{
	return pubKey;
}

const X509_PUBKEY * PKI_CSR::GetX509_PUBKEY() const
{
	return req->req_info->pubkey;
}
