/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/


// PkiClient.h: interface for the PkiClient class.
//
//////////////////////////////////////////////////////////////////////

#ifndef PKICLIENT_H
#define PKICLIENT_H

#include <stdio.h>
#include <openssl/ssl.h>
#include <time.h>

#include <ASN1/ASN1_req.h>
#include <ASN1/ASN1_resp.h>

#include <PKI_ERR.h>
#include <NewPKI.h>
#include <UnixWinSock.h>
#include <newpki_threads.h>
#include <HashTable/HashTable_String.h>
#include <HashTable/HashTable_Dn.h>
#include <PKI_CERT.h>

#if defined(_WIN32) && !defined(NO_BIO_MS_SSL)
	#include <bio_ms_ssl.h>
#else
	#define PCCERT_CONTEXT void *
#endif
#include <mString.h>
#include <vector>
using namespace std;


typedef void (PROC_WAIT_CALLBACK) ();

#ifndef DOXYGEN_SHOULD_SKIP_THIS

class THREAD_ERROR
{
public:
	THREAD_ERROR();
	THREAD_ERROR(const THREAD_ERROR & other);
	~THREAD_ERROR();
	bool operator=(const THREAD_ERROR & other);
	int get_Line();
	void set_Line(int Line);

	int get_Lib();
	void set_Lib(int Lib);

	int get_Function();
	void set_Function(int Function);

	int get_Code();
	void set_Code(int Code);

	const mString & get_File();
	void set_File(const mString & File);

	const mString & get_Datas();
	void set_Datas(const mString & Datas);

	bool FromThreadErr();
private:
	int m_line;
	mString m_file;
	int m_lib;
	int m_function;
	int m_code;
	mString m_datas;
};

#endif




class PkiClient
{
public:
	PkiClient(PROC_WAIT_CALLBACK * WaitFunction);
	virtual ~PkiClient();
	bool Connect(char * Server, unsigned int Port, const PKI_CERT & AuthCert, PCCERT_CONTEXT MsAuthCert);
	bool Reconnect();
	void CloseConnection();
	bool UserLogin(char * entity, char *username, char *password, int & UserType);
	void GetPeerCertificate(PKI_CERT * PeerCert);
	void GetEntityCert(PKI_CERT * entity_cert);
	unsigned int GetRemotePort();
	char * GetRemoteServer();
	char * GetError();
	void SetReadTimeOut(int ReadTimeOut);
	static void CancelAllConnections();
	static unsigned long GetSentBytes();
	static unsigned long GetRecvBytes();
	
	bool SearchLdap(const char * Filters, STACK_OF(LDAP_RESULT) ** Results);
	bool DeletePkcs12(ASN1_INTEGER * CertId);
	bool GetCaCert(PKI_CERT & cert);
	bool SynchObjects(OBJECTS_LIST_REQUEST * KnownObjects, OBJECTS_LIST_RESPONSE ** RepKnownObjects);
	bool GenerateCRL();
	bool RequestCertificate(REQUEST_CERT* cert_request, unsigned long & ReqId);
	bool GetLogsType(STACK_OF(ASN1_INTEGER) ** LogsType);
	bool GetEntityLogsType(int EntityType, STACK_OF(ASN1_INTEGER) ** LogsType);
	bool SetRepositoriesList(REPS_INFO * repositories);
	bool GetRepositoriesList(REPS_INFO ** repositories);
	bool InitEntity(ENTITY_INIT_REQ* init_entity);
	bool GetConfiguration(EXPORTED_PKI_CONF ** conf);
	bool GetMyACL(ASN1_BIT_STRING **acl);
	bool GetOfflineState(bool & offline);
	bool SetOfflineState(bool offline);
	bool PushConfiguration(EXPORTED_PKI_CONF * conf);
	bool RevokeEntityCert(unsigned long serial);
	bool RevokeUserCert(unsigned long serial);
	bool CreatePkiUser(X509_REQ * req, int PkiAdmin, STACK_OF(X509) ** certs);
	bool SignEntity(ENTITY_SIGNATURE_REQ * sign_entity, ENTITY_SIGNATURE_RESP ** resp);
	bool ChangePassword(CHANGE_PASSWD * Password);
	bool ChangeUserPassword(CHANGE_USER_PASSWD * Password);
	bool DeleteEntity(const char * Name);
	bool CreateEntity(ENTITY_CREATION_REQ * cre_entity, ENTITY_CREATION_RESP ** creation_resp);
	bool EnumEntities(STACK_OF(ENTITY_ENTRY_INFO) ** entities);
	bool EnumLOG(ADMIN_REQ_ENUM_LOGS * Filters, STACK_OF(LOG_ENTRY) ** Logs);
	bool GetLogsCount(ADMIN_REQ_ENUM_LOGS * Filters, unsigned long & LogsCount);
	bool EnumUsers(STACK_OF(USER_ENTRY) ** users);
	bool EnumUsersCert(STACK_OF(USER_ENTRY_CERT) ** users_cert);
	bool CreateUser(USER_INFO * UserInfo);
	bool UpdateUser(USER_ENTRY * UserEntry);
	bool GetMyConf(ENTITY_CONF_CRYPTED **myConf);
	bool GetEntitiesLinks(ENTITIES_LINKS ** links);
	bool SetEntitiesLinks(ENTITIES_LINKS * links);

	bool GetLocalConf(ENTITY_CONF ** EntityConf);
	bool ImportProfile(NEWPKI_PROFILE * Profile, unsigned long & ProfileId);
	bool CheckLogsIntegrity();
	bool GetEntityConf(X509 * entity_cert, ENTITY_CONF_BODY ** conf);
	bool SetEntityConf(X509 * entity_cert, ENTITY_CONF_BODY * conf);
	bool GetEntityAcl(X509 * entity_cert, X509_ACL ** acl);
	bool SetEntityAcl(X509 *entity_cert, STACK_OF(ACL_ENTRY) * acl_entries);
	bool GetEntityAudits(X509 *entity_cert, ENTITY_AUDIT **audits);
	bool SetEntityAudits(X509 *entity_cert, ENTITY_AUDIT * audits);
	bool GetEntityMailConf(X509 *entity_cert, EMAIL_CONF ** mail_conf);
	bool SetEntityMailConf(X509 *entity_cert, EMAIL_CONF * mail_conf);
	bool ResourceUnlock(char * ResourceName);
	bool ResourceLock(char * ResourceName);
	bool AdminSendMail(MAIL_INFO *mail);
	bool GetMailConf(EMAIL_CONF ** mail_conf);
	bool GetCaStatus(ASN1_INTEGER ** status);
	bool CreateCaRoot(REQ_CREATE_ROOT_CA * ca_info);
	bool EnumCERT(long index, long num, CERT_STATE state, STACK_OF(INTERNAL_CA_CERT) ** certs);
	bool EnumCrls(long index, long num, STACK_OF(X509_CRL) ** crls);
	bool SignCSR(X509_REQ * req, int Days, PKCS7 **certs);
	bool CreateCaChild(REQ_CREATE_CHILD_CA * ca_info, X509_REQ ** csr);
	bool ImportChildCaCert(PKCS7 * certs);
	bool RenameGroup(unsigned long id, char * Name);
	bool DelGroup(unsigned long id);
	bool AddGroup(char * Name);
	bool SetGroups(USERS_GROUPS * Groups);
	bool GetGroups(USERS_GROUPS ** Groups);
	bool RevokeCert(unsigned long serial);
	bool PushWaitingObject(WAITING_NEWPKI_OBJECT * objs);
	bool SendRequest(CRYPTED_NEWPKI_REQUEST * reqs, ASN1_OCTET_STRING ** transactionId);
	bool GetMyRequests(TRANSACTION_IDS * transactionIds, STORED_NEWPKI_REQUESTS ** objects);
	bool SendResponse(STORED_NEWPKI_RESPONSE * resp);
	bool GetMyResponses(TRANSACTION_IDS * transactionIds, STORED_NEWPKI_RESPONSES ** objects);
	bool EnumProfiles(long index, long num, STACK_OF(NEWPKI_PROFILE_DATAS) ** Profiles);

private:
	static void AddRecvBytes(unsigned long bytes);
	static void AddSentBytes(unsigned long bytes);
	void Disconnect();
	void ClearErrors();
	static int verify_callback(int ok, X509_STORE_CTX *ctx);
	typedef struct md_PKI_CLIENT_PARAM
	{
		PkiClient * me_this;
		int command;
		void * a;
		bool Finished;
		bool Success;
		ADMIN_RESPONSE * response;
	}PKI_CLIENT_PARAM;
	
	typedef struct md_PKI_CONNECTION_PARAM
	{
		PkiClient * me_this;
		char Server[50];
		unsigned short Port;
		bool Finished;
		bool Success;
		PKI_CERT m_AuthCert;
		PCCERT_CONTEXT m_MsAuthCert;
	}PKI_CONNECTION_PARAM;


	#define INIT_NET_PARAM_TYPE(name,type) \
	PKI_CLIENT_PARAM ClientParam; \
	type * ttvalue = type##_new(); \
	if(!ttvalue) \
	{ \
		NEWPKIerr(CLIENT_ERROR_TXT, ERROR_MALLOC); \
		PackThreadErrors(); \
		return false; \
	} \
	ClientParam.command = name; \
	ClientParam.a = ttvalue; \
	ClientParam.me_this = this; \
	ClientParam.response = NULL;
	
	#define INIT_NET_PARAM_ANY(name) \
	PKI_CLIENT_PARAM ClientParam; \
	ClientParam.command = name; \
	ClientParam.me_this = this; \
	ClientParam.response = NULL; \
	ClientParam.a = NULL;
	
	#define INIT_NET_PARAM_DUP(name,type,stsource) \
	PKI_CLIENT_PARAM ClientParam; \
	type * ttvalue; \
	if(stsource) \
	{ \
		ttvalue = (type*)ASN1_item_dup(ASN1_ITEM_rptr(type), stsource); \
		if(!ttvalue) \
		{ \
			NEWPKIerr(CLIENT_ERROR_TXT, ERROR_MALLOC); \
			PackThreadErrors(); \
			return false; \
		} \
	} \
	else \
		ttvalue = NULL; \
	ClientParam.command = name; \
	ClientParam.a = ttvalue; \
	ClientParam.me_this = this; \
	ClientParam.response = NULL;


	#define INIT_NET_PARAM_STATIC(name,stsource) \
	PKI_CLIENT_PARAM ClientParam; \
	ClientParam.command = name; \
	ClientParam.a = stsource; \
	ClientParam.me_this = this; \
	ClientParam.response = NULL;


	#define FREE_NET_PARAM(type) \
	if(ttvalue) type##_free(ttvalue); \
	FREE_NET_PARAM_ANY();
	
	#define FREE_NET_PARAM_ANY() \
	if(ClientParam.response) ADMIN_RESPONSE_free(ClientParam.response)
	
	#define NET_PARAM() ClientParam

	#define NET_PARAM_RESPONSE() ClientParam.response
	
	#define NET_PARAM_SET_STR(stdest, stsource) \
		ASN1_STRING_SET(&(ttvalue->stdest), stsource)

	#define NET_PARAM_SET_INT(stdest, stsource) \
		ASN1_INTEGER_SET(&(ttvalue->stdest), stsource)

	
	
	PROC_WAIT_CALLBACK * m_WaitFunction;

	bool DoNetworkExchange(PKI_CLIENT_PARAM * ClientParam);
	static void ThreadProcSocket(const NewpkiThread * Thread, void * param);
	bool DoBufferSend(PKI_CLIENT_PARAM * ClientParam);

	bool DoNetworkConnection(PKI_CONNECTION_PARAM * ConnParam);
	static void ThreadProcConnection(const NewpkiThread * Thread, void *param);
	bool DoConnection(PKI_CONNECTION_PARAM * ConnParam);

	BIO * ssl_bio;
	SSL_CTX * ctx;
	SSL * ssl;
	SOCKET m_laSocket;
	PKI_CONNECTION_PARAM m_ConnParam;
	X509 * m_entity_cert;
	vector<THREAD_ERROR> errors;

	void PackThreadErrors();
	char * str_Error;
	int m_ReadTimeOut;
	static vector<SOCKET> ListAllConnections;
	static CriticalSection ListAllConnectionsLock;
	static CriticalSection DatasCounterLock;
	static unsigned long m_RcvDatas;
	static unsigned long m_SndDatas;
};

#endif // !defined(PKICLIENT_H)
