/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/


// SQL.cpp: implementation of the SQL class.
//
//////////////////////////////////////////////////////////////////////

#include "SQL.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

SQL::SQL(const SQL_Connection * DB_Connection)
{
	DB_Connection->Lock();
	this->res=NULL;
	sqlHND=DB_Connection->sqlHND;
	Connection = DB_Connection;
}

SQL::~SQL()
{
	if(this->res) mysql_free_result(this->res);
	Connection->Unlock();
}

bool SQL::Execute(const mString & req)
{
	if(this->res)
	{
		mysql_free_result(this->res);
		this->res=NULL;
	}
		
	if(mysql_query(this->sqlHND, req.c_str()))
	{
		NEWPKIerr(SQL_ERROR_TXT, ERROR_DB_SELECT);
		ERR_add_error_data(1, mysql_error(this->sqlHND));
		return false;
	}

	if(mysql_field_count(this->sqlHND))
	{
		this->res = mysql_store_result(this->sqlHND);
		if(!this->res)
		{
			NEWPKIerr(SQL_ERROR_TXT, ERROR_DB_SELECT);
			ERR_add_error_data(1, mysql_error(this->sqlHND));
			return false;
		}
	}
	return true;
}

bool SQL::NumRows(long *num_rows)
{
	if(!res || !this->sqlHND)
	{
		*num_rows=0;
		return false;
	}

	(*num_rows)=(long)mysql_num_rows(this->res);

	if(!(*num_rows) && mysql_errno(this->sqlHND))
	{
		NEWPKIerr(SQL_ERROR_TXT, ERROR_DB_NUMROWS);
		ERR_add_error_data(1, mysql_error(this->sqlHND));
		return false;
	}

	return true;
}

bool SQL::Value(int RowNum, mString Name, mString & result)
{
	int num_fields;
	int i;
	MYSQL_FIELD *fields;
	long num_rows;
	MYSQL_ROW row;

	if(!NumRows(&num_rows))
	{
		NEWPKIerr(SQL_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!this->res || !this->sqlHND || RowNum<0 || RowNum>=num_rows)
	{
		return false;
	}

	
	mysql_data_seek(res, RowNum);

	row=mysql_fetch_row(res);
	if(!row)
	{
		NEWPKIerr(SQL_ERROR_TXT, ERROR_DB_FETCHROWS);
		ERR_add_error_data(1, mysql_error(this->sqlHND));
		return false;
	}

	num_fields = mysql_num_fields(this->res);
	if(!num_fields && mysql_errno(this->sqlHND))
	{
		NEWPKIerr(SQL_ERROR_TXT, ERROR_DB_NUMFIELDS);
		ERR_add_error_data(1, mysql_error(this->sqlHND));
		return false;
	}


	fields = mysql_fetch_fields(this->res);
	if(!fields)
	{
		NEWPKIerr(SQL_ERROR_TXT, ERROR_DB_FETCHFIELD);
		ERR_add_error_data(1, mysql_error(this->sqlHND));
		return false;
	}


	for(i=0;i<num_fields;i++)
	{
		if(stricmp(fields[i].name, Name.c_str())==0)
		{
			result = row[i];
			return true;
		}
	}

	NEWPKIerr(SQL_ERROR_TXT, ERROR_DB_FIELDNAME);
	return false;
}

mString SQL::FormatString(mString String)
{
	mString Buffer;

	if(!String.size())
		return "";
	Buffer.resize((String.size()*2) + 1);
	mysql_real_escape_string(this->sqlHND, &Buffer[0], String.c_str(), String.size());

	return Buffer;
}

long SQL::GetLastID()
{
	return (long)mysql_insert_id(this->sqlHND);
}


bool SQL::CreateDatabase(mString DataBaseName)
{
	mString req = "CREATE DATABASE ";
	req += DataBaseName;
	if(!Execute(req))
	{
		NEWPKIerr(SQL_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!Connection->SelectDatabase(DataBaseName))
	{
		NEWPKIerr(SQL_ERROR_TXT, ERROR_ABORT);
		return false;;
	}
	return true;
}

bool SQL::DropDatabase(mString DataBaseName)
{
	mString req = "DROP DATABASE ";
	req += DataBaseName;

	if(!Execute(req))
	{
		NEWPKIerr(SQL_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool SQL::Value(int RowNum, int ColNum, mString & result)
{
	int num_fields;
	long num_rows;
	MYSQL_ROW row;

	if(!NumRows(&num_rows))
	{
		NEWPKIerr(SQL_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!this->res || !this->sqlHND || RowNum<0 || RowNum>=num_rows)
	{
		return false;
	}

	
	mysql_data_seek(res, RowNum);

	row=mysql_fetch_row(res);
	if(!row)
	{
		NEWPKIerr(SQL_ERROR_TXT, ERROR_DB_FETCHROWS);
		ERR_add_error_data(1, mysql_error(this->sqlHND));
		return false;
	}

	num_fields = mysql_num_fields(this->res);
	if(!num_fields && mysql_errno(this->sqlHND))
	{
		NEWPKIerr(SQL_ERROR_TXT, ERROR_DB_NUMFIELDS);
		ERR_add_error_data(1, mysql_error(this->sqlHND));
		return false;
	}

	if(ColNum >= num_fields)
	{
		NEWPKIerr(SQL_ERROR_TXT, ERROR_DB_NUMFIELDS);
		ERR_add_error_data(1, mysql_error(this->sqlHND));
		return false;
	}
	
	result = row[ColNum];
	return true;
}
