/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/

// mString.h: interface for the mString class.
//
//////////////////////////////////////////////////////////////////////

#ifndef MSTRING_H
#define MSTRING_H

#include <openssl/asn1.h>

#include <string>
#include <vector>
using namespace std;
#include <ctype.h>


class mBuffer
{
public:
	mBuffer()
	{
		m_buffer = NULL;
		m_bufferlen = 0;
	}
	virtual ~mBuffer()
	{
		Clear();
	}
	bool Copy(unsigned char * buffer, size_t bufferlen)
	{
		Clear();
		m_buffer = (unsigned char *)malloc(bufferlen);
		if(!m_buffer)
			return false;
		memcpy(m_buffer, buffer, bufferlen);
		m_bufferlen = bufferlen;
		return true;
	}

	void Assign(unsigned char * buffer, size_t bufferlen)
	{
		Clear();
		m_bufferlen = bufferlen;
		m_buffer = buffer;
	}
	const unsigned char * get_Buffer() const
	{
		return m_buffer;
	}
	size_t get_BufferLen() const
	{
		return m_bufferlen;
	}
	void Clear()
	{
		if(m_buffer)
		{
			free(m_buffer);
			m_buffer = NULL;
			m_bufferlen = 0;
		}
	}
private:
	unsigned char * m_buffer;
	size_t m_bufferlen;
};

class mString
{
public:
	mString();
	mString( const mString & other );
	mString( const string & other );
	mString( const char * other );
	mString( const unsigned char * other );
	mString( const ASN1_STRING * other );
	mString( const int other );
	mString( const unsigned int other );
	mString( const long other );
	mString( const unsigned long other );
	virtual ~mString();

	void TrimRight(const char * filters);
	void TrimLeft(const char * filters);
	mString Left(int count) const;
	mString Right(int count) const;
	bool left_compare(const char * cmp) const;
	void Split(const char * sep, vector<mString> & results) const;

	bool FromDER(const unsigned char * indatas, const int indataslen);
	bool ToDER(unsigned char ** outdatas, int * outdataslen) const;
	bool ToDER(mBuffer & outdatas) const;

	int sprintf(const char *format, ...);
	int vsprintf(const char *format, va_list args);
	int find(const char *str, int pos) const;
	void replace(int pos, int num, const char *datas);
	void resize(unsigned int size);
	unsigned int capacity() const;
	unsigned int size() const;
	int compare(const char *other) const;
	int compare(unsigned char *other) const;
	int compare(mString & other) const;
	int compare(string & other) const;
	int compare(int from, int max, char *other) const;
	int compare(int from, int max, unsigned char *other) const;
	int compare(int from, int max, mString & other) const;
	int compare(int from, int max, string & other) const;
	
	bool operator<(const mString& rhs) const;
	char & operator[](unsigned int pos);
	const char operator[](unsigned int pos) const;

	const char * c_str() const;
	int c_int() const;
	unsigned int c_uint() const;
	long c_lng() const;
	unsigned long c_ulng() const;
	const string & c_string() const;
	string & c_string();

	bool c_ASN1_INTEGER(ASN1_INTEGER ** dest) const;
	bool c_ASN1_UTF8STRING(ASN1_UTF8STRING ** dest) const;

	mString & operator=( const mString & other );
	mString & operator=( const string & other );
	mString & operator=( const char * other );
	mString & operator=( const unsigned char * other );
	mString & operator=( const ASN1_STRING * other );
	mString & operator=( const int other );
	mString & operator=( const unsigned int other );
	mString & operator=( const long other );
	mString & operator=( const unsigned long other );

	mString & operator+=( const mString & other );
	mString & operator+=( const string & other );
	mString & operator+=( const char * other );
	mString & operator+=( const unsigned char * other );
	mString & operator+=( const ASN1_STRING * other );
	mString & operator+=( const int other );
	mString & operator+=( const unsigned int other );
	mString & operator+=( const long other );
	mString & operator+=( const unsigned long other );

	bool operator==( const mString & other ) const;
	bool operator==( const string & other ) const;
	bool operator==( const unsigned char * other ) const;
	bool operator==( const char * other ) const;
	bool operator==( const ASN1_STRING * other ) const;
	bool operator==( const int other ) const;
	bool operator==( const unsigned int other ) const;
	bool operator==( const long other ) const;
	bool operator==( const unsigned long other ) const;

private:
	/* format read states */
	#define DP_S_DEFAULT    0
	#define DP_S_FLAGS      1
	#define DP_S_MIN        2
	#define DP_S_DOT        3
	#define DP_S_MAX        4
	#define DP_S_MOD        5
	#define DP_S_CONV       6
	#define DP_S_DONE       7

	/* format flags - Bits */
	#define DP_F_MINUS      (1 << 0)
	#define DP_F_PLUS       (1 << 1)
	#define DP_F_SPACE      (1 << 2)
	#define DP_F_NUM        (1 << 3)
	#define DP_F_ZERO       (1 << 4)
	#define DP_F_UP         (1 << 5)
	#define DP_F_UNSIGNED   (1 << 6)

	/* conversion flags */
	#define DP_C_SHORT      1
	#define DP_C_LONG       2
	#define DP_C_LDOUBLE    3
	#define DP_C_LLONG      4

	/* some handy macros */
	#define char_to_int(p) (p - '0')
	#define OSSL_MAX(p,q) ((p >= q) ? p : q)

	void fmtstr(const char *, int, int, int);
	void fmtint(long, int, int, int, int);
	void fmtfp(double, int, int, int);
	double abs_val(double value);
	double pow10(int exp);
	long roundv(double value);

	char buffer_lng[sizeof(long) * 5];
	string buffer;
	void Reset();
};

#endif
