/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/


// AsynchJobs.cpp: implementation of the AsynchJobs class.
//
//////////////////////////////////////////////////////////////////////

#include "AsynchJobs.h"
#include <MIME.h>
#include "svintl.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////



ASN1_SEQUENCE(STACK_REQUESTS) = {
	ASN1_SIMPLE(STACK_REQUESTS, id, ASN1_INTEGER),
	ASN1_SIMPLE(STACK_REQUESTS, req, CRYPTED_NEWPKI_REQUEST),
}ASN1_SEQUENCE_END(STACK_REQUESTS)
IMPLEMENT_ASN1_FUNCTIONS(STACK_REQUESTS)


ASN1_SEQUENCE(MAIL_QUEUE_ENTRY) = {
	ASN1_SIMPLE(MAIL_QUEUE_ENTRY, mail, MAIL_INFO),
	ASN1_SIMPLE(MAIL_QUEUE_ENTRY, Author, ASN1_UTF8STRING),
	ASN1_SEQUENCE_OF(MAIL_QUEUE_ENTRY, Recipients, ASN1_UTF8STRING),
	ASN1_SIMPLE(MAIL_QUEUE_ENTRY, AdminMail, ASN1_INTEGER),
}ASN1_SEQUENCE_END(MAIL_QUEUE_ENTRY)
IMPLEMENT_ASN1_FUNCTIONS(MAIL_QUEUE_ENTRY)
IMPLEMENT_PEM_CONVERT(MAIL_QUEUE_ENTRY, "NEWPKI MAIL")


AsynchJobs::AsynchJobs()
{
	m_MailerConf = NULL;
	m_EntityParentCerts = NULL;
	AdminMailQueue = NULL;
	m_Repositories = NULL;
	m_transactionIdReqs = NULL;
	m_transactionIdResps = NULL;
	waitingResps = NULL;
	waitingReqs = NULL;
	m_PkiConf = NULL;
	m_AdminMails = NULL;
	m_DbConn = NULL;

	hThreadMailQueue.Create(ThreadMailQueue, this);

	hThreadConfigurationGet.Create(ThreadConfigurationGet, this);

	hThreadConfigurationPush.Create(ThreadConfigurationPush, this);

}

AsynchJobs::~AsynchJobs()
{
	StopAll();
	RepositoriesAccessLock.LockWrite();
	if(m_Repositories)
	{
		REPS_INFO_free(m_Repositories);
		m_Repositories = NULL;
	}
	RepositoriesAccessLock.Unlock();
	if(m_DbConn)
		delete m_DbConn;
}

bool AsynchJobs::StartMailer()
{
	AdminMailQueue = SKM_sk_new_null(MAIL_QUEUE_ENTRY);
	if(!AdminMailQueue)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		throw ExceptionNewPKI();
	}
	
	m_EntityParentCerts = SKM_sk_new_null(X509);
	if(!m_EntityParentCerts)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	
	if(!m_AdminMails)
	{
		m_AdminMails = SKM_sk_new_null(ASN1_UTF8STRING);
		if(!m_AdminMails)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
			return false;
		}
	}

	if(!hThreadMailQueue.Start())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}

	return true;
}

void AsynchJobs::StopMailer()
{
	hThreadMailQueue.Stop();

	AdminMailQueueLock.Lock();
	if(AdminMailQueue)
	{
		sk_X509_pop_free(AdminMailQueue, MAIL_QUEUE_ENTRY_free);
		AdminMailQueue = NULL;
	}

	if(m_MailerConf)
	{
		EMAIL_CONF_free(m_MailerConf);
		m_MailerConf = NULL;
	}

	if(m_EntityParentCerts)
	{
		SKM_sk_pop_free(X509, m_EntityParentCerts, X509_free);
		m_EntityParentCerts = NULL;
	}
	
	if(m_AdminMails)
	{
		SKM_sk_pop_free(ASN1_UTF8STRING, m_AdminMails, ASN1_UTF8STRING_free);
		m_AdminMails = NULL;
	}
	AdminMailQueueLock.Unlock();
}

bool AsynchJobs::StartConfSync()
{
	m_transactionIdResps = TRANSACTION_IDS_new();
	if(!m_transactionIdResps)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	m_transactionIdReqs = TRANSACTION_IDS_new();
	if(!m_transactionIdReqs)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	waitingResps = STORED_NEWPKI_RESPONSES_new();
	if(!waitingResps)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	waitingReqs = SKM_sk_new_null(STACK_REQUESTS);
	if(!waitingReqs)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	if(!hThreadConfigurationGet.Start())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}

	return true;
}

void AsynchJobs::StopConfSync()
{
	hThreadConfigurationGet.Stop();

	ConfGetLock.Lock();
	if(m_transactionIdReqs)
	{
		TRANSACTION_IDS_free(m_transactionIdReqs);
		m_transactionIdReqs = NULL;
	}
	if(m_transactionIdResps)
	{
		TRANSACTION_IDS_free(m_transactionIdResps);
		m_transactionIdResps = NULL;
	}
	if(waitingResps)
	{
		STORED_NEWPKI_RESPONSES_free(waitingResps);
		waitingResps = NULL;
	}
	if(waitingReqs)
	{
		SKM_sk_pop_free(STACK_REQUESTS, waitingReqs, STACK_REQUESTS_free);
		waitingReqs = NULL;
	}
	ConfGetLock.Unlock();
}

bool AsynchJobs::StartConfPush(bool PushToAllRep)
{
	m_PushToAllRep = PushToAllRep;
	if(!hThreadConfigurationPush.Start())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}
	return true;
}

void AsynchJobs::StopConfPush()
{
	hThreadConfigurationPush.Stop();
	ConfPushLock.Lock();
	if(m_PkiConf)
	{
		EXPORTED_PKI_CONF_free(m_PkiConf);
		m_PkiConf = NULL;
	}
	ConfPushLock.Unlock();
}

void AsynchJobs::StopAll()
{
	StopMailer();
	StopConfSync();
	StopConfPush();
}

void AsynchJobs::SetEventHandler(AsynchJobsEvent * event)
{
	m_event = event;
}

void AsynchJobs::SetEntityCert(const PKI_CERT & cert)
{
	m_EntityCert = cert;
}

void AsynchJobs::SetEntityName(const mString & EntityName)
{
	m_EntityName = EntityName;
}

bool AsynchJobs::SetEntityParentsCert(const STACK_OF(X509) * EntityParentCerts)
{
	AdminMailQueueLock.Lock();
	X509 * currCert;
	int i;
	while( (currCert = SKM_sk_pop(X509, m_EntityParentCerts)) )
	{
		X509_free(currCert);
	}
	for(i=0; i<SKM_sk_num(X509, EntityParentCerts); i++)
	{
		currCert = SKM_sk_value(X509, EntityParentCerts, i);
		if(!currCert) continue;
		currCert = X509_dup(currCert);
		if(!currCert) continue;
		if(SKM_sk_push(X509, m_EntityParentCerts, currCert) < 0)
		{
			X509_free(currCert);
		}
	}
	AdminMailQueueLock.Unlock();
	return true;
}

void AsynchJobs::SetLogger(EntityLog *log)
{
	m_Logging = log;
}

bool AsynchJobs::SetAdminMails(const PKI_ADMIN_SERIALS * Admins)
{
	ASN1_UTF8STRING * newMail;
	PKI_ADMIN_ENTRY * currEntry;
	int i;
	
	AdminMailQueueLock.Lock();

	if(!m_AdminMails)
	{
		m_AdminMails = SKM_sk_new_null(ASN1_UTF8STRING);
		if(!m_AdminMails)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
			AdminMailQueueLock.Unlock();
			return false;
		}
	}
		
	while( (newMail = SKM_sk_pop(ASN1_UTF8STRING, m_AdminMails)) )
	{
		ASN1_UTF8STRING_free(newMail);
	}
	
	for(i=0; i<SKM_sk_num(PKI_ADMIN_ENTRY, Admins->admins); i++)
	{
		currEntry = SKM_sk_value(PKI_ADMIN_ENTRY, Admins->admins, i);
		if(!currEntry) continue;
		
		newMail = NULL;
		ASN1_STRING_SET(&newMail, ASN1_STRING_GET(currEntry->email));
		if(!newMail)
			continue;
			
		if(SKM_sk_push(ASN1_UTF8STRING, m_AdminMails, newMail) < 0)
			ASN1_UTF8STRING_free(newMail);
	}
	AdminMailQueueLock.Unlock();
	return true;
}

bool AsynchJobs::SetMailerInfo(mString MailFrom, const EMAIL_CONF * conf)
{
	AdminMailQueueLock.Lock();
	EMAIL_CONF_cpy(&m_MailerConf, conf);
	AdminMailQueueLock.Unlock();

	if(!m_Mailer.SetInfo(MailFrom, ASN1_STRING_GET(m_MailerConf->server), ASN1_INTEGER_GET(m_MailerConf->port), m_Logging))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}



bool AsynchJobs::SendMail(const mString & Author, const MailInfo & Mail, bool AdminMail) const
{
	ASN1_UTF8STRING * sRecipient;
	MAIL_QUEUE_ENTRY *q;

	if(!Mail)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	q = MAIL_QUEUE_ENTRY_new();
	if(!q)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	if(!Mail.get_Subject().c_ASN1_UTF8STRING(&q->mail->Subject))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		MAIL_QUEUE_ENTRY_free(q);
		return false;
	}
	if(!Mail.get_Body().c_ASN1_UTF8STRING(&q->mail->Body))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		MAIL_QUEUE_ENTRY_free(q);
		return false;
	}


	ASN1_INTEGER_SET(&q->mail->SignMail, Mail.get_SignMail()?1:0);

	if(!Author.c_ASN1_UTF8STRING(&q->Author))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		MAIL_QUEUE_ENTRY_free(q);
		return false;
	}

	ASN1_INTEGER_SET(&q->AdminMail, AdminMail?1:0);


	// If it's not an admin mail we get the recipient from
	// the mail
	if(!AdminMail)
	{
		if(!Mail.get_MailTo().size())
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			MAIL_QUEUE_ENTRY_free(q);
			return false;
		}
		sRecipient = NULL;
		if(!Mail.get_MailTo().c_ASN1_UTF8STRING(&sRecipient))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			MAIL_QUEUE_ENTRY_free(q);
			return false;
		}
		if(SKM_sk_push(ASN1_UTF8STRING, q->Recipients, sRecipient) < 0)
		{
			ASN1_UTF8STRING_free(sRecipient);
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			MAIL_QUEUE_ENTRY_free(q);
			return false;
		}
	}
	else
	{
		// we copy the recipients list !
		if(!ASN1_UTF8STRING_stack_cpy(&q->Recipients, m_AdminMails))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			MAIL_QUEUE_ENTRY_free(q);
			return false;
		}
	}

	
	// Set attachement
	if(Mail.c_get_Attach().get_Buffer() && Mail.c_get_Attach().get_BufferLen())
		ASN1_STRING_set(q->mail->Attach, Mail.c_get_Attach().get_Buffer(), Mail.c_get_Attach().get_BufferLen());
	if(!Mail.get_AttachName().c_ASN1_UTF8STRING(&q->mail->AttachName))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		MAIL_QUEUE_ENTRY_free(q);
		return false;
	}
	if(!Mail.get_AttachType().c_ASN1_UTF8STRING(&q->mail->AttachType))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		MAIL_QUEUE_ENTRY_free(q);
		return false;
	}

	if(!InsertMail(q))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		MAIL_QUEUE_ENTRY_free(q);
		return false;
	}

	return true;
}


bool AsynchJobs::InsertMail(MAIL_QUEUE_ENTRY *q) const
{
	char * pem = MAIL_QUEUE_ENTRY_To_PEM(q);
	if(!pem)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	SQL sql(m_DbConn);
	mString req;

	req.sprintf(MAIL_INSERT, pem);
	free(pem);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	q->id = sql.GetLastID();
	
	AdminMailQueueLock.Lock();
	if(SKM_sk_push(MAIL_QUEUE_ENTRY, AdminMailQueue, q) < 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		AdminMailQueueLock.Unlock();
		return false;
	}
	AdminMailQueueLock.Unlock();
	return true;
}

void AsynchJobs::ThreadMailQueue(const NewpkiThread * Thread, void *param)
{
	MIME mimeEncoder;
	AsynchJobs * me_this = (AsynchJobs *)param;
	MAIL_QUEUE_ENTRY * currMail;
	bool ret;
	mString Body;
	mString mime;
	STACK_OF(ASN1_UTF8STRING) * Recipients;
	time_t lastLoad = 0;;
	time_t currTime;

	do 
	{
		me_this->AdminMailQueueLock.Lock();
		do
		{
			time(&currTime);
			if( (currTime - lastLoad) >= MAILS_DB_RELOAD )
			{
				me_this->LoadMails();
				lastLoad = currTime;
			}
			me_this->AdminMailQueueLock.Unlock();
			NewpkiThread::Sleep(100);
			me_this->AdminMailQueueLock.Lock();
		}
		while(!Thread->ShouldStop() && (!SKM_sk_num(MAIL_QUEUE_ENTRY, me_this->AdminMailQueue) || !me_this->m_Mailer || !me_this->m_EntityParentCerts));
		
		if(Thread->ShouldStop())
		{
			me_this->AdminMailQueueLock.Unlock();
			return;
		}
		
		currMail = SKM_sk_pop(MAIL_QUEUE_ENTRY, me_this->AdminMailQueue);
		me_this->AdminMailQueueLock.Unlock();

		Body = _sv("Entity: ");
		Body += me_this->m_EntityName;
		Body += _sv("\nAuthor: ");
		Body += ASN1_STRING_GET(currMail->Author);
		Body += "\n\n";
		Body += ASN1_STRING_GET(currMail->mail->Body);

		if(SKM_sk_num(ASN1_UTF8STRING, currMail->Recipients))
			Recipients = currMail->Recipients;
		else
			Recipients = me_this->m_AdminMails;

		if( (ret = mimeEncoder.GenerateMIME(mime, Body.c_str(), currMail->mail->Attach->data, currMail->mail->Attach->length, ASN1_STRING_GET(currMail->mail->AttachName), ASN1_STRING_GET(currMail->mail->AttachType))) )
		{
			if(ASN1_INTEGER_GET(currMail->mail->SignMail))
				ret = mimeEncoder.GenerateSMIME(mime, mime.c_str(), me_this->m_EntityCert.GetX509(), me_this->m_EntityCert.GetPrivateKey().GetRsaKey(), me_this->m_EntityParentCerts);
		}
		if(ret)
		{
			ERR_clear_error();
			ret = me_this->m_Mailer.Send((ASN1_INTEGER_GET(currMail->AdminMail) == 1), Recipients, ASN1_STRING_GET(currMail->mail->Subject), (char*)mime.c_str());
		}

		// The email was successfuly sent we can 
		// now remove it from the DB
		if(ret)
		{
			me_this->RemoveMail(currMail->id);
		}
		MAIL_QUEUE_ENTRY_free(currMail);
	}
	while(true);
}


void AsynchJobs::RemoveMail(unsigned long id)
{
	SQL sql(m_DbConn);
	mString req;
	req.sprintf(MAIL_DELETE, id);
	sql.Execute(req);
}


void AsynchJobs::LoadMails()
{
	SQL sql(m_DbConn);
	mString pem;
	long NumRows;
	long i;
	int j;
	MAIL_QUEUE_ENTRY * newEntry;
	MAIL_QUEUE_ENTRY * currEntry;
	if(!sql.Execute(MAIL_SELECT))
	{
		return;
	}
	if(!sql.NumRows(&NumRows))
	{
		return;
	}
	for(i=0; i<NumRows; i++)
	{
		if(!sql.Value(i, "mail", pem))
		{
			return;
		}

		newEntry = NULL;
		if(MAIL_QUEUE_ENTRY_From_PEM(&newEntry, pem.c_str()) <= 0)
		{
			return;
		}
		if(!sql.Value(i, "mail_id", pem))
		{
			return;
		}
		newEntry->id = pem.c_ulng();

		// Do we already have this mail in the stack ?
		for(j=0; j<SKM_sk_num(MAIL_QUEUE_ENTRY, AdminMailQueue); j++)
		{
			currEntry = SKM_sk_value(MAIL_QUEUE_ENTRY, AdminMailQueue, j);
			if(!currEntry)
				continue;
			if(currEntry->id == newEntry->id)
				break;
		}
		// The mail wasn't found
		if(j == SKM_sk_num(MAIL_QUEUE_ENTRY, AdminMailQueue))
		{
			if(SKM_sk_push(MAIL_QUEUE_ENTRY, AdminMailQueue, newEntry) < 0)
			{
				MAIL_QUEUE_ENTRY_free(newEntry);
				return;
			}
		}
		else
		{
			MAIL_QUEUE_ENTRY_free(newEntry);
		}
	}
}

const SmtpClient & AsynchJobs::GetMailer() const
{
	return m_Mailer;
}

void AsynchJobs::ThreadConfigurationGet(const NewpkiThread * Thread, void *param)
{
	AsynchJobs * me_this = (AsynchJobs *)param;
	time_t currTime;
	time_t lastTime;
	time_t waitTime;
	ENTITY_CONF_CRYPTED * newConf;
	STORED_NEWPKI_REQUESTS * reqs;
	STORED_NEWPKI_RESPONSES * resps;
	char * err;
	char * strerr;

	//We wait a bit (2 mins)
	time(&lastTime);
	do
	{
		NewpkiThread::Sleep(500);
		time(&currTime);
	}
	while( ((currTime - lastTime) < ASYNCH_JOBS_INITIAL_WAIT_TIME) && !Thread->ShouldStop());

	while(!Thread->ShouldStop())
	{
		waitTime = ASYNCH_JOBS_INTERVAL;
		reqs = NULL;
		resps = NULL;
		newConf = NULL;
		me_this->RepositoriesAccessLock.LockRead();
		if(me_this->m_Repositories && me_this->Private_GetMyConfiguration(Thread, &newConf, &reqs, &resps))
		{
			me_this->RepositoriesAccessLock.Unlock();
			if(newConf)
			{
				if(!me_this->m_event->OnNewConf(newConf))
				{
					err = ERR_to_string();
					if(err)
						strerr = err;
					else
						strerr = "";
					me_this->m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_ENTITY_GET_MY_CONF, 0, me_this->m_EntityCert.GetStringName(), LOG_NO_OBJECTID, "Conf", strerr);
					if(err)
						free(err);
				}
			}
			if(reqs && SKM_sk_num(STORED_NEWPKI_REQUEST, reqs->requests))
			{
				if(!me_this->m_event->OnNewRequests(reqs))
				{
					err = ERR_to_string();
					if(err)
						strerr = err;
					else
						strerr = "";
					me_this->m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_ENTITY_GET_MY_CONF, 0, me_this->m_EntityCert.GetStringName(), LOG_NO_OBJECTID, "Conf", strerr);
					if(err)
						free(err);
				}
			}
			if(resps && SKM_sk_num(STORED_NEWPKI_RESPONSE, resps->responses))
			{
				if(!me_this->m_event->OnNewResponses(resps))
				{
					err = ERR_to_string();
					if(err)
						strerr = err;
					else
						strerr = "";
					me_this->m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_ENTITY_GET_MY_CONF, 0, me_this->m_EntityCert.GetStringName(), LOG_NO_OBJECTID, "Conf", strerr);
					if(err)
						free(err);
				}
			}
		}
		else
		{
			me_this->RepositoriesAccessLock.Unlock();
		}

		if(newConf)
			ENTITY_CONF_CRYPTED_free(newConf);
		if(reqs)
			STORED_NEWPKI_REQUESTS_free(reqs);
		if(resps)
			STORED_NEWPKI_RESPONSES_free(resps);

		time(&lastTime);

		do
		{
			NewpkiThread::Sleep(500);
			time(&currTime);
		}
		while( ((currTime - lastTime) < waitTime) && !Thread->ShouldStop());
	}
}

bool AsynchJobs::Private_GetMyConfiguration(const NewpkiThread * Thread, ENTITY_CONF_CRYPTED ** myConf, STORED_NEWPKI_REQUESTS **reqs, STORED_NEWPKI_RESPONSES **resps)
{
	int i;
	STORED_NEWPKI_RESPONSE * currResp;
	STACK_REQUESTS * currReq;
	ASN1_OCTET_STRING * transactionId;
	mString err;

	REP_ENTRY_INFO * currRep;
	PkiClient ClientPki(NULL);

	*resps = NULL;
	*myConf = NULL;
	*reqs = NULL;

	ConfGetLock.Lock();
	for(i=0; i<SKM_sk_num(REP_ENTRY_INFO, m_Repositories->list); i++)
	{
		if(Thread->ShouldStop())
		{
			ConfGetLock.Unlock();
			return false;
		}
		currRep = SKM_sk_value(REP_ENTRY_INFO, m_Repositories->list, i);
		if(!currRep) continue;

		//If the destination is me we ignore it
		if(m_EntityCert == currRep->RepositorySSL)
		{
			continue;
		}

		//We connect to the repository
		if(!ConnectToRepository(currRep, &ClientPki, 30, err))
		{
			NewpkiDebug(LOG_LEVEL_WARNING, m_EntityName.c_str(), _sv("Failed to get my conf and my objects from repository %s (%s:%ld)\nReason:%s"), ASN1_STRING_GET(currRep->Name), ASN1_STRING_GET(currRep->Address), ASN1_INTEGER_GET(currRep->Port), err.c_str());
			m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_ENTITY_GET_MY_CONF, 0, NULL, LOG_NO_OBJECTID, ASN1_STRING_GET(currRep->Name), err.c_str());
			continue;
		}
		if(Thread->ShouldStop())
		{
			ConfGetLock.Unlock();
			return false;
		}
		
		//Send each request
		while( !Thread->ShouldStop() && (currReq = SKM_sk_pop(STACK_REQUESTS, waitingReqs)) )
		{
			if(!ClientPki.SendRequest(currReq->req, &transactionId))
			{			
				NewpkiDebug(LOG_LEVEL_WARNING, m_EntityName.c_str(), _sv("Failed to get my conf and my objects from repository %s (%s:%ld)\nReason:%s"), ASN1_STRING_GET(currRep->Name), ASN1_STRING_GET(currRep->Address), ASN1_INTEGER_GET(currRep->Port), ClientPki.GetError());
				m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_ENTITY_GET_MY_CONF, 0, NULL, LOG_NO_OBJECTID, ASN1_STRING_GET(currRep->Name), ClientPki.GetError());
				STACK_REQUESTS_free(currReq);
				continue;
			}
			else
			{
				m_event->OnRequestSent(ASN1_INTEGER_GET(currReq->id), transactionId, currReq->req);
				ASN1_OCTET_STRING_free(transactionId);
			}
			STACK_REQUESTS_free(currReq);
		}
		if(Thread->ShouldStop())
		{
			ConfGetLock.Unlock();
			return false;
		}
		// An error occured !
		if(SKM_sk_num(STACK_REQUESTS, waitingReqs))
		{
			continue;
		}

		

		//Send each response
		while( (currResp = SKM_sk_pop(STORED_NEWPKI_RESPONSE, waitingResps->responses)) )
		{
			if(!ClientPki.SendResponse(currResp))
			{			
				NewpkiDebug(LOG_LEVEL_WARNING, m_EntityName.c_str(), _sv("Failed to get my conf and my objects from repository %s (%s:%ld)\nReason:%s"), ASN1_STRING_GET(currRep->Name), ASN1_STRING_GET(currRep->Address), ASN1_INTEGER_GET(currRep->Port), ClientPki.GetError());
				m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_ENTITY_GET_MY_CONF, 0, NULL, LOG_NO_OBJECTID, ASN1_STRING_GET(currRep->Name), ClientPki.GetError());
				STORED_NEWPKI_RESPONSE_free(currResp);
				continue;
			}
			else
			{
				m_event->OnResponseSent(currResp);
			}
			STORED_NEWPKI_RESPONSE_free(currResp);
		}
		if(Thread->ShouldStop())
		{
			ConfGetLock.Unlock();
			return false;
		}
		// An error occured !
		if(SKM_sk_num(STORED_NEWPKI_RESPONSE, waitingResps->responses))
		{
			continue;
		}

		// Get entity conf
		if(!*myConf && !ClientPki.GetMyConf(myConf))
		{			
			NewpkiDebug(LOG_LEVEL_WARNING, m_EntityName.c_str(), _sv("Failed to get my conf and my objects from repository %s (%s:%ld)\nReason:%s"), ASN1_STRING_GET(currRep->Name), ASN1_STRING_GET(currRep->Address), ASN1_INTEGER_GET(currRep->Port), ClientPki.GetError());
			m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_ENTITY_GET_MY_CONF, 0, NULL, LOG_NO_OBJECTID, ASN1_STRING_GET(currRep->Name), ClientPki.GetError());
			*myConf = NULL;
			continue;
		}
		if(Thread->ShouldStop())
		{
			ConfGetLock.Unlock();
			return false;
		}

		// Get entity requests
		if(m_transactionIdReqs && !*reqs)
		{
			while( (transactionId = SKM_sk_pop(ASN1_OCTET_STRING, m_transactionIdReqs->transactionIds)) )
			{
				ASN1_OCTET_STRING_free(transactionId);
			}
			transactionId = NULL;
			if(!Thread->ShouldStop() && m_event->GetProcessedTransactionIds(m_transactionIdReqs))
			{
				if(!Thread->ShouldStop() && !ClientPki.GetMyRequests(m_transactionIdReqs, reqs))
				{
					NewpkiDebug(LOG_LEVEL_WARNING, m_EntityName.c_str(), _sv("Failed to get my conf and my objects from repository %s (%s:%ld)\nReason:%s"), ASN1_STRING_GET(currRep->Name), ASN1_STRING_GET(currRep->Address), ASN1_INTEGER_GET(currRep->Port), ClientPki.GetError());
					m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_ENTITY_GET_MY_CONF, 0, NULL, LOG_NO_OBJECTID, ASN1_STRING_GET(currRep->Name), ClientPki.GetError());
					*reqs = NULL;
					continue;
				}
			}
		}
		if(Thread->ShouldStop())
		{
			ConfGetLock.Unlock();
			return false;
		}
		
		
		// Get entity responses
		if(m_transactionIdResps && !*resps)
		{
			while( (transactionId = SKM_sk_pop(ASN1_OCTET_STRING, m_transactionIdResps->transactionIds)) )
			{
				ASN1_OCTET_STRING_free(transactionId);
			}
			transactionId = NULL;
			if(!Thread->ShouldStop() && m_event->GetWaitingTransactionIds(m_transactionIdResps) && SKM_sk_num(ASN1_OCTET_STRING, m_transactionIdResps->transactionIds))
			{
				if(!Thread->ShouldStop() && !ClientPki.GetMyResponses(m_transactionIdResps, resps))
				{
					NewpkiDebug(LOG_LEVEL_WARNING, m_EntityName.c_str(), _sv("Failed to get my conf and my objects from repository %s (%s:%ld)\nReason:%s"), ASN1_STRING_GET(currRep->Name), ASN1_STRING_GET(currRep->Address), ASN1_INTEGER_GET(currRep->Port), ClientPki.GetError());
					m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_ENTITY_GET_MY_CONF, 0, NULL, LOG_NO_OBJECTID, ASN1_STRING_GET(currRep->Name), ClientPki.GetError());
					*resps = NULL;
					continue;
				}
			}
		}
		if(Thread->ShouldStop())
		{
			ConfGetLock.Unlock();
			return false;
		}
		
		NewpkiDebug(LOG_LEVEL_INFO, m_EntityName.c_str(), _sv("Successfully got my conf and my objects from repository %s (%s:%ld)"), ASN1_STRING_GET(currRep->Name), ASN1_STRING_GET(currRep->Address), ASN1_INTEGER_GET(currRep->Port));
		ClientPki.CloseConnection();
		ConfGetLock.Unlock();
		return true;
	}
	ClientPki.CloseConnection();
	ConfGetLock.Unlock();
	return false;
}


bool AsynchJobs::SetRepositories(const REPS_INFO *Repositories)
{
	RepositoriesAccessLock.LockWrite();
	if(!REPS_INFO_cpy(&m_Repositories, Repositories))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		RepositoriesAccessLock.Unlock();
		return false;
	}
	RepositoriesAccessLock.Unlock();
	return true;
}

bool AsynchJobs::InsertRequest(unsigned long id, const CRYPTED_NEWPKI_REQUEST *req) const
{
	ConfGetLock.Lock();
	STACK_REQUESTS * newReq;
	newReq = STACK_REQUESTS_new();
	if(!newReq)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ConfGetLock.Unlock();
		return false;
	}
	ASN1_INTEGER_SET(&newReq->id, id);
	if(!CRYPTED_NEWPKI_REQUEST_cpy(&newReq->req, req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfGetLock.Unlock();
		return false;
	}

	if(SKM_sk_push(STACK_REQUESTS, waitingReqs, newReq) < 0)
	{
		STACK_REQUESTS_free(newReq);
	}

	ConfGetLock.Unlock();
	return true;
}

bool AsynchJobs::InsertResponse(const STORED_NEWPKI_RESPONSE *resp) const
{
	STORED_NEWPKI_RESPONSE * lresp;
	ConfGetLock.Lock();
	lresp = STORED_NEWPKI_RESPONSE_cpy(NULL, resp);
	if(!lresp)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfGetLock.Unlock();
		return false;
	}
	if(SKM_sk_push(STORED_NEWPKI_RESPONSE, waitingResps->responses, lresp) < 0)
	{
		STORED_NEWPKI_RESPONSE_free(lresp);
	}

	ConfGetLock.Unlock();
	return true;
}



void AsynchJobs::SpawnStop()
{
	hThreadConfigurationGet.SpawStop();
	hThreadConfigurationPush.SpawStop();
	hThreadMailQueue.SpawStop();
}



void AsynchJobs::ThreadConfigurationPush(const NewpkiThread * Thread, void *param)
{
	long lastVersion;
	AsynchJobs * me_this = (AsynchJobs *)param;
	time_t currTime;
	time_t lastTime;
	time_t waitTime;
	time_t lastUpdate;

	//We do this to always push the configuration at startup
	//Which forces a full synchronization of all repositories
	lastVersion = 0;

	//We wait a bit (2 mins)
	time(&lastTime);
	do
	{
		NewpkiThread::Sleep(500);
		time(&currTime);
	}
	while( ((currTime - lastTime) < ASYNCH_JOBS_INITIAL_WAIT_TIME) && !Thread->ShouldStop());

	time(&lastTime);
	time(&lastUpdate);

	while(!Thread->ShouldStop())
	{
		waitTime = 15;

		me_this->ConfPushLock.Lock();
		me_this->RepositoriesAccessLock.LockRead();
		//Do we have a conf ?
		if(!me_this->m_PkiConf || !me_this->m_Repositories)
		{
			me_this->ConfPushLock.Unlock();
			me_this->RepositoriesAccessLock.Unlock();
			goto wait_next;
		}		

		// We deploy new configuration on all known repositories
		time(&lastUpdate);
		if(!me_this->Private_PushConfiguration(Thread))
		{
			waitTime = 300;
		}
		EXPORTED_PKI_CONF_free(me_this->m_PkiConf);
		me_this->m_PkiConf = NULL;
		me_this->RepositoriesAccessLock.Unlock();
		me_this->ConfPushLock.Unlock();

wait_next:
		do
		{
			NewpkiThread::Sleep(500);
			time(&currTime);
		}
		while( ((currTime - lastTime) < waitTime) && !Thread->ShouldStop());
		lastTime = currTime;

		//Every hour we force an update
		if( (currTime - lastUpdate) > 3600)
			lastVersion = 0;
	}
}

bool AsynchJobs::Private_PushConfiguration(const NewpkiThread * Thread)
{
	int i;
	REP_ENTRY_INFO * currRep;
	bool Success = false;
	PkiClient ClientPki(NULL);
	mString err;

	// We deploy conf on first available repository
	// the repository will be in charge on deploying it
	// on all other repositories
	for(i=0; i<SKM_sk_num(REP_ENTRY_INFO, m_Repositories->list); i++)
	{
		if(Thread->ShouldStop())
		{
			return true;
		}
		currRep = SKM_sk_value(REP_ENTRY_INFO, m_Repositories->list, i);
		if(!currRep) continue;
		
		//If the current rep is me or is firewalled we ignore it
		if(m_EntityCert == currRep->RepositorySSL ||
			ASN1_BIT_STRING_get_bit(currRep->Flags, REP_ENTRY_INFO_FLAG_FIREWALLED))
			continue;

		ERR_clear_error();

		m_Logging->LogMessage(LOG_STATUS_TYPE_REQUEST, LOG_MESSAGE_TYPE_CONFIG_PUSH, 0, NULL, LOG_NO_OBJECTID, ASN1_STRING_GET(currRep->Name));
		
		//We connect to the repository
		if(!ConnectToRepository(currRep, &ClientPki, 30, err))
		{
			NewpkiDebug(LOG_LEVEL_WARNING, m_EntityName.c_str(), _sv("Failed to push conf on entity %s (%s:%ld)\nReason:%s"), ASN1_STRING_GET(currRep->Name), ASN1_STRING_GET(currRep->Address), ASN1_INTEGER_GET(currRep->Port), err.c_str());
			m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_CONFIG_PUSH, 0, NULL, LOG_NO_OBJECTID, ASN1_STRING_GET(currRep->Name), err.c_str());
			continue;
		}
		if(!ClientPki.PushConfiguration(m_PkiConf))
		{			
			NewpkiDebug(LOG_LEVEL_WARNING, m_EntityName.c_str(), _sv("Failed to push conf on entity %s (%s:%ld)\nReason:%s"), ASN1_STRING_GET(currRep->Name), ASN1_STRING_GET(currRep->Address), ASN1_INTEGER_GET(currRep->Port), ClientPki.GetError());
			m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_CONFIG_PUSH, 0, NULL, LOG_NO_OBJECTID, ASN1_STRING_GET(currRep->Name), ClientPki.GetError());
			continue;
		}
		
		NewpkiDebug(LOG_LEVEL_INFO, m_EntityName.c_str(), _sv("Successfully pushed conf on entity %s (%s:%ld)"), ASN1_STRING_GET(currRep->Name), ASN1_STRING_GET(currRep->Address), ASN1_INTEGER_GET(currRep->Port));
		m_Logging->LogMessage(LOG_STATUS_TYPE_SUCCESS, LOG_MESSAGE_TYPE_CONFIG_PUSH, 0, NULL, LOG_NO_OBJECTID, ASN1_STRING_GET(currRep->Name));

		if(!m_PushToAllRep)
		{
			ClientPki.CloseConnection();
			return true;
		}
		Success = true;
	}
	ClientPki.CloseConnection();
	return Success;
}


bool AsynchJobs::SetPkiConf(const EXPORTED_PKI_CONF *PkiConf)
{
	ConfPushLock.Lock();
	if(!EXPORTED_PKI_CONF_cpy(&m_PkiConf, PkiConf))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfPushLock.Unlock();
		return false;
	}
	ConfPushLock.Unlock();
	return true;
}

bool AsynchJobs::ConnectToRepository(const REP_ENTRY_INFO * currRep, PkiClient * ClientPki, int TimeOut, mString & err) const
{
	PKI_CERT peerCert;
	int UserType;
	char * str_err;
	long Pos;
	const char * cn;

	err="";

	ClientPki->CloseConnection();

	ClientPki->SetReadTimeOut(TimeOut);

	//We connect to the repository
	if(!ClientPki->Connect(ASN1_STRING_GET(currRep->Address), ASN1_INTEGER_GET(currRep->Port), m_EntityCert, NULL))
	{
		err = ClientPki->GetError();
		return false;
	}
	if(!ClientPki->UserLogin(ASN1_STRING_GET(currRep->Name), NULL, NULL, UserType))
	{
		err = ClientPki->GetError();
		return false;
	}
	
	//Does the cert correspond to the one expected ?
	ClientPki->GetEntityCert(&peerCert);
	if(!peerCert)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNEXPECTED_CERT);
		str_err = ERR_to_string();
		if(str_err)
		{
			err = str_err;
			free(str_err);
		}
		return false;
	}

	// Is a cert signed by the internal CA ?
	if(m_AclValidator->ValidateCert(peerCert) != INTERNAL_CA_TYPE_ENTITY)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNEXPECTED_CERT);
		str_err = ERR_to_string();
		if(str_err)
		{
			err = str_err;
			free(str_err);
		}
		return false;
	}

	if(peerCert != currRep->RepositorySSL)
	{
		// Is the CN the same ?
		Pos = peerCert.GetCertDN().SeekEntryName("commonName", HASHTABLE_NOT_FOUND);
		if(Pos != HASHTABLE_NOT_FOUND)
		{
			cn = peerCert.GetCertDN().Get(Pos);
			if(cn)
			{
				if(strcmp(ASN1_STRING_GET(currRep->Name), cn) == 0)
				{
					// Yes it the same name, we issue a warning
					NewpkiDebug(LOG_LEVEL_WARNING, m_EntityName.c_str(), _sv("Repository certificate changed (%s)!"), ASN1_STRING_GET(currRep->Name));
					return true;
				}
			}
		}

		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNEXPECTED_CERT);
		str_err = ERR_to_string();
		if(str_err)
		{
			err = str_err;
			free(str_err);
		}
		return false;
	}
	return true;
}

void AsynchJobs::SetX509_ACL_Validator(const X509_ACL_Validator *AclValidator)
{
	m_AclValidator = AclValidator;
}

bool AsynchJobs::CreateTables(SQL_Connection *DbConn)
{
	SQL sql(DbConn);
	long i;
	char * CommonCreates[] = {MAIL_JOBS_STACK, NULL};
	mString req;
	//We execute each request
	for(i=0; CommonCreates[i]; i++)
	{
		if(!sql.Execute(CommonCreates[i]))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}

	return true;
}

bool AsynchJobs::SetDbConn(SQL_Connection *DbConn)
{
	if(m_DbConn)
	{
		delete m_DbConn;
		m_DbConn = NULL;
	}
	try
	{
		m_DbConn = DbConn->Clone();
	}
	catch(ExceptionNewPKI e)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!m_DbConn)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	return true;
}
