/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/


// AsynchJobs.h: interface for the AsynchJobs class.
//
//////////////////////////////////////////////////////////////////////

#ifndef ASYNCHJOBS_H
#define ASYNCHJOBS_H

#include <newpki_threads.h>
#include "EntityLog.h"
#include <mString.h>
#include <SmtpClient.h>
#include <PkiClient.h>
#include "X509_ACL_Validator.h"
#include "MailInfo.h"



/*!
	This class is the declaration of an event handler for asynch jobs
*/
class AsynchJobsEvent
{
public:
	/*! \brief Event triggered when a new conf has been seeked from a repository.
	 *  \param newConf [IN] The new conf.
	 *  \return true on success, false on failure.
	 */
	virtual bool OnNewConf(const ENTITY_CONF_CRYPTED * newConf)=0;
	
	/*! \brief Event triggered when new requests have been seeked from a repository.
	 *  \param reqs [IN] The new requests.
	 *  \return true on success, false on failure.
	 */
	virtual bool OnNewRequests(const STORED_NEWPKI_REQUESTS * reqs)=0;
	
	/*! \brief Event triggered when new responses have been seeked from a repository.
	 *  \param resps [IN] The new responses.
	 *  \return true on success, false on failure.
	 */
	virtual bool OnNewResponses(const STORED_NEWPKI_RESPONSES * resps)=0;
	
	/*! \brief Event triggered when a response has been successfuly sent to a repository.
	 *  \param resp [IN] The sent response.
	 *  \return true on success, false on failure.
	 */
	virtual bool OnResponseSent(const STORED_NEWPKI_RESPONSE * resp)=0;
	
	/*! \brief Event triggered when a request has been successfuly sent to a repository.
	 *  \param RequestId [IN] The ID of the request.
	 *  \param transactionId [IN] The transaction Id returned by the repository.
	 *  \param req [IN] The sent request.
	 *  \return true on success, false on failure.
	 */
	virtual bool OnRequestSent(unsigned long RequestId, const ASN1_OCTET_STRING * transactionId, const CRYPTED_NEWPKI_REQUEST * req)=0;
	
	/*! \brief Call back function used to get the list of request that haven't yet been responded.
	 *  \param transactionIDs [OUT] The un-responded transaction Ids.
	 *  \return true on success, false on failure.
	 */
	virtual bool GetWaitingTransactionIds(TRANSACTION_IDS * transactionIDs)=0;
	
	/*! \brief Call back function used to get the list of known responses.
	 *  \param transactionIDs [OUT] The responded transaction Ids.
	 *  \return true on success, false on failure.
	 */
	virtual bool GetProcessedTransactionIds(TRANSACTION_IDS * transactionIDs)=0;
	
	/*! \brief Call back function used to get requests that should have been sent but haven't because of an error, this function is called at startup and every hour.
	 *  \param Requests [OUT] The un-sent requests.
	 *  \return true on success, false on failure.
	 */
	virtual bool GetWaitingRequests(STACK_OF(CRYPTED_NEWPKI_REQUEST) * Requests)=0;
	
	/*! \brief Call back function used to get responses that should have been sent but haven't because of an error, this function is called at startup and every hour.
	 *  \param Responses [OUT] The un-sent responses.
	 *  \return true on success, false on failure.
	 */
	virtual bool GetWaitingResponses(STACK_OF(STORED_NEWPKI_RESPONSE) * Responses)=0;
};

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/*!
	This structure represent a request to be sent
*/
typedef struct st_STACK_REQUESTS
{
	ASN1_INTEGER * id;				//!< The ID of the request
	CRYPTED_NEWPKI_REQUEST * req;	//!< The request
}STACK_REQUESTS;
DECLARE_ASN1_FUNCTIONS(STACK_REQUESTS)


/*!
	This structure represent an email to be sent
*/
typedef struct st_MAIL_QUEUE_ENTRY
{
	STACK_OF(ASN1_UTF8STRING) * Recipients;	//!< The recipient(s)
	MAIL_INFO * mail;						//!< The mail
	ASN1_UTF8STRING * Author;				//!< The author
	ASN1_INTEGER * AdminMail;				//!< Is is an admin mail?
	unsigned long id;						//!< The internal ID
}MAIL_QUEUE_ENTRY;
DECLARE_ASN1_FUNCTIONS(MAIL_QUEUE_ENTRY)
DECLARE_PEM_CONVERT(MAIL_QUEUE_ENTRY)
#endif

#ifndef ASYNCH_JOBS_INTERVAL
	#define ASYNCH_JOBS_INTERVAL 60 //!< The time to wait after a job is done, to restart it
#endif
#ifndef ASYNCH_JOBS_INITIAL_WAIT_TIME
	#define ASYNCH_JOBS_INITIAL_WAIT_TIME 30 //!< The initial time to wait before starting a job
#endif
#ifndef MAILS_DB_RELOAD
	#define MAILS_DB_RELOAD 1800  //!< The time to wait before reloading all the mails from the DB
#endif

/*!
	This class holds the jobs that need to be done asynchronously
*/
class AsynchJobs  
{
public:
	/*! \brief This is the constructor.
	 */
	AsynchJobs();

	/*! \brief This is the destructor.
	 */
	virtual ~AsynchJobs();

	/*! \brief This function stops all the asynchron jobs.
	 */
	void StopAll();

	/*! \brief This function creates the tables needed to work.
	 *  \param DbConn [IN] A SQL connection.
	 *  \return true on success, false on failure.
	 */
	bool CreateTables(SQL_Connection * DbConn);

	/*! \brief This function sets the SQL connection.
	 *  \param DbConn [IN] A SQL connection.
	 *  \return true on success, false on failure.
	 */
	bool SetDbConn(SQL_Connection * DbConn);

	/*! \brief This function starts the configuration push job.
	 *  \param PushToAllRep [IN] When a new conf is declared, should it be sent to all known repositories?
	 *  \return true on success, false on failure.
	 */
	bool StartConfPush(bool PushToAllRep);

	/*! \brief This function stops the configuration push job.
	 */
	void StopConfPush();
	
	/*! \brief This function starts the mailer job.
	 *  \return true on success, false on failure.
	 */
	bool StartMailer();

	/*! \brief This function stops the mailer job.
	 */
	void StopMailer();

	/*! \brief This function starts the configuration retrieval job.
	 *  \return true on success, false on failure.
	 */
	bool StartConfSync();

	/*! \brief This function stops the configuration retrieval job.
	 */
	void StopConfSync();

	/*! \brief Yiels to start to stop.
	 */
	void SpawnStop();

	/*! \brief This function sets the name of the entity using this class.
	 *  \param EntityName [IN] The name of the entity.
	 */
	void SetEntityName(const mString & EntityName);

	/*! \brief This function sets the certificate of the entity using this class.
	 *  \param cert [IN] The certificate of the entity.
	 */
	void SetEntityCert(const PKI_CERT & cert);
	
	/*! \brief This function sets the list of parents' certificate.
	 *  \param EntityParentCerts [IN] The list.
	 *  \return true on success, false on failure.
	 */
	bool SetEntityParentsCert(const STACK_OF(X509) * EntityParentCerts);
	
	/*! \brief This function sets the instance that should receive the events.
	 *  \param event [IN] The event instance.
	 */
	void SetEventHandler(AsynchJobsEvent * event);
	
	/*! \brief This function sets the logging class.
	 *  \param log [IN] The logging class.
	 */
	void SetLogger(EntityLog * log);
	
	/*! \brief This function sets the mailer's info.
	 *  \param MailFrom [IN] The mail from.
	 *  \param conf [IN] The server configuration.
	 *  \return true on success, false on failure.
	 */
	bool SetMailerInfo(mString MailFrom, const EMAIL_CONF * conf);
	
	/*! \brief This function sets the list of admin mails.
	 *  \param Admins [IN] The list.
	 *  \return true on success, false on failure.
	 */
	bool SetAdminMails(const PKI_ADMIN_SERIALS * Admins);
	
	/*! \brief This function sets the list of repositories.
	 *  \param Repositories [IN] The list.
	 *  \return true on success, false on failure.
	 */
	bool SetRepositories(const REPS_INFO *Repositories);
	
	/*! \brief This function sets the validator.
	 *  \param AclValidator [IN] The validator.
	 */
	void SetX509_ACL_Validator(const X509_ACL_Validator * AclValidator);
	
	/*! \brief This function is called to declare a new PKI conf, it will sent it to the repositories.
	 *  \param PkiConf [IN] The PKI conf.
	 *  \return true on success, false on failure.
	 */
	bool SetPkiConf(const EXPORTED_PKI_CONF * PkiConf);
	
	/*! \brief This function is used to send a inter-entity-response, it will sent it to the repositories.
	 *  \param resp [IN] The response.
	 *  \return true on success, false on failure.
	 */
	bool InsertResponse(const STORED_NEWPKI_RESPONSE * resp) const;
	
	/*! \brief This function is used to send a inter-entity-request, it will sent it to the repositories.
	 *  \param id [IN] An internal id used by requester.
	 *  \param req [IN] The request.
	 *  \return true on success, false on failure.
	 */
	bool InsertRequest(unsigned long id, const CRYPTED_NEWPKI_REQUEST * req) const;

	/*! \brief This function returns the mailer.
	 *  \return The mailer.
	 */
	const SmtpClient & GetMailer() const;
	
	/*! \brief This function is used to send an email.
	 *  \param Author [IN] The author.
	 *  \param Mail [IN] The email.
	 *  \param AdminMail [IN] Is it an admin mail?
	 *  \return true on success, false on failure.
	 */
	bool SendMail(const mString & Author, const MailInfo & Mail, bool AdminMail) const;

	/*! \brief This function is to connect to a Repository.
	 *  \param currRep [IN] The repository info.
	 *  \param ClientPki [IN] The instance of PkiClient to use.
	 *  \param TimeOut [IN] TimeOut>0: the connection timeout.
	 *  \param err [OUT] The error.
	 *  \return true on success, false on failure.
	 */
	bool ConnectToRepository(const REP_ENTRY_INFO * currRep, PkiClient * ClientPki, int TimeOut, mString & err) const;
private:
	void LoadMails();
	void RemoveMail(unsigned long id);
	bool InsertMail(MAIL_QUEUE_ENTRY * q) const;
	#define MAIL_JOBS_STACK "create table mails (mail_id INT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY, mail LONGBLOB NOT NULL);"
	#define MAIL_INSERT "insert into mails (mail) values ('%s');"
	#define MAIL_DELETE "delete from mails where mail_id=%ld;"
	#define MAIL_SELECT "select * from mails;"

	AsynchJobsEvent * m_event;
	PKI_CERT m_EntityCert;
	mString m_EntityName;
	EntityLog * m_Logging;


	ReadersWriter RepositoriesAccessLock;
	CriticalSection ConfGetLock;
	REPS_INFO * m_Repositories;
	STORED_NEWPKI_RESPONSES * waitingResps;
	STACK_OF(STACK_REQUESTS) * waitingReqs;
	

	EXPORTED_PKI_CONF * m_PkiConf;
	CriticalSection ConfPushLock;
	NewpkiThread hThreadConfigurationPush;
	bool m_PushToAllRep;
	static void ThreadConfigurationPush(const NewpkiThread * Thread, void * param);
	bool Private_PushConfiguration(const NewpkiThread * Thread);

	
	TRANSACTION_IDS * m_transactionIdReqs;
	TRANSACTION_IDS * m_transactionIdResps;
	static void ThreadConfigurationGet(const NewpkiThread * Thread, void *param);
	NewpkiThread hThreadConfigurationGet;
	bool Private_GetMyConfiguration(const NewpkiThread * Thread, ENTITY_CONF_CRYPTED ** myConf, STORED_NEWPKI_REQUESTS **reqs, STORED_NEWPKI_RESPONSES **resps);

	SQL_Connection * m_DbConn;
	STACK_OF(X509) * m_EntityParentCerts;
	EMAIL_CONF * m_MailerConf;
	SmtpClient m_Mailer;
	STACK_OF(MAIL_QUEUE_ENTRY) * AdminMailQueue;
	CriticalSection AdminMailQueueLock;
	NewpkiThread hThreadMailQueue;
	static void ThreadMailQueue(const NewpkiThread * Thread, void * param);
	const X509_ACL_Validator * m_AclValidator;
	STACK_OF(ASN1_UTF8STRING) * m_AdminMails;
};

#endif
