/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/


// AutoSynchLDAP.cpp: implementation of the AutoSynchLDAP class.
//
//////////////////////////////////////////////////////////////////////

#include "AutoSynchLDAP.h"
#include <FileLog.h>
#include "svintl.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

AutoSynchLDAP::AutoSynchLDAP()
{
	m_Logging = NULL;
	m_event = NULL;
	m_group_id = 0;
	m_DnSpecs = NULL;
	hThreadSynchLdap.Create(ThreadSynchLdap, this);
	MustNotProcessUIDs = NULL;
	KnownUIDs = NULL;
}

AutoSynchLDAP::~AutoSynchLDAP()
{
	Stop();
	if(m_DnSpecs)
	{
		SKM_sk_pop_free(DN_SPECS, m_DnSpecs, DN_SPECS_free);
		m_DnSpecs = NULL;

	}
}

bool AutoSynchLDAP::Start(ASN1_UTF8STRING *ldap_server, ASN1_INTEGER *ldap_port, ASN1_UTF8STRING *ldap_username, ASN1_UTF8STRING *ldap_password, ASN1_UTF8STRING *ldap_base, ASN1_UTF8STRING *ldap_attr_name)
{
	Stop();

	ConfAccessLock.Lock();
	m_ldap_attr_name = ldap_attr_name;	
	if(ldap_username)
	{
		m_ldap_username = ldap_username;
		if(!m_ldap_username.size())
			m_ldap_username = "anonymous";
	}
	else
		m_ldap_username = "anonymous";

	if(!m_LdapClient.Connect(ldap_server, ldap_port, ldap_username, ldap_password, ldap_base, ldap_attr_name))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	KnownUIDs = SKM_sk_new_null(ASN1_UTF8STRING);
	if(!KnownUIDs)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ConfAccessLock.Unlock();
		return false;
	}

	MustNotProcessUIDs = SKM_sk_new_null(ASN1_UTF8STRING);
	if(!MustNotProcessUIDs)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ConfAccessLock.Unlock();
		return false;
	}
	
	if(!hThreadSynchLdap.Start())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}

void AutoSynchLDAP::Stop()
{
	hThreadSynchLdap.Stop();

	ConfAccessLock.Lock();
	if(MustNotProcessUIDs)
	{
		SKM_sk_pop_free(ASN1_UTF8STRING, MustNotProcessUIDs, ASN1_UTF8STRING_free);
		MustNotProcessUIDs = NULL;
	}
	if(KnownUIDs)
	{
		SKM_sk_pop_free(ASN1_UTF8STRING, KnownUIDs, ASN1_UTF8STRING_free);
		KnownUIDs = NULL;
	}
	ConfAccessLock.Unlock();
}

void AutoSynchLDAP::SetEventHandler(AutoSynchLDAPEvent *event)
{
	m_event = event;
}

bool AutoSynchLDAP::DoSynch()
{
	ASN1_UTF8STRING * currUID;
	ASN1_UTF8STRING * notUID;
	mString Filters;
	STACK_OF(LDAP_RESULT) * Results;
	LDAP_RESULT * currResult;
	char * err;
	unsigned long ferr;
	bool retval;


	while( (currUID = SKM_sk_pop(ASN1_UTF8STRING, KnownUIDs)) )
	{
		ASN1_UTF8STRING_free(currUID);
	}

	if(!ASN1_UTF8STRING_stack_cpy(&KnownUIDs, MustNotProcessUIDs))
	{
		return false;
	}
	
	ERR_clear_error();

	// Get known profiles
	if(!m_event->GetKnownUIDs(KnownUIDs))
	{
		err = ERR_to_string();
		if(err)
		{
			NewpkiDebug(LOG_LEVEL_ERROR, _sv("LDAP Synchronization"), _sv("Failed to get known ids: %s"), err);
			free(err);
		}
		else
		{
			NewpkiDebug(LOG_LEVEL_ERROR, _sv("LDAP Synchronization"), _sv("Failed to get known ids: %s"), _sv("Unknown"));
		}
		return false;
	}

	// Set up the filters
	Filters = "(&";
	Filters += m_ldap_filters;
	while( (currUID = SKM_sk_pop(ASN1_UTF8STRING, KnownUIDs)) )
	{
		Filters += "(!(";
		Filters += m_ldap_attr_name;
		Filters += "=";
		Filters += ASN1_STRING_GET(currUID);
		Filters += "))";

		ASN1_UTF8STRING_free(currUID);
	}
	Filters += ")";

	// Search in LDAP
	if( !(Results = m_LdapClient.Search(Filters, 30, 30)) )
	{
		err = ERR_to_string();
		if(err)
		{
			NewpkiDebug(LOG_LEVEL_ERROR, _sv("LDAP Synchronization"), _sv("Failed to search in LDAP: %s"), err);
			free(err);
		}
		else
		{
			NewpkiDebug(LOG_LEVEL_ERROR, _sv("LDAP Synchronization"), _sv("Failed to search in LDAP: %s"), _sv("Unknown"));
		}
		return false;
	}

	NewpkiDebug(LOG_LEVEL_INFO, _sv("LDAP Synchronization"), _sv("LDAP search returned %d new profile(s)"), Results->num);

	// If we found some entries we don't wait so long
	if(!Results->num)
	{
		SKM_sk_pop_free(LDAP_RESULT, Results, LDAP_RESULT_free);
		return false;
	}
	
	retval = true;
	while( (currResult = SKM_sk_pop(LDAP_RESULT, Results)) )
	{
		ConfAccessLock.Lock();
		if(!ProcessResult(currResult))
		{
			err = ERR_to_string();
			if(err)
			{
				NewpkiDebug(LOG_LEVEL_ERROR, _sv("LDAP Synchronization"), _sv("Failed to process new profile: %s"), err);
				free(err);
			}
			else
			{
				NewpkiDebug(LOG_LEVEL_ERROR, _sv("LDAP Synchronization"), _sv("Failed to process new profile: %s"), _sv("Unknown"));
			}

			// Is this rdn already known ? if yes we add it to the list
			// of "must-not-be-processed"
			if( (ferr = ERR_peek_error()) && ERR_GET_LIB(ferr) == ERR_LIB_NEWPKI && ERR_GET_REASON(ferr) == ERROR_NOT_UNIQUE_PROFILE)
			{
				notUID = NULL;
				ASN1_STRING_SET(&notUID, ASN1_STRING_GET(currResult->uid));
				if(notUID && SKM_sk_push(ASN1_UTF8STRING, MustNotProcessUIDs, notUID) < 0)
					ASN1_UTF8STRING_free(notUID);
			}
			else
			{
				retval = false;
			}
		}
		else
		{
			NewpkiDebug(LOG_LEVEL_INFO, _sv("LDAP Synchronization"), _sv("Successfuly processed new profile: %s"), ASN1_STRING_GET(currResult->rdn));
		}
		LDAP_RESULT_free(currResult);
	}
	SKM_sk_pop_free(LDAP_RESULT, Results, LDAP_RESULT_free);

	return retval;
}

void AutoSynchLDAP::ThreadSynchLdap(const NewpkiThread * Thread, void *param)
{
	time_t currTime;
	time_t lastTime;
	AutoSynchLDAP * me_this = (AutoSynchLDAP *)param;
	time_t waitTime;

	// We wait a bit before starting the sychro
	time(&lastTime);
	do
	{
		NewpkiThread::Sleep(500);
		time(&currTime);
	}
	while( ((currTime - lastTime) < 120) && !Thread->ShouldStop());


	NewpkiDebug(LOG_LEVEL_DEBUG, _sv("LDAP Synchronization"), _sv("Waiting for LDAP Synchronization Intialization..."));
	while(!Thread->ShouldStop() && (
		!me_this->m_LdapClient || 
		!me_this->m_event || 
		!me_this->m_group_id || 
		!me_this->m_DnSpecs || 
		!me_this->m_Policies.EntriesCount()))
	{
		NewpkiThread::Sleep(500);
	}
	if(Thread->ShouldStop())
	{
		NewpkiDebug(LOG_LEVEL_DEBUG, _sv("LDAP Synchronization"), _sv("LDAP Synchronization Stopped"));
		return;
	}
	NewpkiDebug(LOG_LEVEL_DEBUG, _sv("LDAP Synchronization"), _sv("LDAP Synchronization Intialized"));


	do
	{
		me_this->ConfAccessLock.Lock();
		if(me_this->DoSynch())
		{
			waitTime = 10;
		}
		else
		{
			waitTime = LDAP_SYNCH_WAIT_TIME;
		}
		me_this->ConfAccessLock.Unlock();


		time(&lastTime);
		do
		{
			NewpkiThread::Sleep(500);
			time(&currTime);
		}
		while( ((currTime - lastTime) < waitTime) && !Thread->ShouldStop());
	}
	while(!Thread->ShouldStop());

	NewpkiDebug(LOG_LEVEL_DEBUG, _sv("LDAP Synchronization"), _sv("LDAP Synchronization Stopped"));
}


void AutoSynchLDAP::SpawnStop()
{
	hThreadSynchLdap.SpawStop();
}

bool AutoSynchLDAP::SetOptions(bool RemoveUnknownFields, STACK_OF(DN_SPECS) * DnSpecs, const HashTable_String & Policies, unsigned long group_id, ASN1_UTF8STRING * ldap_filters)
{
	ConfAccessLock.Lock();

	m_ldap_filters = ldap_filters;

	m_group_id = group_id;
	m_RemoveUnknownFields = RemoveUnknownFields;

	if(!DN_SPECS_stack_cpy(&m_DnSpecs, DnSpecs))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if( !(m_Policies = Policies) )
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}


bool AutoSynchLDAP::ProcessResult(LDAP_RESULT *Result)
{
	HashTable_Dn Dn;
	HashTable_Dn IndexDn;
	long i;
	long j;
	const char * Name;
	const char * Value;
	DN_SPECS * currDnSpecs;
	LDAP_OBJECT * currObject;

	// We need to have it !
	if(!ASN1_STRING_GET(Result->uid))
	{
		return false;
	}


	for(i=0; i<SKM_sk_num(LDAP_OBJECT, Result->objects); i++)
	{
		currObject = SKM_sk_value(LDAP_OBJECT, Result->objects, i);
		if(!currObject)
			continue;
		
		Name = ASN1_STRING_GET(currObject->name);
		Value = ASN1_STRING_GET(currObject->value);
		if(!Name || !Value)
			continue;
		Dn.Add(Name, Value);
	}


	for(i=0; i<SKM_sk_num(DN_SPECS, m_DnSpecs); i++)
	{
		currDnSpecs = SKM_sk_value(DN_SPECS, m_DnSpecs, i);
		if(!currDnSpecs) continue;

		Name = ASN1_STRING_GET(currDnSpecs->Name);
		if(!Name)
			continue;

		j = Dn.SeekEntryName(Name, HASHTABLE_NOT_FOUND);
		if(j != HASHTABLE_NOT_FOUND)
		{
			Value = Dn.Get(j);
			if(!Value)
			{
				Value = ASN1_STRING_GET(currDnSpecs->Default);
			}
		}
		else
		{
			Value = ASN1_STRING_GET(currDnSpecs->Default);
		}
		if(!Value)
			Value = "";

		// Does it respect the length contraint ?
		if(!ASN1_INTEGER_GET(currDnSpecs->Min) || strlen(Value) >= (unsigned int)ASN1_INTEGER_GET(currDnSpecs->Min))
		{
			if(!ASN1_INTEGER_GET(currDnSpecs->Max) || strlen(Value) <= (unsigned int)ASN1_INTEGER_GET(currDnSpecs->Max))
			{
				IndexDn.Add(Name, Value);
			}
		}
		if(j != HASHTABLE_NOT_FOUND)
			Dn.Delete(j);
	}

	// Are we allowed to have some fields that are not 
	// present in the DN specs ?
	if(!m_RemoveUnknownFields)
	{
		for(i=0; i<Dn.EntriesCount(); i++)
		{
			Name = Dn.GetName(i);
			if(!Name) continue;

			Value = Dn.Get(i);
			if(!Value) continue;

			IndexDn.Add(Name, Value);
		}
	}

	if(!IndexDn.ValidateAgainstPolicy(m_Policies))
	{
		return false;
	}

	NEWPKI_PROFILE * newProfile;
	newProfile = NEWPKI_PROFILE_new();
	if(!newProfile)
	{
		return false;
	}
	ASN1_STRING_SET(&newProfile->ldap_uid, ASN1_STRING_GET(Result->uid));
	if(!IndexDn.To_X509_NAME(newProfile->dn))
	{
		NEWPKI_PROFILE_free(newProfile);
		return false;
	}
	ASN1_INTEGER_SET(&newProfile->EE_Validation, 0);
	ASN1_INTEGER_SET(&newProfile->owner_group_serial, m_group_id);

	if(m_event->OnNewProfile(newProfile))
	{
		if(m_Logging)
		{
			m_Logging->LogMessage(LOG_STATUS_TYPE_SUCCESS, LOG_MESSAGE_TYPE_LDAP_AUTO_SYNCH_IMPRT, 0, m_ldap_username.c_str(), LOG_NO_OBJECTID, ASN1_STRING_GET(newProfile->ldap_uid));
		}
	}
	else
	{
		NEWPKI_PROFILE_free(newProfile);
		return false;
	}
	NEWPKI_PROFILE_free(newProfile);

	return true;
}

void AutoSynchLDAP::SetLogging(EntityLog *Logging)
{
	m_Logging = Logging;
}
