/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/



// CA_Handler.cpp: implementation of the CA_Handler class.
//
//////////////////////////////////////////////////////////////////////

#include "CA_Handler.h"


ASN1_SEQUENCE(INTERNAL_CA_CERT_SIG) = {
	ASN1_SIMPLE(INTERNAL_CA_CERT_SIG, signature, ASN1_BIT_STRING),
	ASN1_SIMPLE(INTERNAL_CA_CERT_SIG, sig_alg, X509_ALGOR),
}ASN1_SEQUENCE_END(INTERNAL_CA_CERT_SIG)
IMPLEMENT_ASN1_FUNCTIONS(INTERNAL_CA_CERT_SIG)
IMPLEMENT_PEM_CONVERT(INTERNAL_CA_CERT_SIG, "NEWPKI CERT SIGNATURE");


ASN1_SEQUENCE(INTERNAL_CA_BODY) = {
	ASN1_SIMPLE(INTERNAL_CA_BODY, ca_serial, ASN1_INTEGER),
	ASN1_SIMPLE(INTERNAL_CA_BODY, ca_type, ASN1_INTEGER),
	ASN1_EXP_OPT(INTERNAL_CA_BODY, lastCrl, X509_CRL, 0),
}ASN1_SEQUENCE_END(INTERNAL_CA_BODY)
IMPLEMENT_ASN1_FUNCTIONS(INTERNAL_CA_BODY)
IMPLEMENT_ASN1_ENCRYPT_SIGN(INTERNAL_CA_BODY)

ASN1_CHOICE(INTERNAL_CA_KEY) = {
	ASN1_EXP(INTERNAL_CA_KEY, d.privkey, RSAPrivateKey, INTERNAL_CA_KEY_TYPE_KEY),
	ASN1_EXP(INTERNAL_CA_KEY, d.keyid, ASN1_UTF8STRING, INTERNAL_CA_KEY_TYPE_ENGINE),
} ASN1_CHOICE_END(INTERNAL_CA_KEY)
IMPLEMENT_ASN1_FUNCTIONS(INTERNAL_CA_KEY)
IMPLEMENT_ASN1_CPY_FUNCTIONS(INTERNAL_CA_KEY)

ASN1_SEQUENCE(INTERNAL_CA) = {
	ASN1_SIMPLE(INTERNAL_CA, ca_cert, X509),
	ASN1_SIMPLE(INTERNAL_CA, ca_key, INTERNAL_CA_KEY),
	ASN1_SIMPLE(INTERNAL_CA, crypted_body, ASN1_ENCRYPT_SIGN),
}ASN1_SEQUENCE_END(INTERNAL_CA)
IMPLEMENT_ASN1_FUNCTIONS(INTERNAL_CA)
IMPLEMENT_PEM_CONVERT(INTERNAL_CA, "NEWPKI INTERNAL CA")




//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CA_Handler::CA_Handler(int CaType)
{
	m_InternCaBody = NULL;
	m_InternCa = NULL;
	m_DbConn = NULL;
	m_CaType = CaType;
}

CA_Handler::~CA_Handler()
{
	if(m_DbConn) delete m_DbConn;
	Reset();
}

bool CA_Handler::CreateTables(SQL_Connection * DbConn)
{
	SQL sql(DbConn);
	long i;
	char * CommonCreates[] = {CA_HANDLER_CREATE_1, CA_HANDLER_CREATE_2, CA_HANDLER_CREATE_3, NULL};

	//We execute each request
	for(i=0; CommonCreates[i]; i++)
	{
		if(!sql.Execute(CommonCreates[i]))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}

	return true;
}

PKI_CERT * CA_Handler::get_CaCert()
{
	if(!*this)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_CA_NOT_CONFIGURED);
		return NULL;
	}
	return &m_CaCert;
}

bool CA_Handler::Create(const PKI_CERT & cert, const INTERNAL_CA_KEY * privkey)
{
	Reset();

	m_CaKey = cert.GetPrivateKey();
	
	if(!m_CaKey)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	
	m_InternCa = INTERNAL_CA_new();
	if(!m_InternCa)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	
	m_InternCaBody = INTERNAL_CA_BODY_new();
	if(!m_InternCaBody)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	
	if(!INTERNAL_CA_KEY_cpy(&m_InternCa->ca_key, privkey))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	if(!X509_cpy(&m_InternCa->ca_cert, cert.GetX509()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	ASN1_INTEGER_SET(&m_InternCaBody->ca_type, m_CaType);
	ASN1_INTEGER_SET(&m_InternCaBody->ca_serial, 1);

	//Flush new datas into database	
	if(!Flush())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//Load datas
	if(!Load())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool CA_Handler::Flush()
{
	SQL sql(m_DbConn);
	mString req;
	char * ca_pem;
	
	if(!m_InternCaBody || !m_InternCa)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}

	//Encrypt the body and generate signature
	if( INTERNAL_CA_BODY_sign_encrypt(m_InternCa->crypted_body, m_InternCaBody, m_CaKey.GetRsaKey(), m_CaKey.GetRsaKey(), EVP_sha1(), EVP_des_ede3_cbc()) <= 0) 
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	//Convert to PEM	
	ca_pem = INTERNAL_CA_To_PEM(m_InternCa);
	if(!ca_pem)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		CleanMemory();
		return false;
	}
	//Save some memory
	CleanMemory();
	
	if(req.sprintf(CA_HANDLER_INSERT_CA, m_CaType, ca_pem) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		free(ca_pem);
		return false;
	}
	free(ca_pem);

	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	return true;
}

bool CA_Handler::Load()
{
	Reset();
	long NumRows;
	
	if(!m_DbConn)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}
	
	//We load the CA datas	
	SQL sql(m_DbConn);
	mString req;
	mString datas;

	if(req.sprintf(CA_HANDLER_GET_CA, m_CaType) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}
	
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	// It can be loaded latter
	if(!NumRows)
	{
		return true;
	}
	
	if(!sql.Value(0, "datas", datas))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	//We convert the datas from the PEM blob
	if(!INTERNAL_CA_From_PEM(&m_InternCa, datas.c_str()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}
	
	//Setting the CA cert
	if(!m_CaCert.SetCert(m_InternCa->ca_cert))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//Setting the CA key
	if(!INTERNAL_CA_KEY_load(m_InternCa->ca_key, m_CaKey, m_Engine))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	//Associate private key and cert
	if(!m_CaCert.SetPrivateKey(m_CaKey))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}



	if(INTERNAL_CA_BODY_decrypt_verify(m_InternCa->crypted_body, &m_InternCaBody, m_CaKey.GetRsaKey(), m_CaKey.GetRsaKey()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
		
	//Now we verify that the two ca_types match
	if(ASN1_INTEGER_GET(m_InternCaBody->ca_type) != m_CaType)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}

	return true;
}


void CA_Handler::Reset()
{
	CaLock.Lock();
	if(m_InternCaBody)
	{
		INTERNAL_CA_BODY_free(m_InternCaBody);
		m_InternCaBody = NULL;
	}
	if(m_InternCa)
	{
		INTERNAL_CA_free(m_InternCa);
		m_InternCa = NULL;
	}
	CaLock.Unlock();
}

void CA_Handler::set_ENGINE(ENGINE *e)
{
	CaLock.Lock();
	m_Engine = e;
	CaLock.Unlock();
}

ENGINE * CA_Handler::get_ENGINE()
{
	return m_Engine;
}

bool CA_Handler::Sign(const PKI_CSR & csr, PKI_CERT & ResultCert, int Days, const HashTable_String *Exts, bool check_sig)
{
	INTERNAL_CA_CERT * currCert;
	time_t currTime;
	X509_REQ * reqcsr;
	char * strDn;	
	mString dn;
	time_t begin_date;
	time_t end_date;
	unsigned long serial;
	
	SQL sql(m_DbConn);
	mString req;
	long NumRows;
	char * signature;
	
	if(!*this)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_CA_NOT_CONFIGURED);
		return false;
	}
	
	//Search if we have a certificate with same DN !
	reqcsr = csr.GetX509_REQ();
	time_gmt(&currTime);

	strDn = X509_NAME_oneline(reqcsr->req_info->subject, NULL, 0);
	if(!strDn)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	dn = sql.FormatString(strDn);
	if(!dn.size())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		free(strDn);
		return false;
	}
	

	if(req.sprintf(CA_HANDLER_GET_CERT_UNIQUE, dn.c_str(), CERT_STATE_VALID, currTime, m_CaType) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		free(strDn);
		return false;
	}
	
	CaLock.Lock();
	
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		CaLock.Unlock();
		free(strDn);
		return false;
	}
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		CaLock.Unlock();
		free(strDn);
		return false;
	}
	if(NumRows)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_CERT_ALREADY_SIGNED_DN);
		CaLock.Unlock();
		free(strDn);
		return false;
	}


	
	serial = ASN1_INTEGER_GET(m_InternCaBody->ca_serial);
	//We sign the request
	if(!m_CaCert.SignCSR(ResultCert, csr, Exts, Days, serial, "sha1", check_sig))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		CaLock.Unlock();
		free(strDn);
		return false;
	}
	
	currCert = INTERNAL_CA_CERT_new();
	if(!currCert)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		CaLock.Unlock();
		free(strDn);
		return false;
	}
	
	
	
	//Preparing signature
	
	begin_date = ResultCert.GetStartDate();
	end_date = ResultCert.GetEndDate();

	ASN1_INTEGER_SET(&currCert->ca_type, m_CaType);
	ASN1_INTEGER_SET(&currCert->serial, serial);
	ASN1_INTEGER_SET(&currCert->state, CERT_STATE_VALID);
	ASN1_STRING_SET(&currCert->dn, strDn);
	ASN1_INTEGER_SET(&currCert->begin_date, begin_date);
	ASN1_INTEGER_SET(&currCert->end_date, end_date);
	ASN1_INTEGER_SET(&currCert->rev_date, 0);
	free(strDn);
	

	if(!X509_cpy(&currCert->cert, ResultCert.GetX509()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		INTERNAL_CA_CERT_free(currCert);
		CaLock.Unlock();
		return false;
	}

	signature = get_InternalCertSig(currCert);
	if(!signature)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		INTERNAL_CA_CERT_free(currCert);
		CaLock.Unlock();
		return false;
	}
	//Not needed anymore
	INTERNAL_CA_CERT_free(currCert);

	 
	//Ok let's do the insertion in database
	if(req.sprintf(CA_HANDLER_INSERT_CERT, m_CaType, serial, CERT_STATE_VALID, dn.c_str(), begin_date, end_date, ResultCert.GetCertPEM().c_str(), signature) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		INTERNAL_CA_CERT_free(currCert);
		free(signature);
		CaLock.Unlock();
		return false;
	}
	dn = "";
	free(signature);
	

	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		CaLock.Unlock();
		return false;
	}
	

	//Increment serial
	serial++;
	ASN1_INTEGER_SET(&m_InternCaBody->ca_serial, serial);
	Flush();

	CaLock.Unlock();
	return true;
}

CA_Handler::operator int()
{
	if(!m_InternCaBody || !m_CaCert || !m_CaKey || !m_DbConn)
		return 0;
	else
		return 1;
}

void CA_Handler::CleanMemory()
{
	if(!m_InternCa) return;
	CaLock.Lock();
	ASN1_OCTET_STRING_set(m_InternCa->crypted_body->crypteddatas, (unsigned char*)"", 1);
	ASN1_OCTET_STRING_set(m_InternCa->crypted_body->sessionkey, (unsigned char*)"", 1);
	ASN1_OCTET_STRING_set(m_InternCa->crypted_body->sig->digest, (unsigned char*)"", 1);
	CaLock.Unlock();
}

STACK_OF(INTERNAL_CA_CERT) * CA_Handler::get_Certs(CERT_STATE state, long index, long num)
{
	if(!*this)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_CA_NOT_CONFIGURED);
		return NULL;
	}

	STACK_OF(INTERNAL_CA_CERT) * resStack;	
	INTERNAL_CA_CERT * res;
	long NumRows;
	SQL sql(m_DbConn);
	mString req;
	long i;

	
	if(state)
	{
		if(req.sprintf(CA_HANDLER_GET_CERTS_BY_STATE, state, m_CaType) <= 0)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			return NULL;
		}
	}
	else
	{
		if(req.sprintf(CA_HANDLER_GET_CERTS, m_CaType) <= 0)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			return NULL;
		}
	}
	if(num)
	{
		req += " LIMIT ";
		req += index;
		req += ",";
		req += num;
	}
	req += ";";
	
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	
	resStack = SKM_sk_new_null(INTERNAL_CA_CERT);
	if(!resStack)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return NULL;
	}
	
	for(i=0; i<NumRows; i++)
	{
		//Convert the sql result to the internal struct	
		res = SQL2InternalCert(&sql, i);
		if(!res)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			SKM_sk_pop_free(INTERNAL_CA_CERT, resStack, INTERNAL_CA_CERT_free);
			return NULL;
		}
		if(SKM_sk_push(INTERNAL_CA_CERT, resStack, res) < 0)
		{
			INTERNAL_CA_CERT_free(res);
		}
	}
	
	return resStack;
}

STACK_OF(X509_CRL) * CA_Handler::get_Crls(long index, long num)
{
	if(!*this)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_CA_NOT_CONFIGURED);
		return NULL;
	}

	STACK_OF(X509_CRL) * resStack;	
	mString crl_pem;
	long NumRows;
	SQL sql(m_DbConn);
	mString req;
	long i;
	PKI_CRL crl;
	X509_CRL * dupCrl;

	
	if(req.sprintf(CA_HANDLER_GET_CRLS, m_CaType) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return NULL;
	}
	if(num)
	{
		req += " LIMIT ";
		req += index;
		req += ",";
		req += num;
	}
	req += ";";
	
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	
	resStack = SKM_sk_new_null(X509_CRL);
	if(!resStack)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return NULL;
	}
	
	for(i=0; i<NumRows; i++)
	{
		//Convert the sql result to the internal struct	
		if(!sql.Value(i, "crl", crl_pem))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return NULL;
		}
		if(!crl.SetCRL((char*)crl_pem.c_str())) continue;
		if(! (dupCrl = crl.GetX509_CRL(true)) ) continue;

		if(SKM_sk_push(X509_CRL, resStack, dupCrl) < 0)
		{
			X509_CRL_free(dupCrl);
		}
	}
	
	return resStack;
}

INTERNAL_CA_CERT * CA_Handler::get_Cert(unsigned long serial)
{
	if(!*this)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_CA_NOT_CONFIGURED);
		return NULL;
	}
	
	INTERNAL_CA_CERT * res;
	long NumRows;
	SQL sql(m_DbConn);
	mString req;
	

	//Get cert
	if(req.sprintf(CA_HANDLER_GET_CERT, serial, m_CaType) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return NULL;
	}
	
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!NumRows)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NO_CERT);
		return NULL;
	}

	//Convert the sql result to the internal struct	
	res = SQL2InternalCert(&sql, 0);
	if(!res)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	
	return res;
}

bool CA_Handler::Revoke(unsigned long serial, PKI_CERT & Cert, CERT_STATE & CertStatus, time_t & rev_date)
{
	INTERNAL_CA_CERT * currCert;
	char * signature;
	mString req;
	int tmp_enum;
	time_t revdate;

	if(!*this)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_CA_NOT_CONFIGURED);
		CertStatus = CERT_STATE_UNKNOWN;
		return false;
	}
	
	//Might be a, initial CRL generation
	if(!serial)
	{
		CertStatus = CERT_STATE_UNKNOWN;
		return true;
	}
	
	currCert = get_Cert(serial);
	if(!currCert)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		CertStatus = CERT_STATE_UNKNOWN;
		return false;
	}

	if(!Cert.SetCert(currCert->cert))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		INTERNAL_CA_CERT_free(currCert);
		CertStatus = CERT_STATE_UNKNOWN;
		return false;
	}
	
	//We verify that the certificate hasn't already been revoked
	if(ASN1_INTEGER_GET(currCert->state) != CERT_STATE_VALID)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_CERT_STATE);
		CertStatus = (CERT_STATE)ASN1_INTEGER_GET(currCert->state);
		INTERNAL_CA_CERT_free(currCert);
		return false;
	}
	
	//We update its status
	ASN1_INTEGER_SET(&currCert->state, CERT_STATE_REVOKED);
	time_gmt(&revdate);
	rev_date = revdate;
	ASN1_INTEGER_SET(&currCert->rev_date, rev_date);
	
	signature = get_InternalCertSig(currCert);
	if(!signature)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		INTERNAL_CA_CERT_free(currCert);
		CertStatus = CERT_STATE_VALID;
		return false;
	}
	
	SQL sql(m_DbConn);
		
	//We update it in the database	
	tmp_enum = (int)CERT_STATE_REVOKED;
	if(req.sprintf(CA_HANDLER_REV_CERT, tmp_enum, signature, rev_date, serial, m_CaType) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		INTERNAL_CA_CERT_free(currCert);
		free(signature);
		CertStatus = CERT_STATE_VALID;
		return false;
	}
	free(signature);
	INTERNAL_CA_CERT_free(currCert);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		CertStatus = CERT_STATE_VALID;
		return false;
	}
	CertStatus = CERT_STATE_REVOKED;
	return true;
}

bool CA_Handler::Generate_CRL(PKI_CRL & resCrl, int Days, int Hours, const HashTable_String * Exts, const char * md)
{
	mString tmpdatas;
	mString rev_date;
	REVOCATION_INFO rev_inf;
	SQL sql(m_DbConn);
	long NumRows;
	long i;
	mString req;

	if(!*this)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_CA_NOT_CONFIGURED);
		return false;
	}
	if((!Days && !Hours) || ! md)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	
	resCrl.Clear();

	
	//Get revoked certs
	if(req.sprintf(CA_HANDLER_GET_CERT_BY_STATE, CERT_STATE_REVOKED, m_CaType) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	//Add them to the CRL
	for(i=0; i < NumRows; i++)
	{
		if(!sql.Value(i, "serial", tmpdatas))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		rev_inf.serial = tmpdatas.c_ulng();

		if(!sql.Value(i, "rev_date", tmpdatas))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		rev_inf.rev_date = tmpdatas.c_ulng();
		
		resCrl.AddRevokedCert(&rev_inf);
	}

	CaLock.Lock();
	if(!resCrl.Generate(m_CaCert, md, (Days * 24) + Hours, Exts))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		CaLock.Unlock();
		return false;
	}
	//Save last crl
	if(!X509_CRL_cpy(&m_InternCaBody->lastCrl, resCrl.GetX509_CRL()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		CaLock.Unlock();
		return false;
	}
	if(!Flush())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		CaLock.Unlock();
		return false;
	}
	CaLock.Unlock();
	
	
	//We insert the new CRL in the database
	if(req.sprintf(CA_HANDLER_INSERT_CRL, m_CaType, resCrl.GetPemCRL().c_str()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

char * CA_Handler::get_InternalCertSig(INTERNAL_CA_CERT * cert)
{
	INTERNAL_CA_CERT_SIG * sig;
	const EVP_PKEY * privateKey;
	char * sig_pem;
	
	privateKey = m_CaKey.GetRsaKey();
	if(!privateKey)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return NULL;
	}
		
	sig = INTERNAL_CA_CERT_SIG_new();
	if(!sig)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return NULL;
	}
	
	//We calculate the cert signature
	if(INTERNAL_CA_CERT_sign(cert, sig, (EVP_PKEY*)privateKey) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		INTERNAL_CA_CERT_SIG_free(sig);
		return NULL;
	}
	
	//We convert the signature to PEM
	sig_pem = INTERNAL_CA_CERT_SIG_To_PEM(sig);
	if(!sig_pem)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		INTERNAL_CA_CERT_SIG_free(sig);
		return NULL;
	}
	INTERNAL_CA_CERT_SIG_free(sig);
	
	return sig_pem;
}

bool CA_Handler::verify_InternalCertSig(INTERNAL_CA_CERT * cert, char * sig_pem)
{
	INTERNAL_CA_CERT_SIG * sig = NULL;
	const EVP_PKEY * privateKey;
	privateKey = m_CaKey.GetRsaKey();
	
	if(!privateKey)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}
		
	//We convert the signature from PEM
	if(INTERNAL_CA_CERT_SIG_From_PEM(&sig, sig_pem) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	//We verify the cert signature
	if(INTERNAL_CA_CERT_verify(cert, sig, (EVP_PKEY *)privateKey) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		INTERNAL_CA_CERT_SIG_free(sig);
		return false;
	}
	
	INTERNAL_CA_CERT_SIG_free(sig);
	
	return true;
}



INTERNAL_CA_CERT * CA_Handler::SQL2InternalCert(SQL * sql, long i)
{
	INTERNAL_CA_CERT * res;
	mString strdatas;
	mString dn;
	PKI_CERT cert;
	
	res = INTERNAL_CA_CERT_new();
	if(!res)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return NULL;
	}
		
	if(!sql->Value(i, "ca_type", strdatas))
	{
		INTERNAL_CA_CERT_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!strdatas.c_ASN1_INTEGER(&res->ca_type))
	{
		INTERNAL_CA_CERT_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return NULL;
	}

	if(!sql->Value(i, "serial", strdatas))
	{
		INTERNAL_CA_CERT_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!strdatas.c_ASN1_INTEGER(&res->serial))
	{
		INTERNAL_CA_CERT_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return NULL;
	}

	if(!sql->Value(i, "state", strdatas))
	{
		INTERNAL_CA_CERT_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!strdatas.c_ASN1_INTEGER(&res->state))
	{
		INTERNAL_CA_CERT_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return NULL;
	}

	if(!sql->Value(i, "dn", strdatas))
	{
		INTERNAL_CA_CERT_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!strdatas.c_ASN1_UTF8STRING(&res->dn))
	{
		INTERNAL_CA_CERT_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return NULL;
	}

	if(!sql->Value(i, "begin_date", strdatas))
	{
		INTERNAL_CA_CERT_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!strdatas.c_ASN1_INTEGER(&res->begin_date))
	{
		INTERNAL_CA_CERT_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return NULL;
	}

	if(!sql->Value(i, "end_date", strdatas))
	{
		INTERNAL_CA_CERT_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!strdatas.c_ASN1_INTEGER(&res->end_date))
	{
		INTERNAL_CA_CERT_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return NULL;
	}
	
	if(!sql->Value(i, "rev_date", strdatas))
	{
		INTERNAL_CA_CERT_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!strdatas.c_ASN1_INTEGER(&res->rev_date))
	{
		INTERNAL_CA_CERT_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return NULL;
	}
	
	if(!sql->Value(i, "cert", strdatas))
	{
		INTERNAL_CA_CERT_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!cert.SetCert(strdatas.c_str()))
	{
		INTERNAL_CA_CERT_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!X509_cpy(&res->cert, cert.GetX509()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		INTERNAL_CA_CERT_free(res);
		return NULL;
	}

	//now we verify the signature of the thing !
	if(!sql->Value(i, "signature", strdatas))
	{
		INTERNAL_CA_CERT_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!verify_InternalCertSig(res, (char*)strdatas.c_str()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		INTERNAL_CA_CERT_free(res);
		return NULL;
	}
	
	return res;
}


bool CA_Handler::GEN_PRIVATE_KEY_load(const GEN_PRIVATE_KEY * KeyGen, PKI_RSA & Key, INTERNAL_CA_KEY *InternalKey, ENGINE * e)
{
	char * keyid;
	int rsalen;

	switch(KeyGen->type)
	{
		case GEN_PRIVATE_KEY_TYPE_ENGINE:
			keyid = ASN1_STRING_GET(KeyGen->d.keyid);
			if(!keyid)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				return false;
			}
			if(!Key.SetKey(keyid, e))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				return false;
			}
			InternalKey->type = INTERNAL_CA_KEY_TYPE_ENGINE;
			ASN1_STRING_SET(&InternalKey->d.keyid, keyid);
			break;
	
		case GEN_PRIVATE_KEY_TYPE_KEYLEN:
			rsalen = (int)ASN1_INTEGER_GET(KeyGen->d.keylen);
			if(!rsalen)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				return false;
			}
			if(!Key.GenerateKey(rsalen, e))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				return false;
			}
			InternalKey->type = INTERNAL_CA_KEY_TYPE_KEY;
			InternalKey->d.privkey = Key.GetRSA(true);
			break;

		default:
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
			return false;
			break;
	}

	return true;
}

bool CA_Handler::INTERNAL_CA_KEY_load(const INTERNAL_CA_KEY *ca_key, PKI_RSA & key, ENGINE * e)
{
	switch(ca_key->type)
	{
		case INTERNAL_CA_KEY_TYPE_KEY:
			if(!key.SetKey(ca_key->d.privkey))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				return false;
			}
			break;
		case INTERNAL_CA_KEY_TYPE_ENGINE:
			if(!key.SetKey(ASN1_STRING_GET(ca_key->d.keyid), e))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				return false;
			}
			break;
		default:
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			return false;
	}
	return true;
}

bool CA_Handler::SetDbConn(const SQL_Connection *DbConn)
{
	if(m_DbConn)
	{
		delete m_DbConn;
		m_DbConn = NULL;
	}

	try
	{
		m_DbConn = DbConn->Clone();
	}
	catch(ExceptionNewPKI e)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!m_DbConn)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	return true;
}

bool CA_Handler::GetLastCrl(PKI_CRL & crl)
{
	bool res;
	
	if(!*this)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_CA_NOT_CONFIGURED);
		return false;
	}

	CaLock.Lock();
	if(m_InternCaBody->lastCrl)
	{
		res = crl.SetCRL(m_InternCaBody->lastCrl);
	}
	else
	{
		res = false;
	}
	CaLock.Unlock();
	return res;
}

