/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/



// CA_Handler.h: interface for the CA_Handler class.
//
//////////////////////////////////////////////////////////////////////

#ifndef CA_HANDLER_H
#define CA_HANDLER_H

#include <PKI_CERT.h>
#include <PKI_CRL.h>
#include <CriticalSection.h>
#include <SQL/SQL.h>
#include <SQL/SQL_CMD.h>
#include <ASN1/ASN1_resp.h>


#ifndef DOXYGEN_SHOULD_SKIP_THIS
typedef struct st_INTERNAL_CA_BODY
{
	ASN1_INTEGER * ca_type;
	ASN1_INTEGER * ca_serial;
	X509_CRL * lastCrl;
}INTERNAL_CA_BODY;
DECLARE_ASN1_FUNCTIONS(INTERNAL_CA_BODY)
DECLARE_ASN1_ENCRYPT_SIGN(INTERNAL_CA_BODY)

typedef struct st_INTERNAL_CA_CERT_SIG
{
	ASN1_BIT_STRING * signature;	//!< The signature
	X509_ALGOR * sig_alg;			//!< The signature algo
}INTERNAL_CA_CERT_SIG;
DECLARE_ASN1_FUNCTIONS(INTERNAL_CA_CERT_SIG)
DECLARE_PEM_CONVERT(INTERNAL_CA_CERT_SIG)

#define INTERNAL_CA_CERT_sign(d, x,pkey) \
	ASN1_sign((int (*)())i2d_INTERNAL_CA_CERT, x->sig_alg, NULL, x->signature, (char *)d, pkey, EVP_sha1())

#define INTERNAL_CA_CERT_verify(d, x,pkey) \
	ASN1_verify((int (*)())i2d_INTERNAL_CA_CERT,x->sig_alg, x->signature,(char *)d, pkey)

#endif

#define INTERNAL_CA_KEY_TYPE_KEY 0		//!< The private key is a software key
#define INTERNAL_CA_KEY_TYPE_ENGINE 1	//!< The private key is a key id stored in an engine

/*!
	This structure is the representation of a private key
*/
typedef struct st_INTERNAL_CA_KEY
{
	int type;	//!< The type of key
	union
	{
		RSA * privkey;				//!< A software key
		ASN1_UTF8STRING * keyid;	//!< A key Id
	}d;
}INTERNAL_CA_KEY;
DECLARE_ASN1_FUNCTIONS(INTERNAL_CA_KEY)
DECLARE_ASN1_CPY_FUNCTIONS(INTERNAL_CA_KEY)


#ifndef DOXYGEN_SHOULD_SKIP_THIS
typedef struct st_INTERNAL_CA
{
	X509 * ca_cert;
	INTERNAL_CA_KEY * ca_key;
	ASN1_ENCRYPT_SIGN * crypted_body;
}INTERNAL_CA;
DECLARE_ASN1_FUNCTIONS(INTERNAL_CA)
DECLARE_PEM_CONVERT(INTERNAL_CA)
#endif



/*!
	This class represents one or more CAs
*/
class CA_Handler  
{
public:	
	/*! \brief This is the constructor.
	 *  \param CaType [IN] Since there is only one table for all the CAs, CaType is used to know on which one we're working.
	 */
	CA_Handler(int CaType);

	/*! \brief This is the destructor.
	 */
	virtual ~CA_Handler();

	/*! \brief This function creates the tables needed to work.
	 *  \param DbConn [IN] A SQL connection.
	 *  \return true on success, false on failure.
	 */
	static bool CreateTables(SQL_Connection * DbConn);

	/*! \brief This function sets the SQL connection.
	 *  \param DbConn [IN] A SQL connection.
	 *  \return true on success, false on failure.
	 */
	bool SetDbConn(const SQL_Connection * DbConn);

	/*! \brief This function generate a private key.
	 *  \param KeyGen [IN] The private key info.
	 *  \param Key [OUT] The private key.
	 *  \param InternalKey [OUT] The private key, in a parsable format.
	 *  \param e [IN] An optional ENGINE.
	 *  \return true on success, false on failure.
	 */
	static bool GEN_PRIVATE_KEY_load(const GEN_PRIVATE_KEY * KeyGen, PKI_RSA & Key, INTERNAL_CA_KEY * InternalKey, ENGINE * e);

	/*! \brief This function loads a private key.
	 *  \param ca_key [IN] The private key, in a parsable format.
	 *  \param key [OUT] The private key.
	 *  \param e [IN] An optional ENGINE.
	 *  \return true on success, false on failure.
	 */
	static bool INTERNAL_CA_KEY_load(const INTERNAL_CA_KEY * ca_key, PKI_RSA & key, ENGINE * e);

	/*! \brief This function creates the CA initial datas.
	 *  \param cert [IN] The CA certificate.
	 *  \param privkey [IN] The CA private key.
	 *  \return true on success, false on failure.
	 */
	bool Create(const PKI_CERT & cert, const INTERNAL_CA_KEY * privkey);

	/*! \brief This function loads the intial datas.
	 *  \return true on success, false on failure.
	 */
	bool Load();
	
	/*! \brief This function returns the CA certificate.
	 *  \return The CA certificate, or NULL if CA hasn't been intializaed yet.
	 */
	PKI_CERT * get_CaCert();
	
	/*! \brief This function returns the ENGINE.
	 *  \return The ENGINE, or NULL if ENGINE wasn't set.
	 */
	ENGINE * get_ENGINE();

	/*! \brief This function sets the ENGINE.
	 *  \param e [IN] The ENGINE.
	 */
	void set_ENGINE(ENGINE * e);
	
	/*! \brief This function returns the list of certificates.
	 *  \param state [IN] The state of certificates to list.
	 *  \param index [IN] The index for the enumeration.
	 *  \param num [IN] The maximum number of entries to return.
	 *  \return The list of certificates, or NULL on failure.
	 */
	STACK_OF(INTERNAL_CA_CERT) * get_Certs(CERT_STATE state=(CERT_STATE)0, long index=0, long num=0);

	/*! \brief This function returns the list of CRLs.
	 *  \param index [IN] The index for the enumeration.
	 *  \param num [IN] The maximum number of entries to return.
	 *  \return The list of CRLs, or NULL on failure.
	 */
	STACK_OF(X509_CRL) * get_Crls(long index, long num);
	
	/*! \brief This function signs a CSR.
	 *  \param csr [IN] The certificate request.
	 *  \param ResultCert [OUT] The resulting certificate.
	 *  \param Days [IN] The number of days for the certificate validity.
	 *  \param Exts [IN] The available extensions for the certificate.
	 *  \param check_sig [IN] Whether to check the CSR signature.
	 *  \return true on success, false on failure.
	 */
	bool Sign(const PKI_CSR & csr, PKI_CERT & ResultCert, int Days, const HashTable_String *Exts, bool check_sig);

	/*! \brief This function revokes a certificate.
	 *  \param serial [IN] The serial of the certificate to revoke.
	 *  \param Cert [OUT] The revoked certificate.
	 *  \param CertStatus [OUT] The status of the certificate.
	 *  \param rev_date [OUT] The revocation date.
	 *  \return true on success, false on failure.
	 */
	bool Revoke(unsigned long serial, PKI_CERT & Cert, CERT_STATE & CertStatus, time_t & rev_date);

	/*! \brief This function generates a CRL.
	 *  \param resCrl [OUT] The new CRL.
	 *  \param Days [IN] The number of days for validity.
	 *  \param Hours [IN] The number of hours for validity.
	 *  \param Exts [IN] The extensions to include to the CRL.
	 *  \param md [IN] The hash algorithm to use for the signature.
	 *  \return true on success, false on failure.
	 */
	bool Generate_CRL(PKI_CRL & resCrl, int Days, int Hours, const HashTable_String * Exts, const char * md);

	/*! \brief This function retieves the last CRL.
	 *  \param crl [OUT] The last CRL.
	 *  \return true on success, false on failure.
	 */
	bool GetLastCrl(PKI_CRL & crl);

	/*! \brief This function return 1 when the CA is operable and 0 when not.
	 *  \return return 1 when the CA is operable and 0 when not.
	 */
	operator int();
private:
	#define CA_HANDLER_CREATE_1				"create table internal_cas (ca_type INT UNSIGNED NOT NULL PRIMARY KEY, datas LONGBLOB NOT NULL);"
	#define CA_HANDLER_CREATE_2				"create table internal_certs (unique_id INT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY, ca_type INT UNSIGNED NOT NULL, serial INT UNSIGNED NOT NULL, state INT UNSIGNED NOT NULL, dn BLOB NOT NULL, begin_date INT UNSIGNED NOT NULL, end_date INT UNSIGNED NOT NULL, rev_date INT UNSIGNED NOT NULL, pubkey LONGBLOB NOT NULL, cert LONGBLOB NOT NULL, signature LONGBLOB NOT NULL);"
	#define CA_HANDLER_CREATE_3				"create table internal_crls (unique_id INT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY, ca_type INT UNSIGNED NOT NULL, crl LONGBLOB NOT NULL);"

	#define CA_HANDLER_INSERT_CA			"REPLACE INTO internal_cas (ca_type, datas) VALUES (%d, '%s');"
	#define CA_HANDLER_GET_CA				"SELECT * FROM internal_cas WHERE ca_type=%d;"
	
	#define CA_HANDLER_INSERT_CERT			"INSERT INTO internal_certs (ca_type, serial, state, dn, begin_date, end_date, rev_date, cert, signature) VALUES (%d, %ld, %d, '%s', %ld, %ld, 0, '%s', '%s');"
	#define CA_HANDLER_GET_CERT				"SELECT * FROM internal_certs WHERE serial=%ld AND ca_type=%d;"
	#define CA_HANDLER_GET_CERT_UNIQUE		"SELECT serial FROM internal_certs WHERE dn='%s' AND state=%d AND end_date > %ld AND ca_type=%d;"
	#define CA_HANDLER_GET_CERT_BY_STATE	"SELECT serial, rev_date FROM internal_certs WHERE state=%d AND ca_type=%d;"
	#define CA_HANDLER_REV_CERT				"UPDATE internal_certs SET state=%d, signature='%s', rev_date=%ld WHERE serial=%ld AND ca_type=%d;"
	#define CA_HANDLER_GET_CERTS			"SELECT * FROM internal_certs WHERE ca_type=%d ORDER BY begin_date DESC"
	#define CA_HANDLER_GET_CERTS_BY_STATE	"SELECT * FROM internal_certs WHERE state=%d AND ca_type=%d ORDER BY begin_date DESC"
	
	#define CA_HANDLER_INSERT_CRL			"INSERT INTO internal_crls (ca_type, crl) VALUES (%d, '%s');"
	#define CA_HANDLER_GET_CRLS				"SELECT * FROM internal_crls WHERE ca_type=%d ORDER BY unique_id DESC"

	bool Flush();
	void Reset();
	void CleanMemory();
	char * get_InternalCertSig(INTERNAL_CA_CERT * cert);
	bool verify_InternalCertSig(INTERNAL_CA_CERT * cert, char * sig_pem);
	INTERNAL_CA_CERT * get_Cert(unsigned long serial);
	
	INTERNAL_CA_CERT * SQL2InternalCert(SQL * sql, long i);
	
	CriticalSection CaLock;
	
	PKI_CERT m_CaCert;
	PKI_RSA m_CaKey;
	INTERNAL_CA_BODY * m_InternCaBody;
	INTERNAL_CA * m_InternCa;
	ENGINE * m_Engine;
	SQL_Connection * m_DbConn;
	int m_CaType;
};

#endif
