/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/



// ClientLDAP.cpp: implementation of the ClientLDAP class.
//
//////////////////////////////////////////////////////////////////////

#include "ClientLDAP.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

#ifndef _WIN32
	#define stricmp strcasecmp
	#define strupr(x) \
	{ \
		size_t dtlen; \
		for(dtlen=0; dtlen<strlen(x); dtlen++) (x)[dtlen]=toupper((x)[dtlen]); \
	}
	#define strlwr(x) \
	{ \
		size_t dtlen; \
		for(dtlen=0; dtlen<strlen(x); dtlen++) (x)[dtlen]=tolower((x)[dtlen]); \
	}
#endif

ClientLDAP::ClientLDAP()
{
	m_Connection = NULL;
}

ClientLDAP::~ClientLDAP()
{
	Disconnect();
}

bool ClientLDAP::Connect(ASN1_UTF8STRING *ldap_server, ASN1_INTEGER *ldap_port, ASN1_UTF8STRING *ldap_username, ASN1_UTF8STRING *ldap_password, ASN1_UTF8STRING *ldap_base, ASN1_UTF8STRING *ldap_attr_name)
{
	m_ldap_server = ldap_server;
	m_ldap_username = ldap_username;
	m_ldap_password = ldap_password;
	m_ldap_base = ldap_base;
	m_ldap_port = ASN1_INTEGER_GET(ldap_port);
	m_ldap_attr_name = ldap_attr_name;

	return Reconnect();
}

void ClientLDAP::Disconnect()
{
	if(m_Connection)
	{
		ldap_unbind_s(m_Connection);
		m_Connection = NULL;
	}
}

STACK_OF(LDAP_RESULT) *	ClientLDAP::Search(const mString & SearchString, int MaxResults, int MaxTime)
{
	LDAPMessage * pMsg;
	LDAPMessage * currMsg;
	STACK_OF(LDAP_RESULT) * Result;

	LDAP_RC_TYPE LdapRet;
	
	// Try to reconnect
	if(!m_Connection && !Reconnect())
	{
		return NULL;
	}

	ldap_set_option(m_Connection, LDAP_OPT_SIZELIMIT, (void *)&MaxResults);
	ldap_set_option(m_Connection, LDAP_OPT_TIMELIMIT, (void *)&MaxTime);

	pMsg=NULL;
	if((LdapRet = ldap_search_s(m_Connection, (char*)m_ldap_base.c_str(), LDAP_SCOPE_SUBTREE, (char*)SearchString.c_str(), NULL, 0, &pMsg)) != LDAP_SUCCESS)
	{
		// Should we try to reconnect
		switch(LdapRet)
		{
			case LDAP_SERVER_DOWN:
				// Retry the search
				if(!Reconnect())
				{
					return NULL;
				}
				ldap_set_option(m_Connection, LDAP_OPT_SIZELIMIT, (void *)&MaxResults);
				ldap_set_option(m_Connection, LDAP_OPT_TIMELIMIT, (void *)&MaxTime);
				if((LdapRet = ldap_search_s(m_Connection, (char*)m_ldap_base.c_str(), LDAP_SCOPE_SUBTREE, (char*)SearchString.c_str(), NULL, 0, &pMsg)) != LDAP_SUCCESS)
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_LDAP);
					ERR_add_error_data(2, "ldap_search_ext_s : ", ldap_err2string(LdapRet));
					return NULL;
				}
				break;
			case LDAP_TIMELIMIT_EXCEEDED:
			case LDAP_SIZELIMIT_EXCEEDED:
				break;
			default:
				NEWPKIerr(PKI_ERROR_TXT, ERROR_LDAP);
				ERR_add_error_data(2, "ldap_search_ext_s : ", ldap_err2string(LdapRet));
				return NULL;
		}
	}


	Result = SKM_sk_new_null(LDAP_RESULT);
	if(!Result)
	{
		ldap_msgfree(pMsg);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return NULL;
	}

	for(currMsg = ldap_first_entry(m_Connection, pMsg); 
		currMsg; 
		currMsg = ldap_next_entry(m_Connection, currMsg))
	{
		AddCurrentMessage(Result, currMsg);
	}

	ldap_msgfree(pMsg);

	return Result;
}


void ClientLDAP::AddCurrentMessage(STACK_OF(LDAP_RESULT) * Result, LDAPMessage * currMsg)
{
	LDAP_RESULT * newResult;
	char * Name;
	char * attrName;
	char ** attrValue;
	BerElement* ptr;


	newResult = LDAP_RESULT_new();
	if(!newResult)
	{
		return;
	}

	// The DN
	Name = ldap_get_dn(m_Connection, currMsg);
	if(!Name)
	{
		return;
	}
	ASN1_STRING_SET(&newResult->rdn, Name);
	ldap_memfree(Name);


	// We add the attributes
	for(	attrName = ldap_first_attribute(m_Connection, currMsg, &ptr);
			attrName;
			attrName = ldap_next_attribute(m_Connection, currMsg, ptr)
		)
	{
		if(!strstr(attrName, ";binary"))
		{
			attrValue = ldap_get_values(m_Connection, currMsg, attrName);
			if(attrValue)
			{
				// Did we already find the attribute UID and is this one
				// the attribute UID ?
				if(!ASN1_STRING_GET(newResult->uid) && m_ldap_attr_name == attrName)
				{
					ASN1_STRING_SET(&newResult->uid, *attrValue);
				}
				AddObject(newResult->objects, attrName, *attrValue);
				ldap_value_free(attrValue);
			}
		}
		ldap_memfree(attrName);
	}

	if(SKM_sk_push(LDAP_RESULT, Result, newResult) < 0)
		LDAP_RESULT_free(newResult);
}

bool ClientLDAP::Reconnect()
{
	LDAP_RC_TYPE LdapRet;
	Disconnect();

	m_Connection = ldap_init((char*)m_ldap_server.c_str(), m_ldap_port);
	if(!m_Connection)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_LDAP);
		ERR_add_error_data(2, "ldap_init : ", ldap_err2string(LDAP_LAST_ERROR));
		return false;
	}

	if(m_ldap_username.size())
	{
		LdapRet = ldap_bind_s(m_Connection, (char*)m_ldap_username.c_str(), (char*)m_ldap_password.c_str(), LDAP_AUTH_SIMPLE);
	}
	else
	{
		LdapRet = ldap_bind_s(m_Connection, NULL, NULL, LDAP_AUTH_SIMPLE);
	}

	if(LdapRet != LDAP_SUCCESS)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_LDAP);
		ERR_add_error_data(2, "ldap_bind_s : ", ldap_err2string(LdapRet));
		Disconnect();
		return false;
	}
	return true;
}

ClientLDAP::operator int()
{
	if(!m_Connection)
		return 0;
	else
		return 1;
}

int ClientLDAP::GetNid(char *Obj)
{
	int nid;

	if(stricmp(Obj, "mail")==0) return NID_pkcs9_emailAddress;

	nid = OBJ_txt2nid(Obj);
	if(nid == NID_undef)
	{
		strupr((char*)Obj);
		nid = OBJ_txt2nid(Obj);
		if(nid == NID_undef)
		{
			strlwr((char*)Obj);
			((char*)Obj)[0] = toupper(Obj[0]);
			nid = OBJ_txt2nid(Obj);
		}
	}

	return nid;
}

void ClientLDAP::AddObject(STACK_OF(LDAP_OBJECT) * Objects, char * Name, char * Value)
{
	LDAP_OBJECT * newObject;
	int nid;

	nid = GetNid(Name);
	if(nid == NID_undef)
		return;

	Name = (char*)OBJ_nid2ln(nid);
	if(!Name)
		return;

	if(ObjectAlreadyKnown(Objects, Name, Value))
	{
		return;
	}

	newObject = LDAP_OBJECT_new();
	if(!newObject)
		return;

	ASN1_STRING_SET(&newObject->name, Name);
	if(!newObject->name)
	{
		LDAP_OBJECT_free(newObject);
		return;
	}
	ASN1_STRING_SET(&newObject->value, Value);
	if(!newObject->value)
	{
		LDAP_OBJECT_free(newObject);
		return;
	}

	if(SKM_sk_push(LDAP_OBJECT, Objects, newObject) < 0)
	{
		LDAP_OBJECT_free(newObject);
	}
}


bool ClientLDAP::ObjectAlreadyKnown(STACK_OF(LDAP_OBJECT) * Objects, char * Name, char * Value)
{
	LDAP_OBJECT * currObject;
	int i;
	char * currName;
	char * currValue;

	for(i=0; i<SKM_sk_num(LDAP_OBJECT, Objects); i++)
	{
		currObject = SKM_sk_value(LDAP_OBJECT, Objects, i);
		if(!currObject)
			continue;
		
		currName = ASN1_STRING_GET(currObject->name);
		currValue = ASN1_STRING_GET(currObject->value);
		if(!currName || !currValue)
			continue;

		if(strcmp(currName, Name) != 0)
			continue;
		if(strcmp(currValue, Value) != 0)
			continue;

		return true;
	}

	return false;
}

