/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/



// Entity.cpp: implementation of the Entity class.
//
//////////////////////////////////////////////////////////////////////

#include "Entity.h"
#include "svintl.h"
#include "MailInfo.h"


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Entity::Entity(ENTITY_CONSTRUCTOR_PARAMETERS)
{
	m_Logging = NULL;
	m_DbConn = NULL;
	entered_client = 0;
	EntityParentCerts = NULL;
	fullInit = false;
	m_Type = Type;

	
	EntityParentCerts = sk_X509_new_null();
	if(!EntityParentCerts)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		throw ExceptionNewPKI();
	}
	
	
	m_Conf = Conf;
	m_AuthModule = authModule;
	m_Engine = e;

	
	
	m_EntityName = EntityName;

	//Doing initial connection to server
	try
	{
		m_DbConn = new SQL_Connection("", m_Conf.get_SqlServer(), m_Conf.get_SqlPort(), m_Conf.get_SqlUsername(), m_Conf.get_SqlPassword());
	}
	catch(ExceptionNewPKI e)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		throw ExceptionNewPKI();
	}
	if(!m_DbConn)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		throw ExceptionNewPKI();
	}

	m_Jobs.SetEventHandler(this);
}

Entity::~Entity()
{
	m_Jobs.StopAll();
	if(m_Logging) delete m_Logging;
	if(m_DbConn) delete m_DbConn;

	sk_X509_pop_free(EntityParentCerts, X509_free);
}


bool Entity::EntityExists(bool &Exists)
{
	SQL sql(m_DbConn);
	mString db_name;
	long NumRows;
	long i;

	//We select the list of databases
	if(!sql.Execute(SELECT_DATABASE) )
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//We verify each of them
	for(i=0; i<NumRows; i++)
	{
		if(!sql.Value(i, SELECT_DATABASE_RES, db_name)) continue;
		
		if(db_name == m_EntityName)
		{
			Exists = true;
			return true;
		}
	}
	Exists = false;
	return true;
}

bool Entity::Private_CreateEntity(const GEN_PRIVATE_KEY* genkey, NewPKIStore * EntityStore, const char ** InitialCreates)
{
	SQL sql(m_DbConn);
	long i;
	char * CommonCreates[] = {COMMON_CREATE_1, COMMON_CREATE_2, COMMON_CREATE_3, COMMON_CREATE_4, NULL};
	bool Exists;
	const char * keyPem;
	mString req;

	//We check if there is another DB with same name
	if(!EntityExists(Exists))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(Exists)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ENTITY_EXISTS);
		return false;
	}

	//We create the database and select it
	if(!sql.CreateDatabase(m_EntityName))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//We execute each request
	for(i=0; CommonCreates[i]; i++)
	{
		if(!sql.Execute(CommonCreates[i]))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			Destroy();
			return false;
		}
	}

	if(InitialCreates)
	{
		for(i=0; InitialCreates[i]; i++)
		{
			if(!sql.Execute(InitialCreates[i]))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				Destroy();
				return false;
			}
		}
	}

	if(!m_Jobs.CreateTables(m_DbConn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}

	// The store
	if(EntityStore && !EntityStore->CreateTables(m_DbConn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}

	//Entity key
	if(genkey)
	{
		switch(genkey->type)
		{
			case GEN_PRIVATE_KEY_TYPE_ENGINE:
				keyPem = ASN1_STRING_GET(genkey->d.keyid);
				if(!m_EntityKey.SetKey(keyPem, m_Engine))
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					Destroy();
					return false;
				}
			case GEN_PRIVATE_KEY_TYPE_KEYLEN:
				if(!m_EntityKey.GenerateKey(ASN1_INTEGER_GET(genkey->d.keylen), m_Engine))
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					Destroy();
					return false;
				}
				keyPem = m_EntityKey.GetRsaKeyPem().c_str();
				break;
			default:
				NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
				Destroy();
				return false;
				break;
		}
		//Insert key in DB
		if(req.sprintf(ENTITY_SET_CERT, "", keyPem) <= 0)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
			Destroy();
			return false;
		}
		if(!sql.Execute(req))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			Destroy();
			return false;
		}
	}

	return true;
}


bool Entity::Private_InitEntity(const X509 *Cert, ENTITY_CONF ** CurrentConf, const ENTITY_SIGNATURE_RESP *init_datas, NewPKIStore * EntityStore)
{
	mString key;
	long NumRows;
	SQL sql(m_DbConn);
	mString req;

	// We get the entity key
	if(!sql.Execute(ENTITY_GET_CERT))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!NumRows)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}

	if(!sql.Value(0, "entity_key", key))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!m_EntityCert.SetCert(Cert))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!m_EntityKey.SetKey(key.c_str(), m_Engine))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!m_EntityCert.SetPrivateKey(m_EntityKey))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//We now insert the Entity cert in the database
	if(req.sprintf(ENTITY_SET_CERT, m_EntityCert.GetCertPEM().c_str(), key.c_str()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(CurrentConf && init_datas)
	{
		if(!ExtractMyConf(init_datas->conf, CurrentConf, init_datas->cas, init_datas->cas->PkiCert, true))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}

	if(EntityStore)
		EntityStore->SetEntityCert(m_EntityCert);

	return true;
}


mString & Entity::GetName()
{
	return m_EntityName;
}

bool Entity::Private_CommonLoad(NewPKIStore * Store, bool LoadConf)
{
	SQL sql(m_DbConn);
	mString cert;
	mString key;
	mString pem_conf;
	long NumRows;

	//We select the database
	if(!m_DbConn->SelectDatabase(m_EntityName))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(Store)
	{
		if(!Store->SetDbConn(m_DbConn))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}

	// We get the entity certificate
	if(!sql.Execute(ENTITY_GET_CERT))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!NumRows)
	{
		return true;
	}

	if(!sql.Value(0, "entity_cert", cert) || !sql.Value(0, "entity_key", key))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//Did we get our entity certificate yet ?
	if(!cert.size())
	{
		NewpkiDebug(LOG_LEVEL_INFO, m_EntityName.c_str(), _sv("Entity is not initialized"));
		return true;
	}


	if(!m_EntityCert.SetCert(cert.c_str()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!m_EntityKey.SetKey(key.c_str(), m_Engine))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!m_EntityCert.SetPrivateKey(m_EntityKey))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(Store)
		Store->SetEntityCert(m_EntityCert);


	if(m_Logging)
	{
		delete m_Logging;
	}

	try
	{
		m_Logging = new EntityLog(m_DbConn, &m_EntityCert);
	}
	catch(ExceptionNewPKI e)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!m_Logging)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}


	m_Logging->SetAuditCalllBack(OnAudit, this);


	const char *EntityMail;
	int Pos;

	Pos = m_EntityCert.GetCertDN().SeekEntryName("emailAddress", -1);
	if(Pos == HASHTABLE_NOT_FOUND)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		ERR_set_error_data(_sv("No emailAddress"), ERR_TXT_STRING);
		return false;
	}

	EntityMail = m_EntityCert.GetCertDN().Get(Pos);
	if(!EntityMail)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		ERR_set_error_data(_sv("No emailAddress"), ERR_TXT_STRING);
		return false;
	}

	m_EntityMail = EntityMail;

	if(!m_Jobs.SetDbConn(m_DbConn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	m_Jobs.SetEntityName(m_EntityName);
	m_Jobs.SetEntityCert(m_EntityCert);
	m_Jobs.SetLogger(m_Logging);
	m_Jobs.SetX509_ACL_Validator(&AclValidator);


	if(LoadConf)
	{
		// We get the personnal conf
		if(!sql.Execute(ENTITY_GET_CONF))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		if(!sql.NumRows(&NumRows))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		if(!NumRows)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			return false;
		}
		if(!sql.Value(0, "conf", pem_conf))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}

		// Last thing is to ask the entity to load its personnal conf
		if(!PersonnalConfFromPEM(pem_conf))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}

	// We have been initialized fully
	fullInit = true;

	return true;
}

void Entity::ClientEnter()
{
	ClientsReadLock.Lock();
		entered_client++;
	ClientsReadLock.Unlock();
}

void Entity::ClientLeave()
{
	ClientsReadLock.Lock();
		entered_client--;
	ClientsReadLock.Unlock();
}

void Entity::WaitForNoClients()
{
	ClientsReadLock.Lock();
		time_t start_t;
		time_t curr_t;
		time(&start_t);

		while(entered_client)
		{
			NewpkiThread::Sleep(10000);
			time(&curr_t);
			if( (curr_t-start_t) > 1200)
			{
				ClientsReadLock.Unlock();
				return;
			}
		}
	ClientsReadLock.Unlock();
}

bool Entity::Destroy()
{
	SQL sql(m_DbConn);

	//We select the list of databases
	if(!sql.DropDatabase(m_EntityName) )
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}


bool Entity::ADMIN_RESPONSE_init(ADMIN_RESPONSE_BODY * response, int Type, void * Datas)
{
	if(!Datas)
	{
		response->d.other = ASN1_NULL_new();
		if(!response->d.other)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
			return false;
		}
	}
	else
	{
		response->d.other = (ASN1_NULL*)Datas;
	}
	response->type = Type;
	
	return true;
}

void Entity::ERR_to_ADMIN_RESPONSE(ADMIN_RESPONSE_BODY * response)
{
	STACK_OF(ERROR_ENTRY) * errors = ERR_to_ERROR_ENTRIES();
	if(!errors)
	{
		return;
	}
	
	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_ERRORS, errors))
	{
		SKM_sk_pop_free(ERROR_ENTRY, errors, ERROR_ENTRY_free);
	}
}

bool Entity::ResourceLock(const char * ResourceName)
{
	ResourcesListLock.Lock();
	if(ResourcesList.find(ResourceName) != ResourcesList.end())
	{
		ResourcesListLock.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_RESOURCE_BUSY);
		return false;
	}
	ResourcesList[ResourceName] = NewpkiThread::CurrentThreadId();
	ResourcesListLock.Unlock();
	return true;
}

void Entity::ResourceUnlock(const char *ResourceName)
{
	std::map<mString, NEWPKI_THREAD_ID>::iterator rsc;
	ResourcesListLock.Lock();
	
	// Search resource lock
	rsc = ResourcesList.find(ResourceName);

	// Did we find the resource and am I the owner ?
	if(rsc == ResourcesList.end() ||
		rsc->second != NewpkiThread::CurrentThreadId())
	{
		ResourcesListLock.Unlock();
		return;
	}
	ResourcesList.erase(rsc);
	ResourcesListLock.Unlock();
}

void Entity::ResourcesClearUser()
{
	std::map<mString, NEWPKI_THREAD_ID>::iterator i;
	NEWPKI_THREAD_ID id = NewpkiThread::CurrentThreadId();

	ResourcesListLock.Lock();
LblResourcesClearUser:
	for(i=ResourcesList.begin(); i != ResourcesList.end(); i++)
	{
		if(i->second == id)
		{
			ResourcesList.erase(i);
			goto LblResourcesClearUser;
		}
	}
	ResourcesListLock.Unlock();
}

bool Entity::ResourceOwnerIsMe(const char *ResourceName)
{
	std::map<mString, NEWPKI_THREAD_ID>::iterator rsc;
	ResourcesListLock.Lock();
	
	// Search resource lock
	rsc = ResourcesList.find(ResourceName);

	// Did we find the resource and am I the owner ?
	if(rsc == ResourcesList.end() ||
		rsc->second != NewpkiThread::CurrentThreadId())
	{
		ResourcesListLock.Unlock();
		return false;
	}
	ResourcesListLock.Unlock();
	return true;
}

void Entity::GetEntityCertificate(PKI_CERT & cEntityCert)
{
	cEntityCert = m_EntityCert;
}




void Entity::OnAudit(void *Param, LOG_MESSAGE_STATUS Status, LOG_MESSAGE_TYPE Message, const mString & User, const mString & Object, time_t log_timegmt, const mString & Error)
{
	MailInfo mail;
	mString Body;
	char ret[50];
	struct tm * tmval;
	Entity * me_this = (Entity *)Param;



	tmval = localtime(&log_timegmt);
	sprintf(ret, "%.2d/%.2d/%.2d %.2d:%.2d:%.2d GMT", tmval->tm_mday, tmval->tm_mon + 1, tmval->tm_year+1900, tmval->tm_hour, tmval->tm_min, tmval->tm_sec);


	Body = _sv("Type: ");
	Body += ASN1_logs_get_TypeString(Message);
	Body += _sv("\nStatus: ");
	Body += ASN1_logs_get_StatusString(Status);
	Body += _sv("\nObject: ");
	Body += Object;
	Body += _sv("\nUser: ");
	Body += User;
	Body += _sv("\nDate: ");
	Body += ret;
	if(Status == LOG_STATUS_TYPE_FAILURE && Error.size())
	{
		Body += _sv("\nError: ");
		Body += Error;
	}
	mail.set_Body(Body);
	mail.set_Subject(_sv("Audit Event"));
	mail.set_SignMail(true);

	me_this->m_Jobs.SendMail(me_this->m_EntityCert.GetStringName(), mail, true);
}

void Entity::OnUserLogout()
{
	NEWPKI_THREAD_ID user_id;
	user_id = NewpkiThread::CurrentThreadId();
	// Clear any resources to user might 
	// have forgoten to unlock
	ResourcesClearUser();

	AuthUsersLock.Lock();
	LogoutUser(AuthenticatedUsers[user_id]);
	AuthenticatedUsers.erase(user_id);
	AuthUsersLock.Unlock();
}

bool Entity::UserLogin(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	//We add it to the list of logged-in users
	ASN1_INTEGER * usertype;
	int TypeUser;
	UserHandle m_hUser;
	NEWPKI_THREAD_ID user_id;

	//User is not supposed to be known yet
	if(hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	m_hUser.SetUserName(body->d.login->username);
	m_hUser.SetPassword(body->d.login->password);

	if(!m_hUser.SetUserCert(UserCert))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!LoginUser(m_hUser, TypeUser))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	usertype = ASN1_INTEGER_new();
	if(!usertype)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	ASN1_INTEGER_SET(&usertype, TypeUser);

	user_id = NewpkiThread::CurrentThreadId();
	AuthUsersLock.Lock();
	AuthenticatedUsers[user_id] = m_hUser;

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_USER_TYPE, usertype))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		AuthenticatedUsers.erase(user_id);
		ASN1_INTEGER_free(usertype);
		AuthUsersLock.Unlock();
		return false;
	}
	AuthUsersLock.Unlock();
	return true;
}

const UserHandle & Entity::GetUserHandle()
{
	NEWPKI_THREAD_ID user_id;
	user_id = NewpkiThread::CurrentThreadId();
	AuthUsersLock.Lock();
	const UserHandle & ret = AuthenticatedUsers[user_id];
	AuthUsersLock.Unlock();
	return ret;
}

bool Entity::ExtractMyConf(const ENTITY_CONF_CRYPTED * c_conf, ENTITY_CONF ** CurrentConf, const INTERNAL_PKI_CA * cas, const X509 * PkiCert, bool UpdateDb)
{
	EVP_PKEY * PkiPubKey;
	ENTITY_CONF * tmpConf = NULL;

	ConfAccessLock.LockWrite();

	//Get PkiCert public key for signature verification
	PkiPubKey = X509_get_pubkey((X509*)PkiCert);
	if(!PkiPubKey)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	//Decrypt the body of our conf
	if(ENTITY_CONF_decrypt_verify(c_conf->crypted, &tmpConf, PkiPubKey, m_EntityKey.GetRsaKey()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		EVP_PKEY_free(PkiPubKey);
		ConfAccessLock.Unlock();
		return false;
	}
	EVP_PKEY_free(PkiPubKey);

	if(*CurrentConf)
	{
		//Do we need to update the conf in the DB ?
		if(ASN1_INTEGER_GET(tmpConf->version) <= ASN1_INTEGER_GET((*CurrentConf)->version))
		{
			ConfAccessLock.Unlock();
			ENTITY_CONF_free(tmpConf);
			return true;
		}
		ENTITY_CONF_free(*CurrentConf);
	}
	*CurrentConf = tmpConf;

	NewpkiDebug(LOG_LEVEL_INFO, m_EntityName.c_str(), _sv("New personnal conf version %ld"), ASN1_INTEGER_GET(tmpConf->version));

	if(!CommonConfLoad(tmpConf, cas))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ParseNewConf(c_conf, cas))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if(UpdateDb)
	{
		if(!WritePersonnalConf())
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ConfAccessLock.Unlock();
			return false;
		}
	}

	ConfAccessLock.Unlock();

	return true;
}

bool Entity::CommonConfLoad(const ENTITY_CONF *Conf, const INTERNAL_PKI_CA * cas)
{

	if(!m_Jobs.SetRepositories(Conf->Repositories))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
		
	if(!m_Jobs.SetAdminMails(Conf->acls->AdminSerials))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(m_Logging)
	{
		//Update the mail conf
		if(!m_Jobs.SetMailerInfo(m_EntityMail, Conf->mail_conf))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		//Update the audits
		if(!m_Logging->SetAudit(Conf->audits))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}

	if(!AclValidator.SetGroups(Conf->groups))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!AclValidator.SetACL(Conf->acls))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!AclValidator.SetCRL(Conf->crls))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(cas && !AclValidator.SetCA(cas))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity::IsFullyInit()
{
	return fullInit;
}


bool Entity::WritePersonnalConf()
{
	mString pem_conf;

	if(!PersonnalConfToPem(pem_conf))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	SQL sql(m_DbConn);
	mString req;

	if(req.sprintf(ENTITY_INSERT_CONF, pem_conf.c_str()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity::CheckLogsIntegrity(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!hUser.GetPassword().size() && !AclValidator.CanUserPerform(UserCert, ACL_TYPE_VIEW_LOGS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}	

	if(!m_Logging->CheckLogsIntegrity())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}


bool Entity::AdminSendMail(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	MailInfo mail;
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!hUser.GetPassword().size() && !AclValidator.CanUserPerform(UserCert, ACL_TYPE_SEND_ADMIN_MAIL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!m_Jobs.GetMailer())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MAIL_SERVER_NOT_SET);
		return false;
	}

	if(!mail.FromMAIL_INFO(body->d.admin_mail))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!m_Jobs.SendMail(UserCert.GetStringName(), mail, true))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity::GetLogsCount(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	unsigned long Count;
	ASN1_INTEGER * ctr;

	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	if(!hUser.GetPassword().size() && !AclValidator.CanUserPerform(UserCert, ACL_TYPE_VIEW_LOGS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!m_Logging->GetLogsCount(Count, body->d.enum_logs))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	ctr = NULL;
	ASN1_INTEGER_SET(&ctr, Count);
	if(!ctr)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_LOGS_COUNT, ctr))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity::EnumLogs(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	STACK_OF(LOG_ENTRY) * logs;

	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	if(!hUser.GetPassword().size() && !AclValidator.CanUserPerform(UserCert, ACL_TYPE_VIEW_LOGS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	logs = sk_new_null();
	if(!logs)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	if(!m_Logging->GetLogs(logs, body->d.enum_logs))
	{
		SKM_sk_pop_free(LOG_ENTRY, logs, LOG_ENTRY_free);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_ENUM_LOGS, logs))
	{
		SKM_sk_pop_free(LOG_ENTRY, logs, LOG_ENTRY_free);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity::GetLogsType(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	STACK_OF(ASN1_INTEGER) * logstype;
	size_t i;
	ASN1_INTEGER * type;
	
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	if(!hUser.GetPassword().size() && !AclValidator.CanUserPerform(UserCert, ACL_TYPE_VIEW_LOGS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	logstype = sk_ASN1_INTEGER_new_null();
	if(!logstype)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	for(i=0; i<LogsType.size(); i++)
	{
		type = NULL;
		ASN1_INTEGER_SET(&type, LogsType[i]);
		if(type)
		{
			sk_ASN1_INTEGER_push(logstype, type);
		}
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_LOGS_TYPE, logstype))
	{
		sk_ASN1_INTEGER_pop_free(logstype, ASN1_INTEGER_free);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity::GetMyACL(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(AclValidator.ValidateCert(UserCert) != INTERNAL_CA_TYPE_USER)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	ConfAccessLock.LockRead();
	ASN1_BIT_STRING * my_acl;
	unsigned char full_acl[] = {(char)255, (char)255, (char)255, (char)255, (char)255, (char)255, (char)255, (char)255, (char)255, (char)255, (char)255, (char)255, (char)0};

	//PKI Administrator
	if(AclValidator.IsPkiAdministrator(UserCert))
	{
		my_acl = ASN1_BIT_STRING_new();
		if(!my_acl)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
			ConfAccessLock.Unlock();
			return false;
		}

		ASN1_BIT_STRING_set(my_acl, full_acl, strlen((char*)full_acl));
	}
	else
	{
		//Get the user ACL
		my_acl = NULL;
		if(!AclValidator.GetUserRights(&my_acl, UserCert.GetSerial()))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ConfAccessLock.Unlock();
			return false;
		}
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_MY_ACL, my_acl))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}



void Entity::ERR_to_STORED_NEWPKI_RESPONSE(NewPKIStore & Store, ASN1_OCTET_STRING * transactionID, X509_PUBKEY * recipient)
{
	NEWPKI_RESPONSE * response;

	response = NEWPKI_RESPONSE_new();
	if(!response)
	{
		return;
	}

	response->type = NEWPKI_RESPONSE_TYPE_ERR;
	response->d.errors = ERR_to_ERROR_ENTRIES();
	if(!response->d.errors)
	{
		NEWPKI_RESPONSE_free(response);
		return;
	}

	SendResponse(Store, transactionID, response, recipient);
	NEWPKI_RESPONSE_free(response);
}

void Entity::SendResponse(NewPKIStore & Store, ASN1_OCTET_STRING * transactionID, NEWPKI_RESPONSE *Response, X509_PUBKEY *Recipient)
{
	STORED_NEWPKI_RESPONSE * s_resp;
	EVP_PKEY * cryptkey;

	s_resp = STORED_NEWPKI_RESPONSE_new();
	if(!s_resp)
	{
		return;
	}
	if(!ASN1_OCTET_STRING_cpy(&s_resp->transactionID, transactionID))
	{
		STORED_NEWPKI_RESPONSE_free(s_resp);
		return;
	}
	if(!X509_PUBKEY_cpy(&s_resp->c_info->sender, m_EntityKey.GetPublicKey()))
	{
		STORED_NEWPKI_RESPONSE_free(s_resp);
		return;
	}
	if(!X509_PUBKEY_cpy(&s_resp->c_info->recipient, Recipient))
	{
		STORED_NEWPKI_RESPONSE_free(s_resp);
		return;
	}

	cryptkey = X509_PUBKEY_get(Recipient);
	if(!cryptkey)
	{
		STORED_NEWPKI_RESPONSE_free(s_resp);
		return;
	}

	if(NEWPKI_RESPONSE_sign_encrypt(s_resp->c_info->c_response, Response, m_EntityKey.GetRsaKey(), cryptkey, EVP_sha1(), EVP_des_ede3_cbc()) <= 0)
	{
		STORED_NEWPKI_RESPONSE_free(s_resp);
		EVP_PKEY_free(cryptkey);
		return;
	}
	EVP_PKEY_free(cryptkey);

	// The response will be removed once it has been sent
	Store.Responder_SetResponse(s_resp);
	m_Jobs.InsertResponse(s_resp);
	STORED_NEWPKI_RESPONSE_free(s_resp);
}


bool Entity::SignResponse(ADMIN_RESPONSE *response)
{
	//We don't have the signer cert yet
	if(!response->signing_cert)
	{
		response->signing_cert = m_EntityCert.GetX509(true);
	}
	if(!response->signing_cert)
	{
		return false;
	}
	if(ASN1_sign((int (*)(void))i2d_ADMIN_RESPONSE_BODY, response->sig_algo, NULL, response->signature, (char*)response->body, (EVP_PKEY*)m_EntityKey.GetRsaKey(), EVP_sha1()) <= 0)
	{
		return false;
	}

	return true;
}

int Entity::GetType()
{
	return m_Type;
}
