/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/



// Entity.h: interface for the Entity class.
//
//////////////////////////////////////////////////////////////////////

#ifndef ENTITY_H
#define ENTITY_H

#include <NewPKI.h>
#include <ASN1/ASN1_req.h>
#include <ASN1/ASN1_resp.h>
#include <Conf.h>
#include <newpki_threads.h>
#include <CriticalSection.h>
#include <ReadersWriter.h>

#include <openssl/engine.h>


#include <HashTable/HashTable_String.h>

#include <SQL/SQL_Conn.h>
#include <SQL/SQL.h>
#include <SQL/SQL_CMD.h>

#include <PKI_CERT.h>

#include "AuthModule.h"
#include <EntityLog.h>
#include <PkiClient.h>
#include <FileLog.h>
#include <ReadersWriter.h>
#include "UserHandle.h"


#include "X509_ACL_Validator.h"
#include "SmtpClient.h"

#include "Entity_ASN1.h"
#include <mString.h>
#include "AsynchJobs.h"

#include "NewPKIStore.h"

#include <map>
#include <vector>
using namespace std;


#define ENTITY_CONSTRUCTOR_PARAMETERS const Config & Conf, mString & EntityName, ENGINE * e, AuthModule * authModule, int Type
#define ENTITY_CONSTRUCTOR_PARAM_PASSTHRU Conf, EntityName, e, authModule, Type

class Entity : public AsynchJobsEvent
{
public:
	/*! \brief This is the constructor.
	 *  \exception ExceptionNewPKI An error occured.
	 */
	Entity(ENTITY_CONSTRUCTOR_PARAMETERS);

	/*! \brief This is the destructor.
	 */
	virtual ~Entity();

	int GetType();

	/*! \brief This function will lock until there are no more clients using the class.
	 */
	void WaitForNoClients();
	
	/*! \brief This function must be called before calling any public function of this class.
	 */
	void ClientEnter();

	/*! \brief This function must be called after calling any public function of this class.
	 */
	void ClientLeave();

	/*! \brief This function destroys the entity.
	 *  \return true on success, false on failure.
	 */
	bool Destroy();

	/*! \brief This function returns the name of the entity.
	 *  \return The name of the entity.
	 */
	mString & GetName();

	/*! \brief Verifies if an entity doesn't already exists.
	 *  \param Exists [IN] true when exists, false when doesn't exist.
	 *  \return true on success, false on failure.
	 */
	bool EntityExists(bool & Exists);

	/*! \brief Yiels to start to stop.
	 */
	virtual void SpawnStop()=0;

	/*! \brief This function allows to get the certificate of the entity.
	 *  \param cEntityCert [OUT] The certificate.
	 */
	virtual void GetEntityCertificate(PKI_CERT & cEntityCert);

	/*! \brief This function creates the entity.
	 *  \param Params [IN] The creation parameters.
	 *  \param response [OUT] The creation response.
	 *  \return true on success, false on failure.
	 */
	virtual bool Create(const ENTITY_CREATION_DATAS * Params, ADMIN_RESPONSE_BODY * response)=0;

	/*! \brief This function intializes the entity.
	 *  \param init_datas [IN] The intialization parameters.
	 *  \return true on success, false on failure.
	 */
	virtual bool Init(const ENTITY_SIGNATURE_RESP * init_datas)=0;

	/*! \brief This function loads the entity.
	 *  \return true on success, false on failure.
	 */
	virtual bool Load()=0;

	/*! \brief This function upgrades the entity.
	 *  \param Version [IN] The current version.
	 *  \return true on success, false on failure.
	 */
	virtual bool Upgrade(const char * Version)=0;

	/*! \brief This function signs an response using the entity's certificate.
	 *  \param response [IN/OUT] The response.
	 *  \return true on success, false on failure.
	 */
	bool SignResponse(ADMIN_RESPONSE* response);

	/*! \brief This function is called when a new request is received.
	 *  \param response [OUT] The response.
	 *  \param ClientCert [IN] The end user certificate.
	 *  \param AdminRequest [IN] The request.
	 *  \return true on success, false on failure.
	 */
	virtual bool ParseAdminCommand(ADMIN_RESPONSE_BODY * response, const PKI_CERT & ClientCert, const ADMIN_REQUEST * AdminRequest)=0;

	/*! \brief This function is called when the connection to a user is closed.
	 */
	void OnUserLogout();

	/*! \brief This function checks if the entity has been fully intialized.
	 *  \return true when initialized, false when not initialized.
	 */
	bool IsFullyInit();
protected:
	#define DECLARE_COMMAND_PARSER(clname) \
		typedef bool (clname::*FUNCTION_COMMAND_##clname)(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser); \
		typedef struct st_internal_##clname \
		{ \
			LOG_MESSAGE_TYPE LogType; \
			char * ObjectName; \
			unsigned long ObjectId; \
			int CommandId; \
			FUNCTION_COMMAND_##clname Function; \
		}internal_##clname;
	
	#define PARSER_COMMAND_BEGIN(clname, resp, userid, streq, usercert, clparam, execcmd, thisptr, cldstlogstype) \
		internal_##clname CommandEntry={(LOG_MESSAGE_TYPE)0, NULL, LOG_NO_OBJECTID, 0, NULL}; \
		ADMIN_REQUEST_BODY * body = (streq)?(streq)->body:NULL; \
		const PKI_CERT & m_UserCert = (usercert); \
		ADMIN_RESPONSE_BODY * internal_resp = (resp); \
		unsigned long cl_user_id = (userid); \
		const UserHandle & cccchUser = (clparam); \
		bool func_ret; \
		char * err; \
		ERR_clear_error(); \
		vector<LOG_MESSAGE_TYPE> & dstLogsType = cldstlogstype; \
		clname * cl_me_this = (clname*)(thisptr); \
		bool clExecCmd = (execcmd); \
		size_t numLogs = dstLogsType.size(); \
		CommandEntry.Function = NULL;

	#define PARSER_ADD_LOG_ENTRY(logtype) \
		if(!numLogs) \
		{ \
			dstLogsType.push_back(logtype); \
		}

	#define PARSER_COMMAND_ENTRY_LOG(cmdid, cmdfunc, logtype, objectname, objectid) \
		if( clExecCmd && body->type == cmdid ) \
		{ \
			CommandEntry.LogType = logtype; \
			CommandEntry.ObjectName = objectname; \
			CommandEntry.ObjectId = objectid; \
			CommandEntry.CommandId = cmdid; \
			CommandEntry.Function = & cmdfunc; \
		} \
		PARSER_ADD_LOG_ENTRY(logtype)


	#define PARSER_COMMAND_ENTRY(cmdid, cmdfunc) \
		PARSER_COMMAND_ENTRY_LOG(cmdid, cmdfunc, (LOG_MESSAGE_TYPE)0, NULL, 0)
	

	#define PARSER_COMMAND_END(clname) \
		if(!clExecCmd) return true; \
		if(!CommandEntry.Function) \
		{ \
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PKI_COMMAND); \
			cl_me_this->ERR_to_ADMIN_RESPONSE(internal_resp); \
			return false; \
		} \
		if(CommandEntry.LogType) \
		{ \
			if(!CommandEntry.ObjectName && CommandEntry.ObjectId == LOG_NO_OBJECTID) \
			{ \
				NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PKI_COMMAND); \
				cl_me_this->ERR_to_ADMIN_RESPONSE(internal_resp); \
				return false; \
			} \
			cl_me_this->m_Logging->LogMessage(LOG_STATUS_TYPE_REQUEST, CommandEntry.LogType, cl_user_id, ((m_UserCert)?m_UserCert.GetStringName():NULL), CommandEntry.ObjectId, CommandEntry.ObjectName); \
		} \
		func_ret = (cl_me_this->*((FUNCTION_COMMAND_##clname) (CommandEntry.Function)))( internal_resp, CommandEntry.CommandId, m_UserCert, body, cccchUser); \
		if(CommandEntry.LogType) \
		{ \
			if(!func_ret) \
			{ \
				err = ERR_to_string(); \
				cl_me_this->m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, CommandEntry.LogType, cl_user_id, ((m_UserCert)?m_UserCert.GetStringName():NULL), CommandEntry.ObjectId, CommandEntry.ObjectName, err); \
				if(err) free(err); \
			} \
			else \
				cl_me_this->m_Logging->LogMessage(LOG_STATUS_TYPE_SUCCESS, CommandEntry.LogType, cl_user_id, ((m_UserCert)?m_UserCert.GetStringName():NULL), CommandEntry.ObjectId, CommandEntry.ObjectName); \
		} \
		if(!func_ret) \
		{ \
			cl_me_this->ERR_to_ADMIN_RESPONSE(internal_resp); \
			return false; \
		} \
		else \
			return true;


	void ERR_to_STORED_NEWPKI_RESPONSE(NewPKIStore & Store, ASN1_OCTET_STRING * transactionID, X509_PUBKEY * recipient);
	void SendResponse(NewPKIStore & Store, ASN1_OCTET_STRING * transactionID, NEWPKI_RESPONSE *Response, X509_PUBKEY *Recipient);
	

	bool Private_CreateEntity(const GEN_PRIVATE_KEY* genkey, NewPKIStore * EntityStore, const char ** InitialCreates);
	bool CommonConfLoad(const ENTITY_CONF * Conf, const INTERNAL_PKI_CA * cas);
	bool Private_InitEntity(const X509 *Cert, ENTITY_CONF ** CurrentConf = NULL, const ENTITY_SIGNATURE_RESP *init_datas = NULL, NewPKIStore * EntityStore = NULL);
	bool Private_CommonLoad(NewPKIStore * Store, bool LoadConf=true);
	bool WritePersonnalConf();
	bool ExtractMyConf(const ENTITY_CONF_CRYPTED * c_conf, ENTITY_CONF ** CurrentConf, const INTERNAL_PKI_CA * cas, const X509 * PkiCert, bool UpdateDb);

	bool ADMIN_RESPONSE_init(ADMIN_RESPONSE_BODY * response, int Type, void * Datas);
	void ERR_to_ADMIN_RESPONSE(ADMIN_RESPONSE_BODY * response);
	
	void ResourceUnlock(const char * ResourceName);
	bool ResourceLock(const char * ResourceName);
	bool ResourceOwnerIsMe(const char * ResourceName);
	const UserHandle & GetUserHandle();

	bool GetMyACL(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	bool AdminSendMail(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	
	bool CheckLogsIntegrity(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	bool EnumLogs(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	bool GetLogsCount(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	bool GetLogsType(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);

	bool Private_GetMyACL(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	bool UserLogin(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);

	

	STACK_OF(X509) * EntityParentCerts;
	X509_ACL_Validator AclValidator;
	ENGINE * m_Engine;
	mString m_EntityName;
	mString m_EntityMail;
	SQL_Connection * m_DbConn;
	Config m_Conf;
	AuthModule * m_AuthModule;
	EntityLog * m_Logging;

	PKI_CERT m_EntityCert;
	PKI_RSA m_EntityKey;
	ReadersWriter ConfAccessLock;
	vector<LOG_MESSAGE_TYPE> LogsType;
	AsynchJobs m_Jobs;
	
private:

	bool fullInit;

	virtual bool LoginUser(UserHandle & hUser, int & UserType)=0;
	virtual void LogoutUser(const UserHandle & hUser)=0;

	virtual bool ParseNewConf(const ENTITY_CONF_CRYPTED * newConf, const INTERNAL_PKI_CA * cas)=0;
	virtual bool PersonnalConfFromPEM(const mString & conf_pem)=0;
	virtual bool PersonnalConfToPem(mString & conf_pem)=0;


	bool isStopped;
	CriticalSection ClientsReadLock;		//!< The clients ctr MUTEX
	unsigned long entered_client;			//!< The number of active client
	void ResourcesClearUser();
	map<mString, NEWPKI_THREAD_ID> ResourcesList;
	CriticalSection ResourcesListLock;

	ENTITY_CONF ** m_pmyConf;

	static void OnAudit(void *Param, LOG_MESSAGE_STATUS Status, LOG_MESSAGE_TYPE Message, const mString & User, const mString & Object, time_t log_timegmt, const mString & Error);

	map<NEWPKI_THREAD_ID, UserHandle> AuthenticatedUsers;
	CriticalSection AuthUsersLock;
	int m_Type;
};

#endif
