/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/


// EntityLog.cpp: implementation of the EntityLog class.
//
//////////////////////////////////////////////////////////////////////
#include <NewPKI.h>
#include "EntityLog.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////




EntityLog::EntityLog(const SQL_Connection * DbConn, const PKI_CERT * EntityCert)
{
	DB_Conn = NULL;
	m_Audit = NULL;

	try
	{
		DB_Conn = DbConn->Clone();
	}
	catch(ExceptionNewPKI e)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		throw ExceptionNewPKI();
	}
	if(!DB_Conn)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		throw ExceptionNewPKI();
	}

	if(EntityCert)
		m_EntityCert = *EntityCert;

	// We load the HashCorrelation
	LogsHash.SetCert(m_EntityCert);

	mString CurrentHash;
	SQL sql(DB_Conn);
	long NumRows;
	if(!sql.Execute(LOG_SELECT_HASH))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		throw ExceptionNewPKI();
	}
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		throw ExceptionNewPKI();
	}
	if(NumRows)
	{
		if(!sql.Value(0, "hash", CurrentHash))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			throw ExceptionNewPKI();
		}
		if(CurrentHash.size())
		{
			if(!LogsHash.SetCurrentHash(CurrentHash))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				throw ExceptionNewPKI();
			}
		}
	}
}

EntityLog::~EntityLog()
{
	if(m_Audit) ENTITY_AUDIT_free(m_Audit);
	if(DB_Conn) delete DB_Conn;
}



void EntityLog::LogMessage(LOG_MESSAGE_STATUS Status, LOG_MESSAGE_TYPE Message, unsigned long UserId, const char * UserDN, unsigned long ObjectId, const char * ObjectName, const char * Error) const 
{
	int i;
	ENTITY_AUDIT_ENTRY * currAudit;
	LOG_ENTRY * log;
	mString object;
	mString user;
	mString error;
	time_t tim;
	const EVP_PKEY * privateKey;
	mString req;
	char * signature;
	char * sig;

	if(!DB_Conn)
		return;

	if(m_EntityCert.GetPrivateKey())
		privateKey = m_EntityCert.GetPrivateKey().GetRsaKey();
	else
		privateKey = NULL;
	
	if(ObjectId != LOG_NO_OBJECTID)
		object = ObjectId;
	else
		object = ObjectName;

	if(UserDN)
		user = UserDN;
	else
		user = UserId;

	error = Error;

	time_gmt(&tim);


	//Check if the entry correspond to an audit
	AuditLock.LockRead();
	if(m_Audit && m_OnAudit)
	{
		for(i=0; i<SKM_sk_num(ENTITY_AUDIT_ENTRY, m_Audit->audits); i++)
		{
			currAudit = SKM_sk_value(ENTITY_AUDIT_ENTRY, m_Audit->audits, i);
			if(!currAudit) continue;

			//Check type
			if(currAudit->type)
			{
				if((LOG_MESSAGE_TYPE)ASN1_INTEGER_GET(currAudit->type) != Message) continue;
			}

			//Check status
			if(currAudit->status)
			{
				if((LOG_MESSAGE_STATUS)ASN1_INTEGER_GET(currAudit->status) != Status) continue;
			}

			//Check object
			if(currAudit->object)
			{
				if( !(object == currAudit->object) ) continue;
			}
			
			//Check user
			if(currAudit->user)
			{
				if( !(user == currAudit->user) ) continue;
			}

			//Ok we have an audit !!
			m_OnAudit(m_Param, Status, Message, user, object, tim, error);
		}
	}
	AuditLock.Unlock();

	log = LOG_ENTRY_new();
	if(!log)
	{
		return;
	}

	ASN1_INTEGER_SET(&log->body->log_status, Status);
	ASN1_INTEGER_SET(&log->body->log_type, Message);
	if(!user.c_ASN1_UTF8STRING(&log->body->user))
	{
		LOG_ENTRY_free(log);
		return;
	}
	if(!object.c_ASN1_UTF8STRING(&log->body->object_name))
	{
		LOG_ENTRY_free(log);
		return;
	}
	if(!error.c_ASN1_UTF8STRING(&log->body->error))
	{
		LOG_ENTRY_free(log);
		return;
	}
	ASN1_INTEGER_SET(&log->body->log_date, tim);

	//If we don't have the private key to sign we still log
	//it's better than nothing
	if(privateKey)
	{
		//Now we generated the signature
		LOG_ENTRY_sign(log, (EVP_PKEY*)privateKey);
	}

	SQL sql(DB_Conn);

	object = sql.FormatString(object);
	user = sql.FormatString(user);
	error = sql.FormatString(error);

	signature = LOG_ENTRY_SIG_To_PEM(log->sig);
	sig = ( (signature)?signature:(char*)"" );


	if(req.sprintf(LOG_INSERT, (int)ASN1_INTEGER_GET(log->body->log_status), (int)ASN1_INTEGER_GET(log->body->log_type), user.c_str(), object.c_str(), error.c_str(), ASN1_INTEGER_GET(log->body->log_date), sig) <= 0)
	{
		if(signature) free(signature);
		LOG_ENTRY_free(log);
		return;
	}

	//On va maintenant l'ins�er dans la base de donn�s
	if(!sql.Execute(req))
	{
		LOG_ENTRY_free(log);
		if(signature) free(signature);
		return;
	}

	HashLock.Lock();
	if(signature)
	{
		// Compute the new hash for logs
		if(!LogsHash.AddEntry(sql.GetLastID(), (char*)log->sig->signature->data, log->sig->signature->length))
		{
			//We don't want the hash to be broken we remove the log entry
			req.sprintf(LOG_DELETE, sql.GetLastID());
			sql.Execute(req);
		}
		else
		{
			//Calculate the new hash
			if(!LogsHash.GetHash(object))
			{
				//We don't want the hash to be broken we remove the log entry
				req.sprintf(LOG_DELETE, sql.GetLastID());
				sql.Execute(req);
			}
			else
			{
				//Insert the new hash in db
				req.sprintf(LOG_INSERT_HASH, object.c_str());
				sql.Execute(req);
			}
		}
		free(signature);
	}
	HashLock.Unlock();

	LOG_ENTRY_free(log);
}


bool EntityLog::GetLogs(STACK_OF(LOG_ENTRY) * Logs, const ADMIN_REQ_ENUM_LOGS * Filters) const
{
	SQL sql(DB_Conn);
	mString req;
	long NumRows;
	long i;
	LOG_ENTRY * Value;
	mString where;
	int Offset;
	int Max;

	if(!DB_Conn || !Logs)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	where = FormatFiltersSQL(Filters);


	if(Filters->index)
		Offset=ASN1_INTEGER_GET(Filters->index);
	else
		Offset=0;

	if(Filters->max)
		Max=ASN1_INTEGER_GET(Filters->max);
	else
		Max=0;

	if(req.sprintf(LOG_SELECT, where.c_str(), Offset, Max) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}

	//We execute the request
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	for(i=0; i<NumRows; i++)
	{
		Value = SqlToLOG_ENTRY(sql, i);
		if(!Value)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		if(SKM_sk_push(LOG_ENTRY, Logs, Value) < 0)
		{
			LOG_ENTRY_free(Value);
		}
	}

	return true;
}

mString EntityLog::FormatFiltersSQL(const ADMIN_REQ_ENUM_LOGS * Filters) const
{
	if(!Filters) return "";

	mString where;
	int entries_count=0;
	char * object = NULL;
	char * user = NULL;
	mString tt_user;

	#define ADD_FILTER(name, operat, value) \
		if(entries_count) \
			where += "AND "; \
		where += name; \
		where += " "; \
		where += operat; \
		where += " '"; \
		where += value; \
		where += "' "; \
		entries_count++;



	where = "WHERE ";
	entries_count = 0;
	object = ASN1_STRING_GET(Filters->object);
	if( object )
	{
		ADD_FILTER("object_name", "=", object);
	}

	if( Filters->type )
	{
		ADD_FILTER("log_type", "=", Filters->type);
	}

	if( Filters->date_from )
	{
		ADD_FILTER("log_date", ">", Filters->date_from);
	}

	if( Filters->date_to )
	{
		ADD_FILTER("log_date", "<", Filters->date_to);
	}

	if( Filters->status )
	{
		ADD_FILTER("log_status", "=", Filters->status);
	}

	user = ASN1_STRING_GET(Filters->user);
	if( user )
	{
		tt_user = "%%";
		tt_user += user;
		tt_user += "%%";
		ADD_FILTER("LOWER(user)", "LIKE", tt_user);
	}

	if(!entries_count)
		return "";
	else
		return where;
}

void EntityLog::SetAuditCalllBack(PROC_ON_AUDIT_CALLBACK *OnAudit, void *Param)
{
	AuditLock.LockWrite();
	m_OnAudit = OnAudit;
	m_Param = Param;
	AuditLock.Unlock();
}

bool EntityLog::SetAudit(const ENTITY_AUDIT *Audit)
{
	AuditLock.LockWrite();
	if(!ENTITY_AUDIT_cpy(&m_Audit, Audit))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	AuditLock.Unlock();
	return true;
}

bool EntityLog::CheckLogsIntegrity() const
{
	SQL sql(DB_Conn);
	mString req;
	long NumRows;
	unsigned Offset = 0;
	int i;
	LOG_ENTRY * Value;
	HashCorrelation LogsCorrelation;

	LogsCorrelation.SetCert(m_EntityCert);


	do
	{
		if(req.sprintf(LOG_GET_LOGS_LIST, Offset*200, 200) <= 0)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			return false;
		}
		if(!sql.Execute(req))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		if(!sql.NumRows(&NumRows))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		if(NumRows)
		{
			for(i=0; i<NumRows; i++)
			{
				Value = SqlToLOG_ENTRY(sql, i);
				if(!Value)
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					return false;
				}

				//We get its log id
				if(!sql.Value(i, "log_id", req))
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					LOG_ENTRY_free(Value);
					return false;
				}
				
				//We verify the signature of the current entry
				if(!LOG_ENTRY_verify(Value, (EVP_PKEY *)m_EntityCert.GetPublicKey()))
				{
					ERR_clear_error();
					NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_ENTRY_SIGNATURE);
					ERR_set_error_data(strdup(req.c_str()), ERR_TXT_STRING | ERR_TXT_MALLOCED);
					LOG_ENTRY_free(Value);
					return false;
				}

				if(!LogsCorrelation.AddEntry(req.c_ulng(), (char*)Value->sig->signature->data, Value->sig->signature->length))
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					LOG_ENTRY_free(Value);
					return false;
				}
				LOG_ENTRY_free(Value);
			}
		}
		Offset++;
	}
	while(NumRows);

	//We now compare the two logs hash to make sure they're identical
	if(LogsCorrelation == LogsHash)
		return true;

	NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_ENTRY_SIGNATURE);
	ERR_set_error_data("Unknown", ERR_TXT_STRING);
	return false;
}

LOG_ENTRY * EntityLog::SqlToLOG_ENTRY(SQL &sql, long i) const
{
	mString str_enum;
	LOG_ENTRY * Value;

	Value = LOG_ENTRY_new();
	if(!Value)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return NULL;
	}

	if(!sql.Value(i, "log_id", str_enum))
	{
		LOG_ENTRY_free(Value);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!str_enum.c_ASN1_INTEGER(&Value->log_id))
	{
		LOG_ENTRY_free(Value);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return NULL;
	}

	if(!sql.Value(i, "log_status", str_enum))
	{
		LOG_ENTRY_free(Value);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!str_enum.c_ASN1_INTEGER(&Value->body->log_status))
	{
		LOG_ENTRY_free(Value);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return NULL;
	}

	if(!sql.Value(i, "log_type", str_enum))
	{
		LOG_ENTRY_free(Value);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!str_enum.c_ASN1_INTEGER(&Value->body->log_type))
	{
		LOG_ENTRY_free(Value);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return NULL;
	}

	if(!sql.Value(i, "user", str_enum))
	{
		LOG_ENTRY_free(Value);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!str_enum.c_ASN1_UTF8STRING(&Value->body->user))
	{
		LOG_ENTRY_free(Value);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return NULL;
	}

	if(!sql.Value(i, "object_name", str_enum))
	{
		LOG_ENTRY_free(Value);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!str_enum.c_ASN1_UTF8STRING(&Value->body->object_name))
	{
		LOG_ENTRY_free(Value);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}

	if(!sql.Value(i, "error", str_enum))
	{
		LOG_ENTRY_free(Value);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!str_enum.c_ASN1_UTF8STRING(&Value->body->error))
	{
		LOG_ENTRY_free(Value);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return NULL;
	}

	
	if(!sql.Value(i, "log_date", str_enum))
	{
		LOG_ENTRY_free(Value);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!str_enum.c_ASN1_INTEGER(&Value->body->log_date))
	{
		LOG_ENTRY_free(Value);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return NULL;
	}

	if(!sql.Value(i, "signature", str_enum))
	{
		LOG_ENTRY_free(Value);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!LOG_ENTRY_SIG_From_PEM(&Value->sig, (char*)str_enum.c_str()))
	{
		LOG_ENTRY_free(Value);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return NULL;
	}
	return Value;
}

bool EntityLog::GetLogsCount(unsigned long &Count, const ADMIN_REQ_ENUM_LOGS *Filters) const
{
	SQL sql(DB_Conn);
	mString req;
	mString where;

	if(!DB_Conn)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	where = FormatFiltersSQL(Filters);

	if(req.sprintf(LOG_SELECT_COUNT, where.c_str()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}

	//We execute the request
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!sql.Value(0, "ctr", req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	Count = req.c_ulng();
	return true;
}
