/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/

// Entity_CA.cpp: implementation of the Entity_CA class.
//
//////////////////////////////////////////////////////////////////////

#include "Entity_CA.h"
#include "NewPKIStore.h"
#include "svintl.h"
#include <PKI_P7B.h>



//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Entity_CA::Entity_CA(ENTITY_CONSTRUCTOR_PARAMETERS):Entity(ENTITY_CONSTRUCTOR_PARAM_PASSTHRU)
{
	myConf = NULL;
	c_myConf = NULL;
	m_InternalCa = NULL;
	m_PlugCA = NULL;

	try
	{
		m_InternalCa = new CA_Handler(0);
	}
	catch(ExceptionNewPKI e)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		throw ExceptionNewPKI();		
	}

	m_InternalCa->set_ENGINE(m_Engine);

	hThreadWorker.Create(ThreadWorker, this);
	hThreadGenerateCRL.Create(ThreadGenerateCRL, this);
	hThreadRepublish.Create(ThreadRepublish, this);
}

Entity_CA::~Entity_CA()
{
	hThreadWorker.Stop();
	hThreadGenerateCRL.Stop();
	hThreadRepublish.Stop();

	ConfAccessLock.LockWrite();
	if(myConf)
		ENTITY_CONF_free(myConf);
	if(c_myConf)
		LOCAL_CA_CONF_free(c_myConf);
	ConfAccessLock.Unlock();

	if(m_InternalCa)
		delete m_InternalCa;
	if(m_PlugCA)
		delete m_PlugCA;
}

bool Entity_CA::Create(const ENTITY_CREATION_DATAS * Params, ADMIN_RESPONSE_BODY * response)
{
	if(!Params)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	if(Params->type != ENTITY_TYPE_CA)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	ENTITY_CREATION_RESP * resp;

	if(!Params->d.entity_key)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}


	//We create the database
	if(!Private_CreateEntity(Params->d.entity_key, &CaPublicationStore, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!CA_Handler::CreateTables(m_DbConn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}

	if(!m_InternalCa->SetDbConn(m_DbConn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}


	resp = ENTITY_CREATION_RESP_new();
	if(!resp)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		Destroy();
		return false;
	}
	resp->type = ENTITY_TYPE_CA;
	if(X509_PUBKEY_set(&resp->d.entity_pub_key, (EVP_PKEY*)m_EntityKey.GetRsaKey()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		Destroy();
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_CREATE_ENTITY, resp))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_CA::Load()
{
	if(!Private_CommonLoad(&CaPublicationStore))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!IsFullyInit())
	{
		return true;
	}

	if(!m_InternalCa->SetDbConn(m_DbConn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	sk_X509_push(EntityParentCerts, X509_dup(c_myConf->cas->RootCa));
	sk_X509_push(EntityParentCerts, X509_dup(c_myConf->cas->EntitiesCa));

	//Start the conf synch
	if(!m_Jobs.StartConfSync())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	//Start the mailer
	if(!m_Jobs.StartMailer())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!m_Jobs.SetEntityParentsCert(EntityParentCerts))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//Load the CA
	if(!m_InternalCa->Load())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!hThreadWorker.Start())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}

	if(!hThreadGenerateCRL.Start())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}
/*
	if(!hThreadRepublish.Start())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}
*/
	return true;
}


bool Entity_CA::Init(const ENTITY_SIGNATURE_RESP * init_datas)
{
	if(IsFullyInit())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!init_datas)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	if(init_datas->body->type != ENTITY_TYPE_CA)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	if(!Private_InitEntity(init_datas->body->d.EntityCert, &myConf, init_datas, &CaPublicationStore))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_CA::PersonnalConfFromPEM(const mString &conf_pem)
{
	//Convert from PEM
	ASN1_ENCRYPT_SIGN * encrypt;
	encrypt = NULL;

	if(ASN1_ENCRYPT_SIGN_From_PEM(&encrypt, conf_pem.c_str()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(c_myConf)
	{
		LOCAL_CA_CONF_free(c_myConf);
		c_myConf = NULL;
	}

	//Decrypt the conf
	if(LOCAL_CA_CONF_decrypt_verify(encrypt, &c_myConf, m_EntityKey.GetRsaKey(), m_EntityKey.GetRsaKey()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ASN1_ENCRYPT_SIGN_free(encrypt);
		return false;
	}
	ASN1_ENCRYPT_SIGN_free(encrypt);
	
	//Extract the conf
	if(myConf)
	{
		ENTITY_CONF_free(myConf);
		myConf = NULL;
	}
	if(!ExtractMyConf(c_myConf->conf, &myConf, c_myConf->cas, c_myConf->cas->PkiCert, false))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_CA::PersonnalConfToPem(mString & conf_pem)
{
	char * pem;
	ASN1_ENCRYPT_SIGN * crypted_conf;

	ConfAccessLock.LockRead();

	//Sign/Crypt the conf
	crypted_conf = ASN1_ENCRYPT_SIGN_new();
	if(!crypted_conf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ConfAccessLock.Unlock();
		return false;
	}

	if(LOCAL_CA_CONF_sign_encrypt(crypted_conf, c_myConf, m_EntityKey.GetRsaKey(), m_EntityKey.GetRsaKey(), EVP_sha1(), EVP_des_ede3_cbc()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ASN1_ENCRYPT_SIGN_free(crypted_conf);
		ConfAccessLock.Unlock();
		return false;
	}

	//Convert it to PEM
	pem = ASN1_ENCRYPT_SIGN_To_PEM(crypted_conf);
	ASN1_ENCRYPT_SIGN_free(crypted_conf);
	if(!pem)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
		
	conf_pem = pem;
	free(pem);

	ConfAccessLock.Unlock();
	return true;
}


bool Entity_CA::LoginUser(UserHandle & hUser, int & UserType)
{
	if(!AclValidator.CanUserPerform(hUser.GetUserCert(), ACL_TYPE_AUTHENTICATE_ON_ENTITY))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	UserType = USER_TYPE_CA;
	return true;
}

void Entity_CA::LogoutUser(const UserHandle & hUser)
{
}

bool Entity_CA::ParseNewConf(const ENTITY_CONF_CRYPTED * newConf, const INTERNAL_PKI_CA * cas)
{
	mString LibraryPath;

	//Load the extensions from the conf
	m_Exts.Lock();
	if(!m_Exts.From_EXTENSION_VALUE(myConf->MY_CA_CONF_PTR->Exts))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		m_Exts.Unlock();
		return false;
	}
	m_Exts.Unlock();

	//Load the CRL extensions from the conf
	m_CrlExts.Lock();
	if(!m_CrlExts.From_EXTENSION_VALUE(myConf->MY_CA_CONF_PTR->CrlExts))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		m_CrlExts.Unlock();
		return false;
	}
	m_CrlExts.Unlock();

	//Load the CRL validity len
	m_crldays = ASN1_INTEGER_GET(myConf->MY_CA_CONF_PTR->CrlValidityDays);
	m_crlhours = ASN1_INTEGER_GET(myConf->MY_CA_CONF_PTR->CrlValidityHours);

	//Load the plug options from the conf
	m_PlugOptions.Lock();
	if(!m_PlugOptions.From_PLUG_OPTION(myConf->MY_CA_CONF_PTR->PlugCa->Options))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		m_PlugOptions.Unlock();
		return false;
	}
	m_PlugOptions.Unlock();


	LibraryPath = myConf->MY_CA_CONF_PTR->PlugCa->LibraryPath;

	//Do we need to update the PlugCA ?
	if(m_PlugCA)
	{
		// Has the lib name changed ?
		if(!(LibraryPath == m_PlugCA->GetLibraryPath()) || !LibraryPath.size())
		{
			delete m_PlugCA;
			m_PlugCA = NULL;
		}
	}
	//Need to load a new ca plug
	if(LibraryPath.size())
	{
		try
		{
			m_PlugCA = new PlugCA(LibraryPath);
		}
		catch(ExceptionNewPKI e)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		if(!m_PlugCA)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
			return false;
		}
	}

	// Validate the plug options
	if(m_PlugCA && !m_PlugCA->ValidateOptions(m_PlugOptions))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//Should we initialize the local conf ?
	if(!c_myConf)
		c_myConf = LOCAL_CA_CONF_new();

	if(!c_myConf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	//Should we update the internal cas certs ?
	if(cas)
	{
		if(!INTERNAL_PKI_CA_cpy(&c_myConf->cas, cas))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}

	//Save the updated conf
	if(!ENTITY_CONF_CRYPTED_cpy(&c_myConf->conf, newConf))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_CA::Upgrade(const char * Version)
{
	return true;
}


void Entity_CA::GetACL_List(STACK_OF(ASN1_INTEGER) *acl_list)
{
	ASN1_INTEGER * acl_id;
	int i;
	static ACL_TYPE list_acls[] =
	{
		ACL_TYPE_VIEW_LOGS,
		ACL_TYPE_CREATE_ROOT_CA, 
		ACL_TYPE_SEND_ADMIN_MAIL, 
		ACL_TYPE_CREATE_CHILD_CA,
		ACL_TYPE_CA_SIGN_CERT,
		ACL_TYPE_CA_REVOKE_CERT,
		ACL_TYPE_CA_GENERATE_CRL,
		(ACL_TYPE)0
	};
	for(i=0; list_acls[i]; i++)
	{
		acl_id = NULL;
		ASN1_INTEGER_SET(&acl_id, list_acls[i]);
		if(acl_id)
		{
			if(SKM_sk_push(ASN1_INTEGER, acl_list, acl_id) < 0)
			{
				ASN1_INTEGER_free(acl_id);
			}
		}
	}	
}

bool Entity_CA::ParseAdminCommand(ADMIN_RESPONSE_BODY * response, const PKI_CERT & ClientCert, const ADMIN_REQUEST * AdminRequest)
{
	//We only accept SSLv3 connection	
	if(!ClientCert)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ERR_to_ADMIN_RESPONSE(response);
		return false;
	}
	return Private_ParseAdminCommand(true, this, LogsType, response, ClientCert, AdminRequest, GetUserHandle());
}

bool Entity_CA::Private_ParseAdminCommand(bool ExecuteCmd, Entity * me_this, vector<LOG_MESSAGE_TYPE> & mLogsType, ADMIN_RESPONSE_BODY * response, const PKI_CERT & ClientCert, const ADMIN_REQUEST * AdminRequest, const UserHandle & hUser)
{
	mString str;
	char * Dn;

	if(AdminRequest && AdminRequest->body->type == ADMIN_REQ_TYPE_SIGN_CSR)
	{
		Dn = X509_NAME_oneline(AdminRequest->body->d.sign_csr->request->req_info->subject, NULL, 0);
		if(Dn)
		{
			str = Dn;
			free(Dn);
		}
	}

	PARSER_COMMAND_BEGIN(Entity_CA, response, 0, AdminRequest, ClientCert, hUser, ExecuteCmd, me_this, mLogsType)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_ENTITY_GET_MY_CONF)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_SEND_ADMIN_MAIL)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_SEND_MAIL)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_NEW_REQUET)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_CA_CERT)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_CA_REV)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_CERT_PUBLICATION)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_REV_PUBLICATION)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_CRL_PUBLICATION)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_LOGIN,					Entity_CA::UserLogin,				LOG_MESSAGE_TYPE_USER_LOGIN, (ClientCert)?(char*)ClientCert.GetStringName():NULL, LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_ENUM_LOGS,				Entity_CA::EnumLogs)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_LOGS_COUNT,			Entity_CA::GetLogsCount)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_CREATE_ROOT_CA,			Entity_CA::CreateRootCa,			LOG_MESSAGE_TYPE_CREATE_ROOT_CA, _sv("none"), LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_CREATE_CHILD_CA,			Entity_CA::CreateChildCa,			LOG_MESSAGE_TYPE_CREATE_CHILD_CA, _sv("none"), LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_IMPORT_CHILD_CA_CERT,	Entity_CA::ImportChildCaCert,		LOG_MESSAGE_TYPE_IMPORT_CHILD_CA_CERT, _sv("none"), LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_CA_STATUS,			Entity_CA::GetStatus)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_CA_CERT,				Entity_CA::GetCaCert)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_MY_ACL,				Entity_CA::GetMyACL)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_LOGS_TYPE,			Entity_CA::GetLogsType)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_SEND_ADMIN_MAIL,			Entity_CA::AdminSendMail,			LOG_MESSAGE_TYPE_ADD_ADMIN_MAIL_QUEUE,	_sv("none"), LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_ENUM_CERTS,				Entity_CA::EnumCerts)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_ENUM_CRLS,				Entity_CA::EnumCrls)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_SIGN_CSR,				Entity_CA::SignCSR,					LOG_MESSAGE_TYPE_SIGN_CSR,	(char*)str.c_str(), LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_REVOKE_CERT,				Entity_CA::RevokeCert,				LOG_MESSAGE_TYPE_CERT_REVOCATION,	NULL, ASN1_INTEGER_GET(AdminRequest->body->d.serial))
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_CHECK_LOGS,				Entity_CA::CheckLogsIntegrity)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_GENERATE_CRL,			Entity_CA::GenerateCRL,			LOG_MESSAGE_TYPE_GEN_CRL,	"CRL", LOG_NO_OBJECTID)
	PARSER_COMMAND_END(Entity_CA)
}

void Entity_CA::LogsTypeGet(vector<LOG_MESSAGE_TYPE> & cLogsType)
{
	Private_ParseAdminCommand(false, NULL, cLogsType, NULL, PKI_CERT(), NULL, UserHandle());
}

void Entity_CA::ThreadRepublish(const NewpkiThread * Thread, void *param)
{
	Entity_CA * me_this = (Entity_CA*)param;
	time_t currTime;
	time_t lastTime;
	STACK_OF(INTERNAL_CA_CERT) * certs;
	STACK_OF(X509_CRL) * crls;
	long index;
	int i;
	INTERNAL_CA_CERT * currCert;
	X509_CRL * currCrl;
	PKI_CERT cert;
	PKI_CRL crl;

	//We wait a bit (2 mins)
	time(&lastTime);
	do
	{
		NewpkiThread::Sleep(500);
		time(&currTime);
	}
	while( ((currTime - lastTime) < 120) && !Thread->ShouldStop());

	// We resend all the known certificates/revocation/crl every 12 hours
	while(!Thread->ShouldStop())
	{

		index = 0;
		do
		{
			certs = me_this->m_InternalCa->get_Certs(CERT_STATE_VALID, index, 50);
			if(certs)
			{
				index += SKM_sk_num(INTERNAL_CA_CERT, certs);
				for(i=0; !Thread->ShouldStop() && i<SKM_sk_num(INTERNAL_CA_CERT, certs); i++)
				{
					currCert = SKM_sk_value(INTERNAL_CA_CERT, certs, i);
					if(!currCert) continue;
					if(!cert.SetCert(currCert->cert)) continue;
					me_this->PublishCert(NULL, cert);
				}
				SKM_sk_pop_free(INTERNAL_CA_CERT, certs, INTERNAL_CA_CERT_free);
			}
		}
		while(!Thread->ShouldStop() && certs && SKM_sk_num(INTERNAL_CA_CERT, certs));

		index = 0;
		do
		{
			certs = me_this->m_InternalCa->get_Certs(CERT_STATE_REVOKED, index, 50);
			if(certs)
			{
				index += SKM_sk_num(INTERNAL_CA_CERT, certs);
				for(i=0; !Thread->ShouldStop() && i<SKM_sk_num(INTERNAL_CA_CERT, certs); i++)
				{
					currCert = SKM_sk_value(INTERNAL_CA_CERT, certs, i);
					if(!currCert) continue;
					if(!cert.SetCert(currCert->cert)) continue;
					me_this->PublishRevocation(NULL, cert, ASN1_INTEGER_GET(currCert->rev_date));
				}
				SKM_sk_pop_free(INTERNAL_CA_CERT, certs, INTERNAL_CA_CERT_free);
			}
		}
		while(!Thread->ShouldStop() && certs && SKM_sk_num(INTERNAL_CA_CERT, certs));


		index = 0;
		do
		{
			crls = me_this->m_InternalCa->get_Crls(index, 50);
			if(crls)
			{
				index += SKM_sk_num(X509_CRL, crls);
				for(i=0; !Thread->ShouldStop() && i<SKM_sk_num(X509_CRL, crls); i++)
				{
					currCrl = SKM_sk_value(X509_CRL, crls, i);
					if(!currCrl) continue;
					if(!crl.SetCRL(currCrl)) continue;
					me_this->PublishCrl(crl);
				}
				SKM_sk_pop_free(X509_CRL, crls, X509_CRL_free);
			}
		}
		while(!Thread->ShouldStop() && crls && SKM_sk_num(X509_CRL, crls));

		time(&lastTime);
		do
		{
			NewpkiThread::Sleep(500);
			time(&currTime);
		}
		while( ((currTime - lastTime) < (12 * 60 * 60)) && !Thread->ShouldStop());
	}
}


void Entity_CA::ThreadGenerateCRL(const NewpkiThread * Thread, void *param)
{
	Entity_CA * me_this = (Entity_CA*)param;
	PKI_CRL lastCRL;
	time_t currTime;
	long crldays;
	long crlhours;
	char * err;
	char * strerr;

	if(!me_this) return;

	while(!Thread->ShouldStop())
	{
		//Get CA conf
		me_this->ConfAccessLock.LockRead();
		//Do we have a conf ?
		if(!me_this->myConf)
		{
			me_this->ConfAccessLock.Unlock();
			goto wait_next;
		}
		//We get validity length
		if(!me_this->m_crldays && !me_this->m_crlhours)
		{
			me_this->ConfAccessLock.Unlock();
			goto wait_next;
		}
		crldays = me_this->m_crldays;
		crlhours = me_this->m_crlhours;

		me_this->ConfAccessLock.Unlock();

		//We get last CRL
		if(!me_this->m_InternalCa->GetLastCrl(lastCRL))
			goto wait_next;

		
		time_gmt(&currTime);

		//Is the last CRL about to expire ?
		//we add a tolerance of 10% before the actual
		//crl expiration, this should be enough to
		//do the generation and the publication
		if( ((lastCRL.GetEndTime() - ((lastCRL.GetEndTime() - lastCRL.GetStartTime()) * 0.1)) > currTime) || (lastCRL.GetEndTime() - currTime) > (crlhours*3600 + crldays * 86400) )
			goto wait_next;

		me_this->m_Logging->LogMessage(LOG_STATUS_TYPE_REQUEST, LOG_MESSAGE_TYPE_GEN_CRL, 0, me_this->m_EntityCert.GetStringName(), LOG_NO_OBJECTID, "CRL");

		if(!me_this->Private_GenerateCRL(lastCRL))
		{
			err = ERR_to_string();
			if(err)
				strerr = err;
			else
				strerr = "";
			me_this->m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_GEN_CRL, 0, me_this->m_EntityCert.GetStringName(), LOG_NO_OBJECTID, "CRL", strerr);
			if(err)
				free(err);
		}
		else
		{
			me_this->m_Logging->LogMessage(LOG_STATUS_TYPE_SUCCESS, LOG_MESSAGE_TYPE_GEN_CRL, 0, me_this->m_EntityCert.GetStringName(), LOG_NO_OBJECTID, "CRL");
		}

wait_next:
		NewpkiThread::Sleep(800);
	}
}


bool Entity_CA::Private_GenerateCRL(PKI_CRL & Crl)
{
	m_CrlExts.Lock();
	if(!m_InternalCa->Generate_CRL(Crl, m_crldays, m_crlhours, &m_CrlExts, "sha1"))
	{
		m_CrlExts.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	m_CrlExts.Unlock();

	if(!OnNewCRL(Crl))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}


bool Entity_CA::Private_SignCSR(const char * ldap_uid, X509_REQ * reqCSR, unsigned long validity, PKI_CERT & Cert)
{
	PKI_CSR csr;
	CERT_STATE CertStatus;

	if(!csr.SetCSR(reqCSR))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!csr.CheckSignature())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_CSR_SIGNATURE);
		return false;
	}

	if(!OnNewRequest(csr))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	m_Exts.Lock();
	if(!m_InternalCa->Sign(csr, Cert, validity, &m_Exts, true))
	{
		m_Exts.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	m_Exts.Unlock();

	if(!OnNewCertificate(ldap_uid, Cert))
	{
		Private_RevokeCERT(ldap_uid, Cert.GetSerial(), CertStatus);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_CA::Private_RevokeCERT(const char * ldap_uid, unsigned long serial, CERT_STATE & CertStatus)
{
	PKI_CERT Cert;
	time_t rev_date;
	if(!m_InternalCa->Revoke(serial, Cert, CertStatus, rev_date))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!OnNewRevocation(ldap_uid, Cert, rev_date))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

void Entity_CA::ThreadWorker(const NewpkiThread * Thread, void *param)
{
	Entity_CA * me_this = (Entity_CA *)param;
	time_t currTime;
	time_t lastTime;
	WORKING_CA_REQUEST * workingReq;
	LOG_MESSAGE_TYPE LogType;
	mString TransactionId;
	mString StrErr;
	STACK_OF(WORKING_CA_REQUEST) * WorkingReqs = NULL;


	//We wait a bit (2 mins)
	time(&lastTime);
	do
	{
		NewpkiThread::Sleep(500);
		time(&currTime);
	}
	while( ((currTime - lastTime) < 120) && !Thread->ShouldStop());


	do
	{
		NewpkiThread::Sleep(500);
		WorkingReqs = SKM_sk_new_null(WORKING_CA_REQUEST);
	}
	while( !WorkingReqs && !Thread->ShouldStop());


	while(!Thread->ShouldStop())
	{
		do
		{
			NewpkiThread::Sleep(1000);
		}
		while(!me_this->CaPublicationStore.Responder_GetWaitingRequests(WorkingReqs) && !Thread->ShouldStop());

		//I have a new request
		while( (workingReq = SKM_sk_pop(WORKING_CA_REQUEST, WorkingReqs))  && !Thread->ShouldStop())
		{
			if(!NewPKIStore::transactionIDtoString(workingReq->transactionID, TransactionId))
				TransactionId = "Unknown";

			StrErr = "";

			//Start to proceed with request
			if(me_this->ProceedWithRequest(workingReq, LogType, StrErr))
			{
				if(StrErr.size())
					me_this->m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LogType, 0, ASN1_STRING_GET(workingReq->SenderName), LOG_NO_OBJECTID, TransactionId.c_str(), StrErr.c_str());
				else
					me_this->m_Logging->LogMessage(LOG_STATUS_TYPE_SUCCESS, LogType, 0, ASN1_STRING_GET(workingReq->SenderName), LOG_NO_OBJECTID, TransactionId.c_str());
			}
			WORKING_CA_REQUEST_free(workingReq);
		}
	}

	if(WorkingReqs)
		SKM_sk_pop_free(WORKING_CA_REQUEST, WorkingReqs, WORKING_CA_REQUEST_free);
}

bool Entity_CA::ProceedWithCertification(NEWPKI_CERT_REQUEST * cert_request, NEWPKI_RESPONSE * currResponse)
{
	char * ldap_uid;
	PKI_CERT cert;
	X509 * currCert;
	int i;
	PKI_CERT * myCert;
	PKI_CRL Crl;
	CERT_STATE CertStatus;

	ASN1_INTEGER_SET(&currResponse->d.cert_response->id, ASN1_INTEGER_GET(cert_request->id));

	ldap_uid = ASN1_STRING_GET(cert_request->ldap_uid);

	if(currResponse->d.cert_response->last_crl)
	{
		X509_CRL_free(currResponse->d.cert_response->last_crl);
		currResponse->d.cert_response->last_crl = NULL;
	}
	// The CRL is optional, so we don't trigger an error
	if(m_InternalCa->GetLastCrl(Crl))
	{
		currResponse->d.cert_response->last_crl = Crl.GetX509_CRL(true);
		if(!currResponse->d.cert_response->last_crl)
			ERR_clear_error();

		Crl.Clear();
	}
	else
	{
		ERR_clear_error();
	}

	
	//Proceed with signature
	if(!Private_SignCSR(ldap_uid, cert_request->request, ASN1_INTEGER_GET(cert_request->validity), cert))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		goto certification_err;
	}

	// Build the answer
	
	if(currResponse->d.cert_response->certificate)
		X509_free(currResponse->d.cert_response->certificate);
	currResponse->d.cert_response->certificate = cert.GetX509(true);
	if(!currResponse->d.cert_response->certificate)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		goto certification_err;
	}

	currResponse->d.cert_response->ParentCerts = SKM_sk_new_null(X509);
	if(!currResponse->d.cert_response->ParentCerts)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		goto certification_err;
	}

	//Copy the parent certs
	if(c_myConf->parentCerts)
	{		
		for(i=0; i<SKM_sk_num(X509, c_myConf->parentCerts); i++)
		{
			// Get the current cert
			currCert = SKM_sk_value(X509, c_myConf->parentCerts, i);
			if(!currCert)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
				goto certification_err;
			}
			// Duplicate it
			CRYPTO_add(&currCert->references,1,CRYPTO_LOCK_X509);
			// Add it to the returned list
			if(SKM_sk_push(X509, currResponse->d.cert_response->ParentCerts, currCert) < 0)
			{
				X509_free(currCert);
				goto certification_err;
			}
		}
	}

	// Add my own cert as well
	myCert = m_InternalCa->get_CaCert();
	if(!myCert)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		goto certification_err;
	}
	currCert = myCert->GetX509();
	if(!currCert)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		goto certification_err;
	}
	// Duplicate it
	CRYPTO_add(&currCert->references,1,CRYPTO_LOCK_X509);
	// Add it to the returned list
	if(SKM_sk_push(X509, currResponse->d.cert_response->ParentCerts, currCert) < 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		X509_free(currCert);
		goto certification_err;
	}

	ASN1_INTEGER_SET(&currResponse->d.cert_response->status, CERT_RESPONSE_OK);

	return true;



certification_err:
	ASN1_INTEGER_SET(&currResponse->d.cert_response->status, CERT_RESPONSE_ERROR);
	currResponse->d.cert_response->errors = ERR_to_ERROR_ENTRIES();
	if(currResponse->d.cert_response->certificate)
	{
		X509_free(currResponse->d.cert_response->certificate);
		currResponse->d.cert_response->certificate = NULL;
	}
	if(currResponse->d.cert_response->ParentCerts)
	{
		SKM_sk_pop_free(X509, currResponse->d.cert_response->ParentCerts, X509_free);
		currResponse->d.cert_response->ParentCerts = NULL;
	}
	if(cert)
		Private_RevokeCERT(ldap_uid, cert.GetSerial(), CertStatus);

	return false;
}


bool Entity_CA::ProceedWithRequest(WORKING_CA_REQUEST * req, LOG_MESSAGE_TYPE & LogType, mString & Err)
{
	NEWPKI_RESPONSE * currResponse;
	char * StrError;
	char * ldap_uid;
	CERT_STATE CertStatus;
	PKI_CRL Crl;

	currResponse = NEWPKI_RESPONSE_new();
	if(!currResponse)
	{
		return false;
	}

	//Proceed with the request
	switch(req->request->type)
	{
		case NEWPKI_REQUEST_TYPE_REV:
			LogType = LOG_MESSAGE_TYPE_CA_REV;

			currResponse->type = NEWPKI_RESPONSE_TYPE_REV;
			currResponse->d.rev_response = NEWPKI_REV_RESPONSE_new();
			if(!currResponse->d.rev_response)
			{
				NEWPKI_RESPONSE_free(currResponse);
				return false;
			}

			if(currResponse->d.rev_response->last_crl)
			{
				X509_CRL_free(currResponse->d.rev_response->last_crl);
				currResponse->d.rev_response->last_crl = NULL;
			}
			// The CRL is optional, so we don't trigger an error
			if(m_InternalCa->GetLastCrl(Crl))
			{
				currResponse->d.rev_response->last_crl = Crl.GetX509_CRL(true);
				if(!currResponse->d.rev_response->last_crl)
					ERR_clear_error();

				Crl.Clear();
			}
			else
			{
				ERR_clear_error();
			}

			ldap_uid = ASN1_STRING_GET(req->request->d.rev_request->ldap_uid);

			ASN1_INTEGER_SET(&currResponse->d.rev_response->id, ASN1_INTEGER_GET(req->request->d.rev_request->id));

			if(!Private_RevokeCERT(ldap_uid, ASN1_INTEGER_GET(req->request->d.rev_request->serial), CertStatus))
			{
				ASN1_INTEGER_SET(&currResponse->d.rev_response->status, REV_RESPONSE_ERROR);
				currResponse->d.rev_response->errors = ERR_to_ERROR_ENTRIES();
			}
			else
			{
				ASN1_INTEGER_SET(&currResponse->d.rev_response->status, REV_RESPONSE_OK);
			}
			ASN1_INTEGER_SET(&currResponse->d.rev_response->cert_status, CertStatus);
			break;

		case NEWPKI_REQUEST_TYPE_CERT:
			LogType = LOG_MESSAGE_TYPE_CA_CERT;

			currResponse->type = NEWPKI_RESPONSE_TYPE_CERT;
			currResponse->d.cert_response = NEWPKI_CERT_RESPONSE_new();
			if(!currResponse->d.cert_response)
			{
				NEWPKI_RESPONSE_free(currResponse);
				return false;
			}

			if(!ProceedWithCertification(req->request->d.cert_request, currResponse))
			{
				StrError = ERR_to_string();
				if(StrError)
				{
					Err = StrError;
					free(StrError);
				}
				else
				{
					Err = _sv("Unknown");
				}
			}
			break;

		default:
			LogType = LOG_MESSAGE_TYPE_CA_CERT;

			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			ERR_to_STORED_NEWPKI_RESPONSE(CaPublicationStore, req->transactionID, req->sender);

			StrError = ERR_to_string();
			if(StrError)
			{
				Err = StrError;
				free(StrError);
			}
			else
			{
				Err = _sv("Unknown");
			}
			return false;
			break;
	}

	SendResponse(CaPublicationStore, req->transactionID, currResponse, req->sender);
	NEWPKI_RESPONSE_free(currResponse);
	
	return true;
}


bool Entity_CA::GetStatus(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(AclValidator.ValidateCert(UserCert) != INTERNAL_CA_TYPE_USER)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	ASN1_INTEGER * status = NULL;

	if(*m_InternalCa)
	{
		ASN1_INTEGER_SET(&status, STATUS_CA_OK);
	}
	else if(c_myConf->privkey)
	{
		ASN1_INTEGER_SET(&status, STATUS_CA_AWAITING_CERT);
	}
	else
	{
		ASN1_INTEGER_SET(&status, STATUS_CA_NO_CERT);
	}
	
	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_STATUS, status))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}


bool Entity_CA::GetCaCert(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(AclValidator.ValidateCert(UserCert) != INTERNAL_CA_TYPE_USER)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!*m_InternalCa)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_CA_NOT_CONFIGURED);
		return false;
	}
	
	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_CERT, m_InternalCa->get_CaCert()->GetX509(true)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}


bool Entity_CA::CreateRootCa(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_CREATE_ROOT_CA))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	//The internal CA has already been generated
	if(*m_InternalCa)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	INTERNAL_CA_KEY * privKey;
	PKI_RSA tmpPrivKey;
	PKI_CERT tmpCert;
	PKI_CSR tmpCsr;
	HashTable_String Exts;
	HashTable_Dn Dn;

	//Load the extensions
	if(!Exts.From_EXTENSION_VALUE(body->d.create_root_ca->extensions))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}

	//Load the DN
	if(!Dn.From_X509_NAME(body->d.create_root_ca->dn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}


	//Load/generate private key
	privKey = INTERNAL_CA_KEY_new();
	if(!privKey)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	if(!CA_Handler::GEN_PRIVATE_KEY_load(body->d.create_root_ca->privKey, tmpPrivKey, privKey, m_Engine))
	{
		INTERNAL_CA_KEY_free(privKey);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//We generate the CSR
	if(!tmpCsr.GenerateCSR(Dn, tmpPrivKey))
	{
		INTERNAL_CA_KEY_free(privKey);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//We now generate the certificate
	if(!tmpCert.CreateSelfSigned(tmpCsr, &Exts, ASN1_INTEGER_GET(body->d.create_root_ca->validity), 0))
	{
		INTERNAL_CA_KEY_free(privKey);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!tmpCert.SetPrivateKey(tmpPrivKey))
	{
		INTERNAL_CA_KEY_free(privKey);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//Initialize the CA
	if(!m_InternalCa->Create(tmpCert, privKey))
	{
		INTERNAL_CA_KEY_free(privKey);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	INTERNAL_CA_KEY_free(privKey);

	if(!m_InternalCa->Load())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_CA::CreateChildCa(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_CREATE_CHILD_CA))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	//The internal CA has already been generated
	if(*m_InternalCa)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	PKI_RSA tmpPrivKey;
	PKI_CSR tmpCsr;
	HashTable_Dn Dn;

	//Load the DN
	if(!Dn.From_X509_NAME(body->d.create_child_ca->dn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}


	//Load/generate private key
	ConfAccessLock.LockWrite();

	//Did we already generate the key ?
	if(c_myConf->privkey)
	{
		ConfAccessLock.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}		
	c_myConf->privkey = INTERNAL_CA_KEY_new();
	if(!c_myConf->privkey)
	{
		ConfAccessLock.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	if(!CA_Handler::GEN_PRIVATE_KEY_load(body->d.create_child_ca->privKey, tmpPrivKey, c_myConf->privkey, m_Engine))
	{
		INTERNAL_CA_KEY_free(c_myConf->privkey);
		c_myConf->privkey = NULL;
		ConfAccessLock.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//We generate the CSR
	if(!tmpCsr.GenerateCSR(Dn, tmpPrivKey))
	{
		ConfAccessLock.Unlock();
		INTERNAL_CA_KEY_free(c_myConf->privkey);
		c_myConf->privkey = NULL;
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//Wa save the new key in database
	if(!WritePersonnalConf())
	{
		ConfAccessLock.Unlock();
		INTERNAL_CA_KEY_free(c_myConf->privkey);
		c_myConf->privkey = NULL;
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	ConfAccessLock.Unlock();


	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_CSR, tmpCsr.GetX509_REQ(true)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}


bool Entity_CA::ImportChildCaCert(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	PKI_P7B p7b;
	int i;
	X509 * currCert;
	PKI_RSA tmpPrivKey;
	PKI_CERT tmpCert;
	X509 * myCert;
	
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_CREATE_ROOT_CA))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	//The internal CA has already been generated
	if(*m_InternalCa)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	ConfAccessLock.LockWrite();
	//Do we have the key ?
	if(!c_myConf->privkey)
	{
		ConfAccessLock.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	// We load the P7B
	if(!p7b.Load(body->d.p7b))
	{
		ConfAccessLock.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	//Load saved private key 
	if(!CA_Handler::INTERNAL_CA_KEY_load(c_myConf->privkey, tmpPrivKey, m_Engine))
	{
		ConfAccessLock.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	

	// Empty the parent cert list from a previous attempt
	while( (currCert = SKM_sk_pop(X509, c_myConf->parentCerts)) )
	{
		X509_free(currCert);
	}
	
	// Find our certificate and our parent's
	myCert = NULL;
	for(i=0; i<p7b.GetNumCert(); i++)
	{
		tmpCert.Clear();
		
		currCert = p7b.GetCert(i);
		if(!currCert)
		{
			ConfAccessLock.Unlock();
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	
		//Load certificate
		if(!tmpCert.SetCert(currCert))
		{
			ConfAccessLock.Unlock();
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		
		// Dothe two pubkeys correspond ?
		if(tmpCert == tmpPrivKey.GetPublicKey())
		{
			myCert = currCert;
		}
		else
		{
			// Insert the parent cert in list
			CRYPTO_add(&currCert->references,1,CRYPTO_LOCK_X509);
			if(SKM_sk_push(X509, c_myConf->parentCerts, currCert) < 0)
			{
				ConfAccessLock.Unlock();
				NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
				X509_free(currCert);
				return false;
			}
		}
	}
	ERR_clear_error();
	
	// Didn't find my cert !!!
	if(!myCert)
	{
		ConfAccessLock.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}
	// initialize my cert
	if(!tmpCert.SetCert(myCert))
	{
		ConfAccessLock.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!tmpCert.SetPrivateKey(tmpPrivKey))
	{
		ConfAccessLock.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	
	//Initialize the CA
	if(!m_InternalCa->Create(tmpCert, c_myConf->privkey))
	{
		ConfAccessLock.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	
	//Update conf without the key
	INTERNAL_CA_KEY_free(c_myConf->privkey);
	c_myConf->privkey = NULL;
	if(!WritePersonnalConf())
	{
		ConfAccessLock.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	ConfAccessLock.Unlock();

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}


bool Entity_CA::EnumCerts(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	STACK_OF(INTERNAL_CA_CERT) * certs;
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	switch(ASN1_INTEGER_GET(body->d.enum_objects->state))
	{
		case CERT_STATE_VALID:
			if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_CA_REVOKE_CERT) && !AclValidator.CanUserPerform(UserCert, ACL_TYPE_CA_SIGN_CERT))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
				return false;
			}
			break;
		case CERT_STATE_REVOKED:
			if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_CA_REVOKE_CERT))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
				return false;
			}
			break;
		default:
			NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
			return false;
			break;
	}

	if(!*m_InternalCa)
	{
		certs = SKM_sk_new_null(INTERNAL_CA_CERT);
		if(!certs)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
			return false;
		}
	}
	else
	{
		certs = m_InternalCa->get_Certs((CERT_STATE)ASN1_INTEGER_GET(body->d.enum_objects->state), ASN1_INTEGER_GET(body->d.enum_objects->index), ASN1_INTEGER_GET(body->d.enum_objects->num));
		if(!certs)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}


	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_CERTS, certs))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		SKM_sk_pop_free(INTERNAL_CA_CERT, certs, INTERNAL_CA_CERT_free);
		return false;
	}

	return true;
}

bool Entity_CA::EnumCrls(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	STACK_OF(X509_CRL) * crls;
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!*m_InternalCa)
	{
		crls = SKM_sk_new_null(X509_CRL);
		if(!crls)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
			return false;
		}
	}
	else
	{
		crls = m_InternalCa->get_Crls(ASN1_INTEGER_GET(body->d.enum_objects->index), ASN1_INTEGER_GET(body->d.enum_objects->num));
		if(!crls)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}


	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_CRLS, crls))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		SKM_sk_pop_free(X509_CRL, crls, X509_CRL_free);
		return false;
	}

	return true;
}

bool Entity_CA::SignCSR(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	PKI_CERT cert;
	PKI_P7B p7b;
	X509 * currCert;
	int i;
	PKI_CERT * parentCert;
	CERT_STATE CertStatus;

	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_CA_SIGN_CERT))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!*m_InternalCa)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_CA_NOT_CONFIGURED);
		return false;
	}

	if(!Private_SignCSR(NULL, body->d.sign_csr->request, ASN1_INTEGER_GET(body->d.sign_csr->days), cert))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	
	// Let's generate the P7B
	ConfAccessLock.LockRead();
	if(c_myConf->parentCerts)
	{
		// We first add all the parents
		for(i=0; i<SKM_sk_num(X509, c_myConf->parentCerts); i++)
		{
			currCert = SKM_sk_value(X509, c_myConf->parentCerts, i);
			if(!currCert)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
				ConfAccessLock.Unlock();
				Private_RevokeCERT(NULL, cert.GetSerial(), CertStatus);
				return false;
			}
			if(!p7b.AddCert(currCert))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				ConfAccessLock.Unlock();
				Private_RevokeCERT(NULL, cert.GetSerial(), CertStatus);
				return false;
			}
		}
	}	
	ConfAccessLock.Unlock();
	
	// We now add this CA cert
	parentCert = m_InternalCa->get_CaCert();
	if(!parentCert)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		Private_RevokeCERT(NULL, cert.GetSerial(), CertStatus);
		return false;
	}
	if(!p7b.AddCert(parentCert->GetX509()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Private_RevokeCERT(NULL, cert.GetSerial(), CertStatus);
		return false;
	}
	
	// We now add the final user cert !
	if(!p7b.AddCert(cert.GetX509()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Private_RevokeCERT(NULL, cert.GetSerial(), CertStatus);
		return false;
	}
	
	// We now really generate the P7B
	if(!p7b.Generate())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Private_RevokeCERT(NULL, cert.GetSerial(), CertStatus);
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_P7B, p7b.GetPKCS7(true)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Private_RevokeCERT(NULL, cert.GetSerial(), CertStatus);
		return false;
	}

	return true;
}


bool Entity_CA::RevokeCert(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	CERT_STATE CertStatus;

	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_CA_REVOKE_CERT))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!*m_InternalCa)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_CA_NOT_CONFIGURED);
		return false;
	}

	if(!Private_RevokeCERT(NULL, ASN1_INTEGER_GET(body->d.serial), CertStatus))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}


bool Entity_CA::OnNewRequest(const PKI_CSR & Csr)
{
	if(m_PlugCA)
	{
		m_PlugOptions.Lock();
		if(!m_PlugCA->OnNewRequest(m_PlugOptions, m_EntityCert, m_Jobs.GetMailer(), Csr))
		{
			m_PlugOptions.Unlock();
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		m_PlugOptions.Unlock();
	}
	return true;
}


bool Entity_CA::OnNewCertificate(const char * ldap_uid, const PKI_CERT & Cert)
{
	if(m_PlugCA)
	{
		m_PlugOptions.Lock();
		if(!m_PlugCA->OnNewCertificate(m_PlugOptions, m_EntityCert, m_Jobs.GetMailer(), Cert))
		{
			m_PlugOptions.Unlock();
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		m_PlugOptions.Unlock();
	}

	if(!PublishCert(ldap_uid, Cert))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}


bool Entity_CA::OnNewRevocation(const char * ldap_uid, const PKI_CERT & Cert, time_t rev_date)
{
	if(m_PlugCA)
	{
		m_PlugOptions.Lock();
		if(!m_PlugCA->OnNewRevocation(m_PlugOptions, m_EntityCert, m_Jobs.GetMailer(), Cert.GetSerial()))
		{
			m_PlugOptions.Unlock();
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		m_PlugOptions.Unlock();
	}

	if(!PublishRevocation(ldap_uid, Cert, rev_date))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}


bool Entity_CA::OnNewCRL(const PKI_CRL & crl)
{
	if(m_PlugCA)
	{
		m_PlugOptions.Lock();
		if(!m_PlugCA->OnNewCRL(m_PlugOptions, m_EntityCert, m_Jobs.GetMailer(), crl))
		{
			m_PlugOptions.Unlock();
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		m_PlugOptions.Unlock();
	}

	if(!PublishCrl(crl))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

void Entity_CA::SpawnStop()
{
	hThreadWorker.SpawStop();
	hThreadGenerateCRL.SpawStop();
	hThreadWorker.SpawStop();
	m_Jobs.SpawnStop();
}

bool Entity_CA::OnNewConf(const ENTITY_CONF_CRYPTED * newConf)
{
	if(!ExtractMyConf(newConf, &myConf, NULL, c_myConf->cas->PkiCert, true))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_CA::OnNewRequests(const STORED_NEWPKI_REQUESTS * reqs)
{
	STORED_NEWPKI_REQUEST * currReq;
	int i;
	mString Sender;
	mString TransactionId;
	char * Err;
	char * strErr;
	bool ret;
	bool IsKnown;

	// We don't continue if we have no requests
	// For each transaction, search if it's not being processed
	// or has been processed yet
	for(i=0; i<SKM_sk_num(STORED_NEWPKI_REQUEST, reqs->requests); i++ )
	{
		currReq = SKM_sk_value(STORED_NEWPKI_REQUEST, reqs->requests, i);
		if(!currReq) continue;

		// Is this transaction known ?
		if(!CaPublicationStore.Responder_IsRequestKnown(currReq->transactionID, IsKnown))
			continue;
		if(IsKnown)
			continue;


		// Create a working request from this request
		ret = AddNewSTORED_NEWPKI_REQUEST(currReq, Sender);

		if(!NewPKIStore::transactionIDtoString(currReq->transactionID, TransactionId))
			TransactionId = "Unknown";

		if(!Sender.size())
			Sender = _sv("Unknown");

		if(!ret)
		{
			if(! (Err = ERR_to_string()) )
				strErr= _sv("Unknown");
			else
				strErr = Err;

			m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_NEW_REQUET, 0, Sender.c_str(), LOG_NO_OBJECTID, TransactionId.c_str(), strErr);

			if(Err)
				free(Err);
		}
		else
		{
			m_Logging->LogMessage(LOG_STATUS_TYPE_SUCCESS, LOG_MESSAGE_TYPE_NEW_REQUET, 0, Sender.c_str(), LOG_NO_OBJECTID, TransactionId.c_str());
		}
	}

	return true;
}

bool Entity_CA::OnNewResponses(const STORED_NEWPKI_RESPONSES * resps)
{
	STORED_NEWPKI_RESPONSE * currResp;
	NEWPKI_RESPONSE * resp;
	EVP_PKEY * signerkey;
	int i, j;
	LOG_MESSAGE_TYPE LogType;
	LOG_MESSAGE_STATUS Status;
	char * StrError;
	mString Error;
	mString ObjectName;
	mString SenderName;
	PUB_ENTRY_INFO * currPub;

	for(i=0; i<SKM_sk_num(STORED_NEWPKI_RESPONSE, resps->responses); i++)
	{
		currResp = SKM_sk_value(STORED_NEWPKI_RESPONSE, resps->responses, i);
		if(!currResp) continue;

		CaPublicationStore.Requester_RequestProcessed(currResp->transactionID);

		signerkey = X509_PUBKEY_get(currResp->c_info->sender);
		if(!signerkey) continue;

		if(!NEWPKI_RESPONSE_decrypt_verify(currResp->c_info->c_response, &resp, signerkey, m_EntityKey.GetRsaKey()))
		{
			EVP_PKEY_free(signerkey);
			continue;
		}
		EVP_PKEY_free(signerkey);

		switch(resp->type)
		{
			case NEWPKI_RESPONSE_TYPE_PUB:
				ConfAccessLock.LockRead();

				// Now search the Pub
				SenderName = _sv("Unknown");
				for(j=0; j < SKM_sk_num(PUB_ENTRY_INFO, myConf->Publications->list); j++)
				{
					currPub = SKM_sk_value(PUB_ENTRY_INFO, myConf->Publications->list, j);
					if(!currPub) continue;

					// Are the public keys the same ?
					if(currPub->PubSSL->cert_info->key->public_key->length ==
						currResp->c_info->sender->public_key->length &&
						memcmp(currPub->PubSSL->cert_info->key->public_key->data,
						currResp->c_info->sender->public_key->data, 
						currResp->c_info->sender->public_key->length) == 0)
					{
						SenderName = currPub->Name;
						break;
					}
				}
				ConfAccessLock.Unlock();

				switch(ASN1_INTEGER_GET(resp->d.pub_response->type))
				{
					case NEWPKI_PUB_REQUEST_CERT:
						LogType = LOG_MESSAGE_TYPE_CERT_PUBLICATION;
						break;
					case NEWPKI_PUB_REQUEST_REV:
						LogType = LOG_MESSAGE_TYPE_REV_PUBLICATION;
						break;
					case NEWPKI_PUB_REQUEST_CRL:
						LogType = LOG_MESSAGE_TYPE_CRL_PUBLICATION;
						break;
					default:
						NEWPKI_RESPONSE_free(resp);
						continue;
				}
				if(ASN1_INTEGER_GET(resp->d.pub_response->status) == PUB_RESPONSE_PUBLISHED)
				{	
					Status = LOG_STATUS_TYPE_SUCCESS;
				}
				else
				{
					Status = LOG_STATUS_TYPE_FAILURE;
					StrError = ERROR_ENTRIES_to_string(resp->d.pub_response->errors);
					if(StrError)
					{
						Error = StrError;
						free(StrError);
					}
					else
					{
						Error = "Unknown Error";
					} // if(ASN1_INTEGER_GET(resp->d.pub_response->type) == PUB_RESPONSE_PUBLISHED)
				} //switch(ASN1_INTEGER_GET(resp->d.pub_response->type))

				ObjectName = resp->d.pub_response->object;
				break;
			
			default:
				NEWPKI_RESPONSE_free(resp);
				continue;
		} // switch(resp->type)

		if(Status == LOG_STATUS_TYPE_SUCCESS)
		{
			m_Logging->LogMessage(LOG_STATUS_TYPE_SUCCESS, LogType, 0, SenderName.c_str(), LOG_NO_OBJECTID, ObjectName.c_str());
		}
		else if(Status == LOG_STATUS_TYPE_FAILURE)
		{
			m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LogType, 0, SenderName.c_str(), LOG_NO_OBJECTID, ObjectName.c_str(), Error.c_str());
		}

		NEWPKI_RESPONSE_free(resp);
	}

	return true;
}


bool Entity_CA::OnResponseSent(const STORED_NEWPKI_RESPONSE * resp)
{
	return CaPublicationStore.Responder_ResponseSent(resp->transactionID);
}

bool Entity_CA::OnRequestSent(unsigned long RequestId, const ASN1_OCTET_STRING * transactionId, const CRYPTED_NEWPKI_REQUEST * req)
{
	return CaPublicationStore.Requester_RequestSent(RequestId, transactionId);
}

bool Entity_CA::GetWaitingTransactionIds(TRANSACTION_IDS *transactionIDs)
{
	return CaPublicationStore.Requester_GetWaitingTransactionIds(transactionIDs);
}

bool Entity_CA::GetProcessedTransactionIds(TRANSACTION_IDS *transactionIDs)
{
	return CaPublicationStore.Responder_GetTransactionIds(transactionIDs);
}

bool Entity_CA::GetWaitingRequests(STACK_OF(CRYPTED_NEWPKI_REQUEST) * Requests)
{
	return CaPublicationStore.Requester_GetWaitingRequests(Requests);
}

bool Entity_CA::GetWaitingResponses(STACK_OF(STORED_NEWPKI_RESPONSE) * Responses)
{
	return CaPublicationStore.Responder_GetUnsentResponses(Responses);
}

bool Entity_CA::AddNewSTORED_NEWPKI_REQUEST(STORED_NEWPKI_REQUEST * req, mString & Sender)
{
	NEWPKI_REQUEST * Request;
	WORKING_CA_REQUEST * w_req;
	EVP_PKEY * signerkey;


	//Verify the request
	if(!STORED_NEWPKI_REQUEST_check(req, Sender))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ERR_to_STORED_NEWPKI_RESPONSE(CaPublicationStore, req->transactionID, req->c_info->sender);
		return false;
	}

	w_req = WORKING_CA_REQUEST_new();
	if(!w_req)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	// Copy the sender's name
	ASN1_STRING_SET(&w_req->SenderName, Sender.c_str());
	
	//Copy transaction ID
	if(!ASN1_OCTET_STRING_cpy(&w_req->transactionID, req->transactionID))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		WORKING_CA_REQUEST_free(w_req);
		return false;
	}
	//Copy the sender public key
	if(!X509_PUBKEY_cpy(&w_req->sender, req->c_info->sender))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		WORKING_CA_REQUEST_free(w_req);
		return false;
	}

	signerkey = X509_PUBKEY_get(req->c_info->sender);
	if(!signerkey)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		WORKING_CA_REQUEST_free(w_req);
		return false;
	}

	// Decrypt the request
	if(NEWPKI_REQUEST_decrypt_verify(req->c_info->c_request, &Request, signerkey, m_EntityKey.GetRsaKey()) <= 0)
	{
		WORKING_CA_REQUEST_free(w_req);
		EVP_PKEY_free(signerkey);
		//We generate an error
		ERR_clear_error();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		ERR_to_STORED_NEWPKI_RESPONSE(CaPublicationStore, req->transactionID, req->c_info->sender);
		return false;
	}
	EVP_PKEY_free(signerkey);

	if(w_req->request)
		NEWPKI_REQUEST_free(w_req->request);
	w_req->request = Request;

	ERR_clear_error();
	if(!CaPublicationStore.Responder_InsertRequest(w_req))
	{
		WORKING_CA_REQUEST_free(w_req);
		//We generate an error
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ERR_to_STORED_NEWPKI_RESPONSE(CaPublicationStore, req->transactionID, req->c_info->sender);
		return false;
	}
	WORKING_CA_REQUEST_free(w_req);

	return true;
}

bool Entity_CA::STORED_NEWPKI_REQUEST_check(STORED_NEWPKI_REQUEST * req, mString & Sender)
{
	int i;
	RA_ENTRY_INFO * currRa;

	// We need to search that the RA has the right 
	// to request a certificate from us
	for(i=0; i < SKM_sk_num(RA_ENTRY_INFO, myConf->Ras->list); i++)
	{
		currRa = SKM_sk_value(RA_ENTRY_INFO, myConf->Ras->list, i);
		if(!currRa) continue;

		//Is it the same public key
		if(currRa->RaSSL->cert_info->key->public_key->length ==
			req->c_info->sender->public_key->length &&
			memcmp(currRa->RaSSL->cert_info->key->public_key->data,
			req->c_info->sender->public_key->data, 
			req->c_info->sender->public_key->length) == 0)
		{
			Sender = currRa->Name;
			break;
		}
	}
	if(i == SKM_sk_num(RA_ENTRY_INFO, myConf->Ras->list))
	{
		//This RA is not allowed to request a certificate
		//from us
		return false;
	}

	
	// Am I the recipient
	if(! (m_EntityCert == req->c_info->recipient) )
	{
		return false;
	}

	return true;
}


bool Entity_CA::GenerateCRL(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	PKI_CRL crl;
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_CA_GENERATE_CRL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!*m_InternalCa)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_CA_NOT_CONFIGURED);
		return false;
	}

	if(!Private_GenerateCRL(crl))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_CA::PublishToEntity(PUB_ENTRY_INFO * PubEntity, const NEWPKI_REQUEST * request)
{
	CRYPTED_NEWPKI_REQUEST * c_Request;
	EVP_PKEY * CryptKey;
	unsigned long req_id;
	
	c_Request = CRYPTED_NEWPKI_REQUEST_new();
	if(!c_Request)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	if(!X509_PUBKEY_cpy(&c_Request->recipient, PubEntity->PubSSL->cert_info->key))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		CRYPTED_NEWPKI_REQUEST_free(c_Request);
		return false;
	}
	if(!X509_PUBKEY_cpy(&c_Request->sender, m_EntityCert.GetX509_PUBKEY()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		CRYPTED_NEWPKI_REQUEST_free(c_Request);
		return false;
	}

	CryptKey = X509_PUBKEY_get(c_Request->recipient);
	if(!CryptKey)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		CRYPTED_NEWPKI_REQUEST_free(c_Request);
		return false;
	}
	if(NEWPKI_REQUEST_sign_encrypt(c_Request->c_request, request, m_EntityCert.GetPrivateKey().GetRsaKey(), CryptKey, EVP_sha1(), EVP_des_ede3_cbc()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		CRYPTED_NEWPKI_REQUEST_free(c_Request);
		EVP_PKEY_free(CryptKey);
		return false;
	}
	EVP_PKEY_free(CryptKey);
	
	if(!CaPublicationStore.Requester_InsertRequest(c_Request, 0, req_id))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		CRYPTED_NEWPKI_REQUEST_free(c_Request);
		return false;
	}
	
	if(!m_Jobs.InsertRequest(req_id, c_Request))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		CRYPTED_NEWPKI_REQUEST_free(c_Request);
		CaPublicationStore.Requester_DeleteTransaction(req_id);
		return false;
	}
	CRYPTED_NEWPKI_REQUEST_free(c_Request);
	
	return true;
}

bool Entity_CA::PublishDatas(const char * ldap_uid, const char * object, NEWPKI_PUB_REQUEST_BODY *PubBody)
{
	NEWPKI_PUB_REQUEST * PubReq;
	NEWPKI_REQUEST * Request;
	PUB_ENTRY_INFO * currPub;
	int i;
	X509 * currCert;

	PubReq = NEWPKI_PUB_REQUEST_new();
	if(!PubReq)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		NEWPKI_PUB_REQUEST_BODY_free(PubBody);
		return false;
	}

	if(ldap_uid)
		ASN1_STRING_SET(&PubReq->ldap_uid, ldap_uid);
	
	ASN1_STRING_SET(&PubReq->object, object);

	NEWPKI_PUB_REQUEST_BODY_free(PubReq->body);
	PubReq->body = PubBody;

	Request = NEWPKI_REQUEST_new();
	if(!Request)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		NEWPKI_PUB_REQUEST_free(PubReq);
		return false;
	}
	Request->type = NEWPKI_REQUEST_TYPE_PUB;
	Request->d.pub_request = PubReq;


	// Setting the parent certs
	for(i=0; i<SKM_sk_num(X509, c_myConf->parentCerts); i++)
	{
		currCert = SKM_sk_value(X509, c_myConf->parentCerts, i);
		if(!currCert)
			continue;

		currCert = X509_dup(currCert);
		if(!currCert)
			continue;

		if(SKM_sk_push(X509, PubReq->ParentCerts, currCert) < 0)
		{
			X509_free(currCert);
		}
	}
	currCert = m_InternalCa->get_CaCert()->GetX509(true);
	if(currCert)
	{
		if(SKM_sk_push(X509, PubReq->ParentCerts, currCert) < 0)
		{
			X509_free(currCert);
		}
	}

	ConfAccessLock.LockRead();
	for(i=0; i<SKM_sk_num(PUB_ENTRY_INFO, myConf->Publications->list); i++)
	{
		currPub = SKM_sk_value(PUB_ENTRY_INFO, myConf->Publications->list, i);
		if(!currPub)
			continue;

		if(!PublishToEntity(currPub, Request))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			NEWPKI_REQUEST_free(Request);
			return false;
		}
	}

	ConfAccessLock.Unlock();
	NEWPKI_REQUEST_free(Request);

	return true;
}

bool Entity_CA::PublishCert(const char *ldap_uid, const PKI_CERT &Cert)
{
	NEWPKI_PUB_REQUEST_BODY * PubBody;
	mString Object;

	PubBody = NEWPKI_PUB_REQUEST_BODY_new();
	if(!PubBody)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	PubBody->d.cert = Cert.GetX509(true);
	if(!PubBody->d.cert)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		NEWPKI_PUB_REQUEST_BODY_free(PubBody);
		return false;
	}
	PubBody->type = NEWPKI_PUB_REQUEST_CERT;

	// PubBody is freed by PublishDatas
	Object.sprintf("[%ld] - %s", Cert.GetSerial(), Cert.GetStringName());
	if(!PublishDatas(ldap_uid, Object.c_str(), PubBody))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_CA::PublishCrl(const PKI_CRL &crl)
{
	NEWPKI_PUB_REQUEST_BODY * PubBody;
	mString Object;

	PubBody = NEWPKI_PUB_REQUEST_BODY_new();
	if(!PubBody)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	PubBody->d.crl = crl.GetX509_CRL(true);
	if(!PubBody->d.crl)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		NEWPKI_PUB_REQUEST_BODY_free(PubBody);
		return false;
	}
	PubBody->type = NEWPKI_PUB_REQUEST_CRL;
	Object.sprintf("Start date [%ld] - End date [%ld]", crl.GetStartTime(), crl.GetEndTime());

	// PubBody is freed by PublishDatas
	if(!PublishDatas(NULL, Object.c_str(), PubBody))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	return true;
}

bool Entity_CA::PublishRevocation(const char *ldap_uid, const PKI_CERT &Cert, time_t rev_date)
{
	NEWPKI_PUB_REQUEST_BODY * PubBody;
	mString Object;

	PubBody = NEWPKI_PUB_REQUEST_BODY_new();
	if(!PubBody)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	PubBody->type = NEWPKI_PUB_REQUEST_REV;
	PubBody->d.rev = NEWPKI_PUB_REQUEST_BODY_REV_new();
	if(!PubBody->d.rev)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		NEWPKI_PUB_REQUEST_BODY_free(PubBody);
		return false;
	}

	// Setting the revoked certificate
	if(PubBody->d.rev->cert)
		X509_free(PubBody->d.rev->cert);		
	PubBody->d.rev->cert = Cert.GetX509(true);
	if(!PubBody->d.rev->cert)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		NEWPKI_PUB_REQUEST_BODY_free(PubBody);
		return false;
	}

	// Setting the revocation date
	if(!ASN1_TIME_set_localtime(PubBody->d.rev->revdate, rev_date))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		NEWPKI_PUB_REQUEST_BODY_free(PubBody);
		return false;
	}

	Object.sprintf("[%ld] - %s", Cert.GetSerial(), Cert.GetStringName());

	// PubBody is freed by PublishDatas
	if(!PublishDatas(ldap_uid, Object.c_str(), PubBody))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

