/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/

// Entity_CA.h: interface for the Entity_CA class.
//
//////////////////////////////////////////////////////////////////////

#ifndef ENTITY_CA_H
#define ENTITY_CA_H


#include "Entity.h"
#include "Entity_CA_ASN1.h"
#include "CaStore.h"


#include "CA_Handler.h"
#include "PlugCA.h"

/*!
	This class represents a CA Entity
*/
class Entity_CA : public Entity  
{
public:
	Entity_CA(ENTITY_CONSTRUCTOR_PARAMETERS);
	virtual ~Entity_CA();
	bool GetWaitingResponses(STACK_OF(STORED_NEWPKI_RESPONSE) * Responses);
	bool GetWaitingRequests(STACK_OF(CRYPTED_NEWPKI_REQUEST) * Requests);
	bool GetProcessedTransactionIds(TRANSACTION_IDS * transactionIDs);
	bool GetWaitingTransactionIds(TRANSACTION_IDS * transactionIDs);
	bool OnRequestSent(unsigned long RequestId, const ASN1_OCTET_STRING * transactionId, const CRYPTED_NEWPKI_REQUEST * req);
	bool OnResponseSent(const STORED_NEWPKI_RESPONSE * resp);
	bool OnNewResponses(const STORED_NEWPKI_RESPONSES * resps);
	bool OnNewRequests(const STORED_NEWPKI_REQUESTS * reqs);
	bool OnNewConf(const ENTITY_CONF_CRYPTED * newConf);

	bool Load();
	bool Create(const ENTITY_CREATION_DATAS * Params, ADMIN_RESPONSE_BODY * response);
	bool Upgrade(const char * Version);
	bool Init(const ENTITY_SIGNATURE_RESP * init_datas);
	bool ParseAdminCommand(ADMIN_RESPONSE_BODY * response, const PKI_CERT & ClientCert, const ADMIN_REQUEST * AdminRequest);
	void SpawnStop();

	/*! \brief This function returns the list of types of logs the entity generates.
	 *  \param cLogsType [OUT] The list.
	 */
	static void LogsTypeGet(vector<LOG_MESSAGE_TYPE> & cLogsType);

	/*! \brief This function returns the list of types of ACL the entity handles.
	 *  \param acl_list [OUT] The list.
	 */
	static void GetACL_List(STACK_OF(ASN1_INTEGER) *acl_list);
private:
	bool PublishRevocation(const char * ldap_uid, const PKI_CERT & Cert, time_t rev_date);
	bool PublishCrl(const PKI_CRL & crl);
	bool PublishCert(const char * ldap_uid, const PKI_CERT & Cert);
	bool PublishDatas(const char * ldap_uid, const char * object, NEWPKI_PUB_REQUEST_BODY * PubBody);
	bool PublishToEntity(PUB_ENTRY_INFO * PubEntity, const NEWPKI_REQUEST * request);

	bool GenerateCRL(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	STORED_NEWPKI_RESPONSE * MakeResponse(NEWPKI_RESPONSE * Response, X509_PUBKEY * Recipient);
	bool LoginUser(UserHandle & hUser, int & UserType);
	void LogoutUser(const UserHandle & hUser);
	
	static bool Private_ParseAdminCommand(bool ExecuteCmd, Entity * me_this, vector<LOG_MESSAGE_TYPE> &  mLogsType, ADMIN_RESPONSE_BODY * response, const PKI_CERT & ClientCert, const ADMIN_REQUEST * AdminRequest, const UserHandle & hUser);

	bool GetStatus(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	bool GetCaCert(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	bool CreateRootCa(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	bool CreateChildCa(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	bool ImportChildCaCert(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	bool EnumCerts(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	bool EnumCrls(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	bool SignCSR(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	bool RevokeCert(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);

	
	bool ParseNewConf(const ENTITY_CONF_CRYPTED * newConf, const INTERNAL_PKI_CA * cas);
	bool PersonnalConfFromPEM(const mString & conf_pem);
	bool PersonnalConfToPem(mString & conf_pem);

	bool Private_SignCSR(const char * ldap_uid, X509_REQ * reqCSR, unsigned long validity, PKI_CERT & Cert);
	bool Private_GenerateCRL(PKI_CRL & Crl);
	bool Private_RevokeCERT(const char * ldap_uid, unsigned long serial, CERT_STATE & CertStatus);

	bool OnNewRequest(const PKI_CSR & Csr);
	bool OnNewCertificate(const char * ldap_uid, const PKI_CERT & Cert);
	bool OnNewRevocation(const char * ldap_uid, const PKI_CERT & Cert, time_t rev_date);
	bool OnNewCRL(const PKI_CRL & crl);

	static void ThreadWorker(const NewpkiThread * Thread, void *param);
	static void ThreadGenerateCRL(const NewpkiThread * Thread, void *param);
	static void ThreadRepublish(const NewpkiThread * Thread, void *param);
	bool ProceedWithRequest(WORKING_CA_REQUEST * req, LOG_MESSAGE_TYPE & LogType, mString & Err);
	bool ProceedWithCertification(NEWPKI_CERT_REQUEST * cert_request, NEWPKI_RESPONSE * currResponse);
	
	bool STORED_NEWPKI_REQUEST_check(STORED_NEWPKI_REQUEST * req, mString & Sender);
	bool AddNewSTORED_NEWPKI_REQUEST(STORED_NEWPKI_REQUEST * req, mString & Sender);

	LOCAL_CA_CONF * c_myConf;
	ENTITY_CONF * myConf;

	NewpkiThread hThreadWorker;
	NewpkiThread hThreadGenerateCRL;
	NewpkiThread hThreadRepublish;
	
	CA_Handler * m_InternalCa;
	HashTable_String m_PlugOptions;
	HashTable_String m_Exts;
	HashTable_String m_CrlExts;
	long m_crldays;
	long m_crlhours;
	PlugCA * m_PlugCA;

	CaStore CaPublicationStore;

#define MY_CA_CONF_PTR body->d.ca_conf->d.CA_CONF_PTR
	DECLARE_COMMAND_PARSER(Entity_CA);
};

#endif
