/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/

 // Entity_KEYSTORE.cpp: implementation of the Entity_KEYSTORE class.
//
//////////////////////////////////////////////////////////////////////

#include "Entity_KEYSTORE.h"
#include "svintl.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Entity_KEYSTORE::Entity_KEYSTORE(ENTITY_CONSTRUCTOR_PARAMETERS):Entity(ENTITY_CONSTRUCTOR_PARAM_PASSTHRU)
{
	myConf = NULL;
	c_myConf = NULL;
}

Entity_KEYSTORE::~Entity_KEYSTORE()
{
	ConfAccessLock.LockWrite();
	if(myConf)
		ENTITY_CONF_free(myConf);
	if(c_myConf)
		LOCAL_KEYSTORE_CONF_free(c_myConf);
	ConfAccessLock.Unlock();
}

bool Entity_KEYSTORE::Create(const ENTITY_CREATION_DATAS * Params, ADMIN_RESPONSE_BODY * response)
{
	if(!Params)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	if(Params->type != ENTITY_TYPE_KEY_STORE)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	ENTITY_CREATION_RESP * resp;

	if(!Params->d.entity_key)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	//We create the database
	if(!Private_CreateEntity(Params->d.entity_key, NULL, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	resp = ENTITY_CREATION_RESP_new();
	if(!resp)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		Destroy();
		return false;
	}
	resp->type = ENTITY_TYPE_KEY_STORE;
	if(X509_PUBKEY_set(&resp->d.entity_pub_key, (EVP_PKEY*)m_EntityKey.GetRsaKey()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		Destroy();
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_CREATE_ENTITY, resp))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_KEYSTORE::Load()
{
	if(!Private_CommonLoad(NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!IsFullyInit())
	{
		return true;
	}

	sk_X509_push(EntityParentCerts, X509_dup(c_myConf->cas->RootCa));
	sk_X509_push(EntityParentCerts, X509_dup(c_myConf->cas->EntitiesCa));

	//Start the conf synch
	if(!m_Jobs.StartConfSync())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	//Start the mailer
	if(!m_Jobs.StartMailer())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!m_Jobs.SetEntityParentsCert(EntityParentCerts))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_KEYSTORE::Init(const ENTITY_SIGNATURE_RESP * init_datas)
{
	if(IsFullyInit())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!init_datas)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	if(init_datas->body->type != ENTITY_TYPE_KEY_STORE)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	if(!Private_InitEntity(init_datas->body->d.EntityCert, &myConf, init_datas))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_KEYSTORE::PersonnalConfFromPEM(const mString &conf_pem)
{
	//Convert from PEM
	ASN1_ENCRYPT_SIGN * encrypt;
	encrypt = NULL;

	if(ASN1_ENCRYPT_SIGN_From_PEM(&encrypt, conf_pem.c_str()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//Decrypt the conf
	if(c_myConf)
	{
		LOCAL_KEYSTORE_CONF_free(c_myConf);
		c_myConf = NULL;
	}

	if(LOCAL_KEYSTORE_CONF_decrypt_verify(encrypt, &c_myConf, m_EntityKey.GetRsaKey(), m_EntityKey.GetRsaKey()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ASN1_ENCRYPT_SIGN_free(encrypt);
		return false;
	}
	ASN1_ENCRYPT_SIGN_free(encrypt);
	
	//Extract the conf
	if(myConf)
	{
		ENTITY_CONF_free(myConf);
		myConf = NULL;
	}
	if(!ExtractMyConf(c_myConf->conf, &myConf, c_myConf->cas, c_myConf->cas->PkiCert, false))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_KEYSTORE::PersonnalConfToPem(mString & conf_pem)
{
	char * pem;
	ASN1_ENCRYPT_SIGN * crypted_conf;

	ConfAccessLock.LockRead();

	//Sign/Crypt the conf
	crypted_conf = ASN1_ENCRYPT_SIGN_new();
	if(!crypted_conf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ConfAccessLock.Unlock();
		return false;
	}

	if(LOCAL_KEYSTORE_CONF_sign_encrypt(crypted_conf, c_myConf, m_EntityKey.GetRsaKey(), m_EntityKey.GetRsaKey(), EVP_sha1(), EVP_des_ede3_cbc()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ASN1_ENCRYPT_SIGN_free(crypted_conf);
		ConfAccessLock.Unlock();
		return false;
	}

	//Convert it to PEM
	pem = ASN1_ENCRYPT_SIGN_To_PEM(crypted_conf);
	ASN1_ENCRYPT_SIGN_free(crypted_conf);
	if(!pem)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	conf_pem = pem;
	free(pem);
	
	ConfAccessLock.Unlock();
	return true;
}

bool Entity_KEYSTORE::LoginUser(UserHandle & hUser, int & UserType)
{
	if(!AclValidator.CanUserPerform(hUser.GetUserCert(), ACL_TYPE_AUTHENTICATE_ON_ENTITY))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	UserType = USER_TYPE_KEYSTORE;
	return true;
}

void Entity_KEYSTORE::LogoutUser(const UserHandle & hUser)
{
}

bool Entity_KEYSTORE::ParseNewConf(const ENTITY_CONF_CRYPTED * newConf, const INTERNAL_PKI_CA * cas)
{
	//Should we initialize the local conf ?
	if(!c_myConf)
		c_myConf = LOCAL_KEYSTORE_CONF_new();

	if(!c_myConf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	//Should we update the internal cas certs ?
	if(cas)
	{
		if(!INTERNAL_PKI_CA_cpy(&c_myConf->cas, cas))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}

	//Save the updated conf
	if(!ENTITY_CONF_CRYPTED_cpy(&c_myConf->conf, newConf))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_KEYSTORE::Upgrade(const char * Version)
{
	return true;
}


void Entity_KEYSTORE::GetACL_List(STACK_OF(ASN1_INTEGER) *acl_list)
{
	ASN1_INTEGER * acl_id;
	int i;
	static ACL_TYPE list_acls[] =
	{
		ACL_TYPE_VIEW_LOGS,
		(ACL_TYPE)0
	};
	for(i=0; list_acls[i]; i++)
	{
		acl_id = NULL;
		ASN1_INTEGER_SET(&acl_id, list_acls[i]);
		if(acl_id)
		{
			if(SKM_sk_push(ASN1_INTEGER, acl_list, acl_id) < 0)
			{
				ASN1_INTEGER_free(acl_id);
			}
		}
	}	
}


bool Entity_KEYSTORE::ParseAdminCommand(ADMIN_RESPONSE_BODY * response, const PKI_CERT & ClientCert, const ADMIN_REQUEST * AdminRequest)
{
	//We only accept SSLv3 connection	
	if(!ClientCert)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ERR_to_ADMIN_RESPONSE(response);
		return false;
	}
	return Private_ParseAdminCommand(true, this, LogsType, response, ClientCert, AdminRequest, GetUserHandle());
}

bool Entity_KEYSTORE::Private_ParseAdminCommand(bool ExecuteCmd, Entity * me_this, vector<LOG_MESSAGE_TYPE> &  mLogsType, ADMIN_RESPONSE_BODY * response, const PKI_CERT & ClientCert, const ADMIN_REQUEST * AdminRequest, const UserHandle & hUser)
{

	PARSER_COMMAND_BEGIN(Entity_KEYSTORE, response, 0, AdminRequest, ClientCert, hUser, ExecuteCmd, me_this, mLogsType)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_ENTITY_GET_MY_CONF)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_SEND_ADMIN_MAIL)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_SEND_MAIL)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_LOGIN,					Entity_KEYSTORE::UserLogin,				LOG_MESSAGE_TYPE_USER_LOGIN, (ClientCert)?(char*)ClientCert.GetStringName():NULL, LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_ENUM_LOGS,				Entity_KEYSTORE::EnumLogs)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_LOGS_COUNT,			Entity_KEYSTORE::GetLogsCount)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_MY_ACL,				Entity_KEYSTORE::GetMyACL)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_LOGS_TYPE,			Entity_KEYSTORE::GetLogsType)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_SEND_ADMIN_MAIL,			Entity_KEYSTORE::AdminSendMail,			LOG_MESSAGE_TYPE_ADD_ADMIN_MAIL_QUEUE,	_sv("none"), LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_CHECK_LOGS,				Entity_KEYSTORE::CheckLogsIntegrity)
	PARSER_COMMAND_END(Entity_KEYSTORE)
}

void Entity_KEYSTORE::LogsTypeGet(vector<LOG_MESSAGE_TYPE> &  cLogsType)
{
	Private_ParseAdminCommand(false, NULL, cLogsType, NULL, PKI_CERT(), NULL, UserHandle());
}


void Entity_KEYSTORE::SpawnStop()
{
	m_Jobs.SpawnStop();
}



bool Entity_KEYSTORE::OnNewConf(const ENTITY_CONF_CRYPTED * newConf)
{
	if(!ExtractMyConf(newConf, &myConf, NULL, c_myConf->cas->PkiCert, true))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_KEYSTORE::OnNewRequests(const STORED_NEWPKI_REQUESTS * reqs)
{
	return true;
}

bool Entity_KEYSTORE::OnRequestSent(unsigned long RequestId, const ASN1_OCTET_STRING * transactionId, const CRYPTED_NEWPKI_REQUEST * req)
{
	return true;
}


bool Entity_KEYSTORE::OnResponseSent(const STORED_NEWPKI_RESPONSE * resp)
{
	return true;
}


bool Entity_KEYSTORE::OnNewResponses(const STORED_NEWPKI_RESPONSES * resps)
{
	return true;
}



bool Entity_KEYSTORE::GetWaitingTransactionIds(TRANSACTION_IDS *transactionIDs)
{
	return false;
}

bool Entity_KEYSTORE::GetProcessedTransactionIds(TRANSACTION_IDS *transactionIDs)
{
	return false;
}

bool Entity_KEYSTORE::GetWaitingRequests(STACK_OF(CRYPTED_NEWPKI_REQUEST) * Requests)
{
	return true;
}

bool Entity_KEYSTORE::GetWaitingResponses(STACK_OF(STORED_NEWPKI_RESPONSE) * Responses)
{
	return true;
}

