/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/



// Entity_PKI.cpp: implementation of the Entity_PKI class.
//
//////////////////////////////////////////////////////////////////////

#include "Entity_CA.h"
#include "Entity_RA.h"
#include "Entity_KEYSTORE.h"
#include "Entity_PUBLICATION.h"
#include "Entity_REPOSITORY.h"

#include <PKI_P7B.h>
#include <PKI_CRL.h>


#include "Entity_PKI.h"
#include "svintl.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


Entity_PKI::Entity_PKI(ENTITY_CONSTRUCTOR_PARAMETERS):Entity(ENTITY_CONSTRUCTOR_PARAM_PASSTHRU)
{
	pki_conf = NULL;	
	RootCa = NULL;
	UsersCa = NULL;
	EntitiesCa = NULL;
	OcspCa = NULL;
	
	try
	{
		RootCa = new CA_Handler(INTERNAL_CA_TYPE_ROOT);
	}
	catch(ExceptionNewPKI e)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		throw ExceptionNewPKI();		
	}
	
	
	try
	{
		UsersCa = new CA_Handler(INTERNAL_CA_TYPE_USER);
	}
	catch(ExceptionNewPKI e)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		throw ExceptionNewPKI();		
	}
	
	try
	{
		EntitiesCa = new CA_Handler(INTERNAL_CA_TYPE_ENTITY);
	}
	catch(ExceptionNewPKI e)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		throw ExceptionNewPKI();		
	}
	
	try
	{
		OcspCa = new CA_Handler(INTERNAL_CA_TYPE_OCSP);
	}
	catch(ExceptionNewPKI e)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		throw ExceptionNewPKI();		
	}
	
	if(!RootCa || !UsersCa || !EntitiesCa || !OcspCa)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		throw ExceptionNewPKI();
	}
	
	RootCa->set_ENGINE(m_Engine);
	UsersCa->set_ENGINE(m_Engine);
	EntitiesCa->set_ENGINE(m_Engine);
	OcspCa->set_ENGINE(m_Engine);
}

Entity_PKI::~Entity_PKI()
{
	Free_PkiConf();

	if(RootCa) delete RootCa;
	if(UsersCa) delete UsersCa;
	if(EntitiesCa) delete EntitiesCa;
	if(OcspCa) delete OcspCa;
}

bool Entity_PKI::Create(const ENTITY_CREATION_DATAS * Params, ADMIN_RESPONSE_BODY * response)
{
	if(!Params)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	if(Params->type != ENTITY_TYPE_PKI)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	PKI_CREATION_REQ * cre_params = Params->d.pki_create;
	PKI_CERT certAdmin;
	mString Cn;
	ENTITY_CREATION_RESP * cre_entity;
	ENTITY_LINKS * link;

	if(!cre_params)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}


	//We create the database
	if(!Private_CreateEntity(cre_params->entity_key, NULL, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!CA_Handler::CreateTables(m_DbConn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}

	if(!RootCa->SetDbConn(m_DbConn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}
	if(!UsersCa->SetDbConn(m_DbConn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}
	if(!EntitiesCa->SetDbConn(m_DbConn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}
	if(!OcspCa->SetDbConn(m_DbConn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}


	// We now create the internal CAs and
	// the PKI admin certificate
	pki_conf = PKI_CONF_new();
	if(!pki_conf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		Destroy();
		return false;
	}
	ASN1_STRING_SET(&pki_conf->my_conf->Name, m_EntityName.c_str());

	//Initialize it link struct
	link = ENTITY_LINKS_new();
	if(!link)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		Destroy();
		return false;
	}
	ASN1_STRING_SET(&link->src->Name, m_EntityName.c_str());
	ASN1_INTEGER_SET(&link->src->Type, ENTITY_TYPE_PKI);
	if(SKM_sk_push(ENTITY_LINKS, pki_conf->links->links, link) < 0)
	{
		ENTITY_LINKS_free(link);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}


	//Initialize the pki conf
	pki_conf->my_conf->body->type = ENTITY_TYPE_PKI;
	pki_conf->my_conf->body->d.pki_conf = ENTITY_CONF_BODY_PKI_new();
	if(!pki_conf->my_conf->body->d.pki_conf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		Destroy();
		return false;
	}
	pki_conf->my_conf->body->d.pki_conf->type = PKI_CONF_VERSION;
	pki_conf->my_conf->body->d.pki_conf->d.ptr = PKI_CONF_STRUCT_new();
	if(!pki_conf->my_conf->body->d.pki_conf->d.ptr)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		Destroy();
		return false;
	}
	ASN1_INTEGER_SET(&MY_CONF_PTR->offline, 0);
	ASN1_INTEGER_SET(&pki_conf->my_conf->version, 0);

	
	//Create ROOT CA
	Cn = m_EntityName;
	Cn += " - Internal ROOT CA";
	if(!InternalCA_Create(INTERNAL_CA_TYPE_ROOT, cre_params->root_ca, cre_params->dn, (char*)Cn.c_str(), ASN1_INTEGER_GET(cre_params->validity)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}
	//We create initial CRL
	if(!InternalCA_Revoke(INTERNAL_CA_TYPE_ROOT, 0))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}
	if(!X509_cpy(&pki_conf->cas->RootCa, RootCa->get_CaCert()->GetX509()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		Destroy();
		return false;
	}
	
	//Create USERS CA
	Cn = m_EntityName;
	Cn += " - Internal USERS CA";
	if(!InternalCA_Create(INTERNAL_CA_TYPE_USER, cre_params->users_ca, cre_params->dn, (char*)Cn.c_str(), ASN1_INTEGER_GET(cre_params->validity)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}
	//We create initial CRL
	if(!InternalCA_Revoke(INTERNAL_CA_TYPE_USER, 0))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}
	if(!X509_cpy(&pki_conf->cas->UsersCa, UsersCa->get_CaCert()->GetX509()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		Destroy();
		return false;
	}
	
	//Create ENTITIES CA
	Cn = m_EntityName;
	Cn += " - Internal ENTITIES CA";
	if(!InternalCA_Create(INTERNAL_CA_TYPE_ENTITY, cre_params->entities_ca, cre_params->dn, (char*)Cn.c_str(), ASN1_INTEGER_GET(cre_params->validity)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}
	//We create initial CRL
	if(!InternalCA_Revoke(INTERNAL_CA_TYPE_ENTITY, 0))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}
	if(!X509_cpy(&pki_conf->cas->EntitiesCa, EntitiesCa->get_CaCert()->GetX509()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		Destroy();
		return false;
	}

	//Create OCSP CA
	Cn = m_EntityName;
	Cn += " - Internal OCSP CA";
	if(!InternalCA_Create(INTERNAL_CA_TYPE_OCSP, cre_params->ocsp_ca, cre_params->dn, (char*)Cn.c_str(), ASN1_INTEGER_GET(cre_params->validity)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}
	//We create initial CRL
	if(!InternalCA_Revoke(INTERNAL_CA_TYPE_OCSP, 0))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}
	if(!X509_cpy(&pki_conf->cas->OcspCa, OcspCa->get_CaCert()->GetX509()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		Destroy();
		return false;
	}



	//We now create the PKI Entity certificate
	//Load entity's private key
	
	// m_EntityKey has been loaded/created by Private_CreateEntity

	if(!InternalCA_SignForceDn(INTERNAL_CA_TYPE_ENTITY, m_EntityKey.GetPublicKey(), m_EntityName.c_str(), ASN1_STRING_GET(cre_params->Email), m_EntityCert))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}
	if(!X509_cpy(&pki_conf->cas->PkiCert, m_EntityCert.GetX509()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		Destroy();
		return false;
	}

	//We now create the PKI administrator certificate
	Cn = m_EntityName;
	Cn += " - PKI Administrator";
	if(!InternalCA_SignForceDn(INTERNAL_CA_TYPE_USER, cre_params->AdminPubKey, (char*)Cn.c_str(), ASN1_STRING_GET(cre_params->AdminEmail), certAdmin))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}

	if(!DeclarePkiAdminCert(certAdmin, ASN1_STRING_GET(cre_params->AdminEmail)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}

	// We init the entity
	if(!Private_InitEntity(m_EntityCert.GetX509()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}


	cre_entity = ENTITY_CREATION_RESP_new();
	if(!cre_entity)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		Destroy();
		return false;
	}

	cre_entity->type = ENTITY_TYPE_PKI;
	cre_entity->d.pki_create = sk_X509_new_null();
	if(!cre_entity->d.pki_create)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ENTITY_CREATION_RESP_free(cre_entity);
		Destroy();
		return false;
	}

	sk_X509_push(cre_entity->d.pki_create, RootCa->get_CaCert()->GetX509(true));
	sk_X509_push(cre_entity->d.pki_create, UsersCa->get_CaCert()->GetX509(true));
	sk_X509_push(cre_entity->d.pki_create, certAdmin.GetX509(true));

	if(!WritePersonnalConf())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ENTITY_CREATION_RESP_free(cre_entity);
		Destroy();
		return false;
	}
	
	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_CREATE_ENTITY, cre_entity))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ENTITY_CREATION_RESP_free(cre_entity);
		Destroy();
		return false;
	}

	Free_PkiConf();
	return true;
}

bool Entity_PKI::Init(const ENTITY_SIGNATURE_RESP * init_datas)
{
	NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
	return false;
}

bool Entity_PKI::Load()
{
	if(!Private_CommonLoad(NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!IsFullyInit())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}
		
	//We load the CAs
	if(!RootCa->SetDbConn(m_DbConn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!UsersCa->SetDbConn(m_DbConn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!EntitiesCa->SetDbConn(m_DbConn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!OcspCa->SetDbConn(m_DbConn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!RootCa->Load())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!UsersCa->Load())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!EntitiesCa->Load())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!OcspCa->Load())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	sk_X509_push(EntityParentCerts, X509_dup(pki_conf->cas->RootCa));
	sk_X509_push(EntityParentCerts, X509_dup(pki_conf->cas->EntitiesCa));

	//Start the mailer
	if(!m_Jobs.StartMailer())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!m_Jobs.SetEntityParentsCert(EntityParentCerts))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	//Start the conf synch
	if(!m_Jobs.StartConfPush(false))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	// Synch the conf
	ConfAccessLock.LockRead();
	if(!m_Jobs.SetPkiConf(pki_conf->crypt_confs))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();

	return true;
}

bool Entity_PKI::Upgrade(const char * Version)
{
	return true;
}


bool Entity_PKI::InternalCA_Create(INTERNAL_CA_TYPE Type, const GEN_PRIVATE_KEY * KeyGen, const X509_NAME * Dn, const char * Cn, int days)
{
	PKI_CERT ResultCert;
	PKI_CSR csr;
	PKI_RSA privKey;
	HashTable_Dn dn;
	HashTable_String Exts;
	Exts.AllowDuplicateNames();
	CA_Handler * currCa;
	INTERNAL_CA_KEY * caKey;


	if(!KeyGen || !Dn || !Cn || !days)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	if(!dn.From_X509_NAME(Dn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	if(!dn.Add("commonName", Cn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}

	caKey = INTERNAL_CA_KEY_new();
	if(!caKey)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!CA_Handler::GEN_PRIVATE_KEY_load(KeyGen, privKey, caKey, m_Engine))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		INTERNAL_CA_KEY_free(caKey);
		return false;
	}

	if(!csr.GenerateCSR(dn, privKey))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		INTERNAL_CA_KEY_free(caKey);
		return false;
	}


	if(Type != INTERNAL_CA_TYPE_ROOT)
	{
		//This is a sub CA not ROOT CA
		if(!InternalCA_Sign(INTERNAL_CA_TYPE_ROOT, csr, ResultCert, days, true))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			INTERNAL_CA_KEY_free(caKey);
			return false;
		}
	}
	else
	{
		Exts.Add("basicConstraints", "critical, CA:TRUE");
		Exts.Add("nsComment", "NewPKI Generated CA Certificate");
		Exts.Add("subjectKeyIdentifier", "hash");
		Exts.Add("authorityKeyIdentifier", "keyid:always");
		Exts.Add("nsCertType", "sslCA, emailCA, objCA");
		Exts.Add("keyUsage", "keyCertSign, cRLSign, digitalSignature, nonRepudiation, keyEncipherment");
		
		if(!ResultCert.CreateSelfSigned(csr, &Exts, days, 1))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			INTERNAL_CA_KEY_free(caKey);
			return false;
		}
	}
	if(!ResultCert.SetPrivateKey(privKey))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		INTERNAL_CA_KEY_free(caKey);
		return false;
	}
	

	//We have the key and the cert we can create the corresponding
	//database entry
	currCa = GetInternalCA(Type);
	if(!currCa)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		INTERNAL_CA_KEY_free(caKey);
		return false;
	}

	//We physicaly "create" it
	if(!currCa->Create(ResultCert, caKey))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		INTERNAL_CA_KEY_free(caKey);
		return false;
	}
	INTERNAL_CA_KEY_free(caKey);

	return true;
}

bool Entity_PKI::InternalCA_SignForceDn(INTERNAL_CA_TYPE Type, const X509_PUBKEY * pubKey, const char *cn, const char * email, PKI_CERT & ResultCert)
{
	time_t currTime;
	CA_Handler * currCa;
	PKI_CSR csrK;
	PKI_CERT * caCert;
	int days;

	currCa = GetInternalCA(Type);
	if(!currCa)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	caCert = currCa->get_CaCert();
	if(!caCert)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	const PKI_RSA & caKey = caCert->GetPrivateKey();
	if(!caKey)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//We format the CSR DN with the parent DN

	if(!Format_CSR(cn, email, pubKey, csrK, caCert->GetX509()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	// Calculate maximum life time for the cert
	time_gmt(&currTime);
	days = (caCert->GetEndDate() - currTime) / (3600*24) - 1;

	//We sign the cert
	if(!InternalCA_Sign(Type, csrK, ResultCert, days, false))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_PKI::Format_CSR(const char * strCn, const char * strEmail, const X509_PUBKEY * pubKey, PKI_CSR & csrK, const X509 * parent_cert)
{
	X509_REQ * csr;
	HashTable_Dn dn;
	int pos;
	X509_NAME * tmpDN;
	EVP_PKEY * tmpKey;

	csr = X509_REQ_new();
	if(!csr)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	tmpKey = X509_PUBKEY_get((X509_PUBKEY *)pubKey);
	if(!tmpKey)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		X509_REQ_free(csr);
		return false;
	}
	if(X509_PUBKEY_set(&csr->req_info->pubkey, tmpKey) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		EVP_PKEY_free(tmpKey);
		X509_REQ_free(csr);
		return false;
	}
	EVP_PKEY_free(tmpKey);


	dn.From_X509_NAME(parent_cert->cert_info->subject);

	pos = -1;
	while( (pos = dn.SeekEntryName("commonName", pos)) != HASHTABLE_NOT_FOUND)
	{
		dn.Delete(pos);
	}
	dn.Add("commonName", strCn);

	pos = -1;
	while( (pos = dn.SeekEntryName("emailAddress", pos)) != HASHTABLE_NOT_FOUND)
	{
		dn.Delete(pos);
	}
	if(strEmail)
	{
		dn.Add("emailAddress", strEmail);
	}


	tmpDN = X509_NAME_new();
	if(!tmpDN)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		X509_REQ_free(csr);
		return false;
	}
	if(!dn.To_X509_NAME(tmpDN))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		X509_NAME_free(tmpDN);
		X509_REQ_free(csr);
		return false;
	}

	if(X509_NAME_set(&csr->req_info->subject, tmpDN) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		X509_NAME_free(tmpDN);
		X509_REQ_free(csr);
		return false;
	}
	X509_NAME_free(tmpDN);

	if(!csrK.SetCSR(csr))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		X509_REQ_free(csr);
		return false;
	}
	X509_REQ_free(csr);

	return true;
}

bool Entity_PKI::InternalCA_Sign(INTERNAL_CA_TYPE Type, const PKI_CSR & csr, PKI_CERT & ResultCert, int Days, bool check_sig)
{

	HashTable_String Exts;
	Exts.AllowDuplicateNames();
	CA_Handler * currCa;

	switch(Type)
	{
		case INTERNAL_CA_TYPE_ROOT:
			Exts.Add("basicConstraints", "critical, CA:TRUE");
			Exts.Add("nsComment", "NewPKI Internal CA Certificate");
			Exts.Add("subjectKeyIdentifier", "hash");
			Exts.Add("authorityKeyIdentifier", "keyid:always");
			Exts.Add("nsCertType", "sslCA, emailCA, objCA");
			Exts.Add("keyUsage", "keyCertSign, cRLSign, digitalSignature, nonRepudiation, keyEncipherment");
			break;
		case INTERNAL_CA_TYPE_USER:
			Exts.Add("basicConstraints", "CA:FALSE");
			Exts.Add("nsComment", "NewPKI Internal CA EndUser Certificate");
			Exts.Add("subjectKeyIdentifier", "hash");
			Exts.Add("authorityKeyIdentifier", "keyid:always");
			Exts.Add("nsCertType", "client");
			Exts.Add("keyUsage", "digitalSignature, nonRepudiation, keyEncipherment");
			Exts.Add("extendedKeyUsage", "clientAuth");
			break;
		case INTERNAL_CA_TYPE_ENTITY:
			Exts.Add("basicConstraints", "CA:FALSE");
			Exts.Add("nsComment", "NewPKI Internal CA Entity Certificate");
			Exts.Add("subjectKeyIdentifier", "hash");
			Exts.Add("authorityKeyIdentifier", "keyid:always");
			Exts.Add("nsCertType", "client, email");
			Exts.Add("keyUsage", "digitalSignature, nonRepudiation, keyEncipherment");
			Exts.Add("extendedKeyUsage", "clientAuth, emailProtection");
			break;
		case INTERNAL_CA_TYPE_OCSP:
			Exts.Add("basicConstraints", "CA:FALSE");
			Exts.Add("nsComment", "NewPKI Internal CA OCSP-Server Certificate");
			Exts.Add("subjectKeyIdentifier", "hash");
			Exts.Add("authorityKeyIdentifier", "keyid:always");
			Exts.Add("extendedKeyUsage", "OCSPSigning");
			break;
		default:
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
			return false;
			break;
			
	}

	//Search corresponding CA_Handler
	currCa = GetInternalCA(Type);
	if(!currCa)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	if(!currCa->Sign(csr, ResultCert, Days, &Exts, check_sig))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_PKI::InternalCA_Revoke(INTERNAL_CA_TYPE Type, unsigned long serial)
{
	CA_Handler * currCa;
	X509_CRL ** dstCrl;
	PKI_CRL resCrl;
	PKI_CERT revCert;
	CERT_STATE CertState;
	time_t revdate;

	currCa = GetInternalCA(Type);
	if(!currCa)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	if(!currCa->Revoke(serial, revCert, CertState, revdate))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!currCa->Generate_CRL(resCrl, 30, 0, NULL, "sha1"))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	ConfAccessLock.LockWrite();
	switch(Type)
	{
		case INTERNAL_CA_TYPE_ROOT:
			dstCrl = &pki_conf->my_conf->crls->RootCaCrl;
			break;
		case INTERNAL_CA_TYPE_USER:
			dstCrl = &pki_conf->my_conf->crls->UsersCaCrl;
			break;
		case INTERNAL_CA_TYPE_ENTITY:
			dstCrl = &pki_conf->my_conf->crls->EntitiesCaCrl;
			break;
		case INTERNAL_CA_TYPE_OCSP:
			dstCrl = &pki_conf->my_conf->crls->OcspCaCrl;
			break;
		default:
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
			ConfAccessLock.Unlock();
			return false;
			break;
	}
	//Copy result CRL into destination
	if(!X509_CRL_cpy(dstCrl, resCrl.GetX509_CRL()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	ConfAccessLock.Unlock();
	
	return true;
}


bool Entity_PKI::PersonnalConfFromPEM(const mString & conf_pem)
{
	ASN1_ENCRYPT_SIGN * crypted_conf;
	
	ConfAccessLock.LockWrite();

	Free_PkiConf();

	crypted_conf = NULL;
	if(!ASN1_ENCRYPT_SIGN_From_PEM(&crypted_conf, conf_pem.c_str()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		ASN1_ENCRYPT_SIGN_free(crypted_conf);
		ConfAccessLock.Unlock();
		return false;
	}

	//We now have the conf, with there encrypted body
	//let's decrypt it, decrypt the body and verify signature
	if(PKI_CONF_decrypt_verify(crypted_conf, &pki_conf, m_EntityKey.GetRsaKey(), m_EntityKey.GetRsaKey()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ASN1_ENCRYPT_SIGN_free(crypted_conf);
		ConfAccessLock.Unlock();
		return false;
	}
	ASN1_ENCRYPT_SIGN_free(crypted_conf);

	if(!CommonConfLoad(pki_conf->my_conf, pki_conf->cas))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	ConfAccessLock.Unlock();
	return true;
}

bool Entity_PKI::PersonnalConfToPem(class mString & pem_conf)
{
	ASN1_ENCRYPT_SIGN * crypted_conf;
	char * pem;

	ConfAccessLock.LockWrite();
	//We increase the version of the global conf

	ASN1_INTEGER_SET(&pki_conf->crypt_confs->confs->version, ASN1_INTEGER_GET(pki_conf->crypt_confs->confs->version)+1);
	
	//Ok, we have crypted and sign every conf
	//let's sign the whole thing with the PKI Entity Key
	if(ASN1_sign((int(*)())i2d_EXPORTED_PKI_CONF_BODY, pki_conf->crypt_confs->sig->algor, NULL, pki_conf->crypt_confs->sig->digest, (char*)pki_conf->crypt_confs->confs, (EVP_PKEY*)m_EntityKey.GetRsaKey(), EVP_sha1()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	crypted_conf = ASN1_ENCRYPT_SIGN_new();	
	if(!crypted_conf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ConfAccessLock.Unlock();
		return false;
	}

	//Ok, now we're all ready to sign and crypt for ourself the whole think
	//Encrypt the body and generate signature
	if( !PKI_CONF_sign_encrypt(crypted_conf, pki_conf, m_EntityKey.GetRsaKey(), m_EntityKey.GetRsaKey(), EVP_sha1(), EVP_des_ede3_cbc()) ) 
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ASN1_ENCRYPT_SIGN_free(crypted_conf);
		ConfAccessLock.Unlock();
		return false;
	}

	//We have everything we can now convert it to PEM
	pem = ASN1_ENCRYPT_SIGN_To_PEM(crypted_conf);
	if(!pem)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ASN1_ENCRYPT_SIGN_free(crypted_conf);
		ConfAccessLock.Unlock();
		return false;
	}
	ASN1_ENCRYPT_SIGN_free(crypted_conf);

	// Synch the conf if online
	if(!ASN1_INTEGER_GET(MY_CONF_PTR->offline))
	{
		if(!m_Jobs.SetPkiConf(pki_conf->crypt_confs))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ConfAccessLock.Unlock();
			free(pem);
			return false;
		}
	}
	ConfAccessLock.Unlock();

	pem_conf = pem;
	free(pem);

	return true;
}


void Entity_PKI::Free_PkiConf()
{
	if(pki_conf)
	{
		ConfAccessLock.LockWrite();
		PKI_CONF_free(pki_conf);
		pki_conf = NULL;
		ConfAccessLock.Unlock();
	}
}

bool Entity_PKI::ParseAdminCommand(ADMIN_RESPONSE_BODY * response, const PKI_CERT & ClientCert, const ADMIN_REQUEST * AdminRequest)
{
	int pos;
	const char * cn = NULL;
	PKI_CSR csr;


	//We only accept SSLv3 connection	
	if(!ClientCert)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ERR_to_ADMIN_RESPONSE(response);
		return false;
	}
	
	if(AdminRequest->body->type == ADMIN_REQ_TYPE_CREATE_PKI_USER)
	{
		if(csr.SetCSR(AdminRequest->body->d.create_pki_user->csr))
		{
			if(((pos = csr.GetRequestDN().SeekEntryName("commonName", -1)) != HASHTABLE_NOT_FOUND) )
				cn = csr.GetRequestDN().Get(pos);
		}
	}

	
	PARSER_COMMAND_BEGIN(Entity_PKI, response, 0, AdminRequest, ClientCert, GetUserHandle(), true, this, LogsType)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_SEND_ADMIN_MAIL)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_SEND_MAIL)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_CONFIG_PUSH)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_LOGIN,					Entity_PKI::UserLogin,				LOG_MESSAGE_TYPE_USER_LOGIN, (ClientCert)?(char*)ClientCert.GetStringName():NULL, LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_ENUM_LOGS,				Entity_PKI::EnumLogs)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_LOGS_COUNT,			Entity_PKI::GetLogsCount)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_ENUM_ENTITIES,			Entity_PKI::EnumEntities)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_SIGN_ENTITY,				Entity_PKI::SignEntity,				LOG_MESSAGE_TYPE_SIGN_ENTITY, (AdminRequest->body->d.sign_entity)?ASN1_STRING_GET(AdminRequest->body->d.sign_entity->Name):NULL, LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_ENUM_USERS,				Entity_PKI::EnumUsers)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_CREATE_PKI_USER,			Entity_PKI::CreateUser,				LOG_MESSAGE_TYPE_USER_ADD, (char*)cn, LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_REVOKE_USER,				Entity_PKI::RevokeUserCert,			LOG_MESSAGE_TYPE_USER_REVOCATION, NULL, ASN1_INTEGER_GET(AdminRequest->body->d.serial))
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_REVOKE_ENTITY,			Entity_PKI::RevokeEntityCert,		LOG_MESSAGE_TYPE_ENTITY_REVOCATION, NULL, ASN1_INTEGER_GET(AdminRequest->body->d.serial))
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_SET_OFFLINE,				Entity_PKI::SetOfflineState,		LOG_MESSAGE_TYPE_SET_OFFLINE, NULL, ASN1_INTEGER_GET(AdminRequest->body->d.offline))
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_OFFLINE,				Entity_PKI::GetOfflineState)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_MY_ACL,				Entity_PKI::GetMyACL)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_GET_CONFIG,				Entity_PKI::GetConfiguration,		LOG_MESSAGE_TYPE_GET_CONFIGURATION, _sv("none"), LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_ENTITY_CONF,			Entity_PKI::GetEntityConf)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_SET_ENTITY_CONF,			Entity_PKI::SetEntityConf,			LOG_MESSAGE_TYPE_SET_ENTITY_CONF, (AdminRequest->body->d.entity_conf->entity_cert)?AdminRequest->body->d.entity_conf->entity_cert->name:NULL, LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_REPOSITORIES,		Entity_PKI::GetRepositoriesList)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_SET_REPOSITORIES,		Entity_PKI::SetRepositoriesList,	LOG_MESSAGE_TYPE_SET_REPOSITORIES, _sv("none"), LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_LOGS_TYPE,			Entity_PKI::GetLogsType)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_LOCK_RESOURCE,			Entity_PKI::UserLockResourse,		LOG_MESSAGE_TYPE_LOCK_RESOURCE, (AdminRequest->body->d.resource_name)?ASN1_STRING_GET(AdminRequest->body->d.resource_name):NULL, LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_UNLOCK_RESOURCE,			Entity_PKI::UserUnlockResourse,		LOG_MESSAGE_TYPE_UNLOCK_RESOURCE, (AdminRequest->body->d.resource_name)?ASN1_STRING_GET(AdminRequest->body->d.resource_name):NULL, LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_ENTITY_ACL,			Entity_PKI::GetEntityAcl)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_SET_ENTITY_ACL,			Entity_PKI::SetEntityAcl,			LOG_MESSAGE_TYPE_SET_ENTITY_ACL, (AdminRequest->body->d.entity_acl->entity_cert)?AdminRequest->body->d.entity_acl->entity_cert->name:(char*)"PKI Entity", LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_ENTITY_MAIL_CONF,	Entity_PKI::GetEntityMailConf)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_SET_ENTITY_MAIL_CONF,	Entity_PKI::SetEntityMailConf,		LOG_MESSAGE_TYPE_SET_ENTITY_MAIL_CONF, (AdminRequest->body->d.entity_mail_conf->entity_cert)?AdminRequest->body->d.entity_mail_conf->entity_cert->name:(char*)"PKI Entity", LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_SEND_ADMIN_MAIL,			Entity_PKI::AdminSendMail,			LOG_MESSAGE_TYPE_ADD_ADMIN_MAIL_QUEUE,	_sv("none"), LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_ENTITY_AUDITS,		Entity_PKI::GetEntityAudits)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_SET_ENTITY_AUDITS,		Entity_PKI::SetEntityAudits,		LOG_MESSAGE_TYPE_SET_ENTITY_AUDITS, (AdminRequest->body->d.entity_audits->entity_cert)?AdminRequest->body->d.entity_audits->entity_cert->name:NULL, LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_ENTITY_LOGS_TYPE,	Entity_PKI::GetEntityLogsType)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_ENTITIES_LINKS,		Entity_PKI::GetEntitiesLinks)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_SET_ENTITIES_LINKS,		Entity_PKI::SetEntitiesLinks,		LOG_MESSAGE_TYPE_SET_ENTITIES_LINKS, _sv("none"), LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_GROUPS,				Entity_PKI::GetGroups)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_SET_GROUPS,				Entity_PKI::SetGroups,				LOG_MESSAGE_TYPE_SET_GROUPS, _sv("none"), LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_ADD_GROUP,				Entity_PKI::AddGroup,				LOG_MESSAGE_TYPE_ADD_GROUP, ASN1_STRING_GET(body->d.group_name), LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_DEL_GROUP,				Entity_PKI::DelGroup,				LOG_MESSAGE_TYPE_DEL_GROUP, NULL, ASN1_INTEGER_GET(body->d.group_id))
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_REN_GROUP,				Entity_PKI::RenameGroup,			LOG_MESSAGE_TYPE_REN_GROUP, NULL, ASN1_INTEGER_GET(body->d.rename_group->id))
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_CHECK_LOGS,				Entity_PKI::CheckLogsIntegrity)
	PARSER_COMMAND_END(Entity_PKI)
}


bool Entity_PKI::LoginUser(UserHandle & hUser, int & UserType)
{
	if(!AclValidator.CanUserPerform(hUser.GetUserCert(), ACL_TYPE_AUTHENTICATE_ON_ENTITY))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	UserType = USER_TYPE_PKI;
	return true;
}

void Entity_PKI::LogoutUser(const UserHandle & hUser)
{
}


bool Entity_PKI::SignEntity(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	ENTITY_SIGNATURE_RESP * sign_entity;
	PKI_CERT resCert;
	X509_PUBKEY * pubKey;
	REP_ENTRY_INFO * repository;
	PUB_ENTRY_INFO * publication;
	ENTITY_CONF_CRYPTED * ExportedConf;
	int EntityType;
	unsigned long entity_cert_serial=0;
	unsigned long ocsp_cert_serial=0;
	X509 * EntityCert = NULL;
	char * EntityName;
	char * EntityEmail;

	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_MANAGE_ENTITIES))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	EntityName = ASN1_STRING_GET(body->d.sign_entity->Name);
	EntityEmail = ASN1_STRING_GET(body->d.sign_entity->Email);

	if(!EntityEmail || !EntityName || !*EntityEmail || !*EntityName)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	EntityType = body->d.sign_entity->body->type;
	//Doing basic verification
	switch(EntityType)
	{
		case ENTITY_TYPE_KEY_STORE:
		case ENTITY_TYPE_RA:
		case ENTITY_TYPE_CA:
			//We need to make sur there is a repository available
			if(!SKM_sk_num(REP_ENTRY_INFO, pki_conf->Repositories->list))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_NO_REPOSITORY);
				return false;
			}
			pubKey = body->d.sign_entity->body->d.EntityPubKey;
			break;
		case ENTITY_TYPE_PUBLICATION:
			//We need to make sur there is a repository available
			if(!SKM_sk_num(REP_ENTRY_INFO, pki_conf->Repositories->list))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_NO_REPOSITORY);
				return false;
			}
			pubKey = body->d.sign_entity->body->d.sign_pub->EntityPubKey;
		case ENTITY_TYPE_REPOSITORY:
			pubKey = body->d.sign_entity->body->d.sign_rep->EntityPubKey;
			break;
		default:
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
			return false;
			break;
	}


	ConfAccessLock.LockWrite();
	
	//Generate entity cert
	if(!InternalCA_SignForceDn(INTERNAL_CA_TYPE_ENTITY, pubKey, EntityName, EntityEmail, resCert))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	entity_cert_serial = resCert.GetSerial();

	sign_entity = ENTITY_SIGNATURE_RESP_new();
	if(!sign_entity)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		NewPKI_LockErr();
		Private_RevokeEntityCert(entity_cert_serial); 
		NewPKI_UnlockErr();
		ConfAccessLock.Unlock();
		return false;
	}

	//Generate the real signature response
	sign_entity->body->type = EntityType;
	switch(EntityType)
	{
		case ENTITY_TYPE_RA:
		case ENTITY_TYPE_CA:
		case ENTITY_TYPE_REPOSITORY:
		case ENTITY_TYPE_KEY_STORE:
			sign_entity->body->d.EntityCert = resCert.GetX509(true);
			EntityCert = sign_entity->body->d.EntityCert;
			break;

		case ENTITY_TYPE_PUBLICATION:
			sign_entity->body->d.sign_pub = ENTITY_SIGNATURE_RESP_PUB_new();
			if(!sign_entity->body->d.sign_pub)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
				ENTITY_SIGNATURE_RESP_free(sign_entity);
				NewPKI_LockErr();
				Private_RevokeEntityCert(entity_cert_serial); 
				NewPKI_UnlockErr();
				ConfAccessLock.Unlock();
				return false;
			}
			if(sign_entity->body->d.sign_pub->EntityCert)
				X509_free(sign_entity->body->d.sign_pub->EntityCert);
			sign_entity->body->d.sign_pub->EntityCert = resCert.GetX509(true);
			EntityCert = sign_entity->body->d.sign_pub->EntityCert;

			//Now we generate the OCSP cert
			if(!InternalCA_SignForceDn(INTERNAL_CA_TYPE_OCSP, body->d.sign_entity->body->d.sign_pub->OcspPubKey, EntityName, EntityEmail, resCert))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				ENTITY_SIGNATURE_RESP_free(sign_entity);
				NewPKI_LockErr();
				Private_RevokeEntityCert(entity_cert_serial); 
				NewPKI_UnlockErr();
				ConfAccessLock.Unlock();
				return false;
			}
			ocsp_cert_serial = resCert.GetSerial();

			if(sign_entity->body->d.sign_pub->OcspCert)
				X509_free(sign_entity->body->d.sign_pub->OcspCert);
			sign_entity->body->d.sign_pub->OcspCert = resCert.GetX509(true);
			break;
	}


	//If we just created a publication, we add it to the list
	if(EntityType == ENTITY_TYPE_PUBLICATION)
	{
		publication = PUB_ENTRY_INFO_new();
		if(!publication)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
			ENTITY_SIGNATURE_RESP_free(sign_entity);
			NewPKI_LockErr();
			Private_RevokeEntityCert(entity_cert_serial); 
			Private_RevokeEntityCert(ocsp_cert_serial); 
			NewPKI_UnlockErr();
			ConfAccessLock.Unlock();
			return false;
		}
		ASN1_STRING_SET(&publication->Name, ASN1_STRING_GET(body->d.sign_entity->Name));
		
		if(publication->PubOCSP)
			X509_free(publication->PubOCSP);
		CRYPTO_add(&sign_entity->body->d.sign_pub->OcspCert->references, 1, CRYPTO_LOCK_X509);
		publication->PubOCSP = sign_entity->body->d.sign_pub->OcspCert;

		
		if(publication->PubSSL)
			X509_free(publication->PubSSL);
		CRYPTO_add(&sign_entity->body->d.sign_pub->EntityCert->references, 1, CRYPTO_LOCK_X509);
		publication->PubSSL = sign_entity->body->d.sign_pub->EntityCert;

		if(SKM_sk_push(PUB_ENTRY_INFO, pki_conf->Publications->list, publication) < 0)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
			PUB_ENTRY_INFO_free(publication);
			ENTITY_SIGNATURE_RESP_free(sign_entity);
			NewPKI_LockErr();
			Private_RevokeEntityCert(entity_cert_serial); 
			Private_RevokeEntityCert(ocsp_cert_serial); 
			NewPKI_UnlockErr();
			ConfAccessLock.Unlock();
			return false;
		}
	}
	else if(EntityType == ENTITY_TYPE_REPOSITORY)
	{
		//If we just created a repository, we add it to the list
		repository = REP_ENTRY_INFO_new();
		if(!repository)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
			ENTITY_SIGNATURE_RESP_free(sign_entity);
			NewPKI_LockErr();
			Private_RevokeEntityCert(entity_cert_serial); 
			NewPKI_UnlockErr();
			ConfAccessLock.Unlock();
			return false;
		}
		ASN1_STRING_SET(&repository->Name, ASN1_STRING_GET(body->d.sign_entity->Name));
		ASN1_STRING_SET(&repository->Address, ASN1_STRING_GET(body->d.sign_entity->body->d.sign_rep->Address));
		ASN1_INTEGER_SET(&repository->Port, ASN1_INTEGER_GET(body->d.sign_entity->body->d.sign_rep->Port));
		
		if(repository->RepositorySSL)
			X509_free(repository->RepositorySSL);
		CRYPTO_add(&sign_entity->body->d.EntityCert->references, 1, CRYPTO_LOCK_X509);
		repository->RepositorySSL = sign_entity->body->d.EntityCert;

		if(SKM_sk_push(REP_ENTRY_INFO, pki_conf->Repositories->list, repository) < 0)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
			REP_ENTRY_INFO_free(repository);
			ENTITY_SIGNATURE_RESP_free(sign_entity);
			NewPKI_LockErr();
			Private_RevokeEntityCert(entity_cert_serial); 
			NewPKI_UnlockErr();
			ConfAccessLock.Unlock();
			return false;
		}
	}
	
	//Creating a default conf for the new entity
	if(!Add_ENTITY_CONF(EntityName, EntityCert, EntityType))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ENTITY_SIGNATURE_RESP_free(sign_entity);
		NewPKI_LockErr();
		Private_RevokeEntityCert(entity_cert_serial); 
		if(ocsp_cert_serial)
			Private_RevokeEntityCert(ocsp_cert_serial); 
		NewPKI_UnlockErr();
		ConfAccessLock.Unlock();
		return false;
	}

	//writing conf to Database
	if(!WritePersonnalConf())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ENTITY_SIGNATURE_RESP_free(sign_entity);
		NewPKI_LockErr();
		Private_RevokeEntityCert(entity_cert_serial); 
		if(ocsp_cert_serial)
			Private_RevokeEntityCert(ocsp_cert_serial); 
		NewPKI_UnlockErr();
		ConfAccessLock.Unlock();
		return false;
	}

	//We copy the exported conf to the response
	ExportedConf = EXPORTED_PKI_CONF_get_conf(EntityCert);
	if(!ExportedConf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		ENTITY_SIGNATURE_RESP_free(sign_entity);
		NewPKI_LockErr();
		Private_RevokeEntityCert(entity_cert_serial); 
		if(ocsp_cert_serial)
			Private_RevokeEntityCert(ocsp_cert_serial); 
		NewPKI_UnlockErr();
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ENTITY_CONF_CRYPTED_cpy(&sign_entity->conf, ExportedConf))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		ENTITY_SIGNATURE_RESP_free(sign_entity);
		NewPKI_LockErr();
		Private_RevokeEntityCert(entity_cert_serial); 
		if(ocsp_cert_serial)
			Private_RevokeEntityCert(ocsp_cert_serial); 
		NewPKI_UnlockErr();
		ConfAccessLock.Unlock();
		return false;
	}

	//We copy the internal ca certs to the signature response
	if(!INTERNAL_PKI_CA_cpy(&sign_entity->cas, pki_conf->cas))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ENTITY_SIGNATURE_RESP_free(sign_entity);
		NewPKI_LockErr();
		Private_RevokeEntityCert(entity_cert_serial); 
		if(ocsp_cert_serial)
			Private_RevokeEntityCert(ocsp_cert_serial); 
		NewPKI_UnlockErr();
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_SIGN_ENTITY, sign_entity))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ENTITY_SIGNATURE_RESP_free(sign_entity);
		NewPKI_LockErr();
		Private_RevokeEntityCert(entity_cert_serial); 
		if(ocsp_cert_serial)
			Private_RevokeEntityCert(ocsp_cert_serial); 
		NewPKI_UnlockErr();
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}


bool Entity_PKI::EnumEntities(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	ENTITY_ENTRY_INFO * entity;
	STACK_OF(ENTITY_ENTRY_INFO) * entities;
	int i;
	STORED_ENTITY_CONF * currConf;

	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	entities = sk_new_null();
	if(!entities)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	ConfAccessLock.LockRead();

	for(i=0; i<SKM_sk_num(STORED_ENTITY_CONF, pki_conf->confs); i++)
	{
		currConf = SKM_sk_value(STORED_ENTITY_CONF, pki_conf->confs, i);
		if(!currConf) continue;

		entity = ENTITY_ENTRY_INFO_new(); 
		if(!entity) continue;

		ASN1_STRING_SET(&entity->Name, ASN1_STRING_GET(currConf->conf->Name));
		ASN1_INTEGER_SET(&entity->Type, currConf->conf->body->type);
		if(!X509_cpy(&entity->certificate, currConf->certificate)) continue;
		if(SKM_sk_push(ENTITY_ENTRY_INFO, entities, entity) < 0)
		{
			ENTITY_ENTRY_INFO_free(entity);
		}
	}
	ConfAccessLock.Unlock();

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_ENUM_ENTITIES, entities))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		SKM_sk_pop_free(ENTITY_ENTRY_INFO, entities, ENTITY_ENTRY_INFO_free);
		return false;
	}
	return true;
}


bool Entity_PKI::EnumUsers(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	int i;
	CA_Handler * currCa;
	STACK_OF(USER_ENTRY_CERT) * users;
	USER_ENTRY_CERT * user;
	STACK_OF(INTERNAL_CA_CERT) * certs;
	INTERNAL_CA_CERT * currCert;
	bool IsAdmin;


	IsAdmin = AclValidator.IsPkiAdministrator(UserCert);

	currCa = GetInternalCA(INTERNAL_CA_TYPE_USER);
	if(!currCa)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	certs = currCa->get_Certs();
	if(!certs)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	users = sk_new_null();
	if(!users)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		SKM_sk_pop_free(INTERNAL_CA_CERT, certs, INTERNAL_CA_CERT_free);
		return false;
	}

	for(i=0; i<SKM_sk_num(INTERNAL_CA_CERT, certs) ; i++)
	{
		currCert = SKM_sk_value(INTERNAL_CA_CERT, certs, i);
		if(!currCert)
			continue;

		if(ASN1_INTEGER_GET(currCert->state) == CERT_STATE_REVOKED)
			continue;

		// Only a PKI Admin user can see other PKI Admin users
		if(!IsAdmin && AclValidator.IsPkiAdministrator(ASN1_INTEGER_GET(currCert->serial)))
			continue;

		user = USER_ENTRY_CERT_new();
		if(!user)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
			SKM_sk_pop_free(USER_ENTRY_CERT, users, USER_ENTRY_CERT_free);
			SKM_sk_pop_free(INTERNAL_CA_CERT, certs, INTERNAL_CA_CERT_free);
			return false;
		}

		if(!X509_cpy(&user->user_cert, currCert->cert))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
			SKM_sk_pop_free(USER_ENTRY_CERT, users, USER_ENTRY_CERT_free);
			SKM_sk_pop_free(INTERNAL_CA_CERT, certs, INTERNAL_CA_CERT_free);
			USER_ENTRY_CERT_free(user);
			return false;
		}
		
		if(SKM_sk_push(USER_ENTRY_CERT, users, user) < 0)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			SKM_sk_pop_free(USER_ENTRY_CERT, users, USER_ENTRY_CERT_free);
			SKM_sk_pop_free(INTERNAL_CA_CERT, certs, INTERNAL_CA_CERT_free);
			USER_ENTRY_CERT_free(user);
			return false;
		}
	}
	SKM_sk_pop_free(INTERNAL_CA_CERT, certs, INTERNAL_CA_CERT_free);

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_ENUM_USERS_CERT, users))
	{
		SKM_sk_pop_free(USER_ENTRY_CERT, users, USER_ENTRY_CERT_free);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_PKI::CreateUser(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	HashTable_Dn Dn;
	PKI_CSR csr;
	PKI_CERT cert;
	CA_Handler * currCa;
	int i;
	const char * value;
	const char * email;
	STACK_OF(X509) * certs;
	int What2Update = ENTITY_CONF_UPDATE_ACL;


	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_MANAGE_USERS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	//Only a PKI Admin can create another PKI Admin
	if(!AclValidator.IsPkiAdministrator(UserCert) && ASN1_INTEGER_GET(body->d.create_pki_user->pkiadmin))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}



	currCa = GetInternalCA(INTERNAL_CA_TYPE_USER);
	if(!currCa)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	if(!csr.SetCSR(body->d.create_pki_user->csr))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!csr.CheckSignature())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_CSR_SIGNATURE);
		return false;
	}


	//Get CN
	i = csr.GetRequestDN().SeekEntryName("commonName", -1);
	if( i == HASHTABLE_NOT_FOUND )
	{
		ERR_set_error_data("commonName", ERR_TXT_STRING);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}
	value = csr.GetRequestDN().Get(i);
	if(!value)
	{
		ERR_set_error_data("commonName", ERR_TXT_STRING);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}
	
	i = csr.GetRequestDN().SeekEntryName("emailAddress", -1);
	if( i == HASHTABLE_NOT_FOUND )
	{
		ERR_set_error_data("emailAddress", ERR_TXT_STRING);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}
	email = csr.GetRequestDN().Get(i);
	if(!email)
	{
		ERR_set_error_data("emailAddress", ERR_TXT_STRING);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}
	
	
	if(!InternalCA_SignForceDn(INTERNAL_CA_TYPE_USER, csr.GetX509_PUBKEY(), value, email, cert))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}
	
	//Create response
	certs = sk_X509_new_null();
	if(!certs)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		InternalCA_Revoke(INTERNAL_CA_TYPE_USER, cert.GetSerial());
		return false;
	}

	sk_X509_push(certs, RootCa->get_CaCert()->GetX509(true));
	sk_X509_push(certs, UsersCa->get_CaCert()->GetX509(true));
	sk_X509_push(certs, cert.GetX509(true));

	ConfAccessLock.LockWrite();
	if(ASN1_INTEGER_GET(body->d.create_pki_user->pkiadmin))
	{
		if(!DeclarePkiAdminCert(cert, email))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			sk_X509_pop_free(certs, X509_free);
			InternalCA_Revoke(INTERNAL_CA_TYPE_USER, cert.GetSerial());
			return false;
		}

		//Update my own ACL handler
		if(!AclValidator.SetACL(pki_conf->my_conf->acls))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			sk_X509_pop_free(certs, X509_free);
			InternalCA_Revoke(INTERNAL_CA_TYPE_USER, cert.GetSerial());
			return false;
		}
	
		//Yell to other entities the new admin
		What2Update |= ENTITY_CONF_UPDATE_ADMINS;
	}
	ConfAccessLock.Unlock();



	//Yell to other entities the new ACL
	if(!UpdateEntityConfs(What2Update))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		sk_X509_pop_free(certs, X509_free);
		InternalCA_Revoke(INTERNAL_CA_TYPE_USER, cert.GetSerial());
		return false;
	}

	if(!WritePersonnalConf())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		sk_X509_pop_free(certs, X509_free);
		InternalCA_Revoke(INTERNAL_CA_TYPE_USER, cert.GetSerial());
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_CREATE_PKI_USER, certs))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		sk_X509_pop_free(certs, X509_free);
		InternalCA_Revoke(INTERNAL_CA_TYPE_USER, cert.GetSerial());
		return false;
	}

	return true;
}

bool Entity_PKI::RevokeUserCert(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	PKI_ADMIN_ENTRY * currEntry;
	int i;
	int What2Update = ENTITY_CONF_UPDATE_ACL | ENTITY_CONF_UPDATE_CRL | ENTITY_CONF_UPDATE_GROUPS;
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_MANAGE_USERS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	// Only a PKI admin can revoke another PKI Admin
	if(!AclValidator.IsPkiAdministrator(UserCert.GetSerial()) && 
		AclValidator.IsPkiAdministrator(ASN1_INTEGER_GET(body->d.serial)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	ConfAccessLock.LockRead();
	// We need to have at least one PKI Admin
	if(AclValidator.IsPkiAdministrator(ASN1_INTEGER_GET(body->d.serial))  &&
		SKM_sk_num(ASN1_INTEGER, pki_conf->my_conf->acls->AdminSerials->admins) <= 1)
	{
		ConfAccessLock.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	ConfAccessLock.Unlock();

	if(!InternalCA_Revoke(INTERNAL_CA_TYPE_USER, ASN1_INTEGER_GET(body->d.serial)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(AclValidator.IsPkiAdministrator(ASN1_INTEGER_GET(body->d.serial)))
	{
		//We remove it from the list
		ConfAccessLock.LockWrite();
		for(i=0; i < SKM_sk_num(ASN1_INTEGER, pki_conf->my_conf->acls->AdminSerials->admins) ; i++)
		{
			currEntry = SKM_sk_value(PKI_ADMIN_ENTRY, pki_conf->my_conf->acls->AdminSerials->admins, i);
			if(!currEntry) continue;

			if(ASN1_INTEGER_cmp(currEntry->serial, body->d.serial) == 0)
			{
				SKM_sk_delete(PKI_ADMIN_ENTRY, pki_conf->my_conf->acls->AdminSerials->admins, i);
				PKI_ADMIN_ENTRY_free(currEntry);
				break;
			}
		}
		ConfAccessLock.Unlock();
		What2Update |= ENTITY_CONF_UPDATE_ADMINS;
	}

	// Yell to other entities the new ACL and the new CRL
	if(!UpdateEntityConfs(What2Update))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	//Remove user from the ACLs
	if(!WritePersonnalConf())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_PKI::Private_RevokeEntityCert(unsigned long serial)
{
	int i;
	int j;
	int k;
	bool FoundRep;
	REP_ENTRY_INFO * currRep;
	PUB_ENTRY_INFO * currPub;
	STORED_ENTITY_CONF * currConf;
	ENTITY_CONF_CRYPTED * c_currConf;
	ENTITY_ENTRY_INFO * currDstLink;
	ENTITY_LINKS * currLink;
	int What2Update = ENTITY_CONF_UPDATE_LINKS;

	// Search in publications list for this entity cert
	// if we find it then we revoke the OCSP cert as well
	// and we remove it from the publications list
	for(i=0; i<SKM_sk_num(PUB_ENTRY_INFO, pki_conf->Publications->list); i++)
	{
		currPub = SKM_sk_value(PUB_ENTRY_INFO, pki_conf->Publications->list, i);
		if(!currPub) continue;

		//Does the publication cert have our serial ?
		if((unsigned long)ASN1_INTEGER_GET(X509_get_serialNumber(currPub->PubSSL)) == serial)
		{
			//We have to revoke the OCSP cert and remove it from the list
			if(!InternalCA_Revoke(INTERNAL_CA_TYPE_OCSP, ASN1_INTEGER_GET(X509_get_serialNumber(currPub->PubOCSP))))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				ConfAccessLock.Unlock();
				return false;
			}

			PUB_ENTRY_INFO_free(currPub);
			SKM_sk_delete(PUB_ENTRY_INFO, pki_conf->Publications->list, i);
			break;
		}
	}
	// Did we find it as in the publication list ?
	if(i == SKM_sk_num(PUB_ENTRY_INFO, pki_conf->Publications->list))
	{	
		// Search in repositories list for this entity cert
		// if we find it then we remove it from the repositories list
		for(i=0; i<SKM_sk_num(REP_ENTRY_INFO, pki_conf->Repositories->list); i++)
		{
			currRep = SKM_sk_value(REP_ENTRY_INFO, pki_conf->Repositories->list, i);
			if(!currRep) continue;

			//Does the repository cert have our serial ?
			if((unsigned long)ASN1_INTEGER_GET(X509_get_serialNumber(currRep->RepositorySSL)) == serial)
			{
				// We check that we can revoke it, meaning, is it 
				// the only repository for some of the others entities ?
				for(j=0; j<SKM_sk_num(ENTITY_LINKS, pki_conf->links->links); j++)
				{
					currLink = SKM_sk_value(ENTITY_LINKS, pki_conf->links->links, j);
					if(!currLink) continue;

					// The PKI Entity can ignore all reps
					if(ASN1_INTEGER_GET(currLink->src->Type) == ENTITY_TYPE_PKI ||
						ASN1_INTEGER_GET(currLink->src->Type) == ENTITY_TYPE_REPOSITORY)
						continue;

					// Check to see if the respository to revoke 
					// is the only one of the current conf !
					FoundRep = false;
					for(k=0;k<SKM_sk_num(ENTITY_ENTRY_INFO, currLink->dsts);k++)
					{
						currDstLink = SKM_sk_value(ENTITY_ENTRY_INFO, currLink->dsts, k);
						if(!currDstLink) continue;

						// We only want repositories
						if(ASN1_INTEGER_GET(currDstLink->Type) != ENTITY_TYPE_REPOSITORY)
							continue;

						//Is it the seeked one ?
						if(ASN1_STRING_cmp(currDstLink->Name, currRep->Name) != 0)
						{
							FoundRep = true;
							break;
						}
						FoundRep = true;
					}

					//Did we find another repository ?
					if(FoundRep && k == SKM_sk_num(ENTITY_ENTRY_INFO, currLink->dsts))
					{
						NEWPKIerr(PKI_ERROR_TXT, ERROR_LAST_ENTITY_REPOSITORY);
						ERR_set_error_data(strdup(ASN1_STRING_GET(currLink->src->Name)), ERR_TXT_STRING | ERR_TXT_MALLOCED);
						return false;
					}
				}

				REP_ENTRY_INFO_free(currRep);
				SKM_sk_delete(REP_ENTRY_INFO, pki_conf->Repositories->list, i);
				break;
			}
		}
	}

	//We need to delete the conf for this entity
	for(i=0; i<SKM_sk_num(STORED_ENTITY_CONF, pki_conf->confs); i++)
	{
		currConf = SKM_sk_value(STORED_ENTITY_CONF, pki_conf->confs, i);
		if(!currConf) continue;
		
		//Does the conf cert have our serial ?
		if((unsigned long)ASN1_INTEGER_GET(X509_get_serialNumber(currConf->certificate)) == serial)
		{
			//We need to delete the crypted conf for this entity
			for(j=0; j<SKM_sk_num(ENTITY_CONF_CRYPTED, pki_conf->crypt_confs->confs->all_confs); j++)
			{
				c_currConf = SKM_sk_value(ENTITY_CONF_CRYPTED, pki_conf->crypt_confs->confs->all_confs, j);
				if(!c_currConf) continue;
				
				//Is it the seeked one ?
				if(currConf->certificate->cert_info->key->public_key->length !=
					c_currConf->recipient->public_key->length ||
					memcmp(currConf->certificate->cert_info->key->public_key->data,
					c_currConf->recipient->public_key->data, 
					c_currConf->recipient->public_key->length) != 0)
					continue;


				ENTITY_CONF_CRYPTED_free(c_currConf);
				SKM_sk_delete(ENTITY_CONF_CRYPTED, pki_conf->crypt_confs->confs->all_confs, j);
				break;
			}

			//We remove its entry in the links
			for(j=0; j < SKM_sk_num(ENTITY_LINKS, pki_conf->links->links); j++)
			{
				currLink = SKM_sk_value(ENTITY_LINKS, pki_conf->links->links, j);
				if(!currLink) continue;
				if(ASN1_STRING_cmp(currLink->src->Name, currConf->conf->Name) != 0) 
					continue;

				ENTITY_LINKS_free(currLink);
				SKM_sk_delete(ENTITY_LINKS, pki_conf->links->links, j);
				break;
			}

			STORED_ENTITY_CONF_free(currConf);
			SKM_sk_delete(STORED_ENTITY_CONF, pki_conf->confs, i);
			break;
		}
	}

	//We revoke the entity cert
	if(!InternalCA_Revoke(INTERNAL_CA_TYPE_ENTITY, serial))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	
	//Yell to other entities the new CRL
	What2Update |= ENTITY_CONF_UPDATE_CRL;
	if(!UpdateEntityConfs(What2Update))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	// We flush the new conf
	if(!WritePersonnalConf())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_PKI::RevokeEntityCert(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{

	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_MANAGE_ENTITIES))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}


	ConfAccessLock.LockWrite();

	if(!Private_RevokeEntityCert(ASN1_INTEGER_GET(body->d.serial)))
	{
		ConfAccessLock.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	ConfAccessLock.Unlock();

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

CA_Handler * Entity_PKI::GetInternalCA(INTERNAL_CA_TYPE Type)
{
	switch(Type)
	{
		case INTERNAL_CA_TYPE_ROOT:
			return RootCa;
			break;
		case INTERNAL_CA_TYPE_USER:
			return UsersCa;
			break;
		case INTERNAL_CA_TYPE_ENTITY:
			return EntitiesCa;
			break;
		case INTERNAL_CA_TYPE_OCSP:
			return OcspCa;
			break;
		default:
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
			return NULL;
			break;
	}
}

bool Entity_PKI::GetOfflineState(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{

	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_MODIFY_PKI_STATUS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_STATUS, ASN1_INTEGER_dup(MY_CONF_PTR->offline)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_PKI::SetOfflineState(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{

	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_MODIFY_PKI_STATUS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}


	ConfAccessLock.LockWrite();
	ASN1_INTEGER_SET(&MY_CONF_PTR->offline, ASN1_INTEGER_GET(body->d.offline));
	if(!WritePersonnalConf())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

void Entity_PKI::GetACL_List(STACK_OF(ASN1_INTEGER) *acl_list)
{
	ASN1_INTEGER * acl_id;
	int i;
	static ACL_TYPE list_acls[] =
	{
		ACL_TYPE_MANAGE_USERS,
		ACL_TYPE_MANAGE_ENTITIES,
		ACL_TYPE_VIEW_LOGS,
		ACL_TYPE_MANAGE_ACLS,
		ACL_TYPE_MODIFY_PKI_STATUS, 
		ACL_TYPE_SEND_ADMIN_MAIL,
		(ACL_TYPE)0
	};
	for(i=0; list_acls[i]; i++)
	{
		acl_id = NULL;
		ASN1_INTEGER_SET(&acl_id, list_acls[i]);
		if(acl_id)
		{
			if(SKM_sk_push(ASN1_INTEGER, acl_list, acl_id) < 0)
			{
				ASN1_INTEGER_free(acl_id);
			}
		}
	}
}

bool Entity_PKI::GetConfiguration(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	EXPORTED_PKI_CONF * conf = NULL;

	ConfAccessLock.LockRead();
	if(!EXPORTED_PKI_CONF_cpy(&conf, pki_conf->crypt_confs))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();


	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_CONF, conf))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	return true;
}

bool Entity_PKI::GetEntityConf(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	STORED_ENTITY_CONF * conf = NULL;
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_MANAGE_ENTITIES))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	if(!ResourceOwnerIsMe(body->d.entity_cert->name))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	ConfAccessLock.LockRead();
	
	conf = STORED_ENTITY_CONF_get(pki_conf->confs, body->d.entity_cert);
	if(!conf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN_ENTITY);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_ENTITY_CONF, ENTITY_CONF_BODY_cpy(NULL, conf->conf->body)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}

bool Entity_PKI::SetEntityConf(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	STORED_ENTITY_CONF * conf = NULL;
	bool ret;

	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_MANAGE_ENTITIES))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!ResourceOwnerIsMe(body->d.entity_conf->entity_cert->name))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	ConfAccessLock.LockWrite();
	
	conf = STORED_ENTITY_CONF_get(pki_conf->confs, body->d.entity_conf->entity_cert);
	if(!conf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN_ENTITY);
		ConfAccessLock.Unlock();
		return false;
	}

	//Verify entity conf
	switch(conf->conf->body->type)
	{
		case ENTITY_TYPE_CA:
			ret = CheckCaConf(body->d.entity_conf->conf->d.ca_conf);
			break;
		case ENTITY_TYPE_RA:
			ret = CheckRaConf(body->d.entity_conf->conf->d.ra_conf);
			break;
		case ENTITY_TYPE_REPOSITORY:
			ret = CheckRepositoryConf(body->d.entity_conf->conf->d.rep_conf);
			break;
		default:
			ret = true;
			break;
	}
	
	if(!ENTITY_CONF_BODY_cpy(&conf->conf->body, body->d.entity_conf->conf))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	//We update the crypted conf
	if(!UpdateEntityConfs(0, conf->conf, true))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!WritePersonnalConf())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	ConfAccessLock.Unlock();
	return true;
}

bool Entity_PKI::CheckRepositoryConf(ENTITY_CONF_BODY_REP *Conf)
{
	return true;
}

bool Entity_PKI::CheckRaConf(ENTITY_CONF_BODY_RA *Conf)
{
	int i, j;
	DN_SPECS * CurrField;
	POLICY_VALUE * CurrPolicy;
	bool ret;

	//We check the validity of the DN fields
	ret = true;
	for(i=0; i<SKM_sk_num(DN_SPECS, Conf->d.RA_CONF_PTR->DnSpecs); i++)
	{
		CurrField = SKM_sk_value(DN_SPECS, Conf->d.RA_CONF_PTR->DnSpecs, i);
		if(!CurrField) continue;

		if(!IsValidObject(ASN1_STRING_GET(CurrField->Name)))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_GET_DN_SPECIF);
			ERR_set_error_data(strdup(ASN1_STRING_GET(CurrField->Name)), ERR_TXT_STRING | ERR_TXT_MALLOCED);
			ret = false;
		}
	}

	//We remove any policy field that cannot be found in the DN specs
	for(i=0; i<SKM_sk_num(POLICY_VALUE, Conf->d.RA_CONF_PTR->Policies); i++)
	{
		CurrPolicy = SKM_sk_value(POLICY_VALUE, Conf->d.RA_CONF_PTR->Policies, i);
		if(!CurrPolicy) continue;

		if(HashTable_String::IsValidPolicyField(ASN1_STRING_GET(CurrPolicy->name), ASN1_STRING_GET(CurrPolicy->value)) == DN_TYPE_UNKNOWN)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}

		for(j=0; j<SKM_sk_num(DN_SPECS, Conf->d.RA_CONF_PTR->DnSpecs); j++)
		{
			CurrField = SKM_sk_value(DN_SPECS, Conf->d.RA_CONF_PTR->DnSpecs, j);
			if(!CurrField) continue;

			if(ASN1_STRING_cmp(CurrPolicy->name, CurrField->Name) == 0)
				break;
		}
		//The DN spec is not known, remove the policy entry
		if(j == SKM_sk_num(DN_SPECS, Conf->d.RA_CONF_PTR->DnSpecs))
		{
			SKM_sk_delete(POLICY_VALUE, Conf->d.RA_CONF_PTR->Policies, i);
			POLICY_VALUE_free(CurrPolicy);
			i--;
		}
	}

	// We set all the DN fields not found
	// in the policy to "optional"
	for(i=0; i<SKM_sk_num(DN_SPECS, Conf->d.RA_CONF_PTR->DnSpecs); i++)
	{
		CurrField = SKM_sk_value(DN_SPECS, Conf->d.RA_CONF_PTR->DnSpecs, i);
		if(!CurrField) continue;

		for(j=0; j<SKM_sk_num(POLICY_VALUE, Conf->d.RA_CONF_PTR->Policies); j++)
		{
			CurrPolicy = SKM_sk_value(POLICY_VALUE, Conf->d.RA_CONF_PTR->Policies, j);
			if(!CurrPolicy) continue;

			if(ASN1_STRING_cmp(CurrPolicy->name, CurrField->Name) == 0)
				break;
		}
		//The DN field is not known, add a policy entry
		if(j == SKM_sk_num(POLICY_VALUE, Conf->d.RA_CONF_PTR->Policies))
		{
			CurrPolicy = POLICY_VALUE_new();
			if(CurrPolicy)
			{
				ASN1_STRING_SET(&CurrPolicy->name, ASN1_STRING_GET(CurrField->Name));
				ASN1_STRING_SET(&CurrPolicy->value, STR_OPTIONAL);
				if(SKM_sk_push(POLICY_VALUE, Conf->d.RA_CONF_PTR->Policies, CurrPolicy) < 0)
				{
					POLICY_VALUE_free(CurrPolicy);
				}
			}
		}
	}


	return ret;
}


bool Entity_PKI::CheckCaConf(ENTITY_CONF_BODY_CA *Conf)
{
	int i;
	EXTENSION_VALUE * CurrExt;
	bool ret;

	//We check the validity of the extensions
	ret = true;
	for(i=0; i<SKM_sk_num(EXTENSION_VALUE, Conf->d.CA_CONF_PTR->Exts); i++)
	{
		CurrExt = SKM_sk_value(EXTENSION_VALUE, Conf->d.CA_CONF_PTR->Exts, i);
		if(!CurrExt) continue;

		if(!PKI_EXT::VerifyExtensionValue(ASN1_STRING_GET(CurrExt->name), ASN1_STRING_GET(CurrExt->value)))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_GET_DN_SPECIF);
			ERR_add_error_data(2, "Cert Exts:", ASN1_STRING_GET(CurrExt->name));
			ret = false;
		}
	}

	for(i=0; i<SKM_sk_num(EXTENSION_VALUE, Conf->d.CA_CONF_PTR->CrlExts); i++)
	{
		CurrExt = SKM_sk_value(EXTENSION_VALUE, Conf->d.CA_CONF_PTR->CrlExts, i);
		if(!CurrExt) continue;

		if(!PKI_EXT::VerifyExtensionValue(ASN1_STRING_GET(CurrExt->name), ASN1_STRING_GET(CurrExt->value)))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_GET_DN_SPECIF);
			ERR_add_error_data(2, "Crl Exts:", ASN1_STRING_GET(CurrExt->name));
			ret = false;
		}
	}

	return ret;
}


bool Entity_PKI::GetRepositoriesList(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	REPS_INFO * repositories = NULL;
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_MANAGE_ENTITIES))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	if(!ResourceOwnerIsMe(PKI_REPOSITORIES_RESOURCE))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	ConfAccessLock.LockRead();
	if(!REPS_INFO_cpy(&repositories, pki_conf->Repositories))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	
	
	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_GET_REPOSITORIES, repositories))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	return true;
}

bool Entity_PKI::SetRepositoriesList(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	int i;
	int j;
	REP_ENTRY_INFO * rep1;
	REP_ENTRY_INFO * rep2;
	REPS_INFO * tmpRep1 = NULL;
	REPS_INFO * tmpRep2 = NULL;
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_MANAGE_ENTITIES))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	if(!ResourceOwnerIsMe(PKI_REPOSITORIES_RESOURCE))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	ConfAccessLock.LockWrite();
	
	if(! (tmpRep1 = REPS_INFO_cpy(NULL, pki_conf->Repositories)) )
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	if(! (tmpRep2 = REPS_INFO_cpy(NULL, body->d.repositories)) )
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		REPS_INFO_free(tmpRep1);
		ConfAccessLock.Unlock();
		return false;
	}

	// We validate the repositories, to make sure
	// the user didn't send something phoney
	for(i=0; i<SKM_sk_num(REP_ENTRY_INFO, tmpRep1->list); i++)
	{
		rep1 = SKM_sk_value(REP_ENTRY_INFO, tmpRep1->list, i);
		if(!rep1) continue;

		for(j=0; j<SKM_sk_num(REP_ENTRY_INFO, tmpRep2->list); j++)
		{
			rep2 = SKM_sk_value(REP_ENTRY_INFO, tmpRep2->list, j);
			if(!rep2) continue;

			if(X509_cmp(rep1->RepositorySSL, rep2->RepositorySSL) == 0 &&
				ASN1_STRING_cmp(rep1->Name, rep2->Name) == 0)
			{
				SKM_sk_delete(REP_ENTRY_INFO, tmpRep1->list, i);
				SKM_sk_delete(REP_ENTRY_INFO, tmpRep2->list, j);
				i--;
				REP_ENTRY_INFO_free(rep1);
				REP_ENTRY_INFO_free(rep2);
				break;
			}
		}		
	}

	if(SKM_sk_num(REP_ENTRY_INFO, tmpRep1->list) > 0 ||
		SKM_sk_num(REP_ENTRY_INFO, tmpRep2->list) > 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		REPS_INFO_free(tmpRep1);
		REPS_INFO_free(tmpRep2);
		ConfAccessLock.Unlock();
		return false;
	}
	REPS_INFO_free(tmpRep1);
	REPS_INFO_free(tmpRep2);

	if(!REPS_INFO_cpy(&pki_conf->Repositories, body->d.repositories))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	//Yell to other entities the new repositories list
	if(!UpdateEntityConfs(ENTITY_CONF_UPDATE_LINKS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	
	if(!WritePersonnalConf())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	
	
	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	return true;
}

X509_ACL_INIT Entity_PKI::X509_ACL_init(X509_ACL * acls, int EntityType)
{
	bool Modified = false;
	int i;
	int j;
	CA_Handler * usersCa;
	INTERNAL_CA_CERT * currCert;
	USERS_GROUP * currGroup;
	ACL_ENTRY * currAcl;
	bool Found;
	ASN1_INTEGER * currAclType;
	STACK_OF(INTERNAL_CA_CERT) * certs;

	usersCa = GetInternalCA(INTERNAL_CA_TYPE_USER);
	if(!usersCa)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return X509_ACL_INIT_ERROR;
	}

	certs = usersCa->get_Certs();
	if(!certs)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return X509_ACL_INIT_ERROR;
	}

	for(i=0; i<SKM_sk_num(INTERNAL_CA_CERT, certs); i++)
	{
		currCert = SKM_sk_value(INTERNAL_CA_CERT, certs, i);
		if(!currCert) continue;

		//We never add the an admin cert in the ACLs, it has all the rights
		if(X509_ACL_Validator::Static_IsPkiAdministrator(currCert->cert, acls->AdminSerials))
			continue;
		

		// We add the new certs and delete the revoked ones
		Found = false;
		for(j=0; j<SKM_sk_num(ACL_ENTRY, acls->acl_entries); j++)
		{
			currAcl = SKM_sk_value(ACL_ENTRY, acls->acl_entries, j);

			if(!currAcl || ASN1_INTEGER_GET(currAcl->type) != ACL_ENTRY_TYPE_USER) continue;

			//Is it the certificate
			if(ASN1_INTEGER_cmp(X509_get_serialNumber(currCert->cert), currAcl->serial) == 0)
			{
				//Is it revoked
				if(ASN1_INTEGER_GET(currCert->state) == CERT_STATE_REVOKED)
				{
					ACL_ENTRY_free(currAcl);
					SKM_sk_delete(ACL_ENTRY, acls->acl_entries, j);
					j--;
					Modified=true;
				}
				Found = true;
			}
		}

		//If we don't know the user, we add it to the list
		if(!Found && (ASN1_INTEGER_GET(currCert->state) == CERT_STATE_VALID))
		{
			//We add the user in the ACL, with no rights
			currAcl = ACL_ENTRY_new();
			if(!currAcl)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
				SKM_sk_pop_free(INTERNAL_CA_CERT, certs, INTERNAL_CA_CERT_free);
				return X509_ACL_INIT_ERROR;
			}
			ASN1_INTEGER_SET(&currAcl->type, ACL_ENTRY_TYPE_USER);
			ASN1_INTEGER_SET(&currAcl->serial, ASN1_INTEGER_GET(X509_get_serialNumber(currCert->cert)));
			ASN1_STRING_SET(&currAcl->name, currCert->cert->name);
			Modified = true;
			
			if(SKM_sk_push(ACL_ENTRY, acls->acl_entries, currAcl) < 0)
			{
				ACL_ENTRY_free(currAcl);
			}
		}
	}
	SKM_sk_pop_free(INTERNAL_CA_CERT, certs, INTERNAL_CA_CERT_free);


	//Update the ACL with the groups
	for(i=0; i<SKM_sk_num(USERS_GROUP, pki_conf->my_conf->groups->list); i++)
	{
		currGroup = SKM_sk_value(USERS_GROUP, pki_conf->my_conf->groups->list, i);
		if(!currGroup) continue;

		// We add the new groups
		Found = false;
		for(j=0; j<SKM_sk_num(ACL_ENTRY, acls->acl_entries); j++)
		{
			currAcl = SKM_sk_value(ACL_ENTRY, acls->acl_entries, j);
			if(!currAcl || ASN1_INTEGER_GET(currAcl->type) != ACL_ENTRY_TYPE_GROUP) continue;

			//Is it the group
			if(ASN1_INTEGER_cmp(currGroup->serial, currAcl->serial) == 0)
			{
				Found = true;
				break;
			}
		}

		//If we don't know the user, we add it to the list
		if(!Found)
		{
			//We add the user in the ACL, with no rights
			currAcl = ACL_ENTRY_new();
			if(!currAcl)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
				SKM_sk_pop_free(INTERNAL_CA_CERT, certs, INTERNAL_CA_CERT_free);
				return X509_ACL_INIT_ERROR;
			}
			ASN1_INTEGER_SET(&currAcl->type, ACL_ENTRY_TYPE_GROUP);
			ASN1_INTEGER_SET(&currAcl->serial, ASN1_INTEGER_GET(currGroup->serial));
			ASN1_STRING_SET(&currAcl->name, ASN1_STRING_GET(currGroup->name));
			Modified = true;
			
			if(SKM_sk_push(ACL_ENTRY, acls->acl_entries, currAcl) < 0)
			{
				ACL_ENTRY_free(currAcl);
			}
		}
	}

	//We now delete the removed groups
	for(i=0; i<SKM_sk_num(ACL_ENTRY, acls->acl_entries); i++)
	{
		currAcl = SKM_sk_value(ACL_ENTRY, acls->acl_entries, i);
		if(!currAcl || ASN1_INTEGER_GET(currAcl->type) != ACL_ENTRY_TYPE_GROUP) continue;

		//Search corresponding group
		Found = false;
		for(j=0; j<SKM_sk_num(USERS_GROUP, pki_conf->my_conf->groups->list); j++)
		{
			currGroup = SKM_sk_value(USERS_GROUP, pki_conf->my_conf->groups->list, j);
			if(!currGroup) continue;

			if(ASN1_STRING_cmp(currGroup->name, currAcl->name) == 0)
			{
				Found = true;
				break;
			}
		}

		//Group was no found we remove it
		if(!Found)
		{
			ACL_ENTRY_free(currAcl);
			SKM_sk_delete(ACL_ENTRY, acls->acl_entries, i);
			i--;
			Modified=true;
		}
	}

	
	
	
	//We initialize the available ACL types
	while( (currAclType = SKM_sk_pop(ASN1_INTEGER, acls->acl_type)) )
	{
		ASN1_INTEGER_free(currAclType);
	}

	switch(EntityType)
	{
		case ENTITY_TYPE_PKI:
			Entity_PKI::GetACL_List(acls->acl_type);
			break;
			
		case ENTITY_TYPE_REPOSITORY:
			Entity_REPOSITORY::GetACL_List(acls->acl_type);
			break;
			
		case ENTITY_TYPE_RA:
			Entity_RA::GetACL_List(acls->acl_type);
			break;
			
		case ENTITY_TYPE_KEY_STORE:
			Entity_KEYSTORE::GetACL_List(acls->acl_type);
			break;
			
		case ENTITY_TYPE_PUBLICATION:
			Entity_PUBLICATION::GetACL_List(acls->acl_type);
			break;
			
		case ENTITY_TYPE_CA:
			Entity_CA::GetACL_List(acls->acl_type);
			break;
	}

	if(Modified)
		return X509_ACL_INIT_MODIFIED;
	else
		return X509_ACL_INIT_SAME;
}

bool Entity_PKI::Add_ENTITY_CONF(char * EntityName, X509 *EntityCert, int Type)
{
	STORED_ENTITY_CONF * newConf;
	ENTITY_LINKS * link;
	int i, ConfIndex, LinkIndex;
	REP_ENTRY_INFO * currRep;
	ENTITY_ENTRY_INFO * currLinkEntry;
	ENTITY_LINKS * currLink;
	DN_SPECS * newDnSpecs;

	newConf = STORED_ENTITY_CONF_new();
	if(!newConf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	
	//Copy entity cert
	if(!X509_cpy(&newConf->certificate, EntityCert))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		STORED_ENTITY_CONF_free(newConf);
		return false;
	}

	ASN1_STRING_SET(&newConf->conf->Name, EntityName);
	
	//Copy default PKI email conf
	if(!EMAIL_CONF_cpy(&newConf->conf->mail_conf, pki_conf->my_conf->mail_conf))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		STORED_ENTITY_CONF_free(newConf);
		return false;
	}

	//Allocate a default conf
	newConf->conf->body->type = Type;
	switch(Type)
	{
		case ENTITY_TYPE_CA:
			newConf->conf->body->d.ca_conf = ENTITY_CONF_BODY_CA_new();
			if(!newConf->conf->body->d.ca_conf)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
				STORED_ENTITY_CONF_free(newConf);
				return false;
			}
			newConf->conf->body->d.ca_conf->type = CA_CONF_VERSION;
			newConf->conf->body->d.ca_conf->d.ptr = CA_CONF_STRUCT_new();
			if(!newConf->conf->body->d.ca_conf->d.ptr)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
				STORED_ENTITY_CONF_free(newConf);
				return false;
			}
			break;
		case ENTITY_TYPE_RA:
			newConf->conf->body->d.ra_conf = ENTITY_CONF_BODY_RA_new();
			if(!newConf->conf->body->d.ra_conf)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
				STORED_ENTITY_CONF_free(newConf);
				return false;
			}

			newConf->conf->body->d.ra_conf->type = RA_CONF_VERSION;
			newConf->conf->body->d.ra_conf->d.ptr = RA_CONF_STRUCT_new();
			if(!newConf->conf->body->d.ra_conf->d.ptr)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
				STORED_ENTITY_CONF_free(newConf);
				return false;
			}

			newDnSpecs = DN_SPECS_new();
			if(newDnSpecs)
			{
				ASN1_STRING_SET(&newDnSpecs->Name, "commonName");
				ASN1_STRING_SET(&newDnSpecs->Default, "");
				ASN1_STRING_SET(&newDnSpecs->Comment, "Common Name (eg, YOUR name)");
				ASN1_INTEGER_SET(&newDnSpecs->Min, 0);
				ASN1_INTEGER_SET(&newDnSpecs->Max, 64);
				if(SKM_sk_push(DN_SPECS, newConf->conf->body->d.ra_conf->d.RA_CONF_PTR->DnSpecs, newDnSpecs) < 0)
				{
					DN_SPECS_free(newDnSpecs);
				}
			}

			newDnSpecs = DN_SPECS_new();
			if(newDnSpecs)
			{
				ASN1_STRING_SET(&newDnSpecs->Name, "emailAddress");
				ASN1_STRING_SET(&newDnSpecs->Default, "");
				ASN1_STRING_SET(&newDnSpecs->Comment, "Email Address");
				ASN1_INTEGER_SET(&newDnSpecs->Min, 2);
				ASN1_INTEGER_SET(&newDnSpecs->Max, 40);
				if(SKM_sk_push(DN_SPECS, newConf->conf->body->d.ra_conf->d.RA_CONF_PTR->DnSpecs, newDnSpecs) < 0)
				{
					DN_SPECS_free(newDnSpecs);
				}
			}

			newDnSpecs = DN_SPECS_new();
			if(newDnSpecs)
			{
				ASN1_STRING_SET(&newDnSpecs->Name, "countryName");
				ASN1_STRING_SET(&newDnSpecs->Default, "");
				ASN1_STRING_SET(&newDnSpecs->Comment, "Country Name");
				ASN1_INTEGER_SET(&newDnSpecs->Min, 2);
				ASN1_INTEGER_SET(&newDnSpecs->Max, 2);
				if(SKM_sk_push(DN_SPECS, newConf->conf->body->d.ra_conf->d.RA_CONF_PTR->DnSpecs, newDnSpecs) < 0)
				{
					DN_SPECS_free(newDnSpecs);
				}
			}

			newDnSpecs = DN_SPECS_new();
			if(newDnSpecs)
			{
				ASN1_STRING_SET(&newDnSpecs->Name, "stateOrProvinceName");
				ASN1_STRING_SET(&newDnSpecs->Default, "");
				ASN1_STRING_SET(&newDnSpecs->Comment, "State or Province Name (full name)");
				ASN1_INTEGER_SET(&newDnSpecs->Min, 0);
				ASN1_INTEGER_SET(&newDnSpecs->Max, 0);
				if(SKM_sk_push(DN_SPECS, newConf->conf->body->d.ra_conf->d.RA_CONF_PTR->DnSpecs, newDnSpecs) < 0)
				{
					DN_SPECS_free(newDnSpecs);
				}
			}

			newDnSpecs = DN_SPECS_new();
			if(newDnSpecs)
			{
				ASN1_STRING_SET(&newDnSpecs->Name, "localityName");
				ASN1_STRING_SET(&newDnSpecs->Default, "");
				ASN1_STRING_SET(&newDnSpecs->Comment, "Locality Name (eg, city)");
				ASN1_INTEGER_SET(&newDnSpecs->Min, 0);
				ASN1_INTEGER_SET(&newDnSpecs->Max, 0);
				if(SKM_sk_push(DN_SPECS, newConf->conf->body->d.ra_conf->d.RA_CONF_PTR->DnSpecs, newDnSpecs) < 0)
				{
					DN_SPECS_free(newDnSpecs);
				}
			}

			newDnSpecs = DN_SPECS_new();
			if(newDnSpecs)
			{
				ASN1_STRING_SET(&newDnSpecs->Name, "organizationName");
				ASN1_STRING_SET(&newDnSpecs->Default, "");
				ASN1_STRING_SET(&newDnSpecs->Comment, "Organization Name (eg, company)");
				ASN1_INTEGER_SET(&newDnSpecs->Min, 0);
				ASN1_INTEGER_SET(&newDnSpecs->Max, 0);
				if(SKM_sk_push(DN_SPECS, newConf->conf->body->d.ra_conf->d.RA_CONF_PTR->DnSpecs, newDnSpecs) < 0)
				{
					DN_SPECS_free(newDnSpecs);
				}
			}

			newDnSpecs = DN_SPECS_new();
			if(newDnSpecs)
			{
				ASN1_STRING_SET(&newDnSpecs->Name, "organizationalUnitName");
				ASN1_STRING_SET(&newDnSpecs->Default, "");
				ASN1_STRING_SET(&newDnSpecs->Comment, "Organizational Unit Name (eg, section)");
				ASN1_INTEGER_SET(&newDnSpecs->Min, 0);
				ASN1_INTEGER_SET(&newDnSpecs->Max, 0);
				if(SKM_sk_push(DN_SPECS, newConf->conf->body->d.ra_conf->d.RA_CONF_PTR->DnSpecs, newDnSpecs) < 0)
				{
					DN_SPECS_free(newDnSpecs);
				}
			}
			CheckRaConf(newConf->conf->body->d.ra_conf);

			ASN1_INTEGER_SET(&newConf->conf->body->d.ra_conf->d.RA_CONF_PTR->default_validity, 365);
			ASN1_INTEGER_SET(&newConf->conf->body->d.ra_conf->d.RA_CONF_PTR->MinKeyLen, 1024);
			ASN1_INTEGER_SET(&newConf->conf->body->d.ra_conf->d.RA_CONF_PTR->MinPwdLen, 6);

			ASN1_BIT_STRING_set_bit(newConf->conf->body->d.ra_conf->d.RA_CONF_PTR->Flags, ALLOW_SERVER_SIDE_RA_KEY_GEN, 1);
			ASN1_BIT_STRING_set_bit(newConf->conf->body->d.ra_conf->d.RA_CONF_PTR->Flags, ALLOW_PKCS10_IMPORT, 1);
			ASN1_BIT_STRING_set_bit(newConf->conf->body->d.ra_conf->d.RA_CONF_PTR->Flags, ALLOW_RA_OP_SC_KEY_GEN, 1);
			ASN1_BIT_STRING_set_bit(newConf->conf->body->d.ra_conf->d.RA_CONF_PTR->Flags, PUSH_KEYS_TO_KEY_ESCROW, 1);
			ASN1_BIT_STRING_set_bit(newConf->conf->body->d.ra_conf->d.RA_CONF_PTR->Flags, REMOVE_UNKNOWN_DN_FIELDS, 1);

			break;
		case ENTITY_TYPE_PUBLICATION:
			newConf->conf->body->d.pub_conf = ENTITY_CONF_BODY_PUB_new();
			if(!newConf->conf->body->d.pub_conf)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
				STORED_ENTITY_CONF_free(newConf);
				return false;
			}
			newConf->conf->body->d.pub_conf->type = PUB_CONF_VERSION;
			newConf->conf->body->d.pub_conf->d.ptr = PUB_CONF_STRUCT_new();
			if(!newConf->conf->body->d.pub_conf->d.ptr)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
				STORED_ENTITY_CONF_free(newConf);
				return false;
			}
			ASN1_INTEGER_SET(&newConf->conf->body->d.pub_conf->d.PUB_CONF_PTR->OcspEnabled, 0);
			break;
		case ENTITY_TYPE_KEY_STORE:
			newConf->conf->body->d.conf = ASN1_OCTET_STRING_new();
			if(!newConf->conf->body->d.conf)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
				STORED_ENTITY_CONF_free(newConf);
				return false;
			}
			break;
		case ENTITY_TYPE_REPOSITORY:
			newConf->conf->body->d.rep_conf = ENTITY_CONF_BODY_REP_new();
			if(!newConf->conf->body->d.rep_conf)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
				STORED_ENTITY_CONF_free(newConf);
				return false;
			}
			newConf->conf->body->d.rep_conf->type = REP_CONF_VERSION;
			newConf->conf->body->d.rep_conf->d.ptr = REP_CONF_STRUCT_new();
			if(!newConf->conf->body->d.rep_conf->d.ptr)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
				STORED_ENTITY_CONF_free(newConf);
				return false;
			}
			ASN1_INTEGER_SET(&newConf->conf->body->d.rep_conf->d.REP_CONF_PTR->Enabled, 0);
			break;
	}
	
	ASN1_INTEGER_SET(&newConf->conf->version, 0);

	
	//Initialize its link struct
	link = ENTITY_LINKS_new();
	if(!link)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		STORED_ENTITY_CONF_free(newConf);
		return false;
	}
	ASN1_STRING_SET(&link->src->Name, EntityName);
	ASN1_INTEGER_SET(&link->src->Type, Type);

	// We add by default all the repositories to the list, it
	// will allow the newly created entity to go seek
	// its conf
	for(i=0; i < SKM_sk_num(REP_ENTRY_INFO, pki_conf->Repositories->list); i++)
	{
		currRep = SKM_sk_value(REP_ENTRY_INFO, pki_conf->Repositories->list, i);
		if(!currRep) continue;

		//Is it me ?
		if(strcmp(EntityName, ASN1_STRING_GET(currRep->Name)) == 0)
			continue;

		currLinkEntry = ENTITY_ENTRY_INFO_new();
		if(!currLinkEntry)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
			STORED_ENTITY_CONF_free(newConf);
			ENTITY_LINKS_free(link);
			return false;
		}

		ASN1_STRING_SET(&currLinkEntry->Name, ASN1_STRING_GET(currRep->Name));
		ASN1_INTEGER_SET(&currLinkEntry->Type, ENTITY_TYPE_REPOSITORY);

		if(SKM_sk_push(ENTITY_ENTRY_INFO, link->dsts, currLinkEntry) < 0)
		{
			ENTITY_ENTRY_INFO_free(currLinkEntry);
		}
	}

	//Are we creating a repository? if yes then 
	//we let all other reps know we exists
	if(Type == ENTITY_TYPE_REPOSITORY)
	{
		for(i=0; i < SKM_sk_num(ENTITY_LINKS, pki_conf->links->links); i++)
		{
			currLink = SKM_sk_value(ENTITY_LINKS, pki_conf->links->links, i);
			if(!currLink) continue;

			//Is it a rep
			if(ASN1_INTEGER_GET(currLink->src->Type) != ENTITY_TYPE_REPOSITORY)
				continue;
			//Is it me ?
			if(strcmp(EntityName, ASN1_STRING_GET(currLink->src->Name)) == 0)
				continue;

			currLinkEntry = ENTITY_ENTRY_INFO_new();
			if(!currLinkEntry)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
				STORED_ENTITY_CONF_free(newConf);
				ENTITY_LINKS_free(link);
				return false;
			}

			ASN1_STRING_SET(&currLinkEntry->Name, EntityName);
			ASN1_INTEGER_SET(&currLinkEntry->Type, ENTITY_TYPE_REPOSITORY);

			if(SKM_sk_push(ENTITY_ENTRY_INFO, currLink->dsts, currLinkEntry) < 0)
			{
				ENTITY_ENTRY_INFO_free(currLinkEntry);
			}
		}
	}

	//Adding the new entity to the list of available one
	ConfIndex = SKM_sk_push(STORED_ENTITY_CONF, pki_conf->confs, newConf);
	if( ConfIndex < 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		STORED_ENTITY_CONF_free(newConf);
		ENTITY_LINKS_free(link);

		// This will delete the added link entries
		if(Type == ENTITY_TYPE_REPOSITORY)
			UpdateEntityConfs(ENTITY_CONF_UPDATE_LINKS);
		return false;
	}

	//add new link entry to list
	LinkIndex = SKM_sk_push(ENTITY_LINKS, pki_conf->links->links, link);
	if( LinkIndex < 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		SKM_sk_delete(STORED_ENTITY_CONF, pki_conf->confs, ConfIndex);
		STORED_ENTITY_CONF_free(newConf);
		ENTITY_LINKS_free(link);

		// This will delete the added link entries
		if(Type == ENTITY_TYPE_REPOSITORY)
			UpdateEntityConfs(ENTITY_CONF_UPDATE_LINKS);
		return false;
	}

	if(!UpdateEntityConfs(ENTITY_CONF_UPDATE_ALL, newConf->conf))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		SKM_sk_delete(STORED_ENTITY_CONF, pki_conf->confs, ConfIndex);
		SKM_sk_delete(ENTITY_LINKS, pki_conf->links->links, LinkIndex);
		STORED_ENTITY_CONF_free(newConf);
		ENTITY_LINKS_free(link);

		// This will delete the added link entries
		if(Type == ENTITY_TYPE_REPOSITORY)
			UpdateEntityConfs(ENTITY_CONF_UPDATE_LINKS);

		return false;
	}

	// We synchronize the links of the modified entities
	if(Type == ENTITY_TYPE_REPOSITORY && !UpdateEntityConfs(ENTITY_CONF_UPDATE_LINKS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		SKM_sk_delete(STORED_ENTITY_CONF, pki_conf->confs, ConfIndex);
		SKM_sk_delete(ENTITY_LINKS, pki_conf->links->links, LinkIndex);
		STORED_ENTITY_CONF_free(newConf);
		ENTITY_LINKS_free(link);
		return false;
	}

	return true;
}


bool Entity_PKI::UserLockResourse(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_MANAGE_ENTITIES))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!ResourceLock(ASN1_STRING_GET(body->d.resource_name)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_PKI::UserUnlockResourse(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_MANAGE_ENTITIES))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	ResourceUnlock(ASN1_STRING_GET(body->d.resource_name));

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_PKI::GetEntityAcl(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	ENTITY_CONF * conf = NULL;
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	ConfAccessLock.LockRead();

	conf = GetEntityConfiguration(body->d.entity_cert);
	if(!conf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN_ENTITY);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, (conf != pki_conf->my_conf?ACL_TYPE_MANAGE_ENTITIES:ACL_TYPE_MODIFY_PKI_STATUS)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}
	
	if(!ResourceOwnerIsMe((char*)(conf != pki_conf->my_conf?body->d.entity_cert->name:PKI_ACL_RESOURCE)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_ENTITY_ACL, X509_ACL_cpy(NULL, conf->acls)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}

bool Entity_PKI::SetEntityAcl(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	ENTITY_CONF * conf = NULL;
	int i, j;	
	ACL_ENTRY * currAcl;
	ACL_ENTRY * userAcl;

	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	ConfAccessLock.LockWrite();

	conf = GetEntityConfiguration(body->d.entity_acl->entity_cert);
	if(!conf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN_ENTITY);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, (conf != pki_conf->my_conf?ACL_TYPE_MANAGE_ENTITIES:ACL_TYPE_MANAGE_ACLS)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ResourceOwnerIsMe((char*)(conf != pki_conf->my_conf?body->d.entity_acl->entity_cert->name:PKI_ACL_RESOURCE)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}
	
	for(i=0; i<SKM_sk_num(ACL_ENTRY, conf->acls->acl_entries); i++)
	{
		currAcl = SKM_sk_value(ACL_ENTRY, conf->acls->acl_entries, i);
		if(!currAcl) continue;

		// Get the associated ACL from the request
		for(j=0; j<SKM_sk_num(ACL_ENTRY, body->d.entity_acl->acl_entries);j++)
		{
			userAcl = SKM_sk_value(ACL_ENTRY, body->d.entity_acl->acl_entries, j);
			if(!userAcl) continue;

			if(ASN1_INTEGER_cmp(userAcl->serial, currAcl->serial) == 0 &&
				ASN1_INTEGER_cmp(userAcl->type, currAcl->type) == 0)
			{
				ASN1_BIT_STRING_cpy(&currAcl->acls, userAcl->acls);
				break;
			}
		}
	}

	//Is it me ?
	if(conf == pki_conf->my_conf)
	{
		if(!AclValidator.SetACL(conf->acls))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ConfAccessLock.Unlock();
			return false;
		}
	}
	else
	{
		//We update the crypted conf
		if(!UpdateEntityConfs(0, conf, true))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ConfAccessLock.Unlock();
			return false;
		}
	}


	if(!WritePersonnalConf())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}

bool Entity_PKI::GetEntityMailConf(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	ENTITY_CONF * conf = NULL;
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	ConfAccessLock.LockRead();

	conf = GetEntityConfiguration(body->d.entity_cert);
	if(!conf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN_ENTITY);
		ConfAccessLock.Unlock();
		return false;
	}
		
	if(!AclValidator.CanUserPerform(UserCert, (conf != pki_conf->my_conf?ACL_TYPE_MANAGE_ENTITIES:ACL_TYPE_MODIFY_PKI_STATUS)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}
	
	if(!ResourceOwnerIsMe((char*)(conf != pki_conf->my_conf?body->d.entity_cert->name:PKI_EMAIL_RESOURCE)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_MAIL_CONF, EMAIL_CONF_cpy(NULL, conf->mail_conf)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}

bool Entity_PKI::SetEntityMailConf(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	ENTITY_CONF * conf = NULL;
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	ConfAccessLock.LockWrite();

	conf = GetEntityConfiguration(body->d.entity_mail_conf->entity_cert);
	if(!conf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN_ENTITY);
		ConfAccessLock.Unlock();
		return false;
	}


	if(!AclValidator.CanUserPerform(UserCert, (conf != pki_conf->my_conf?ACL_TYPE_MANAGE_ENTITIES:ACL_TYPE_MODIFY_PKI_STATUS)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ResourceOwnerIsMe((char*)(conf != pki_conf->my_conf?body->d.entity_mail_conf->entity_cert->name:PKI_EMAIL_RESOURCE)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}
	
	if(!EMAIL_CONF_cpy(&conf->mail_conf, body->d.entity_mail_conf->mail_conf))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	//ASN1_INTEGER_SET(&conf->version, ASN1_INTEGER_GET(conf->version) + 1);

	//Is it me ?
	if(conf == pki_conf->my_conf)
	{
		if(!m_Jobs.SetMailerInfo(m_EntityMail, pki_conf->my_conf->mail_conf))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ConfAccessLock.Unlock();
			return false;
		}
	}
	else
	{
		//We update the crypted conf
		if(!UpdateEntityConfs(0, conf, true))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ConfAccessLock.Unlock();
			return false;
		}
	}


	if(!WritePersonnalConf())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}

bool Entity_PKI::STORED_ENTITY_CONF_to_name_x509(const STORED_ENTITY_CONF * conf, ASN1_UTF8STRING ** name, X509 ** cert)
{
	ASN1_STRING_SET(name, ASN1_STRING_GET(conf->conf->Name));

	if(!X509_cpy(cert, conf->certificate))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	return true;
}


bool Entity_PKI::SynchronizeLinks(ENTITY_CONF ** currConf)
{
	int i;
	int j;
	ENTITY_LINKS * currLink;
	ENTITY_ENTRY_INFO * currLinkEntry;
	REP_ENTRY_INFO * currRep;
	CA_ENTRY_INFO * currCa;
	RA_ENTRY_INFO * currRa;
	KEY_ENTRY_INFO * currKey;
	PUB_ENTRY_INFO * currPub;
	STORED_ENTITY_CONF * currSConf;
	ENTITY_CONF * savedConf = NULL;
	bool Found;
	PUBLICATION_ENTRY * currPubMeth;


	//We save it, not to affect the conf
	//if an error occurs
	if(!ENTITY_CONF_cpy(&savedConf, *currConf))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}


	//Seek for corresponding entity in the links conf
	currLink = NULL;
	for(i=0; i < SKM_sk_num(ENTITY_LINKS, pki_conf->links->links); i++)
	{
		currLink = SKM_sk_value(ENTITY_LINKS, pki_conf->links->links, i);
		if(!currLink) continue;

		if(ASN1_STRING_cmp(currLink->src->Name, savedConf->Name) == 0)
		{
			break;
		}
		currLink = NULL;
	}
	if(!currLink)
	{
		ENTITY_CONF_free(savedConf);
		return true;
	}

	//We first empty the REP, CA, RA, Publication and Key Store list
	while( (currRep = SKM_sk_pop(REP_ENTRY_INFO, savedConf->Repositories->list)) )
	{
		REP_ENTRY_INFO_free(currRep);
	}
	while( (currCa = SKM_sk_pop(CA_ENTRY_INFO, savedConf->Cas->list)) )
	{
		CA_ENTRY_INFO_free(currCa);
	}
	while( (currRa = SKM_sk_pop(RA_ENTRY_INFO, savedConf->Ras->list)) )
	{
		RA_ENTRY_INFO_free(currRa);
	}
	while( (currKey = SKM_sk_pop(KEY_ENTRY_INFO, savedConf->KeyStores->list)) )
	{
		KEY_ENTRY_INFO_free(currKey);
	}
	while( (currPub = SKM_sk_pop(PUB_ENTRY_INFO, savedConf->Publications->list)) )
	{
		PUB_ENTRY_INFO_free(currPub);
	}

	// Let's do the synchro !
	for(i=0; i < SKM_sk_num(ENTITY_ENTRY_INFO, currLink->dsts); i++)
	{
		currLinkEntry = SKM_sk_value(ENTITY_ENTRY_INFO, currLink->dsts, i);
		if(!currLinkEntry) continue;

		Found = false;
		//Update repositories for all kind of entities

		switch(ASN1_INTEGER_GET(currLinkEntry->Type))
		{
			case ENTITY_TYPE_REPOSITORY:
				//Seek corresponding respositories info
				for(j=0; j < SKM_sk_num(REP_ENTRY_INFO, pki_conf->Repositories->list); j++)
				{
					currRep = SKM_sk_value(REP_ENTRY_INFO, pki_conf->Repositories->list, j);
					if(!currRep)
						continue;
					//Is the repository we're looking for ?
					if(ASN1_STRING_cmp(currRep->Name, currLinkEntry->Name) != 0) continue;

					//Yes, add it to the list for this entity
					currRep = REP_ENTRY_INFO_cpy(NULL, currRep);
					if(!currRep)
					{
						ENTITY_CONF_free(savedConf);
						NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
						return false;
					}

					// Copy the flags from the link
					if(!ASN1_BIT_STRING_cpy(&currRep->Flags, currLinkEntry->flags))
					{
						ENTITY_CONF_free(savedConf);
						REP_ENTRY_INFO_free(currRep);
						NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
						return false;
					}

					if(SKM_sk_push(REP_ENTRY_INFO, savedConf->Repositories->list, currRep) < 0)
					{
						REP_ENTRY_INFO_free(currRep);
						ENTITY_CONF_free(savedConf);
						NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
						return false;
					}
					Found = true;
					break;
				}
				break;
				
			case  ENTITY_TYPE_RA:
				//Seek corresponding RA info
				for(j=0; j < SKM_sk_num(STORED_ENTITY_CONF, pki_conf->confs); j++)
				{
					currSConf = SKM_sk_value(STORED_ENTITY_CONF, pki_conf->confs, j);
					if(!currSConf) continue;
					if(currSConf->conf->body->type != ENTITY_TYPE_RA) continue;

					//Is it the seeked one ?
					if(ASN1_STRING_cmp(currSConf->conf->Name, currLinkEntry->Name) != 0) continue;

					currRa = RA_ENTRY_INFO_new();
					if(!currRa)
					{
						NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
						ENTITY_CONF_free(savedConf);
						return false;
					}
					if(!STORED_ENTITY_CONF_to_name_x509(currSConf, &currRa->Name, &currRa->RaSSL))
					{
						RA_ENTRY_INFO_free(currRa);
						ENTITY_CONF_free(savedConf);
						NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
						return false;
					}
					if(SKM_sk_push(RA_ENTRY_INFO, savedConf->Ras->list, currRa) < 0)
					{
						RA_ENTRY_INFO_free(currRa);
						ENTITY_CONF_free(savedConf);
						NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
						return false;
					}
					Found=true;
					break;
				}
				break;

			case  ENTITY_TYPE_CA:
				//Seek corresponding CA info
				for(j=0; j < SKM_sk_num(STORED_ENTITY_CONF, pki_conf->confs); j++)
				{
					currSConf = SKM_sk_value(STORED_ENTITY_CONF, pki_conf->confs, j);
					if(!currSConf) continue;
					if(currSConf->conf->body->type != ENTITY_TYPE_CA) continue;

					//Is it the seeked one ?
					if(ASN1_STRING_cmp(currSConf->conf->Name, currLinkEntry->Name) != 0) continue;

					currCa = CA_ENTRY_INFO_new();
					if(!currCa)
					{
						NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
						ENTITY_CONF_free(savedConf);
						return false;
					}
					if(!STORED_ENTITY_CONF_to_name_x509(currSConf, &currCa->Name, &currCa->CaSSL))
					{
						CA_ENTRY_INFO_free(currCa);
						ENTITY_CONF_free(savedConf);
						NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
						return false;
					}
					if(SKM_sk_push(CA_ENTRY_INFO, savedConf->Cas->list, currCa) < 0)
					{
						CA_ENTRY_INFO_free(currCa);
						ENTITY_CONF_free(savedConf);
						NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
						return false;
					}
					Found=true;
					break;
				}
				break;

			case  ENTITY_TYPE_KEY_STORE:
				//Seek corresponding Key Store info
				for(j=0; j < SKM_sk_num(STORED_ENTITY_CONF, pki_conf->confs); j++)
				{
					currSConf = SKM_sk_value(STORED_ENTITY_CONF, pki_conf->confs, j);
					if(!currSConf) continue;
					if(currSConf->conf->body->type != ENTITY_TYPE_KEY_STORE) continue;

					//Is it the seeked one ?
					if(ASN1_STRING_cmp(currSConf->conf->Name, currLinkEntry->Name) != 0) continue;

					currKey = KEY_ENTRY_INFO_new();
					if(!currKey)
					{
						NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
						ENTITY_CONF_free(savedConf);
						return false;
					}
					if(!STORED_ENTITY_CONF_to_name_x509(currSConf, &currKey->Name, &currKey->KeySSL))
					{
						KEY_ENTRY_INFO_free(currKey);
						ENTITY_CONF_free(savedConf);
						NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
						return false;
					}
					if(SKM_sk_push(KEY_ENTRY_INFO, savedConf->KeyStores->list, currKey) < 0)
					{
						KEY_ENTRY_INFO_free(currKey);
						ENTITY_CONF_free(savedConf);
						NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
						return false;
					}
					Found=true;
					break;
				}
				break;

			case  ENTITY_TYPE_PUBLICATION:
				//Seek corresponding Pub Store info
				for(j=0; j < SKM_sk_num(PUB_ENTRY_INFO, pki_conf->Publications->list); j++)
				{
					currPub = SKM_sk_value(PUB_ENTRY_INFO, pki_conf->Publications->list, j);
					if(!currPub) continue;

					//Is the publication we're looking for ?
					if(ASN1_STRING_cmp(currPub->Name, currLinkEntry->Name) != 0) continue;

					//Yes, add it to the list for this entity
					currPub = PUB_ENTRY_INFO_cpy(NULL, currPub);
					if(!currPub)
					{
						ENTITY_CONF_free(savedConf);
						NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
						return false;
					}

					if(SKM_sk_push(PUB_ENTRY_INFO, savedConf->Publications->list, currPub) < 0)
					{
						PUB_ENTRY_INFO_free(currPub);
						ENTITY_CONF_free(savedConf);
						NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
						return false;
					}
					Found = true;
					break;
				}
				break;

			default:
				ENTITY_CONF_free(savedConf);
				NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
				return false;
		}

		// If we didn't find the corresponding entry
		// we erase the link
		if(!Found)
		{
			ENTITY_ENTRY_INFO_free(currLinkEntry);
			SKM_sk_delete(ENTITY_ENTRY_INFO, currLink->dsts, i);
			i--;
		}
	}

	//We copy the working conf to the real conf
	ENTITY_CONF_free(*currConf);
	*currConf = savedConf;

	// We need to synchronize the Publication method 
	// when working on a Publication entity

	if(savedConf->body->type == ENTITY_TYPE_PUBLICATION)
	{
		// First delete the CA that have been 
		// removed from our link
		for(i=0; i<SKM_sk_num(PUBLICATION_ENTRY, savedConf->body->d.pub_conf->d.PUB_CONF_PTR->Publications); i++)
		{
			currPubMeth = SKM_sk_value(PUBLICATION_ENTRY, savedConf->body->d.pub_conf->d.PUB_CONF_PTR->Publications, i);
			if(!currPubMeth) continue;

			for(j=0; j<SKM_sk_num(CA_ENTRY_INFO, savedConf->Cas->list); j++)
			{
				currCa = SKM_sk_value(CA_ENTRY_INFO, savedConf->Cas->list, j);
				if(!currCa) continue;

				if(ASN1_STRING_cmp(currPubMeth->CaName, currCa->Name) == 0)
					break;
			}
			// Not found, has been removed
			if(j == SKM_sk_num(CA_ENTRY_INFO, savedConf->Cas->list))
			{
				SKM_sk_delete(PUBLICATION_ENTRY, savedConf->body->d.pub_conf->d.PUB_CONF_PTR->Publications, i);
				PUBLICATION_ENTRY_free(currPubMeth);
			}
		}

		// Now add the new CAs that have been linked to me
		for(i=0; i<SKM_sk_num(CA_ENTRY_INFO, savedConf->Cas->list); i++)
		{
			currCa = SKM_sk_value(CA_ENTRY_INFO, savedConf->Cas->list, i);
			if(!currCa) continue;

			for(j=0; j<SKM_sk_num(PUBLICATION_ENTRY, savedConf->body->d.pub_conf->d.PUB_CONF_PTR->Publications); j++)
			{
				currPubMeth = SKM_sk_value(PUBLICATION_ENTRY, savedConf->body->d.pub_conf->d.PUB_CONF_PTR->Publications, j);
				if(!currPubMeth) continue;


				if(ASN1_STRING_cmp(currPubMeth->CaName, currCa->Name) == 0)
					break;
			}
			// Not found, need to be added
			if(j == SKM_sk_num(PUBLICATION_ENTRY, savedConf->body->d.pub_conf->d.PUB_CONF_PTR->Publications))
			{
				currPubMeth = PUBLICATION_ENTRY_new();
				if(currPubMeth)
				{
					ASN1_STRING_SET(&currPubMeth->CaName, ASN1_STRING_GET(currCa->Name));
					if( SKM_sk_push(PUBLICATION_ENTRY, savedConf->body->d.pub_conf->d.PUB_CONF_PTR->Publications, currPubMeth) <0 )
					{
						PUBLICATION_ENTRY_free(currPubMeth);
					}
				}
			}
		}
	}

	return true;
}


bool Entity_PKI::UpdateEntityConfs(int What, ENTITY_CONF * OneConf, bool ForceCryptGen)
{
	int i;
	int j;
	int k;
	CA_Handler * usersCa;
	INTERNAL_CA_CERT * currCert;
	USERS_GROUP * currGroup;
	ENTITY_CONF_CRYPTED * newConf;
	STORED_ENTITY_CONF * currConf;
	EVP_PKEY * repPubKey;
	bool ConfChanged;
	ASN1_INTEGER * currSerial;
	STACK_OF(INTERNAL_CA_CERT) * certs;

	ConfAccessLock.LockWrite();

	if(!pki_conf->crypt_confs->confs->all_confs)
		pki_conf->crypt_confs->confs->all_confs = sk_new_null();
	if(!pki_conf->crypt_confs->confs->all_confs)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!OneConf || OneConf == pki_conf->my_conf)
	{
		// We update for the local pki conf 
		// with what was requested

		//Shall we synchronize the groups ?
		if(What & ENTITY_CONF_UPDATE_ALL || What & ENTITY_CONF_UPDATE_GROUPS)
		{
			//We remove any revoke cert
			usersCa = GetInternalCA(INTERNAL_CA_TYPE_USER);
			if(!usersCa)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				ConfAccessLock.Unlock();
				return false;
			}

			certs = usersCa->get_Certs();
			if(!certs)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				ConfAccessLock.Unlock();
				return false;
			}

			for(i=0; i<SKM_sk_num(INTERNAL_CA_CERT, certs); i++)
			{
				currCert = SKM_sk_value(INTERNAL_CA_CERT, certs, i);
				if(!currCert) continue;

				//We ignore the Admin cert
				if(X509_ACL_Validator::Static_IsPkiAdministrator(currCert->cert, pki_conf->my_conf->acls->AdminSerials))
					continue;

				//We're only interessed by the revoked ones
				if(ASN1_INTEGER_GET(currCert->state) != CERT_STATE_REVOKED)
					continue;
				

				// We remove the cert form all the groups we can find it in
				for(j=0; j<SKM_sk_num(USERS_GROUP, pki_conf->my_conf->groups->list); j++)
				{
					currGroup = SKM_sk_value(USERS_GROUP, pki_conf->my_conf->groups->list, j);
					if(!currGroup) continue;
					
					for(k=0; k<SKM_sk_num(ASN1_INTEGER, currGroup->users_serial); k++)
					{
						currSerial = SKM_sk_value(ASN1_INTEGER, currGroup->users_serial, k);
						if(!currSerial) continue;

						if(ASN1_INTEGER_cmp(currSerial, X509_get_serialNumber(currCert->cert)) == 0)
						{
							SKM_sk_delete(ASN1_INTEGER, currGroup->users_serial, k);
							ASN1_INTEGER_free(currSerial);
							k--;
						}
					}
				}
			}
			SKM_sk_pop_free(INTERNAL_CA_CERT, certs, INTERNAL_CA_CERT_free);
		}


		//Shall we update the ACL ?
		if(What & ENTITY_CONF_UPDATE_ALL || What & ENTITY_CONF_UPDATE_ACL || What & ENTITY_CONF_UPDATE_GROUPS)
		{
			switch(X509_ACL_init(pki_conf->my_conf->acls, ENTITY_TYPE_PKI))
			{
				case X509_ACL_INIT_ERROR:
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					ConfAccessLock.Unlock();
					return false;
					break;
				case X509_ACL_INIT_MODIFIED:
					if(!AclValidator.SetACL(pki_conf->my_conf->acls))
					{
						NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
						ConfAccessLock.Unlock();
						return false;
					}
					break;
				default:
					break;
			}
		}
		//Shall we update the links
		if(What & ENTITY_CONF_UPDATE_ALL || What & ENTITY_CONF_UPDATE_LINKS)
		{
			if(!SynchronizeLinks(&pki_conf->my_conf))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				ConfAccessLock.Unlock();
				return false;			
			}
			
			if(!m_Jobs.SetRepositories(pki_conf->my_conf->Repositories))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				ConfAccessLock.Unlock();
				return false;
			}
		}
		
		// Update admins mails
		if(What & ENTITY_CONF_UPDATE_ALL || What & ENTITY_CONF_UPDATE_ADMINS)
		{
			if(!m_Jobs.SetAdminMails(pki_conf->my_conf->acls->AdminSerials))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				ConfAccessLock.Unlock();
				return false;
			}
		}
		
		// Update CRLs
		if(What & ENTITY_CONF_UPDATE_ALL || What & ENTITY_CONF_UPDATE_CRL)
		{
			if(!AclValidator.SetCRL(pki_conf->my_conf->crls))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				ConfAccessLock.Unlock();
				return false;
			}
		}
	}



	//For each conf we update what was requested
	for(i=0; i<SKM_sk_num(STORED_ENTITY_CONF, pki_conf->confs); i++)
	{
		currConf = SKM_sk_value(STORED_ENTITY_CONF, pki_conf->confs, i);
		if(!currConf) continue;

		if(OneConf && OneConf != currConf->conf) continue;
		
		ConfChanged = ForceCryptGen;

		//Shall we update the CRL
		if(What & ENTITY_CONF_UPDATE_ALL || What & ENTITY_CONF_UPDATE_CRL)
		{
			if(!INTERNAL_PKI_CRL_cpy(&currConf->conf->crls, pki_conf->my_conf->crls))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				ConfAccessLock.Unlock();
				return false;			
			}
			ConfChanged = true;
		}

		//Shall we update the PKI admins list ?
		if(What & ENTITY_CONF_UPDATE_ALL || What & ENTITY_CONF_UPDATE_ADMINS)
		{
			if(!PKI_ADMIN_SERIALS_cpy(&currConf->conf->acls->AdminSerials, pki_conf->my_conf->acls->AdminSerials))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				ConfAccessLock.Unlock();
				return false;			
			}
			ConfChanged = true;
		}

		//Shall we update the groups ?
		if(What & ENTITY_CONF_UPDATE_ALL || What & ENTITY_CONF_UPDATE_GROUPS)
		{
			if(!USERS_GROUPS_cpy(&currConf->conf->groups, pki_conf->my_conf->groups))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				ConfAccessLock.Unlock();
				return false;			
			}
			ConfChanged = true;
		}


		//Shall we update the ACL
		if(What & ENTITY_CONF_UPDATE_ALL || What & ENTITY_CONF_UPDATE_ACL || What & ENTITY_CONF_UPDATE_GROUPS)
		{
			switch(X509_ACL_init(currConf->conf->acls, currConf->conf->body->type))
			{
				case X509_ACL_INIT_MODIFIED:
					ConfChanged = true;
					break;

				case X509_ACL_INIT_ERROR:
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					ConfAccessLock.Unlock();
					return false;
					break;
				
				default:
					break;
			}
		}

		//Shall we update the Repositories list or the links
		if(What & ENTITY_CONF_UPDATE_ALL || What & ENTITY_CONF_UPDATE_LINKS)
		{
			if(!SynchronizeLinks(&currConf->conf))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				ConfAccessLock.Unlock();
				return false;			
			}
			ConfChanged = true;
		}
			
		if(ConfChanged)
		{
			ASN1_INTEGER_SET(&currConf->conf->version, ASN1_INTEGER_GET(currConf->conf->version) + 1);

			newConf = NULL;
			//Seek crypted conf to update it id we don't find it we add it
			for(j=0; j<SKM_sk_num(ENTITY_CONF_CRYPTED, pki_conf->crypt_confs->confs->all_confs); j++)
			{
				newConf = SKM_sk_value(ENTITY_CONF_CRYPTED, pki_conf->crypt_confs->confs->all_confs, j);
				if(!newConf)
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
					ConfAccessLock.Unlock();
					return false;
				}

				//Is it the seeked one ?
				if(currConf->certificate->cert_info->key->public_key->length ==
					newConf->recipient->public_key->length &&
					memcmp(currConf->certificate->cert_info->key->public_key->data,
					newConf->recipient->public_key->data, 
					newConf->recipient->public_key->length) == 0)
					break;
			}
			//Did we find it ?
			if(j == SKM_sk_num(ENTITY_CONF_CRYPTED, pki_conf->crypt_confs->confs->all_confs))
			{
				//No
				newConf = ENTITY_CONF_CRYPTED_new();
				if(!newConf)
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
					ConfAccessLock.Unlock();
					return false;
				}
				//Set public key
				if(!X509_PUBKEY_cpy(&newConf->recipient, currConf->certificate->cert_info->key))
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					ConfAccessLock.Unlock();
					return false;
				}
			}

			//Sign and crypt conf
			repPubKey = X509_get_pubkey(currConf->certificate);
			if(!repPubKey)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				ConfAccessLock.Unlock();
				return false;
			}
			//Encrypt the body with the entity public key and generate signature with our private key
			if( ENTITY_CONF_sign_encrypt(newConf->crypted, currConf->conf, m_EntityKey.GetRsaKey(), repPubKey, EVP_sha1(), EVP_des_ede3_cbc()) <= 0) 
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				EVP_PKEY_free(repPubKey);
				ConfAccessLock.Unlock();
				return false;
			}
			EVP_PKEY_free(repPubKey);

			//Do we need to insert it ?
			if(j == SKM_sk_num(ENTITY_CONF_CRYPTED, pki_conf->crypt_confs->confs->all_confs))
			{
				if(SKM_sk_push(ENTITY_CONF_CRYPTED, pki_conf->crypt_confs->confs->all_confs, newConf) < 0)
				{
					ENTITY_CONF_CRYPTED_free(newConf);
				}
			}
		}
	}

	ConfAccessLock.Unlock();
	return true;
}



bool Entity_PKI::GetEntityAudits(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	ENTITY_CONF * conf;
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	ConfAccessLock.LockRead();

	conf = GetEntityConfiguration(body->d.entity_cert);
	if(!conf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN_ENTITY);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, (conf != pki_conf->my_conf?ACL_TYPE_MANAGE_ENTITIES:ACL_TYPE_MODIFY_PKI_STATUS)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}
	
	if(!ResourceOwnerIsMe((char*)(conf != pki_conf->my_conf?body->d.entity_cert->name:PKI_AUDITS_RESOURCE)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}
	
	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_AUDITS, ENTITY_AUDIT_cpy(NULL, conf->audits)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}

bool Entity_PKI::SetEntityAudits(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	ENTITY_CONF * conf = NULL;
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	ConfAccessLock.LockWrite();

	conf = GetEntityConfiguration(body->d.entity_audits->entity_cert);
	if(!conf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN_ENTITY);
		ConfAccessLock.Unlock();
		return false;
	}


	if(!AclValidator.CanUserPerform(UserCert, (conf != pki_conf->my_conf?ACL_TYPE_MANAGE_ENTITIES:ACL_TYPE_MODIFY_PKI_STATUS)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ResourceOwnerIsMe((char*)(conf != pki_conf->my_conf?body->d.entity_audits->entity_cert->name:PKI_AUDITS_RESOURCE)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}

	
	if(!ENTITY_AUDIT_cpy(&conf->audits, body->d.entity_audits->audits))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	//ASN1_INTEGER_SET(&conf->version, ASN1_INTEGER_GET(conf->version) + 1);

	if(conf == pki_conf->my_conf)
	{
		if(!m_Logging->SetAudit(pki_conf->my_conf->audits))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ConfAccessLock.Unlock();
			return false;
		}
	}
	else
	{
		//We update the crypted conf
		if(!UpdateEntityConfs(0, conf, true))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ConfAccessLock.Unlock();
			return false;
		}
	}


	if(!WritePersonnalConf())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}

bool Entity_PKI::GetEntityLogsType(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_VIEW_LOGS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	size_t i;
	ASN1_INTEGER * type;
	STACK_OF(ASN1_INTEGER) * logstype;
	vector<LOG_MESSAGE_TYPE> cLogsType;


	switch(ASN1_INTEGER_GET(body->d.entity_type))
	{
		case ENTITY_TYPE_PKI:
			cLogsType = LogsType;
			break;
			
		case ENTITY_TYPE_REPOSITORY:
			Entity_REPOSITORY::LogsTypeGet(cLogsType);
			break;
			
		case ENTITY_TYPE_RA:
			Entity_RA::LogsTypeGet(cLogsType);
			break;
			
		case ENTITY_TYPE_PUBLICATION:
			Entity_PUBLICATION::LogsTypeGet(cLogsType);
			break;
			
		case ENTITY_TYPE_KEY_STORE:
			Entity_KEYSTORE::LogsTypeGet(cLogsType);
			break;
			
		case ENTITY_TYPE_CA:
			Entity_CA::LogsTypeGet(cLogsType);
			break;

		default:
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
			return false;
	}


	
	logstype = sk_ASN1_INTEGER_new_null();
	if(!logstype)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	for(i=0; i<cLogsType.size(); i++)
	{
		type = NULL;
		ASN1_INTEGER_SET(&type, cLogsType[i]);
		if(type)
		{
			sk_ASN1_INTEGER_push(logstype, type);
		}
	}
	
	
	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_LOGS_TYPE, logstype))
	{
		sk_ASN1_INTEGER_pop_free(logstype, ASN1_INTEGER_free);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}



ENTITY_CONF * Entity_PKI::GetEntityConfiguration(X509 *EntityCert)
{
	STORED_ENTITY_CONF * conf = NULL;
	if(X509_cmp(EntityCert, pki_conf->cas->PkiCert) == 0)
	{
		return pki_conf->my_conf;
	}

	conf = STORED_ENTITY_CONF_get(pki_conf->confs, EntityCert);
	if(!conf)
	{
		return NULL;
	}
	return conf->conf;
}


bool Entity_PKI::GetEntitiesLinks(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_MODIFY_PKI_STATUS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	if(!ResourceOwnerIsMe(PKI_ENTITIES_LINKS_RESOURCE))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	ConfAccessLock.LockRead();

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_ENTITIES_LINKS, ENTITIES_LINKS_cpy(NULL, pki_conf->links)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}

ENTITY_LINKS * Entity_PKI::get_EntityLink(const ENTITIES_LINKS * Links, const char * Name)
{
	int i;
	ENTITY_LINKS * currLink;

	for(i=0; i<SKM_sk_num(ENTITY_LINKS, Links->links); i++)
	{
		currLink = SKM_sk_value(ENTITY_LINKS, Links->links, i);
		if(!currLink) continue;

		if(strcmp(Name, ASN1_STRING_GET(currLink->src->Name)) == 0)
		{
			return currLink;
		}
	}

	return NULL;
}

ENTITY_ENTRY_INFO * Entity_PKI::get_LinkToEntity(const ENTITY_LINKS * EntityLink, const char * Name)
{
	int i;
	ENTITY_ENTRY_INFO * currLinkEntry;

	for(i=0; i<SKM_sk_num(ENTITY_ENTRY_INFO, EntityLink->dsts); i++)
	{
		currLinkEntry = SKM_sk_value(ENTITY_ENTRY_INFO, EntityLink->dsts, i);
		if(!currLinkEntry) continue;

		if(strcmp(Name, ASN1_STRING_GET(currLinkEntry->Name)) == 0)
		{
			return currLinkEntry;
		}
	}

	return NULL;
}


bool Entity_PKI::SetEntitiesLinks(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_MODIFY_PKI_STATUS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!ResourceOwnerIsMe(PKI_ENTITIES_LINKS_RESOURCE))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	//Let's do some integrity check !
	int i;
	int j;
	int k;
	int FoundCtr;

	STORED_ENTITY_CONF * currConf;
	ENTITY_LINKS * currLink;
	ENTITY_ENTRY_INFO * currLinkEntry;
	ENTITY_LINKS * currLink2;
	ENTITY_ENTRY_INFO * currLinkEntry2;

	ConfAccessLock.LockWrite();

	//Check that all conf are there
	FoundCtr = 0;
	for(i=0; i<SKM_sk_num(STORED_ENTITY_CONF, pki_conf->confs); i++)
	{
		currConf = SKM_sk_value(STORED_ENTITY_CONF, pki_conf->confs, i);
		if(!currConf) continue;

		//Is the conf present in the links ?
		if( (currLink = get_EntityLink(body->d.entities_links,  ASN1_STRING_GET(currConf->conf->Name))) &&
			ASN1_INTEGER_GET(currLink->src->Type) == currConf->conf->body->type)
		{
			FoundCtr++;
		}
	}
	if(FoundCtr != (SKM_sk_num(STORED_ENTITY_CONF, pki_conf->confs)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		ConfAccessLock.Unlock();
		return false;
	}

	//Is the PKI present in the links ?
	if( !(currLink = get_EntityLink(body->d.entities_links,  ASN1_STRING_GET(pki_conf->my_conf->Name))) ||
		ASN1_INTEGER_GET(currLink->src->Type) != ENTITY_TYPE_PKI)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		ConfAccessLock.Unlock();
		return false;
	}

	// Check that the user didn't add some unknown entities
	// And that it links to some known entities
	for(i=0; i<SKM_sk_num(ENTITY_LINKS, body->d.entities_links->links); i++)
	{
		currLink = SKM_sk_value(ENTITY_LINKS, body->d.entities_links->links, i);
		if(!currLink) continue;

		if(ASN1_STRING_cmp(pki_conf->my_conf->Name, currLink->src->Name) != 0 || ASN1_INTEGER_GET(currLink->src->Type) != pki_conf->my_conf->body->type)
		{
			//Check corresponding entity
			for(j=0; j<SKM_sk_num(STORED_ENTITY_CONF, pki_conf->confs); j++)
			{
				currConf = SKM_sk_value(STORED_ENTITY_CONF, pki_conf->confs, j);
				if(!currConf) continue;

				if(ASN1_STRING_cmp(currConf->conf->Name, currLink->src->Name) == 0 && ASN1_INTEGER_GET(currLink->src->Type) == currConf->conf->body->type)
				{
					break;
				}
			}
			// The conf of the entity doesn't exist, the user added
			// a link to a unknown entity
			if(j == SKM_sk_num(STORED_ENTITY_CONF, pki_conf->confs))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
				ConfAccessLock.Unlock();
				return false;
			}
		}

		//Search for the links
		FoundCtr = 0;
		for(k=0; k<SKM_sk_num(ENTITY_ENTRY_INFO, currLink->dsts); k++)
		{
			currLinkEntry = SKM_sk_value(ENTITY_ENTRY_INFO, currLink->dsts, k);
			if(!currLinkEntry) continue;

			if(ASN1_STRING_cmp(pki_conf->my_conf->Name, currLinkEntry->Name) != 0 || ASN1_INTEGER_GET(currLinkEntry->Type) != pki_conf->my_conf->body->type)
			{
				for(j=0; j<SKM_sk_num(STORED_ENTITY_CONF, pki_conf->confs); j++)
				{
					currConf = SKM_sk_value(STORED_ENTITY_CONF, pki_conf->confs, j);
					if(!currConf) continue;

					if(ASN1_STRING_cmp(currConf->conf->Name, currLinkEntry->Name) == 0 && ASN1_INTEGER_GET(currLinkEntry->Type) == currConf->conf->body->type)
					{
						break;
					}
				}
				// The conf of the entity doesn't exist, the user added
				// a link to a unknown entity
				if(j == SKM_sk_num(STORED_ENTITY_CONF, pki_conf->confs))
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
					ConfAccessLock.Unlock();
					return false;
				}
			}
			// Is a repository ?
			if(ASN1_INTEGER_GET(currLinkEntry->Type) == ENTITY_TYPE_REPOSITORY)
			{
				FoundCtr++;
			}
		}
		// If we're dealing with a CA or a RA it must be connected to
		// at least one repositoy
		if( (ASN1_INTEGER_GET(currLink->src->Type) == ENTITY_TYPE_CA || 
			ASN1_INTEGER_GET(currLink->src->Type) == ENTITY_TYPE_RA ||
			ASN1_INTEGER_GET(currLink->src->Type) == ENTITY_TYPE_KEY_STORE ||
			ASN1_INTEGER_GET(currLink->src->Type) == ENTITY_TYPE_PUBLICATION) && 
			!FoundCtr)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_RA_CA_LINK_REP);
			ERR_set_error_data(strdup(ASN1_STRING_GET(currLink->src->Name)), ERR_TXT_MALLOCED | ERR_TXT_STRING);
			ConfAccessLock.Unlock();
			return false;
		}
	}

	// Check the flags of the repositories.
	for(i=0; i<SKM_sk_num(ENTITY_LINKS, body->d.entities_links->links); i++)
	{
		currLink = SKM_sk_value(ENTITY_LINKS, body->d.entities_links->links, i);
		if(!currLink) continue;
		if(ASN1_INTEGER_GET(currLink->src->Type) != ENTITY_TYPE_REPOSITORY)
			continue;

		for(j=0; j<SKM_sk_num(ENTITY_ENTRY_INFO, currLink->dsts); j++)
		{
			currLinkEntry = SKM_sk_value(ENTITY_ENTRY_INFO, currLink->dsts, j);
			if(!currLinkEntry) continue;

			if(ASN1_INTEGER_GET(currLinkEntry->Type) != ENTITY_TYPE_REPOSITORY)
			{
				ASN1_BIT_STRING_set_bit(currLinkEntry->flags, REP_ENTRY_INFO_FLAG_FIREWALLED, 0);
				ASN1_BIT_STRING_set_bit(currLinkEntry->flags, REP_ENTRY_INFO_FLAG_SHOULD_FULL_SYNCHRO, 0);
				continue;
			}
			else
			{
				// Get the full entry for the entity
				currLink2 = get_EntityLink(body->d.entities_links, ASN1_STRING_GET(currLinkEntry->Name));
				if(!currLink2)
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
					ConfAccessLock.Unlock();
					return false;
				}

				// Get its link to me
				currLinkEntry2 = get_LinkToEntity(currLink2, ASN1_STRING_GET(currLink->src->Name));
				if(!currLinkEntry2)
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
					ConfAccessLock.Unlock();
					return false;
				}

				// Now verify/reset the flags
				if(ASN1_BIT_STRING_get_bit(currLinkEntry->flags, REP_ENTRY_INFO_FLAG_SHOULD_FULL_SYNCHRO))
				{
					// If I should full shynchro, then it means I have to be firewalled for it
					ASN1_BIT_STRING_set_bit(currLinkEntry2->flags, REP_ENTRY_INFO_FLAG_FIREWALLED, 1);
					ASN1_BIT_STRING_set_bit(currLinkEntry2->flags, REP_ENTRY_INFO_FLAG_SHOULD_FULL_SYNCHRO, 0);

					ASN1_BIT_STRING_set_bit(currLinkEntry->flags, REP_ENTRY_INFO_FLAG_FIREWALLED, 0);
				}
				else if(ASN1_BIT_STRING_get_bit(currLinkEntry->flags, REP_ENTRY_INFO_FLAG_FIREWALLED))
				{
					// If its firewalled to me then it should full synchro for me
					ASN1_BIT_STRING_set_bit(currLinkEntry2->flags, REP_ENTRY_INFO_FLAG_FIREWALLED, 0);
					ASN1_BIT_STRING_set_bit(currLinkEntry2->flags, REP_ENTRY_INFO_FLAG_SHOULD_FULL_SYNCHRO, 1);
				}
				else
				{
					// both repositories can access each other
					ASN1_BIT_STRING_set_bit(currLinkEntry2->flags, REP_ENTRY_INFO_FLAG_FIREWALLED, 0);
					ASN1_BIT_STRING_set_bit(currLinkEntry2->flags, REP_ENTRY_INFO_FLAG_SHOULD_FULL_SYNCHRO, 0);
				}
			}
		}
	}

	
	if(!ENTITIES_LINKS_cpy(&pki_conf->links, body->d.entities_links))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	//Yell to other entities the new repositories list
	if(!UpdateEntityConfs(ENTITY_CONF_UPDATE_LINKS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!WritePersonnalConf())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}

void Entity_PKI::SpawnStop()
{
	m_Jobs.SpawnStop();
}

bool Entity_PKI::ParseNewConf(const ENTITY_CONF_CRYPTED * newConf, const INTERNAL_PKI_CA * cas)
{
	return true;
}

bool Entity_PKI::DeclarePkiAdminCert(const PKI_CERT & cert, const char * Email)
{
	PKI_ADMIN_ENTRY * entry;
	
	entry = PKI_ADMIN_ENTRY_new();
	if(!entry)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	
	ASN1_INTEGER_SET(&entry->serial, cert.GetSerial());
	ASN1_STRING_SET(&entry->email, Email);

	if(SKM_sk_push(ASN1_INTEGER, pki_conf->my_conf->acls->AdminSerials->admins, entry) < 0)
	{
		PKI_ADMIN_ENTRY_free(entry);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_PKI::GetGroups(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_MANAGE_USERS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	if(!ResourceOwnerIsMe(PKI_ENTITIES_GROUPS_RESOURCE))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	ConfAccessLock.LockRead();

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_GROUPS, USERS_GROUPS_cpy(NULL, pki_conf->my_conf->groups)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}

bool Entity_PKI::SetGroups(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	int i, j;
	USERS_GROUP * currGroup;
	USERS_GROUP * userGroup;
	ASN1_INTEGER * currSerial;

	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	ConfAccessLock.LockWrite();

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_MANAGE_USERS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ResourceOwnerIsMe(PKI_ENTITIES_GROUPS_RESOURCE))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}
	
	for(i=0; i<SKM_sk_num(USERS_GROUP, pki_conf->my_conf->groups->list); i++)
	{
		currGroup = SKM_sk_value(USERS_GROUP, pki_conf->my_conf->groups->list, i);
		if(!currGroup) continue;

		// Get the associated group from the request
		for(j=0; j<SKM_sk_num(USERS_GROUP, body->d.groups->list);j++)
		{
			userGroup = SKM_sk_value(USERS_GROUP, body->d.groups->list, j);
			if(!userGroup) continue;

			// We found the group
			if(ASN1_INTEGER_cmp(currGroup->serial, userGroup->serial) == 0)
			{
				//Empty it
				while( (currSerial = SKM_sk_pop(ASN1_INTEGER, currGroup->users_serial)) )
				{
					ASN1_INTEGER_free(currSerial);
				}

				//Copy it to the local one
				while( (currSerial = SKM_sk_pop(ASN1_INTEGER, userGroup->users_serial)) )
				{
					if(SKM_sk_push(ASN1_INTEGER, currGroup->users_serial, currSerial) < 0)
					{
						ASN1_INTEGER_free(currSerial);
					}
				}
				break;
			}
		}
	}

	if(!UpdateEntityConfs(ENTITY_CONF_UPDATE_GROUPS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!WritePersonnalConf())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}

bool Entity_PKI::AddGroup(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	int i;
	USERS_GROUP * currGroup;
	unsigned long newSerial;

	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	ConfAccessLock.LockWrite();

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_MANAGE_USERS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ResourceOwnerIsMe(PKI_ENTITIES_GROUPS_RESOURCE))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}
	
	newSerial = 0;
	// Get the highest serial
	for(i=0; i<SKM_sk_num(USERS_GROUP, pki_conf->my_conf->groups->list); i++)
	{
		currGroup = SKM_sk_value(USERS_GROUP, pki_conf->my_conf->groups->list, i);
		if(!currGroup) continue;
		if(ASN1_STRING_cmp(currGroup->name, body->d.group_name) == 0)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_GROUP_ALREADY_EXISTS);
			ConfAccessLock.Unlock();
			return false;
		}
		if((unsigned long)ASN1_INTEGER_GET(currGroup->serial) > newSerial)
			newSerial = ASN1_INTEGER_GET(currGroup->serial);
	}
	newSerial++;

	currGroup = USERS_GROUP_new();
	if(!currGroup)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ConfAccessLock.Unlock();
		return false;
	}
	ASN1_STRING_SET(&currGroup->name, ASN1_STRING_GET(body->d.group_name));
	ASN1_INTEGER_SET(&currGroup->serial, newSerial);

	if(SKM_sk_push(USERS_GROUP, pki_conf->my_conf->groups->list, currGroup) < 0)
	{
		USERS_GROUP_free(currGroup);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!UpdateEntityConfs(ENTITY_CONF_UPDATE_GROUPS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!WritePersonnalConf())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}

bool Entity_PKI::DelGroup(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	int i;
	int j;
	USERS_GROUP * currGroup;
	STORED_ENTITY_CONF * currConf;

	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	ConfAccessLock.LockWrite();

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_MANAGE_USERS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ResourceOwnerIsMe(PKI_ENTITIES_GROUPS_RESOURCE))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}
	
	currGroup = NULL;
	for(i=0; i<SKM_sk_num(USERS_GROUP, pki_conf->my_conf->groups->list); i++)
	{
		currGroup = SKM_sk_value(USERS_GROUP, pki_conf->my_conf->groups->list, i);
		if(!currGroup) continue;

		//Is it the seeked group ?
		if(ASN1_INTEGER_cmp(currGroup->serial, body->d.group_id) == 0)
		{
			break;
		}
	}
	if(i == SKM_sk_num(USERS_GROUP, pki_conf->my_conf->groups->list))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_GROUP_UNKNOWN);
		ConfAccessLock.Unlock();
		return false;
	}

	// We found the group, now we need to make sure this group
	// isn't used in an RA configuration for the auto ldap synch
	for(j=0; j<SKM_sk_num(STORED_ENTITY_CONF, pki_conf->confs); j++)
	{
		currConf = SKM_sk_value(STORED_ENTITY_CONF, pki_conf->confs, j);
		if(!currConf)
			continue;

		// Is it a RA
		if(currConf->conf->body->type != ENTITY_TYPE_RA)
			continue;

		// Is the LDAP auto synch activated
		if(!ASN1_BIT_STRING_get_bit(currConf->conf->body->d.ra_conf->d.RA_CONF_PTR->Flags, ALLOW_LDAP_AUTO_SYNCH))
			continue;

		// Is the group the one we delete
		if(ASN1_INTEGER_cmp(currGroup->serial, currConf->conf->body->d.ra_conf->d.RA_CONF_PTR->ldap_profiles_group) == 0)
		{
			// ERROR !
			NEWPKIerr(PKI_ERROR_TXT, ERROR_GROUP_USED_RA_LDAP_SYNCH);
			ERR_add_error_data(1, ASN1_STRING_GET(currConf->conf->Name));
			ConfAccessLock.Unlock();
			return false;
		}

	}

	USERS_GROUP_free(currGroup);
	SKM_sk_delete(USERS_GROUP, pki_conf->my_conf->groups->list, i);


	if(!UpdateEntityConfs(ENTITY_CONF_UPDATE_GROUPS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!WritePersonnalConf())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}

bool Entity_PKI::RenameGroup(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	int i;
	USERS_GROUP * currGroup;
	USERS_GROUP * foundGroup;

	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	ConfAccessLock.LockWrite();

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_MANAGE_USERS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ResourceOwnerIsMe(PKI_ENTITIES_GROUPS_RESOURCE))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}
	
	foundGroup = NULL;
	for(i=0; i<SKM_sk_num(USERS_GROUP, pki_conf->my_conf->groups->list); i++)
	{
		currGroup = SKM_sk_value(USERS_GROUP, pki_conf->my_conf->groups->list, i);
		if(!currGroup) continue;

		//Is it the seeked group ?
		if(ASN1_INTEGER_cmp(currGroup->serial, body->d.rename_group->id) == 0)
		{
			foundGroup = currGroup;
		}
		else
		{
			// A group already has this name ?
			if(ASN1_STRING_cmp(currGroup->name, body->d.rename_group->name) == 0)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_GROUP_ALREADY_EXISTS);
				ConfAccessLock.Unlock();
				return false;
			}
		}
	}
	if(!foundGroup)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_GROUP_UNKNOWN);
		ConfAccessLock.Unlock();
		return false;
	}

	ASN1_STRING_SET(&foundGroup->name, ASN1_STRING_GET(body->d.rename_group->name));

	if(!UpdateEntityConfs(ENTITY_CONF_UPDATE_GROUPS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!WritePersonnalConf())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}



bool Entity_PKI::OnNewConf(const ENTITY_CONF_CRYPTED * newConf)
{
	return true;
}

bool Entity_PKI::OnNewRequests(const STORED_NEWPKI_REQUESTS * reqs)
{
	return true;
}

bool Entity_PKI::OnRequestSent(unsigned long RequestId, const ASN1_OCTET_STRING * transactionId, const CRYPTED_NEWPKI_REQUEST * req)
{
	return true;
}


bool Entity_PKI::OnResponseSent(const STORED_NEWPKI_RESPONSE * resp)
{
	return true;
}


bool Entity_PKI::OnNewResponses(const STORED_NEWPKI_RESPONSES * resps)
{
	return true;
}



bool Entity_PKI::GetWaitingTransactionIds(TRANSACTION_IDS *transactionIDs)
{
	return false;
}

bool Entity_PKI::GetProcessedTransactionIds(TRANSACTION_IDS *transactionIDs)
{
	return false;
}

bool Entity_PKI::GetWaitingRequests(STACK_OF(CRYPTED_NEWPKI_REQUEST) * Requests)
{
	return true;
}

bool Entity_PKI::GetWaitingResponses(STACK_OF(STORED_NEWPKI_RESPONSE) * Responses)
{
	return true;
}

ENTITY_CONF_CRYPTED * Entity_PKI::EXPORTED_PKI_CONF_get_conf(X509 * cert)
{
	int i;
	ENTITY_CONF_CRYPTED * currConf;
	for(i=0; i< SKM_sk_num(ENTITY_CONF_CRYPTED, pki_conf->crypt_confs->confs->all_confs); i++)
	{
		currConf = SKM_sk_value(ENTITY_CONF_CRYPTED, pki_conf->crypt_confs->confs->all_confs, i);
		if(!currConf) continue;

		if(cert->cert_info->key->public_key->length ==
			currConf->recipient->public_key->length &&
			memcmp(cert->cert_info->key->public_key->data,
			currConf->recipient->public_key->data, 
			currConf->recipient->public_key->length) == 0)
		{
			return currConf;
		}
	}
	return NULL;
}

