/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/

 // Entity_PUBLICATION.cpp: implementation of the Entity_PUBLICATION class.
//
//////////////////////////////////////////////////////////////////////

#include "Entity_PUBLICATION.h"
#include "svintl.h"


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Entity_PUBLICATION::Entity_PUBLICATION(ENTITY_CONSTRUCTOR_PARAMETERS):Entity(ENTITY_CONSTRUCTOR_PARAM_PASSTHRU)
{
	myConf = NULL;
	c_myConf = NULL;
	hThreadWorker.Create(ThreadWorker, this);
	m_OcspServersHandler = NULL;
}

Entity_PUBLICATION::~Entity_PUBLICATION()
{
	hThreadWorker.Stop();

	ConfAccessLock.LockWrite();
	if(myConf)
		ENTITY_CONF_free(myConf);
	if(c_myConf)
		LOCAL_PUBLICATION_CONF_free(c_myConf);
	ConfAccessLock.Unlock();
	UnloadAllPub();
}

bool Entity_PUBLICATION::Create(const ENTITY_CREATION_DATAS * Params, ADMIN_RESPONSE_BODY * response)
{
	if(!Params)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	if(Params->type != ENTITY_TYPE_PUBLICATION)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	PUB_CREATION_REQ * cre_params = Params->d.pub_create;

	ENTITY_CREATION_RESP * resp;

	if(!cre_params)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	const char * InitialCreates[]={PUB_ENTITY_CREATE_1, NULL};

	//We create the database
	if(!Private_CreateEntity(cre_params->entity_key, &PublicationStore, InitialCreates))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	const char * keyPem;
	SQL sql(m_DbConn);
	mString req;
	

	//OCSP key
	switch(cre_params->ocsp_key->type)
	{
		case GEN_PRIVATE_KEY_TYPE_ENGINE:
			keyPem = ASN1_STRING_GET(cre_params->ocsp_key->d.keyid);
			if(!m_OcspKey.SetKey(keyPem, m_Engine))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				Destroy();
				return false;
			}
		case GEN_PRIVATE_KEY_TYPE_KEYLEN:
			if(!m_OcspKey.GenerateKey(ASN1_INTEGER_GET(cre_params->ocsp_key->d.keylen), m_Engine))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				Destroy();
				return false;
			}
			keyPem = m_OcspKey.GetRsaKeyPem().c_str();
			break;
		default:
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			Destroy();
			return false;
			break;
	}
	//Insert key in DB
	if(req.sprintf(PUB_ENTITY_SET_OCSP, "", keyPem) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		Destroy();
		return false;
	}
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		Destroy();
		return false;
	}


	resp = ENTITY_CREATION_RESP_new();
	if(!resp)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		Destroy();
		return false;
	}
	resp->type = ENTITY_TYPE_PUBLICATION;
	resp->d.pub_create = PUB_ENTITY_CREATION_RESP_new();
	if(!resp->d.pub_create)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ENTITY_CREATION_RESP_free(resp);
		Destroy();
		return false;
	}

	if(X509_PUBKEY_set(&resp->d.pub_create->entity_key, (EVP_PKEY*)m_EntityKey.GetRsaKey()) <= 0)
	{
		ENTITY_CREATION_RESP_free(resp);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		Destroy();
		return false;
	}
	if(X509_PUBKEY_set(&resp->d.pub_create->ocsp_key, (EVP_PKEY*)m_OcspKey.GetRsaKey()) <= 0)
	{
		ENTITY_CREATION_RESP_free(resp);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		Destroy();
		return false;
	}
	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_CREATE_ENTITY, resp))
	{
		ENTITY_CREATION_RESP_free(resp);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_PUBLICATION::Load()
{
	if(!Private_CommonLoad(&PublicationStore))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!IsFullyInit())
	{
		return true;
	}


	SQL sql(m_DbConn);
	mString cert;
	mString key;
	long NumRows;
	
	// We get the ocsp certificate
	if(!sql.Execute(PUB_ENTITY_GET_OCSP))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(NumRows)
	{
		if(!sql.Value(0, "ocsp_cert", cert) || !sql.Value(0, "ocsp_key", key))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		//Did we get our ocsp certificate yet ?
		if(cert.size())
		{
			if(!m_OcspCert.SetCert(cert.c_str()))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				return false;
			}
			if(!m_OcspKey.SetKey(key.c_str(), m_Engine))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				return false;
			}
			if(!m_OcspCert.SetPrivateKey(m_OcspKey))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				return false;
			}
		}
	}
	
	sk_X509_push(EntityParentCerts, X509_dup(c_myConf->cas->RootCa));
	sk_X509_push(EntityParentCerts, X509_dup(c_myConf->cas->EntitiesCa));

	//Start the conf synch
	if(!m_Jobs.StartConfSync())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	//Start the mailer
	if(!m_Jobs.StartMailer())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!m_Jobs.SetEntityParentsCert(EntityParentCerts))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!hThreadWorker.Start())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}

	return true;
}

bool Entity_PUBLICATION::Init(const ENTITY_SIGNATURE_RESP * init_datas)
{
	if(IsFullyInit())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!init_datas)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	if(init_datas->body->type != ENTITY_TYPE_PUBLICATION)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	if(!Private_InitEntity(init_datas->body->d.sign_pub->EntityCert, &myConf, init_datas, &PublicationStore))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	// Insert the OCSP Cert
	mString key;
	long NumRows;
	SQL sql(m_DbConn);
	mString req;

	// We get the ocsp key
	if(!sql.Execute(PUB_ENTITY_GET_OCSP))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!NumRows)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}
	if(!sql.Value(0, "ocsp_key", key))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!m_OcspCert.SetCert(init_datas->body->d.sign_pub->OcspCert))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!m_OcspKey.SetKey(key.c_str(), m_Engine))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!m_OcspCert.SetPrivateKey(m_OcspKey))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//We now insert the OCSP cert in the database
	if(req.sprintf(PUB_ENTITY_SET_OCSP, m_OcspCert.GetCertPEM().c_str(), key.c_str()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_PUBLICATION::PersonnalConfFromPEM(const mString &conf_pem)
{
	//Convert from PEM
	ASN1_ENCRYPT_SIGN * encrypt;
	encrypt = NULL;

	if(ASN1_ENCRYPT_SIGN_From_PEM(&encrypt, conf_pem.c_str()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//Decrypt the conf
	if(c_myConf)
	{
		LOCAL_PUBLICATION_CONF_free(c_myConf);
		c_myConf = NULL;
	}

	if(LOCAL_PUBLICATION_CONF_decrypt_verify(encrypt, &c_myConf, m_EntityKey.GetRsaKey(), m_EntityKey.GetRsaKey()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ASN1_ENCRYPT_SIGN_free(encrypt);
		return false;
	}
	ASN1_ENCRYPT_SIGN_free(encrypt);
	
	//Extract the conf
	if(myConf)
	{
		ENTITY_CONF_free(myConf);
		myConf = NULL;
	}
	if(!ExtractMyConf(c_myConf->conf, &myConf, c_myConf->cas, c_myConf->cas->PkiCert, false))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_PUBLICATION::PersonnalConfToPem(mString & conf_pem)
{
	char * pem;
	ASN1_ENCRYPT_SIGN * crypted_conf;

	ConfAccessLock.LockRead();

	//Sign/Crypt the conf
	crypted_conf = ASN1_ENCRYPT_SIGN_new();
	if(!crypted_conf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ConfAccessLock.Unlock();
		return false;
	}

	if(LOCAL_PUBLICATION_CONF_sign_encrypt(crypted_conf, c_myConf, m_EntityKey.GetRsaKey(), m_EntityKey.GetRsaKey(), EVP_sha1(), EVP_des_ede3_cbc()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ASN1_ENCRYPT_SIGN_free(crypted_conf);
		ConfAccessLock.Unlock();
		return false;
	}

	//Convert it to PEM
	pem = ASN1_ENCRYPT_SIGN_To_PEM(crypted_conf);
	ASN1_ENCRYPT_SIGN_free(crypted_conf);
	if(!pem)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	conf_pem = pem;
	free(pem);
	ConfAccessLock.Unlock();
	return true;
}


bool Entity_PUBLICATION::LoginUser(UserHandle & hUser, int & UserType)
{
	if(!AclValidator.CanUserPerform(hUser.GetUserCert(), ACL_TYPE_AUTHENTICATE_ON_ENTITY))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	UserType = USER_TYPE_KEYSTORE;
	return true;
}

void Entity_PUBLICATION::LogoutUser(const UserHandle & hUser)
{
}

bool Entity_PUBLICATION::ParseNewConf(const ENTITY_CONF_CRYPTED * newConf, const INTERNAL_PKI_CA * cas)
{
	PUBLICATION_ENTRY * currEntry;
	PUBLICATION_METHOD * currMethod;
	PublicationMethod * newMethod;
	int i, j;
	char * Err;
	char * StrErr;
	mString Name;

	//Should we initialize the local conf ?
	if(!c_myConf)
	{
		c_myConf = LOCAL_PUBLICATION_CONF_new();
		if(!c_myConf)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
			return false;
		}
	}

	//Should we update the internal cas certs ?
	if(cas)
	{
		if(!INTERNAL_PKI_CA_cpy(&c_myConf->cas, cas))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}

	//Save the updated conf
	if(!ENTITY_CONF_CRYPTED_cpy(&c_myConf->conf, newConf))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	// We load the OCSP
	LoadOcsp();
	
	// We now load all the publicatins methods
	ListPublicationsLock.LockWrite();
	UnloadAllPub();

	for(i=0; i<SKM_sk_num(PUBLICATION_ENTRY, myConf->body->d.pub_conf->d.PUB_CONF_PTR->Publications); i++)
	{
		currEntry = SKM_sk_value(PUBLICATION_ENTRY, myConf->body->d.pub_conf->d.PUB_CONF_PTR->Publications, i);
		if(!currEntry)
			continue;

		for(j=0; j<SKM_sk_num(PUBLICATION_METHOD, currEntry->Methods); j++)
		{
			currMethod = SKM_sk_value(PUBLICATION_METHOD, currEntry->Methods, j);
			if(!currMethod)
				continue;

			newMethod = NULL;
			try
			{
				newMethod = new PublicationMethod(ASN1_STRING_GET(currEntry->CaName), currMethod);
				if(!newMethod)
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
					Name.sprintf("%s - %s", ASN1_STRING_GET(currEntry->CaName), ASN1_STRING_GET(currMethod->Name));
					StrErr = ERR_to_string();
					if(!StrErr)
						Err = "Unknown Error";
					else
						Err = StrErr;

					m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_LOADING_PUB_METH, 0, _sv("none"), LOG_NO_OBJECTID, Name.c_str(), StrErr);
					if(StrErr)
						free(StrErr);
				}
			}
			catch(ExceptionNewPKI e)
			{
				Name.sprintf("%s - %s", ASN1_STRING_GET(currEntry->CaName), ASN1_STRING_GET(currMethod->Name));
				StrErr = ERR_to_string();
				if(!StrErr)
					Err = "Unknown Error";
				else
					Err = StrErr;

				m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_LOADING_PUB_METH, 0, _sv("none"), LOG_NO_OBJECTID, Name.c_str(), StrErr);
				if(StrErr)
					free(StrErr);
			}
			if(newMethod)
			{
				ListPublications.push_back(newMethod);
			}
		}
	}

	ListPublicationsLock.Unlock();

	return true;
}

bool Entity_PUBLICATION::Upgrade(const char * Version)
{
	return true;
}


void Entity_PUBLICATION::GetACL_List(STACK_OF(ASN1_INTEGER) *acl_list)
{
	ASN1_INTEGER * acl_id;
	int i;
	static ACL_TYPE list_acls[] =
	{
		ACL_TYPE_VIEW_LOGS,
		(ACL_TYPE)0
	};
	for(i=0; list_acls[i]; i++)
	{
		acl_id = NULL;
		ASN1_INTEGER_SET(&acl_id, list_acls[i]);
		if(acl_id)
		{
			if(SKM_sk_push(ASN1_INTEGER, acl_list, acl_id) < 0)
			{
				ASN1_INTEGER_free(acl_id);
			}
		}
	}	
}



bool Entity_PUBLICATION::ParseAdminCommand(ADMIN_RESPONSE_BODY * response, const PKI_CERT & ClientCert, const ADMIN_REQUEST * AdminRequest)
{
	//We only accept SSLv3 connection	
	if(!ClientCert)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ERR_to_ADMIN_RESPONSE(response);
		return false;
	}
	return Private_ParseAdminCommand(true, this, LogsType, response, ClientCert, AdminRequest, GetUserHandle());
}

bool Entity_PUBLICATION::Private_ParseAdminCommand(bool ExecuteCmd, Entity * me_this, vector<LOG_MESSAGE_TYPE> &  mLogsType, ADMIN_RESPONSE_BODY * response, const PKI_CERT & ClientCert, const ADMIN_REQUEST * AdminRequest, const UserHandle & hUser)
{

	PARSER_COMMAND_BEGIN(Entity_PUBLICATION, response, 0, AdminRequest, ClientCert, hUser, ExecuteCmd, me_this, mLogsType)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_ENTITY_GET_MY_CONF)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_SEND_ADMIN_MAIL)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_SEND_MAIL)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_NEW_REQUET)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_LOADING_PUB_METH)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_CERT_PUBLICATION)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_REV_PUBLICATION)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_CRL_PUBLICATION)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_OCSP_REQUEST)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_OCSP_STOP)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_OCSP_START)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_LOGIN,					Entity_PUBLICATION::UserLogin,				LOG_MESSAGE_TYPE_USER_LOGIN, (ClientCert)?(char*)ClientCert.GetStringName():NULL, LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_ENUM_LOGS,				Entity_PUBLICATION::EnumLogs)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_LOGS_COUNT,			Entity_PUBLICATION::GetLogsCount)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_MY_ACL,				Entity_PUBLICATION::GetMyACL)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_LOGS_TYPE,			Entity_PUBLICATION::GetLogsType)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_SEND_ADMIN_MAIL,			Entity_PUBLICATION::AdminSendMail,			LOG_MESSAGE_TYPE_ADD_ADMIN_MAIL_QUEUE,	_sv("none"), LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_CHECK_LOGS,				Entity_PUBLICATION::CheckLogsIntegrity)
	PARSER_COMMAND_END(Entity_PUBLICATION)
}

void Entity_PUBLICATION::LogsTypeGet(vector<LOG_MESSAGE_TYPE> &  cLogsType)
{
	Private_ParseAdminCommand(false, NULL, cLogsType, NULL, PKI_CERT(), NULL, UserHandle());
}


void Entity_PUBLICATION::SpawnStop()
{
	hThreadWorker.SpawStop();
	m_Jobs.SpawnStop();
}



bool Entity_PUBLICATION::OnNewConf(const ENTITY_CONF_CRYPTED * newConf)
{
	if(!ExtractMyConf(newConf, &myConf, NULL, c_myConf->cas->PkiCert, true))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_PUBLICATION::OnNewRequests(const STORED_NEWPKI_REQUESTS * reqs)
{
	STORED_NEWPKI_REQUEST * currReq;
	int i;
	mString Sender;
	mString TransactionId;
	char * Err;
	char * strErr;
	bool ret;
	bool IsKnown;

	// We don't continue if we have no requests
	// For each transaction, search if it's not being processed
	// or has been processed yet
	for(i=0; i<SKM_sk_num(STORED_NEWPKI_REQUEST, reqs->requests); i++ )
	{
		currReq = SKM_sk_value(STORED_NEWPKI_REQUEST, reqs->requests, i);
		if(!currReq) continue;

		// Is this transaction known ?
		if(!PublicationStore.Responder_IsRequestKnown(currReq->transactionID, IsKnown))
			continue;
		if(IsKnown)
			continue;


		// Create a working request from this request
		ret = AddNewSTORED_NEWPKI_REQUEST(currReq, Sender);

		if(!NewPKIStore::transactionIDtoString(currReq->transactionID, TransactionId))
			TransactionId = _sv("Unknown");

		if(!Sender.size())
			Sender = _sv("Unknown");

		if(!ret)
		{
			if(! (Err = ERR_to_string()) )
				strErr= _sv("Unknown");
			else
				strErr = Err;

			m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_NEW_REQUET, 0, Sender.c_str(), LOG_NO_OBJECTID, TransactionId.c_str(), strErr);

			if(Err)
				free(Err);
		}
		else
		{
			m_Logging->LogMessage(LOG_STATUS_TYPE_SUCCESS, LOG_MESSAGE_TYPE_NEW_REQUET, 0, Sender.c_str(), LOG_NO_OBJECTID, TransactionId.c_str());
		}
	}

	return true;
}

bool Entity_PUBLICATION::OnRequestSent(unsigned long RequestId, const ASN1_OCTET_STRING * transactionId, const CRYPTED_NEWPKI_REQUEST * req)
{
	return true;
}


bool Entity_PUBLICATION::OnResponseSent(const STORED_NEWPKI_RESPONSE * resp)
{
	return PublicationStore.Responder_ResponseSent(resp->transactionID);
}


bool Entity_PUBLICATION::OnNewResponses(const STORED_NEWPKI_RESPONSES * resps)
{
	return true;
}



bool Entity_PUBLICATION::GetWaitingTransactionIds(TRANSACTION_IDS *transactionIDs)
{
	return true;
}

bool Entity_PUBLICATION::GetProcessedTransactionIds(TRANSACTION_IDS *transactionIDs)
{
	return PublicationStore.Responder_GetTransactionIds(transactionIDs);
}

bool Entity_PUBLICATION::GetWaitingRequests(STACK_OF(CRYPTED_NEWPKI_REQUEST) * Requests)
{
	return true;
}

bool Entity_PUBLICATION::GetWaitingResponses(STACK_OF(CRYPTED_NEWPKI_RESPONSE) * Responses)
{
	return PublicationStore.Responder_GetUnsentResponses(Responses);
}

void Entity_PUBLICATION::UnloadAllPub()
{
	int i;
	PublicationMethod * currMethod;

	ListPublicationsLock.LockWrite();
	
	for(i=0; i<(int)ListPublications.size(); i++)
	{
		currMethod = ListPublications.at(i);
		if(currMethod)
			delete currMethod;
	}
	ListPublications.clear();
	ListPublicationsLock.Unlock();

}

void Entity_PUBLICATION::ThreadWorker(const NewpkiThread * Thread, void *param)
{
	Entity_PUBLICATION * me_this = (Entity_PUBLICATION *)param;
	time_t currTime;
	time_t lastTime;
	WORKING_CA_REQUEST * workingReq;
	STACK_OF(WORKING_CA_REQUEST) * WorkingReqs = NULL;


	//We wait a bit (2 mins)
	time(&lastTime);
	do
	{
		NewpkiThread::Sleep(500);
		time(&currTime);
	}
	while( ((currTime - lastTime) < 120) && !Thread->ShouldStop());


	do
	{
		NewpkiThread::Sleep(500);
		WorkingReqs = SKM_sk_new_null(WORKING_CA_REQUEST);
	}
	while( !WorkingReqs && !Thread->ShouldStop());


	while(!Thread->ShouldStop())
	{
		do
		{
			NewpkiThread::Sleep(1000);
		}
		while(!me_this->PublicationStore.Responder_GetWaitingRequests(WorkingReqs) && !Thread->ShouldStop());

		//I have a new request
		while( (workingReq = SKM_sk_pop(WORKING_CA_REQUEST, WorkingReqs))  && !Thread->ShouldStop())
		{
			//Start to proceed with request
			ERR_clear_error();
			me_this->ProceedWithRequest(workingReq);
		}
	}

	if(WorkingReqs)
		SKM_sk_pop_free(WORKING_CA_REQUEST, WorkingReqs, WORKING_CA_REQUEST_free);
}

bool Entity_PUBLICATION::ProceedWithRequest(WORKING_CA_REQUEST * req)
{
	PKI_CERT cert;
	PKI_CRL crl;
	NEWPKI_RESPONSE * currResponse;
	int i, j;
	PublicationMethod * currMethod;
	char * StrError;
	mString ldap_uid;
	char * Err;
	PKI_P7B p7b;
	X509 * currCert;
	LOG_MESSAGE_TYPE LogType;
	bool PublicationResult;
	mString PubName;
	time_t revdate;

	//Proceed with the request
	if(req->request->type != NEWPKI_REQUEST_TYPE_PUB)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		ERR_to_STORED_NEWPKI_RESPONSE(PublicationStore, req->transactionID, req->sender);

		StrError = ERR_to_string();
		if(StrError)
		{
			Err = StrError;
		}
		else
		{
			Err = "Unknown Error";
		}
		m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_CERT_PUBLICATION, 0, ASN1_STRING_GET(req->SenderName), LOG_NO_OBJECTID, _sv("Unknown"), Err);
		if(StrError)
			free(StrError);
		return false;
	}

	// Proceed with publication request

	ldap_uid = req->request->d.pub_request->ldap_uid;


	AddUnknownCAs(req->request->d.pub_request->ParentCerts);


	// We make the certificate info available to the OCSP responder
	// even if we don't have a publication method for this CA
	switch(req->request->d.pub_request->body->type)
	{
		case NEWPKI_PUB_REQUEST_REV:
			if(cert.SetCert(req->request->d.pub_request->body->d.rev->cert))
			{
				currCert = GetIssuer(req->request->d.pub_request->ParentCerts, cert.GetX509(), NULL);
				if(!currCert)
				{
					// TODO insrer un log spcifique
				}
				else
				{
					revdate = ASN1_TIME_to_timet((char*)req->request->d.pub_request->body->d.rev->revdate->data);
					if(!PublicationStore.OnNewRevocation(cert.GetSerial(), revdate, currCert->cert_info->subject, currCert->cert_info->key))
					{
						// TODO insrer un log spcifique
					}
				}
			}
			else
			{
				// TODO insrer un log spcifique
			}
			break;
		case NEWPKI_PUB_REQUEST_CERT:
			if(cert.SetCert(req->request->d.pub_request->body->d.cert))
			{
				currCert = GetIssuer(req->request->d.pub_request->ParentCerts, cert.GetX509(), NULL);
				if(!currCert)
				{
					// TODO insrer un log spcifique
				}
				else
				{
					if(!PublicationStore.OnNewCertificate(cert.GetSerial(), currCert->cert_info->subject, currCert->cert_info->key))
					{
						// TODO insrer un log spcifique
					}
				}
			}
			else
			{
				// TODO insrer un log spcifique
			}
			break;
		case NEWPKI_PUB_REQUEST_CRL:
			if(crl.SetCRL(req->request->d.pub_request->body->d.crl))
			{
				currCert = GetIssuer(req->request->d.pub_request->ParentCerts, NULL, crl.GetX509_CRL());
				if(!currCert)
				{
					// TODO insrer un log spcifique
				}
				else
				{
					if(!PublicationStore.OnNewCrl(crl, currCert->cert_info->subject, currCert->cert_info->key))
					{
						// TODO insrer un log spcifique
					}
				}
			}
			else
			{
				// TODO insrer un log spcifique
			}
			break;
	}
	

	PublicationResult = true;
	ListPublicationsLock.LockRead();
	for(i=0; i<(int)ListPublications.size(); i++)
	{
		currMethod = ListPublications.at(i);
		if(!currMethod)
			continue;

		// Is it the right entity
		if( strcmp(currMethod->GetCaName().c_str(), ASN1_STRING_GET(req->SenderName) ) != 0)
			continue;

		ERR_clear_error();
		PubName.sprintf("%s : %s", currMethod->GetName().c_str(), ASN1_STRING_GET(req->request->d.pub_request->object));
		
		// Is it the right kind ?
		switch(req->request->d.pub_request->body->type)
		{
			case NEWPKI_PUB_REQUEST_CERT:
				if(currMethod->GetType() != PUBLICATION_METHOD_TYPE_CERT)
					continue;

				LogType = LOG_MESSAGE_TYPE_CERT_PUBLICATION;

				// Generating the PKCS7
				if(SKM_sk_num(X509, req->request->d.pub_request->ParentCerts))
				{
					for(j=0; j<SKM_sk_num(X509, req->request->d.pub_request->ParentCerts); j++)
					{
						currCert = SKM_sk_value(X509, req->request->d.pub_request->ParentCerts, j);
						if(!currCert)
							continue;
						p7b.AddCert(currCert);
					}
					if(!p7b.AddCert(cert.GetX509()))
					{
						PublicationResult = false;
						NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					}
					if(!p7b.Generate())
					{
						PublicationResult = false;
						NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					}
					else
					{
						PublicationResult = currMethod->OnNewCertificate(cert, p7b, ldap_uid);
						if(!PublicationResult)
						{
							NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
						}
					}
				}
				else
				{
					PublicationResult = false;
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				}

				break;

			case NEWPKI_PUB_REQUEST_REV:
				if(currMethod->GetType() != PUBLICATION_METHOD_TYPE_CERT)
					continue;

				LogType = LOG_MESSAGE_TYPE_REV_PUBLICATION;

				PublicationResult = currMethod->OnNewRevocation(cert, ldap_uid);
				if(!PublicationResult)
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				}
				break;

			case NEWPKI_PUB_REQUEST_CRL:
				if(currMethod->GetType() != PUBLICATION_METHOD_TYPE_CRL)
					continue;

				LogType = LOG_MESSAGE_TYPE_CRL_PUBLICATION;

				PublicationResult = currMethod->OnNewCrl(crl);
				if(!PublicationResult)
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				}
				break;
		
			default:
				continue;
				break;
		}

		if(PublicationResult)
		{
			m_Logging->LogMessage(LOG_STATUS_TYPE_SUCCESS, LogType, 0, ASN1_STRING_GET(req->SenderName), LOG_NO_OBJECTID, PubName.c_str());
		}
		else
		{
			StrError = ERR_to_string();
			if(StrError)
			{
				Err = StrError;
			}
			else
			{
				Err = "Unknown Error";
			}
			m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LogType, 0, ASN1_STRING_GET(req->SenderName), LOG_NO_OBJECTID, PubName.c_str(), Err);
			if(StrError)
				free(StrError);
		}
	}
	ListPublicationsLock.Unlock();

	currResponse = NEWPKI_RESPONSE_new();
	if(!currResponse)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);

		StrError = ERR_to_string();
		if(StrError)
		{
			Err = StrError;
		}
		else
		{
			Err = "Unknown Error";
		}
		m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_CERT_PUBLICATION, 0, ASN1_STRING_GET(req->SenderName), LOG_NO_OBJECTID, _sv("Unknown"), Err);
		if(StrError)
			free(StrError);
		return false;
	}

	currResponse->type = NEWPKI_RESPONSE_TYPE_PUB;
	currResponse->d.pub_response = NEWPKI_PUB_RESPONSE_new();
	if(!currResponse->d.pub_response)
	{
		NEWPKI_RESPONSE_free(currResponse);
		return false;
	}
	ASN1_INTEGER_SET(&currResponse->d.pub_response->type, req->request->d.pub_request->body->type);
	ASN1_STRING_SET(&currResponse->d.pub_response->object, ASN1_STRING_GET(req->request->d.pub_request->object));

	if(PublicationResult)
	{
		ASN1_INTEGER_SET(&currResponse->d.pub_response->status, PUB_RESPONSE_PUBLISHED);
	}
	else
	{
		ASN1_INTEGER_SET(&currResponse->d.pub_response->status, PUB_RESPONSE_ERROR);
		currResponse->d.pub_response->errors = ERR_to_ERROR_ENTRIES();
	}
	
	SendResponse(PublicationStore, req->transactionID, currResponse, req->sender);
	NEWPKI_RESPONSE_free(currResponse);
	
	return true;
}


bool Entity_PUBLICATION::AddNewSTORED_NEWPKI_REQUEST(STORED_NEWPKI_REQUEST * req, mString & Sender)
{
	NEWPKI_REQUEST * Request;
	WORKING_CA_REQUEST * w_req;
	EVP_PKEY * signerkey;


	//Verify the request
	if(!STORED_NEWPKI_REQUEST_check(req, Sender))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ERR_to_STORED_NEWPKI_RESPONSE(PublicationStore, req->transactionID, req->c_info->sender);
		return false;
	}

	w_req = WORKING_CA_REQUEST_new();
	if(!w_req)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	// Copy the sender's name
	ASN1_STRING_SET(&w_req->SenderName, Sender.c_str());
	
	//Copy transaction ID
	if(!ASN1_OCTET_STRING_cpy(&w_req->transactionID, req->transactionID))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		WORKING_CA_REQUEST_free(w_req);
		return false;
	}
	//Copy the sender public key
	if(!X509_PUBKEY_cpy(&w_req->sender, req->c_info->sender))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		WORKING_CA_REQUEST_free(w_req);
		return false;
	}

	signerkey = X509_PUBKEY_get(req->c_info->sender);
	if(!signerkey)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		WORKING_CA_REQUEST_free(w_req);
		return false;
	}

	// Decrypt the request
	if(NEWPKI_REQUEST_decrypt_verify(req->c_info->c_request, &Request, signerkey, m_EntityKey.GetRsaKey()) <= 0)
	{
		WORKING_CA_REQUEST_free(w_req);
		EVP_PKEY_free(signerkey);
		//We generate an error
		ERR_clear_error();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		ERR_to_STORED_NEWPKI_RESPONSE(PublicationStore, req->transactionID, req->c_info->sender);
		return false;
	}
	EVP_PKEY_free(signerkey);

	if(w_req->request)
		NEWPKI_REQUEST_free(w_req->request);
	w_req->request = Request;

	ERR_clear_error();
	if(!PublicationStore.Responder_InsertRequest(w_req))
	{
		WORKING_CA_REQUEST_free(w_req);
		//We generate an error
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ERR_to_STORED_NEWPKI_RESPONSE(PublicationStore, req->transactionID, req->c_info->sender);
		return false;
	}
	WORKING_CA_REQUEST_free(w_req);

	return true;
}

bool Entity_PUBLICATION::STORED_NEWPKI_REQUEST_check(STORED_NEWPKI_REQUEST * req, mString & Sender)
{
	int i;
	CA_ENTRY_INFO * currCa;

	ConfAccessLock.LockRead();

	// We need to search that the RA has the right 
	// to request a certificate from us
	for(i=0; i < SKM_sk_num(CA_ENTRY_INFO, myConf->Cas->list); i++)
	{
		currCa = SKM_sk_value(CA_ENTRY_INFO, myConf->Cas->list, i);
		if(!currCa) continue;

		//Is it the same public key
		if(currCa->CaSSL->cert_info->key->public_key->length ==
			req->c_info->sender->public_key->length &&
			memcmp(currCa->CaSSL->cert_info->key->public_key->data,
			req->c_info->sender->public_key->data, 
			req->c_info->sender->public_key->length) == 0)
		{
			Sender = currCa->Name;
			break;
		}
	}
	if(i == SKM_sk_num(CA_ENTRY_INFO, myConf->Cas->list))
	{
		//This RA is not allowed to request a certificate
		//from us
		ConfAccessLock.Unlock();
		return false;
	}

	
	// Am I the recipient
	if(! (m_EntityCert == req->c_info->recipient) )
	{
		ConfAccessLock.Unlock();
		return false;
	}

	ConfAccessLock.Unlock();
	return true;
}

bool Entity_PUBLICATION::GetOcspCertStatus(OCSP_CERTID *cid, int & OcspStatus, time_t & rev_date, mString & CaName, unsigned long & RequestSerial)
{
	int i;
	X509 * currCa;
	const EVP_MD *dgst;
	char oid[50];
	OCSP_CERTID *ca_id;

	if(!cid->serialNumber)
	{
		NEWPKIerr(OCSP_ERROR_TXT, ERROR_BAD_DATAS_FORMAT);
		ERR_add_error_data(1, "serialNumber");
		OcspStatus = V_OCSP_CERTSTATUS_UNKNOWN;
		return false;
	}
	dgst = EVP_get_digestbyobj(cid->hashAlgorithm->algorithm);
	if(!dgst)
	{
		NEWPKIerr(OCSP_ERROR_TXT, ERROR_BAD_DATAS_FORMAT);
		if(OBJ_obj2txt(oid, sizeof(oid), cid->hashAlgorithm->algorithm, 1) <= 0)
		{
			strcpy(oid, "unknown");
		}
		ERR_add_error_data(2, "hashAlgorithm->algorithm : ", oid);
		OcspStatus = V_OCSP_CERTSTATUS_UNKNOWN;
		return false;
	}


	ConfAccessLock.LockRead();

	// We search if we know this CA (might come latter)
	for(i=0; i < SKM_sk_num(X509, c_myConf->known_cas); i++)
	{
		currCa = SKM_sk_value(X509, c_myConf->known_cas, i);
		if(!currCa)
			continue;

		if( !(ca_id = OCSP_cert_to_id(dgst, NULL, currCa)) )
			continue;
		
		// Is it for this CA ?
		if (OCSP_id_issuer_cmp(ca_id, cid) != 0)
		{
			OCSP_CERTID_free(ca_id);
			continue;
		}
		OCSP_CERTID_free(ca_id);

		RequestSerial = ASN1_INTEGER_GET(cid->serialNumber);
		if(PublicationStore.GetCertStatus(RequestSerial, currCa->cert_info->subject, currCa->cert_info->key, OcspStatus, rev_date))
		{
			ConfAccessLock.Unlock();
			return true;
		}
		else
		{
			NEWPKIerr(OCSP_ERROR_TXT, ERROR_ABORT);
			ConfAccessLock.Unlock();
			return false;
		}
	}
	ConfAccessLock.Unlock();

	OcspStatus = V_OCSP_CERTSTATUS_UNKNOWN;
	return true;
}

void Entity_PUBLICATION::AddUnknownCAs(STACK_OF(X509) *CAs)
{
	int i, j;
	X509 * currCa1;
	X509 * currCa2;
	int ctr;

	// We add to our list all the unknown CAs
	ConfAccessLock.LockWrite();

	ctr = SKM_sk_num(X509, c_myConf->known_cas);

	for(i=0; i < SKM_sk_num(X509, CAs); i++)
	{
		currCa1 = SKM_sk_value(X509, CAs, i);
		if(!currCa1)
			continue;

		for(j=0; j < SKM_sk_num(X509, c_myConf->known_cas); j++)
		{
			currCa2 = SKM_sk_value(X509, c_myConf->known_cas, j);
			if(!currCa2)
				continue;

			if(X509_cmp(currCa1, currCa2) == 0)
				break;
		}

		// The CA is not known
		if(j == SKM_sk_num(X509, c_myConf->known_cas))
		{
			currCa1 = X509_dup(currCa1);
			if(currCa1)
			{
				if(SKM_sk_push(X509, c_myConf->known_cas, currCa1) < 0)
				{
					X509_free(currCa1);
				}
			}
		}
	}

	// Did we add some entries ?
	if(ctr != SKM_sk_num(X509, c_myConf->known_cas))
	{
		ConfAccessLock.Unlock();
		WritePersonnalConf();
		return;
	}

	ConfAccessLock.Unlock();
}

X509 * Entity_PUBLICATION::GetIssuer(STACK_OF(X509) * ParentCerts, X509 *Cert, X509_CRL * Crl)
{
	int i;
	X509 * currCa;
	EVP_PKEY * pubkey;

	if(!Cert && !Crl)
		return NULL;

	for(i=SKM_sk_num(X509, ParentCerts)-1; i >= 0 ; i--)
	{
		currCa = SKM_sk_value(X509, ParentCerts, i);
		if(!currCa)
			continue;

		if( (pubkey = X509_get_pubkey(currCa)) )
		{
			if(Cert)
			{
				if(X509_verify(Cert, pubkey) == 1)
				{
					EVP_PKEY_free(pubkey);
					return currCa;
				}
			}
			else if(Crl)
			{
				if(X509_CRL_verify(Crl, pubkey) == 1)
				{
					EVP_PKEY_free(pubkey);
					return currCa;
				}
			}
			EVP_PKEY_free(pubkey);
		}
	}

	return NULL;
}

bool Entity_PUBLICATION::OnNewOCSP(const char * Ip, const OCSP_REQUEST *request, OCSP_RESPONSE **response)
{
	char * Err;
	char * strErr;
	if(!Private_OnNewOCSP(Ip, request, response))
	{
		if(! (Err = ERR_to_string()) )
			strErr= _sv("Unknown");
		else
			strErr = Err;
		m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_OCSP_REQUEST, 0, _sv("none"), LOG_NO_OBJECTID, m_EntityName.c_str(), strErr);
		if(Err)
			free(Err);
		return false;
	}
	return true;
}

bool Entity_PUBLICATION::Private_OnNewOCSP(const char * Ip, const OCSP_REQUEST *request, OCSP_RESPONSE **response)
{
	ASN1_TIME *thisupd;
	OCSP_CERTID *cid;
	OCSP_BASICRESP *bs;
	int i, id_count;
	int cert_status;
	ASN1_TIME *revtm;
	OCSP_ONEREQ *one;
	STACK_OF(X509) * ParentCerts;
	time_t rev_date;
	mString CaName;
	unsigned long RequestSerial;


	id_count = OCSP_request_onereq_count((OCSP_REQUEST*)request);
	if (id_count <= 0)
	{
		*response = OCSP_response_create(OCSP_RESPONSE_STATUS_MALFORMEDREQUEST, NULL);
		NEWPKIerr(OCSP_ERROR_TXT, ERROR_BAD_DATAS_FORMAT);
		return false;
	}

	bs = OCSP_BASICRESP_new();
	if(!bs)
	{
		*response = OCSP_response_create(OCSP_RESPONSE_STATUS_INTERNALERROR, NULL);
		NEWPKIerr(OCSP_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	thisupd = X509_gmtime_adj(NULL, 0);
	if(!thisupd)
	{
		*response = OCSP_response_create(OCSP_RESPONSE_STATUS_INTERNALERROR, NULL);
		NEWPKIerr(OCSP_ERROR_TXT, ERROR_MALLOC);
		OCSP_BASICRESP_free(bs);
		return false;
	}

	// Examine each certificate id in the request
	for (i = 0; i < id_count; i++)
	{
		// Get request
		one = OCSP_request_onereq_get0((OCSP_REQUEST*)request, i);
		if(!one)
		{
			*response = OCSP_response_create(OCSP_RESPONSE_STATUS_INTERNALERROR, NULL);
			NEWPKIerr(OCSP_ERROR_TXT, ERROR_ABORT);
			OCSP_BASICRESP_free(bs);
			ASN1_TIME_free(thisupd);
			return false;
		}
		// Get certificate info
		cid = OCSP_onereq_get0_id(one);
		if(!cid)
		{
			*response = OCSP_response_create(OCSP_RESPONSE_STATUS_INTERNALERROR, NULL);
			NEWPKIerr(OCSP_ERROR_TXT, ERROR_ABORT);
			OCSP_BASICRESP_free(bs);
			ASN1_TIME_free(thisupd);
			return false;
		}

		// Get its state
		if(!GetOcspCertStatus(cid, cert_status, rev_date, CaName, RequestSerial))
		{
			*response = OCSP_response_create(OCSP_RESPONSE_STATUS_INTERNALERROR, NULL);
			NEWPKIerr(OCSP_ERROR_TXT, ERROR_ABORT);
			OCSP_BASICRESP_free(bs);
			ASN1_TIME_free(thisupd);
			return false;
		}

		CaName += ":";
		CaName += RequestSerial;

		m_Logging->LogMessage(LOG_STATUS_TYPE_REQUEST, LOG_MESSAGE_TYPE_OCSP_REQUEST, 0, Ip, LOG_NO_OBJECTID, CaName.c_str());

		if(cert_status == V_OCSP_CERTSTATUS_REVOKED)
		{
			revtm = ASN1_TIME_set_localtime(NULL, rev_date);
			if(!revtm)
			{
				*response = OCSP_response_create(OCSP_RESPONSE_STATUS_INTERNALERROR, NULL);
				NEWPKIerr(OCSP_ERROR_TXT, ERROR_ABORT);
				OCSP_BASICRESP_free(bs);
				ASN1_TIME_free(thisupd);
				return false;
			}
		}
		else
		{
			revtm = NULL;
		}

		// Add the status for the current certificate
		if(!OCSP_basic_add1_status(bs, cid, cert_status, 0, revtm, thisupd, NULL))
		{
			*response = OCSP_response_create(OCSP_RESPONSE_STATUS_INTERNALERROR, NULL);
			NEWPKIerr(OCSP_ERROR_TXT, ERROR_ABORT);
			OCSP_BASICRESP_free(bs);
			ASN1_TIME_free(thisupd);
			if(revtm)
				ASN1_TIME_free(revtm);
			return false;
		}
		if(revtm)
			ASN1_TIME_free(revtm);
	}
	ASN1_TIME_free(thisupd);
	

	if(OCSP_copy_nonce(bs, (OCSP_REQUEST*)(OCSP_REQUEST*)request) <= 0)
	{
		*response = OCSP_response_create(OCSP_RESPONSE_STATUS_INTERNALERROR, NULL);
		NEWPKIerr(OCSP_ERROR_TXT, ERROR_ABORT);
		OCSP_BASICRESP_free(bs);
		return false;
	}

	ConfAccessLock.LockRead();

	X509 * currParentCert;
	ParentCerts = SKM_sk_new_null(X509);
	if(ParentCerts)
	{
		currParentCert = X509_dup(c_myConf->cas->RootCa);
		if(currParentCert)
		{
			if(SKM_sk_push(X509, ParentCerts, currParentCert) < 0)
			{
				X509_free(currParentCert);
			}
		}
		currParentCert = X509_dup(c_myConf->cas->OcspCa);
		if(currParentCert)
		{
			if(SKM_sk_push(X509, ParentCerts, currParentCert) < 0)
			{
				X509_free(currParentCert);
			}
		}
	}

	if(OCSP_basic_sign(bs, m_OcspCert.GetX509(), (EVP_PKEY*)m_OcspKey.GetRsaKey(), EVP_sha1(), ParentCerts, 0) <= 0)
	{
		*response = OCSP_response_create(OCSP_RESPONSE_STATUS_INTERNALERROR, NULL);
		NEWPKIerr(OCSP_ERROR_TXT, ERROR_ABORT);
		OCSP_BASICRESP_free(bs);
		ConfAccessLock.Unlock();
		SKM_sk_pop_free(X509, ParentCerts, X509_free);
		return false;
	}
	SKM_sk_pop_free(X509, ParentCerts, X509_free);
	ConfAccessLock.Unlock();

	*response = OCSP_response_create(OCSP_RESPONSE_STATUS_SUCCESSFUL, bs);

	OCSP_BASICRESP_free(bs);
	return true;
}


void Entity_PUBLICATION::SetOcspServersHandler(OcspServers *OcspServersHandler)
{
	m_OcspServersHandler = OcspServersHandler;
	ConfAccessLock.LockRead();
	LoadOcsp();
	ConfAccessLock.Unlock();
}

void Entity_PUBLICATION::LoadOcsp()
{
	ServerConf Conf;
	mString Name;
	char * Err;
	char * strErr;

	if(!m_OcspServersHandler || !myConf || !m_Logging)
		return;

	if(!ASN1_INTEGER_GET(myConf->body->d.pub_conf->d.PUB_CONF_PTR->OcspEnabled))
	{
		if(m_OcspServersHandler->IsRunning(m_EntityName))
		{
			// Unload the server
			m_OcspServersHandler->StopServer(m_EntityName);
			m_Logging->LogMessage(LOG_STATUS_TYPE_SUCCESS, LOG_MESSAGE_TYPE_OCSP_STOP, 0, _sv("none"), LOG_NO_OBJECTID, m_EntityName.c_str());
		}
	}
	else
	{
		if(ASN1_STRING_GET(myConf->body->d.pub_conf->d.PUB_CONF_PTR->BindAddress))
			Conf.set_BindAddress(ASN1_INTEGER_GET(myConf->body->d.pub_conf->d.PUB_CONF_PTR->BindAddress));

		Conf.set_LocalPort(ASN1_INTEGER_GET(myConf->body->d.pub_conf->d.PUB_CONF_PTR->Port));
		Name.sprintf("OCSP - %s", m_EntityName.c_str());
		Conf.set_ServerName(Name);
		if(!m_OcspServersHandler->StartServer(m_EntityName, Conf))
		{
			if(! (Err = ERR_to_string()) )
				strErr= _sv("Unknown");
			else
				strErr = Err;
			m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_OCSP_START, 0, _sv("none"), LOG_NO_OBJECTID, m_EntityName.c_str(), strErr);
			if(Err)
				free(Err);
		}
		else
		{
			m_Logging->LogMessage(LOG_STATUS_TYPE_SUCCESS, LOG_MESSAGE_TYPE_OCSP_START, 0, _sv("none"), LOG_NO_OBJECTID, m_EntityName.c_str());
		}
	}
}

