/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/

// Entity_PUBLICATION.h: interface for the Entity_PUBLICATION class.
//
//////////////////////////////////////////////////////////////////////

#ifndef ENTITY_PUBLICATION_H
#define ENTITY_PUBLICATION_H


#include "Entity.h"
#include "Entity_PUBLICATION_ASN1.h"
#include "PublicationMethod.h"
#include "PubStore.h"
#include <PKI_P7B.h>
#include <openssl/ocsp.h>

#include "OcspServers.h"
#include <vector>
using namespace std;


/*!
	This class represents a Publication Entity
*/
class Entity_PUBLICATION : public Entity  
{
public:
	Entity_PUBLICATION(ENTITY_CONSTRUCTOR_PARAMETERS);
	virtual ~Entity_PUBLICATION();

	bool GetWaitingResponses(STACK_OF(STORED_NEWPKI_RESPONSE) * Responses);
	bool GetWaitingRequests(STACK_OF(CRYPTED_NEWPKI_REQUEST) * Requests);
	bool GetProcessedTransactionIds(TRANSACTION_IDS * transactionIDs);
	bool GetWaitingTransactionIds(TRANSACTION_IDS * transactionIDs);
	bool OnRequestSent(unsigned long RequestId, const ASN1_OCTET_STRING * transactionId, const CRYPTED_NEWPKI_REQUEST * req);
	bool OnResponseSent(const STORED_NEWPKI_RESPONSE * resp);
	bool OnNewResponses(const STORED_NEWPKI_RESPONSES * resps);
	bool OnNewRequests(const STORED_NEWPKI_REQUESTS * reqs);
	bool OnNewConf(const ENTITY_CONF_CRYPTED * newConf);

	bool Load();
	bool Create(const ENTITY_CREATION_DATAS * Params, ADMIN_RESPONSE_BODY * response);
	bool Upgrade(const char * Version);
	bool Init(const ENTITY_SIGNATURE_RESP * init_datas);
	bool ParseAdminCommand(ADMIN_RESPONSE_BODY * response, const PKI_CERT & ClientCert, const ADMIN_REQUEST * AdminRequest);
	void SpawnStop();

	/*! \brief This function returns the list of types of logs the entity generates.
	 *  \param cLogsType [OUT] The list.
	 */
	static void LogsTypeGet(vector<LOG_MESSAGE_TYPE> & cLogsType);

	/*! \brief This function returns the list of types of ACL the entity handles.
	 *  \param acl_list [OUT] The list.
	 */
	static void GetACL_List(STACK_OF(ASN1_INTEGER) *acl_list);

	/*! \brief This function is called to initialize the OCSP responder.
	 *  \param OcspServersHandler [IN] The OCSP servers provider.
	 */
	void SetOcspServersHandler(OcspServers * OcspServersHandler);

	/*! \brief This function is called when a new OCSP request arrives.
	 *  \param Ip [IN] The IP of the requester.
	 *  \param request [IN] The OCSP request.
	 *  \param response [OUT] The OCSP response.
	 *  \return true on success, false on failure.
	 */
	bool OnNewOCSP(const char * Ip, const OCSP_REQUEST *request, OCSP_RESPONSE **response);
private:
	void LoadOcsp();
	bool Private_OnNewOCSP(const char * Ip, const OCSP_REQUEST *request, OCSP_RESPONSE **response);
	X509 * GetIssuer(STACK_OF(X509) * ParentCerts, X509 * Cert, X509_CRL * Crl);
	void AddUnknownCAs(STACK_OF(X509) * CAs);
	bool GetOcspCertStatus(OCSP_CERTID *cid, int & OcspStatus, time_t & rev_date, mString & CaName, unsigned long & RequestSerial);
	void UnloadAllPub();
	bool LoginUser(UserHandle & hUser, int & UserType);
	void LogoutUser(const UserHandle & hUser);
	
	static bool Private_ParseAdminCommand(bool ExecuteCmd, Entity * me_this, vector<LOG_MESSAGE_TYPE> &  mLogsType, ADMIN_RESPONSE_BODY * response, const PKI_CERT & ClientCert, const ADMIN_REQUEST * AdminRequest, const UserHandle & hUser);

	bool ParseNewConf(const ENTITY_CONF_CRYPTED * newConf, const INTERNAL_PKI_CA * cas);
	bool PersonnalConfFromPEM(const mString & conf_pem);
	bool PersonnalConfToPem(mString & conf_pem);

	static void ThreadWorker(const NewpkiThread * Thread, void *param);
	bool ProceedWithRequest(WORKING_CA_REQUEST * req);
	
	bool STORED_NEWPKI_REQUEST_check(STORED_NEWPKI_REQUEST * req, mString & Sender);
	bool AddNewSTORED_NEWPKI_REQUEST(STORED_NEWPKI_REQUEST * req, mString & Sender);

	LOCAL_PUBLICATION_CONF * c_myConf;
	ENTITY_CONF * myConf;

	ReadersWriter ListPublicationsLock;
	vector<PublicationMethod*> ListPublications;
	PKI_CERT m_OcspCert;
	PKI_RSA m_OcspKey;

	PubStore PublicationStore;
	OcspServers * m_OcspServersHandler;

	NewpkiThread hThreadWorker;

#define MY_PUBLICATION_CONF_PTR body->d.conf
	DECLARE_COMMAND_PARSER(Entity_PUBLICATION);
};

#endif
