/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/

// Entity_RA.cpp: implementation of the Entity_RA class.
//
//////////////////////////////////////////////////////////////////////

#include "Entity_RA.h"
#include "svintl.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Entity_RA::Entity_RA(ENTITY_CONSTRUCTOR_PARAMETERS):Entity(ENTITY_CONSTRUCTOR_PARAM_PASSTHRU)
{
	myConf = NULL;
	c_myConf = NULL;

	LdapSynch.SetEventHandler(&RaProfilesStore);
}

Entity_RA::~Entity_RA()
{
	ConfAccessLock.LockWrite();
	if(myConf)
		ENTITY_CONF_free(myConf);
	if(c_myConf)
		LOCAL_RA_CONF_free(c_myConf);
	ConfAccessLock.Unlock();
}

bool Entity_RA::Create(const ENTITY_CREATION_DATAS * Params, ADMIN_RESPONSE_BODY * response)
{
	if(!Params)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	if(Params->type != ENTITY_TYPE_RA)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	ENTITY_CREATION_RESP * resp;

	if(!Params->d.entity_key)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	//We create the database
	if(!Private_CreateEntity(Params->d.entity_key, &RaProfilesStore, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	resp = ENTITY_CREATION_RESP_new();
	if(!resp)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		Destroy();
		return false;
	}
	resp->type = ENTITY_TYPE_RA;
	if(X509_PUBKEY_set(&resp->d.entity_pub_key, (EVP_PKEY*)m_EntityKey.GetRsaKey()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		Destroy();
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_CREATE_ENTITY, resp))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_RA::Load()
{
	if(!Private_CommonLoad(&RaProfilesStore))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!IsFullyInit())
	{
		return true;
	}

	LdapSynch.SetLogging(m_Logging);

	sk_X509_push(EntityParentCerts, X509_dup(c_myConf->cas->RootCa));
	sk_X509_push(EntityParentCerts, X509_dup(c_myConf->cas->EntitiesCa));

	//Start the conf synch
	if(!m_Jobs.StartConfSync())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	//Start the mailer
	if(!m_Jobs.StartMailer())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!m_Jobs.SetEntityParentsCert(EntityParentCerts))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//Load repository
	if(!RaProfilesStore.Load())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	RaProfilesStore.SetAclValidator(&AclValidator);

	return true;
}

bool Entity_RA::Init(const ENTITY_SIGNATURE_RESP * init_datas)
{
	if(IsFullyInit())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!init_datas)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	if(init_datas->body->type != ENTITY_TYPE_RA)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	if(!Private_InitEntity(init_datas->body->d.EntityCert, &myConf, init_datas, &RaProfilesStore))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!RaProfilesStore.Create())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_RA::PersonnalConfFromPEM(const mString &conf_pem)
{
	//Convert from PEM
	ASN1_ENCRYPT_SIGN * encrypt;
	encrypt = NULL;

	if(ASN1_ENCRYPT_SIGN_From_PEM(&encrypt, conf_pem.c_str()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//Decrypt the conf
	if(c_myConf)
	{
		LOCAL_RA_CONF_free(c_myConf);
		c_myConf = NULL;
	}

	if(LOCAL_RA_CONF_decrypt_verify(encrypt, &c_myConf, m_EntityKey.GetRsaKey(), m_EntityKey.GetRsaKey()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ASN1_ENCRYPT_SIGN_free(encrypt);
		return false;
	}
	ASN1_ENCRYPT_SIGN_free(encrypt);
	
	//Extract the conf
	if(myConf)
	{
		ENTITY_CONF_free(myConf);
		myConf = NULL;
	}
	if(!ExtractMyConf(c_myConf->conf, &myConf, c_myConf->cas, c_myConf->cas->PkiCert, false))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_RA::PersonnalConfToPem(mString & conf_pem)
{
	char * pem;
	ASN1_ENCRYPT_SIGN * crypted_conf;

	ConfAccessLock.LockRead();

	//Sign/Crypt the conf
	crypted_conf = ASN1_ENCRYPT_SIGN_new();
	if(!crypted_conf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ConfAccessLock.Unlock();
		return false;
	}

	if(LOCAL_RA_CONF_sign_encrypt(crypted_conf, c_myConf, m_EntityKey.GetRsaKey(), m_EntityKey.GetRsaKey(), EVP_sha1(), EVP_des_ede3_cbc()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ASN1_ENCRYPT_SIGN_free(crypted_conf);
		ConfAccessLock.Unlock();
		return false;
	}

	//Convert it to PEM
	pem = ASN1_ENCRYPT_SIGN_To_PEM(crypted_conf);
	ASN1_ENCRYPT_SIGN_free(crypted_conf);
	if(!pem)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	conf_pem = pem;
	free(pem);
	ConfAccessLock.Unlock();
	return true;
}

bool Entity_RA::LoginUser(UserHandle & hUser, int & UserType)
{
	if(!AclValidator.CanUserPerform(hUser.GetUserCert(), ACL_TYPE_AUTHENTICATE_ON_ENTITY))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	UserType = USER_TYPE_RA;
	return true;
}

void Entity_RA::LogoutUser(const UserHandle & hUser)
{
}

bool Entity_RA::ParseNewConf(const ENTITY_CONF_CRYPTED * newConf, const INTERNAL_PKI_CA * cas)
{
	//Should we initialize the local conf ?
	if(!c_myConf)
		c_myConf = LOCAL_RA_CONF_new();

	if(!c_myConf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	//Should we update the internal cas certs ?
	if(cas)
	{
		if(!INTERNAL_PKI_CA_cpy(&c_myConf->cas, cas))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}

	if(!m_Policies.From_POLICY_VALUE(myConf->body->d.ra_conf->d.RA_CONF_PTR->Policies))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//Save the updated conf
	if(!ENTITY_CONF_CRYPTED_cpy(&c_myConf->conf, newConf))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!RaProfilesStore.SetGroups(myConf->groups))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	RaProfilesStore.SetMinPasswdLen(ASN1_INTEGER_GET(myConf->body->d.ra_conf->d.RA_CONF_PTR->MinPwdLen));

	char * Err;
	char * StrErr;
	
	if(ASN1_BIT_STRING_get_bit(myConf->body->d.ra_conf->d.RA_CONF_PTR->Flags, ALLOW_LDAP_PROFILE_IMPORT))
	{
		if(!LdapClient.Connect(myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_server, myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_port, myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_username, myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_password, myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_base, myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_attr_name))
		{
			Err = ERR_to_string();
			if(Err)
				StrErr=Err;
			else
				StrErr=_sv("Unknown");
			m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_LDAP_CONNECTION, 0, ASN1_STRING_GET(myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_username), LOG_NO_OBJECTID, ASN1_STRING_GET(myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_server), StrErr);
			if(Err)
				free(Err);
		}
		else
		{
			m_Logging->LogMessage(LOG_STATUS_TYPE_SUCCESS, LOG_MESSAGE_TYPE_LDAP_CONNECTION, 0, ASN1_STRING_GET(myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_username), LOG_NO_OBJECTID, ASN1_STRING_GET(myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_server));
		}

		if(ASN1_BIT_STRING_get_bit(myConf->body->d.ra_conf->d.RA_CONF_PTR->Flags, ALLOW_LDAP_AUTO_SYNCH))
		{
			if(!LdapSynch.SetOptions(
				(ASN1_BIT_STRING_get_bit(myConf->body->d.ra_conf->d.RA_CONF_PTR->Flags, REMOVE_UNKNOWN_DN_FIELDS) == 1),
				myConf->body->d.ra_conf->d.RA_CONF_PTR->DnSpecs,
				m_Policies,
				ASN1_INTEGER_GET(myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_profiles_group),
				myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_filters))
			{
				Err = ERR_to_string();
				if(Err)
					StrErr=Err;
				else
					StrErr=_sv("Unknown");
				m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_LDAP_AUTO_SYNCH_START, 0, ASN1_STRING_GET(myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_username), LOG_NO_OBJECTID, ASN1_STRING_GET(myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_server), StrErr);
				if(Err)
					free(Err);
			}
			else
			{
				if(!LdapSynch.Start(myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_server, myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_port, myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_username, myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_password, myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_base, myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_attr_name))
				{
					Err = ERR_to_string();
					if(Err)
						StrErr=Err;
					else
						StrErr=_sv("Unknown");
					m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_LDAP_AUTO_SYNCH_START, 0, ASN1_STRING_GET(myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_username), LOG_NO_OBJECTID, ASN1_STRING_GET(myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_server), StrErr);
					if(Err)
						free(Err);
				}
				else
				{
					m_Logging->LogMessage(LOG_STATUS_TYPE_SUCCESS, LOG_MESSAGE_TYPE_LDAP_AUTO_SYNCH_START, 0, ASN1_STRING_GET(myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_username), LOG_NO_OBJECTID, ASN1_STRING_GET(myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_server));
				}
			}
		}
		else
		{
			LdapSynch.Stop();
		}
	}

	
	return true;
}

bool Entity_RA::Upgrade(const char * Version)
{
	return true;
}


void Entity_RA::GetACL_List(STACK_OF(ASN1_INTEGER) *acl_list)
{
	ASN1_INTEGER * acl_id;
	int i;
	static ACL_TYPE list_acls[] =
	{
		ACL_TYPE_VIEW_LOGS,
		ACL_TYPE_SEND_ADMIN_MAIL,
		ACL_TYPE_RA_IMPORT_PROFILE,
		ACL_TYPE_RA_DELEGATE_PROFILE_OWNERSHIP,
		ACL_TYPE_RA_REQUEST_CERT,
		ACL_TYPE_RA_REVOKE_CERT,
		(ACL_TYPE)0
	};
	for(i=0; list_acls[i]; i++)
	{
		acl_id = NULL;
		ASN1_INTEGER_SET(&acl_id, list_acls[i]);
		if(acl_id)
		{
			if(SKM_sk_push(ASN1_INTEGER, acl_list, acl_id) < 0)
			{
				ASN1_INTEGER_free(acl_id);
			}
		}
	}	
}


bool Entity_RA::ParseAdminCommand(ADMIN_RESPONSE_BODY * response, const PKI_CERT & ClientCert, const ADMIN_REQUEST * AdminRequest)
{
	//We only accept SSLv3 connection	
	if(!ClientCert)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ERR_to_ADMIN_RESPONSE(response);
		return false;
	}
	return Private_ParseAdminCommand(true, this, LogsType, response, ClientCert, AdminRequest, GetUserHandle());
}

bool Entity_RA::Private_ParseAdminCommand(bool ExecuteCmd, Entity * me_this, vector<LOG_MESSAGE_TYPE> &  mLogsType, ADMIN_RESPONSE_BODY * response, const PKI_CERT & ClientCert, const ADMIN_REQUEST * AdminRequest, const UserHandle & hUser)
{
	PARSER_COMMAND_BEGIN(Entity_RA, response, 0, AdminRequest, ClientCert, hUser, ExecuteCmd, me_this, mLogsType)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_ENTITY_GET_MY_CONF)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_SEND_ADMIN_MAIL)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_SEND_MAIL)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_CERT_SIG_RESP)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_CERT_REV_RESP)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_CERT_RNW_RESP)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_LDAP_AUTO_SYNCH_START)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_LDAP_AUTO_SYNCH_IMPRT)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_LDAP_CONNECTION)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_LOGIN,					Entity_RA::UserLogin,				LOG_MESSAGE_TYPE_USER_LOGIN, (ClientCert)?(char*)ClientCert.GetStringName():NULL, LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_ENUM_LOGS,				Entity_RA::EnumLogs)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_LOGS_COUNT,			Entity_RA::GetLogsCount)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_MY_ACL,				Entity_RA::GetMyACL)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_LOGS_TYPE,			Entity_RA::GetLogsType)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_SEND_ADMIN_MAIL,			Entity_RA::AdminSendMail,			LOG_MESSAGE_TYPE_ADD_ADMIN_MAIL_QUEUE,	_sv("none"), LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_CHECK_LOGS,				Entity_RA::CheckLogsIntegrity)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_GET_LOCAL_CONF,			Entity_RA::GetLocalConf,			LOG_MESSAGE_TYPE_GET_LOCAL_CONF,	_sv("none"), LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_GROUPS,				Entity_RA::GetUsersGroups)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_IMPORT_PROFILE,			Entity_RA::ImportProfile,			LOG_MESSAGE_TYPE_IMPORT_PROFILE,	_sv("none"), LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_ENUM_PROFILES,			Entity_RA::EnumProfiles)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_REQUEST_CERT,			Entity_RA::RequestCert,				LOG_MESSAGE_TYPE_CERT_REQUEST,	NULL, ASN1_INTEGER_GET(body->d.cert_request->profile_id))
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_REVOKE_CERT,				Entity_RA::RevokeCert,				LOG_MESSAGE_TYPE_CERT_REVOCATION,	NULL, ASN1_INTEGER_GET(body->d.serial))
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_DELETE_PKCS12,			Entity_RA::DeletePkcs12,			LOG_MESSAGE_TYPE_DELETE_PKCS12,	NULL, ASN1_INTEGER_GET(body->d.serial))
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_SEARCH_LDAP,				Entity_RA::SearchLdap)
	PARSER_COMMAND_END(Entity_RA)
}

void Entity_RA::LogsTypeGet(vector<LOG_MESSAGE_TYPE> &  cLogsType)
{
	Private_ParseAdminCommand(false, NULL, cLogsType, NULL, PKI_CERT(), NULL, UserHandle());
}


void Entity_RA::SpawnStop()
{
	m_Jobs.SpawnStop();
	LdapSynch.SpawnStop();
}


bool Entity_RA::GetLocalConf(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	ENTITY_CONF * conf;
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(AclValidator.ValidateCert(UserCert) != INTERNAL_CA_TYPE_USER)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	ConfAccessLock.LockRead();
	conf = ENTITY_CONF_cpy(NULL, myConf);
	if(!conf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();

	// We don't send the LDAP info, those are very private datas
	ASN1_STRING_SET(&conf->body->d.ra_conf->d.RA_CONF_PTR->ldap_server, "");
	ASN1_STRING_SET(&conf->body->d.ra_conf->d.RA_CONF_PTR->ldap_username, "");
	ASN1_STRING_SET(&conf->body->d.ra_conf->d.RA_CONF_PTR->ldap_password, "");
	ASN1_STRING_SET(&conf->body->d.ra_conf->d.RA_CONF_PTR->ldap_base, "");
	ASN1_INTEGER_SET(&conf->body->d.ra_conf->d.RA_CONF_PTR->ldap_port, 0);
	ASN1_INTEGER_SET(&conf->body->d.ra_conf->d.RA_CONF_PTR->ldap_profiles_group, 0);
	ASN1_STRING_SET(&conf->body->d.ra_conf->d.RA_CONF_PTR->ldap_filters, "");
	ASN1_STRING_SET(&conf->body->d.ra_conf->d.RA_CONF_PTR->ldap_attr_name, "");

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_LOCAL_ENTITY_CONF, conf))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_RA::GetUsersGroups(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(AclValidator.ValidateCert(UserCert) != INTERNAL_CA_TYPE_USER)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	ConfAccessLock.LockRead();
	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_GROUPS, USERS_GROUPS_cpy(NULL, myConf->groups)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}


bool Entity_RA::ImportProfile(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_RA_IMPORT_PROFILE))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	// Did the user delegate ownership of profile,
	// and is he allowed to ?
	if(ASN1_INTEGER_GET(body->d.profile->owner_group_serial) &&  
		!AclValidator.CanUserPerform(UserCert, ACL_TYPE_RA_DELEGATE_PROFILE_OWNERSHIP))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	long ProfileId;
	ASN1_INTEGER * Id = NULL;
	HashTable_Dn Dn;

	if(!Dn.From_X509_NAME(body->d.profile->dn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//Let's check the policy
	ConfAccessLock.LockRead();
	if(!Dn.ValidateAgainstPolicy(m_Policies))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();

	// Make sure we have an email address for this profile
	// so the EE can send the email
	if(ASN1_INTEGER_GET(body->d.profile->EE_Validation))
	{
		if(Dn.SeekEntryName("emailAddress", -1) == HASHTABLE_NOT_FOUND)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_DN_FIELD_UNPRESENT);
			ERR_set_error_data("emailAddress", ERR_TXT_STRING);
			return false;
		}
	}

	// TODO: Reorder the DN !

	if(!RaProfilesStore.InsertProfile(UserCert, body->d.profile, ProfileId))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	// If we require user validation on this profile
	// then we must send the profile to the EE
	if(ASN1_INTEGER_GET(body->d.profile->EE_Validation))
	{

		// TODO: Crer une requete de publication du profil sur l'EE
/*
		if((!SendEE_Req())
		{
		}
		  
*/
	}
	
	ASN1_INTEGER_SET(&Id, ProfileId);

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_PROFILE_ID, Id))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}


bool Entity_RA::EnumProfiles(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	STACK_OF(NEWPKI_PROFILE_DATAS) * profiles;
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_RA_IMPORT_PROFILE))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	profiles = RaProfilesStore.EnumProfiles(UserCert, ASN1_INTEGER_GET(body->d.enum_objects->index), ASN1_INTEGER_GET(body->d.enum_objects->num));
	if(!profiles)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_PROFILES, profiles))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}


bool Entity_RA::SearchLdap(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	mString Filters;
	STACK_OF(LDAP_RESULT) * Results;
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_RA_IMPORT_PROFILE))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!LdapClient)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_LDAP_NOT_CONFIGURED);
		return false;
	}

	ConfAccessLock.LockRead();
	if(ASN1_STRING_GET(myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_filters))
	{
		Filters.sprintf("(&%s%s)", ASN1_STRING_GET(myConf->body->d.ra_conf->d.RA_CONF_PTR->ldap_filters), ASN1_STRING_GET(body->d.ldap_search));
	}
	else
	{
		Filters = body->d.ldap_search;
	}
	ConfAccessLock.Unlock();

	Results = LdapClient.Search(Filters);
	if(!Results)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_LDAP_RESULTS, Results))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}


bool Entity_RA::RequestCert(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_RA_REQUEST_CERT))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	mString ttDatas;
	PKI_CSR req;
	PKI_RSA privKey;
	int keylen;
	NEWPKI_PROFILE_DATAS * profile;
	NEWPKI_REQUEST * Request;
	CA_ENTRY_INFO * currCa;
	int i;
	unsigned long CertReqId;
	HashTable_Dn Dn;
	RA_STORE_PRIVDATAS * PrivDatas = NULL;
	X509_PUBKEY * recipient=NULL;
	const char * admin_mail;

	// Is the type set ?
	if(!ASN1_INTEGER_GET(body->d.cert_request->type))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}

	ConfAccessLock.LockRead();
	// Are we allowed to request certs from this CA ?
	for(i=0; i < SKM_sk_num(CA_ENTRY_INFO, myConf->Cas->list); i++)
	{
		currCa = SKM_sk_value(CA_ENTRY_INFO, myConf->Cas->list, i);
		if(!currCa) continue;

		if(ASN1_STRING_cmp(currCa->Name, body->d.cert_request->ca_name) == 0)
		{
			recipient = X509_PUBKEY_cpy(NULL, currCa->CaSSL->cert_info->key);
			if(!recipient)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				ConfAccessLock.Unlock();
				return false;
			}
			break;
		}
	}
	// Did we find the CA ?
	if(i == SKM_sk_num(CA_ENTRY_INFO, myConf->Cas->list))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}


	
		
	// Get admin mail
	i = UserCert.GetCertDN().SeekEntryName("emailAddress", HASHTABLE_NOT_FOUND);
	if(i == HASHTABLE_NOT_FOUND)
	{
		admin_mail = "";
	}
	else
	{
		admin_mail = UserCert.GetCertDN().Get(i);
		if(!admin_mail)
			admin_mail = "";
	}
	if(ASN1_BIT_STRING_get_bit(body->d.cert_request->flags, REQUEST_CERT_FLAG_SEND_ADMIN_NOTIFICATION))
	{
		if(!*admin_mail)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			ERR_set_error_data(_sv("emailAddress not found in admin certificate"), ERR_TXT_STRING);
			X509_PUBKEY_free(recipient);
			ConfAccessLock.Unlock();
			return false;
		}
		if(!m_Jobs.GetMailer())
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_MAIL_SERVER_NOT_SET);
			X509_PUBKEY_free(recipient);
			ConfAccessLock.Unlock();
			return false;
		}
	}
	
	if((ASN1_BIT_STRING_get_bit(body->d.cert_request->flags, REQUEST_CERT_FLAG_SEND_CERT_TO_USER) ||
		ASN1_BIT_STRING_get_bit(body->d.cert_request->flags, REQUEST_CERT_FLAG_RA_GENERATE_P12_PASSWD) ||
		ASN1_BIT_STRING_get_bit(body->d.cert_request->flags, REQUEST_CERT_FLAG_SEND_P12_PASSWD_TO_USER)) &&
		!m_Jobs.GetMailer())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MAIL_SERVER_NOT_SET);
		X509_PUBKEY_free(recipient);
		ConfAccessLock.Unlock();
		return false;
	}
	

	// We now verify if the request respects the policy constraints
	switch(body->d.cert_request->request->type)
	{
		case REQUEST_CERT_BODY_PKCS10:
			if(!ASN1_BIT_STRING_get_bit(myConf->body->d.ra_conf->d.RA_CONF_PTR->Flags, ALLOW_PKCS10_IMPORT))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
				X509_PUBKEY_free(recipient);
				ConfAccessLock.Unlock();
				return false;
			}
			if(!req.SetCSR(body->d.cert_request->request->d.pkcs10))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				X509_PUBKEY_free(recipient);
				ConfAccessLock.Unlock();
				return false;
			}
			keylen = EVP_PKEY_size((EVP_PKEY*)req.GetPublicKey())*8;
			break;
		case REQUEST_CERT_BODY_PKCS12:
			if(!ASN1_BIT_STRING_get_bit(myConf->body->d.ra_conf->d.RA_CONF_PTR->Flags, ALLOW_SERVER_SIDE_RA_KEY_GEN))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
				X509_PUBKEY_free(recipient);
				ConfAccessLock.Unlock();
				return false;
			}

			// If we're supposed to generate the PKCS#12 paswword we don't
			// do a length verification
			if(!ASN1_BIT_STRING_get_bit(body->d.cert_request->flags, REQUEST_CERT_FLAG_RA_GENERATE_P12_PASSWD))
			{
				ttDatas = body->d.cert_request->request->d.p12_privdatas->passwd;
				//Let's verify the password length
				if(ttDatas.size() < (unsigned int)ASN1_INTEGER_GET(myConf->body->d.ra_conf->d.RA_CONF_PTR->MinPwdLen) )
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PASSWORD_LEN);
					ttDatas = myConf->body->d.ra_conf->d.RA_CONF_PTR->MinPwdLen;
					ERR_set_error_data(strdup(ttDatas.c_str()), ERR_TXT_MALLOCED | ERR_TXT_STRING);
					X509_PUBKEY_free(recipient);
					ConfAccessLock.Unlock();
					return false;
				}
			}
			keylen = ASN1_INTEGER_GET(body->d.cert_request->request->d.p12_privdatas->keylen);
			break;
		default:
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			X509_PUBKEY_free(recipient);
			ConfAccessLock.Unlock();
			return false;
	}
	//Let's verify the key length
	if(keylen < ASN1_INTEGER_GET(myConf->body->d.ra_conf->d.RA_CONF_PTR->MinKeyLen))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_KEY_LEN);
		ttDatas = myConf->body->d.ra_conf->d.RA_CONF_PTR->MinKeyLen;
		ERR_set_error_data(strdup(ttDatas.c_str()), ERR_TXT_MALLOCED | ERR_TXT_STRING);
		X509_PUBKEY_free(recipient);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();

	
	//Let's get the profile
	profile = RaProfilesStore.GetProfile(UserCert, ASN1_INTEGER_GET(body->d.cert_request->profile_id));
	if(!profile)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		X509_PUBKEY_free(recipient);
		return false;
	}
	// Is the profile ready for signature ?
	if(ASN1_INTEGER_GET(profile->State) != NEWPKI_PROFILE_STATE_VALIDATED)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		NEWPKI_PROFILE_DATAS_free(profile);
		X509_PUBKEY_free(recipient);
		return false;
	}

	// Are we supposed to send the certificate to the user ?
	// REQUEST_CERT_FLAG_RA_GENERATE_P12_PASSWD means as well that 
	// the password is automatically sent to the user !
	if(ASN1_BIT_STRING_get_bit(body->d.cert_request->flags, REQUEST_CERT_FLAG_SEND_CERT_TO_USER) ||
		ASN1_BIT_STRING_get_bit(body->d.cert_request->flags, REQUEST_CERT_FLAG_RA_GENERATE_P12_PASSWD) ||
		ASN1_BIT_STRING_get_bit(body->d.cert_request->flags, REQUEST_CERT_FLAG_SEND_P12_PASSWD_TO_USER))
	{
		// Does the user's DN have a emailAddress field ?
		if(!Dn.From_X509_NAME(profile->Profile->dn))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			NEWPKI_PROFILE_DATAS_free(profile);
			X509_PUBKEY_free(recipient);
			return false;
		}
		if(Dn.SeekEntryName("emailAddress", HASHTABLE_NOT_FOUND) == HASHTABLE_NOT_FOUND)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			ERR_set_error_data(_sv("emailAddress not found in user's DN"), ERR_TXT_STRING);
			NEWPKI_PROFILE_DATAS_free(profile);
			X509_PUBKEY_free(recipient);
			return false;
		}
	}


	switch(body->d.cert_request->request->type)
	{
		case REQUEST_CERT_BODY_PKCS10:
			PrivDatas = NULL;
			break;

		case REQUEST_CERT_BODY_PKCS12:
			// Let's generate the PKCS#10
			if(!Dn.From_X509_NAME(profile->Profile->dn))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				NEWPKI_PROFILE_DATAS_free(profile);
				X509_PUBKEY_free(recipient);
				return false;
			}
			if(!privKey.GenerateKey(ASN1_INTEGER_GET(body->d.cert_request->request->d.p12_privdatas->keylen), m_Engine))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				NEWPKI_PROFILE_DATAS_free(profile);
				X509_PUBKEY_free(recipient);
				return false;
			}
			if(!req.GenerateCSR(Dn, privKey))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				NEWPKI_PROFILE_DATAS_free(profile);
				X509_PUBKEY_free(recipient);
				return false;
			}
			PrivDatas = RA_STORE_PRIVDATAS_new();
			if(!PrivDatas)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				NEWPKI_PROFILE_DATAS_free(profile);
				X509_PUBKEY_free(recipient);
				return false;
			}

			// Copying the password
			if(!ASN1_BIT_STRING_get_bit(body->d.cert_request->flags, REQUEST_CERT_FLAG_RA_GENERATE_P12_PASSWD))
			{
				ASN1_STRING_SET(&PrivDatas->passwd, ASN1_STRING_GET(body->d.cert_request->request->d.p12_privdatas->passwd));
			}
			// Copying the private key 
			RSA_free(PrivDatas->privkey);
			PrivDatas->privkey = privKey.GetRSA(true);
			if(!PrivDatas->privkey)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
				NEWPKI_PROFILE_DATAS_free(profile);
				RA_STORE_PRIVDATAS_free(PrivDatas);
				X509_PUBKEY_free(recipient);
				return false;
			}
			break;
	}


	// We get the LDAP RDN !
	ttDatas = profile->Profile->ldap_uid;

	NEWPKI_PROFILE_DATAS_free(profile);

	// Insert the request in the store, and get the id
	if(!RaProfilesStore.InsertCertReq(ASN1_STRING_GET(body->d.cert_request->ca_name), ASN1_INTEGER_GET(body->d.cert_request->profile_id), ASN1_INTEGER_GET(body->d.cert_request->type), req, PrivDatas, body->d.cert_request->flags, admin_mail, CertReqId))
	{
		if(PrivDatas) RA_STORE_PRIVDATAS_free(PrivDatas);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		X509_PUBKEY_free(recipient);
		return false;
	}
	if(PrivDatas)
		RA_STORE_PRIVDATAS_free(PrivDatas);
	
	
	// Generate the requests
	Request = NEWPKI_REQUEST_new();
	if(!Request)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		X509_PUBKEY_free(recipient);
		RaProfilesStore.DeleteCertReq(CertReqId);
		return false;
	}
	Request->type = NEWPKI_REQUEST_TYPE_CERT;
	Request->d.cert_request = NEWPKI_CERT_REQUEST_new();
	if(!Request->d.cert_request)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		NEWPKI_REQUEST_free(Request);
		X509_PUBKEY_free(recipient);
		RaProfilesStore.DeleteCertReq(CertReqId);
		return false;
	}

	// We set the LDAP RDN
	ASN1_STRING_SET(&Request->d.cert_request->ldap_uid, ttDatas.c_str());
	ttDatas="";

	// We set the request id
	ASN1_INTEGER_SET(&Request->d.cert_request->id, CertReqId);

	// We copy the PKCS#10
	X509_REQ_free(Request->d.cert_request->request);
	Request->d.cert_request->request = req.GetX509_REQ(true);
	if(!Request->d.cert_request->request)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NEWPKI_REQUEST_free(Request);
		X509_PUBKEY_free(recipient);
		RaProfilesStore.DeleteCertReq(CertReqId);
		return false;
	}
	// We set the validity len
	ASN1_INTEGER_SET(&Request->d.cert_request->validity, ASN1_INTEGER_GET(body->d.cert_request->validity));

	
	
	ASN1_INTEGER * Id = NULL;
	ASN1_INTEGER_SET(&Id, CertReqId);
	if(!Id)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		X509_PUBKEY_free(recipient);
		NEWPKI_REQUEST_free(Request);
		RaProfilesStore.DeleteCertReq(CertReqId);
		return false;
	}
	
	if(!SendRequest(CertReqId, Request, recipient))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		X509_PUBKEY_free(recipient);
		NEWPKI_REQUEST_free(Request);
		RaProfilesStore.DeleteCertReq(CertReqId);
		return false;
	}
	
	NEWPKI_REQUEST_free(Request);
	X509_PUBKEY_free(recipient);
	
	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_REQ_ID, Id))
	{
		ASN1_INTEGER_free(Id);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		RaProfilesStore.DeleteCertReq(CertReqId);
		return false;
	}

	return true;
}


bool Entity_RA::OnNewConf(const ENTITY_CONF_CRYPTED * newConf)
{
	if(!ExtractMyConf(newConf, &myConf, NULL, c_myConf->cas->PkiCert, true))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_RA::OnNewRequests(const STORED_NEWPKI_REQUESTS * reqs)
{
	return true;
}

bool Entity_RA::OnRequestSent(unsigned long RequestId, const ASN1_OCTET_STRING * transactionId, const CRYPTED_NEWPKI_REQUEST * req)
{
	RaProfilesStore.Requester_RequestSent(RequestId, transactionId);
	return true;
}


bool Entity_RA::OnResponseSent(const STORED_NEWPKI_RESPONSE * resp)
{
	return true;
}


bool Entity_RA::OnNewResponses(const STORED_NEWPKI_RESPONSES * resps)
{
	STORED_NEWPKI_RESPONSE * currResp;
	NEWPKI_RESPONSE * resp;
	EVP_PKEY * signerkey;
	NEWPKI_PROFILE_DATAS_CERT * cert;
	CA_ENTRY_INFO * currCa;
	int i, j;
	mString ca_name;

	for(i=0; i<SKM_sk_num(STORED_NEWPKI_RESPONSE, resps->responses); i++)
	{
		currResp = SKM_sk_value(STORED_NEWPKI_RESPONSE, resps->responses, i);
		if(!currResp) continue;

		signerkey = X509_PUBKEY_get(currResp->c_info->sender);
		if(!signerkey) continue;

		if(!NEWPKI_RESPONSE_decrypt_verify(currResp->c_info->c_response, &resp, signerkey, m_EntityKey.GetRsaKey()))
		{
			EVP_PKEY_free(signerkey);
			continue;
		}
		EVP_PKEY_free(signerkey);

		cert = RaProfilesStore.GetCertReq(currResp->transactionID);
		if(!cert)
		{
			NEWPKI_RESPONSE_free(resp);
			continue;
		}
		
		ConfAccessLock.LockRead();
		// Now search the CA, to make sure it is the real signer !
		for(j=0; j < SKM_sk_num(CA_ENTRY_INFO, myConf->Cas->list); j++)
		{
			currCa = SKM_sk_value(CA_ENTRY_INFO, myConf->Cas->list, j);
			if(!currCa) continue;

			// Is it the CA ?
			if(ASN1_STRING_cmp(currCa->Name, cert->ca_name) != 0) continue;

			// Are the public keys the same ?
			if(currCa->CaSSL->cert_info->key->public_key->length ==
				currResp->c_info->sender->public_key->length &&
				memcmp(currCa->CaSSL->cert_info->key->public_key->data,
				currResp->c_info->sender->public_key->data, 
				currResp->c_info->sender->public_key->length) == 0)
			{
				// Yep
				ca_name = currCa->Name;
				RaProfilesStore.InsertResponse(ca_name, currResp->transactionID, resp, m_Jobs, m_Logging);
			}
			break;
		}
		ConfAccessLock.Unlock();
		NEWPKI_RESPONSE_free(resp);
		NEWPKI_PROFILE_DATAS_CERT_free(cert);
	}

	return true;
}


bool Entity_RA::GetWaitingTransactionIds(TRANSACTION_IDS *transactionIDs)
{
	return RaProfilesStore.Requester_GetWaitingTransactionIds(transactionIDs);
}

bool Entity_RA::GetProcessedTransactionIds(TRANSACTION_IDS *transactionIDs)
{
	// The RA never gets requests from other entities
	return false;
}

bool Entity_RA::RevokeCert(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_RA_REVOKE_CERT))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	mString ttDatas;
	PKI_CSR req;
	PKI_RSA privKey;
	NEWPKI_REQUEST * Request;
	CA_ENTRY_INFO * currCa;
	int i;
	HashTable_Dn Dn;
	X509_PUBKEY * recipient=NULL;

	unsigned long cert_serial;
	mString ca_name;
	mString ldap_uid;



	// Suspend the cert and get usefull info
	if(!RaProfilesStore.SuspendCertificate(UserCert, ASN1_INTEGER_GET(body->d.serial), cert_serial, ca_name, ldap_uid))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	ConfAccessLock.LockRead();
	// Are we allowed to request certs from this CA ?
	for(i=0; i < SKM_sk_num(CA_ENTRY_INFO, myConf->Cas->list); i++)
	{
		currCa = SKM_sk_value(CA_ENTRY_INFO, myConf->Cas->list, i);
		if(!currCa) continue;

		if(ca_name == currCa->Name)
		{
			recipient = X509_PUBKEY_cpy(NULL, currCa->CaSSL->cert_info->key);
			if(!recipient)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				ConfAccessLock.Unlock();
				return false;
			}
			break;
		}
	}
	// Did we find the CA ?
	if(i == SKM_sk_num(CA_ENTRY_INFO, myConf->Cas->list))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	
	// Generate the request
	Request = NEWPKI_REQUEST_new();
	if(!Request)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		X509_PUBKEY_free(recipient);
		return false;
	}
	Request->type = NEWPKI_REQUEST_TYPE_REV;
	Request->d.rev_request = NEWPKI_REV_REQUEST_new();
	if(!Request->d.rev_request)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		NEWPKI_REQUEST_free(Request);
		X509_PUBKEY_free(recipient);
		return false;
	}

	ASN1_STRING_SET(&Request->d.rev_request->ldap_uid, ldap_uid.c_str());

	ASN1_INTEGER_SET(&Request->d.rev_request->id, ASN1_INTEGER_GET(body->d.serial));
	ASN1_INTEGER_SET(&Request->d.rev_request->reason, 0);
	ASN1_INTEGER_SET(&Request->d.rev_request->serial, cert_serial);


	if(!SendRequest(ASN1_INTEGER_GET(body->d.serial), Request, recipient))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NEWPKI_REQUEST_free(Request);
		X509_PUBKEY_free(recipient);
		return false;
	}
	NEWPKI_REQUEST_free(Request);
	X509_PUBKEY_free(recipient);

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_RA::DeletePkcs12(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_RA_REQUEST_CERT))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!RaProfilesStore.DeletePKCS12(UserCert, ASN1_INTEGER_GET(body->d.serial)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_RA::SendRequest(unsigned long CertReqId, NEWPKI_REQUEST * Request, X509_PUBKEY * recipient)
{
	CRYPTED_NEWPKI_REQUEST * c_Request;
	EVP_PKEY * CryptKey;
	unsigned long req_id;
	
	c_Request = CRYPTED_NEWPKI_REQUEST_new();
	if(!c_Request)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	if(!X509_PUBKEY_cpy(&c_Request->recipient, recipient))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		CRYPTED_NEWPKI_REQUEST_free(c_Request);
		return false;
	}
	if(!X509_PUBKEY_cpy(&c_Request->sender, m_EntityCert.GetX509_PUBKEY()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		CRYPTED_NEWPKI_REQUEST_free(c_Request);
		return false;
	}

	CryptKey = X509_PUBKEY_get(c_Request->recipient);
	if(!CryptKey)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		CRYPTED_NEWPKI_REQUEST_free(c_Request);
		return false;
	}
	if(NEWPKI_REQUEST_sign_encrypt(c_Request->c_request, Request, m_EntityCert.GetPrivateKey().GetRsaKey(), CryptKey, EVP_sha1(), EVP_des_ede3_cbc()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		CRYPTED_NEWPKI_REQUEST_free(c_Request);
		EVP_PKEY_free(CryptKey);
		return false;
	}
	EVP_PKEY_free(CryptKey);
	
	if(!RaProfilesStore.Requester_InsertRequest(c_Request, CertReqId, req_id))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		CRYPTED_NEWPKI_REQUEST_free(c_Request);
		return false;
	}
	
	if(!m_Jobs.InsertRequest(req_id, c_Request))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		CRYPTED_NEWPKI_REQUEST_free(c_Request);
		RaProfilesStore.Requester_DeleteTransaction(req_id);
		return false;
	}
	CRYPTED_NEWPKI_REQUEST_free(c_Request);
	
	return true;
}

bool Entity_RA::GetWaitingRequests(STACK_OF(CRYPTED_NEWPKI_REQUEST) * Requests)
{
	return RaProfilesStore.Requester_GetWaitingRequests(Requests);
}

bool Entity_RA::GetWaitingResponses(STACK_OF(CRYPTED_NEWPKI_RESPONSE) * Responses)
{
	return true;
}

