/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/



// Entity_REPOSITORY.cpp: implementation of the Entity_REPOSITORY class.
//
//////////////////////////////////////////////////////////////////////

#include "Entity_REPOSITORY.h"
#include "svintl.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


Entity_REPOSITORY::Entity_REPOSITORY(ENTITY_CONSTRUCTOR_PARAMETERS):Entity(ENTITY_CONSTRUCTOR_PARAM_PASSTHRU)
{
	myConf = NULL;
	c_myConf = NULL;
	pki_conf = NULL;
	WaitingObjects = NULL;
	hThreadCaObjectsPush.Create(ThreadCaObjectsPush, this);

	WaitingObjects = SKM_sk_new_null(WAITING_NEWPKI_OBJECT);
	if(!WaitingObjects)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		throw ExceptionNewPKI();
	}
}

Entity_REPOSITORY::~Entity_REPOSITORY()
{
	hThreadCaObjectsPush.Stop();


	ConfAccessLock.LockWrite();
	if(myConf)
		ENTITY_CONF_free(myConf);
	if(pki_conf)
		EXPORTED_PKI_CONF_free(pki_conf);
	if(c_myConf)
		LOCAL_REPOSITORY_CONF_free(c_myConf);
	if(WaitingObjects)
		SKM_sk_pop_free(WAITING_NEWPKI_OBJECT, WaitingObjects, WAITING_NEWPKI_OBJECT_free);
	ConfAccessLock.Unlock();
}

bool Entity_REPOSITORY::Create(const ENTITY_CREATION_DATAS * Params, ADMIN_RESPONSE_BODY * response)
{
	if(!Params)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	if(Params->type != ENTITY_TYPE_REPOSITORY)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	ENTITY_CREATION_RESP * resp;

	if(!Params->d.entity_key)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}


	const char * InitialCreates[]={REP_ENTITY_CREATE_1, NULL};

	//We create the database
	if(!Private_CreateEntity(Params->d.entity_key, &RepositoryStore, InitialCreates))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}


	resp = ENTITY_CREATION_RESP_new();
	if(!resp)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		Destroy();
		return false;
	}
	resp->type = ENTITY_TYPE_REPOSITORY;
	if(X509_PUBKEY_set(&resp->d.entity_pub_key, (EVP_PKEY*)m_EntityKey.GetRsaKey()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		Destroy();
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_CREATE_ENTITY, resp))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_REPOSITORY::Init(const ENTITY_SIGNATURE_RESP * init_datas)
{
	if(IsFullyInit())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!init_datas)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	if(init_datas->body->type != ENTITY_TYPE_REPOSITORY)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	if(!Private_InitEntity(init_datas->body->d.EntityCert, &myConf, init_datas, &RepositoryStore))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!RepositoryStore.Create())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_REPOSITORY::Load()
{
	if(!Private_CommonLoad(&RepositoryStore))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!IsFullyInit())
	{
		return true;
	}

	sk_X509_push(EntityParentCerts, X509_dup(c_myConf->cas->RootCa));
	sk_X509_push(EntityParentCerts, X509_dup(c_myConf->cas->EntitiesCa));
	
	SQL sql(m_DbConn);
	mString req;
	mString pem_conf;
	long NumRows;

	// We get the PKI conf
	if(!sql.Execute(REP_ENTITY_GET_G_CONF))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(NumRows)
	{
		if(!sql.Value(0, "conf", pem_conf))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}

		if(!EXPORTED_PKI_CONF_From_PEM(&pki_conf, pem_conf.c_str()))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}

	if(!RepositoryStore.Load())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!hThreadCaObjectsPush.Start())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}

	//Start the mailer
	if(!m_Jobs.StartMailer())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!m_Jobs.SetEntityParentsCert(EntityParentCerts))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//Start the conf synch
	if(!m_Jobs.StartConfPush(true))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(pki_conf)
	{
		// Synch the conf
		ConfAccessLock.LockRead();
		if(!m_Jobs.SetPkiConf(pki_conf))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ConfAccessLock.Unlock();
			return false;
		}
		ConfAccessLock.Unlock();
	}

	return true;
}


bool Entity_REPOSITORY::PersonnalConfFromPEM(const mString &conf_pem)
{
	//Convert from PEM
	ASN1_ENCRYPT_SIGN * encrypt;
	encrypt = NULL;

	if(ASN1_ENCRYPT_SIGN_From_PEM(&encrypt, conf_pem.c_str()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//Decrypt the conf
	if(c_myConf)
	{
		LOCAL_REPOSITORY_CONF_free(c_myConf);
		c_myConf = NULL;
	}

	if(LOCAL_REPOSITORY_CONF_decrypt_verify(encrypt, &c_myConf, m_EntityKey.GetRsaKey(), m_EntityKey.GetRsaKey()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ASN1_ENCRYPT_SIGN_free(encrypt);
		return false;
	}
	ASN1_ENCRYPT_SIGN_free(encrypt);
	
	//Extract the conf
	if(myConf)
	{
		ENTITY_CONF_free(myConf);
		myConf = NULL;
	}
	if(!ExtractMyConf(c_myConf->conf, &myConf, c_myConf->cas, c_myConf->cas->PkiCert, false))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_REPOSITORY::PersonnalConfToPem(mString & conf_pem)
{
	char * pem;
	ASN1_ENCRYPT_SIGN * crypted_conf;

	ConfAccessLock.LockRead();
	//Sign/Crypt the conf
	crypted_conf = ASN1_ENCRYPT_SIGN_new();
	if(!crypted_conf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ConfAccessLock.Unlock();
		return false;
	}

	if(LOCAL_REPOSITORY_CONF_sign_encrypt(crypted_conf, c_myConf, m_EntityKey.GetRsaKey(), m_EntityKey.GetRsaKey(), EVP_sha1(), EVP_des_ede3_cbc()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ASN1_ENCRYPT_SIGN_free(crypted_conf);
		ConfAccessLock.Unlock();
		return false;
	}

	//Convert it to PEM
	pem = ASN1_ENCRYPT_SIGN_To_PEM(crypted_conf);
	ASN1_ENCRYPT_SIGN_free(crypted_conf);
	if(!pem)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	conf_pem = pem;
	free(pem);
	ConfAccessLock.Unlock();
	return true;
}


bool Entity_REPOSITORY::ParseNewConf(const ENTITY_CONF_CRYPTED * newConf, const INTERNAL_PKI_CA * cas)
{
	//Should we initialize the local conf ?
	if(!c_myConf)
		c_myConf = LOCAL_REPOSITORY_CONF_new();

	if(!c_myConf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	//Should we update the internal cas certs ?
	if(cas)
	{
		if(!INTERNAL_PKI_CA_cpy(&c_myConf->cas, cas))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}

	//Save the updated conf
	if(!ENTITY_CONF_CRYPTED_cpy(&c_myConf->conf, newConf))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

void Entity_REPOSITORY::ThreadCaObjectsPush(const NewpkiThread * Thread, void *param)
{
	Entity_REPOSITORY * me_this = (Entity_REPOSITORY *)param;
	time_t currTime;
	time_t lastTime;
	WAITING_NEWPKI_OBJECT * WaitingObject = NULL;
	int i;
	int j;
	REP_ENTRY_INFO * currRep;
	PkiClient ClientPki(NULL);
	mString err;
	STACK_OF(WAITING_NEWPKI_OBJECT) * localObjects = NULL;
	REPS_INFO * Repositories = NULL;
	map<mString, time_t> RepositoriesSynchroTime;
	X509_PUBKEY * pubKey;

	
	//We wait a bit (2 mins)
	time(&lastTime);
	do
	{
		NewpkiThread::Sleep(500);
		time(&currTime);
	}
	while( ((currTime - lastTime) < 120) && !Thread->ShouldStop());

	
	time(&lastTime);
	while(!Thread->ShouldStop())
	{
		me_this->ConfAccessLock.LockWrite();
		//Do we have the repositories list ?
		if(!me_this->c_myConf || !me_this->myConf)
		{
			me_this->ConfAccessLock.Unlock();
			goto wait_next;
		}
		//We copy the repositories not to block the rest of the program
		//while accessing it, since we don't know if the
		//remote host won't block us
		if(!(Repositories = REPS_INFO_cpy(NULL, me_this->myConf->Repositories)))
		{
			me_this->ConfAccessLock.Unlock();
			goto wait_next;
		}
		me_this->ConfAccessLock.Unlock();


		// Copy the waiting objects
		me_this->WaitingObjectsLock.LockWrite();
		if(!(localObjects = SKM_sk_new_null(WAITING_NEWPKI_OBJECT)))
		{
			me_this->WaitingObjectsLock.Unlock();
			goto wait_next;
		}

		while( (WaitingObject = SKM_sk_pop(WAITING_NEWPKI_OBJECT, me_this->WaitingObjects)) )
		{
			// If we fail to add ot to our list, put back to the orginal list
			// if the insertion fails again, destroy the object, we will
			// synchronize it later during a full synchro
			if(SKM_sk_push(WAITING_NEWPKI_OBJECT, localObjects, WaitingObject) < 0 && 
				SKM_sk_push(WAITING_NEWPKI_OBJECT, me_this->WaitingObjects, WaitingObject) < 0)
			{
				WAITING_NEWPKI_OBJECT_free(WaitingObject);
			}
		}
		me_this->WaitingObjectsLock.Unlock();

		//For each repository we do a full synchro
		for(i=0; i<SKM_sk_num(REP_ENTRY_INFO, Repositories->list); i++)
		{
			if(Thread->ShouldStop())
				break;

			currRep = SKM_sk_value(REP_ENTRY_INFO, Repositories->list, i);
			if(!currRep) continue;

			//If the current rep is me or is firewalled we ignore it
			if(me_this->m_EntityCert == currRep->RepositorySSL ||
				ASN1_BIT_STRING_get_bit(currRep->Flags, REP_ENTRY_INFO_FLAG_FIREWALLED))
				continue;

			// We synchronize if the repository has never been synchronized or
			// if the repository was synchronized more than 2 hours ago
			time(&currTime);
			if(RepositoriesSynchroTime.find(currRep->Name) == RepositoriesSynchroTime.end() ||
				(currTime - RepositoriesSynchroTime[currRep->Name]) >= 7200)
			{
				// Do synchro
				if(me_this->SynchronizeWithRepository(&ClientPki, currRep))
					RepositoriesSynchroTime[currRep->Name] = currTime;
			}
			else
			{
				if(ASN1_BIT_STRING_get_bit(currRep->Flags, REP_ENTRY_INFO_FLAG_SHOULD_FULL_SYNCHRO))
				{
					// We always do a full synchro, this allows us to get its new objects
					me_this->SynchronizeWithRepository(&ClientPki, currRep);
				}
				else
				{
					// We just push the new object
					for(j=0; j<SKM_sk_num(WAITING_NEWPKI_OBJECT, localObjects); j++)
					{
						if(Thread->ShouldStop())
							break;

						WaitingObject = SKM_sk_value(WAITING_NEWPKI_OBJECT, localObjects, j);
						if(!WaitingObject) continue;

						// We ignore the object if the repository already had it
						if(me_this->RepositoryKnowsObject(WaitingObject, currRep->RepositorySSL))
							continue;

						// Connect to repository
						if(!me_this->m_Jobs.ConnectToRepository(currRep, &ClientPki, 30, err))
						{
							NewpkiDebug(LOG_LEVEL_WARNING, me_this->m_EntityName.c_str(), _sv("Failed to push NewPKI object on repository %s (%s:%ld)\nReason:%s"), ASN1_STRING_GET(currRep->Name), ASN1_STRING_GET(currRep->Address), ASN1_INTEGER_GET(currRep->Port), err.c_str());
							me_this->m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_OBJECTS_PUSH, 0, NULL, LOG_NO_OBJECTID, ASN1_STRING_GET(currRep->Name), err.c_str());
							continue;
						}
						
						if(!ClientPki.PushWaitingObject(WaitingObject))
						{			
							NewpkiDebug(LOG_LEVEL_WARNING, me_this->m_EntityName.c_str(), _sv("Failed to push NewPKI object on repository %s (%s:%ld)\nReason:%s"), ASN1_STRING_GET(currRep->Name), ASN1_STRING_GET(currRep->Address), ASN1_INTEGER_GET(currRep->Port), ClientPki.GetError());
							me_this->m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_OBJECTS_PUSH, 0, NULL, LOG_NO_OBJECTID, ASN1_STRING_GET(currRep->Name), ClientPki.GetError());
							continue;
						}

						NewpkiDebug(LOG_LEVEL_INFO, me_this->m_EntityName.c_str(), _sv("Successfully pushed NewPKI object on repository %s (%s:%ld)"), ASN1_STRING_GET(currRep->Name), ASN1_STRING_GET(currRep->Address), ASN1_INTEGER_GET(currRep->Port));
						me_this->m_Logging->LogMessage(LOG_STATUS_TYPE_SUCCESS, LOG_MESSAGE_TYPE_OBJECTS_PUSH, 0, NULL, LOG_NO_OBJECTID, ASN1_STRING_GET(currRep->Name));

						//Add this repository to the path
						pubKey = X509_PUBKEY_cpy(NULL, currRep->RepositorySSL->cert_info->key);
						if(pubKey)
						{
							if(SKM_sk_push(X509_PUBKEY, WaitingObject->rep_path, pubKey) < 0)
							{
								X509_PUBKEY_free(pubKey);
							}
						}						
					}
				}
			}
		}

wait_next:
		if(Repositories)
		{
			REPS_INFO_free(Repositories);
			Repositories = NULL;
		}
		if(localObjects)
		{
			SKM_sk_pop_free(WAITING_NEWPKI_OBJECT, localObjects, WAITING_NEWPKI_OBJECT_free);
			localObjects = NULL;
		}
		do
		{
			NewpkiThread::Sleep(500);
			time(&currTime);
		}
		while( ((currTime - lastTime) < 60) && !Thread->ShouldStop());
		lastTime = currTime;
	}
}


bool Entity_REPOSITORY::ParseAdminCommand(ADMIN_RESPONSE_BODY * response, const PKI_CERT & ClientCert, const ADMIN_REQUEST * AdminRequest)
{
	//We only accept SSLv3 connection	
	if(!ClientCert)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		ERR_to_ADMIN_RESPONSE(response);
		return false;
	}
	
	return Private_ParseAdminCommand(true, this, LogsType, response, ClientCert, AdminRequest, UserHandle());
}

bool Entity_REPOSITORY::Private_ParseAdminCommand(bool ExecuteCmd, Entity * me_this, vector<LOG_MESSAGE_TYPE> &  mLogsType, ADMIN_RESPONSE_BODY * response, const PKI_CERT & ClientCert, const ADMIN_REQUEST * AdminRequest, const UserHandle & hUser)
{
	int pos;
	const char * cn = NULL;


	if(ClientCert && AdminRequest)
	{
		switch(AdminRequest->body->type)
		{
			case ADMIN_REQ_TYPE_GET_MY_REQS:
			case ADMIN_REQ_TYPE_GET_MY_RESPS:
			case ADMIN_REQ_TYPE_PUSH_REQ:
			case ADMIN_REQ_TYPE_PUSH_RESP:
				if(((pos = ClientCert.GetCertDN().SeekEntryName("commonName", -1)) != HASHTABLE_NOT_FOUND) )
					cn = ClientCert.GetCertDN().Get(pos);
				break;

			default:
				break;
		}
	}
	PARSER_COMMAND_BEGIN(Entity_REPOSITORY, response, 0, AdminRequest, ClientCert, hUser, ExecuteCmd, me_this, mLogsType)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_CONFIG_PUSH)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_SEND_ADMIN_MAIL)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_SEND_MAIL)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_LOGIN,					Entity_REPOSITORY::UserLogin,				LOG_MESSAGE_TYPE_USER_LOGIN, (ClientCert)?(char*)ClientCert.GetStringName():NULL, LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_PUSH_CONFIG,				Entity_REPOSITORY::PushConfiguration,		LOG_MESSAGE_TYPE_CONFIG_IMPORT, NULL, ASN1_INTEGER_GET(AdminRequest->body->d.conf->confs->version))
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_GET_MY_CONFIG,			Entity_REPOSITORY::GetMyConf,				LOG_MESSAGE_TYPE_ENTITY_CONFIG_GET, (ClientCert)?(char*)ClientCert.GetStringName():NULL, LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_ENUM_LOGS,				Entity_REPOSITORY::EnumLogs)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_LOGS_COUNT,			Entity_REPOSITORY::GetLogsCount)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_LOGS_TYPE,			Entity_REPOSITORY::GetLogsType)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_PUSH_NEWPKI_OBJECTS,		Entity_REPOSITORY::PushWaitingObject,		LOG_MESSAGE_TYPE_OBJECTS_IMPORT, _sv("none"), LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_GET_MY_REQS,				Entity_REPOSITORY::GetMyRequests,			LOG_MESSAGE_TYPE_GET_MY_REQS, (char*)cn, LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_GET_MY_RESPS,			Entity_REPOSITORY::GetMyResponses,			LOG_MESSAGE_TYPE_GET_MY_RESPS, (char*)cn, LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_CHECK_LOGS,				Entity_REPOSITORY::CheckLogsIntegrity)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_PUSH_REQ,				Entity_REPOSITORY::ImportRequest,			LOG_MESSAGE_TYPE_INSERT_REQ, (char*)cn, LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_PUSH_RESP,				Entity_REPOSITORY::ImportResponse,			LOG_MESSAGE_TYPE_INSERT_RESP, (char*)cn, LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_SYNCH_OBJECTS,			Entity_REPOSITORY::SynchronizeObjects)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_GET_CONFIG,				Entity_REPOSITORY::GetConfiguration,		LOG_MESSAGE_TYPE_GET_CONFIGURATION, _sv("none"), LOG_NO_OBJECTID)
	PARSER_COMMAND_END(Entity_REPOSITORY)
}

void Entity_REPOSITORY::LogsTypeGet(vector<LOG_MESSAGE_TYPE> &  cLogsType)
{
	Private_ParseAdminCommand(false, NULL, cLogsType, NULL, PKI_CERT(), NULL, UserHandle());
}

bool Entity_REPOSITORY::PushConfiguration(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	// Figure out what kind of certificate it is:
	// At this point we can get:
	//	- the PKI entity certificate
	//	- A user certificate
	//	- An entity certificate
	// If its a user certificate we have to check
	// the acls to see if he can do this action

	switch(AclValidator.ValidateCert(UserCert))
	{
		//The PKI entity or another entity
		case INTERNAL_CA_TYPE_ENTITY:
		case 0:
			break;

		//A user
		case INTERNAL_CA_TYPE_USER:
			if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_IMPORT_CONF, false))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
				return false;
			}
			break;

		//Not allowed
		default:
			NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
			return false;
			break;
	}


	if(!OnNewPkiConf(body->d.conf))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	return true;
}

bool Entity_REPOSITORY::GetMyConf(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	int i;
	ENTITY_CONF_CRYPTED * currConf;
	
	
	//Is this a valid entity cert ?
	if(AclValidator.ValidateCert(UserCert) != INTERNAL_CA_TYPE_ENTITY)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	ConfAccessLock.LockRead();

	if(pki_conf)
	{
		//Find entity conf
		for(i=0; i<SKM_sk_num(ENTITY_CONF_CRYPTED, pki_conf->confs->all_confs); i++)
		{
			currConf = SKM_sk_value(ENTITY_CONF_CRYPTED, pki_conf->confs->all_confs, i);
			if(!currConf) continue;

			if( UserCert == currConf->recipient)
			{
				currConf = ENTITY_CONF_CRYPTED_cpy(NULL, currConf);
				if(!currConf)
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
					ConfAccessLock.Unlock();
					return false;				
				}

				if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_MY_CONF, currConf))
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					ConfAccessLock.Unlock();
					return false;
				}

				ConfAccessLock.Unlock();
				return true;
			}
		}

		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN_ENTITY);
		ConfAccessLock.Unlock();
		return false;
	}
	
	NEWPKIerr(PKI_ERROR_TXT, ERROR_ENTITY_NOT_READY);
	ConfAccessLock.Unlock();
	return false;
}

bool Entity_REPOSITORY::LoginUser(UserHandle & hUser, int & UserType)
{
	switch(AclValidator.ValidateCert(hUser.GetUserCert()))
	{
		case 0:	//The PKI Entity Cert
		case INTERNAL_CA_TYPE_ENTITY: // Another entity
			break;

		case INTERNAL_CA_TYPE_USER:	//A user
			if(!AclValidator.CanUserPerform(hUser.GetUserCert(), ACL_TYPE_AUTHENTICATE_ON_ENTITY, false))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
				return false;
			}
			break;

		default:
			NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
			return false;
	}
	UserType = USER_TYPE_REPOSITORY;
	return true;
}

void Entity_REPOSITORY::LogoutUser(const UserHandle & hUser)
{
}

void Entity_REPOSITORY::GetACL_List(STACK_OF(ASN1_INTEGER) *acl_list)
{
	ASN1_INTEGER * acl_id;
	int i;
	static ACL_TYPE list_acls[] =
	{
		ACL_TYPE_VIEW_LOGS,
		ACL_TYPE_IMPORT_CONF, 
		ACL_TYPE_IMPORT_NEWPKI_OBJECTS, 
		(ACL_TYPE)0
	};
	for(i=0; list_acls[i]; i++)
	{
		acl_id = NULL;
		ASN1_INTEGER_SET(&acl_id, list_acls[i]);
		if(acl_id)
		{
			if(SKM_sk_push(ASN1_INTEGER, acl_list, acl_id) < 0)
			{
				ASN1_INTEGER_free(acl_id);
			}
		}
	}	
}

bool Entity_REPOSITORY::Upgrade(const char * Version)
{
	return true;
}

bool Entity_REPOSITORY::PushWaitingObject(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	// Figure out what kind of certificate it is:
	// At this point we can get:
	//	- A user certificate
	//	- An entity certificate
	// If its a user certificate we have to check
	// the acls to see if he can do this action

	switch(AclValidator.ValidateCert(UserCert))
	{
		//The PKI entity or another entity
		case INTERNAL_CA_TYPE_ENTITY:
			break;

		case INTERNAL_CA_TYPE_USER:
			if(!AclValidator.CanUserPerform(UserCert, ACL_TYPE_IMPORT_NEWPKI_OBJECTS))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
				return false;
			}
			break;
		//Not allowed
		case 0:
		default:
			NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
			return false;
			break;
	}

	if(!OnNewObject(body->d.waiting_obj))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	return true;
}


bool Entity_REPOSITORY::GetMyRequests(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	STORED_NEWPKI_REQUESTS * reqs;

	// Figure out what kind of certificate it is:
	// At this point we can only get an entity certificate

	switch(AclValidator.ValidateCert(UserCert))
	{
		//The PKI entity or another entity
		case INTERNAL_CA_TYPE_ENTITY:
			break;
		//Not allowed
		case 0:
		default:
			NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
			return false;
			break;
	}

	//Get the requests from the RepStore
	reqs = RepositoryStore.GetRequests(UserCert.GetPublicKey(), body->d.transactions_ids);
	if(!reqs)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_REQS, reqs))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	return true;
}

bool Entity_REPOSITORY::GetMyResponses(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	STORED_NEWPKI_RESPONSES * resps;

	// Figure out what kind of certificate it is:
	// At this point we can only get an entity certificate

	switch(AclValidator.ValidateCert(UserCert))
	{
		//The PKI entity or another entity
		case INTERNAL_CA_TYPE_ENTITY:
			break;
		//Not allowed
		case 0:
		default:
			NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
			return false;
			break;
	}

	//Get the responses from the RepStore
	resps = RepositoryStore.GetResponses(UserCert.GetPublicKey(), body->d.transactions_ids);
	if(!resps)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_RESPS, resps))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	return true;
}

bool Entity_REPOSITORY::ImportRequest(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	ASN1_OCTET_STRING * transactionId;
	STORED_NEWPKI_REQUEST * s_req;

	// Figure out what kind of certificate it is:
	// At this point we can only get an entity certificates
	if(AclValidator.ValidateCert(UserCert) != INTERNAL_CA_TYPE_ENTITY)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	//Insert the request in the RepStore
	s_req = RepositoryStore.PublishRequest(body->d.new_req);
	if(!s_req)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	// Save transaction id
	transactionId = ASN1_OCTET_STRING_cpy(NULL, s_req->transactionID);
	if(!transactionId)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!CreateNewObject(s_req, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ASN1_OCTET_STRING_free(transactionId);
		return false;
	}
	
	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_TRANSACTION_ID, transactionId))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	return true;
}

bool Entity_REPOSITORY::ImportResponse(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	STORED_NEWPKI_RESPONSE  *resp;
	bool exists;

	// Figure out what kind of certificate it is:
	// At this point we can only get an entity certificates
	if(AclValidator.ValidateCert(UserCert) != INTERNAL_CA_TYPE_ENTITY)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	resp = STORED_NEWPKI_RESPONSE_cpy(NULL, body->d.new_resp);
	if(!resp)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//Insert the response in the RepStore
	if(!RepositoryStore.InsertResponse(resp, exists))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		STORED_NEWPKI_RESPONSE_free(resp);
		return false;
	}

	if(exists)
	{
		return true;
	}

	if(!CreateNewObject(NULL, resp))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
		
	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	return true;
}

void Entity_REPOSITORY::SpawnStop()
{
	hThreadCaObjectsPush.SpawStop();
	m_Jobs.SpawnStop();
}



bool Entity_REPOSITORY::OnNewConf(const ENTITY_CONF_CRYPTED * newConf)
{
	return true;
}

bool Entity_REPOSITORY::OnNewRequests(const STORED_NEWPKI_REQUESTS * reqs)
{
	return true;
}

bool Entity_REPOSITORY::OnRequestSent(unsigned long RequestId, const ASN1_OCTET_STRING * transactionId, const CRYPTED_NEWPKI_REQUEST * req)
{
	return true;
}


bool Entity_REPOSITORY::OnResponseSent(const STORED_NEWPKI_RESPONSE * resp)
{
	return true;
}

bool Entity_REPOSITORY::OnNewResponses(const STORED_NEWPKI_RESPONSES * resps)
{
	return true;
}

bool Entity_REPOSITORY::GetWaitingTransactionIds(TRANSACTION_IDS *transactionIDs)
{
	return false;
}

bool Entity_REPOSITORY::GetProcessedTransactionIds(TRANSACTION_IDS *transactionIDs)
{
	return false;
}

bool Entity_REPOSITORY::SynchronizeWithRepository(PkiClient * ClientPki, const REP_ENTRY_INFO * Repository)
{
	OBJECTS_LIST_REQUEST * mylist;
	OBJECTS_LIST_RESPONSE * known_list;
	STORED_NEWPKI_REQUESTS * reqs;
	STORED_NEWPKI_RESPONSES * resps;
	STORED_NEWPKI_REQUEST * currReq;
	STORED_NEWPKI_RESPONSE * currResp;
	WAITING_NEWPKI_OBJECT * object;
	mString err;
	EXPORTED_PKI_CONF * conf;

	mylist = OBJECTS_LIST_REQUEST_new();
	if(!mylist)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	if(!RepositoryStore.GetKnownRequests(mylist->requests))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		OBJECTS_LIST_REQUEST_free(mylist);
		return false;
	}
	if(!RepositoryStore.GetKnownResponses(mylist->responses))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		OBJECTS_LIST_REQUEST_free(mylist);
		return false;
	}

	// We send it the list of requests and responses we know, 
	// it will return us a list of objects
	// it doesn't know, along with the objects we don't known, 
	
	//We connect to the repository
	if(!m_Jobs.ConnectToRepository(Repository, ClientPki, 30, err))
	{
		NewpkiDebug(LOG_LEVEL_WARNING, m_EntityName.c_str(), _sv("Failed to synchronize objects on respository %s (%s:%ld)\nReason:%s"), ASN1_STRING_GET(Repository->Name), ASN1_STRING_GET(Repository->Address), ASN1_INTEGER_GET(Repository->Port), err.c_str());
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		OBJECTS_LIST_REQUEST_free(mylist);
		return false;
	}
	if(!ClientPki->SynchObjects(mylist, &known_list))
	{			
		NewpkiDebug(LOG_LEVEL_WARNING, m_EntityName.c_str(), _sv("Failed to synchronize objects on respository %s (%s:%ld)\nReason:%s"), ASN1_STRING_GET(Repository->Name), ASN1_STRING_GET(Repository->Address), ASN1_INTEGER_GET(Repository->Port), ClientPki->GetError());
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		OBJECTS_LIST_REQUEST_free(mylist);
		return false;
	}
	OBJECTS_LIST_REQUEST_free(mylist);

	// The repository sent my a list of requests I didn't know
	// I need to integrate them
	while( (object = SKM_sk_pop(WAITING_NEWPKI_OBJECT, known_list->objects)) )
	{
		if(!OnNewObject(object))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			WAITING_NEWPKI_OBJECT_free(object);
			OBJECTS_LIST_RESPONSE_free(known_list);
			return false;
		}
		WAITING_NEWPKI_OBJECT_free(object);
	}

	//Get the requests from our store, the repository doesn' know
	reqs = RepositoryStore.GetRequests(NULL, known_list->requests);
	if(!reqs)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		OBJECTS_LIST_RESPONSE_free(known_list);
		return false;
	}
	while( (currReq = SKM_sk_pop(STORED_NEWPKI_REQUEST, reqs->requests)) )
	{
		object = CreateWaitingObject(currReq, NULL);
		if(!object) continue;
		if(!ClientPki->PushWaitingObject(object))
		{
			NewpkiDebug(LOG_LEVEL_WARNING, m_EntityName.c_str(), _sv("Failed to synchronize with respository %s (%s:%ld)\nReason:%s"), ASN1_STRING_GET(Repository->Name), ASN1_STRING_GET(Repository->Address), ASN1_INTEGER_GET(Repository->Port), ClientPki->GetError());
		}
		WAITING_NEWPKI_OBJECT_free(object);
	}
	STORED_NEWPKI_REQUESTS_free(reqs);

	//Get the responses from our store, the repository doesn' know
	resps = RepositoryStore.GetResponses(NULL, known_list->responses, false);
	if(!resps)
	{
		OBJECTS_LIST_RESPONSE_free(known_list);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		OBJECTS_LIST_REQUEST_free(mylist);
		return false;
	}
	while( (currResp = SKM_sk_pop(STORED_NEWPKI_RESPONSE, resps->responses)) )
	{
		object = CreateWaitingObject(NULL, currResp);
		if(!object) continue;
		if(!ClientPki->PushWaitingObject(object))
		{
			NewpkiDebug(LOG_LEVEL_WARNING, m_EntityName.c_str(), _sv("Failed to synchronize with respository %s (%s:%ld)\nReason:%s"), ASN1_STRING_GET(Repository->Name), ASN1_STRING_GET(Repository->Address), ASN1_INTEGER_GET(Repository->Port), ClientPki->GetError());
		}
		WAITING_NEWPKI_OBJECT_free(object);
	}
	STORED_NEWPKI_RESPONSES_free(resps);
	OBJECTS_LIST_RESPONSE_free(known_list);

	// Now we're going to get its PKI conf
	conf = NULL;
	if(!ClientPki->GetConfiguration(&conf))
	{
		NewpkiDebug(LOG_LEVEL_WARNING, m_EntityName.c_str(), _sv("Failed to synchronize with respository %s (%s:%ld)\nReason:%s"), ASN1_STRING_GET(Repository->Name), ASN1_STRING_GET(Repository->Address), ASN1_INTEGER_GET(Repository->Port), ClientPki->GetError());
		return false;
	}
	if(!OnNewPkiConf(conf))
	{
		NewpkiDebug(LOG_LEVEL_WARNING, m_EntityName.c_str(), _sv("Failed to synchronize with respository %s (%s:%ld)\nReason:%s"), ASN1_STRING_GET(Repository->Name), ASN1_STRING_GET(Repository->Address), ASN1_INTEGER_GET(Repository->Port), ClientPki->GetError());
		EXPORTED_PKI_CONF_free(conf);
		return false;
	}
	EXPORTED_PKI_CONF_free(conf);

	NewpkiDebug(LOG_LEVEL_INFO, m_EntityName.c_str(), _sv("Successfully synchronized with respository %s (%s:%ld)"), ASN1_STRING_GET(Repository->Name), ASN1_STRING_GET(Repository->Address), ASN1_INTEGER_GET(Repository->Port));

	ClientPki->CloseConnection();

	return true;
}

bool Entity_REPOSITORY::CreateNewObject(STORED_NEWPKI_REQUEST *request, STORED_NEWPKI_RESPONSE *response)
{
	ConfAccessLock.LockWrite();
	WAITING_NEWPKI_OBJECT * object;

	// Set the synchro struct
	object = CreateWaitingObject(request, response);
	if(!object)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	//Add it to the list
	if(SKM_sk_push(WAITING_NEWPKI_OBJECT, WaitingObjects, object) < 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		WAITING_NEWPKI_OBJECT_free(object);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();

	return true;
}

WAITING_NEWPKI_OBJECT * Entity_REPOSITORY::CreateWaitingObject(STORED_NEWPKI_REQUEST *request, STORED_NEWPKI_RESPONSE *response)
{
	WAITING_NEWPKI_OBJECT * object;
	X509_PUBKEY * myPubKey;
	object = WAITING_NEWPKI_OBJECT_new();
	if(!object)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		if(request) STORED_NEWPKI_REQUEST_free(request);
		if(response) STORED_NEWPKI_RESPONSE_free(response);
		return NULL;
	}

	if(request)
	{
		object->object->type = NEWPKI_OBJECTS_TYPE_REQ;
		object->object->d.request = request;
	}
	else if(response)
	{
		object->object->type = NEWPKI_OBJECTS_TYPE_RESP;
		object->object->d.response = response;
	}
	else
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		WAITING_NEWPKI_OBJECT_free(object);
		if(request) STORED_NEWPKI_REQUEST_free(request);
		if(response) STORED_NEWPKI_RESPONSE_free(response);
		return NULL;
	}

	myPubKey = X509_PUBKEY_cpy(NULL, m_EntityKey.GetPublicKey());
	if(!myPubKey)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		WAITING_NEWPKI_OBJECT_free(object);
		return NULL;
	}
	if(SKM_sk_push(X509_PUBKEY, object->rep_path, myPubKey) < 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		X509_PUBKEY_free(myPubKey);
		WAITING_NEWPKI_OBJECT_free(object);
		return NULL;
	}
	
	return object;
}


bool Entity_REPOSITORY::SynchronizeObjects(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	int i;
	ASN1_OCTET_STRING * currTID;
	OBJECTS_LIST_RESPONSE * objects_resp;
	STORED_NEWPKI_REQUESTS * reqs;
	STORED_NEWPKI_RESPONSES * resps;
	STORED_NEWPKI_REQUEST * currReq;
	STORED_NEWPKI_RESPONSE * currResp;
	WAITING_NEWPKI_OBJECT * object;
	bool Exists;

	// Figure out what kind of certificate it is
	if(AclValidator.ValidateCert(UserCert) != INTERNAL_CA_TYPE_ENTITY)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	objects_resp = OBJECTS_LIST_RESPONSE_new();
	if(!objects_resp)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	// We now generate the list of requests the requesting
	// repository doesn't know of
	reqs = RepositoryStore.GetRequests(NULL, body->d.known_objects->requests);
	if(!reqs)
	{
		OBJECTS_LIST_RESPONSE_free(objects_resp);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	while( (currReq = SKM_sk_pop(STORED_NEWPKI_REQUEST, reqs->requests)) )
	{
		object = CreateWaitingObject(currReq, NULL);
		if(!object) continue;
		if(SKM_sk_push(WAITING_NEWPKI_OBJECT, objects_resp->objects, object) < 0)
		{
			WAITING_NEWPKI_OBJECT_free(object);
		}
	}
	STORED_NEWPKI_REQUESTS_free(reqs);

	// We now generate the list of responses the requesting
	// repository doesn't know of
	resps = RepositoryStore.GetResponses(NULL, body->d.known_objects->responses, false);
	if(!resps)
	{
		OBJECTS_LIST_RESPONSE_free(objects_resp);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	while( (currResp = SKM_sk_pop(STORED_NEWPKI_RESPONSE, resps->responses)) )
	{
		object = CreateWaitingObject(NULL, currResp);
		if(!object) continue;
		if(SKM_sk_push(WAITING_NEWPKI_OBJECT, objects_resp->objects, object) < 0)
		{
			WAITING_NEWPKI_OBJECT_free(object);
		}
	}
	STORED_NEWPKI_RESPONSES_free(resps);

	
	// Add to the response the list of objects we know of
	for(i=0; i<SKM_sk_num(ASN1_OCTET_STRING, body->d.known_objects->requests->transactionIds); i++)
	{
		currTID = SKM_sk_value(ASN1_OCTET_STRING, body->d.known_objects->requests->transactionIds, i);
		if(!currTID) continue;

		if(RepositoryStore.RequestExists(currTID, Exists) && Exists)
		{
			if( (currTID = ASN1_OCTET_STRING_cpy(NULL, currTID)) )
			{
				if(SKM_sk_push(ASN1_OCTET_STRING, objects_resp->requests->transactionIds, currTID) < 0)
				{
					ASN1_OCTET_STRING_free(currTID);
				}
			}
		}
	}	

	// Add to the response the list of objects we know of
	for(i=0; i<SKM_sk_num(ASN1_OCTET_STRING, body->d.known_objects->responses->transactionIds); i++)
	{
		currTID = SKM_sk_value(ASN1_OCTET_STRING, body->d.known_objects->responses->transactionIds, i);
		if(!currTID) continue;

		if(RepositoryStore.ResponseExists(currTID, Exists) && Exists)
		{
			if( (currTID = ASN1_OCTET_STRING_cpy(NULL, currTID)) )
			{
				if(SKM_sk_push(ASN1_OCTET_STRING, objects_resp->responses->transactionIds, currTID) < 0)
				{
					ASN1_OCTET_STRING_free(currTID);
				}
			}
		}
	}

	// We add to the list of known requests/responses the ones 
	// the peer doesn't known of
	for(i=0; i<SKM_sk_num(WAITING_NEWPKI_OBJECT, objects_resp->objects); i++)
	{
		object = SKM_sk_value(WAITING_NEWPKI_OBJECT, objects_resp->objects, i);
		if(!object) continue;

		if(object->object->type == NEWPKI_OBJECTS_TYPE_REQ)
		{
			if( (currTID = ASN1_OCTET_STRING_cpy(NULL, object->object->d.request->transactionID)) )
			{
				if(SKM_sk_push(ASN1_OCTET_STRING, objects_resp->requests->transactionIds, currTID) < 0)
				{
					ASN1_OCTET_STRING_free(currTID);
				}
			}
		}
		else if(object->object->type == NEWPKI_OBJECTS_TYPE_RESP)
		{
			if( (currTID = ASN1_OCTET_STRING_cpy(NULL, object->object->d.response->transactionID)) )
			{
				if(SKM_sk_push(ASN1_OCTET_STRING, objects_resp->responses->transactionIds, currTID) < 0)
				{
					ASN1_OCTET_STRING_free(currTID);
				}
			}
		}
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_KNOWN_OBJECTS, objects_resp))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		OBJECTS_LIST_RESPONSE_free(objects_resp);
		return false;
	}
	
	return true;
}

bool Entity_REPOSITORY::GetWaitingRequests(STACK_OF(CRYPTED_NEWPKI_REQUEST) * Requests)
{
	return true;
}

bool Entity_REPOSITORY::GetWaitingResponses(STACK_OF(CRYPTED_NEWPKI_RESPONSE) * Responses)
{
	return true;
}

bool Entity_REPOSITORY::OnNewObject(WAITING_NEWPKI_OBJECT *object)
{
	X509_PUBKEY * currPubkey;
	bool ret;
	bool Exists;

	// Verify that the NewPKI object hasn't already 
	// been pushed on me
	if(RepositoryKnowsObject(object, m_EntityCert.GetX509()))
	{
		return true;
	}

	//Insert new obects to repository
	switch(object->object->type)
	{
		case NEWPKI_OBJECTS_TYPE_REQ:
			ret = RepositoryStore.InsertRequest(object->object->d.request, Exists);
			break;
		case NEWPKI_OBJECTS_TYPE_RESP:
			ret = RepositoryStore.InsertResponse(object->object->d.response, Exists);
			break;
		default:
			NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
			return false;
			break;
	}
	if(!ret)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(Exists)	// We ignore it, if it already exists
	{
		return true;
	}

	object = WAITING_NEWPKI_OBJECT_cpy(NULL, object);
	if(!object)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//Add my-self to the rep_path
	currPubkey = X509_PUBKEY_cpy(NULL, m_EntityCert.GetX509_PUBKEY());
	if(!currPubkey)
	{
		WAITING_NEWPKI_OBJECT_free(object);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(SKM_sk_push(X509_PUBKEY, object->rep_path, currPubkey) < 0)
	{
		X509_PUBKEY_free(currPubkey);
		WAITING_NEWPKI_OBJECT_free(object);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	WaitingObjectsLock.LockWrite();
	//Add it to the list
	if(SKM_sk_push(WAITING_NEWPKI_OBJECT, WaitingObjects, object) < 0)
	{
		WaitingObjectsLock.Unlock();
		WAITING_NEWPKI_OBJECT_free(object);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	WaitingObjectsLock.Unlock();

	return true;
}

bool Entity_REPOSITORY::RepositoryKnowsObject(const WAITING_NEWPKI_OBJECT* object, const X509 * Cert)
{
	int j;
	X509_PUBKEY * currPubKey;

	for(j=0; j<SKM_sk_num(X509_PUBKEY, object->rep_path); j++)
	{
		currPubKey = SKM_sk_value(X509_PUBKEY, object->rep_path, j);
		if(!currPubKey) continue;

		if(Cert->cert_info->key->public_key->length ==
			currPubKey->public_key->length &&
			memcmp(Cert->cert_info->key->public_key->data,
			currPubKey->public_key->data, 
			currPubKey->public_key->length) == 0)
		{
			return true;
		}
	}

	return false;
}

bool Entity_REPOSITORY::OnNewPkiConf(const EXPORTED_PKI_CONF *conf)
{
	EVP_PKEY * pubKey;
	int i;
	ENTITY_CONF_CRYPTED * currConf;
	mString req;
	char * conf_pem;

	ConfAccessLock.LockWrite();
	//Check configuration validity, has it been signed by the PKI?
	pubKey = X509_get_pubkey(c_myConf->cas->PkiCert);
	if(!pubKey)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if(ASN1_verify((int (*)())i2d_EXPORTED_PKI_CONF_BODY, conf->sig->algor, conf->sig->digest, (char*)conf->confs, pubKey) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		EVP_PKEY_free(pubKey);
		ConfAccessLock.Unlock();
		return false;
	}
	EVP_PKEY_free(pubKey);
	

	//Check configuration version, we ignore it if we have a higher version
	if(pki_conf && (ASN1_INTEGER_GET(pki_conf->confs->version) >= ASN1_INTEGER_GET(conf->confs->version)) )
	{
		ConfAccessLock.Unlock();
		return true;
	}

	// Synch the conf
	if(!m_Jobs.SetPkiConf(conf))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	//Copy conf
	if(!EXPORTED_PKI_CONF_cpy(&pki_conf, conf))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	//Insert global conf into database
	conf_pem = EXPORTED_PKI_CONF_To_PEM(pki_conf);
	if(!conf_pem)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
		
	SQL sql(m_DbConn);

	if(req.sprintf(REP_ENTITY_INSERT_G_CONF, conf_pem) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		free(conf_pem);
		ConfAccessLock.Unlock();
		return false;
	}

	free(conf_pem);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	NewpkiDebug(LOG_LEVEL_INFO, m_EntityName.c_str(), _sv("New PKI conf version %ld"), ASN1_INTEGER_GET(pki_conf->confs->version));


	//Find my own personal conf
	for(i=0; i<SKM_sk_num(ENTITY_CONF_CRYPTED, pki_conf->confs->all_confs); i++)
	{
		currConf = SKM_sk_value(ENTITY_CONF_CRYPTED, pki_conf->confs->all_confs, i);
		if(!currConf) continue;

		//Is it my public key ?
		if(m_EntityCert == currConf->recipient)
		{
			if(!ExtractMyConf(currConf, &myConf, NULL, c_myConf->cas->PkiCert, true))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				ConfAccessLock.Unlock();
				return false;
			}
			break;
		}
	}
	ConfAccessLock.Unlock();

	return true;
}


bool Entity_REPOSITORY::GetConfiguration(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	EXPORTED_PKI_CONF * conf = NULL;

	ConfAccessLock.LockRead();
	if(!pki_conf)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ENTITY_NOT_READY);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!EXPORTED_PKI_CONF_cpy(&conf, pki_conf))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();


	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_CONF, conf))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	return true;
}

