/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/



// Entity_SERVER.cpp: implementation of the Entity_SERVER class.
//
//////////////////////////////////////////////////////////////////////
#include "Entity_CA.h"
#include "Entity_RA.h"
#include "Entity_PKI.h"
#include "Entity_REPOSITORY.h"
#include "Entity_PUBLICATION.h"
#include "Entity_KEYSTORE.h"
#include <PKI_PASSWD.h>

#include "Entity_SERVER.h"
#include "svintl.h"


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Entity_SERVER::Entity_SERVER(ENTITY_CONSTRUCTOR_PARAMETERS):Entity(ENTITY_CONSTRUCTOR_PARAM_PASSTHRU)
{
	m_ThreadDeleteEntity.Create(ThreadDeleteEntity, this);
}

Entity_SERVER::~Entity_SERVER()
{
	std::map<mString, ENTITY_INFO>::iterator i;

	m_ThreadDeleteEntity.Stop();

	m_EntitiesLock.LockWrite();
	for(i=m_EntitiesList.begin(); i != m_EntitiesList.end(); i++)
	{
		NewpkiDebug(LOG_LEVEL_DEBUG, _sv("NewPKI Server"), _sv("Unloading entity %s"), i->first.c_str());
		i->second.entity->WaitForNoClients();
		delete i->second.entity;
		NewpkiDebug(LOG_LEVEL_DEBUG, _sv("NewPKI Server"), _sv("Successfuly unloaded entity %s"), i->first.c_str());
	}
	m_EntitiesList.clear();
	m_EntitiesLock.Unlock();
}


Entity * Entity_SERVER::Entity_new(mString & Name, int Type)
{
	switch(Type)
	{
		case ENTITY_TYPE_RA:
			return new Entity_RA(m_Conf, Name, m_Engine, m_AuthModule, Type);
			break;
		case ENTITY_TYPE_CA:
			return new Entity_CA(m_Conf, Name, m_Engine, m_AuthModule, Type);
			break;
		case ENTITY_TYPE_PKI:
			return new Entity_PKI(m_Conf, Name, m_Engine, m_AuthModule, Type);
			break;
		case ENTITY_TYPE_REPOSITORY:
			return new Entity_REPOSITORY(m_Conf, Name, m_Engine, m_AuthModule, Type);
			break;
		case ENTITY_TYPE_KEY_STORE:
			return new Entity_KEYSTORE(m_Conf, Name, m_Engine, m_AuthModule, Type);
			break;
		case ENTITY_TYPE_PUBLICATION:
			return new Entity_PUBLICATION(m_Conf, Name, m_Engine, m_AuthModule, Type);
			break;
		default:
			return NULL;
	}
}

bool Entity_SERVER::Create(const ENTITY_CREATION_DATAS * Params, ADMIN_RESPONSE_BODY * response)
{
	GEN_PRIVATE_KEY * entity_key;
	const char * InitialCreates[]={INITIAL_CREATE_1, INITIAL_CREATE_2, INITIAL_CREATE_3, INITIAL_CREATE_4, INITIAL_CREATE_5, NULL};

	entity_key = GEN_PRIVATE_KEY_new();
	if(!entity_key)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	if(m_Conf.get_SSL_KeyId().size())
	{
		entity_key->type = GEN_PRIVATE_KEY_TYPE_ENGINE;
		ASN1_STRING_SET(&entity_key->d.keyid, m_Conf.get_SSL_KeyId().c_str());
	}
	else
	{
		entity_key->type = GEN_PRIVATE_KEY_TYPE_KEYLEN;
		ASN1_INTEGER_SET(&entity_key->d.keylen, 2048);
	}



	NewpkiDebug(LOG_LEVEL_DEBUG, _sv("NewPKI Server"), _sv("Creating NewPKI database..."));
	//We create the database
	if(!Private_CreateEntity(entity_key, NULL, InitialCreates))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NewpkiDebug(LOG_LEVEL_ERROR, _sv("NewPKI Server"), _sv("Failed to create NewPKI database"));
		GEN_PRIVATE_KEY_free(entity_key);
		return false;
	}
	GEN_PRIVATE_KEY_free(entity_key);
	NewpkiDebug(LOG_LEVEL_DEBUG, _sv("NewPKI Server"), _sv("Successfuly created NewPKI database"));


	// Creation of the SSL certificate
	NewpkiDebug(LOG_LEVEL_DEBUG, _sv("NewPKI Server"), _sv("Creating NewPKI server's SSL certificate..."));
	if(!CreateSSL_CERT())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NewpkiDebug(LOG_LEVEL_ERROR, _sv("NewPKI Server"), _sv("Failed to create NewPKI server's SSL certificate"));
		Destroy();
		return false;
	}

	if(!Private_InitEntity(m_EntityCert.GetX509()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NewpkiDebug(LOG_LEVEL_ERROR, _sv("NewPKI Server"), _sv("Failed to create NewPKI server's SSL certificate"));
		Destroy();
		return false;
	}

	NewpkiDebug(LOG_LEVEL_DEBUG, _sv("NewPKI Server"), _sv("Successfuly created NewPKI server's SSL certificate"));

	return true;
}

bool Entity_SERVER::Init(const ENTITY_SIGNATURE_RESP * init_datas)
{
	NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
	return false;
}

bool Entity_SERVER::Load()
{
	SQL sql(m_DbConn);
	long NumRows;
	long i;
	mString entity_name;
	mString str_entity_type;	
	char * err;
	

	if(!Private_CommonLoad(NULL, false))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!IsFullyInit())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}
	

	// Getting available entities and loading each of them
	if(!sql.Execute(INITIAL_GET_ENTITIES) )
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	for(i=0; i<NumRows; i++)
	{
		if(!sql.Value(i, "entity_name", entity_name))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}

		if(!sql.Value(i, "entity_type", str_entity_type))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}

		NewpkiDebug(LOG_LEVEL_DEBUG, entity_name.c_str(), _sv("Loading entity"));
		if(!LoadEntity(entity_name, str_entity_type.c_int()))
		{
			err = ERR_to_string();
			m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_ENTITY_LOAD, 0, NULL, LOG_NO_OBJECTID, entity_name.c_str(), err);
			NewpkiDebug(LOG_LEVEL_ERROR, entity_name.c_str(), _sv("Entity failed to load: %s"), err);
			if(err) free(err);
		}
		else
		{
			m_Logging->LogMessage(LOG_STATUS_TYPE_SUCCESS, LOG_MESSAGE_TYPE_ENTITY_LOAD, 0, NULL, LOG_NO_OBJECTID, entity_name.c_str());
			NewpkiDebug(LOG_LEVEL_DEBUG, entity_name.c_str(), _sv("Entity loaded"));
		}
	}

	if(!m_ThreadDeleteEntity.Start())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_SERVER::PersonnalConfFromPEM(const mString & conf_pem)
{
	return false;
}

bool Entity_SERVER::PersonnalConfToPem(mString & conf_pem)
{
	return false;
}

bool Entity_SERVER::LoadEntity(mString & Name, int Type)
{
	Entity * entity;
	ENTITY_INFO entity_info;

	//We create the entity
	try
	{
		entity = Entity_new(Name, Type);
	}
	catch(ExceptionNewPKI e)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!entity)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//We upgrade it
	if(!entity->Upgrade(NEWPKI_VERSION))
	{
		delete entity;
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NewpkiDebug(LOG_LEVEL_ERROR, _sv("NewPKI Server"), _sv("Failed to upgrade Entity %s"), Name.c_str());
		return false;
	}

	//We load it
	if(!entity->Load())
	{
		delete entity;
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NewpkiDebug(LOG_LEVEL_ERROR, _sv("NewPKI Server"), _sv("Failed to load Entity %s"), Name.c_str());
		return false;
	}


	entity_info.Type = Type;
	entity_info.entity = entity;

	//Add it to the list
	m_EntitiesLock.LockWrite();
	m_EntitiesList[Name] = entity_info;
	m_EntitiesLock.Unlock();
	return true;
}

void Entity_SERVER::OnADMIN_REQUEST(ADMIN_RESPONSE_BODY * response, const char * ClientIp, const PKI_CERT & ClientCert, const ADMIN_REQUEST *AdminRequest)
{
	mString EntityName;
	char * err;

	if(ParseAdminRequest(response, ClientCert, AdminRequest, EntityName))
	{
		m_Logging->LogMessage(LOG_STATUS_TYPE_SUCCESS, LOG_MESSAGE_TYPE_ADMIN_REQUEST, 0, ClientIp, LOG_NO_OBJECTID, EntityName.c_str());
	}
	else
	{
		err = ERR_to_string();
		m_Logging->LogMessage(LOG_STATUS_TYPE_FAILURE, LOG_MESSAGE_TYPE_ADMIN_REQUEST, 0, ClientIp, LOG_NO_OBJECTID, EntityName.c_str(), err?err:_sv("Unknown"));
		if(err) free(err);
	}
}

bool Entity_SERVER::ParseAdminRequest(ADMIN_RESPONSE_BODY * response, const PKI_CERT & ClientCert, const ADMIN_REQUEST *AdminRequest, mString & EntityName)
{
	Entity * entity;
	char * entity_name;
	bool ret;

	ERR_clear_error();

	m_EntitiesLock.LockRead();

	entity = GetEntity();

	if(AdminRequest->body->type == ADMIN_REQ_TYPE_LOGIN)
	{
		// User wants to login but he already did
		if(entity)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PKI_COMMAND);
			m_EntitiesLock.Unlock();
			ERR_to_ADMIN_RESPONSE(response);
			EntityName = entity->GetName();
			return false;
		}

		//Get entity user intents to log on
		entity_name = ASN1_STRING_GET(AdminRequest->body->d.login->entity);
		
		//It's for the local server entity (myself)
		if(!entity_name || !*entity_name)
		{
			entity_name = INITIAL_DB;
		}

		// we add the user to the list of known users
		GlobalUsersLock.Lock();
		GlobalUsers[NewpkiThread::CurrentThreadId()] = entity_name;
		GlobalUsersLock.Unlock();
		
		entity = GetEntity();
	}
	if(!entity)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN_ENTITY);
		m_EntitiesLock.Unlock();
		ERR_to_ADMIN_RESPONSE(response);
		EntityName = _sv("Unknown");
		return false;
	}
	EntityName = entity->GetName();

	entity->ClientEnter();
	m_EntitiesLock.Unlock();
	if(!entity->IsFullyInit())
	{// I don't take commands for an entity when it's not initialized
		entity->ClientLeave();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ENTITY_NOT_READY);
		ERR_to_ADMIN_RESPONSE(response);
		return false;
	}
	else
	{
		ret = entity->ParseAdminCommand(response, ClientCert, AdminRequest);
	}
	entity->ClientLeave();

	return ret;
}

void Entity_SERVER::OnConnectionClosed()
{
	Entity * entity;
	std::map<NEWPKI_THREAD_ID, mString>::iterator user;

	m_EntitiesLock.LockRead();
	if( (entity = GetEntity()) )
	{
		GlobalUsersLock.Lock();

		user = GlobalUsers.find(NewpkiThread::CurrentThreadId());
		if(user != GlobalUsers.end())
		{
			GlobalUsers.erase(user);
		}
		GlobalUsersLock.Unlock();
		entity->ClientEnter();
		entity->OnUserLogout();
		entity->ClientLeave();
	}
	m_EntitiesLock.Unlock();
}



Entity * Entity_SERVER::GetEntity()
{
	mString entity_name;
	std::map<NEWPKI_THREAD_ID, mString>::iterator user;

	GlobalUsersLock.Lock();
	//Is user known ?
	user = GlobalUsers.find(NewpkiThread::CurrentThreadId());
	if(user == GlobalUsers.end())
	{
		GlobalUsersLock.Unlock();
		return NULL;
	}
	entity_name = user->second;
	GlobalUsersLock.Unlock();

	//If it's me we do internal processing
	if(entity_name == INITIAL_DB)
	{
		return this;
	}
	else
	{
		//We get the associated entity
		if( m_EntitiesList.find(entity_name) == m_EntitiesList.end() )
		{
			return NULL;
		}
		return m_EntitiesList[entity_name].entity;
	}
}

bool Entity_SERVER::CreateSSL_CERT()
{
	mString req;
	char localName[500];

	mString cert_pem;
	mString key_pem;

	HashTable_Dn Dn;
	PKI_CSR CsrK;
	mString emailAddress;

	if(gethostname(localName, sizeof(localName)) == -1)
	{
		strcpy(localName, "localhost");
	}
	emailAddress = "localhost@";
	emailAddress += localName;

	Dn.Add("commonName", localName);
	Dn.Add("organizationName", "NewPKI");
	Dn.Add("organizationalUnitName", "SSL Server");
	Dn.Add("emailAddress", (char*)emailAddress.c_str());

	if(!CsrK.GenerateCSR(Dn, m_EntityKey))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	HashTable_String Exts;
	Exts.Add("basicConstraints", "CA:FALSE");
	Exts.Add("nsComment", "NewPKI SSL Server Certificate");
	Exts.Add("subjectKeyIdentifier", "hash");
	Exts.Add("authorityKeyIdentifier", "keyid:always");
	Exts.Add("keyUsage", "digitalSignature, nonRepudiation, keyEncipherment");
	Exts.Add("nsCertType", "server");
	Exts.Add("extendedKeyUsage", "serverAuth");


	if(!m_EntityCert.CreateSelfSigned(CsrK, &Exts, 3650, 1))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool Entity_SERVER::ParseAdminCommand(ADMIN_RESPONSE_BODY * response, const PKI_CERT & ClientCert, const ADMIN_REQUEST * AdminRequest)
{
	unsigned long user_id;
	const UserHandle & hUser = GetUserHandle();

	user_id	= ( (hUser)?hUser.GetUserId():0);

	PARSER_COMMAND_BEGIN(Entity_SERVER, response, user_id, AdminRequest, ClientCert, hUser, true, this, LogsType)
		PARSER_ADD_LOG_ENTRY(LOG_MESSAGE_TYPE_ENTITY_LOAD)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_LOGIN,				Entity_SERVER::UserLogin,			LOG_MESSAGE_TYPE_USER_LOGIN, ( (ASN1_STRING_GET(AdminRequest->body->d.login->username))?ASN1_STRING_GET(AdminRequest->body->d.login->username):((ClientCert)?(char*)ClientCert.GetStringName():NULL)), LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_CREATE_ENTITY,		Entity_SERVER::CreateEntity,		LOG_MESSAGE_TYPE_ENTITY_ADD, ASN1_STRING_GET(AdminRequest->body->d.cre_entity->Name) , LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_DELETE_ENTITY,		Entity_SERVER::DeleteEntity,		LOG_MESSAGE_TYPE_ENTITY_DEL, ASN1_STRING_GET(AdminRequest->body->d.del_entity), LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_ENUM_ENTITIES,		Entity_SERVER::EnumEntities)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_ENUM_LOGS,			Entity_SERVER::EnumLogs)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_LOGS_COUNT,		Entity_SERVER::GetLogsCount)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_ENUM_USERS,			Entity_SERVER::EnumUsers)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_CREATE_USER,			Entity_SERVER::CreateUser,			LOG_MESSAGE_TYPE_USER_ADD, ASN1_STRING_GET(AdminRequest->body->d.cre_user->Name), LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_UPDATE_USER,			Entity_SERVER::UpdateUser,			LOG_MESSAGE_TYPE_USER_MOD, NULL, ASN1_INTEGER_GET(AdminRequest->body->d.upd_user->user_id))
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_CHG_USER_PASSWD,		Entity_SERVER::ChangeUserPassword,	LOG_MESSAGE_TYPE_PWD_MOD, NULL, ASN1_INTEGER_GET(AdminRequest->body->d.chg_user_pwd->user_id))
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_CHG_PASSWD,			Entity_SERVER::ChangePassword,		LOG_MESSAGE_TYPE_PWD_MOD, NULL, hUser.GetUserId())
		PARSER_COMMAND_ENTRY_LOG(	ADMIN_REQ_TYPE_INIT_ENTITY,			Entity_SERVER::InitEntity,			LOG_MESSAGE_TYPE_ENTITY_INIT, ASN1_STRING_GET(AdminRequest->body->d.init_entity->name) , LOG_NO_OBJECTID)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_GET_LOGS_TYPE,		Entity_SERVER::GetLogsType)
		PARSER_COMMAND_ENTRY(		ADMIN_REQ_TYPE_CHECK_LOGS,			Entity_SERVER::CheckLogsIntegrity)
	PARSER_COMMAND_END(Entity_SERVER)
}

bool Entity_SERVER::LoginUser(UserHandle & hUser, int & UserType)
{
	mString req;
	long NumRows;
	mString tmpdatas;
	PKI_PASSWD pki_password;
	mString username;
	mString password;

	username = hUser.GetUserName();
	password = hUser.GetPassword();
	if(!username.size() || !password.size())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PKI_COMMAND);
		return false;
	}


	SQL sql(m_DbConn);

	// We seek the user
	if(req.sprintf( SQL_SELECT_USER, username.c_str()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!NumRows)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_UNKNOWN_USER);
		return false;
	}

	if(!sql.Value(0, "flags", tmpdatas))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	if( (tmpdatas.c_lng() & USER_FLAGS_EXTERNAL_AUTH) )
	{
		if(!m_AuthModule || !m_AuthModule->Login((char*)username.c_str(), (char*)password.c_str()))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_UNKNOWN_USER);
			return false;
		}
	}
	else
	{
		if(!sql.Value(0, "password", tmpdatas))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		if(!pki_password.VerifySHA1Password((char*)password.c_str(), (char*)tmpdatas.c_str()))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_UNKNOWN_USER);
			return false;
		}
	}


	if(!sql.Value(0, "user_id", tmpdatas))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	hUser.SetUserId(tmpdatas.c_ulng());

	if(!sql.Value(0, "activated", tmpdatas))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	// If user has been deactivated
	if(!tmpdatas.c_int())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_UNKNOWN_USER);
		return false;
	}
		
	UserType = USER_TYPE_SERVER;
	return true;
}

void Entity_SERVER::LogoutUser(const UserHandle & hUser)
{
}

bool Entity_SERVER::CreateEntity(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	char * strEntityName;
	mString EntityName;
	ENTITY_INFO entity_info;

	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	
	strEntityName = ASN1_STRING_GET(body->d.cre_entity->Name);
	entity_info.Type = body->d.cre_entity->Datas->type;

	if(!strEntityName)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}
	EntityName = strEntityName;

	//We create the entity
	entity_info.entity = Entity_new(EntityName, entity_info.Type);
	if(!entity_info.entity)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!entity_info.entity->Create(body->d.cre_entity->Datas, response))
	{
		delete entity_info.entity;
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!entity_info.entity->Load())
	{
		ASN1_CHOICE_VALUE_free((ASN1_VALUE*)response, ASN1_ITEM_rptr(ADMIN_RESPONSE_BODY));
		entity_info.entity->Destroy();
		delete entity_info.entity;
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}


	//We now insert it in the database
	SQL sql(m_DbConn);
	mString req;

	if(req.sprintf(INITIAL_INSERT_ENTITY, EntityName.c_str(), entity_info.Type) == -1)
	{
		entity_info.entity->Destroy();
		delete entity_info.entity;
		ASN1_CHOICE_VALUE_free((ASN1_VALUE*)response, ASN1_ITEM_rptr(ADMIN_RESPONSE_BODY));
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	if(!sql.Execute(req))
	{
		entity_info.entity->Destroy();
		delete entity_info.entity;
		ASN1_CHOICE_VALUE_free((ASN1_VALUE*)response, ASN1_ITEM_rptr(ADMIN_RESPONSE_BODY));
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(entity_info.Type == ENTITY_TYPE_PUBLICATION)
	{
		((Entity_PUBLICATION*)entity_info.entity)->SetOcspServersHandler(m_OcspServersHandler);
	}

	m_EntitiesLock.LockWrite();
	m_EntitiesList[EntityName] = entity_info;
	m_EntitiesLock.Unlock();

	return true;
}

bool Entity_SERVER::DeleteEntity(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	const char * entity_name;
	mString req;
	Entity * entity;

	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	
	SQL sql(m_DbConn);

	entity_name = ASN1_STRING_GET(body->d.del_entity);
	if(!entity_name)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}


	m_EntitiesLock.LockWrite();
	if( m_EntitiesList.find(entity_name) == m_EntitiesList.end() )
	{
		m_EntitiesLock.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN_ENTITY);
		return false;
	}
	//Ask the entity to stop any activity
	m_EntitiesList[entity_name].entity->SpawnStop();
	
	//We remove it from the database
	if(req.sprintf(INITIAL_REMOVE_ENTITY, entity_name) <= 0)
	{
		m_EntitiesLock.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	if(!sql.Execute(req))
	{
		m_EntitiesLock.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	entity = m_EntitiesList[entity_name].entity;
	m_EntitiesList.erase(entity_name);
	m_EntitiesLock.Unlock();

	m_DeleteEntitiesLock.Lock();
	m_DeleteEntitiesList.push_back(entity);
	m_DeleteEntitiesLock.Unlock();

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_SERVER::EnumEntities(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	ENTITY_ENTRY_INFO * entity;
	STACK_OF(ENTITY_ENTRY_INFO) * entities;
	PKI_CERT EntityCert;
	std::map<mString, ENTITY_INFO>::iterator i;

	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	entities = sk_new_null();
	if(!entities)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	m_EntitiesLock.LockRead();
	for(i=m_EntitiesList.begin(); i != m_EntitiesList.end(); i++)
	{
		entity = ENTITY_ENTRY_INFO_new(); 
		if(!entity) continue;

		i->second.entity->GetEntityCertificate(EntityCert);

		ASN1_STRING_SET(&entity->Name, i->first.c_str());
		ASN1_INTEGER_SET(&entity->Type, i->second.Type);

		if(EntityCert)
		{
			if(!X509_cpy(&entity->certificate, EntityCert.GetX509()))
			{
				ENTITY_ENTRY_INFO_free(entity);
				continue;
			}
		}
		if(SKM_sk_push(ENTITY_ENTRY_INFO, entities, entity) < 0)
		{
			ENTITY_ENTRY_INFO_free(entity);
			continue;
		}
	}
	m_EntitiesLock.Unlock();

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_ENUM_ENTITIES, entities))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		SKM_sk_pop_free(ENTITY_ENTRY_INFO, entities, ENTITY_ENTRY_INFO_free);
		return false;
	}
	return true;
}

bool Entity_SERVER::EnumUsers(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	if(!Private_EnumUsers(response, hUser))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_SERVER::CreateUser(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	if(!Private_CreateUser(response, hUser, ASN1_STRING_GET(body->d.cre_user->Name), ASN1_INTEGER_GET(body->d.cre_user->Flags), ASN1_INTEGER_GET(body->d.cre_user->Activated)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_SERVER::Private_EnumUsers(ADMIN_RESPONSE_BODY * response, const UserHandle & hUser)
{
	SQL sql(m_DbConn);

	long NumRows;
	long tmplong;
	mString tmpstr;
	STACK_OF(USER_ENTRY) * users;
	USER_ENTRY * user;
	int i;

	if(!sql.Execute(SQL_SELECT_USERS))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	users = sk_new_null();
	if(!users)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	for(i=0; i<NumRows; i++)
	{
		user = USER_ENTRY_new();
		if(!user)
		{
			SKM_sk_pop_free(USER_ENTRY, users, USER_ENTRY_free);
			NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
			return false;
		}
		if(SKM_sk_push(USER_ENTRY, users, user) < 0)
		{
			SKM_sk_pop_free(USER_ENTRY, users, USER_ENTRY_free);
			USER_ENTRY_free(user);
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}


		if(!sql.Value(i, "user_id", tmpstr))
		{
			SKM_sk_pop_free(USER_ENTRY, users, USER_ENTRY_free);
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		sscanf(tmpstr.c_str(), "%ld", &tmplong);
		ASN1_INTEGER_SET(&user->user_id, tmplong);

		if(!sql.Value(i, "username", tmpstr))
		{
			SKM_sk_pop_free(USER_ENTRY, users, USER_ENTRY_free);
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		ASN1_STRING_SET(&user->user_info->Name, tmpstr.c_str());

		if(!sql.Value(i, "activated", tmpstr))
		{
			SKM_sk_pop_free(USER_ENTRY, users, USER_ENTRY_free);
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		sscanf(tmpstr.c_str(), "%ld", &tmplong);
		ASN1_INTEGER_SET(&user->user_info->Activated, tmplong);

		if(!sql.Value(i, "flags", tmpstr))
		{
			SKM_sk_pop_free(USER_ENTRY, users, USER_ENTRY_free);
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		sscanf(tmpstr.c_str(), "%ld", &tmplong);
		ASN1_INTEGER_SET(&user->user_info->Flags, tmplong);
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_ENUM_USERS, users))
	{
		SKM_sk_pop_free(USER_ENTRY, users, USER_ENTRY_free);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_SERVER::Private_CreateUser(ADMIN_RESPONSE_BODY * response, const UserHandle & hUser, char * Username, unsigned long Flags, int Activated)
{
	SQL sql(m_DbConn);
	mString req;
	char * password;
	PKI_PASSWD pwd;
	
	if(!Username)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	//We ask the module to validate the user name
	if( (Flags & USER_FLAGS_EXTERNAL_AUTH) )
	{
		if(!m_AuthModule)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_AUTH_LIB);
			return false;
		}
		if(!m_AuthModule->ValidateUsername(Username))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}

	password = pwd.CalcSHA1Password(Username);
	if(!password)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}

	if(req.sprintf(SQL_INSERT_USER, Username, password, Activated, Flags) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}


bool Entity_SERVER::UpdateUser(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	if(!Private_UpdateUser(response, hUser, ASN1_INTEGER_GET(body->d.upd_user->user_id), ASN1_STRING_GET(body->d.upd_user->user_info->Name), ASN1_INTEGER_GET(body->d.upd_user->user_info->Flags), ASN1_INTEGER_GET(body->d.upd_user->user_info->Activated)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_SERVER::Private_UpdateUser(ADMIN_RESPONSE_BODY * response, const UserHandle & hUser, unsigned long user_id, char * Username, unsigned long Flags, int Activated)
{
	SQL sql(m_DbConn);
	mString req;

	if(!Username)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	//No one ca modify the default admin !
	if(user_id == DEFAULT_ADMIN_ID)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	//We verify thru the module the validity of the username
	if( (Flags & USER_FLAGS_EXTERNAL_AUTH) )
	{
		if(!m_AuthModule)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_AUTH_LIB);
			return false;
		}
		if(!m_AuthModule->ValidateUsername(Username))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}

	if(req.sprintf(SQL_UPDATE_USER, Username, Activated, Flags, user_id) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}



bool Entity_SERVER::ChangeUserPassword(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	if(!Private_ChangeUserPassword(response, hUser, ASN1_INTEGER_GET(body->d.chg_user_pwd->user_id), ASN1_STRING_GET(body->d.chg_user_pwd->Password)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_SERVER::ChangePassword(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}
	if(!Private_ChangeUserPassword(response, hUser, hUser.GetUserId(), ASN1_STRING_GET(body->d.chg_pwd->Password)))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool Entity_SERVER::Private_ChangeUserPassword(ADMIN_RESPONSE_BODY * response, const UserHandle & hUser, unsigned long user_id, char *NewPassword)
{
	SQL sql(m_DbConn);
	mString req;
	PKI_PASSWD pwd;
	char  * password;

	if(!NewPassword)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	// Users can change their own password
	if(user_id != hUser.GetUserId())
	{
		//Only default admin can change his password !
		if(user_id == DEFAULT_ADMIN_ID && hUser.GetUserId() != DEFAULT_ADMIN_ID)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
			return false;
		}
	}
	

	password = pwd.CalcSHA1Password(NewPassword);
	if(!password)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}

	if(req.sprintf(SQL_UPDATE_USER_PWD, password, user_id) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		return false;
	}
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}


bool Entity_SERVER::InitEntity(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser)
{
	char * entity_name;
	if(!hUser)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	entity_name = ASN1_STRING_GET(body->d.init_entity->name);
	if(!entity_name)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	m_EntitiesLock.LockWrite();
	if( m_EntitiesList.find(entity_name) == m_EntitiesList.end() )
	{
		m_EntitiesLock.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN_ENTITY);
		return false;
	}

	if(!m_EntitiesList[entity_name].entity->Init(body->d.init_entity->sign_resp))
	{
		m_EntitiesLock.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!m_EntitiesList[entity_name].entity->Load())
	{
		m_EntitiesLock.Unlock();
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	m_EntitiesLock.Unlock();

	if(!ADMIN_RESPONSE_init(response, ADMIN_RESP_TYPE_NONE, NULL))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}


bool Entity_SERVER::Upgrade(const char * Version)
{
	return true;
}

void Entity_SERVER::SpawnStop()
{
	std::map<mString, ENTITY_INFO>::iterator i;

	m_ThreadDeleteEntity.SpawStop();

	m_EntitiesLock.LockRead();
	for(i=m_EntitiesList.begin(); i != m_EntitiesList.end(); i++)
	{
		i->second.entity->ClientEnter();
		i->second.entity->SpawnStop();
		i->second.entity->ClientLeave();
	}
	m_EntitiesLock.Unlock();
	m_Jobs.SpawnStop();
}

bool Entity_SERVER::ParseNewConf(const ENTITY_CONF_CRYPTED * newConf, const INTERNAL_PKI_CA * cas)
{
	return true;
}

bool Entity_SERVER::OnNewConf(const ENTITY_CONF_CRYPTED * newConf)
{
	return true;
}

bool Entity_SERVER::OnNewRequests(const STORED_NEWPKI_REQUESTS * reqs)
{
	return true;
}

bool Entity_SERVER::OnRequestSent(unsigned long RequestId, const ASN1_OCTET_STRING * transactionId, const CRYPTED_NEWPKI_REQUEST * req)
{
	return true;
}

bool Entity_SERVER::OnResponseSent(const STORED_NEWPKI_RESPONSE * resp)
{
	return true;
}

bool Entity_SERVER::OnNewResponses(const STORED_NEWPKI_RESPONSES * resps)
{
	return true;
}

bool Entity_SERVER::GetWaitingTransactionIds(TRANSACTION_IDS *transactionIDs)
{
	return false;
}

bool Entity_SERVER::GetProcessedTransactionIds(TRANSACTION_IDS *transactionIDs)
{
	return false;
}

bool Entity_SERVER::GetWaitingRequests(STACK_OF(CRYPTED_NEWPKI_REQUEST) * Requests)
{
	return true;
}

bool Entity_SERVER::GetWaitingResponses(STACK_OF(CRYPTED_NEWPKI_RESPONSE) * Responses)
{
	return true;
}

bool Entity_SERVER::EntitySignResponse(ADMIN_RESPONSE *response)
{
	Entity * entity;
	bool Ret;

	m_EntitiesLock.LockRead();
		entity = GetEntity();
		if(!entity)
			entity = this;

	entity->ClientEnter();
	m_EntitiesLock.Unlock();

	// If entity is not fully initialized I sign the response
	if(!entity->IsFullyInit())
	{
		Ret = SignResponse(response);
	}
	else
	{
		Ret = entity->SignResponse(response);
	}
	entity->ClientLeave();

	return Ret;
}

bool Entity_SERVER::OnNewOCSP(const char * Ip, const OCSP_REQUEST *request, OCSP_RESPONSE **response, const mString &EntityName, bool & ShouldStopServer)
{
	bool Ret;
	
	if(EntityName == INITIAL_DB)
	{
		ShouldStopServer = true;
		return false;
	}
	ShouldStopServer = false;

	m_EntitiesLock.LockRead();
	//We get the entity
	if(m_EntitiesList.find(EntityName) == m_EntitiesList.end())
	{
		m_EntitiesLock.Unlock();
		ShouldStopServer = true;
		return false;
	}

	m_EntitiesList[EntityName].entity->ClientEnter();
	m_EntitiesLock.Unlock();

	if(m_EntitiesList[EntityName].Type != ENTITY_TYPE_PUBLICATION)
	{
		m_EntitiesList[EntityName].entity->ClientLeave();
		ShouldStopServer = true;
		return false;
	}

	Ret = ((Entity_PUBLICATION*)m_EntitiesList[EntityName].entity)->OnNewOCSP(Ip, request, response);
	m_EntitiesList[EntityName].entity->ClientLeave();

	return Ret;
}


void Entity_SERVER::SetOcspServersHandler(OcspServers *OcspServersHandler)
{
	std::map<mString, ENTITY_INFO>::iterator i;
	m_OcspServersHandler = OcspServersHandler;

	m_EntitiesLock.LockRead();
	for(i=m_EntitiesList.begin(); i != m_EntitiesList.end(); i++)
	{
		i->second.entity->ClientEnter();
		if(i->second.Type != ENTITY_TYPE_PUBLICATION)
		{
			i->second.entity->ClientLeave();
			continue;
		}
		((Entity_PUBLICATION*)i->second.entity)->SetOcspServersHandler(OcspServersHandler);
		i->second.entity->ClientLeave();
	}
	m_EntitiesLock.Unlock();
}

void Entity_SERVER::ThreadDeleteEntity(const NewpkiThread *Thread, void *param)
{
	Entity_SERVER * me_this = (Entity_SERVER *)param;
	Entity * entity;

	while(!Thread->ShouldStop())
	{
		me_this->m_DeleteEntitiesLock.Lock();
		if(me_this->m_DeleteEntitiesList.size())
		{
			entity = me_this->m_DeleteEntitiesList[0];
			me_this->m_DeleteEntitiesList.erase(me_this->m_DeleteEntitiesList.begin());
			me_this->m_DeleteEntitiesLock.Unlock();
			if(entity)
			{
				entity->Destroy();
				entity->WaitForNoClients();
				delete entity;
			}
		}
		else
		{
			me_this->m_DeleteEntitiesLock.Unlock();
		}
		NewpkiThread::Sleep(500);
	}
}

