/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/



// Entity_SERVER.h: interface for the Entity_SERVER class.
//
//////////////////////////////////////////////////////////////////////

#ifndef ENTITY_SERVER_H
#define ENTITY_SERVER_H

class OcspServers;


#include <newpki_threads.h>
#include <CriticalSection.h>
#include <HashTable/HashTable_String.h>

#include "Entity.h"
#include "ASN1/ASN1_entity.h"
#include <mString.h>
#include <openssl/ocsp.h>

#include <map>
using namespace std;


/*!
	This class represents a Server Entity
*/
class Entity_SERVER : public Entity
{
public:
	Entity_SERVER(ENTITY_CONSTRUCTOR_PARAMETERS);
	virtual ~Entity_SERVER();

	bool Load();
	bool Create(const ENTITY_CREATION_DATAS * Params, ADMIN_RESPONSE_BODY * response);
	bool ParseAdminCommand(ADMIN_RESPONSE_BODY * response, const PKI_CERT & ClientCert, const ADMIN_REQUEST * AdminRequest);
	void SpawnStop();
	bool Upgrade(const char * Version);
	bool Init(const ENTITY_SIGNATURE_RESP * init_datas);

	bool GetWaitingResponses(STACK_OF(CRYPTED_NEWPKI_RESPONSE) * Responses);
	bool GetWaitingRequests(STACK_OF(CRYPTED_NEWPKI_REQUEST) * Requests);
	bool GetProcessedTransactionIds(TRANSACTION_IDS * transactionIDs);
	bool GetWaitingTransactionIds(TRANSACTION_IDS * transactionIDs);
	bool OnRequestSent(unsigned long RequestId, const ASN1_OCTET_STRING * transactionId, const CRYPTED_NEWPKI_REQUEST * req);
	bool OnResponseSent(const STORED_NEWPKI_RESPONSE * resp);
	bool OnNewResponses(const STORED_NEWPKI_RESPONSES * resps);
	bool OnNewRequests(const STORED_NEWPKI_REQUESTS * reqs);
	bool OnNewConf(const ENTITY_CONF_CRYPTED * newConf);

	/*! \brief This function is called when a new request is received, the server entity will send it to the destination entity.
	 *  \param response [OUT] The response.
	 *  \param ClientIp [IN] The end user IP.
	 *  \param ClientCert [IN] The end user certificate.
	 *  \param AdminRequest [IN] The request.
	 *  \return true on success, false on failure.
	 */
	void OnADMIN_REQUEST(ADMIN_RESPONSE_BODY * response, const char * ClientIp, const PKI_CERT & ClientCert, const ADMIN_REQUEST *AdminRequest);

	/*! \brief This function is called when a new OCSP request arrives, it will send it to the right Entity_PUBLICATION.
	 *  \param Ip [IN] The IP of the requester.
	 *  \param request [IN] The OCSP request.
	 *  \param response [OUT] The OCSP response.
	 *  \param EntityName [IN] The name of the Entity_PUBLICATION that should process it.
	 *  \param ShouldStopServer [OUT] Set to true when the Entity_PUBLICATION is not available anymore.
	 *  \return true on success, false on failure.
	 */
	bool OnNewOCSP(const char * Ip, const OCSP_REQUEST * request, OCSP_RESPONSE ** response, const mString & EntityName, bool & ShouldStopServer);

	/*! \brief This function signs an response using the entity's certificate.
	 *  \param response [IN/OUT] The response.
	 *  \return true on success, false on failure.
	 */
	bool EntitySignResponse(ADMIN_RESPONSE* response);

	/*! \brief This function is called to initialize all the OCSP responders.
	 *  \param OcspServersHandler [IN] The OCSP servers provider.
	 */
	void SetOcspServersHandler(OcspServers * OcspServersHandler);

	/*! \brief This function is called when the connection to a user is closed.
	 */
	void OnConnectionClosed();

private:
	typedef struct st_ENTITY_INFO
	{
		int Type;
		Entity * entity;
	}ENTITY_INFO;

	CriticalSection GlobalUsersLock;
	map<NEWPKI_THREAD_ID, mString> GlobalUsers;

	bool ParseAdminRequest(ADMIN_RESPONSE_BODY * response, const PKI_CERT & ClientCert, const ADMIN_REQUEST *AdminRequest, mString & EntityName);
	bool EnumEntities(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	bool DeleteEntity(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	bool CreateEntity(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	bool EnumUsers(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	bool CreateUser(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	bool UpdateUser(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	bool ChangeUserPassword(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	bool ChangePassword(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);


	bool Private_EnumUsers(ADMIN_RESPONSE_BODY * response, const UserHandle & hUser);
	bool Private_CreateUser(ADMIN_RESPONSE_BODY * response, const UserHandle & hUser, char * Username, unsigned long Flags, int Activated);
	bool Private_UpdateUser(ADMIN_RESPONSE_BODY * response, const UserHandle & hUser, unsigned long user_id, char * Username, unsigned long Flags, int Activated);
	bool Private_ChangeUserPassword(ADMIN_RESPONSE_BODY * response, const UserHandle & hUser, unsigned long user_id, char * NewPassword);


	OcspServers * m_OcspServersHandler;
	static bool Private_ParseAdminCommand(bool ExecuteCmd, Entity * me_this, vector<LOG_MESSAGE_TYPE> &  mLogsType, ADMIN_RESPONSE_BODY * response, PKI_CERT * ClientCert, ADMIN_REQUEST * AdminRequest, UserHandle * Param);

	bool InitEntity(ADMIN_RESPONSE_BODY * response, int command, const PKI_CERT & UserCert, const ADMIN_REQUEST_BODY *body, const UserHandle & hUser);
	Entity* GetEntity();


	Entity * Entity_new(mString & Name, int Type);
	bool ParseNewConf(const ENTITY_CONF_CRYPTED * newConf, const INTERNAL_PKI_CA * cas);

	bool LoginUser(UserHandle & hUser, int & UserType);
	void LogoutUser(const UserHandle & hUser);

	bool PersonnalConfFromPEM(const mString & conf_pem);
	bool PersonnalConfToPem(mString & conf_pem);


	bool CreateSSL_CERT();
	void CleanEntities();


	NewpkiThread m_ThreadDeleteEntity;
	static void ThreadDeleteEntity(const NewpkiThread * Thread, void *param);
	
	ReadersWriter m_EntitiesLock;
	map<mString, ENTITY_INFO> m_EntitiesList;

	vector<Entity*> m_DeleteEntitiesList;
	CriticalSection m_DeleteEntitiesLock;
	bool LoadEntity(mString & Name, int Type);

	DECLARE_COMMAND_PARSER(Entity_SERVER);
};

#endif
