 /*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/


// NewPKIStore.cpp: implementation of the NewPKIStore class.
//
//////////////////////////////////////////////////////////////////////

#include "NewPKIStore.h"
#include <PKI_ERR.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ASN1_SEQUENCE(WORKING_CA_REQUEST) = {
	ASN1_SIMPLE(WORKING_CA_REQUEST, SenderName, ASN1_UTF8STRING), 
	ASN1_SIMPLE(WORKING_CA_REQUEST, sender, X509_PUBKEY), 
	ASN1_SIMPLE(WORKING_CA_REQUEST, transactionID, ASN1_OCTET_STRING), 
	ASN1_SIMPLE(WORKING_CA_REQUEST, request, NEWPKI_REQUEST), 
}ASN1_SEQUENCE_END(WORKING_CA_REQUEST)
IMPLEMENT_ASN1_FUNCTIONS(WORKING_CA_REQUEST)
IMPLEMENT_PEM_CONVERT(WORKING_CA_REQUEST, "WORKING NEWPKI REQUEST")


NewPKIStore::NewPKIStore()
{
	m_DbConn = NULL;
}

NewPKIStore::~NewPKIStore()
{
	if(m_DbConn)
		delete m_DbConn;
}

bool NewPKIStore::BINARYtoString(const unsigned char * datas, int len, mString &strId)
{
	//Convert the transactionID to hex
	char tt[STR_LONG_LEN];

	strId = "";
	for(int i=0; i < len; i++)
	{
		sprintf(tt, "%.2x", datas[i]);
		strId += tt;
	}

	return true;
}

unsigned char * NewPKIStore::StringtoBINARY(const mString & strId, int & len)
{
	char tt[3];
	unsigned char c;
	unsigned char * dstBuffer;
	unsigned char * Result;
	const char * srcBuffer;
	int i;

	Result = (unsigned char*)malloc(strId.size()/2 + 1);
	if(!Result)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return NULL;
	}
	if(!strId.size())
	{
		Result[0] = '\0';
		len = 0;
		return Result;
	}

	srcBuffer = strId.c_str();
	len = strId.size();
	tt[2]=0;
	for(i=0; i < len; i=i+2)
	{
		memcpy(tt, &(srcBuffer[i]), 2);
		c = (unsigned char)strtoul(tt, NULL, 16);
		dstBuffer = &Result[i/2];
		memcpy(dstBuffer, &c, sizeof(c));
	}

	len = i/2;
	return Result;
}

ASN1_BIT_STRING * NewPKIStore::StringtoASN1_BIT_STRING(const mString & strId)
{
	unsigned char * Result;
	ASN1_BIT_STRING * ret;
	int len;

	Result = StringtoBINARY(strId, len);
	if(!Result)
		return NULL;

	ret = ASN1_BIT_STRING_new();
	if(!ret)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		free(Result);
		return NULL;
	}
	if(ASN1_BIT_STRING_set(ret, (unsigned char*)Result, len) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		free(Result);
		ASN1_BIT_STRING_free(ret);
		return NULL;
	}
	free(Result);
	return ret;
}

bool NewPKIStore::ASN1_BIT_STRINGtoString(const ASN1_BIT_STRING * bit, mString &strId)
{
	//Convert the transactionID to hex
	return BINARYtoString(bit->data, bit->length, strId);
}


ASN1_OCTET_STRING * NewPKIStore::StringtoTransactionID(const mString & strId)
{
	unsigned char * Result;
	ASN1_OCTET_STRING * ret;
	int len;

	Result = StringtoBINARY(strId, len);
	if(!Result)
		return NULL;
	ret = ASN1_BIT_STRING_new();
	if(!ret)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		free(Result);
		return NULL;
	}
	if(ASN1_OCTET_STRING_set(ret, (unsigned char*)Result, len) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		free(Result);
		ASN1_OCTET_STRING_free(ret);
		return NULL;
	}
	free(Result);
	return ret;
}

bool NewPKIStore::transactionIDtoString(const ASN1_OCTET_STRING *transactionID, mString &strId)
{
	//Convert the transactionID to hex
	if(transactionID->length > MAX_TID_LEN)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return false;
	}
	return BINARYtoString(transactionID->data, transactionID->length, strId);
}

bool NewPKIStore::EVP_PKEYtoString(const EVP_PKEY * pubKey, mString &strKey)
{
	X509_PUBKEY * x_pubkey=NULL;
	if(X509_PUBKEY_set(&x_pubkey, (EVP_PKEY*)pubKey) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!X509_PUBKEYtoString(x_pubkey, strKey))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		X509_PUBKEY_free(x_pubkey);
		return false;
	}
	X509_PUBKEY_free(x_pubkey);
	return true;
}

bool NewPKIStore::X509_PUBKEYtoString(const X509_PUBKEY * pubKey, mString &strKey)
{
	return ASN1_BIT_STRINGtoString(pubKey->public_key, strKey);
}


bool NewPKIStore::X509_NAMEtoString(const X509_NAME * dn, mString &strHash)
{
	unsigned char * datas, * p;
	int len;
	unsigned char sha1[SHA_DIGEST_LENGTH];

	if( (len = i2d_X509_NAME((X509_NAME*)dn, NULL)) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	datas = (unsigned char*)malloc(len);
	if(!datas)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	p = datas;
	if( (len = i2d_X509_NAME((X509_NAME*)dn, &p)) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		free(datas);
		return false;
	}

	

	if(!SHA1(datas, len, sha1))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		free(datas);
		return false;
	}
	free(datas);

	strHash = "";
	return BINARYtoString(sha1, SHA_DIGEST_LENGTH, strHash);
}

void NewPKIStore::SetEntityCert(const PKI_CERT & EntityCert)
{
	m_EntityCert = EntityCert;
}

bool NewPKIStore::CreateRequester(const SQL_Connection * DbConn)
{
	SQL sql(DbConn);
	long i;
	char * CommonCreates[] = {NEWPKISTORE_CREATE_2, NULL};


	//We execute each request
	for(i=0; CommonCreates[i]; i++)
	{
		if(!sql.Execute(CommonCreates[i]))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}

	return true;
}

bool NewPKIStore::CreateResponder(const SQL_Connection * DbConn)
{
	SQL sql(DbConn);
	long i;
	char * CommonCreates[] = {NEWPKISTORE_CREATE_1, NULL};


	//We execute each request
	for(i=0; CommonCreates[i]; i++)
	{
		if(!sql.Execute(CommonCreates[i]))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}

	return true;
}

bool NewPKIStore::SetDbConn(const SQL_Connection * DbConn)
{
	if(m_DbConn)
	{
		delete m_DbConn;
		m_DbConn = NULL;
	}

	try
	{
		m_DbConn = DbConn->Clone();
	}
	catch(ExceptionNewPKI e)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!m_DbConn)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	return true;
}



bool NewPKIStore::Responder_InsertRequest(const WORKING_CA_REQUEST * Request)
{
	SQL sql(m_DbConn);
	mString req;
	char * req_pem;
	mString strTransactionId;

	if(!transactionIDtoString(Request->transactionID, strTransactionId))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	req_pem = WORKING_CA_REQUEST_To_PEM(Request);
	if(!req_pem)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	req.sprintf(RESPONDER_NEWPKISTORE_INSERT_REQ, strTransactionId.c_str(), TRANSACTION_STATUS_WAITING, req_pem);
	free(req_pem);
	
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	return true;
}

bool NewPKIStore::Responder_GetWaitingRequests(STACK_OF(WORKING_CA_REQUEST) * Requests)
{
	long NumRows;
	SQL sql(m_DbConn);
	mString req;
	long i;
	mString strReq;
	WORKING_CA_REQUEST * request;

	req.sprintf(RESPONDER_NEWPKISTORE_GET_REQS_BY_STATE, TRANSACTION_STATUS_WAITING);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	for(i=0; i<NumRows; i++)
	{
		if(!sql.Value(i,"request", strReq))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}

		request = NULL;
		if(!WORKING_CA_REQUEST_From_PEM(&request, strReq.c_str()))
			continue;

		if(SKM_sk_push(WORKING_CA_REQUEST, Requests, request) < 0)
		{
			WORKING_CA_REQUEST_free(request);
		}
	}

	return true;
}

bool NewPKIStore::Responder_GetUnsentResponses(STACK_OF(STORED_NEWPKI_RESPONSE) * Responses)
{
	long NumRows;
	SQL sql(m_DbConn);
	mString req;
	long i;
	mString strReq;
	STORED_NEWPKI_RESPONSE * response;

	req.sprintf(RESPONDER_NEWPKISTORE_GET_RESP_BY_STATE, TRANSACTION_STATUS_PROCESSED);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	for(i=0; i<NumRows; i++)
	{
		if(!sql.Value(i,"response", strReq))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}

		response = NULL;
		if(!STORED_NEWPKI_RESPONSE_From_PEM(&response, strReq.c_str()))
			continue;

		if(SKM_sk_push(STORED_NEWPKI_RESPONSE, Responses, response) < 0)
		{
			STORED_NEWPKI_RESPONSE_free(response);
		}
	}

	return true;
}

bool NewPKIStore::Responder_SetResponse(const STORED_NEWPKI_RESPONSE * Response)
{
	SQL sql(m_DbConn);
	mString req;
	mString strTransactionId;
	char * resp_pem;

	if(!transactionIDtoString(Response->transactionID, strTransactionId))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	resp_pem = STORED_NEWPKI_RESPONSE_To_PEM(Response);
	if(!resp_pem)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	req.sprintf(RESPONDER_NEWPKISTORE_SET_RESPONSE, TRANSACTION_STATUS_PROCESSED, resp_pem, strTransactionId.c_str());
	free(resp_pem);

	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool NewPKIStore::Responder_ResponseSent(const ASN1_OCTET_STRING * TransactionID)
{
	SQL sql(m_DbConn);
	mString req;
	mString strTransactionId;

	if(!transactionIDtoString(TransactionID, strTransactionId))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	req.sprintf(RESPONDER_NEWPKISTORE_SET_STATUS, TRANSACTION_STATUS_SENT, strTransactionId.c_str());
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool NewPKIStore::Responder_IsRequestKnown(const ASN1_OCTET_STRING * TransactionID, bool & IsKnown)
{
	SQL sql(m_DbConn);
	mString req;
	mString strTransactionId;
	long NumRows;

	if(!transactionIDtoString(TransactionID, strTransactionId))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	req.sprintf(RESPONDER_NEWPKISTORE_GET_TID, strTransactionId.c_str());
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!NumRows)
		IsKnown = false;
	else
		IsKnown = true;

	return true;
}


bool NewPKIStore::Responder_GetTransactionIds(TRANSACTION_IDS *transactionIDs)
{
	long NumRows;
	SQL sql(m_DbConn);
	mString req;
	long i;
	mString strTransactionID;
	ASN1_OCTET_STRING * transactionID;

	req.sprintf(RESPONDER_NEWPKISTORE_GET_TIDS);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	for(i=0; i<NumRows; i++)
	{
		if(!sql.Value(i,"transactionID", strTransactionID))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}

		if(strTransactionID.size())
		{
			transactionID = StringtoTransactionID(strTransactionID);
			if(!transactionID) continue;
			if(SKM_sk_push(ASN1_OCTET_STRING, transactionIDs->transactionIds, transactionID) < 0)
			{
				ASN1_OCTET_STRING_free(transactionID);
			}
		}
	}

	return true;
}







bool NewPKIStore::Requester_InsertRequest(const CRYPTED_NEWPKI_REQUEST * Request, unsigned long priv_attr, unsigned long & req_id)
{
	SQL sql(m_DbConn);
	mString req;
	char * req_pem;
	
	req_pem = CRYPTED_NEWPKI_REQUEST_To_PEM(Request);
	if(!req_pem)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}	
	req.sprintf(REQUESTER_NEWPKISTORE_INSERT_REQ, TRANSACTION_STATUS_WAITING, req_pem, priv_attr);
	free(req_pem);
	
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	req_id = sql.GetLastID();
	
	return true;
}

bool NewPKIStore::Requester_RequestSent(unsigned long req_id, const ASN1_OCTET_STRING *transactionID)
{
	SQL sql(m_DbConn);
	mString req;
	mString strTransactionId;

	if(!transactionIDtoString(transactionID, strTransactionId))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	req.sprintf(REQUESTER_NEWPKISTORE_SET_REQ_TID, TRANSACTION_STATUS_SENT, strTransactionId.c_str(), req_id);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool NewPKIStore::Requester_GetWaitingTransactionIds(TRANSACTION_IDS *transactionIDs)
{
	long NumRows;
	SQL sql(m_DbConn);
	mString req;
	long i;
	mString strTransactionID;
	ASN1_OCTET_STRING * transactionID;

	req.sprintf(REQUESTER_NEWPKISTORE_GET_TIDS_BY_STATE, TRANSACTION_STATUS_SENT);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	for(i=0; i<NumRows; i++)
	{
		if(!sql.Value(i,"transactionID", strTransactionID))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}

		if(strTransactionID.size())
		{
			transactionID = StringtoTransactionID(strTransactionID);
			if(!transactionID) continue;
			if(SKM_sk_push(ASN1_OCTET_STRING, transactionIDs->transactionIds, transactionID) < 0)
			{
				ASN1_OCTET_STRING_free(transactionID);
			}
		}
	}

	return true;
}

bool NewPKIStore::Requester_DeleteTransaction(unsigned long req_id)
{
	SQL sql(m_DbConn);
	mString req;
	
	req.sprintf(REQUESTER_NEWPKISTORE_DELETE_REQ, req_id);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool NewPKIStore::Requester_DeleteTransactionPrivAttr(unsigned long priv_attr)
{
	SQL sql(m_DbConn);
	mString req;
	
	req.sprintf(REQUESTER_NEWPKISTORE_DELETE_REQ_PRIV_ATTR, priv_attr);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool NewPKIStore::Requester_GetWaitingRequests(STACK_OF(CRYPTED_NEWPKI_REQUEST) * Requests)
{
	long NumRows;
	SQL sql(m_DbConn);
	mString req;
	long i;
	mString strReq;
	CRYPTED_NEWPKI_REQUEST * request;

	req.sprintf(REQUESTER_NEWPKISTORE_GET_REQS_BY_STATE, TRANSACTION_STATUS_WAITING);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	for(i=0; i<NumRows; i++)
	{
		if(!sql.Value(i,"request", strReq))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}

		request = NULL;
		if(!CRYPTED_NEWPKI_REQUEST_From_PEM(&request, strReq.c_str()))
			continue;

		if(SKM_sk_push(CRYPTED_NEWPKI_REQUEST, Requests, request) < 0)
		{
			CRYPTED_NEWPKI_REQUEST_free(request);
		}
	}

	return true;
}


bool NewPKIStore::Requester_RequestProcessed(const ASN1_OCTET_STRING *transactionID)
{
	SQL sql(m_DbConn);
	mString req;
	mString strTransactionId;

	if(!transactionIDtoString(transactionID, strTransactionId))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	req.sprintf(REQUESTER_NEWPKISTORE_SET_REQ_STATE, TRANSACTION_STATUS_PROCESSED, strTransactionId.c_str());
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool NewPKIStore::GenerateTransactionID(ASN1_OCTET_STRING * transactionID, const X509_PUBKEY *sender, const X509_PUBKEY *recipient, unsigned long counter)
{
	unsigned char sha1[SHA_DIGEST_LENGTH];
	unsigned char new_transactionID[(SHA_DIGEST_LENGTH*3) + sizeof(counter) + 250];
	unsigned char * currPtr;


	currPtr = new_transactionID;
	if(!SHA1(m_EntityCert.GetPrivateKey().GetPublicKey()->public_key->data, m_EntityCert.GetPrivateKey().GetPublicKey()->public_key->length, currPtr))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	currPtr += SHA_DIGEST_LENGTH;
	if(!SHA1(recipient->public_key->data, recipient->public_key->length, currPtr))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	currPtr += SHA_DIGEST_LENGTH;
	if(!SHA1(sender->public_key->data, sender->public_key->length, currPtr))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	currPtr += SHA_DIGEST_LENGTH;
	
	memcpy(currPtr, &counter, sizeof(counter));
	currPtr += sizeof(counter);
	RAND_bytes(currPtr, 250);
	
	// Calculate the SHA1 of the whole thing to be sure it's not to big
	if(!SHA1(new_transactionID, sizeof(new_transactionID), sha1))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(ASN1_OCTET_STRING_set(transactionID, sha1, sizeof(sha1)) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}
