/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/


// NewPKIStore.h: interface for the NewPKIStore class.
//
//////////////////////////////////////////////////////////////////////

#ifndef NEWPKISTORE_H
#define NEWPKISTORE_H

#include <openssl/x509.h>
#include <openssl/evp.h>
#include <mString.h>
#include <ASN1/ASN1_cert.h>
#include <ASN1/ASN1_conf.h>
#include <PKI_CERT.h>
#include <SQL/SQL_Conn.h>


typedef struct st_WORKING_CA_REQUEST
{
	ASN1_UTF8STRING * SenderName;
	X509_PUBKEY * sender;
	ASN1_OCTET_STRING * transactionID;
	NEWPKI_REQUEST * request;
}WORKING_CA_REQUEST;
DECLARE_ASN1_FUNCTIONS(WORKING_CA_REQUEST)
DECLARE_PEM_CONVERT(WORKING_CA_REQUEST)


/*!
	This class holds the basic functionnalities for a requester and a responder
*/
class NewPKIStore  
{
public:
	/*! \brief This is the constructor.
	 */
	NewPKIStore();

	/*! \brief This is the destructor.
	 */
	virtual ~NewPKIStore();

	/*! \brief This function creates the tables needed to work.
	 *  \param DbConn [IN] A SQL connection.
	 *  \return true on success, false on failure.
	 */
	virtual bool CreateTables(const SQL_Connection * DbConn)=0;

	/*! \brief This function converts binary datas to a string.
	 *  \param datas [IN] The binary datas.
	 *  \param len [IN] The binary datas length.
	 *  \param strId [OUT] The result string.
	 *  \return true on success, false on failure.
	 */
	static bool BINARYtoString(const unsigned char * datas, int len, mString &strId);

	/*! \brief This function converts a string to binary datas.
	 *  \param strId [IN] The string.
	 *  \param len [OUT] The binary datas length.
	 *  \return The binary datas or NULL on failure.
	 */
	static unsigned char * StringtoBINARY(const mString & strId, int & len);

	/*! \brief This function converts a ASN1_BIT_STRING to a string.
	 *  \param bit [IN] The ASN1_BIT_STRING.
	 *  \param strId [OUT] The result string.
	 *  \return true on success, false on failure.
	 */
	static bool ASN1_BIT_STRINGtoString(const ASN1_BIT_STRING * bit, mString &strId);

	/*! \brief This function converts a string to a ASN1_BIT_STRING.
	 *  \param strId [IN] The string.
	 *  \return The ASN1_BIT_STRING or NULL on failure.
	 */
	static ASN1_BIT_STRING * StringtoASN1_BIT_STRING(const mString & strId);
	
	/*! \brief This function converts a string to a ASN1_OCTET_STRING.
	 *  \param strId [IN] The string.
	 *  \return The ASN1_OCTET_STRING or NULL on failure.
	 */
	static ASN1_OCTET_STRING * StringtoTransactionID(const mString & strId);
	
	/*! \brief This function converts a ASN1_OCTET_STRING to a string.
	 *  \param transactionID [IN] The ASN1_OCTET_STRING.
	 *  \param strId [OUT] The result string.
	 *  \return true on success, false on failure.
	 */
	static bool transactionIDtoString(const ASN1_OCTET_STRING * transactionID, mString & strId);

	/*! \brief This function converts an EVP_PKEY to a string.
	 *  \param pubKey [IN] The EVP_PKEY.
	 *  \param strKey [OUT] The result string.
	 *  \return true on success, false on failure.
	 */
	static bool EVP_PKEYtoString(const EVP_PKEY * pubKey, mString &strKey);

	/*! \brief This function converts a X509_PUBKEY to a string.
	 *  \param pubKey [IN] The X509_PUBKEY.
	 *  \param strKey [OUT] The result string.
	 *  \return true on success, false on failure.
	 */
	static bool X509_PUBKEYtoString(const X509_PUBKEY * pubKey, mString &strKey);

	/*! \brief This function converts Binary a X509_NAME to a string.
	 *  \param dn [IN] The X509_NAME.
	 *  \param strHash [OUT] The result string.
	 *  \return true on success, false on failure.
	 */
	static bool X509_NAMEtoString(const X509_NAME * dn, mString &strHash);

	
	/*! \brief This function sets the SQL connection.
	 *  \param DbConn [IN] A SQL connection.
	 *  \return true on success, false on failure.
	 */
	bool SetDbConn(const SQL_Connection * DbConn);

	/*! \brief This function sets the certificate of the entity using this class.
	 *  \param EntityCert [IN] The certificate of the entity.
	 */
	void SetEntityCert(const PKI_CERT & EntityCert);

	/*! \brief This function is used to get the list of known responses.
	 *  \param transactionIDs [OUT] The responded transaction Ids.
	 *  \return true on success, false on failure.
	 */
	bool Responder_GetTransactionIds(TRANSACTION_IDS *transactionIDs);


	/*! \brief This function inserts a new request to be treated.
	 *  \param Request [IN] The request.
	 *  \return true on success, false on failure.
	 */
	bool Responder_InsertRequest(const WORKING_CA_REQUEST * Request);

	/*! \brief This function returns the requests that haven't been processed yet.
	 *  \param Requests [OUT] The un-processed requests.
	 *  \return true on success, false on failure.
	 */
	bool Responder_GetWaitingRequests(STACK_OF(WORKING_CA_REQUEST) * Requests);

	/*! \brief This function returns the responses that haven't been sent yet.
	 *  \param Responses [OUT] The unsent responses.
	 *  \return true on success, false on failure.
	 */
	bool Responder_GetUnsentResponses(STACK_OF(STORED_NEWPKI_RESPONSE) * Responses);

	/*! \brief This function set a response.
	 *  \param Response [IN] The new response.
	 *  \return true on success, false on failure.
	 */
	bool Responder_SetResponse(const STORED_NEWPKI_RESPONSE * Response);

	/*! \brief This function marks a response as sent.
	 *  \param TransactionID [IN] The ID of the sent response.
	 *  \return true on success, false on failure.
	 */
	bool Responder_ResponseSent(const ASN1_OCTET_STRING * TransactionID);

	/*! \brief This function check if a response is known.
	 *  \param TransactionID [IN] The ID of the response.
	 *  \param IsKnown [OUT] true when known, false when unknown.
	 *  \return true on success, false on failure.
	 */
	bool Responder_IsRequestKnown(const ASN1_OCTET_STRING * TransactionID, bool & IsKnown);

	/*! \brief This function inserts a new request.
	 *  \param Request [IN] The request.
	 *  \param priv_attr [IN] A private value used by children.
	 *  \param req_id [OUT] The internal ID of the new request.
	 *  \return true on success, false on failure.
	 */
	bool Requester_InsertRequest(const CRYPTED_NEWPKI_REQUEST * Request, unsigned long priv_attr, unsigned long & req_id);

	/*! \brief This functions returns the list of requests for which we're expecting a response.
	 *  \param transactionIDs [OUT] The un-responded transaction Ids.
	 *  \return true on success, false on failure.
	 */
	bool Requester_GetWaitingTransactionIds(TRANSACTION_IDS *transactionIDs);
	
	/*! \brief This function marks the request as sent an assign it the transactionID given by the Respository.
	 *  \param req_id [IN] The internal ID of the request.
	 *  \param transactionID [IN] The transaction Id returned by the Repository.
	 *  \return true on success, false on failure.
	 */
	bool Requester_RequestSent(unsigned long req_id, const ASN1_OCTET_STRING *transactionID);
	
	/*! \brief This function is used to delete a request.
	 *  \param req_id [IN] The internal ID of the request.
	 *  \return true on success, false on failure.
	 */
	bool Requester_DeleteTransaction(unsigned long req_id);
	
	/*! \brief This function is used to delete a request matching priv_attr.
	 *  \param priv_attr [IN] A private value used by children.
	 *  \return true on success, false on failure.
	 */
	bool Requester_DeleteTransactionPrivAttr(unsigned long priv_attr);
	
	/*! \brief This function returns the requests that haven't been sent yet.
	 *  \param Requests [OUT] The unsent requests.
	 *  \return true on success, false on failure.
	 */
	bool Requester_GetWaitingRequests(STACK_OF(CRYPTED_NEWPKI_REQUEST) * Requests);
	
	/*! \brief This function marks a request as having been processed by the responder.
	 *  \param transactionID [IN] The processed request.
	 *  \return true on success, false on failure.
	 */
	bool Requester_RequestProcessed(const ASN1_OCTET_STRING *transactionID);

protected:
	#define RESPONDER_TABLE "responder"
	#define REQUESTER_TABLE "requester"
	#define REQUESTER_PRIV_ATTR "priv_attr"
	#define REQUESTER_TID "transactionID"
	#define MAX_TID_LEN SHA_DIGEST_LENGTH

	PKI_CERT m_EntityCert;
	SQL_Connection * m_DbConn;
	bool CreateRequester(const SQL_Connection * DbConn);
	bool CreateResponder(const SQL_Connection * DbConn);
	bool GenerateTransactionID(ASN1_OCTET_STRING * transactionID, const X509_PUBKEY *sender, const X509_PUBKEY *recipient, unsigned long counter);

private:
	#define NEWPKISTORE_CREATE_1						"create table "RESPONDER_TABLE" (req_id INT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY, transactionID VARCHAR(41) NOT NULL, status INT UNSIGNED NOT NULL, request LONGBLOB NOT NULL, response LONGBLOB NOT NULL);"
	#define NEWPKISTORE_CREATE_2						"create table "REQUESTER_TABLE" (req_id INT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY, "REQUESTER_TID" VARCHAR(41) NOT NULL, "REQUESTER_PRIV_ATTR" INT UNSIGNED NOT NULL, status INT UNSIGNED NOT NULL, request LONGBLOB NOT NULL);"

	#define RESPONDER_NEWPKISTORE_INSERT_REQ			"INSERT INTO "RESPONDER_TABLE" (transactionID, status, request, response) VALUES ('%s', %d, '%s', '');"
	#define RESPONDER_NEWPKISTORE_GET_TIDS				"SELECT transactionID FROM "RESPONDER_TABLE";"
	#define RESPONDER_NEWPKISTORE_SET_RESPONSE			"UPDATE "RESPONDER_TABLE" SET status='%d', response='%s', request='' WHERE transactionID='%s';"
	#define RESPONDER_NEWPKISTORE_SET_STATUS			"UPDATE "RESPONDER_TABLE" SET status='%d', response='' WHERE transactionID='%s';"
	#define RESPONDER_NEWPKISTORE_GET_REQS_BY_STATE		"SELECT request FROM "RESPONDER_TABLE" WHERE status='%d' ORDER BY req_id DESC LIMIT 5;"
	#define RESPONDER_NEWPKISTORE_GET_RESP_BY_STATE		"SELECT response FROM "RESPONDER_TABLE" WHERE status='%d';"
	#define RESPONDER_NEWPKISTORE_GET_TID				"SELECT req_id FROM "RESPONDER_TABLE" WHERE transactionID='%s';"
	
	#define REQUESTER_NEWPKISTORE_INSERT_REQ			"INSERT INTO "REQUESTER_TABLE" ("REQUESTER_TID", status, request, "REQUESTER_PRIV_ATTR") VALUES ('', %d, '%s', '%ld');"
	#define REQUESTER_NEWPKISTORE_GET_TIDS_BY_STATE		"SELECT "REQUESTER_TID" FROM "REQUESTER_TABLE" WHERE status='%d' ORDER BY req_id DESC;"
	#define REQUESTER_NEWPKISTORE_SET_REQ_TID			"UPDATE "REQUESTER_TABLE" SET status='%d', "REQUESTER_TID"='%s', request='' WHERE req_id='%ld';"
	#define REQUESTER_NEWPKISTORE_SET_TID_STATUS		"UPDATE "REQUESTER_TABLE" SET status='%d' WHERE "REQUESTER_TID"='%s';"
	#define REQUESTER_NEWPKISTORE_DELETE_REQ			"DELETE FROM "REQUESTER_TABLE" WHERE req_id='%ld';"
	#define REQUESTER_NEWPKISTORE_GET_REQS_BY_STATE		"SELECT request FROM "REQUESTER_TABLE" WHERE status='%d' ORDER BY req_id DESC;"
	#define REQUESTER_NEWPKISTORE_SET_REQ_STATE			"UPDATE "REQUESTER_TABLE" SET status='%d' WHERE "REQUESTER_TID"='%s';"
	#define REQUESTER_NEWPKISTORE_DELETE_REQ_PRIV_ATTR	"DELETE FROM "REQUESTER_TABLE" WHERE "REQUESTER_PRIV_ATTR"='%ld';"

	typedef enum en_TRANSACTION_STATUS
	{
		TRANSACTION_STATUS_WAITING=1,
		TRANSACTION_STATUS_PROCESSED=2,
		TRANSACTION_STATUS_SENT=3
	}TRANSACTION_STATUS;
};

#endif
