/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/

// RaStore.cpp: implementation of the RaStore class.
//
//////////////////////////////////////////////////////////////////////

#include "RaStore.h"
#include <PKI_PKCS12.h>
#include <PKI_P7B.h>
#include "svintl.h"

#include <vector>
using namespace std;


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ASN1_SEQUENCE(RA_STORE_SIG) = {
	ASN1_SIMPLE(RA_STORE_SIG, signature, ASN1_BIT_STRING),
	ASN1_SIMPLE(RA_STORE_SIG, sig_alg, X509_ALGOR),
}ASN1_SEQUENCE_END(RA_STORE_SIG)
IMPLEMENT_ASN1_FUNCTIONS(RA_STORE_SIG);
IMPLEMENT_PEM_CONVERT(RA_STORE_SIG, "NEWPKI RA STORE SIGNATURE");

ASN1_SEQUENCE(RA_STORE_INFO) = {
	ASN1_SIMPLE(RA_STORE_INFO, hash, ASN1_UTF8STRING),
	ASN1_SIMPLE(RA_STORE_INFO, counter, ASN1_INTEGER),
}ASN1_SEQUENCE_END(RA_STORE_INFO)
IMPLEMENT_ASN1_FUNCTIONS(RA_STORE_INFO);
IMPLEMENT_ASN1_ENCRYPT_SIGN(RA_STORE_INFO);

ASN1_SEQUENCE(RA_STORE_PRIVDATAS) = {
	ASN1_SIMPLE(RA_STORE_PRIVDATAS, passwd, ASN1_UTF8STRING),
	ASN1_SIMPLE(RA_STORE_PRIVDATAS, privkey, RSAPrivateKey),
}ASN1_SEQUENCE_END(RA_STORE_PRIVDATAS)
IMPLEMENT_ASN1_FUNCTIONS(RA_STORE_PRIVDATAS);
IMPLEMENT_ASN1_ENCRYPT_SIGN(RA_STORE_PRIVDATAS);



RaStore::RaStore():NewPKIStore()
{
	ProfileInfo = NULL;
	m_groups = NULL;
	m_AclValidator = NULL;
}

RaStore::~RaStore()
{
	if(ProfileInfo)
		RA_STORE_INFO_free(ProfileInfo);
	if(m_groups)
		USERS_GROUPS_free(m_groups);
}

bool RaStore::CreateTables(const SQL_Connection * DbConn)
{
	SQL sql(DbConn);
	long i;
	char * CommonCreates[] = {RASTORE_CREATE_1, RASTORE_CREATE_2, RASTORE_CREATE_3, NULL};

	if(!CreateRequester(DbConn))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//We execute each request
	for(i=0; CommonCreates[i]; i++)
	{
		if(!sql.Execute(CommonCreates[i]))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}

	return true;
}

bool RaStore::Create()
{
	ProfileHash.SetCert(m_EntityCert);

	if(ProfileInfo)
		RA_STORE_INFO_free(ProfileInfo);
	ProfileInfo =  RA_STORE_INFO_new();
	if(!ProfileInfo)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	ASN1_INTEGER_SET(&ProfileInfo->counter, 0);

	if(!Flush_ProfileInfo())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}



bool RaStore::Load()
{
	SQL sql(m_DbConn);
	mString pem_hash;
	long NumRows;
	ASN1_ENCRYPT_SIGN * crypted_conf;

	ConfAccessLock.LockWrite();

	ProfileHash.SetCert(m_EntityCert);

	//Load the requests info
	if(!sql.Execute(RASTORE_GET_PROFILES_HASH))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	if(!sql.NumRows( &NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	if(!NumRows)
	{
		// TODO crer en fonction des profils deja existants voir RepStore::RegenerateHash
		if(!Create())
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ConfAccessLock.Unlock();
			return false;
		}
	}
	else
	{
		if(!sql.Value(0,"hash", pem_hash))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ConfAccessLock.Unlock();
			return false;
		}

		crypted_conf = NULL;
		if(!ASN1_ENCRYPT_SIGN_From_PEM(&crypted_conf, pem_hash.c_str()))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ASN1_ENCRYPT_SIGN_free(crypted_conf);
			ConfAccessLock.Unlock();
			return false;
		}

		//We now have the hash, with there encrypted body
		//let's decrypt it
		//Decrypt the body and verify signature
		if(ProfileInfo)
		{
			RA_STORE_INFO_free(ProfileInfo);
			ProfileInfo = NULL;

		}

		if(RA_STORE_INFO_decrypt_verify(crypted_conf, &ProfileInfo, m_EntityCert.GetPrivateKey().GetRsaKey(), m_EntityCert.GetPrivateKey().GetRsaKey()) <= 0)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ASN1_ENCRYPT_SIGN_free(crypted_conf);
			ConfAccessLock.Unlock();
			return false;
		}
		ASN1_ENCRYPT_SIGN_free(crypted_conf);
	}
	pem_hash = ProfileInfo->hash;
	if(!ProfileHash.SetCurrentHash(pem_hash))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	ConfAccessLock.Unlock();
	return true;
}


bool RaStore::Flush_ProfileInfo()
{
	ConfAccessLock.LockWrite();
	SQL sql(m_DbConn);
	mString hash;
	mString req;
	char * pem_info;
	ASN1_ENCRYPT_SIGN * crypt;

	if(!ProfileHash.GetHash(hash))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	
	if(ASN1_STRING_set(ProfileInfo->hash, hash.c_str(), hash.size()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	crypt = ASN1_ENCRYPT_SIGN_new();
	if(!crypt)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	if(!RA_STORE_INFO_sign_encrypt(crypt, ProfileInfo, m_EntityCert.GetPrivateKey().GetRsaKey(), m_EntityCert.GetPublicKey(), EVP_sha1(), EVP_des_ede3_cbc()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ASN1_ENCRYPT_SIGN_free(crypt);
		ConfAccessLock.Unlock();
		return false;
	}
	pem_info = ASN1_ENCRYPT_SIGN_To_PEM(crypt);
	if(!pem_info)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ASN1_ENCRYPT_SIGN_free(crypt);
		ConfAccessLock.Unlock();
		return false;
	}
	ASN1_ENCRYPT_SIGN_free(crypt);


	if(req.sprintf(RASTORE_INSERT_PROFILES_HASH, pem_info) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		free(pem_info);
		ConfAccessLock.Unlock();
		return false;
	}
	free(pem_info);

	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}


void RaStore::SetMinPasswdLen(unsigned long MinPasswdLen)
{
	ConfAccessLock.LockWrite();
	m_MinPasswdLen = MinPasswdLen;
	ConfAccessLock.Unlock();
}

bool RaStore::SetGroups(const USERS_GROUPS * groups)
{
	ConfAccessLock.LockWrite();
	if(!USERS_GROUPS_cpy(&m_groups, groups))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}

bool RaStore::DeleteProfile(long ProfileId)
{
	ConfAccessLock.LockWrite();
	if(!ProfileInfo || !ProfileHash)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		ConfAccessLock.Unlock();
		return false;
	}

	SQL sql(m_DbConn);
	long NumRows;
	unsigned Offset = 0;
	int i;
	HashCorrelation tmpCorrelation;
	mString signature;
	mString id;
	mString req;
	RA_STORE_SIG * sig;

	if(req.sprintf(RASTORE_DELETE_PROFILE, ProfileId) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	//We need to recalculate the whole hash
	tmpCorrelation.SetCert(m_EntityCert);

	do
	{
		if(req.sprintf(RASTORE_GET_ALL_PROFILES, Offset*200, 200) <= 0)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			ConfAccessLock.Unlock();
			return false;
		}
		if(!sql.Execute(req))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ConfAccessLock.Unlock();
			return false;
		}
		if(!sql.NumRows(&NumRows))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ConfAccessLock.Unlock();
			return false;
		}
		if(NumRows)
		{
			for(i=0; i<NumRows; i++)
			{
				if(!sql.Value(i,"profile_id", id))
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					ConfAccessLock.Unlock();
					return false;
				}
				if(!sql.Value(i,"signature", signature))
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					ConfAccessLock.Unlock();
					return false;
				}

				sig = NULL;
				if(RA_STORE_SIG_From_PEM(&sig, signature.c_str()) <= 0)
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					ConfAccessLock.Unlock();
					return false;
				}
				if(!tmpCorrelation.AddEntry(id.c_ulng(), (char*)sig->signature->data, sig->signature->length))
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					RA_STORE_SIG_free(sig);
					ConfAccessLock.Unlock();
					return false;
				}
				RA_STORE_SIG_free(sig);
			}
		}
		Offset++;
	}
	while(NumRows);

	if(!tmpCorrelation.GetHash(signature))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ProfileHash.Clear();
	if(!ProfileHash.SetCurrentHash(signature))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	if(!Flush_ProfileInfo())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}

bool RaStore::InsertProfile(const PKI_CERT & UserCert, const NEWPKI_PROFILE * Profile, long & ProfileId)
{
	ConfAccessLock.LockWrite();
	if(!ProfileInfo || !ProfileHash)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		ConfAccessLock.Unlock();
		return false;
	}
	SQL sql(m_DbConn);
	mString req;
	mString strHash;
	mString strDN;
	mString strUID;
	bool Exists;
	char * pem_datas;
	char * pem_signature;
	RA_STORE_SIG * signature;
	char * dn;

	// Search if the profile is unique
	if(!X509_NAMEtoString(Profile->dn, strHash))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	if(!ProfileExists(strHash, Exists))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	if(Exists)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_UNIQUE_PROFILE);
		ConfAccessLock.Unlock();
		return false;
	}

	signature = RA_STORE_SIG_new();
	if(!signature)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	//Get the signature into pem
	if(!NEWPKI_PROFILE_sign(Profile, signature, (EVP_PKEY*)m_EntityCert.GetPrivateKey().GetRsaKey()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		RA_STORE_SIG_free(signature);
		ConfAccessLock.Unlock();
		return false;
	}
	pem_signature = RA_STORE_SIG_To_PEM(signature);
	if(!pem_signature)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		RA_STORE_SIG_free(signature);
		ConfAccessLock.Unlock();
		return false;
	}

	pem_datas = NEWPKI_PROFILE_To_PEM(Profile);
	if(!pem_datas)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		free(pem_signature);
		RA_STORE_SIG_free(signature);
		ConfAccessLock.Unlock();
		return false;
	}

	dn = X509_NAME_oneline(Profile->dn, NULL, 0);
	if(!dn)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		free(pem_signature);
		free(pem_datas);
		RA_STORE_SIG_free(signature);
		ConfAccessLock.Unlock();
		return false;
	}
	strDN = sql.FormatString(dn);
	free(dn);

	dn = ASN1_STRING_GET(Profile->ldap_uid);
	if(dn)
		strUID = sql.FormatString(dn);
	else
		strUID = "";

	unsigned long owner_serial;
	int owner_type;

	owner_serial = ASN1_INTEGER_GET(Profile->owner_group_serial);
	if(owner_serial)
	{
		owner_type = OWNER_TYPE_GROUP;
	}
	else
	{
		owner_serial = UserCert.GetSerial();
		if(!owner_serial)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			free(pem_signature);
			free(pem_datas);
			RA_STORE_SIG_free(signature);
			ConfAccessLock.Unlock();
			return false;
		}
		owner_type = OWNER_TYPE_USER;
	}

	if(req.sprintf(RASTORE_INSERT_PROFILE, owner_type, owner_serial, strHash.c_str(), strDN.c_str(), strUID.c_str(), ASN1_INTEGER_GET(Profile->EE_Validation)?NEWPKI_PROFILE_STATE_WAITING_FOR_VALIDATION:NEWPKI_PROFILE_STATE_VALIDATED, pem_datas, pem_signature) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		free(pem_signature);
		free(pem_datas);
		RA_STORE_SIG_free(signature);
		ConfAccessLock.Unlock();
		return false;
	}
	free(pem_signature);
	free(pem_datas);

	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		RA_STORE_SIG_free(signature);
		ConfAccessLock.Unlock();
		return false;
	}
	ProfileId = sql.GetLastID();

	// Add this entry to the global hash
	if(!ProfileHash.AddEntry(ProfileId, (char*)signature->signature->data, signature->signature->length))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		RA_STORE_SIG_free(signature);
		ConfAccessLock.Unlock();
		return false;
	}
	RA_STORE_SIG_free(signature);

	if(!Flush_ProfileInfo())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}

bool RaStore::ProfileExists(const mString & dn_hash, bool & exists)
{
	mString req;
	SQL sql(m_DbConn);
	long NumRows;

	req.sprintf(RASTORE_SEARCH_PROFILE_DN, dn_hash.c_str());

	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(NumRows)
		exists = true;
	else
		exists = false;

	return true;
}


STACK_OF(NEWPKI_PROFILE_DATAS) * RaStore::EnumProfiles(const PKI_CERT &UserCert, long index, long num)
{
	STACK_OF(NEWPKI_PROFILE_DATAS) * resStack;	
	NEWPKI_PROFILE_DATAS * res;
	long NumRows;
	SQL sql(m_DbConn);
	mString req;
	long i, j;
	mString Where;
	mString tmpDatas;
	USERS_GROUP * currGroup;
	ASN1_INTEGER * currSerial;

	// PKI Admins can see all the profile
	if(!m_AclValidator->IsPkiAdministrator(UserCert))
	{
		Where.sprintf("WHERE (owner_type=%d AND owner_serial=%ld)", OWNER_TYPE_USER, UserCert.GetSerial());

		// Create the WHERE syntax according to user's groups
		ConfAccessLock.LockRead();

		for(i=0; i<SKM_sk_num(USERS_GROUP, m_groups->list); i++)
		{
			currGroup = SKM_sk_value(USERS_GROUP, m_groups->list, i);
			if(!currGroup) continue;

			for(j=0; j<SKM_sk_num(ASN1_INTEGER, currGroup->users_serial); j++)
			{
				currSerial = SKM_sk_value(ASN1_INTEGER, currGroup->users_serial, j);
				if(!currSerial) continue;

				// Add a WHERE entry
				if((unsigned long)ASN1_INTEGER_GET(currSerial) == UserCert.GetSerial())
				{
					tmpDatas.sprintf(" OR (owner_type=%d AND owner_serial=%ld)", OWNER_TYPE_GROUP, ASN1_INTEGER_GET(currGroup->serial));
					Where += tmpDatas;
				}
			}
		}
		ConfAccessLock.Unlock();
	}


	req.sprintf(RASTORE_GET_PROFILES, Where.c_str(), index, num);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	
	resStack = SKM_sk_new_null(NEWPKI_PROFILE_INFO);
	if(!resStack)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return NULL;
	}
	
	for(i=0; i<NumRows; i++)
	{
		res = NEWPKI_PROFILE_DATAS_new();
		if(!res) continue;

		// The profile id
		if(!sql.Value(i, "profile_id", tmpDatas))
		{
			NEWPKI_PROFILE_DATAS_free(res);
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			SKM_sk_pop_free(NEWPKI_PROFILE_DATAS, resStack, NEWPKI_PROFILE_DATAS_free);
			return NULL;
		}
		if(!tmpDatas.c_ASN1_INTEGER(&res->Id))
		{
			NEWPKI_PROFILE_DATAS_free(res);
			SKM_sk_pop_free(NEWPKI_PROFILE_DATAS, resStack, NEWPKI_PROFILE_DATAS_free);
			return NULL;
		}

		// The state
		if(!sql.Value(i, "state", tmpDatas))
		{
			NEWPKI_PROFILE_DATAS_free(res);
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			SKM_sk_pop_free(NEWPKI_PROFILE_DATAS, resStack, NEWPKI_PROFILE_DATAS_free);
			return NULL;
		}
		if(!tmpDatas.c_ASN1_INTEGER(&res->State))
		{
			NEWPKI_PROFILE_DATAS_free(res);
			SKM_sk_pop_free(NEWPKI_PROFILE_DATAS, resStack, NEWPKI_PROFILE_DATAS_free);
			return NULL;
		}

		//The real profile
		if(!sql.Value(i, "datas", tmpDatas))
		{
			NEWPKI_PROFILE_DATAS_free(res);
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			SKM_sk_pop_free(NEWPKI_PROFILE_DATAS, resStack, NEWPKI_PROFILE_DATAS_free);
			return NULL;
		}
		// We need to reconvert if from the PEM
		if(NEWPKI_PROFILE_From_PEM(&res->Profile, tmpDatas.c_str()) <= 0)
		{
			NEWPKI_PROFILE_DATAS_free(res);
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			SKM_sk_pop_free(NEWPKI_PROFILE_DATAS, resStack, NEWPKI_PROFILE_DATAS_free);
			return NULL;
		}

		//Get the certs associated with the profile !
		if(!GetCerts(ASN1_INTEGER_GET(res->Id), res->Certs))
		{
			NEWPKI_PROFILE_DATAS_free(res);
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			SKM_sk_pop_free(NEWPKI_PROFILE_DATAS, resStack, NEWPKI_PROFILE_DATAS_free);
			return NULL;
		}

		if(SKM_sk_push(NEWPKI_PROFILE_DATAS, resStack, res) < 0)
		{
			NEWPKI_PROFILE_DATAS_free(res);
		}
	}
	
	return resStack;
}

bool RaStore::InsertCertReq(const char *CaName, unsigned long ProfileId, int Type, const PKI_CSR & pkcs10, const RA_STORE_PRIVDATAS *PrivDatas, const ASN1_BIT_STRING * flags, const char * admin_mail, unsigned long & CertReqId)
{
	ConfAccessLock.LockWrite();
	SQL sql(m_DbConn);
	mString req;
	mString strFlags;
	ASN1_ENCRYPT_SIGN * crypt;
	char * pem_datas;
	mString strPrivDatas;

	if(!ProfileHash)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		ConfAccessLock.Unlock();
		return false;
	}

	if(!ASN1_BIT_STRINGtoString(flags, strFlags))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}

	if(PrivDatas)
	{
		crypt = ASN1_ENCRYPT_SIGN_new();
		if(!crypt)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ConfAccessLock.Unlock();
			return false;
		}
		if(!RA_STORE_PRIVDATAS_sign_encrypt(crypt, PrivDatas, m_EntityCert.GetPrivateKey().GetRsaKey(), m_EntityCert.GetPrivateKey().GetRsaKey(), EVP_sha1(), EVP_des_ede3_cbc()))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ASN1_ENCRYPT_SIGN_free(crypt);
			ConfAccessLock.Unlock();
			return false;
		}
		pem_datas = ASN1_ENCRYPT_SIGN_To_PEM(crypt);
		if(!pem_datas)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ASN1_ENCRYPT_SIGN_free(crypt);
			ConfAccessLock.Unlock();
			return false;
		}
		ASN1_ENCRYPT_SIGN_free(crypt);
		strPrivDatas = pem_datas;
		free(pem_datas);
	}
	else
	{
		strPrivDatas = "";
	}

	req.sprintf(RASTORE_INSERT_CERT, ProfileId, CaName, pkcs10.GetPemCSR().c_str(), strPrivDatas.c_str(), Type, NEWPKI_PROFILE_CERT_STATE_WAITING, strFlags.c_str(), admin_mail);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	CertReqId = sql.GetLastID();
	if(CertReqId == (unsigned long)-1)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	ConfAccessLock.Unlock();
	return true;
}

bool RaStore::GetCerts(unsigned long ProfileId, STACK_OF(NEWPKI_PROFILE_DATAS_CERT) * Certs)
{
	SQL sql(m_DbConn);
	char * err;
	mString req;
	long NumRows;
	long i;
	NEWPKI_PROFILE_DATAS_CERT * currCert;
	PKI_CERT ttCert;

	req.sprintf(RASTORE_GET_CERTS, ProfileId);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	for(i=0; i<NumRows; i++)
	{
		ERR_clear_error();
		currCert = Sql2Cert(&sql, i);
		if(!currCert)
		{
			currCert = NEWPKI_PROFILE_DATAS_CERT_new();
			if(currCert)
			{
				ASN1_INTEGER_SET(&currCert->Id, 0);
				ASN1_INTEGER_SET(&currCert->flags, 0);
				ASN1_INTEGER_SET(&currCert->state, NEWPKI_PROFILE_CERT_STATE_ERROR);
				ASN1_INTEGER_SET(&currCert->type, NEWPKI_PROFILE_CERT_TYPE_PKCS12);
				ASN1_STRING_SET(&currCert->ca_name, _sv("Unknown"));
				err = ERR_to_string();
				if(!err)
				{
					ASN1_STRING_SET(&currCert->error, _sv("Unknown"));
				}
				else
				{
					ASN1_STRING_SET(&currCert->error, err);
					free(err);
				}
			}
		}
		if(currCert && SKM_sk_push(NEWPKI_PROFILE_DATAS_CERT, Certs, currCert) < 0)
			NEWPKI_PROFILE_DATAS_CERT_free(currCert);

	}
	
	return true;
}

NEWPKI_PROFILE_DATAS * RaStore::GetProfile(const PKI_CERT &UserCert, unsigned long ProfileId)
{
	NEWPKI_PROFILE_DATAS * res;
	long NumRows;
	SQL sql(m_DbConn);
	mString req;
	mString tmpDatas;
	unsigned long owner_type;
	unsigned long owner_serial;


	req.sprintf(RASTORE_GET_PROFILE, ProfileId);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!NumRows)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return NULL;
	}

	res = NEWPKI_PROFILE_DATAS_new();
	if(!res)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return NULL;
	}

	if(!sql.Value(0, "owner_type", tmpDatas))
	{
		NEWPKI_PROFILE_DATAS_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	owner_type = tmpDatas.c_ulng();

	if(!sql.Value(0, "owner_serial", tmpDatas))
	{
		NEWPKI_PROFILE_DATAS_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	owner_serial = tmpDatas.c_ulng();
	

	if(!CanUserAccessProfile(UserCert, owner_serial, owner_type))
	{
		NEWPKI_PROFILE_DATAS_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return NULL;
	}

	// The profile id
	if(!sql.Value(0, "profile_id", tmpDatas))
	{
		NEWPKI_PROFILE_DATAS_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!tmpDatas.c_ASN1_INTEGER(&res->Id))
	{
		NEWPKI_PROFILE_DATAS_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}

	// The state
	if(!sql.Value(0, "state", tmpDatas))
	{
		NEWPKI_PROFILE_DATAS_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!tmpDatas.c_ASN1_INTEGER(&res->State))
	{
		NEWPKI_PROFILE_DATAS_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}

	//The real profile
	if(!sql.Value(0, "datas", tmpDatas))
	{
		NEWPKI_PROFILE_DATAS_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	// We need to reconvert if from the PEM
	if(NEWPKI_PROFILE_From_PEM(&res->Profile, tmpDatas.c_str()) <= 0)
	{
		NEWPKI_PROFILE_DATAS_free(res);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	return res;
}

bool RaStore::DeleteCertReq(unsigned long CertReqId)
{
	SQL sql(m_DbConn);
	mString req;

	// Delete all the transactions linked to this cert	
	if(!Requester_DeleteTransactionPrivAttr(CertReqId))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	// Delete the cert
	req.sprintf(RASTORE_DELETE_CERT, CertReqId);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}
bool RaStore::InsertResponse(const mString & ca_name, const ASN1_OCTET_STRING *transactionID, const NEWPKI_RESPONSE * resp, const AsynchJobs & Jobs, const EntityLog * Logging)
{
	int Status = NEWPKI_PROFILE_CERT_STATE_ERROR;
	mString strError;
	mString strSqlError;
	mString strPrivDatas;
	mString x509;
	SQL sql(m_DbConn);
	mString req;
	long NumRows;
	mString tid;
	mString currStatus;
	char * err;
	unsigned long cert_id;
	mString AdminMail;
	unsigned long serial = 0;
	PKI_CRL Crl;
	ASN1_BIT_STRING * flags;
	bool SendCertificate;
	bool SendPkcs12Pwd;
	bool NotifyAdmin;
	bool GenPkcs12Pwd;
	LOG_MESSAGE_TYPE LogMessage;
	LOG_MESSAGE_STATUS LogStatus = (LOG_MESSAGE_STATUS)0;
	NEWPKI_PROFILE * Profile;
	HashTable_Dn ProfileDn;
	MailInfo CertificateMail;
	MailInfo PasswordMail;



	if(!ProfileHash)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_PARAM);
		return false;
	}

	if(!transactionIDtoString(transactionID, tid))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	req.sprintf(RASTORE_GET_CERT_BY_TID, tid.c_str());
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!NumRows)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!sql.Value(0,"id", req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	cert_id = req.c_ulng();
	if(!cert_id)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!sql.Value(0,"status", currStatus))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!sql.Value(0,"priv_datas", strPrivDatas))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!sql.Value(0,"flags", req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	flags = StringtoASN1_BIT_STRING(req);
	if(!flags)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	NotifyAdmin = (ASN1_BIT_STRING_get_bit(flags, REQUEST_CERT_FLAG_SEND_ADMIN_NOTIFICATION) == 1);
	SendCertificate = (ASN1_BIT_STRING_get_bit(flags, REQUEST_CERT_FLAG_SEND_CERT_TO_USER) == 1);
	GenPkcs12Pwd = (ASN1_BIT_STRING_get_bit(flags, REQUEST_CERT_FLAG_RA_GENERATE_P12_PASSWD) == 1);
	SendPkcs12Pwd = (ASN1_BIT_STRING_get_bit(flags, REQUEST_CERT_FLAG_SEND_P12_PASSWD_TO_USER) || GenPkcs12Pwd);
	ASN1_BIT_STRING_free(flags);


	if(!sql.Value(0,"admin_mail", AdminMail))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	Profile = NEWPKI_PROFILE_new();
	if(!Profile)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	if(!sql.Value(0, "datas", req))
	{
		NEWPKI_PROFILE_free(Profile);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	// We need to reconvert if from the PEM
	if(NEWPKI_PROFILE_From_PEM(&Profile, req.c_str()) <= 0)
	{
		NEWPKI_PROFILE_free(Profile);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!ProfileDn.From_X509_NAME(Profile->dn))
	{
		NEWPKI_PROFILE_free(Profile);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	NEWPKI_PROFILE_free(Profile);



	switch(resp->type)
	{
		case NEWPKI_RESPONSE_TYPE_CERT:
			if(currStatus.c_int() != NEWPKI_PROFILE_CERT_STATE_WAITING)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
				return false;
			}
			
			if(!InsertResponseCert(resp->d.cert_response, Crl, LogMessage, LogStatus, Status, strError, ProfileDn, strPrivDatas, x509, serial, CertificateMail, PasswordMail, SendCertificate, SendPkcs12Pwd, GenPkcs12Pwd))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				return false;
			}

			break;
		case NEWPKI_RESPONSE_TYPE_REV:
			if(currStatus.c_int() != NEWPKI_PROFILE_CERT_STATE_WAITING_REV)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
				return false;
			}

			SendCertificate = false;
			GenPkcs12Pwd = false;
			SendPkcs12Pwd = false;
			LogMessage = LOG_MESSAGE_TYPE_CERT_REV_RESP;

			if(resp->d.rev_response->last_crl)
				Crl.SetCRL(resp->d.rev_response->last_crl);

			// Did everything go right for the rev ?
			switch(ASN1_INTEGER_GET(resp->d.rev_response->status))
			{
				case REV_RESPONSE_OK:
					Status = NEWPKI_PROFILE_CERT_STATE_REVOKED;
					LogStatus = LOG_STATUS_TYPE_SUCCESS;
					strPrivDatas = "";
					break;
				
				case REV_RESPONSE_ERROR:
					LogStatus = LOG_STATUS_TYPE_FAILURE;
					if(ASN1_INTEGER_GET(resp->d.rev_response->cert_status) == CERT_STATE_REVOKED)
					{
						Status = NEWPKI_PROFILE_CERT_STATE_REVOKED;
					}
					else if(ASN1_INTEGER_GET(resp->d.rev_response->cert_status) == CERT_STATE_VALID)
					{
						Status = NEWPKI_PROFILE_CERT_STATE_ACTIVE;
					}
					else if(ASN1_INTEGER_GET(resp->d.rev_response->cert_status) == CERT_STATE_UNKNOWN)
					{
						Status = NEWPKI_PROFILE_CERT_STATE_ACTIVE;
					}

					strPrivDatas = "";
					if(resp->d.rev_response->errors)
					{
						err = ERROR_ENTRIES_to_string(resp->d.rev_response->errors);
						if(err)
						{
							strError = err;
							free(err);
						}
					}
					break;

				default:
					NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
					return false;
					break;
			}
			break;
		case NEWPKI_RESPONSE_TYPE_ERR:
			if(currStatus.c_int() != NEWPKI_PROFILE_CERT_STATE_WAITING && currStatus.c_int() != NEWPKI_PROFILE_CERT_STATE_ACTIVE && currStatus.c_int() != NEWPKI_PROFILE_CERT_STATE_WAITING_REV)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
				return false;
			}
			SendCertificate = false;
			GenPkcs12Pwd = false;
			SendPkcs12Pwd = false;
			Status = NEWPKI_PROFILE_CERT_STATE_ERROR;
			LogMessage = LOG_MESSAGE_TYPE_CERT_SIG_RESP;
			strPrivDatas = "";
			if(resp->d.errors)
			{
				err = ERROR_ENTRIES_to_string(resp->d.errors);
				if(err)
				{
					strError = err;
					free(err);
				}
			}
			break;
		
		default:
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			return false;
			break;
	}

	// Let's do some verification before sending the mails
	if(SendPkcs12Pwd)
	{
		// TODO On ne peut pas laisser cette rponse
		// nous tre repropos indfiniment, il faut demander
		// se rvocation
		if(!PasswordMail || !PasswordMail.get_MailTo().c_str())
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
			return false;
		}
	}
	if(SendCertificate)
	{
		// TODO On ne peut pas laisser cette rponse
		// nous tre repropos indfiniment, il faut demander
		// se rvocation
		if(!CertificateMail || !CertificateMail.get_MailTo().c_str() || !CertificateMail.get_Attach().get_BufferLen())
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
			return false;
		}
	}
	if(NotifyAdmin)
	{
		if(!AdminMail.size())
		{
			// TODO Ajouter une entre de log
			NotifyAdmin = false;
		}
	}

	// Are we supposed to send the PKCS#12 password to the user ?
	if(SendPkcs12Pwd)
	{
		// TODO On ne peut pas laisser cette rponse
		// nous tre repropos indfiniment, il faut demander
		// se rvocation
		if(!Jobs.SendMail(m_EntityCert.GetStringName(), PasswordMail, false))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}


	// Are we supposed to send the certificate to the user ?
	if(SendCertificate)
	{
		// TODO On ne peut pas laisser cette rponse
		// nous tre repropos indfiniment, il faut demander
		// se rvocation
		if(!Jobs.SendMail(m_EntityCert.GetStringName(), CertificateMail, false))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}

	// Are we supposed to notify the admin ?
	if(NotifyAdmin)
	{
		PasswordMail.Clear();
		PasswordMail.set_MailTo(AdminMail);
		PasswordMail.set_Subject(_sv("Received response"));
		req.sprintf(_sv("Received response for request %ld."), cert_id);
		PasswordMail.set_Body(req);
		PasswordMail.set_SignMail(true);

		if(!Jobs.SendMail(m_EntityCert.GetStringName(), PasswordMail, false))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}



	// Set the response in the database
	if(strError.size())
		strSqlError = sql.FormatString(strError);

	req.sprintf(RASTORE_SET_CERT_RESP, Status, x509.c_str(), strSqlError.c_str(), strPrivDatas.c_str(), serial, cert_id);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!Requester_RequestProcessed(transactionID))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(Crl)
		UpdateFromCRL(ca_name, Crl);


	if(LogMessage)
	{
		switch(LogStatus)
		{
			case LOG_STATUS_TYPE_SUCCESS:
				Logging->LogMessage(LogStatus, LogMessage, 0, ca_name.c_str(), cert_id);
				break;
			case LOG_STATUS_TYPE_FAILURE:
				Logging->LogMessage(LogStatus, LogMessage, 0, ca_name.c_str(), cert_id, NULL, strError.c_str());
				break;
			default:
				NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
				return false;
		}
	}

	return true;
}

NEWPKI_PROFILE_DATAS_CERT * RaStore::GetCertReq(const ASN1_OCTET_STRING *transactionID)
{
	SQL sql(m_DbConn);
	mString req;
	long NumRows;
	NEWPKI_PROFILE_DATAS_CERT * currCert;
	mString tid;
	
	if(!transactionIDtoString(transactionID, tid))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}

	req.sprintf(RASTORE_GET_CERT_BY_TID, tid.c_str());
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}
	if(!NumRows)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return NULL;
	}

	currCert = Sql2Cert(&sql, 0);
	if(!currCert)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return currCert;
}

NEWPKI_PROFILE_DATAS_CERT * RaStore::Sql2Cert(SQL *sql, int index)
{
	mString value;
	PKI_CERT ttCert;
	PKI_PKCS12 ttP12;
	NEWPKI_PROFILE_DATAS_CERT * currCert;

	currCert = NEWPKI_PROFILE_DATAS_CERT_new();
	if(!currCert)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return NULL;
	}

	if(!sql->Value(index,"id", value))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NEWPKI_PROFILE_DATAS_CERT_free(currCert);
		return NULL;
	}
	if(!value.c_ASN1_INTEGER(&currCert->Id))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NEWPKI_PROFILE_DATAS_CERT_free(currCert);
		return NULL;
	}

	if(!sql->Value(index,"ca_name", value))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NEWPKI_PROFILE_DATAS_CERT_free(currCert);
		return NULL;
	}
	if(!value.c_ASN1_UTF8STRING(&currCert->ca_name))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NEWPKI_PROFILE_DATAS_CERT_free(currCert);
		return NULL;
	}

	if(!sql->Value(index,"status", value))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NEWPKI_PROFILE_DATAS_CERT_free(currCert);
		return NULL;
	}
	if(!value.c_ASN1_INTEGER(&currCert->state))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NEWPKI_PROFILE_DATAS_CERT_free(currCert);
		return NULL;
	}


	if(!sql->Value(index,"type", value))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NEWPKI_PROFILE_DATAS_CERT_free(currCert);
		return NULL;
	}
	if(!value.c_ASN1_INTEGER(&currCert->type))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NEWPKI_PROFILE_DATAS_CERT_free(currCert);
		return NULL;
	}
	
	
	if(!sql->Value(index,"flags", value))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NEWPKI_PROFILE_DATAS_CERT_free(currCert);
		return NULL;
	}
	if(currCert->flags)
		ASN1_BIT_STRING_free(currCert->flags);
	currCert->flags = StringtoASN1_BIT_STRING(value);
	if(!currCert->flags)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NEWPKI_PROFILE_DATAS_CERT_free(currCert);
		return NULL;
	}
		
	if(!sql->Value(index,"admin_mail", value))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NEWPKI_PROFILE_DATAS_CERT_free(currCert);
		return NULL;
	}
	if(!value.c_ASN1_UTF8STRING(&currCert->admin_mail))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NEWPKI_PROFILE_DATAS_CERT_free(currCert);
		return NULL;
	}


	// Are we in state of having the PKCS#12 ?
	if(ASN1_INTEGER_GET(currCert->state) == NEWPKI_PROFILE_CERT_STATE_ACTIVE
	&& ASN1_INTEGER_GET(currCert->type) == NEWPKI_PROFILE_CERT_TYPE_PKCS12)
	{
		// Do we have the PKCS#12 ?
		if(!sql->Value(index,"priv_datas", value))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			NEWPKI_PROFILE_DATAS_CERT_free(currCert);
			return NULL;
		}
		if(value.size())
		{
			if(!ttP12.Load(value.c_str()))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				NEWPKI_PROFILE_DATAS_CERT_free(currCert);
				return NULL;
			}
			currCert->p12 = ttP12.GetPKCS12(true);
			if(!currCert->p12)
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				NEWPKI_PROFILE_DATAS_CERT_free(currCert);
				return NULL;
			}
		}
	}
	
	
	if(!sql->Value(index,"x509", value))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NEWPKI_PROFILE_DATAS_CERT_free(currCert);
		return NULL;
	}
	if(value.size())
	{
		if(!ttCert.SetCert(value.c_str()))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			NEWPKI_PROFILE_DATAS_CERT_free(currCert);
			return NULL;
		}
		currCert->cert = ttCert.GetX509(true);
		if(!currCert->cert)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			NEWPKI_PROFILE_DATAS_CERT_free(currCert);
			return NULL;
		}
	}

	if(!sql->Value(index,"error", value))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		NEWPKI_PROFILE_DATAS_CERT_free(currCert);
		return NULL;
	}
	if(value.size())
	{
		if(!value.c_ASN1_UTF8STRING(&currCert->error))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			NEWPKI_PROFILE_DATAS_CERT_free(currCert);
			return NULL;
		}
	}

	return currCert;
}

bool RaStore::SuspendCertificate(const PKI_CERT &UserCert, unsigned long id, unsigned long & serial, mString & ca_name, mString & ldap_uid)
{
	SQL sql(m_DbConn);
	mString req;
	long NumRows;
	unsigned long owner_serial;
	unsigned long owner_type;
	PKI_CERT Cert;
	NEWPKI_PROFILE * Profile;


	req.sprintf(RASTORE_GET_CERT_4_REV, id);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!NumRows)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	// Check the current status of the certificate
	if(!sql.Value(0, "status", req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(req.c_int() != NEWPKI_PROFILE_CERT_STATE_ACTIVE)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!sql.Value(0, "owner_type", req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	owner_type = req.c_ulng();

	if(!sql.Value(0, "owner_serial", req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	owner_serial = req.c_ulng();
		
	if(!CanUserAccessProfile(UserCert, owner_serial, owner_type))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}


	// Get the serial of the x509
	if(!sql.Value(0, "x509", req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!Cert.SetCert(req.c_str()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	serial = Cert.GetSerial();
	
	// Get the ca_name
	if(!sql.Value(0, "ca_name", ca_name))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	// Get the LDAP RDN
	if(!sql.Value(0, "datas", req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	Profile = NULL;
	// We need to reconvert if from the PEM
	if(NEWPKI_PROFILE_From_PEM(&Profile, req.c_str()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	ldap_uid = ASN1_STRING_GET(Profile->ldap_uid);
	NEWPKI_PROFILE_free(Profile);



	req.sprintf(RASTORE_SUSPEND_CERT, NEWPKI_PROFILE_CERT_STATE_WAITING_REV, id);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool RaStore::CanUserAccessProfile(const PKI_CERT &UserCert, unsigned long owner_serial, unsigned long owner_type)
{
	USERS_GROUP * currGroup;
	ASN1_INTEGER * currSerial;
	long i, j;

	if(m_AclValidator->IsPkiAdministrator(UserCert))
	{
		return true;
	}

	if(owner_type == OWNER_TYPE_USER)
	{
		if(owner_serial == UserCert.GetSerial())
		{
			return true;
		}
		else
		{
			return false;
		}
	}

	ConfAccessLock.LockRead();
	for(i=0; i<SKM_sk_num(USERS_GROUP, m_groups->list); i++)
	{
		currGroup = SKM_sk_value(USERS_GROUP, m_groups->list, i);
		if(!currGroup) continue;

		// is this the group ?
		if((unsigned long)ASN1_INTEGER_GET(currGroup->serial) == owner_serial)
		{
			for(j=0; j<SKM_sk_num(ASN1_INTEGER, currGroup->users_serial); j++)
			{
				currSerial = SKM_sk_value(ASN1_INTEGER, currGroup->users_serial, j);
				if(!currSerial) continue;

				// is this the user ?
				if((unsigned long)ASN1_INTEGER_GET(currSerial) == UserCert.GetSerial())
				{
					ConfAccessLock.Unlock();
					return true;
				}
			}

			ConfAccessLock.Unlock();
			return false;
		}
	}
	ConfAccessLock.Unlock();
	return false;
}

bool RaStore::DeletePKCS12(const PKI_CERT &UserCert, unsigned long id)
{
	SQL sql(m_DbConn);
	mString req;
	long NumRows;
	unsigned long owner_serial;
	unsigned long owner_type;
	PKI_CERT Cert;


	req.sprintf(RASTORE_GET_CERT_OWNER, id);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!NumRows)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	// Check the current status of the certificate
	if(!sql.Value(0, "status", req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(req.c_int() != NEWPKI_PROFILE_CERT_STATE_ACTIVE)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	if(!sql.Value(0, "owner_type", req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	owner_type = req.c_ulng();

	if(!sql.Value(0, "owner_serial", req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	owner_serial = req.c_ulng();
		
	if(!CanUserAccessProfile(UserCert, owner_serial, owner_type))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_NOT_ALLOWED);
		return false;
	}

	req.sprintf(RASTORE_DELETE_PKCS12, id);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool RaStore::UpdateFromCRL(const mString & ca_name, const PKI_CRL &Crl)
{
	mString req;
	mString where;
	long i;
	const REVOCATION_INFO * currRev;
	SQL sql(m_DbConn);

	for(i=0; i<Crl.RevokedCertsCount(); i++)
	{
		currRev = Crl.GetRevokedCert(i);
		if(!currRev)
			continue;

		where += "serial = '";
		where += currRev->serial;
		where += "'";

		if(i != Crl.RevokedCertsCount()-1)
			where += " OR ";
	}
	req.sprintf(RASTORE_UPDATE_FROM_CRL, NEWPKI_PROFILE_CERT_STATE_REVOKED, ca_name.c_str(), NEWPKI_PROFILE_CERT_STATE_REVOKED, where.c_str());

	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}



bool RaStore::OnNewProfile(const NEWPKI_PROFILE *newProfile)
{
	PKI_CERT EmptyCert;
	long id;
	return InsertProfile(EmptyCert, newProfile, id);
}

bool RaStore::GetKnownUIDs(STACK_OF(ASN1_UTF8STRING) * KnownUIDs)
{
	mString req;
	long i;
	SQL sql(m_DbConn);
	ASN1_UTF8STRING * currUID;
	long NumRows;

	req.sprintf(RASTORE_GET_KNOWN_RDNS);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	for(i=0; i<NumRows; i++)
	{
		// The profile id
		if(!sql.Value(i, "ldap_uid", req))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		currUID = NULL;
		if(!req.c_ASN1_UTF8STRING(&currUID))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
			return false;
		}
		if(SKM_sk_push(ASN1_UTF8STRING, KnownUIDs, currUID) < 0)
		{
			ASN1_UTF8STRING_free(currUID);
		}
	}
	
	return true;
}

void RaStore::GeneratePassword(mString &strP12Password)
{
	unsigned long MinPwdLen;
	char * AllowedChars = {"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890{}()[],.:;!?"};
	size_t AllowedCharsLen = strlen(AllowedChars);
	unsigned long i;
	char tt[2];
	size_t j;

	ConfAccessLock.LockRead();
	MinPwdLen = m_MinPasswdLen;
	ConfAccessLock.Unlock();

	strP12Password = "";
	srand(time(NULL));
	tt[1] = '\0';
	for(i=0; i<MinPwdLen; i++)
	{
		do
		{
			j = rand();
		}
		while(j > AllowedCharsLen);

		tt[0] = AllowedChars[j];
		strP12Password += tt;
	}
}


void RaStore::SetAclValidator(const X509_ACL_Validator *AclValidator)
{
	m_AclValidator = AclValidator;
}

bool RaStore::InsertResponseCert(NEWPKI_CERT_RESPONSE* cert_response, PKI_CRL & Crl, LOG_MESSAGE_TYPE & LogMessage, LOG_MESSAGE_STATUS & LogStatus, int & Status, mString & strError, const HashTable_Dn & ProfileDn, mString & strPrivDatas, mString & x509, unsigned long & serial, MailInfo & CertificateMail, MailInfo & PasswordMail, bool & SendCertificate, bool & SendPkcs12Pwd, bool & GenPkcs12Pwd)
{
	mString strP12Password;
	mString UserMail;
	mString MailBody;
	mString Id;

	PKI_CERT UserCert;
	PKI_PKCS12 p12;	
	PKI_RSA	PrivKey;
	PKI_P7B p7b;
	X509 * currParentCert;
	HashTable_String ParentCerts;
	int i;
	int pos;
	RA_STORE_PRIVDATAS * PrivDatas;
	ASN1_ENCRYPT_SIGN * crypt;
	char * err;

	LogMessage = LOG_MESSAGE_TYPE_CERT_SIG_RESP;

	if(cert_response->last_crl)
		Crl.SetCRL(cert_response->last_crl);

	// Is it an error or the certificate ?
	switch(ASN1_INTEGER_GET(cert_response->status))
	{
		// A certificate
		case CERT_RESPONSE_OK:
			Status = NEWPKI_PROFILE_CERT_STATE_ACTIVE;
			LogStatus = LOG_STATUS_TYPE_SUCCESS;

			if(!UserCert.SetCert(cert_response->certificate))
			{
				NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
				return false;
			}
			x509 = UserCert.GetCertPEM();
			serial = UserCert.GetSerial();

			// Get the user's mail address
			pos = UserCert.GetCertDN().SeekEntryName("emailAddress", HASHTABLE_NOT_FOUND);
			if(pos != HASHTABLE_NOT_FOUND)
				UserMail = UserCert.GetCertDN().Get(pos);
			if(!UserMail.size())
			{
				pos = ProfileDn.SeekEntryName("emailAddress", HASHTABLE_NOT_FOUND);
				if(pos != HASHTABLE_NOT_FOUND)
					UserMail = ProfileDn.Get(pos);
			}

			// Are we supposed to generate a PKCS12 ?
			if(strPrivDatas.size())
			{
				//PKCS12 Generation
				crypt = NULL;
				if(!ASN1_ENCRYPT_SIGN_From_PEM(&crypt, strPrivDatas.c_str()))
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					return false;
				}

				PrivDatas = NULL;
				if(!RA_STORE_PRIVDATAS_decrypt_verify(crypt, &PrivDatas, m_EntityCert.GetPrivateKey().GetRsaKey(), m_EntityCert.GetPrivateKey().GetRsaKey()))
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					ASN1_ENCRYPT_SIGN_free(crypt);
					return false;
				}
				ASN1_ENCRYPT_SIGN_free(crypt);

				// Load the private datas into a more explicit way

				// Are we supposed to generate the PKCS#12 password ?
				if(GenPkcs12Pwd)
				{
					// Generate it
					GeneratePassword(strP12Password);
					MailBody.sprintf("%s%s", _sv("This is your certificate's password:\n\n"), strP12Password.c_str());

					if(SendPkcs12Pwd)
					{
						PasswordMail.set_MailTo(UserMail);

						PasswordMail.set_Subject(_sv("Your certificate's password"));
						PasswordMail.set_Body(MailBody);
						PasswordMail.set_SignMail(true);
					}
				}
				else
				{
					// Get it from the private datas
					strP12Password = PrivDatas->passwd;
				}
				if(!strP12Password.size())
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					RA_STORE_PRIVDATAS_free(PrivDatas);
					return false;
				}

				if(!PrivKey.SetKey(PrivDatas->privkey))
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					RA_STORE_PRIVDATAS_free(PrivDatas);
					return false;
				}
				RA_STORE_PRIVDATAS_free(PrivDatas);

				// Ok, let's generate the PKCS12
				p12.SetEndUserCert(UserCert);
				p12.SetEndUserKey(PrivKey);

				ParentCerts.AllowDuplicateNames();

				if(cert_response->ParentCerts)
				{
					for(i=0; i<SKM_sk_num(X509, cert_response->ParentCerts); i++)
					{
						currParentCert = SKM_sk_value(X509, cert_response->ParentCerts, i);
						if(!currParentCert)
							continue;
						if(!UserCert.SetCert(currParentCert))
							continue;
						
						Id = i;
						if(!ParentCerts.Add(Id.c_str(), UserCert.GetCertPEM().c_str()))
							continue;
					}
					p12.SetParentCerts(ParentCerts);
				}
				if(!p12.Generate((char*)strP12Password.c_str()))
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					return false;
				}
				// The P12 is saved in DB
				strPrivDatas = p12.GetPemPKCS12();

				// Save the P12 in DER
				if(!p12.GetPemPKCS12().ToDER(CertificateMail.get_Attach()))
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					return false;
				}
				p12.Clear();

				if(SendCertificate)
				{
					CertificateMail.set_MailTo(UserMail);

					CertificateMail.set_AttachName("certificate.p12");
					CertificateMail.set_AttachType("application/x-pkcs12");

					CertificateMail.set_Subject(_sv("Your certificate"));
					CertificateMail.set_Body(_sv("Your certificate has been generated."));
					CertificateMail.set_SignMail(true);
				}
			}
			else
			{
				// Generate the PKCS7
				p7b.AddCert(UserCert.GetX509());

				if(cert_response->ParentCerts)
				{
					for(i=0; i<SKM_sk_num(X509, cert_response->ParentCerts); i++)
					{
						currParentCert = SKM_sk_value(X509, cert_response->ParentCerts, i);
						if(!currParentCert)
							continue;

						p7b.AddCert(currParentCert);
					}
				}
				if(!p7b.Generate())
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					return false;
				}
				// Save the p7b in DER
				if(!p7b.GetPemP7B().ToDER(CertificateMail.get_Attach()))
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					return false;
				}
				p7b.Clear();

				CertificateMail.set_AttachName("certificate.p7b");
				CertificateMail.set_AttachType("application/x-pkcs7-certificates");
			}
			break;
		
		// An error
		case CERT_RESPONSE_ERROR:
			Status = NEWPKI_PROFILE_CERT_STATE_ERROR;
			LogStatus = LOG_STATUS_TYPE_FAILURE;
			SendCertificate = false;
			GenPkcs12Pwd = false;
			SendPkcs12Pwd = false;
			strPrivDatas = "";
			if(cert_response->errors)
			{
				err = ERROR_ENTRIES_to_string(cert_response->errors);
				if(err)
				{
					strError = err;
					free(err);
				}
			}
			else
			{
				strError = _sv("Unknown");
			}
			break;

		default:
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			return false;
			break;
	}

	return true;
}
