/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/

// RaStore.h: interface for the RaStore class.
//
//////////////////////////////////////////////////////////////////////

#ifndef RASTORE_H
#define RASTORE_H

#include "NewPKIStore.h"
#include <ReadersWriter.h>
#include <PEM/PEM_DER.h>
#include <HashCorrelation.h>
#include "AsynchJobs.h"
#include "AutoSynchLDAP.h"
#include "MailInfo.h"

#ifndef DOXYGEN_SHOULD_SKIP_THIS
typedef struct st_RA_STORE_SIG
{
	ASN1_BIT_STRING * signature;
	X509_ALGOR * sig_alg;
}RA_STORE_SIG;
DECLARE_ASN1_FUNCTIONS(RA_STORE_SIG)
DECLARE_PEM_CONVERT(RA_STORE_SIG)

typedef struct st_RA_STORE_INFO
{
	ASN1_UTF8STRING * hash;
	ASN1_INTEGER * counter;
}RA_STORE_INFO;
DECLARE_ASN1_FUNCTIONS(RA_STORE_INFO)
DECLARE_ASN1_ENCRYPT_SIGN(RA_STORE_INFO)

typedef struct st_RA_STORE_PRIVDATAS
{
	ASN1_UTF8STRING * passwd;
	RSA * privkey;
}RA_STORE_PRIVDATAS;
DECLARE_ASN1_FUNCTIONS(RA_STORE_PRIVDATAS)
DECLARE_ASN1_ENCRYPT_SIGN(RA_STORE_PRIVDATAS)
#endif


/*!
	This class is the store for a RA
*/
class RaStore : public AutoSynchLDAPEvent, public NewPKIStore
{
public:
	/*! \brief This is the constructor.
	 */
	RaStore();

	/*! \brief This is the destructor.
	 */
	virtual ~RaStore();


	/*! \brief This function creates the intial datas.
	 *  \return true on success, false on failure.
	 */
	bool Create();

	/*! \brief This function loads the intial datas.
	 *  \return true on success, false on failure.
	 */
	bool Load();
	
	bool CreateTables(const SQL_Connection * DbConn);
	
	/*! \brief This function sets the groups info.
	 *  \param groups [IN] The groups info.
	 *  \return true on success, false on failure.
	 */
	bool SetGroups(const USERS_GROUPS * groups);
	
	/*! \brief This function sets the validator.
	 *  \param AclValidator [IN] The validator.
	 */
	void SetAclValidator(const X509_ACL_Validator * AclValidator);
	
	/*! \brief This function sets the minimum PKCS12 password length.
	 *  \param MinPasswdLen [IN] The minimum PKCS12 password length.
	 */
	void SetMinPasswdLen(unsigned long MinPasswdLen);

	bool OnNewProfile(const NEWPKI_PROFILE * newProfile);

	bool GetKnownUIDs(STACK_OF(ASN1_UTF8STRING) * KnownUIDs);

	/*! \brief This function inserts a new profile.
	 *  \param UserCert [IN] The certificate of the creator.
	 *  \param Profile [IN] The new profile.
	 *  \param ProfileId [OUT] The new profile's internal UID.
	 *  \return true on success, false on failure.
	 */
	bool InsertProfile(const PKI_CERT &UserCert, const NEWPKI_PROFILE * Profile, long & ProfileId);

	/*! \brief This function gets the list of profiles.
	 *  \param UserCert [IN] The certificate of the operator requesting the operation.
	 *  \param index [IN] The index for the enumeration.
	 *  \param num [IN] The maximum number of entries to return.
	 *  \return The profiles list or NULL on failure.
	 */
	STACK_OF(NEWPKI_PROFILE_DATAS) * EnumProfiles(const PKI_CERT &UserCert, long index, long num);

	/*! \brief This function deletes a profile.
	 *  \param ProfileId [IN] The profile's internal UID.
	 *  \return true on success, false on failure.
	 */
	bool DeleteProfile(long ProfileId);

	/*! \brief This function gets a profile.
	 *  \param UserCert [IN] The certificate of the operator requesting the operation.
	 *  \param ProfileId [IN] The profile's internal UID.
	 *  \return The profile or NULL on failure.
	 */
	NEWPKI_PROFILE_DATAS * GetProfile(const PKI_CERT &UserCert, unsigned long ProfileId);

	/*! \brief This function deletes the PKCS12 associated with a profile.
	 *  \param UserCert [IN] The certificate of the operator requesting the operation.
	 *  \param id [IN] The profile's internal UID.
	 *  \return true on success, false on failure.
	 */
	bool DeletePKCS12(const PKI_CERT &UserCert, unsigned long id);
	
	/*! \brief This function inserts a new certificate request for a profile.
	 *  \param CaName [IN] The name of the CA on which the certificate was requested.
	 *  \param ProfileId [IN] The profile's internal UID.
	 *  \param Type [IN] The CSR type (see NEWPKI_PROFILE_CERT_TYPE_PKCS10, NEWPKI_PROFILE_CERT_TYPE_PKCS12, etc.).
	 *  \param pkcs10 [IN] The PKCS10.
	 *  \param PrivDatas [IN] The private datas (PKCS12's password, private key, etc.).
	 *  \param flags [IN] The flags (see REQUEST_CERT_FLAG_SEND_ADMIN_NOTIFICATION, REQUEST_CERT_FLAG_SEND_CERT_TO_USER, etc.).
	 *  \param admin_mail [IN] The email of administrator that generated this request.
	 *  \param CertReqId [OUT] The request's internal UID.
	 *  \return true on success, false on failure.
	 */
	bool InsertCertReq(const char *CaName, unsigned long ProfileId, int Type, const PKI_CSR & pkcs10, const RA_STORE_PRIVDATAS *PrivDatas, const ASN1_BIT_STRING * flags, const char * admin_mail, unsigned long & CertReqId);

	/*! \brief This function gets a certificate request.
	 *  \param transactionID [IN] The transactionID associated with the request.
	 *  \return The certificate request or NULL on failure.
	 */
	NEWPKI_PROFILE_DATAS_CERT * GetCertReq(const ASN1_OCTET_STRING * transactionID);

	/*! \brief This function deletes a certificate request.
	 *  \param CertReqId [IN] The certificate request's internal UID.
	 *  \return true on success, false on failure.
	 */
	bool DeleteCertReq(unsigned long CertReqId);

	/*! \brief This function suspends a certificate.
	 *  \param UserCert [IN] The certificate of the operator requesting the operation.
	 *  \param id [IN] The certificate's internal UID.
	 *  \param serial [OUT] The serial of the certificate.
	 *  \param ca_name [OUT] The name of the CA that generated the certificate.
	 *  \param ldap_uid [OUT] The LDAP uid of the associated profile.
	 *  \return true on success, false on failure.
	 */
	bool SuspendCertificate(const PKI_CERT & UserCert, unsigned long id, unsigned long & serial, mString & ca_name, mString & ldap_uid);

	/*! \brief This function suspends a certificate.
	 *  \param ca_name [IN] The name of the CA sending the response.
	 *  \param transactionID [IN] The transaction ID of the response.
	 *  \param resp [IN] The response.
	 *  \param Jobs [IN] The asynch jobs, used to send notifications email.
	 *  \param Logging [IN] The logging class.
	 *  \return true on success, false on failure.
	 */
	bool InsertResponse(const mString & ca_name, const ASN1_OCTET_STRING *transactionID, const NEWPKI_RESPONSE * resp, const AsynchJobs & Jobs, const EntityLog * Logging);

private:
	bool InsertResponseCert(NEWPKI_CERT_RESPONSE* cert_response, PKI_CRL & Crl, LOG_MESSAGE_TYPE & LogMessage, LOG_MESSAGE_STATUS & LogStatus, int & Status, mString & strError, const HashTable_Dn & ProfileDn, mString & strPrivDatas, mString & x509, unsigned long & serial, MailInfo & CertificateMail, MailInfo & PasswordMail, bool & SendCertificate, bool & SendPkcs12Pwd, bool & GenPkcs12Pwd);
	void GeneratePassword(mString & strP12Password);
	bool UpdateFromCRL(const mString & ca_name, const PKI_CRL & Crl);
	bool CanUserAccessProfile(const PKI_CERT & UserCert, unsigned long owner_serial, unsigned long owner_type);
	NEWPKI_PROFILE_DATAS_CERT * Sql2Cert(SQL * sql, int index);
	bool GetCerts(unsigned long ProfileId, STACK_OF(NEWPKI_PROFILE_DATAS_CERT) * Certs);

	#define OWNER_TYPE_USER		1		//!< The owner of the profile is a user
	#define OWNER_TYPE_GROUP	2		//!< The owner of the profile is a group


	bool Flush_ProfileInfo();
	#define RASTORE_CREATE_1				"create table profiles (profile_id INT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY, owner_type INT UNSIGNED NOT NULL, owner_serial INT UNSIGNED NOT NULL, dn_hash VARCHAR(41) NOT NULL, dn LONGBLOB NOT NULL, ldap_uid LONGBLOB NOT NULL, state INT UNSIGNED NOT NULL, datas LONGBLOB NOT NULL, signature LONGBLOB NOT NULL);"
	#define RASTORE_CREATE_2				"create table profiles_hash (id INT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY, hash LONGBLOB NOT NULL);"
	#define RASTORE_CREATE_3				"create table certs (id INT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY, profile_id INT UNSIGNED NOT NULL, ca_name BLOB NOT NULL, pkcs10 LONGBLOB NOT NULL, priv_datas LONGBLOB NOT NULL, status INT UNSIGNED NOT NULL, type INT UNSIGNED NOT NULL, flags LONGBLOB NOT NULL, admin_mail BLOB NOT NULL, error LONGBLOB NOT NULL, serial INT UNSIGNED NOT NULL, x509 LONGBLOB NOT NULL);"
	
	#define RASTORE_INSERT_PROFILE			"INSERT INTO profiles (owner_type, owner_serial, dn_hash, dn, ldap_uid, state, datas, signature) VALUES ('%d', '%ld', '%s', '%s', '%s', '%d', '%s', '%s');"
	#define RASTORE_GET_PROFILES			"SELECT profile_id, state, datas, signature FROM profiles %s ORDER BY profile_id DESC LIMIT %ld,%ld;"
	#define RASTORE_GET_PROFILES_CTRL		"SELECT datas, signature FROM profiles;"
	#define RASTORE_SEARCH_PROFILE			"SELECT profile_id FROM profiles WHERE profile_id='%ld';"
	#define RASTORE_SEARCH_PROFILE_DN		"SELECT profile_id FROM profiles WHERE dn_hash='%s';"
	#define RASTORE_INSERT_PROFILES_HASH	"REPLACE INTO profiles_hash (id, hash) VALUES (1, '%s');"
	#define RASTORE_GET_PROFILES_HASH		"SELECT * FROM profiles_hash WHERE id=1;"
	#define RASTORE_UPDATE_PROFILE_STATE	"UPDATE profiles SET state='%d' WHERE profile_id='%d';"
	#define RASTORE_DELETE_PROFILE			"DELETE FROM profiles WHERE profile_id='%ld';"
	#define RASTORE_GET_ALL_PROFILES		"SELECT profile_id, datas FROM profiles LIMIT %ld,%ld;"
	#define RASTORE_GET_PROFILE				"SELECT * FROM profiles WHERE profile_id='%ld';"
	#define RASTORE_GET_KNOWN_RDNS			"SELECT ldap_uid FROM profiles WHERE ldap_uid!='';"

	#define RASTORE_INSERT_CERT				"INSERT INTO certs (profile_id, ca_name, pkcs10, priv_datas, type, status, flags, admin_mail, error, serial, x509) VALUES ('%ld', '%s', '%s', '%s', '%d', '%d', '%s', '%s', '', 0, '');"
	#define RASTORE_GET_CERTS				"SELECT * FROM certs WHERE profile_id=%ld ORDER BY id DESC;"
	#define RASTORE_UPDATE_CERT_SET_X509	"UPDATE certs SET status=%d, x509='%s' WHERE id='%ld';"
	#define RASTORE_SUSPEND_CERT			"UPDATE certs SET status=%d WHERE id='%ld';"
	#define RASTORE_GET_CERT				"SELECT * FROM certs WHERE id=%ld;"
	#define RASTORE_DELETE_CERT				"DELETE FROM certs WHERE id='%ld';"
	#define RASTORE_GET_CERT_BY_TID			"SELECT certs.*, profiles.datas FROM certs, "REQUESTER_TABLE", profiles WHERE "REQUESTER_TABLE"."REQUESTER_PRIV_ATTR" = certs.id AND profiles.profile_id = certs.profile_id AND "REQUESTER_TABLE"."REQUESTER_TID"='%s';"
	#define RASTORE_SET_CERT_RESP			"UPDATE certs SET pkcs10='', status='%d', x509='%s', error='%s', priv_datas='%s', serial='%ld' WHERE id='%ld';"
	#define RASTORE_GET_CERT_4_REV			"SELECT certs.x509, certs.ca_name, certs.status, profiles.owner_type, profiles.owner_serial, profiles.datas FROM profiles, certs WHERE certs.id ='%ld' AND profiles.profile_id = certs.profile_id;"
	#define RASTORE_GET_CERT_OWNER			"SELECT certs.status, profiles.owner_type, profiles.owner_serial FROM profiles, certs WHERE certs.id ='%ld' AND profiles.profile_id = certs.profile_id;"
	#define RASTORE_DELETE_PKCS12			"UPDATE certs SET priv_datas='' WHERE id='%ld';"
	#define RASTORE_UPDATE_FROM_CRL			"UPDATE certs SET pkcs10='', priv_datas='', x509='', status='%d', serial='0' WHERE ca_name='%s' AND status!='%d' AND (%s);"


	#define NEWPKI_PROFILE_sign(x,sig,pkey) \
		ASN1_sign((int (*)())i2d_NEWPKI_PROFILE, sig->sig_alg, NULL, sig->signature, (char *)x, pkey, EVP_sha1())

	#define NEWPKI_PROFILE_verify(x,sig,pkey) \
		ASN1_verify((int (*)())i2d_NEWPKI_PROFILE,sig->sig_alg, sig->signature,(char *)x, pkey)

	bool ProfileExists(const mString & dn_hash, bool & exists);

	unsigned long m_MinPasswdLen;
	ReadersWriter ConfAccessLock;
	HashCorrelation ProfileHash;
	RA_STORE_INFO * ProfileInfo;
	USERS_GROUPS * m_groups;
	const X509_ACL_Validator * m_AclValidator;
};

#endif
