/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/

// RepStore.cpp: implementation of the RepStore class.
//
//////////////////////////////////////////////////////////////////////

#include "RepStore.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ASN1_SEQUENCE(REP_STORE_SIG) = {
	ASN1_SIMPLE(REP_STORE_SIG, signature, ASN1_BIT_STRING),
	ASN1_SIMPLE(REP_STORE_SIG, sig_alg, X509_ALGOR),
}ASN1_SEQUENCE_END(REP_STORE_SIG)
IMPLEMENT_ASN1_FUNCTIONS(REP_STORE_SIG);
IMPLEMENT_PEM_CONVERT(REP_STORE_SIG, "NEWPKI REP STORE SIGNATURE");

ASN1_SEQUENCE(REP_STORE_INFO) = {
	ASN1_SIMPLE(REP_STORE_INFO, hash, ASN1_UTF8STRING),
	ASN1_SIMPLE(REP_STORE_INFO, counter, ASN1_INTEGER),
}ASN1_SEQUENCE_END(REP_STORE_INFO)
IMPLEMENT_ASN1_FUNCTIONS(REP_STORE_INFO);
IMPLEMENT_ASN1_ENCRYPT_SIGN(REP_STORE_INFO);



RepStore::RepStore():NewPKIStore()
{
	ReqInfo = NULL;
	RespInfo = NULL;
}

RepStore::~RepStore()
{
	if(ReqInfo)
		REP_STORE_INFO_free(ReqInfo);
	if(RespInfo)
		REP_STORE_INFO_free(RespInfo);
}

bool RepStore::CreateTables(const SQL_Connection * DbConn)
{
	SQL sql(DbConn);
	long i;
	char * CommonCreates[] = {REPSTORE_CREATE_1, REPSTORE_CREATE_2, REPSTORE_CREATE_3, REPSTORE_CREATE_4, NULL};


	//We execute each request
	for(i=0; CommonCreates[i]; i++)
	{
		if(!sql.Execute(CommonCreates[i]))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}
	return true;
}

bool RepStore::Create()
{
	ReqHash.SetCert(m_EntityCert);
	RespHash.SetCert(m_EntityCert);

	if(RespInfo)
		REP_STORE_INFO_free(RespInfo);
	if(ReqInfo)
		REP_STORE_INFO_free(ReqInfo);

	RespInfo =  REP_STORE_INFO_new();
	if(!RespInfo)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	ReqInfo =  REP_STORE_INFO_new();
	if(!ReqInfo)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}
	ASN1_INTEGER_SET(&ReqInfo->counter, 0);
	ASN1_INTEGER_SET(&RespInfo->counter, 0);

	if(!Flush_ReqInfo())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(!Flush_RespInfo())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}



bool RepStore::Load()
{
	SQL sql(m_DbConn);
	mString pem_hash;
	long NumRows;
	ASN1_ENCRYPT_SIGN * crypted_conf;

	ConfAccessLock.LockWrite();

	ReqHash.SetCert(m_EntityCert);
	RespHash.SetCert(m_EntityCert);


	//We now have the hash, with there encrypted body
	//let's decrypt it
	//Decrypt the body and verify signature
	if(ReqInfo)
	{
		REP_STORE_INFO_free(ReqInfo);
		ReqInfo = NULL;
	}
	if(RespInfo)
	{
		REP_STORE_INFO_free(RespInfo);
		RespInfo = NULL;
	}



	//Load the requests info
	if(!sql.Execute(REPSTORE_GET_REQ_HASH))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	if(!sql.NumRows( &NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	if(!NumRows)
	{
		// TODO ajouter une entre de log
		// Something weird happended !
		if(!RegenerateHash(true))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			ConfAccessLock.Unlock();
			return false;
		}
	}
	else
	{
		if(!sql.Value(0,"hash", pem_hash))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			ConfAccessLock.Unlock();
			return false;
		}

		crypted_conf = NULL;
		if(!ASN1_ENCRYPT_SIGN_From_PEM(&crypted_conf, pem_hash.c_str()))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			ASN1_ENCRYPT_SIGN_free(crypted_conf);
			ConfAccessLock.Unlock();
			return false;
		}

		if(REP_STORE_INFO_decrypt_verify(crypted_conf, &ReqInfo, m_EntityCert.GetPrivateKey().GetRsaKey(), m_EntityCert.GetPrivateKey().GetRsaKey()) <= 0)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ASN1_ENCRYPT_SIGN_free(crypted_conf);
			ConfAccessLock.Unlock();
			return false;
		}
		ASN1_ENCRYPT_SIGN_free(crypted_conf);
		pem_hash = ReqInfo->hash;
		if(!ReqHash.SetCurrentHash(pem_hash))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ConfAccessLock.Unlock();
			return false;
		}
	}
	
	//Load the responses info
	if(!sql.Execute(REPSTORE_GET_RESP_HASH))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		return false;
	}
	if(!NumRows)
	{
		// TODO ajouter une entre de log
		// Something weird happended !
		if(!RegenerateHash(false))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			ConfAccessLock.Unlock();
			return false;
		}
	}
	else
	{
		if(!sql.Value(0,"hash", pem_hash))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			ConfAccessLock.Unlock();
			return false;
		}

		crypted_conf = NULL;
		if(!ASN1_ENCRYPT_SIGN_From_PEM(&crypted_conf, pem_hash.c_str()))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			ASN1_ENCRYPT_SIGN_free(crypted_conf);
			ConfAccessLock.Unlock();
			return false;
		}

		//We now have the hash, with there encrypted body
		//let's decrypt it
		//Decrypt the body and verify signature
		if(REP_STORE_INFO_decrypt_verify(crypted_conf, &RespInfo, m_EntityCert.GetPrivateKey().GetRsaKey(), m_EntityCert.GetPrivateKey().GetRsaKey()) <= 0)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ASN1_ENCRYPT_SIGN_free(crypted_conf);
			ConfAccessLock.Unlock();
			return false;
		}
		ASN1_ENCRYPT_SIGN_free(crypted_conf);
		pem_hash = RespInfo->hash;
		if(!RespHash.SetCurrentHash(pem_hash))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ConfAccessLock.Unlock();
			return false;
		}
	}

	ConfAccessLock.Unlock();
	return true;
}

bool RepStore::Flush_RespInfo()
{
	SQL sql(m_DbConn);
	mString hash;
	mString req;
	char * pem_info;
	ASN1_ENCRYPT_SIGN * crypt;

	if(!RespHash.GetHash(hash))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	if(ASN1_STRING_set(RespInfo->hash, hash.c_str(), hash.size()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	crypt = ASN1_ENCRYPT_SIGN_new();
	if(!crypt)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!REP_STORE_INFO_sign_encrypt(crypt, RespInfo, m_EntityCert.GetPrivateKey().GetRsaKey(), m_EntityCert.GetPublicKey(), EVP_sha1(), EVP_des_ede3_cbc()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ASN1_ENCRYPT_SIGN_free(crypt);
		return false;
	}
	pem_info = ASN1_ENCRYPT_SIGN_To_PEM(crypt);
	if(!pem_info)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ASN1_ENCRYPT_SIGN_free(crypt);
		return false;
	}
	ASN1_ENCRYPT_SIGN_free(crypt);


	if(req.sprintf(REPSTORE_INSERT_RESP_HASH, pem_info) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		free(pem_info);
		return false;
	}
	free(pem_info);

	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

bool RepStore::Flush_ReqInfo()
{
	SQL sql(m_DbConn);
	mString hash;
	mString req;
	char * pem_info;
	ASN1_ENCRYPT_SIGN * crypt;

	if(!ReqHash.GetHash(hash))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	if(ASN1_STRING_set(ReqInfo->hash, hash.c_str(), hash.size()) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	crypt = ASN1_ENCRYPT_SIGN_new();
	if(!crypt)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!REP_STORE_INFO_sign_encrypt(crypt, ReqInfo, m_EntityCert.GetPrivateKey().GetRsaKey(), m_EntityCert.GetPublicKey(), EVP_sha1(), EVP_des_ede3_cbc()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ASN1_ENCRYPT_SIGN_free(crypt);
		return false;
	}
	pem_info = ASN1_ENCRYPT_SIGN_To_PEM(crypt);
	if(!pem_info)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ASN1_ENCRYPT_SIGN_free(crypt);
		return false;
	}
	ASN1_ENCRYPT_SIGN_free(crypt);


	if(req.sprintf(REPSTORE_INSERT_REQ_HASH, pem_info) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		free(pem_info);
		return false;
	}
	free(pem_info);

	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	return true;
}

STORED_NEWPKI_REQUEST * RepStore::PublishRequest(CRYPTED_NEWPKI_REQUEST * request)
{
	STORED_NEWPKI_REQUEST * s_request;
	unsigned long CurrCtr;
	bool Exists;

	ConfAccessLock.LockWrite();

	s_request = STORED_NEWPKI_REQUEST_new();
	if(!s_request)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		ConfAccessLock.Unlock();
		return NULL;
	}

	// Generate a new transactionID, it has to be
	// unique PKI-Wise
	CurrCtr = ASN1_INTEGER_GET(ReqInfo->counter) + 1;
	if(!GenerateTransactionID(s_request->transactionID, request->sender, request->recipient, CurrCtr))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		STORED_NEWPKI_REQUEST_free(s_request);
		return NULL;
	}
	// Increment the unique counter
	ASN1_INTEGER_SET(&ReqInfo->counter, CurrCtr);

	if(!CRYPTED_NEWPKI_REQUEST_cpy(&s_request->c_info, request))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		STORED_NEWPKI_REQUEST_free(s_request);
		return NULL;
	}

	if(!InsertRequest(s_request, Exists))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		STORED_NEWPKI_REQUEST_free(s_request);
		return NULL;
	}

	ConfAccessLock.Unlock();
	return s_request;
}

bool RepStore::InsertRequest(STORED_NEWPKI_REQUEST *request, bool & Exists)
{
	SQL sql(m_DbConn);
	mString req;
	mString strTransactionID;
	mString	strPubKey;
	char * pem_datas;
	char * pem_signature;
	REP_STORE_SIG * signature;

	//Search if we already have this transaction
	if(!RequestExists(request->transactionID, Exists))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(Exists)
	{
		return true;
	}

	//Convert transaction ID to string
	if(!transactionIDtoString(request->transactionID, strTransactionID))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	// Convert public key to string
	if(!X509_PUBKEYtoString(request->c_info->recipient, strPubKey))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	signature = REP_STORE_SIG_new();
	if(!signature)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//Get the signature into pem
	if(!STORED_NEWPKI_REQUEST_sign(request, signature, (EVP_PKEY*)m_EntityCert.GetPrivateKey().GetRsaKey()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		REP_STORE_SIG_free(signature);
		return false;
	}
	pem_signature = REP_STORE_SIG_To_PEM(signature);
	if(!pem_signature)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		REP_STORE_SIG_free(signature);
		return false;
	}

	pem_datas = STORED_NEWPKI_REQUEST_To_PEM(request);
	if(!pem_datas)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		free(pem_signature);
		REP_STORE_SIG_free(signature);
		return false;
	}

	if(req.sprintf(REPSTORE_INSERT_REQ, strTransactionID.c_str(), strPubKey.c_str(), pem_datas, pem_signature) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		REP_STORE_SIG_free(signature);
		return false;
	}
	free(pem_datas);
	free(pem_signature);

	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		REP_STORE_SIG_free(signature);
		return false;
	}

	// Add this entry to the global hash
	if(!ReqHash.AddEntry((char*)request->transactionID->data, request->transactionID->length, (char*)signature->signature->data, signature->signature->length))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		REP_STORE_SIG_free(signature);
		return false;
	}
	REP_STORE_SIG_free(signature);

	if(!Flush_ReqInfo())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}

bool RepStore::InsertResponse(STORED_NEWPKI_RESPONSE *response, bool & Exists)
{
	SQL sql(m_DbConn);
	mString req;
	mString strTransactionID;
	mString	strPubKey;
	char * pem_datas;
	char * pem_signature;
	REP_STORE_SIG * signature;

	//Search if we already have this transaction
	if(!ResponseExists(response->transactionID, Exists))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(Exists)
	{
		return true;
	}

	//Convert transaction ID to string
	if(!transactionIDtoString(response->transactionID, strTransactionID))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	// Convert public key to string
	if(!X509_PUBKEYtoString(response->c_info->recipient, strPubKey))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	
	signature = REP_STORE_SIG_new();
	if(!signature)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	//Get the signature into pem
	if(!STORED_NEWPKI_RESPONSE_sign(response, signature, (EVP_PKEY*)m_EntityCert.GetPrivateKey().GetRsaKey()))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		REP_STORE_SIG_free(signature);
		return false;
	}
	pem_signature = REP_STORE_SIG_To_PEM(signature);
	if(!pem_signature)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		REP_STORE_SIG_free(signature);
		return false;
	}

	pem_datas = STORED_NEWPKI_RESPONSE_To_PEM(response);
	if(!pem_datas)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		free(pem_signature);
		REP_STORE_SIG_free(signature);
		return false;
	}

	if(req.sprintf(REPSTORE_INSERT_RESP, strTransactionID.c_str(), strPubKey.c_str(), pem_datas, pem_signature) <= 0)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		REP_STORE_SIG_free(signature);
		return false;
	}
	free(pem_datas);
	free(pem_signature);

	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		REP_STORE_SIG_free(signature);
		return false;
	}

	// Add this entry to the global hash
	if(!RespHash.AddEntry((char*)response->transactionID->data, response->transactionID->length, (char*)signature->signature->data, signature->signature->length))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		REP_STORE_SIG_free(signature);
		return false;
	}
	REP_STORE_SIG_free(signature);

	//Flush the hash to database
	if(!Flush_RespInfo())
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	return true;
}


bool RepStore::RequestExists(ASN1_OCTET_STRING * transactionID, bool & exists)
{
	mString strTransactionID;
	mString req;
	SQL sql(m_DbConn);
	long NumRows;

	if(!transactionIDtoString(transactionID, strTransactionID))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	req.sprintf(REPSTORE_SEARCH_REQ, strTransactionID.c_str());

	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(NumRows)
		exists = true;
	else
		exists = false;

	return true;
}

bool RepStore::ResponseExists(ASN1_OCTET_STRING * transactionID, bool & exists)
{
	mString strTransactionID;
	mString req;
	SQL sql(m_DbConn);
	long NumRows;

	if(!transactionIDtoString(transactionID, strTransactionID))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	req.sprintf(REPSTORE_SEARCH_RESP, strTransactionID.c_str());

	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}

	if(NumRows)
		exists = true;
	else
		exists = false;

	return true;
}

STORED_NEWPKI_REQUESTS * RepStore::GetRequests(const EVP_PKEY *pubkey, const TRANSACTION_IDS * notListIds)
{
	int i;
	STORED_NEWPKI_REQUESTS * requests;
	STORED_NEWPKI_REQUEST * request;
	ASN1_OCTET_STRING * currTransactionId;
	mString strPubKey;
	mString strTransactionID;
	mString req;
	mString where;
	SQL sql(m_DbConn);
	long NumRows;

	requests = STORED_NEWPKI_REQUESTS_new();
	if(!requests)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	// We build the WHERE part of the request
	if(pubkey)
	{
		//Convert the recipient key to string
		if(!EVP_PKEYtoString(pubkey, strPubKey))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			STORED_NEWPKI_REQUESTS_free(requests);
			return false;
		}

		where.sprintf("WHERE recipient='%s'", strPubKey.c_str());
		strPubKey="";
	}

	if(notListIds && SKM_sk_num(ASN1_OCTET_STRING, notListIds->transactionIds))
	{
		if(pubkey)
			where += " AND ";
		else
			where = "WHERE ";

		for(i=0; i<SKM_sk_num(ASN1_OCTET_STRING, notListIds->transactionIds); i++)
		{
			currTransactionId = SKM_sk_value(ASN1_OCTET_STRING, notListIds->transactionIds, i);
			if(!currTransactionId)
				continue;
			if(!transactionIDtoString(currTransactionId, strTransactionID))
				continue;

			where += "transactionID <> '";
			where += strTransactionID;
			where += "'";

			if((i+1) < SKM_sk_num(ASN1_OCTET_STRING, notListIds->transactionIds))
			{
				where += " AND ";
			}
		}
	}
	req.sprintf(REPSTORE_GET_REQS, where.c_str());

	ConfAccessLock.LockRead();

	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		STORED_NEWPKI_REQUESTS_free(requests);
		return NULL;
	}
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		STORED_NEWPKI_REQUESTS_free(requests);
		return NULL;
	}
	for(i=0; i<NumRows; i++)
	{
		request = SqlToRequest(i, sql);
		if(!request)
		{
			ERR_clear_error();
			continue;
		}

		if(SKM_sk_push(STORED_NEWPKI_REQUEST, requests->requests, request) < 0)
		{
			STORED_NEWPKI_REQUEST_free(request);
		}
	}
	ConfAccessLock.Unlock();
	return requests;
}

STORED_NEWPKI_REQUEST * RepStore::SqlToRequest(long i, SQL & sql)
{
	mString str_enum;
	STORED_NEWPKI_REQUEST * Value;

	Value = STORED_NEWPKI_REQUEST_new();
	if(!Value)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return NULL;
	}

	if(!sql.Value(i, "request_PEM", str_enum))
	{
		STORED_NEWPKI_REQUEST_free(Value);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}

	if(!STORED_NEWPKI_REQUEST_From_PEM(&Value, (char*)str_enum.c_str()))
	{
		STORED_NEWPKI_REQUEST_free(Value);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return NULL;
	}
	return Value;
}


STORED_NEWPKI_RESPONSES * RepStore::GetResponses(const EVP_PKEY *pubkey, const TRANSACTION_IDS * ListIds, bool inList)
{
	int i;
	STORED_NEWPKI_RESPONSES * responses;
	STORED_NEWPKI_RESPONSE * response;
	ASN1_OCTET_STRING * currTransactionId;
	mString strPubKey;
	mString strTransactionID;
	mString req;
	mString where;
	SQL sql(m_DbConn);
	long NumRows;

	responses = STORED_NEWPKI_RESPONSES_new();
	if(!responses)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return false;
	}

	if(pubkey)
	{
		//Convert the recipient key to string
		if(!EVP_PKEYtoString(pubkey, strPubKey))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			STORED_NEWPKI_RESPONSES_free(responses);
			return false;
		}

		// We build the WHERE part of the response
		where.sprintf("WHERE recipient='%s'", strPubKey.c_str());
		strPubKey="";
	}

	if(ListIds && SKM_sk_num(ASN1_OCTET_STRING, ListIds->transactionIds))
	{
		if(pubkey)
			where += " AND (";
		else
			where = "WHERE (";

		for(i=0; i<SKM_sk_num(ASN1_OCTET_STRING, ListIds->transactionIds); i++)
		{
			currTransactionId = SKM_sk_value(ASN1_OCTET_STRING, ListIds->transactionIds, i);
			if(!currTransactionId)
				continue;
			if(!transactionIDtoString(currTransactionId, strTransactionID))
				continue;

			if(inList)
				where += "transactionID = '";
			else 
				where += "transactionID <> '";
			where += strTransactionID;
			where += "'";

			if((i+1) < SKM_sk_num(ASN1_OCTET_STRING, ListIds->transactionIds))
			{
				if(inList)
					where += " OR ";
				else
					where += " AND ";

			}
		}
		where += ")";
	}
	req.sprintf(REPSTORE_GET_RESPS, where.c_str());

	ConfAccessLock.LockRead();

	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		STORED_NEWPKI_RESPONSES_free(responses);
		return NULL;
	}

	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		ConfAccessLock.Unlock();
		STORED_NEWPKI_RESPONSES_free(responses);
		return NULL;
	}
	for(i=0; i<NumRows; i++)
	{
		response = SqlToResponse(i, sql);
		if(!response)
		{
			ERR_clear_error();
			continue;
		}

		if(SKM_sk_push(STORED_NEWPKI_RESPONSE, responses->responses, response) < 0)
		{
			STORED_NEWPKI_RESPONSE_free(response);
		}
	}
	ConfAccessLock.Unlock();
	return responses;
}

STORED_NEWPKI_RESPONSE * RepStore::SqlToResponse(long i, SQL & sql)
{
	mString str_enum;
	STORED_NEWPKI_RESPONSE * Value;

	Value = STORED_NEWPKI_RESPONSE_new();
	if(!Value)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		return NULL;
	}

	if(!sql.Value(i, "response_PEM", str_enum))
	{
		STORED_NEWPKI_RESPONSE_free(Value);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return NULL;
	}

	if(!STORED_NEWPKI_RESPONSE_From_PEM(&Value, (char*)str_enum.c_str()))
	{
		STORED_NEWPKI_RESPONSE_free(Value);
		NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
		return NULL;
	}
	return Value;
}


bool RepStore::GetKnownRequests(TRANSACTION_IDS *tids)
{
	int i;
	ASN1_OCTET_STRING * currTransactionId;
	mString req;
	SQL sql(m_DbConn);
	long NumRows;

	req.sprintf(REPSTORE_GET_KNOWN_REQS);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	for(i=0; i<NumRows; i++)
	{
		if(!sql.Value(i, "transactionID", req))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		currTransactionId = StringtoTransactionID(req);
		if(!currTransactionId)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}

		if(SKM_sk_push(ASN1_OCTET_STRING, tids->transactionIds, currTransactionId) < 0)
		{
			ASN1_OCTET_STRING_free(currTransactionId);
		}
	}
	return true;
}

bool RepStore::GetKnownResponses(TRANSACTION_IDS *tids)
{
	int i;
	ASN1_OCTET_STRING * currTransactionId;
	mString req;
	SQL sql(m_DbConn);
	long NumRows;

	req.sprintf(REPSTORE_GET_KNOWN_RESPS);
	if(!sql.Execute(req))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	if(!sql.NumRows(&NumRows))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		return false;
	}
	for(i=0; i<NumRows; i++)
	{
		if(!sql.Value(i, "transactionID", req))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		currTransactionId = StringtoTransactionID(req);
		if(!currTransactionId)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}

		if(SKM_sk_push(ASN1_OCTET_STRING, tids->transactionIds, currTransactionId) < 0)
		{
			ASN1_OCTET_STRING_free(currTransactionId);
		}
	}
	return true;
}

bool RepStore::RegenerateHash(bool Request)
{
	SQL sql(m_DbConn);
	mString req;
	long NumRows;
	unsigned Offset = 0;
	int i;
	char * sqltemplate;
	ASN1_OCTET_STRING * transactionID;
	HashCorrelation * DestHash;
	REP_STORE_SIG * signature;
	unsigned long counter = 0;
	REP_STORE_INFO * DestStore;


	if(Request)
	{
		sqltemplate = REPSTORE_GET_REQS_LIMIT;
		DestHash = & ReqHash;

		ReqInfo =  REP_STORE_INFO_new();
		if(!ReqInfo)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
			return false;
		}
		ASN1_INTEGER_SET(&ReqInfo->counter, 0);

		DestStore = ReqInfo;
	}
	else
	{
		sqltemplate = REPSTORE_GET_RESPS_LIMIT;
		DestHash = & RespHash;

		RespInfo =  REP_STORE_INFO_new();
		if(!RespInfo)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
			return false;
		}
		ASN1_INTEGER_SET(&RespInfo->counter, 0);
		DestStore = RespInfo;
	}

	DestHash->Clear();
	do
	{
		if(req.sprintf(sqltemplate, Offset*200, 200) <= 0)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_BAD_DATAS);
			return false;
		}
		if(!sql.Execute(req))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		if(!sql.NumRows(&NumRows))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
		if(NumRows)
		{
			// Increment the counter for when for request
			if(Request)
			{
				counter += NumRows;
			}


			for(i=0; i<NumRows; i++)
			{
				// We get the transaction id
				if(!sql.Value(i, "transactionID", req))
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					return false;
				}

				if(! (transactionID = StringtoTransactionID(req)) ) 
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					return false;
				}

				// We get the signature
				if(!sql.Value(i, "signature", req))
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					ASN1_OCTET_STRING_free(transactionID);
					return false;
				}

				if(! (signature = REP_STORE_SIG_new()) )
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					ASN1_OCTET_STRING_free(transactionID);
					return false;
				}

				if(REP_STORE_SIG_From_PEM(&signature, req.c_str()) <= 0)
				{
					ASN1_OCTET_STRING_free(transactionID);
					REP_STORE_SIG_free(signature);
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					return false;
				}

				if(!DestHash->AddEntry((char*)transactionID->data, transactionID->length, (char*)signature->signature->data, signature->signature->length))
				{
					NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
					ASN1_OCTET_STRING_free(transactionID);
					REP_STORE_SIG_free(signature);
					return false;
				}
				REP_STORE_SIG_free(signature);
				ASN1_OCTET_STRING_free(transactionID);
			}
		}
		Offset++;
	}
	while(NumRows);


	if(Request)
	{
		ASN1_INTEGER_SET(&ReqInfo->counter, counter);
		if(!Flush_ReqInfo())
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}
	else
	{
		if(!Flush_RespInfo())
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			return false;
		}
	}

	return true;
}
