/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/

// RepStore.h: interface for the RepStore class.
//
//////////////////////////////////////////////////////////////////////

#ifndef REPSTORE_H
#define REPSTORE_H

#include "NewPKIStore.h"
#include <ReadersWriter.h>
#include <PEM/PEM_DER.h>
#include <HashCorrelation.h>

#ifndef DOXYGEN_SHOULD_SKIP_THIS
typedef struct st_REP_STORE_SIG
{
	ASN1_BIT_STRING * signature;
	X509_ALGOR * sig_alg;
}REP_STORE_SIG;
DECLARE_ASN1_FUNCTIONS(REP_STORE_SIG)
DECLARE_PEM_CONVERT(REP_STORE_SIG)


typedef struct st_REP_STORE_INFO
{
	ASN1_UTF8STRING * hash;
	ASN1_INTEGER * counter;
}REP_STORE_INFO;
DECLARE_ASN1_FUNCTIONS(REP_STORE_INFO)
DECLARE_ASN1_ENCRYPT_SIGN(REP_STORE_INFO)
#endif

/*!
	This class is the store for a Repository
*/
class RepStore : public NewPKIStore
{
public:
	/*! \brief This is the constructor.
	 */
	RepStore();

	/*! \brief This is the destructor.
	 */
	virtual ~RepStore();

	bool CreateTables(const SQL_Connection * DbConn);

	/*! \brief This function creates the intial datas.
	 *  \return true on success, false on failure.
	 */
	bool Create();

	/*! \brief This function loads the intial datas.
	 *  \return true on success, false on failure.
	 */
	bool Load();

	/*! \brief This function returns all the known responses.
	 *  \param tids [OUT] All the known responses.
	 *  \return true on success, false on failure.
	 */
	bool GetKnownResponses(TRANSACTION_IDS * tids);
	
	/*! \brief This function returns all the known requests.
	 *  \param tids [OUT] All the known requests.
	 *  \return true on success, false on failure.
	 */
	bool GetKnownRequests(TRANSACTION_IDS * tids);
	
	/*! \brief This function function stores a request received from another entity (not another repository).
	 *  \param request [IN] The new request.
	 *  \return The request + transactionId, the transactionId is sent back to the entity, the formated request is sent to all the other repositories.
	 */
	STORED_NEWPKI_REQUEST * PublishRequest(CRYPTED_NEWPKI_REQUEST * request);
	
	/*! \brief This function function stores a request received from another repository.
	 *  \param request [IN] The new request.
	 *  \param Exists [OUT] true when exists, false when doesn't exist.
	 *  \return true on success, false on failure.
	 */
	bool InsertRequest(STORED_NEWPKI_REQUEST * request, bool & Exists);
	
	/*! \brief This function function stores a response received from another entity.
	 *  \param response [IN] The new response.
	 *  \param Exists [OUT] true when exists, false when doesn't exist.
	 *  \return true on success, false on failure.
	 */
	bool InsertResponse(STORED_NEWPKI_RESPONSE *response, bool & Exists);
	
	/*! \brief This function function returns a list of requests.
	 *  \param pubkey [IN] The recipient public key, can be NULL when synchronizing with another repository.
	 *  \param notListIds [IN] A list of requests the recipient doesn't want (because he already knows them).
	 *  \return The list of requests.
	 */
	STORED_NEWPKI_REQUESTS * GetRequests(const EVP_PKEY * pubkey, const TRANSACTION_IDS * notListIds);
	
	/*! \brief This function function returns a list of responses.
	 *  \param pubkey [IN] The recipient public key, can be NULL when synchronizing with another repository.
	 *  \param ListIds [IN] A list of TID, which the user wants or doesn't want, depending on inList.
	 *  \param inList [IN] Is ListIds a list of responses the user want, or doesn't want.
	 *  \return The list of responses.
	 */
	STORED_NEWPKI_RESPONSES * GetResponses(const EVP_PKEY *pubkey, const TRANSACTION_IDS * ListIds, bool inList = true);
	
	/*! \brief This function checks if a request is already known.
	 *  \param transactionID [IN] The request's transaction ID.
	 *  \param exists [OUT] true when exists, false when doesn't exist.
	 *  \return true on success, false on failure.
	 */
	bool RequestExists(ASN1_OCTET_STRING * transactionID, bool & exists);
	
	/*! \brief This function checks if a response is already known.
	 *  \param transactionID [IN] The response's transaction ID.
	 *  \param exists [OUT] true when exists, false when doesn't exist.
	 *  \return true on success, false on failure.
	 */
	bool ResponseExists(ASN1_OCTET_STRING * transactionID, bool & exists);
private:
	bool RegenerateHash(bool Request);
	bool Flush_ReqInfo();
	bool Flush_RespInfo();
	#define REPSTORE_CREATE_1				"create table requests (transactionID VARCHAR(41) NOT NULL PRIMARY KEY, recipient BLOB NOT NULL, request_PEM LONGBLOB NOT NULL, signature LONGBLOB NOT NULL);"
	#define REPSTORE_CREATE_2				"create table responses (transactionID VARCHAR(41) NOT NULL PRIMARY KEY, recipient BLOB NOT NULL, response_PEM LONGBLOB NOT NULL, signature LONGBLOB NOT NULL);"
	#define REPSTORE_CREATE_3				"create table req_hash (id INT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY, hash LONGBLOB NOT NULL);"
	#define REPSTORE_CREATE_4				"create table resp_hash (id INT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY, hash LONGBLOB NOT NULL);"
	
	#define REPSTORE_INSERT_REQ				"INSERT INTO requests (transactionID, recipient, request_PEM, signature) VALUES ('%s', '%s', '%s', '%s');"
	#define REPSTORE_GET_REQS				"SELECT request_PEM FROM requests %s;"
	#define REPSTORE_GET_REQS_CTRL			"SELECT request_PEM, signature FROM requests;"
	#define REPSTORE_SEARCH_REQ				"SELECT transactionID FROM requests WHERE transactionID='%s';"
	#define REPSTORE_INSERT_REQ_HASH		"REPLACE INTO req_hash (id, hash) VALUES (1, '%s');"
	#define REPSTORE_GET_REQ_HASH			"SELECT * FROM req_hash WHERE id=1;"
	#define REPSTORE_GET_KNOWN_REQS			"SELECT transactionID FROM requests;"
	#define REPSTORE_GET_REQS_LIMIT			"SELECT transactionID, signature FROM requests LIMIT %d,%d;;"

	#define REPSTORE_INSERT_RESP			"INSERT INTO responses (transactionID, recipient, response_PEM, signature) VALUES ('%s', '%s', '%s', '%s');"
	#define REPSTORE_GET_RESPS				"SELECT response_PEM FROM responses %s;"
	#define REPSTORE_GET_RESPS_CTRL			"SELECT response_PEM, signature FROM responses;"
	#define REPSTORE_SEARCH_RESP			"SELECT transactionID FROM responses WHERE transactionID='%s';"
	#define REPSTORE_INSERT_RESP_HASH		"REPLACE INTO resp_hash (id, hash) VALUES (1, '%s');"
	#define REPSTORE_GET_RESP_HASH			"SELECT * FROM resp_hash WHERE id=1;"
	#define REPSTORE_GET_KNOWN_RESPS		"SELECT transactionID FROM responses;"
	#define REPSTORE_GET_RESPS_LIMIT		"SELECT transactionID, signature FROM responses LIMIT %d,%d;;"


	#define STORED_NEWPKI_REQUEST_sign(x,sig,pkey) \
		ASN1_sign((int (*)())i2d_STORED_NEWPKI_REQUEST, sig->sig_alg, NULL, sig->signature, (char *)x, pkey, EVP_sha1())

	#define STORED_NEWPKI_REQUEST_verify(x,sig,pkey) \
		ASN1_verify((int (*)())i2d_STORED_NEWPKI_REQUEST,sig->sig_alg, sig->signature,(char *)x, pkey)


	#define STORED_NEWPKI_RESPONSE_sign(x,sig,pkey) \
		ASN1_sign((int (*)())i2d_STORED_NEWPKI_RESPONSE, sig->sig_alg, NULL, sig->signature, (char *)x, pkey, EVP_sha1())

	#define STORED_NEWPKI_RESPONSE_verify(x,sig,pkey) \
		ASN1_verify((int (*)())i2d_STORED_NEWPKI_RESPONSE, sig->sig_alg, sig->signature,(char *)x, pkey)


	STORED_NEWPKI_REQUEST * SqlToRequest(long i, SQL & sql);
	STORED_NEWPKI_RESPONSE * SqlToResponse(long i, SQL & sql);

	
	ReadersWriter ConfAccessLock;
	HashCorrelation RespHash;
	HashCorrelation ReqHash;
	REP_STORE_INFO * ReqInfo;
	REP_STORE_INFO * RespInfo;
};

#endif
