/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/


// SockServerADMIN.cpp: implementation of the SockServerADMIN class.
//
//////////////////////////////////////////////////////////////////////

#include "SockServerADMIN.h"
#include "svintl.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


SockServerADMIN::SockServerADMIN():SOCK_SERVER(MAX_ADMIN_CONNECTIONS)
{
	cl_ctr = 0;
}

SockServerADMIN::~SockServerADMIN()
{
	HandlesList.clear();
}

void SockServerADMIN::OnServerStopped()
{
}

ADMIN_REQUEST * SockServerADMIN::ReadRequest(BIO * ssl_bio, SOCKET hSocket)
{
	ADMIN_REQUEST * req;
	fd_set rfds;
	struct timeval tv;
	int sockt_state;

	do
	{
		FD_ZERO(&rfds);
		FD_SET(hSocket, &rfds);
		tv.tv_sec = 0;
		tv.tv_usec = 100;
		sockt_state = select(hSocket + 1, &rfds, NULL, NULL, &tv);
	}
	while(sockt_state == 0 && !ShouldStop());

	// Where we asked to stop or was there an error ?
	if(ShouldStop() || sockt_state != 1)
	{
		return NULL;
	}

	req = d2i_ADMIN_REQUEST_bio(ssl_bio, NULL);
	return req;
}


void SockServerADMIN::OnConnection(const char * Ip, SOCKET connection)
{
	SSL * ssl;
	BIO * ssl_bio;
	ADMIN_REQUEST * req;
	ADMIN_RESPONSE * response;
	X509 * ClientCert; 
	PKI_CERT UserCert;
	int ret;
	SSL_CTX * ctx;
	char * subject;
	char * issuer;
	char * tmperr;
	char * errorstr;
	const THREAD_SSL_CTX & SslCtx = GetThread_SSL_CTX();
	ctx = SslCtx.GetSSL_CTX();
	if(!ctx)
	{
		return;
	}
	if( !(ssl_bio=BIO_new(BIO_f_ssl())) )
	{
		return;
	}
	if( !(ssl=SSL_new(ctx)) )
	{
		BIO_free_all(ssl_bio);
		return;
	}

	SSL_set_mode(ssl, SSL_MODE_AUTO_RETRY);

	SSL_set_fd(ssl, connection);

	if(BIO_set_ssl(ssl_bio, ssl, BIO_CLOSE) <= 0)
	{
		BIO_free_all(ssl_bio);
		return;
	}

	if(BIO_set_ssl_mode(ssl_bio, 0) <= 0)
	{
		BIO_free_all(ssl_bio);
		return;
	}

	ERR_clear_error();
	if( (ret = SSL_accept(ssl)) <= 0)
	{
		tmperr = NULL;
		switch(SSL_get_error(ssl, ret))
		{
			case SSL_ERROR_SSL:
				tmperr = ERR_to_string();
				errorstr = tmperr;
				break;
			case SSL_ERROR_SYSCALL:
				errorstr = strerror(errno);
				break;
			default:
				errorstr = "Other";
				break;
		}
		NewpkiDebug(LOG_LEVEL_ERROR, m_Conf.get_ServerName().c_str(), _sv("%.20d - SSL_accept failed : %d -> %s\n"), connection, SSL_get_error(ssl, ret), errorstr);
		if(tmperr) free(tmperr);

		BIO_free_all(ssl_bio);
		return;
	}


	ClientCert = SSL_get_peer_certificate(ssl);
	if(ClientCert)
	{
		if(!UserCert.SetCert(ClientCert))
		{
			BIO_free_all(ssl_bio);
			return;
		}


		issuer = X509_NAME_oneline (X509_get_issuer_name(ClientCert),0,0);
		subject = X509_NAME_oneline (X509_get_subject_name(ClientCert),0,0);
		NewpkiDebug(LOG_LEVEL_DEBUG, m_Conf.get_ServerName().c_str(), _sv("%.20d - SSLv3 Connection from: \n\tissuer : %s\n\tsubject : %s\n"), connection, issuer?issuer:_sv("Unknown"), subject?subject:_sv("Unknown"));
		if(issuer)
			OPENSSL_free(issuer);
		if(subject)
			OPENSSL_free(subject);

		X509_free(ClientCert);
	}

	
	response = ADMIN_RESPONSE_new();
	if(!response)
	{
		BIO_free_all(ssl_bio);
		return;
	}
	req = NULL;	
	ERR_clear_error();

	if(response->signing_cert)
	{
		X509_free(response->signing_cert);
		response->signing_cert = NULL;
	}
/*
	unsigned long my_ctr = 0;
	mString tmpfile;
	FILE * fp;
	unsigned long tmp_ctr = 0;
	my_ctr = cl_ctr;
	cl_ctr++;
*/
	while( (req = ReadRequest(ssl_bio, connection)))
	{
/*
		tmpfile.sprintf("d:\\asn1\\%ld - %ld.txt", my_ctr, tmp_ctr);
		fp = fopen(tmpfile.c_str(), "wb");
		if(fp)
		{
			ASN1_i2d_fp((int (*)(void))i2d_ADMIN_REQUEST,fp,(unsigned char *)req);
			fclose(fp);
		}
		tmp_ctr++;
*/		
		m_Central->OnADMIN_REQUEST(response->body, Ip, UserCert, req);
		ADMIN_REQUEST_free(req);
		if(!response->body->d.other)
		{
			break;
		}
		else
		{
			if(!m_Central->SignResponse(response))
			{
				break;
			}
			if(!i2d_ADMIN_RESPONSE_bio(ssl_bio, response))
			{
				break;
			}
		}
		ADMIN_RESPONSE_BODY_free(response->body);
		response->body = ADMIN_RESPONSE_BODY_new();
		if(!response->body)
		{
			break;
		}
	}

	ADMIN_RESPONSE_free(response);	

#ifdef _DEBUG
	ERR_print_errors_fp(stdout);
#endif

	AddRecvBytes(BIO_number_read(ssl_bio));
	AddSentBytes(BIO_number_written(ssl_bio));
	BIO_free_all(ssl_bio);
	m_Central->OnConnectionClosed();
}

void SockServerADMIN::OnServerStarted()
{

}

bool SockServerADMIN::Load(PKIX_Central *Central)
{
	m_Central = Central;
	m_Central->GetSSL_Cert(m_SslCert);
	return true;
}

int SockServerADMIN::verify_callback(int ok, X509_STORE_CTX *ctx)
{
	return 1;
}


const THREAD_SSL_CTX & SockServerADMIN::GetThread_SSL_CTX()
{

	NEWPKI_THREAD_ID hThread = NewpkiThread::CurrentThreadId();

	HandlesListLock.Lock();
	// Is this the first call for this thread ?
	if(HandlesList.find(hThread) != HandlesList.end())
	{
		const THREAD_SSL_CTX & ret1 = HandlesList[hThread];
		HandlesListLock.Unlock();
		return ret1;
	}


	// Create the new item
	if(HandlesList[hThread].SetCert(m_SslCert))
	{
		SSL_CTX_set_verify(HandlesList[hThread].GetSSL_CTX(), SSL_VERIFY_PEER|SSL_VERIFY_CLIENT_ONCE, verify_callback);
	//	SSL_CTX_set_info_callback(HandlesList[hThread].GetSSL_CTX(), info_callback);
	}
	const THREAD_SSL_CTX & ret2 = HandlesList[hThread];
	HandlesListLock.Unlock();
	return ret2;
}


void SockServerADMIN::info_callback(const SSL *s, int where, int ret)
{
    if(where & SSL_CB_LOOP)
		NewpkiDebug(LOG_LEVEL_DEBUG, "SSL INFO", _sv("%.20d - SSL state (%s): %s"), SSL_get_fd((SSL*)s), 
        where & SSL_ST_CONNECT ? "connect" :
        where & SSL_ST_ACCEPT ? "accept" :
        "undefined", SSL_state_string_long(s));
    else if(where & SSL_CB_ALERT)
        NewpkiDebug(LOG_LEVEL_DEBUG, "SSL INFO", _sv("%.20d - SSL alert (%s): %s: %s"), SSL_get_fd((SSL*)s), 
            where & SSL_CB_READ ? "read" : "write",
            SSL_alert_type_string_long(ret),
            SSL_alert_desc_string_long(ret));
/*
    else if(where==SSL_CB_HANDSHAKE_DONE)
        print_stats(s->ctx);*
*/
}

