/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/

// SockServerADMIN.h: interface for the SockServerADMIN class.
//
//////////////////////////////////////////////////////////////////////

#ifndef SOCKSERVERADMIN_H
#define SOCKSERVERADMIN_H

#include "PKIX_Central.h"
#include <PKI_CERT.h>
#include "SOCK_SERVER.h"

#include <map>
using namespace std;

#ifndef MAX_ADMIN_CONNECTIONS
	#define MAX_ADMIN_CONNECTIONS	50
#endif

/*!
	This class is a per thread SSL context
*/
class THREAD_SSL_CTX
{
public:
	/*! \brief This is the constructor.
	 */
	THREAD_SSL_CTX()
	{
		m_ctx = NULL;
	}

	/*! \brief This is the constructor.
	 *  \param other [IN] Load datas from another instance.
	 */
	THREAD_SSL_CTX(const THREAD_SSL_CTX & other)
	{
		m_ctx = NULL;
		*this = other;
	}

	/*! \brief This is the destructor.
	 */
	~THREAD_SSL_CTX()
	{
		if(m_ctx)
		{
			SSL_CTX_free(m_ctx);
			m_ctx = NULL;
		}
	}

	/*! \brief This function set the server's certificate
	 *  \param cert [IN] The server's certificate.
	 *  \return true on success, false on failure.
	 */
	bool SetCert(const PKI_CERT & cert)
	{
		SSL_METHOD * meth;
		if(!cert)
		{
			return false;
		}
		if(! (m_cert = cert))
		{
			return false;
		}

		if(m_ctx)
		{
			SSL_CTX_free(m_ctx);
			m_ctx = NULL;
		}
		meth = SSLv23_server_method();
		if(!meth)
		{
			return false;
		}
		m_ctx = SSL_CTX_new (meth);
		if(!m_ctx)
		{
			return false;
		}
		if (SSL_CTX_use_certificate(m_ctx, m_cert.GetX509()) <= 0)
		{
			SSL_CTX_free(m_ctx);
			m_ctx = NULL;
			return false;
		}
		if (SSL_CTX_use_PrivateKey(m_ctx, (EVP_PKEY*)m_cert.GetPrivateKey().GetRsaKey()) <= 0) 
		{
			SSL_CTX_free(m_ctx);
			m_ctx = NULL;
			return false;
		}
		if (SSL_CTX_check_private_key(m_ctx) <= 0)
		{
			SSL_CTX_free(m_ctx);
			m_ctx = NULL;
			return false;
		}

		SSL_CTX_set_mode(m_ctx, SSL_MODE_AUTO_RETRY);

		return true;
	}

	/*! \brief This function returns the SSL context
	 *  \return The SSL context.
	 */
	SSL_CTX * GetSSL_CTX() const
	{
		return m_ctx;
	}

	/*! \brief This operator copies one THREAD_SSL_CTX into another.
	 *  \param other [IN] The other context to copy from.
	 *  \return true on success, false on failure.
	 */
	bool operator=(const THREAD_SSL_CTX & other)
	{
		return SetCert(other.m_cert);
	}

private:
	PKI_CERT m_cert;
	SSL_CTX * m_ctx;
};


/*!
	This class handles newpki's protocol
*/
class SockServerADMIN : public SOCK_SERVER  
{
public:
	/*! \brief This is the constructor.
	 */
	SockServerADMIN();

	/*! \brief This is the destructor.
	 */
	virtual ~SockServerADMIN();

	/*! \brief This function loads the provider.
	 *  \param Central [IN] The accessor to the entities.
	 */
	bool Load(PKIX_Central *Central);

private:
	ADMIN_REQUEST * ReadRequest(BIO * ssl_bio, SOCKET hSocket);
	static void info_callback(const SSL *s, int where, int ret);
	const THREAD_SSL_CTX & GetThread_SSL_CTX();
	static int verify_callback(int ok, X509_STORE_CTX *ctx);
	void OnServerStarted();
	void OnConnection(const char * Ip, SOCKET connection);
	void OnServerStopped();
	PKIX_Central * m_Central;
	PKI_CERT m_SslCert;
	CriticalSection HandlesListLock;
	unsigned long cl_ctr;
	map<NEWPKI_THREAD_ID, THREAD_SSL_CTX> HandlesList;
};

#endif
