/*
	Copyright (C) 2003 Frdric Giudicelli (contact_nos@yahoo.com). 
	All rights reserved.

	This product includes cryptographic software written by Eric Young
	(eay@cryptsoft.com)

	This program is released under the GPL with the additional exemption that
	compiling, linking, and/or using OpenSSL is allowed.

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation; either version 2 of the License.

	This program is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	this program; if not, write to the Free Software Foundation, Inc., 59 Temple
	Place, Suite 330, Boston, MA 02111-1307 USA
*/

// X509_ACL_Validator.cpp: implementation of the X509_ACL_Validator class.
//
//////////////////////////////////////////////////////////////////////

#include "X509_ACL_Validator.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

X509_ACL_Validator::X509_ACL_Validator()
{
	m_acl = NULL;
	m_groups = NULL;
}

X509_ACL_Validator::~X509_ACL_Validator()
{
	if(m_groups) USERS_GROUPS_free(m_groups);
	if(m_acl) X509_ACL_free(m_acl);
}

bool X509_ACL_Validator::SetGroups(const USERS_GROUPS * groups)
{
	AccessLock.LockWrite();

	if(!USERS_GROUPS_cpy(&m_groups, groups))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		AccessLock.Unlock();
		return false;
	}
	AccessLock.Unlock();

	return true;
}

bool X509_ACL_Validator::SetACL(const X509_ACL *acl)
{
	AccessLock.LockWrite();

	if(!X509_ACL_cpy(&m_acl, acl))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		AccessLock.Unlock();
		return false;
	}
	AccessLock.Unlock();

	return true;
}

bool X509_ACL_Validator::SetCRL(const INTERNAL_PKI_CRL *crl)
{
	AccessLock.LockWrite();
	if(crl->RootCaCrl && crl->RootCaCrl->crl->version && !m_RootCaCrl.SetCRL(crl->RootCaCrl))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		AccessLock.Unlock();
		return false;
	}
	if(crl->EntitiesCaCrl && crl->EntitiesCaCrl->crl->version && !m_EntitiesCaCrl.SetCRL(crl->EntitiesCaCrl))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		AccessLock.Unlock();
		return false;
	}
	if(crl->OcspCaCrl && crl->OcspCaCrl->crl->version && !m_OcspCaCrl.SetCRL(crl->OcspCaCrl))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		AccessLock.Unlock();
		return false;
	}
	if(crl->UsersCaCrl && crl->UsersCaCrl->crl->version && !m_UsersCaCrl.SetCRL(crl->UsersCaCrl))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		AccessLock.Unlock();
		return false;
	}

	AccessLock.Unlock();

	return true;
}

bool X509_ACL_Validator::SetCA(const INTERNAL_PKI_CA *ca)
{
	AccessLock.LockWrite();

	if(!m_RootCa.SetCert(ca->RootCa))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		AccessLock.Unlock();
		return false;
	}
	if(!m_EntitiesCa.SetCert(ca->EntitiesCa))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		AccessLock.Unlock();
		return false;
	}
	if(!m_UsersCa.SetCert(ca->UsersCa))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		AccessLock.Unlock();
		return false;
	}
	if(!m_OcspCa.SetCert(ca->OcspCa))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		AccessLock.Unlock();
		return false;
	}
	if(!m_PkiCert.SetCert(ca->PkiCert))
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
		AccessLock.Unlock();
		return false;
	}

	AccessLock.Unlock();

	return true;
}


int X509_ACL_Validator::ValidateCert(const PKI_CERT & usercert) const
{
	int IssuerType;
	time_t currTime;
	const PKI_CRL * IssuerCrl;

	if(!usercert)
	{
		return -1;
	}

	AccessLock.LockRead();

	//Is it the PKI cert ?
	if(usercert == m_PkiCert)
	{
		AccessLock.Unlock();
		return 0;
	}

	//Is certificate valid ?
	time_gmt(&currTime);
	if( (usercert.GetStartDate() > currTime) || (currTime >= usercert.GetEndDate()))
	{
		AccessLock.Unlock();
		return -1;
	}


	//Get issuer type
	if(X509_verify(usercert.GetX509(), (EVP_PKEY*)m_EntitiesCa.GetPublicKey()) > 0)
	{
		IssuerType = INTERNAL_CA_TYPE_ENTITY;
		IssuerCrl = &m_EntitiesCaCrl;
		if(m_RootCaCrl.IsRevoked(m_EntitiesCa))
		{
			AccessLock.Unlock();
			ERR_clear_error();
			return -1;
		}
	}
	else if(X509_verify(usercert.GetX509(), (EVP_PKEY*)m_UsersCa.GetPublicKey()) > 0)
	{
		IssuerType = INTERNAL_CA_TYPE_USER;
		IssuerCrl = &m_UsersCaCrl;
		if(m_RootCaCrl.IsRevoked(m_UsersCa))
		{
			AccessLock.Unlock();
			ERR_clear_error();
			return -1;
		}
	}
	else if(X509_verify(usercert.GetX509(), (EVP_PKEY*)m_RootCa.GetPublicKey()) > 0)
	{
		IssuerType = INTERNAL_CA_TYPE_ROOT;
		IssuerCrl = &m_RootCaCrl;
	}
	else if(X509_verify(usercert.GetX509(), (EVP_PKEY*)m_OcspCa.GetPublicKey()) > 0)
	{
		IssuerType = INTERNAL_CA_TYPE_OCSP;
		IssuerCrl = &m_OcspCaCrl;
		if(m_RootCaCrl.IsRevoked(m_OcspCa))
		{
			AccessLock.Unlock();
			ERR_clear_error();
			return -1;
		}
	}
	else 
	{
		AccessLock.Unlock();
		ERR_clear_error();
		return -1;
	}
	if(IssuerCrl->IsRevoked(usercert))
	{
		AccessLock.Unlock();
		ERR_clear_error();
		return -1;
	}
	AccessLock.Unlock();

	ERR_clear_error();
	return IssuerType;
}


bool X509_ACL_Validator::CanUserPerform(const PKI_CERT & usercert, int command, bool CheckCert) const
{
	USERS_GROUP * currGroup;
	ASN1_INTEGER * currSerial;
	ACL_ENTRY * currAcl;
	int i;
	int j;

	if(CheckCert && ValidateCert(usercert) != INTERNAL_CA_TYPE_USER)
	{
		return false;
	}

	AccessLock.LockRead();

	//This is the PKI admin
	if(IsPkiAdministrator(usercert))
	{
		AccessLock.Unlock();
		return true;
	}

	currAcl = FindAcl(ACL_ENTRY_TYPE_USER, usercert.GetSerial());
	if(!currAcl)
	{
		AccessLock.Unlock();
		return false;
	}

	//We check the ACL
	if(ASN1_BIT_STRING_get_bit(currAcl->acls, command))
	{
		AccessLock.Unlock();
		return true;
	}
	else
	{

		//Check the groups rights
		for(i=0; i<SKM_sk_num(USERS_GROUP, m_groups->list); i++)
		{
			currGroup = SKM_sk_value(USERS_GROUP, m_groups->list, i);
			if(!currGroup) continue;

			// Get the ACL for this group
			currAcl = FindAcl(ACL_ENTRY_TYPE_GROUP, ASN1_INTEGER_GET(currGroup->serial));
			if(!currAcl) continue;

			// Is the group the right ?
			if(!ASN1_BIT_STRING_get_bit(currAcl->acls, command)) continue;

			// Is the user in the group ?
			for(j=0; j<SKM_sk_num(ASN1_INTEGER, currGroup->users_serial); j++)
			{
				currSerial = SKM_sk_value(ASN1_INTEGER, currGroup->users_serial, j);
				if(!currSerial) continue;

				// The user is the the group !!!
				if((unsigned long)ASN1_INTEGER_GET(currSerial) == usercert.GetSerial())
				{
					AccessLock.Unlock();
					return true;
				}

			}
		}

		AccessLock.Unlock();
		return false;
	}
}

ACL_ENTRY * X509_ACL_Validator::FindAcl(int type, unsigned long serial) const
{
	int i;
	ACL_ENTRY * currAcl;

	for(i=0; i<SKM_sk_num(ACL_ENTRY, m_acl->acl_entries); i++)
	{
		currAcl = SKM_sk_value(ACL_ENTRY, m_acl->acl_entries, i);
		if(!currAcl) continue;
		if(ASN1_INTEGER_GET(currAcl->type) != type) continue;

		if((unsigned long)ASN1_INTEGER_GET(currAcl->serial) == serial)
		{
			return currAcl;
		}
	}
	return NULL;
}

bool X509_ACL_Validator::GetUserRights(ASN1_BIT_STRING ** resAcl, unsigned long serial) const
{
	USERS_GROUP * currGroup;
	ASN1_INTEGER * currSerial;
	int i;
	int j;
	int k;
	int min_len;
	int max_len;
	ACL_ENTRY * acl;
	unsigned char * buffer;

	// we are ORing the user rights and its groups rights
	AccessLock.LockRead();

	//Find the longest ACL in the groups
	max_len = 0;
	for(i=0; i<SKM_sk_num(USERS_GROUP, m_groups->list); i++)
	{
		currGroup = SKM_sk_value(USERS_GROUP, m_groups->list, i);
		if(!currGroup) continue;

		acl = FindAcl(ACL_ENTRY_TYPE_GROUP, ASN1_INTEGER_GET(currGroup->serial));
		if(!acl) continue;		
		if(acl->acls->length > max_len) max_len = acl->acls->length;
	}

	acl = FindAcl(ACL_ENTRY_TYPE_USER, serial);
	if(!acl)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_UNKNOWN);
		AccessLock.Unlock();
		return false;
	}
	if(acl->acls->length > max_len) max_len = acl->acls->length;

	*resAcl = ASN1_BIT_STRING_new();
	if(!*resAcl)
	{
		NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
		AccessLock.Unlock();
		return false;
	}
	if(max_len)
	{
		buffer = (unsigned char*)malloc(max_len);
		if(!buffer)
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_MALLOC);
			ASN1_BIT_STRING_free(*resAcl);
			*resAcl = NULL;
			AccessLock.Unlock();
			return false;
		}
		memset(buffer, 0, max_len);
		if(!ASN1_BIT_STRING_set(*resAcl, buffer, max_len))
		{
			NEWPKIerr(PKI_ERROR_TXT, ERROR_ABORT);
			ASN1_BIT_STRING_free(*resAcl);
			*resAcl = NULL;
			free(buffer);
			AccessLock.Unlock();
			return false;
		}
		free(buffer);
	}


	// Initialize the response acl with the user's
	// by default
	for(i=0; i<acl->acls->length; i++)
	{
		(*resAcl)->data[i] |= acl->acls->data[i];
	}



	for(i=0; i<SKM_sk_num(USERS_GROUP, m_groups->list); i++)
	{
		currGroup = SKM_sk_value(USERS_GROUP, m_groups->list, i);
		if(!currGroup) continue;

		// Is the user in the group ?
		for(j=0; j<SKM_sk_num(ASN1_INTEGER, currGroup->users_serial); j++)
		{
			currSerial = SKM_sk_value(ASN1_INTEGER, currGroup->users_serial, j);
			if(!currSerial) continue;
			if((unsigned long)ASN1_INTEGER_GET(currSerial) != serial) continue;

			acl = FindAcl(ACL_ENTRY_TYPE_GROUP, ASN1_INTEGER_GET(currGroup->serial));
			if(!acl) continue;

			//Let's do the ORing
			min_len = (acl->acls->length < (*resAcl)->length)?acl->acls->length:(*resAcl)->length;
			for(k=0; k<min_len; k++)
			{
				(*resAcl)->data[k] |= acl->acls->data[k];
			}

			//Next group
			break;
		}
	}

	AccessLock.Unlock();
	return true;
}

bool X509_ACL_Validator::IsPkiAdministrator(const PKI_CERT & usercert) const
{
	bool ret;
	AccessLock.LockRead();
	ret = Static_IsPkiAdministrator(usercert.GetSerial(), m_acl->AdminSerials);
	AccessLock.Unlock();
	return ret;
}

bool X509_ACL_Validator::IsPkiAdministrator(long serial) const
{
	bool ret;
	AccessLock.LockRead();
	ret = Static_IsPkiAdministrator(serial, m_acl->AdminSerials);
	AccessLock.Unlock();
	return ret;
}

bool X509_ACL_Validator::Static_IsPkiAdministrator(const X509 *cert, const PKI_ADMIN_SERIALS * AdminSerials)
{
	return Static_IsPkiAdministrator(ASN1_INTEGER_GET(X509_get_serialNumber((X509*)cert)), AdminSerials);
}

bool X509_ACL_Validator::Static_IsPkiAdministrator(long serial, const PKI_ADMIN_SERIALS * AdminSerials)
{
	PKI_ADMIN_ENTRY * currEntry;
	int i;

	for(i=0; i < SKM_sk_num(PKI_ADMIN_ENTRY, AdminSerials->admins) ; i++)
	{
		currEntry = SKM_sk_value(PKI_ADMIN_ENTRY, AdminSerials->admins, i);
		if(!currEntry) continue;

		if(ASN1_INTEGER_GET(currEntry->serial) == serial) 
			return true;
	}
	return false;
}
